/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.Date;

import sbrt.kernel.utilities.Util;
import sbrt.shell.ApplicationException;
import sbrt.shell.ProcessManagerFactory;
import sbrt.shell.io.AbstractSbrtFile;
import sbrt.shell.io.ProcessInputFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.ProcessManager;

/**
 * A class to encapsulate the execution of processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 *
 */
public class ProcessRunner {

    /**
     * The name of the file in which errors will be logged.
     */
    public static final String ERROR_LOG = "sbrt_error_log.txt";

    /**
     * Executes the process specified in the indicated process file.
     * @param processFileName the name of the process file.
     * @param out the object to which informative messages will be written.
     */
    public void run(String processFileName, PrintWriter out) {
        run(new ProcessInputFile(processFileName), out);
    }

    /**
     * Executes the process specified in the indicated process file.
     * @param processFileName the name of the process file.
     * @param out the object to which informative messages will be written.
     */
    public void run(String processFileName, OutputStream out) {
        run(processFileName, new PrintWriter(out, true));
    }

    /**
     * Executes the process specified in the indicated process file.
     * @param processFileName the name of the process file.
     * @param out the object to which informative messages will be written.
     */
    public void run(String processFileName, PrintStream out) {
        run(processFileName, new PrintWriter(out, true));
    }

    /**
     * Executes the process specified in the provided process manager input.
     * @param processInput the input for the process to be executed.
     * @param out the object to which informative messages will be written.
     */
    public void run(ManagerInput<String, String> processInput, OutputStream out) {
        run(processInput, new PrintWriter(out, true));
    }

    /**
     * Executes the process specified in the provided process manager input.
     * @param processInput the input for the process to be executed.
     * @param out the object to which informative messages will be written.
     */
    public void run(ManagerInput<String, String> processInput, PrintStream out) {
        run(processInput, new PrintWriter(out, true));
    }

    /**
     * Executes the process specified in the provided process manager input.
     * @param processInput the input for the process to be executed.
     * @param out the object to which informative messages will be written.
     * @throws FatalException if a fatal error is detected during process execution.
     */
    public void run(ManagerInput<String, String> processInput, PrintWriter out) {

        try {

            executeProcess(processInput, out);

        } catch (Throwable t) {
        	throw interpret(t);
        }

    }

    /**
     * Interprets the provided throwable.
     * @param t an object thrown by the SBRT.
     * @return an exception whose message describes the error that occurred.
     */
    public static FatalException interpret(Throwable t) {
    
    	if (t instanceof FatalException) {
    		return (FatalException) t;
    	}
    	
    	if (t instanceof ApplicationException) {
        
    		// Getting the cause if one exists
    		Throwable cause = t.getCause();

    		if (cause == null || cause instanceof ApplicationException) {
    			return fatalError(t.getMessage());
    		}

    		return fatalError(t.getMessage(), cause);
    	}
    	
    	if (t instanceof NoClassDefFoundError) {

    		String s =
                "A required class is missing. Please ensure that "
                        + "the proper files are located in the SBRT's"
                        + " 'lib' directory.";

    		return fatalError(s, t);
    	}
    	
    	if (t instanceof UnsatisfiedLinkError) {

    		String s =
                "A required library is missing. Please ensure that "
                        + "the proper files are located in the SBRT's"
                        + " 'lib' directory.";

    		return fatalError(s, t);
    	} 
    	
    	if (t instanceof OutOfMemoryError) {

    		String s =
                "The Java Virtual Machine ran out of memory. The maximum "
                        + "memory allocated to the JVM can be set by using the "
                        + "'-Xmx' option. See the User's Guide "
                        + "for additional information.";
    		
    		return fatalError(s, t);
    	} 
    	
        String s =
                "An unknown error has occurred. Please report this "
                        + "problem to Jeremiah Wright at the email address "
                        + "'j.wright@bioc.uzh.ch'. Please attach the error "
                        + "log and all input files used when the error occurred.";

        return fatalError(s, t);
    }
    	
    
    private void executeProcess(ManagerInput<String, String> processInput,
                                PrintWriter out) {

        // Creating process manager
        ProcessManager<String, String> manager =
                ProcessManagerFactory.make(processInput, out);

        // Executing this process
        manager.execute(out);

        // Printing a closing message
        out.println();
        out.println("The program finished successfully.");
    }

    private static FatalException fatalError(String message) {
        String s = Util.widthFormat("ERROR:  " + message, 78);
        return new FatalException(s);
    }

    private static FatalException fatalError(String message, Throwable t) {

        if (t == null) {
            return fatalError(message);
        }

        // Creating field for exit message
        String exitMessage = null;

        try {

            // Opening output file to append this error
            String name = AbstractSbrtFile.getCanonicalName(ERROR_LOG);
            BufferedWriter wtr = new BufferedWriter(new FileWriter(name, true));

            // Writing date
            wtr.write(new Date().toString());
            wtr.newLine();

            // Writing exception
            wtr.write(t.toString());
            wtr.newLine();

            // Writing stack trace
            for (StackTraceElement e : t.getStackTrace()) {
                wtr.write(e.toString());
                wtr.newLine();
            }

            // Writing extra line
            wtr.newLine();

            // Closing
            wtr.flush();
            wtr.close();

            // Setting exit message
            exitMessage =
                    message
                            + "  The error log '"
                            + ERROR_LOG
                            + "' may contain additional information.";

        } catch (IOException e) {

            String newLine = System.getProperty("line.separator");

            exitMessage =
                    "Original Error: "
                            + message
                            + newLine
                            + t
                            + newLine
                            + "Most recent Error: "
                            + e
                            + newLine
                            + newLine
                            + "An unknown error occurred while trying to write "
                            + "an error to the error log (see above).";

        } 
        
        return fatalError(exitMessage);
    }

}
