/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt;

import java.util.LinkedHashMap;
import java.util.Map;

import sbrt.shell.gui.TextPaneGui;
import sbrt.shell.io.ProcessInputFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.MapManagerInput;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.MapFormat;

/**
 * This class is used to execute the Systems Biology Research Tool as an
 * application.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class Sbrt {

    /**
     * Returns the full version number of the Systems Biology Research Tool.
     * @return the full version number of the Systems Biology Research Tool.
     */
    public static String getVersion() {
        return MAJOR + "." + MINOR + "." + REVISION;
    }
    
    /**
     * Indicates the major version number of the Systems Biology Research Tool.
     */
    public static final int MAJOR = 2;
    
    /**
     * Indicates the minor version number of the Systems Biology Research Tool.
     */
    public static final int MINOR = 0;
    
    /**
     * Indicates the revision number of the Systems Biology Research Tool.
     */
    public static final int REVISION = 0;
    
    /**
     * Indicates the typical way in which this program is called from the
     * command line.
     */
    public static final String USAGE = "usage: sbrt process_file.txt";

    /**
     * The argument used to launch the GUI from the command line.
     */
    public static final String GUI_ARG = "-g";

    /**
     * The argument used to display the version number.
     */
    public static final String VERSION_ARG = "-v";

    /**
     * Parses the provided command line arguments and executes the application
     * they specify.
     * 
     * @param args the command line arguments.
     */
    public static void main(String[] args) {

    	
		ProcessRunner runner = new ProcessRunner();
		
    	try {
    		
    		interpretCommandLine(args, runner);
    		
    	} catch (Throwable t) {
    		
    		FatalException e = ProcessRunner.interpret(t);
    		
    		fatalError(e);
    	}
        
    }


	@SuppressWarnings("fallthrough")
	private static void interpretCommandLine(String[] args, ProcessRunner runner) {
    	
        switch (args.length) {

            case 0:
                System.err.println(USAGE);
                System.exit(1);

            case 1:

                if (args[0].equals(GUI_ARG)) {
                    launchGui();
                    break;
                }
                
                if (args[0].equals(VERSION_ARG)) {
                    System.out.println(getVersion());
                    break;
                }

            default: {
            

                ManagerInput<String, String> input = parseCommandLine(args);
                
                runner.run(input, System.out);
                
                // Killing any non-daemon threads that may be running
                System.exit(0);
            }
        }
    }
    
    private static void launchGui() {

        TextPaneGui gui = new TextPaneGui();
        gui.createAndShowWindow();

    }

    private static void fatalError(FatalException e) {
        System.err.println();
        System.err.println();
        System.err.println(e.getMessage());
        System.exit(1);
    }
    
    private static ManagerInput<String, String> parseCommandLine(String[] args) {

        // If only one argument exists, assume it's an input file
        if (args.length == 1) {
            return new ProcessInputFile(args[0]);
        }

        // Creating map entry parser
        MapFormat<String, String> mapFormat = EqualsMapFormatV1.getBasic();

        // Creating storage for inputs
        Map<String, String> input = new LinkedHashMap<String, String>();

        // Iterating thru each command line argument
        for (int i = 0; i < args.length; i++) {

            // Checking for -- at the beginning
            if (!args[i].startsWith("--")) {
                String s =
                        "The command line argument '"
                                + args[i]
                                + "' does not begin with '--'.";
                throw new FormatException(s);
            }

            try {

                // Getting next argument-value pair
                Map.Entry<String, String> entry = mapFormat.parsePair(args[i]);

                // Removing -- from the beginning and
                // replacing underscores with whitespace
                String keyword = entry.getKey().substring(2).replace('_', ' ');

                // Storing
                input.put(keyword, entry.getValue());

            } catch (FormatException e) {
                String s =
                        "The command line is improperly formatted. "
                                + e.getMessage();
                throw new FormatException(s);
            }

        }

        return new MapManagerInput<String, String>(input);
    }

}
