/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.ext.R;

import org.rosuda.JRI.REXP;
import org.rosuda.JRI.Rengine;

/**
 * This class is used to access the R engine safely. To use the engine,
 * it must first be locked. This grants sole access to the one holding the
 * key. The key must be supplied for each evaluation. When finished using the
 * engine, the key-holder should unlock it. The R engine cannot be stopped
 * after it has been started, except by {@linkplain Runtime#exit(int)}.
 *  
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class R {

	private static Rengine re;
    private static Object key;
    
    private R() {
        // Prevents instantiaion
    }
    
    /**
     * Indicates if the R engine is currently locked.
     * @return {@code true} if the R engine is locked, {@code false} othewise.
     */
    public static boolean isLocked() {
        return key != null;
    }
    
    /**
     * Indicates if the provided object is the key to unlock the R engine. 
     * @param potentialKey the potential key object.
     * @return {@code true} if the provided object can unlock the R engine,
     * {@code false} otherwise.
     */
    public static boolean isKey(Object potentialKey) {
        
        if (potentialKey == null) {
            return false;
        }
        
        return R.key == potentialKey;
    }
    
    /**
     * Locks the R engine so that no one can use it without the key.
     * @return the key that can unlock the R engine.
     */
    public static Object lock() {
        
        if (isLocked()) {
            String s = "The R engine is already locked.";
            throw new IllegalStateException(s);
        }
        
        return key = new Object();
    }
    
    /**
     * Unlocks the R engine so that others have access, that is, someone can
     * request a new key. If the provided key does not fit the current lock,
     * no action is taken.
     * 
     * @param key the key that fits the current lock.
     */
    public static void unlock(Object key) {

        if (!isLocked()) {
            return;
        }
        
        if (isKey(key)) {
            R.key = null;
        }
        
    }
    
    /**
     * Evaluates the provided R command.
     * 
     * @param statement the R command to be evaluated.
     * @param key the key required to unlock the R engine.
     * 
     * @return the result of executing the provided R command.
     * 
     * @throws IllegalArgumentException if the provided key does not fit the lock.
     * @throws JriException if the R engine throws an exception.
     */
    public static REXP eval(String statement, Object key) {
        
        if (!isKey(key)) {
            String s = "Not the right key: " + key;
            throw new IllegalArgumentException(s);
        }
        
        startR();
        
        try {
            
            return re.eval(statement);
            
        } catch (Throwable t) {
            throw new JriException(t);
        }
        
    }

    /**
     * Starts the R engine if it is not already running. It is not necessary to
     * call this method, because it is called automatically when required. If the
     * R engine is already running, no action is taken.
     * 
     * @throws JriException if an exception is thrown while attempting to start
     * the R engine.
     */
    public static void startR() {

        if (!isRunning()) {
            makeEngine();
        }

    }

    private static boolean isRunning() {
        
        if (re == null) {
            return false;
        }
        
        return re.waitForR();
    }
    
    private static void makeEngine() {
        
        String[] rArgs = { "--vanilla" };

        DummyConsole c = new DummyConsole();

        try {
            
            re = new Rengine(rArgs, false, c);
            
        } catch(Throwable t) {
            throw new JriException(t);
        }

        if (!re.waitForR()) {
            String s = "R could not be started.";
            throw new IllegalStateException(s);
        }
        
    }

}
