/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.ext.R.stat;

import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.rosuda.JRI.REXP;
import org.rosuda.JRI.RVector;

import sbrt.ext.R.R;
import sbrt.kernel.math.stat.comp.CorrResult;
import sbrt.kernel.math.stat.comp.Correlator;

/**
 * This class computes correlation statistics using {@code R}.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <V> the variable type.
 */
public abstract class RCorrelator<V> implements
        Correlator<V, RCorrelator.Result> {

    /**
     * Used for testing purposes.
     * 
     * @param args ignored.
     */
    public static void main(String[] args) {

        Double[] x = { 10d, 9d, 8d, 7d, 6d, 1d, 2d, 3d, 4d, 5d };
        Double[] y = { 1d, 2d, 3d, 4d, 5d, 6d, 7d, 8d, 9d, 10d };

        System.out.println("Starting R");
        R.startR();

        System.out.println("Pearson");
        System.out.println(new RPearson<Object>().corr(x, y));

        System.out.println("\nSpearman");
        System.out.println(new RSpearman<Object>().corr(x, y));

        System.out.println("\nKendall");
        System.out.println(new RKendall<Object>().corr(x, y));

    }

    /**
     * Returns the name of the correlation statistic computed by this {@code R}
     * correlator. This name should be understood by the {@code R} function
     * {@code cor.test()}; that is, {@code kendall}, {@code spearman}, or
     * {@code pearson} for example.
     * 
     * @return the name of the correlation statistic computed by this {@code R}
     * correlator.
     */
    protected abstract String getCorrName();

    
    
    // The current key for unlocking the R engine
    private Object key;

    
    

    /**
     * Computes and returns the correlation statistic of the provided x and y
     * values. The key sets of the provided maps need not be identical, but only
     * the intersecting values will be considered.
     * 
     * @param xValues variables mapped to their values.
     * @param yValues variables mapped to their values.
     * 
     * @return the results of computing the correlation statistic between the
     * provided x and y values.
     */
    public final Result corr(Map<? extends V, Double> xValues,
                             Map<? extends V, Double> yValues) {

        List<Double> x = new LinkedList<Double>();
        List<Double> y = new LinkedList<Double>();

        for (Map.Entry<? extends V, Double> e : xValues.entrySet()) {

            // Getting the next variable in the x vector
            V key = e.getKey();
            
            // Getting the value of this variable from the y vector,
            // if it exists
            Double yValue = yValues.get(key);
            
            if (yValue == null) {
                continue;
            }
            
            // Getting the value of this variable from the x vector
            Double xValue = e.getValue();

            // Storing them both
            x.add(xValue);
            y.add(yValue);
        }

        return corr(x, y);
    }

    /**
     * Computes and returns the correlation results for the provided values.
     * 
     * @param x an ordered array of values.
     * @param y an ordered array of values.
     * 
     * @return the correlation results for the provided values.
     */
    public final Result corr(Double[] x, Double[] y) {
        return corr(Arrays.asList(x), Arrays.asList(y));
    }

    /**
     * Computes and returns the correlation results for the provided values.
     * 
     * @param x an ordered collection of values.
     * @param y an ordered collection of values.
     * 
     * @return the correlation results for the provided values.
     */
    public final Result corr(Collection<Double> x, Collection<Double> y) {

        // Locking the R engine to acquire sole access
        key = R.lock();

        // Creating x and y vectors in R
        setXY(x, y);

        // Getting the name of the R correlation to be computed
        String name = getCorrName();

        // Computing the correlation coefficient
        REXP rResult = R.eval("cor.test(x, y, , \"" + name + "\")", key);

        // Interpretting the result
        Result result = convertResult(rResult);

        // Unlocking the R engine so others can use it
        R.unlock(key);

        return result;
    }

    private void setXY(Collection<Double> x, Collection<Double> y) {

        int xSize = x.size();
        int ySize = y.size();

        if (xSize != ySize) {
            String s = "x and y are not the same size";
            throw new IllegalArgumentException(s);
        }

        String xString = getRVector("x", x);
        String yString = getRVector("y", y);

        R.eval(xString, key);
        R.eval(yString, key);
    }

    private Result convertResult(REXP corrEvalResult) {

        if (corrEvalResult == null) {
            String s = "R did not perform the computation.";
            throw new IllegalStateException(s);
        }

        RVector v = corrEvalResult.asVector();

        double stat = v.at("statistic").asDouble();
        double estimate = v.at("estimate").asDouble();
        double p = v.at("p.value").asDouble();

        return new Result(stat, estimate, p);
    }

    private String getRVector(String varName, Collection<Double> vector) {

        int size = vector.size();

        StringBuilder vBuilder = new StringBuilder(size * 10);

        vBuilder.append(varName + " <- c(");

        Iterator<Double> i = vector.iterator();

        while (i.hasNext()) {

            Double x = i.next();

            vBuilder.append(x.toString());

            if (i.hasNext()) {
                vBuilder.append(",");
            }

        }

        vBuilder.append(")");

        return vBuilder.toString();
    }

    /**
     * This class is used to store the results of computing a correlation using
     * R.
     * 
     * @author This class was written and documented by Jeremiah Wright while in
     * the Wagner lab.
     * 
     */
    public static final class Result implements CorrResult {

        private final double stat;

        private final double estimate;

        private final double p;

        private Result(double stat, double estimate, double p) {
            this.stat = stat;
            this.estimate = estimate;
            this.p = p;
        }

        /**
         * Returns the corresponding statistic.
         * 
         * @return the corresponding statistic.
         */
        public double getStatistic() {
            return stat;
        }

        /**
         * Returns the computed correlation coefficient.
         * 
         * @return the computed correlation coefficient.
         */
        public double getCoeff() {
            return estimate;
        }

        /**
         * Returns the corresponding P value.
         * 
         * @return the corresponding P value.
         */
        public double getP() {
            return p;
        }

        /**
         * Returns a string representation of this result.
         * 
         * @return a string representation of this result.
         */
        public String toString() {
            return "stat = " + stat + ", coeff = " + estimate + ", p = " + p;
        }
    }
}
