/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.ext.cplex;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.math.optimization.LinearProgramSolver;
import sbrt.kernel.math.optimization.OptSense;
import sbrt.kernel.math.optimization.ProgramSolverException;
import sbrt.kernel.math.optimization.ProgramSolverStatus;
import sbrt.kernel.math.optimization.QuadraticProgramSolver;
import sbrt.kernel.utilities.*;
import ilog.concert.*;
import ilog.cplex.*;

/**
 * This class is a wrapper for the CPLEX package from ILOG.
 *  
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 * @param <R> the row variable type.
 * @param <C> the column variable type.
 */
public final class CplexSolver<R, C> 
	implements LinearProgramSolver<R, C>, QuadraticProgramSolver<R, C> {

    private static final String NAME = "CPLEX";

    /** Optimization model. */
    private IloCplex cplexModel;
    private IloLPMatrix matrix;
    private IloObjective objectiveFunction;

    private ProgramSolverStatus status;
    private double objectiveValue;
    private Map<C, Double> solutionVector;

    /** ColumnIndices (Object) -> their respective IloNumVar. */
    private Map<C, IloNumVar> columns;

    /** RowIndices (Object) -> their respective IloRange. */
    private Map<Object, IloRange> rows;

    /**
     * Constructs a new CPLEX program solver.
     */
    public CplexSolver() {

        columns = new HashMap<C, IloNumVar>();
        rows = new HashMap<Object, IloRange>();

        status = ProgramSolverStatus.UNKNOWN;
        objectiveValue = Double.NaN;
        solutionVector = Collections.emptyMap();

        try {

            cplexModel = new IloCplex();
            cplexModel.setOut(null);
            cplexModel.setParam(IloCplex.BooleanParam.PreInd, false);

            matrix = cplexModel.addLPMatrix();
            objectiveFunction =
                    cplexModel.addObjective(IloObjectiveSense.Maximize);

        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }

    }

    private int getColumnIndex(Object column) {
    	
        int index = -1;
        
        if (!columns.containsKey(column)) {
            return index;
        }
        
        try {
        	
            IloNumVar v = columns.get(column);
            index = matrix.getIndex(v);
            
        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }
        
        return index;
    }

    private int getRowIndex(Object row) {
    	
        int index = -1;
        
        if (!rows.containsKey(row)) {
            return index;
        }
        
        try {
        	
            IloRange r = rows.get(row);
            index = matrix.getIndex(r);
            
        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }
        
        return index;
    }

	/**
	 * Clears all information contained in this program solver. Calling 
	 * this method restores this program solver to its original, pristine 
	 * state.
	 */
    public void clear() {

        columns = new HashMap<C, IloNumVar>();
        rows = new HashMap<Object, IloRange>();

        status = ProgramSolverStatus.UNKNOWN;
        objectiveValue = Double.NaN;
        solutionVector = Collections.emptyMap();

        try {

            cplexModel.clearModel();
            
            matrix = cplexModel.addLPMatrix();
            objectiveFunction =
                    cplexModel.addObjective(IloObjectiveSense.Maximize);

        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }

    }

    /**
     * Sets the element of the program matrix at the provided row and 
     * column variables to the one provided.
     * 
     * @param <S> a type extending the primary row variable type.
     * @param <T> a type extending the primary column variable type.
     * 
     * @param row the row variable.
     * @param column the column variable.
     * @param value the double precision element value.
     */
    public <S extends R, T extends C> void 
    	setMatrixElement(S row, T column, double value) {

        int rowIndex = getRowIndex(row);
        int columnIndex = getColumnIndex(column);

        try {
            // adding row to matrix if not already present
            if (rowIndex < 0) {
                IloRange r = cplexModel.range(Double.NEGATIVE_INFINITY,
                                              Double.POSITIVE_INFINITY,
                                              row.toString());
                rowIndex = matrix.addRow(r);
                rows.put(row, r);
            }

            // adding column to matrix if not already present
            if (columnIndex < 0) {
                IloNumVar v = cplexModel.numVar(Double.NEGATIVE_INFINITY,
                                                Double.POSITIVE_INFINITY,
                                                column.toString());
                columnIndex = matrix.addColumn(v);
                columns.put(column, v);
            }

            // Setting value, even if it's zero
            matrix.setNZ(rowIndex, columnIndex, value);

        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }
    }

    /**
     * Sets a constraint on the provided mathematical expression of
     * column variables.
     * 
     * @param expression a mathematical expression of column variables for
     * which the constraint is set.
     * @param c the constraint.
     */
    public void setConstraint(MathExpr<? extends C> expression, Interval c) {
    	
        if (expression instanceof LinearComb) {
            setConstraint((LinearComb<? extends C>) expression, c);
        }

        else if (expression instanceof QuadraticExpr) {
            setConstraint((QuadraticExpr<? extends C>) expression, c);
        }

        else {
            String s = "CPLEX can only handle linear and quadratic " +
                       "constraints: " + expression;
            throw new IllegalArgumentException(s);
        }
    }

    /**
     * Sets constraints on the provided mathematical expressions of
     * column variables.
     * 
     * @param exprConstraints mathematical expressions of column variables
     * mapped to their corresponding constraint.
     */
    public void setConstraints(
    		Map<? extends MathExpr<? extends C>, Interval> exprConstraints) {

    	for (Map.Entry<? extends MathExpr<? extends C>, Interval> e :
    		exprConstraints.entrySet()) {

            setConstraint(e.getKey(), e.getValue());
        }
        
    }

    /**
     * Sets a constraint on the provided mathematical expression of
     * column variables.
     * 
     * @param linComb a mathematical expression of column variables for
     * which the constraint is set.
     * @param c the constraint.
     */
    public void setConstraint(LinearComb<? extends C> linComb, Interval c) {

    	try {
            IloRange range = rows.get(linComb);

            if (range == null) {
                range = cplexModel.addRange(c.getLower(),
                                            getLinearExpr(linComb),
                                            c.getUpper());
                matrix.addRow(range);
                rows.put(linComb, range);
            }

            else {
                // WARNING: DO NOT CHANGE THIS LINE
                // For some strange reason using
                // 'range.setLB(c.getLower());  range.setUB(c.getUpper());'
                // does NOT work
                range.setBounds(c.getLower(), c.getUpper());
            }

        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }

    }

    /**
     * Sets a constraint on the provided mathematical expression of
     * column variables.
     * 
     * @param expr a mathematical expression of column variables for
     * which the constraint is set.
     * @param c the constraint.
     */
    public void setConstraint(QuadraticExpr<? extends C> expr, Interval c) {
        throw new UnsupportedOperationException("not yet implemented");
    }
    
    /**
     * Sets the constraint on the provided column variable.
     * 
     * @param <T> a type extending the primary column variable type.
     * @param column the column variable for which the constraint is set.
     * @param c the constraint.
     */
    public <T extends C> void setColumnConstraint(T column, Interval c) {
    	
        try {
            IloNumVar var = columns.get(column);

            if (var == null) {
                String s = "Bad column constraint, no such column: " + column;
                throw new IllegalArgumentException(s);
            }

            var.setLB(c.getLower());
            var.setUB(c.getUpper());

        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }
    }

    /**
     * Sets the constraints for the provided column variables.
     * 
     * @param constraints column variables mapped to their corresponding
     * constraint.
     */
    public void setColumnConstraints(Map<? extends C, Interval> constraints) {
    	
    	for (Map.Entry<? extends C, Interval> e : constraints.entrySet()) {
    		setColumnConstraint(e.getKey(), e.getValue());
        }
        
    }

    /**
     * Sets the constraint for the provided row variable.
     * 
     * @param <T> a type extending the primary row variable type.
     * @param row the row variable for which the constraint is set.
     * @param c the constraint.
     */
    public <T extends R> void setRowConstraint(T row, Interval c) {
    	

        try {
        	
            IloRange range = rows.get(row);

            if (range == null) {
                String s = "Bad row constraint, no such row: " + row;
                throw new IllegalArgumentException(s);
            }

            range.setBounds(c.getLower(), c.getUpper());

        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }
    }

    /**
     * Sets the constraints for the provided row variables.
     * 
     * @param rowConstraints row variables mapped to their corresponding
     * constraint.
     */
    public void setRowConstraints(Map<? extends R, Interval> rowConstraints) {

    	for (Map.Entry<? extends R, Interval> e : rowConstraints.entrySet()) {
            setRowConstraint(e.getKey(), e.getValue());
        }
    }

    /**
     * Sets the objective function and optimization sense to the provided
     * values.
     * 
     * @param objective a mathematical expression of column variables from
     * this program solver.
     * @param sense the optimization sense.
     * 
     * @see #setObjectiveFunction(MathExpr)
     * @see #setOptimizationSense(OptSense)
     */
    public void setObjectiveFunction(MathExpr<? extends C> objective,
                                     OptSense sense) {
        setObjectiveFunction(objective);
        setOptimizationSense(sense);
    }

    /**
     * Sets the objective function and optimization sense to the provided
     * values.
     * 
     * @param objective a mathematical expression of column variables from
     * this program solver.
     * @param sense the optimization sense.
     * 
     * @see #setObjectiveFunction(MathExpr)
     * @see #setOptimizationSense(OptSense)
     */
    public void setObjectiveFunction(LinearComb<? extends C> objective,
                                     OptSense sense) {
        setObjectiveFunction(objective);
        setOptimizationSense(sense);
    }

    /**
     * Sets the objective function and optimization sense to the provided
     * values.
     * 
     * @param objective a mathematical expression of column variables from
     * this program solver.
     * @param sense the optimization sense.
     * 
     * @see #setObjectiveFunction(MathExpr)
     * @see #setOptimizationSense(OptSense)
     */
    public void setObjectiveFunction(QuadraticExpr<? extends C> objective,
                                     OptSense sense) {
        setObjectiveFunction(objective);
        setOptimizationSense(sense);
    }

    /**
     * Sets the objective function to the provided mathematical expression of
     * column variables.
     * 
     * @param objective a mathematical expression of column variables from
     * this program solver.
     */
    public void setObjectiveFunction(MathExpr<? extends C> objective) {
    	
        if (objective instanceof LinearComb) {
            setObjectiveFunction((LinearComb<? extends C>) objective);
        }

        else if (objective instanceof QuadraticExpr) {
            setObjectiveFunction((QuadraticExpr<? extends C>) objective);
        }

        else {
            String s = "CPLEX can only handle linear and quadratic " +
                       "objective functions: " + objective;
            throw new IllegalArgumentException(s);
        }
    }

    private IloLinearNumExpr 
    getLinearExpr(LinearComb<? extends C> expression) throws IloException {

        IloLinearNumExpr iloExpression = cplexModel.linearNumExpr();

        for (LinearComb.Term<? extends C> t : expression) {

            C column = t.getVariable();
            IloNumVar var = columns.get(column);

            if (var == null) {
                String s = "Bad linear combination, no such column: " +
                           column;
                throw new IllegalArgumentException(s);
            }

            iloExpression.addTerm(t.getCoefficient(), var);
        }

        iloExpression.setConstant(expression.getConstant());

        return iloExpression;
    }

    /**
     * Sets the objective function to the provided mathematical expression of
     * column variables.
     * 
     * @param objective a mathematical expression of column variables from
     * this program solver.
     */
    public void setObjectiveFunction(LinearComb<? extends C> objective) {

        try {
        	
            IloLinearNumExpr objExpression = getLinearExpr(objective);
            this.objectiveFunction.setExpr(objExpression);
            
        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }

    }
    
    /**
     * Not yet implemented.
     * 
     * @param objective a mathematical expression of column variables from
     * this program solver.
     * 
     * @throws UnsupportedOperationException
     */
	public void setObjectiveFunction(QuadraticExpr<? extends C> objective) {
		throw new UnsupportedOperationException();
	}
    
/*
    public void setObjectiveFunction(QuadraticExpr<? extends C> objectiveFunction) {
    	
        try {
        	
            List terms = new LinkedList();

            for (QuadraticExpr.Term<? extends C> term : objectiveFunction) {

                double coeff = term.getCoefficient();
                
                for (C var : term)
                IloNumVar var1 = (IloNumVar) columns.get(t.getVariable1());
                IloNumVar var2 = (IloNumVar) columns.get(t.getVariable2());

                if (var2 == null) {
                    IloNumExpr e = cplexModel.prod(coeff, var1);
                    terms.add(e);
                } else {
                    IloNumExpr e = cplexModel.prod(coeff, var1, var2);
                    terms.add(e);
                }
            }

        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }
        throw new UnsupportedOperationException("not yet implemented");
    }
*/
    /*
        private IloNumExpr getSum(List terms) {

            if (terms.size() == 1) {
                return (IloNumExpr) terms.get(0);
            }

            Iterator i = terms.iterator();
            IloNumExpr first = (IloNumExpr) i.next();

            IloNumExpr e = null;

        }
     */

    /**
     * Sets the optimization sense.
     * @param sense the optimization sense.
     */
    public void setOptimizationSense(OptSense sense) {
        try {
            if (sense == OptSense.MAX) {
                objectiveFunction.setSense(IloObjectiveSense.Maximize);
            }

            else if (sense == OptSense.MIN) {
                objectiveFunction.setSense(IloObjectiveSense.Minimize);
            }
        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }
    }

    private ProgramSolverStatus determineStatus() throws IloException {

        ProgramSolverStatus status = null;
        IloCplex.Status cplexStatus = cplexModel.getStatus();

        if (cplexStatus.equals(IloCplex.Status.Optimal)) {
            status = ProgramSolverStatus.OPTIMAL;
        }

        else if (cplexStatus.equals(IloCplex.Status.Unbounded)) {
            status = ProgramSolverStatus.UNBOUNDED;
        }

        else if (cplexStatus.equals(IloCplex.Status.Infeasible)) {
            status = ProgramSolverStatus.INFEASIBLE;
        }

        else if (cplexStatus.equals(IloCplex.Status.Feasible)) {
            status = ProgramSolverStatus.FEASIBLE;
        }

        else if (cplexStatus.equals(IloCplex.Status.InfeasibleOrUnbounded)) {
            status = ProgramSolverStatus.INFEASIBLE_OR_UNBOUNDED;
        }

        else if (cplexStatus.equals(IloCplex.Status.Unknown)) {
            status = ProgramSolverStatus.UNKNOWN;
        }

        else if (cplexStatus.equals(IloCplex.Status.Bounded)) {
            status = ProgramSolverStatus.BOUNDED;
        }

        else if (cplexStatus.equals(IloCplex.Status.Error)) {
            status = ProgramSolverStatus.ERROR;
        }

        else {
            String s = "Unknown CPLEX status: " + cplexStatus;
            throw new IllegalStateException(s);
        }

        return status;
    }

    private double determineObjectiveValue() throws IloException {
        return cplexModel.getObjValue();
    }

    private Map<C, Double> determineSolutionVector() throws IloException {
    	
        Map<C, Double> solutionVector = 
        	new LinkedHashMap<C, Double>(columns.size() + 1, 1.0F);

        double[] values = cplexModel.getValues(matrix);
        
        for (C column : columns.keySet()) {
            int index = getColumnIndex(column);
            solutionVector.put(column, values[index]);
        }

        return solutionVector;
    }

    /**
     * Computes the optimal value of the current objective function in the 
     * direction of the current optimization sense.
     * 
     * @return {@code true} if an optimal solution was found; 
     * {@code false} otherwise.
     * 
     * @see #getObjectiveValue()
     * @see #getSolutionVector()
     */
    public boolean optimize() {
        boolean optimal = false;

        try {
            cplexModel.solve();
            status = determineStatus();

            if (status == ProgramSolverStatus.OPTIMAL) {
                optimal = true;
                objectiveValue = determineObjectiveValue();
                solutionVector = determineSolutionVector();
            } else {
                objectiveValue = Double.NaN;
                solutionVector = Collections.emptyMap();
            }

        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }

        return optimal;
    }

    /**
     * Returns the status of the most recently attempted optimization.
     * @return the status of the most recently attempted optimization.
     * @see #optimize()
     */
    public ProgramSolverStatus getStatus() {
        return status;
    }
    
    /**
     * Returns the most recently computed objective value.
     * @return the most recently computed objective value.
     * @see #optimize()
     */
    public double getObjectiveValue() {
        return objectiveValue;
    }
    
    /**
     * Returns the most recently computed solution vector.
     * @return column variables mapped to their corresponding numerical values.
     * @see #optimize()
     */
    public UnmodifiableMap<C, Double> getSolutionVector() {
        return new UnmodifiableMap<C, Double>(solutionVector);
    }

    /**
     * Writes the CPLEX model to a file with the provided name.
     * @param fileName the name of the file to be written.
     */
    public void writeProblem(String fileName) {
        try {
            cplexModel.exportModel(fileName);
        } catch (IloException e) {
            throw new ProgramSolverException(e);
        }
    }

    /**
     * Returns the name of this program solver.
     * @return the name of this program solver.
     */
    public String toString() {
        return NAME;
    }
    
    /**
     * Reads the program solver parameter file with the provided name.
     * This file must be in PRM format. See the CPLEX documentation for further
     * details.
     *  
     * @param prmFileName the name of the PRM file containing parameter values.
     */
    public void setParamFile(String prmFileName) {
    	
    	try {
    		cplexModel.readParam(prmFileName);
    	} catch (IloException e) {
    		throw new ProgramSolverException(e);
    	}
    }
    
    /**
     * Indicates if the provided object is a column variable in the matrix of
     * this program solver.
     * 
     * @param obj the potential column variable.
     * 
     * @return {@code true} if the provided object is a column variable in the
     * matrix of this program solver; {@code false} otherwise.
     */
	public boolean isColumnVar(Object obj) {
		return columns.containsKey(obj);
	}
    
    /**
     * Indicates if the provided object is a row variable in the matrix of
     * this program solver.
     * 
     * @param obj the potential row variable.
     * 
     * @return {@code true} if the provided object is a row variable in the
     * matrix of this program solver; {@code false} otherwise.
     */
	public boolean isRowVar(Object obj) {
		return rows.containsKey(obj);
	}

}
