/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.ext.glpk;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.math.optimization.LinearProgramSolver;
import sbrt.kernel.math.optimization.OptSense;
import sbrt.kernel.math.optimization.ProgramSolverStatus;
import sbrt.kernel.utilities.*;

/**
 * This class is a wrapper for the GLPK package from GNU.
 *  
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 * @param <R> the row variable type.
 * @param <C> the column variable type.
 */
public class GlpkSolver<R, C> implements LinearProgramSolver<R, C> {

	// The name of the linear program solver 
    private static final String name = "GLPK";

    // The linear program solver
    private org.gnu.glpk.GlpkSolver solver;

    // Row variable -> index i
    private Map<Object, Integer> rowIndices;

    // Column variable -> index j
    private Map<C, Integer> columnIndices;

    // The solution status
    private ProgramSolverStatus status;

    // The objective value
    private double objectiveValue;

    // The solution vector
    private Map<C, Double> solutionVector;

    /**
     * Constructs a new GLPK program solver.
     */
    public GlpkSolver() {
        clear();
    }

	/**
	 * Clears all information contained in this program solver. Calling 
	 * this method restores this program solver to its original, pristine 
	 * state.
	 */
    public void clear() {

    	// Creating a new GLPK solver
        solver = new org.gnu.glpk.GlpkSolver();

        // Turning off message printing
        solver.enablePrints(false);

        // Creating storage
        columnIndices = new LinkedHashMap<C, Integer>();
        rowIndices = new LinkedHashMap<Object, Integer>();

        // Initializing fields
        status = ProgramSolverStatus.UNKNOWN;
        objectiveValue = Double.NaN;
        solutionVector = Collections.emptyMap();
    }

    private void checkCoefficient(double x) {
    	
        if (Double.isInfinite(x) || Double.isNaN(x)) {
            String s = "Invalid coefficient: " + x;
            throw new IllegalArgumentException(s);
        }
        
    }

    private void checkColumnVar(Object columnVar) {
    	
        if (!isColumnVar(columnVar)) {
            String s = "column variable does not exist: " + columnVar;
            throw new IllegalArgumentException(s);
        }
    }

    private void checkExpression(MathExpr<? extends C> expression) {
    	
    	// Ensuring each variable exists
    	for (C columnIndex : expression.getVariables()) {
    		checkColumnVar(columnIndex);
        }
    }

    private int getRowIndex(Object rowIndex) {
    	
    	// Getting the integer index for the provided row variable
        Integer i = rowIndices.get(rowIndex);
        
        // If an index does not yet exist for this variable
        if (i == null) {
        	
        	// Creating the new index
            i = rowIndices.size() + 1;

            // Storing it
            rowIndices.put(rowIndex, i);

            // Creating space for it in the GLPK solver
            solver.addRows(1);
        }
        
        return i;
    }

    private <Y extends C> int getColumnIndex(Y columnIndex) {
    	
    	// Getting the integer index for the provided column variable
        Integer j = columnIndices.get(columnIndex);

        // If an index does not yet exist for this variable
        if (j == null) {
        	
        	// Creating the new index
            j = columnIndices.size() + 1;

            // Storing it
            columnIndices.put(columnIndex, j);

            // Creating space for it in the GLPK solver
            solver.addCols(1);
        }
        
        return j;
    }
    
    /**
     * Indicates if the provided object is a row variable in the matrix of
     * this program solver.
     * 
     * @param obj the potential row variable.
     * 
     * @return {@code true} if the provided object is a row variable in the
     * matrix of this program solver; {@code false} otherwise.
     */
    public boolean isRowVar(Object obj) {
    	return rowIndices.get(obj) != null;
    }
    
    /**
     * Indicates if the provided object is a column variable in the matrix of
     * this program solver.
     * 
     * @param obj the potential column variable.
     * 
     * @return {@code true} if the provided object is a column variable in the
     * matrix of this program solver; {@code false} otherwise.
     */
    public boolean isColumnVar(Object obj) {
    	return columnIndices.get(obj) != null;
    }

    /**
     * Sets the element of the program matrix at the provided row and 
     * column variables to the one provided.
     * 
     * @param <X> a type extending the primary row variable type.
     * @param <Y> a type extending the primary column variable type.
     * 
     * @param row the row variable.
     * @param column the column variable.
     * @param value the double precision element value.
     */
    public <X extends R, Y extends C> void 
    	setMatrixElement(X row, Y column, double value) {

    	setElement(row, column, value);
    }

    private <Y extends C> void setElement(Object rowIndex,
    									  Y columnIndex,
    									  double value) {
    	
    	// Checking inputs
    	Util.nullCheck(rowIndex);
        Util.nullCheck(columnIndex);
        checkCoefficient(value);

        // Getting indices (and creating them if they don't yet exist)
        int i = getRowIndex(rowIndex);
        int j = getColumnIndex(columnIndex);

        // Determining an appropriate size
        int size = rowIndices.size() + 1;
        
        // Creating storage for row indices and values
        int[] rowIndices = new int[size];
        double[] values = new double[size];
        
        // Filling row indices and values from column j
        solver.getMatCol(j, rowIndices, values);

        // Modifying the arrays appropriately
        int len = set(i, rowIndices, value, values);

        // Setting the new values for column j
        solver.setMatCol(j, len, rowIndices, values);

//        print(rowIndices, values); System.err.println();
    }
    
    /*
	private static void print(double[] array) {
		
		for (int i = 0; i < array.length; i ++) {
			System.err.print(array[i]);
			if (i < array.length - 1) {
				System.err.print(", ");
			}
		}
		
		System.err.println();
	}
	
	private static void print(int[] array, double[] values) {
		
		for (int i = 0; i < array.length; i ++) {
			
			if (array[i] == 0) {
				continue;
			}
			
			System.err.print("i: " + i);
			System.err.print(", row index: " + array[i]);
			System.err.println(", value: " + values[i]);
		}
		
		System.err.println();
	}
	*/
    
    private static int 
    set(int rowIndex, int[] rowIndices, double value, double[] values) {
    	
    	// Removing previous value if new value is zero
    	if (value == 0d) {
    		return remove(rowIndex, rowIndices, values);
    	}
    	
    	boolean replaced = false;
    	int length = 0;

    	// Replacing the index-value pair if it already exists
    	for (int i = 1; i < rowIndices.length; i ++) {
    		
    		if (rowIndices[i] != 0) {
    			++length;
    		}
    		
    		// Making the replacement
    		if (rowIndices[i] == rowIndex) {
    			values[i] = value;
    			replaced = true;
    		}
    		
    	}
    	
    	// If the replacement was made, nothing more to do
    	if (replaced) {
    		return length;
    	}
    	
    	// Denoting the addition of another element 
    	++length;
    	
    	// Adding to the end if the index-value pair does not already exist
    	rowIndices[length] = rowIndex;
    	values[length] = value;
    	
    	return length;
    }
    
    private static int 
    remove(int rowIndex, int[] rowIndices, double[] values) {
    	
    	int length = 0;
    	
    	for (int i = 1, j = 1; j < rowIndices.length; i ++, j++) {
    		
    		// If a true index exists here
    		if (rowIndices[i] != 0) {
    			++length;
    		}
    		
    		// If it's the index to be removed
    		if (rowIndices[i] == rowIndex) {
    			++j;
    			--length;
    		}
    		
    		// Writing over previous values
    		rowIndices[i] = rowIndices[j];
    		values[i] = values[j];
    	}
    	
    	return length;
    }
    
    /**
     * Sets the constraint on the provided column variable.
     * 
     * @param <Y> a type extending the primary column variable type.
     * @param column the column variable for which the constraint is set.
     * @param c the constraint.
     */
    public <Y extends C> void setColumnConstraint(Y column, Interval c) {

        // Getting integer index for the provided column variable
        Integer index = columnIndices.get(column);

        // If the index does not exist
        if (index == null) {
            String s = "column index does not exist: " + column;
            throw new IllegalArgumentException(s);
        }

        // Setting column bounds
        solver.setColBnds(index.intValue(), 			// index
        				  determineConstraintType(c),  	// variable type
                          c.getLower(),  				// lower bound 
                          c.getUpper());  				// upper bound
    }

    /**
     * Sets the constraints for the provided column variables.
     * 
     * @param columnConstraints column variables mapped to their corresponding
     * constraint.
     */
    public void 
    setColumnConstraints(Map<? extends C, Interval> columnConstraints) {

    	// Iterating thru each variable-constraint pair
    	for (Map.Entry<? extends C, Interval> e : columnConstraints.entrySet()) {
    		setColumnConstraint(e.getKey(), e.getValue());
        }
    }

    /**
     * Sets the constraint for the provided row variable.
     * 
     * @param <X> a type extending the primary row variable type.
     * @param row the row variable for which the constraint is set.
     * @param c the constraint.
     */
    public <X extends R> void setRowConstraint(X row, Interval c) {
    	setRowConst(row, c);
    }
    
    private void setRowConst(Object rowIndex, Interval c) {
    	
    	// Getting integer index for the provided row variable
        Integer index = rowIndices.get(rowIndex);
        
        // If the index does not exist
        if (index == null) {
            String s = "row index does not exist: " + rowIndex;
            throw new IllegalArgumentException(s);
        }

        // Setting row bounds
        solver.setRowBnds(index.intValue(),				// index
        				  determineConstraintType(c),	// variable type
                          c.getLower(), 				// lower bound
                          c.getUpper());				// upper bound
    }

    /**
     * Sets the constraints for the provided row variables.
     * 
     * @param rowConstraints row variables mapped to their corresponding
     * constraint.
     */
    public void setRowConstraints(Map<? extends R, Interval> rowConstraints) {

    	// Iterating thru each row variable-constraint pair
    	for (Map.Entry<? extends R, Interval> e : rowConstraints.entrySet()) {
            setRowConstraint(e.getKey(), e.getValue());
        }
    }

    /**
     * Sets a constraint on the provided mathematical expression of
     * column variables.
     * 
     * @param expression a mathematical expression of column variables for
     * which the constraint is set.
     * @param c the constraint.
     */
    public void setConstraint(MathExpr<? extends C> expression, Interval c) {

    	// Checking input
        if (!(expression instanceof LinearComb)) {
            String s = name + " doesn't understand expressions of type " +
                       expression.getClass().getCanonicalName() + ": " + 
                       expression;
            throw new IllegalArgumentException(s);
        }
        
        // Setting constraint
        setConstraint((LinearComb<? extends C>) expression, c);
    }

    /**
     * Sets a constraint on the provided mathematical expression of
     * column variables.
     * 
     * @param expression a mathematical expression of column variables for
     * which the constraint is set.
     * @param c the constraint.
     */
    public void setConstraint(LinearComb<? extends C> expression, Interval c) {

    	// Iterating thru each term
    	for (LinearComb.Term<? extends C> term : expression) {

    		// Getting the variable
            C column = term.getVariable();

            // Ensuring this variable exists 
            checkColumnVar(column);

            // Setting the term's coefficient as an element, using the 
            // expression as the row variable 
            setElement(expression, column, term.getCoefficient());
        }
    	
    	// Setting the constraints on this expression
        setRowConst(expression, c);
    }

    /**
     * Sets constraints on the provided mathematical expressions of
     * column variables.
     * 
     * @param expressionConstraints mathematical expressions of column variables
     * mapped to their corresponding constraint.
     */
    public void setConstraints(Map<? extends MathExpr<? extends C>, Interval> 
    							expressionConstraints) {

    	// Iterating thru each expression-constraint pair
    	for (Map.Entry<? extends MathExpr<? extends C>, Interval> e : 
    		 			expressionConstraints.entrySet()) {

    		// Setting constraint
            setConstraint(e.getKey(), e.getValue());
        }
    }

    /**
     * Sets the objective function and optimization sense to the provided
     * values.
     * 
     * @param expression a mathematical expression of column variables from
     * this program solver.
     * @param sense the optimization sense.
     * 
     * @see #setObjectiveFunction(MathExpr)
     * @see #setOptimizationSense(OptSense)
     */
    public void setObjectiveFunction(MathExpr<? extends C> expression,
                                     OptSense sense) {
        setObjectiveFunction(expression);
        setOptimizationSense(sense);
    }

    /**
     * Sets the objective function to the provided mathematical expression of
     * column variables.
     * 
     * @param expression a mathematical expression of column variables from
     * this program solver.
     */
    public void setObjectiveFunction(MathExpr<? extends C> expression) {

    	// Checking input
        if (!(expression instanceof LinearComb)) {
            String s = name + " does not understand expressions of type " +
                       expression.getClass() + ": " + expression;
            throw new IllegalArgumentException(s);
        }

        setObjectiveFunction((LinearComb<? extends C>) expression);
    }

    /**
     * Sets the objective function and optimization sense to the provided
     * values.
     * 
     * @param expression a mathematical expression of column variables from
     * this program solver.
     * @param sense the optimization sense.
     * 
     * @see #setObjectiveFunction(MathExpr)
     * @see #setOptimizationSense(OptSense)
     */
    public void setObjectiveFunction(LinearComb<? extends C> expression,
                                     OptSense sense) {
        setObjectiveFunction(expression);
        setOptimizationSense(sense);
    }

    /**
     * Sets the objective function to the provided mathematical expression of
     * column variables.
     * 
     * @param expression a mathematical expression of column variables from
     * this program solver.
     */
    public void setObjectiveFunction(LinearComb<? extends C> expression) {
    	
    	// Checking input
        checkExpression(expression);

        // Clearing previous objective function
        for (int j = 1, size = columnIndices.size(); j <= size; j++) {
            solver.setObjCoef(j, 0);
        }

        // Iterating thru each term
        for (LinearComb.Term<? extends C> term : expression) {

        	// Getting the column index
            int j = columnIndices.get(term.getVariable());

            // Setting coefficient in objective function
            solver.setObjCoef(j, term.getCoefficient());
        }

        // Setting constant in objective function
        solver.setObjCoef(0, expression.getConstant());
    }

    /**
     * Sets the optimization sense.
     * @param sense the optimization sense.
     */
    public void setOptimizationSense(OptSense sense) {

    	// Checking input
    	Util.nullCheck(sense);

        if (sense == OptSense.MAX) {
            solver.setObjDir(org.gnu.glpk.GlpkSolver.LPX_MAX);
        } 
        
        else if (sense == OptSense.MIN) {
            solver.setObjDir(org.gnu.glpk.GlpkSolver.LPX_MIN);
        }
    }

    /**
     * Computes the optimal value of the current objective function in the 
     * current optimization sense.
     * 
     * @return {@code true} if an optimal solution was found; 
     * {@code false} otherwise.
     * 
     * @see #getObjectiveValue()
     * @see #getSolutionVector()
     */
    public boolean optimize() {

    	// Running the simplex method
        solver.simplex();

        // Getting solution status
        status = determineStatus();

        // If an optimal solution was found
        if (status == ProgramSolverStatus.OPTIMAL) {

        	// Saving objective value and solution vector
        	objectiveValue = solver.getObjVal();
            solutionVector = determineSolutionVector();

            // Returning true to indicate optimal solution was found
            return true;
        }

        // Setting objective value and solution vector
        objectiveValue = Double.NaN;
        solutionVector = Collections.emptyMap();

        // Returning false to indicate an optimal solution wasn't found
        return false;
    }

    /**
     * Returns the most recently computed objective value.
     * @return the most recently computed objective value.
     * @see #optimize()
     */
    public double getObjectiveValue() {
        return objectiveValue;
    }
    
    /**
     * Returns the most recently computed solution vector.
     * @return column variables mapped to their corresponding numerical values.
     * @see #optimize()
     */
    public Map<C, Double> getSolutionVector() {
    	
    	// Determining appropriate size
    	int size = columnIndices.size() + 1;

    	// Creating storage
    	Map<C, Double> m = new LinkedHashMap<C, Double>(size, 1.0f);
    	
    	// Creating a defensive copy 
    	m.putAll(solutionVector);
    	
        return m;
    }

    /**
     * Returns the status of the most recently attempted optimization.
     * @return the status of the most recently attempted optimization.
     * @see #optimize()
     */
    public ProgramSolverStatus getStatus() {
        return status;
    }

    private Map<C, Double> determineSolutionVector() {

    	// Determining appropriate size
    	int size = columnIndices.size() + 1;
    	
    	// Creating storage
        Map<C, Double> vector = new LinkedHashMap<C, Double>(size, 1.0F);

    	// Iterating thru each colum variable-integer index pair
        for (Map.Entry<C, Integer> e : columnIndices.entrySet()) {

        	// Getting pair
            C variable = e.getKey();
            Integer index = e.getValue();

            // Getting solution value
            double value = solver.getColPrim(index);

            // Storing
            vector.put(variable, value);
        }

        return vector;
    }

    private ProgramSolverStatus determineStatus() {

        int status = solver.getStatus();

        if (status == org.gnu.glpk.GlpkSolver.LPX_OPT) {
            return ProgramSolverStatus.OPTIMAL;
        }

        if (status == org.gnu.glpk.GlpkSolver.LPX_FEAS) {
            return ProgramSolverStatus.FEASIBLE;
        }

        if (status == org.gnu.glpk.GlpkSolver.LPX_NOFEAS || 
        		status == org.gnu.glpk.GlpkSolver.LPX_INFEAS) {
            return ProgramSolverStatus.INFEASIBLE;
        }

        if (status == org.gnu.glpk.GlpkSolver.LPX_UNBND) {
            return ProgramSolverStatus.UNBOUNDED;
        }

        if (status == org.gnu.glpk.GlpkSolver.LPX_UNDEF) {
            return ProgramSolverStatus.UNDEFINED;
        }

        return ProgramSolverStatus.UNKNOWN;
    }

    private int determineConstraintType(Interval c) {

        double lower = c.getLower();
        double upper = c.getUpper();

        if (Double.isInfinite(lower) && Double.isInfinite(upper)) {
            return org.gnu.glpk.GlpkSolver.LPX_FR;
        } 
        
        else if (Double.isInfinite(lower)) {
            return org.gnu.glpk.GlpkSolver.LPX_UP;
        } 
        
        else if (Double.isInfinite(upper)) {
            return org.gnu.glpk.GlpkSolver.LPX_LO;
        } 
        
        else if (Double.compare(lower, upper) == 0) {
            return org.gnu.glpk.GlpkSolver.LPX_FX;
        }

        return org.gnu.glpk.GlpkSolver.LPX_DB;
    }

    /**
     * Throws {@code UnsupportedOperationException}.
     * 
     * @param fileName the name of the file containing parameter values.
     * 
     * @throws UnsupportedOperationException
     */
    public void setParamFile(String fileName) {
    	throw new UnsupportedOperationException("Not yet implemented.");
    }

}
