/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.ext.mma.algebra;

import java.util.*;

import com.wolfram.jlink.*;

import sbrt.ext.ExternalSoftwareException;
import sbrt.kernel.math.*;

/**
 * This class is intended for use solely by the class {@linkplain MathematicaSolver}.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 *
 */
public class MMaExprInterpretter {

    private static Expr PLUS = new Expr(Expr.SYMBOL, "Plus");

    private static Expr TIMES = new Expr(Expr.SYMBOL, "Times");

    private static Expr LIST = new Expr(Expr.SYMBOL, "List");

    private static Expr RULE = new Expr(Expr.SYMBOL, "Rule");

    /**
     * Converts the provided Mathematica expression to a {@code Map}.
     * @param solution the solution to a system of linear equations.
     * @return a {@code Map} representation of the provided {@code Expr}.
     */
    protected static Map<String, LinearComb<String>> interpretSolution(Expr solution) {
        return new MMaExprInterpretter().makeSolution(solution);
    }
    
    private MMaExprInterpretter() {
        // preventing instantiation
    }
    
    private Map<String, LinearComb<String>> makeSolution(Expr solution) {
        
        // Creating storage for the linear system's solution
        Map<String, LinearComb<String>> sol = 
            new TreeMap<String, LinearComb<String>>();
        
        // Getting the list (or vector) of rules
        List<Expr> rules = getVector(solution);
        
        // Iterating thru each Rule, converting it to a map entry
        for (Expr rule : rules) {
            
            // Converting the Rule to a singleton Map
            Map<Expr, Expr> ruleMap = getRule(rule);
            
            // Getting that maps sole entry
            Map.Entry<Expr, Expr> entry = ruleMap.entrySet().iterator().next();
            
            // Converting the key to a variable name
            String var = getStringFromSymbol(entry.getKey());
            
            // Converting the value to a linear combination
            LinearComb<String> value = getLinComb(entry.getValue());
            
            sol.put(var, value);
        }
        
        return sol;
    }
    
    private Map<Expr, Expr> getRule(Expr rule) {

        // Checking the head
        if (!rule.head().equals(RULE)) {
            String s = rule.toString() + " is not a " + RULE.toString();
            throw new IllegalArgumentException(s);
        }

        // Checking the length
        if (rule.length() != 2) {
            String s = rule.toString() + " does not have 2 arguments.";
            throw new IllegalArgumentException(s);
        }

        return Collections.singletonMap(rule.part(1), rule.part(2));
    }

    private List<Expr> getVector(Expr vector) {
        
        // Checking if vector
        if (!vector.vectorQ()) {
            String s = vector.toString() + " is not a vector";
            throw new IllegalArgumentException(s);
        }
        
        return getList(vector);
    }
    
    private List<Expr> getList(Expr list) {

        // Checking if list
        if (!list.listQ()) {
            String s = list.toString() + " is not a " + LIST.toString();
            throw new IllegalArgumentException(s);
        }

        int length = list.length();

        List<Expr> javaList = new ArrayList<Expr>(length);

        for (int i = 1; i <= length; i++) {
            javaList.add(list.part(i));
        }

        return javaList;
    }

    private LinearComb<String> getLinComb(Expr expr) {

        if (expr.head().equals(PLUS)) {
            return getLinCombFromPlus(expr);
        }
        
        if (expr.head().equals(TIMES)) {
            return getLinCombFromTimes(expr);
        }
        
        if (expr.symbolQ()) {
            return getLinCombFromSymbol(expr);
        }
        
        if (expr.numberQ()) {
            return getLinCombFromNumber(expr);
        }
        
        String s = "The expression " + expr + " was not understood.";
        throw new IllegalArgumentException(s);
    }
    
    private LinearComb<String> getLinCombFromNumber(Expr number) {
        
        double constant = getDouble(number);
        
        List<LinearComb.Term<String>> terms = Collections.emptyList();
        
        return new HashLinearComb<String>(terms, constant, 0d);
    }
    
    private LinearComb<String> getLinCombFromSymbol(Expr symbol) {
        
        LinearComb.Term<String> term = getTermFromSymbol(symbol);
        
        return HashLinearComb.singleTerm(term);
    }
    
    private LinearComb<String> getLinCombFromTimes(Expr times) {
        
        LinearComb.Term<String> term = getTermFromTimes(times);
        
        return HashLinearComb.singleTerm(term);
    }
    
    private LinearComb<String> getLinCombFromPlus(Expr plus) {

        // Checking the head
        if (!plus.head().equals(PLUS)) {
            String s = plus.toString() + " is not a " + PLUS.toString();
            throw new IllegalArgumentException(s);
        }
        
        // Creating storage for all terms
        List<LinearComb.Term<String>> terms = 
            new LinkedList<LinearComb.Term<String>>();
        
        double constant = Double.NaN;
        
        int length = plus.length();
        
        for (int i = 1; i <= length; i ++) {
            
            Expr partI = plus.part(i);
            
            if (partI.numberQ() && Double.isNaN(constant)) {
                constant = getDouble(partI);
            }
            
            else if (partI.numberQ() && !Double.isNaN(constant)) {
                String s = "More than one constant exists in " + plus;
                throw new IllegalArgumentException(s);
            }
            
            else {
                LinearComb.Term<String> t = getTermFromSymbolOrTimes(partI);
                terms.add(t);
            }
            
        }
        
        if (Double.isNaN(constant)) {
            constant = 0d;
        }
        
        return new HashLinearComb<String>(terms, constant, 0d);
    }
    
    private LinearComb.Term<String> getTermFromSymbol(Expr symbol) {
        String var = getStringFromSymbol(symbol);
        return new HashLinearComb.Term<String>(var);
    }
    
    private LinearComb.Term<String> getTermFromSymbolOrTimes(Expr symbolOrTimes) {
        
        if (symbolOrTimes.symbolQ()) {
            return getTermFromSymbol(symbolOrTimes);
        }

        if (symbolOrTimes.head().equals(TIMES)) {
            return getTermFromTimes(symbolOrTimes);
        }
        
        String s = symbolOrTimes + " is neither a Symbol nor " + TIMES;
        throw new IllegalArgumentException(s);
    }
    
    private String getStringFromSymbol(Expr symbol) {
        
        // Checking if symbol
        if (!symbol.symbolQ()) {
            String s = symbol + " is not a symbol.";
            throw new IllegalArgumentException(s);
        }
        
        try {
            return symbol.asString();
        } catch (ExprFormatException e) {
            throw new ExternalSoftwareException(e);
        }
    }
    
    private double getDouble(Expr number) {
        
        // Checking if number
        if (!number.numberQ()) {
            String s = number + " is not a number.";
            throw new IllegalArgumentException(s);
        }
        
        // Checking if complex
        if (number.complexQ()) {
            String s = number + " is a complex number.";
            throw new IllegalArgumentException(s);
        }
        
        try {
            return number.asDouble();
        } catch (ExprFormatException e) {
            throw new ExternalSoftwareException(e);
        }
    }
    
    private LinearComb.Term<String> getTermFromTimes(Expr times) {

        // Checking the head
        if (!times.head().equals(TIMES)) {
            String s = times.toString() + " is not a " + TIMES.toString();
            throw new IllegalArgumentException(s);
        }

        // Checking length
        if (times.length() != 2) {
            String s = times.toString() + " does not have 2 arguments.";
            throw new IllegalArgumentException(s);
        }
        
        double coeff = getDouble(times.part(1));
        String var = getStringFromSymbol(times.part(2));

        return new HashLinearComb.Term<String>(var, coeff);
    }

}
