/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.ext.mma.algebra;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import sbrt.ext.ExternalSoftwareException;
import sbrt.ext.mma.MathematicaException;
import sbrt.kernel.math.HashLinearComb;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.algebra.HashLinearSystemSolution;
import sbrt.kernel.math.algebra.LinearSystemSolution;
import sbrt.kernel.math.algebra.LinearSystemSolver;
import sbrt.kernel.utilities.Util;

import com.wolfram.jlink.Expr;
import com.wolfram.jlink.KernelLink;
import com.wolfram.jlink.MathLinkException;
import com.wolfram.jlink.MathLinkFactory;

/**
 * This class is used to solve systems of linear equations using Mathematica.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <V> the variable type.
 */
public class MathematicaSolver<V> implements LinearSystemSolver<V> {

    private static Expr PLUS = new Expr(Expr.SYMBOL, "Plus");

    private static Expr TIMES = new Expr(Expr.SYMBOL, "Times");

    private static Expr EQUAL = new Expr(Expr.SYMBOL, "Equal");

    private static Expr LIST = new Expr(Expr.SYMBOL, "List");

    /**
     * Used for testing purposes.
     * 
     * @param args ignored.
     */
    public static void main(String[] args) {

        String mathKernel = "/usr/local/bin/math -mathlink";

        LinearComb.Term<String> term1 = new HashLinearComb.Term<String>("x1");
        LinearComb.Term<String> term2 = new HashLinearComb.Term<String>("x2");
        LinearComb.Term<String> term3 = new HashLinearComb.Term<String>("x3");
        LinearComb.Term<String> term4 = new HashLinearComb.Term<String>("x4");

        List<LinearComb.Term<String>> terms =
                new LinkedList<LinearComb.Term<String>>();

        terms.add(term1);
        terms.add(term2);
        terms.add(term3);
        terms.add(term4);

        LinearComb<String> expr1 =
                new HashLinearComb<String>(terms, -1234.141, 0);

        terms.clear();
        terms.add(term2);
        terms.add(term3);

        LinearComb<String> expr2 = new HashLinearComb<String>(terms, 88, 0);

        List<LinearComb<String>> expressions =
                new LinkedList<LinearComb<String>>();
        expressions.add(expr1);
        expressions.add(expr2);

        Map<LinearComb<String>, Double> equations =
                new HashMap<LinearComb<String>, Double>();

        equations.put(expr1, 5d);
        equations.put(expr2, 235.1);

        MathematicaSolver<String> ms =
                new MathematicaSolver<String>(mathKernel);
        LinearSystemSolution<String> solution = ms.solve(equations);

        Util.print(solution.toMap().entrySet());
    }

    private KernelLink mma;

    private Map<V, String> origToMma;

    private Map<String, V> mmaToOrig;

    /**
     * Creates a new linear system solver.
     * 
     * @param kernelLinkArgs the argument for the Mathematica JLink method
     * {@linkplain MathLinkFactory#createKernelLink(String[])}.
     * 
     * @throws ExternalSoftwareException if a {@code MathLinkException} is
     * thrown.
     */
    public MathematicaSolver(String[] kernelLinkArgs) {

        try {

            mma = MathLinkFactory.createKernelLink(kernelLinkArgs);
            mma.connect(10000); // Wait at most 10 seconds
            mma.discardAnswer();

        } catch (MathLinkException e) {

            String s = "An error occurred connecting to the MMa kernel.";
            throw new MathematicaException(s, e);
        }

        origToMma = new HashMap<V, String>();
        mmaToOrig = new TreeMap<String, V>();
    }

    /**
     * Creates a new linear system solver.
     * 
     * @param kernelLinkCmd the argument for the Mathematica JLink method
     * {@linkplain MathLinkFactory#createKernelLink(String)}.
     * 
     * @throws MathematicaException if a {@code MathLinkException} is
     * thrown.
     */
    public MathematicaSolver(String kernelLinkCmd) {

        try {

            mma = MathLinkFactory.createKernelLink(kernelLinkCmd);
            mma.connect(10000); // Wait at most 10 seconds
            mma.discardAnswer();

        } catch (MathLinkException e) {

            String s = "An error occurred connecting to the MMa kernel.";
            throw new MathematicaException(s, e);
        }

        origToMma = new HashMap<V, String>();
        mmaToOrig = new TreeMap<String, V>();
    }

    /**
     * Closes the link to the Mathematica kernel.
     */
    public void close() {
        mma.close();
        mma = null;
    }

    /**
     * Calls {@code close()}.
     * 
     * @throws Throwable
     */
    protected void finalize() throws Throwable {
        close();
        super.finalize();
    }

    /**
     * Computes and returns a solution to the provided system of linear
     * equations.
     * 
     * @param equations linear combinations each mapped to a {@code Double}.
     * Each key-value pair represents the left and right hand sides,
     * respectively, of an equation.
     * 
     * @return a solution to the provided system of linear equations.
     * 
     * @throws MathematicaException if a {@code MathLinkException} is
     * thrown.
     * @throws IllegalStateException if {@code close()} has already been called.
     */
    public LinearSystemSolution<V> solve(Map<LinearComb<V>, Double> equations) {

        if (mma == null) {
            String s = "The link to the Mathematica kernel was already closed.";
            throw new IllegalStateException(s);
        }

        LinearSystemSolution<V> solution = null;

        try {
            solution = getSolution(equations);
        } catch (MathLinkException e) {
            throw new MathematicaException(e);
        }

        // Freeing up memory
        origToMma.clear();
        mmaToOrig.clear();

        return solution;
    }

    private LinearSystemSolution<V> 
    getSolution(Map<LinearComb<V>, Double> equations) throws MathLinkException {

        // Creating a Mathematica-safe variable name for each
        // provided variable
        createMMaVariables(equations.keySet());

        // Creating a Mathematica expression to solve the provided
        // system of linear equations
        Expr solution = getRationalSolution(equations);

        // Calling the Mathematica kernel to solve the system
        mma.evaluate(solution);
        mma.waitForAnswer();

        // Retrieving the answer from the Mathematica kernel
        Expr ans = mma.getExpr();

        // Converting the Mathematica expression into Java objects
        Map<String, LinearComb<String>> javaSolution =
                MMaExprInterpretter.interpretSolution(ans);

        // Converting the Mathematica variables back to their original form
        Map<V, LinearComb<V>> solToOriginal = convertVars(javaSolution);

        // Adding equations for the free variables as well
        // (i.e. free variable = free variable)
        addFreeVariables(solToOriginal);

        // Returning as a linear system solution
        return new HashLinearSystemSolution<V>(solToOriginal);
    }

    private void addFreeVariables(Map<V, LinearComb<V>> solution) {

        Set<V> freeVariables = new HashSet<V>();

        // Gathering free variables
        for (LinearComb<V> expr : solution.values()) {
            freeVariables.addAll(expr.getVariables());
        }

        // Adding them to the solution
        for (V freeVar : freeVariables) {

            // Creating a linear combination
            LinearComb<V> expr = HashLinearComb.singleTerm(freeVar);

            if (solution.put(freeVar, expr) != null) {
                throw new IllegalStateException();
            }
        }

    }

    private Map<V, LinearComb<V>> convertVars(Map<String, LinearComb<String>> equations) {

        // Determing appropriate size
        int size = equations.size() + 1;

        Map<V, LinearComb<V>> m = new HashMap<V, LinearComb<V>>(size, 1f);

        for (Map.Entry<String, LinearComb<String>> e : equations.entrySet()) {

            String mmaVariable = e.getKey();
            LinearComb<String> mmaExpr = e.getValue();

            // Converting Mathematica variables back to variables from
            // the original system
            V origVariable = mmaToOrig.get(mmaVariable);
            LinearComb<V> origExpr = convertVars(mmaExpr);

            m.put(origVariable, origExpr);
        }

        return m;
    }

    private LinearComb<V> convertVars(LinearComb<String> expr) {

        List<LinearComb.Term<V>> terms = new LinkedList<LinearComb.Term<V>>();

        for (LinearComb.Term<String> term : expr) {

            String mmaVar = term.getVariable();
            double coeff = term.getCoefficient();

            V origVar = mmaToOrig.get(mmaVar);

            terms.add(new HashLinearComb.Term<V>(origVar, coeff));
        }

        double constant = expr.getConstant();

        return new HashLinearComb<V>(terms, constant, 0d);
    }

    private void createMMaVariables(Collection<LinearComb<V>> exprs) {

        // Gathering all of the provided variables
        for (LinearComb<V> expr : exprs) {
            for (V variable : expr.getVariables()) {
                origToMma.put(variable, null);
            }
        }

        int count = 1;

        // Creating a Mathematica variable for each provided variable
        for (Map.Entry<V, String> e : origToMma.entrySet()) {

            String mmaVariable = "x" + Integer.toString(count++);

            e.setValue(mmaVariable);

            mmaToOrig.put(mmaVariable, e.getKey());
        }

    }

    private Expr flatten(Expr expr) {

        Expr flatten = new Expr(Expr.SYMBOL, "Flatten");

        // Creating args for Flatten[]
        Expr[] args = { expr };

        return new Expr(flatten, args);
    }

    private Expr getRationalSolution(Map<LinearComb<V>, Double> equations) {

        // Converting provided list to rationalized equations
        Expr eq = getRationalEquations(equations);

        // Creating arguments for Solve[]
        Expr[] args = { eq };

        Expr solve = new Expr(Expr.SYMBOL, "Solve");

        return flatten(new Expr(solve, args));
    }

    private Expr getRationalEquations(Map<LinearComb<V>, Double> equations) {

        // Getting the list of all equations
        Expr list = getEquationList(equations);

        // Creating arguments for Rationalize[]
        Expr[] args = { list, new Expr(0) };

        Expr rationalize = new Expr(Expr.SYMBOL, "Rationalize");

        return new Expr(rationalize, args);
    }

    private Expr getEquationList(Map<LinearComb<V>, Double> equations) {

        // Creating storage for all expressions
        Expr[] allExprs = new Expr[equations.size()];

        int index = 0;

        for (Map.Entry<LinearComb<V>, Double> e : equations.entrySet()) {
            allExprs[index++] = getEquation(e.getKey(), e.getValue());
        }

        return new Expr(LIST, allExprs);
    }

    private Expr getEquation(LinearComb<V> leftSide, Double rightSide) {

        // Creating left and right hand expressions
        Expr leftExpr = getExpression(leftSide);
        Expr rightExpr = new Expr(rightSide);

        // Creating arguments for Equal[]
        Expr[] sides = { leftExpr, rightExpr };

        return new Expr(EQUAL, sides);
    }

    private Expr getExpression(LinearComb<V> expression) {

        // Creating storage for all terms and the constant
        Expr[] allTerms = new Expr[expression.terms() + 1];

        int index = 0;

        // Adding the constant first
        allTerms[index++] = new Expr(expression.getConstant());

        // Adding each term
        for (LinearComb.Term<V> t : expression) {

            // Getting the MMa variable name
            String varName = origToMma.get(t.getVariable());
            double coeff = t.getCoefficient();

            // Creating arguments for Times[]
            Expr[] args = { 
            		new Expr(Expr.SYMBOL, varName),
            		new Expr(coeff) 
            };

            // Multiplying the variable by the coefficient
            Expr term = new Expr(TIMES, args);

            allTerms[index++] = term;
        }

        Expr expr = null;

        // If more than one Expr exists, add them together
        if (allTerms.length > 1) {
            expr = new Expr(PLUS, allTerms);
        }

        // Otherwise, just use the constant
        else {
            expr = allTerms[0];
        }

        return expr;
    }

}
