/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import java.util.Collections;
import java.util.List;

import sbrt.kernel.math.HashLinearComb;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.LinearComb.Term;

/**
 * This class is a skeleton implementation of {@code ExchangeRxn}.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 */
public abstract class AbstractExchangeRxn implements ChemRxn, ExchangeRxn {

    // The associated chemical species of this exchange reaction
    private final String species;

    /**
     * Constructs a new exchange reaction for the provided chemical species.
     * 
     * @param chemSpecies the chemical species associated with this exchange reaction.
     * 
     * @throws IllegalArgumentException if an empty string is provided.
     */
    public AbstractExchangeRxn(String chemSpecies) {
        
        if (chemSpecies.length() == 0) {
        	String s = "An empty string was provided.";
        	throw new IllegalArgumentException(s);
        }
        
        this.species = chemSpecies;
    }

	/**
	 * Returns the chemical species associated with this exchange reaction.
	 * @return the chemical species associated with this exchange reaction.
	 */
    public final String getSpecies() {
        return species;
    }

    /**
     * Returns an empty linear combination.
     * @return an empty linear combination.
     */
    public LinearComb<String> getReactants() {
        
        // Creating empty list of terms
        List<Term<String>> terms = Collections.emptyList();
        
        // Returning empty linear combination
        return new HashLinearComb<String>(terms, 0d, 0d);
    }

    /**
     * Returns an empty linear combination.
     * @return an empty linear combination.
     */
    public LinearComb<String> getProducts() {
        
        // Creating empty list of terms
        List<Term<String>> terms = Collections.emptyList();
        
        // Returning empty linear combination
        return new HashLinearComb<String>(terms, 0d, 0d);
    }

    /**
     * Returns {@code obj == this}.
     * 
     * @param obj the reference object with which to compare.
     * 
     * @return {@code obj == this}.
     */
    public boolean equals(Object obj) {
    	return obj == this;
    }

    /**
     * Returns a content-based hash code for this exchange reaction.
     * @return a content-based hash code for this exchange reaction.
     */
    public int hashCode() {
        return species.hashCode();
    }
    
}
