/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.LinearComb.Term;

/**
 * This class is a skeleton implementation of ChemicalRxn. It performs basic
 * functions that are useful for most classes implementing ChemicalRxn.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 */
public abstract class AbstractRxn implements ChemRxn {

	// The linear combinations of reactants and products
	private final LinearComb<String> reactants;
	private final LinearComb<String> products;

	/**
	 * Constructs a new chemical reaction from the provided reactants
	 * and products with the given name.
	 * 
	 * @param reactants
	 *            the linear combination of chemical species representing the
	 *            "left-hand side" of the reaction.
	 * 
	 * @param products
	 *            the linear combination of chemical species representing the
	 *            "right-hand side" of the reaction.
	 * 
	 * 
	 * @throws NullPointerException
	 *             if the reactants or products are {@code null}, or if any
	 *             chemical species within them is {@code null}.
	 * 
	 * @throws IllegalArgumentException
	 *             if
	 * 
	 * <P>
	 * &bull; the reactants or products contain a chemical species in common,
	 * </P>
	 * <P>
	 * &bull; either linear combination contains a non-zero constant,
	 * </P>
	 * <P>
	 * &bull; a stoichiometric coefficient is NaN, infinite, or &le; 0.
	 * </P>
	 */
	public AbstractRxn(LinearComb<String> reactants,
			           LinearComb<String> products) {

		checkInputs(reactants, products);

		// Saving
		this.reactants = reactants;
		this.products = products;
	}

	/**
	 * Returns the reactants of this reaction.
	 * 
	 * @return the reactants of this reaction.
	 */
	public final LinearComb<String> getReactants() {
		return reactants;
	}

	/**
	 * Returns the products of this reaction.
	 * 
	 * @return the products of this reaction.
	 */
	public final LinearComb<String> getProducts() {
		return products;
	}

	/**
	 * Returns an appropriate string representation of this reaction.
	 * 
	 * @return an appropriate string representation of this reaction.
	 */
	public abstract String toString();

	/**
	 * Returns a content-based hash code for this reaction.
	 * 
	 * @return a content-based hash code for this reaction.
	 */
	public int hashCode() {
		return 37 * reactants.hashCode() ^ products.hashCode();
	}

    /**
     * Returns {@code obj == this}.
     * 
     * @param obj the reference object with which to compare.
     * 
     * @return {@code obj == this}.
     */
	public boolean equals(Object obj) {
		return obj == this;
	}

	/**
	 * Ensures the provided reactants and products are well-formed.
	 */
	private static void checkInputs(LinearComb<String> reactants,
			                        LinearComb<String> products) {

		// Ensuring the provided stoichiometric expressions are well-formed
		checkExpr(reactants);
		checkExpr(products);

		// Iterating thru each reactant
		for (Object species : reactants.getVariables()) {

			// If the products also contain this chemical species
			if (products.hasVariable(species)) {

				// Throw exception
				String s = "reactants [" + reactants + "] and products ["
						+ products + "] contain the same chemical species: "
						+ species;
				throw new IllegalArgumentException(s);
			}
		}
	}

	/**
	 * Ensures the provided stoichiometric expression is well-formed.
	 */
	private static void checkExpr(LinearComb<String> stoichiometricExpr) {

		checkConstant(stoichiometricExpr);

        if (stoichiometricExpr.getVariables().isEmpty()) {
            String s = "empty stoichiometric expression: " + stoichiometricExpr;
            throw new IllegalArgumentException(s);
        }
        
		// Iterating thru each term in the stoichiometric expression
		for (Term<String> term : stoichiometricExpr) {
			checkSpecies(term.getVariable(), stoichiometricExpr);
			checkCoeff(term.getCoefficient(), stoichiometricExpr);
		}
	}

	private static void checkSpecies(String species, LinearComb<String> expr) {

		if (species.length() == 0) {
			String s = "Stoichiometric expression with an empty string: " + expr;
			throw new IllegalArgumentException(s);
		}

	}

	private static void checkConstant(LinearComb<?> expr) {

		if (expr.getConstant() == 0d) {
			return;
		}

		String s = "Stoichiometric expression with non-zero constant: " + expr;
		throw new IllegalArgumentException(s);
	}

	private static void checkCoeff(double coeff, LinearComb<?> expr) {

		if (invalidCoeff(coeff)) {
			String s = "Invalid stoichiometric coefficient (" + coeff
					+ ") in: " + expr;
			throw new IllegalArgumentException(s);
		}

	}

	/**
	 * Returns true if the provided coefficient is invalid; false otherwise.
	 */
	private static boolean invalidCoeff(double x) {
		return x <= 0d || Double.isInfinite(x) || Double.isNaN(x);
	}

}
