/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import java.util.Collection;
import java.util.Map;
import java.util.Set;

import sbrt.kernel.utilities.LogicalRelation;

/**
 * This interface is used to represent fluxomes with catalyzed reactions.
 * 
 * @author This interface was written and documented by Jeremiah Wright while in
 *         the Wagner lab.
 */
public interface CatalyzedFluxome extends Fluxome {

	/**
	 * Returns the chemical reactions contained in this fluxome that are
	 * catalyzed by the provided catalyst.
	 * 
	 * @param catalyst
	 *            the name of the catalyst whose reactions are to be returned.
	 * 
	 * @return reaction names mapped to their respective chemical reactions.
	 */
	public Map<String, ? extends ChemRxn> getCatalyzedRxns(String catalyst);

	/**
	 * Returns the set of names of all chemical reactions contained in this
	 * fluxome that are catalyzed by the provided catalyst.
	 * 
	 * @param catalyst
	 *            the name of the catalyst whose reaction names are to be
	 *            returned.
	 * 
	 * @return the set of names of all chemical reactions contained in this
	 *         fluxome that are catalyzed by the provided catalyst.
	 */
	public Set<String> getCatalyzedRxnNames(String catalyst);

    /**
	 * Returns the boolean logic rule for catalysts of the reaction with the
	 * provided name.
	 * 
	 * @param rxnName
	 *            the name of the reaction whose catalyst rule are to be
	 *            returned.
	 * 
	 * @return the boolean relationship of the catalysts of the specified
	 *         reaction.
	 */
    public LogicalRelation<?> getCatalystRule(String rxnName);

    /**
	 * Returns the boolean logic rules for the catalysts of all reactions in this
	 * fluxome.
	 * 
	 * @return reaction names mapped to the boolean relationship of their
	 *         catalysts.
	 */
    public Map<String, ? extends LogicalRelation<?>> getCatalystRules();
    
	/**
	 * Returns the set of all catalysts contained in this fluxome. 
	 * @return the set of all catalysts contained in this fluxome. 
	 */
	public Set<String> getCatalysts();

	/**
	 * Indicates if the provided string is the name of a catalyst contained in
	 * this fluxome.
	 * 
	 * @param s the string to test.
	 * 
	 * @return {@code true} if the string is a valid catalyst; {@code false}
	 *         otherwise.
	 */
	public boolean isCatalyst(String s);

	/**
	 * Returns the total number of catalysts contained in this fluxome.
	 * @return the total number of catalysts contained in this fluxome.
	 */
	public int catalysts();

	/**
	 * Indicates if the provided catalyst is required for the reaction with the
	 * provided name.
	 * 
	 * @param catalyst the catalyst.
	 * @param rxnName the name of the reaction.
	 * 
	 * @return {@code true} if the provided catalyst is required by the reaction
	 *         with the provided name; {@code false} otherwise.
	 */
    public boolean isReqCatalyst(String catalyst, String rxnName);

    /**
     * Indicates if the reaction with the provided name could function if
     * each of the catalysts in the provided collection were simultaneously
     * absent.
	 * 
	 * @param catalysts the collection of catalysts.
	 * @param rxnName the name of the reaction.
	 * 
	 * @return {@code true} if the provided catalysts are required for the 
	 * specified reaction to function; {@code false} otherwise.
	 * 
	 * @throws IllegalArgumentException
	 *             if {@code rxnName} is not a valid reaction name.
	 */
    public boolean areReqCatalysts(Collection<String> catalysts, String rxnName);
    
	/**
	 * Returns a copy of this fluxome, but with all constraints replaced by
	 * those provided. If a flux constraint for a particular reaction was not
	 * provided, its default value will be used.
	 * 
	 * @param constraints the new set of flux constraints.
	 * 
	 * @param constraintTolerance
	 *            the allowed amount by which constraints can violate their
	 *            maximum defined intervals. These maximum defined intervals are
	 *            implementation-dependent.
	 * 
	 * @return a copy of this fluxome, but with all constraints replaced by
	 *         those provided.
	 */
	public CatalyzedFluxome replaceConstraints(FluxConstraints constraints,
									           double constraintTolerance);

}
