/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import sbrt.kernel.math.*;

/**
 * This interface is used to represent chemical reactions. A chemical reaction
 * is composed of reactants and products, which are represented as
 * linear combinations of chemical species.
 * 
 * @author This interface was written and documented by Jeremiah Wright while in
 *         the Wagner Lab.
 */
public interface ChemRxn {

	/**
	 * Returns the reactants of this chemical reaction.
	 * 
	 * @return the reactants of this chemical reaction.
	 */
    public LinearComb<String> getReactants();

    /**
	 * Returns the products of this chemical reaction.
	 * 
	 * @return the products of this chemical reaction.
	 */
    public LinearComb<String> getProducts();

    /**
	 * Indicates if this chemical reaction is considered reversible.
	 * 
	 * @return {@code true} if this reaction is reversible; {@code false}
	 *         otherwise.
	 */
    public boolean isReversible();
    
    /**
     * Returns an appropriate string representation of this chemical reaction.
     * @return an appropriate string representation of this chemical reaction.
     */
    public String toString();

    /**
	 * Indicates if the provided object is "equal to" this chemical reaction.
	 * 
	 * @param obj the reference object with which to compare.
	 * 
	 * @return {@code true} if this reaction is equal to the provided object;
	 *         {@code false} otherwise.
	 */
    public boolean equals(Object obj);

    /**
     * Returns a content-based hash code for this chemical reaction.
     * @return a content-based hash code for this chemical reaction.
     */
    public int hashCode();

}
