/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;

/**
 * This class is used to group sets of flux constraints.
 * <P>
 * Typically, a constraint in FBA only applies to the flux of a single chemical
 * reaction. This type of constraint is referred to here as a "flux constraint".
 * </P>
 * <P>
 * It is also possible, however, to constrain the value of a mathematical
 * expression containing one or more fluxes. This type of constraint is referred
 * to here as an "expression constraint", or sometimes as a "reaction
 * expression constraint".
 * </P>
 * <p>
 * Instances of this class are immutable.
 * </p>
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner Lab.
 */
public class FluxConstraints {

	// All constraints for fluxes through single reactions
	private final UnmodifiableMap<String, Interval> fluxConstraints;
	
	// All constraints for mathematical expressions of fluxes
	private final UnmodifiableMap<MathExpr<String>, Interval>
		exprConstraints;

	/**
	 * Constructs a set of FBA constraints from the provided arguments.
	 * 
	 * @param fluxConstraints reactions names mapped to their 
	 * corresponding flux constraint.
	 *
	 * @param exprConstraints mathematical expressions of reaction names
	 * mapped to their corresponding constraint.
	 */
	public FluxConstraints(Map<String, Interval> fluxConstraints,
						   Map<? extends MathExpr<String>, Interval> exprConstraints) {
		
		// Checking inputs
		Util.nullCheck(fluxConstraints);
		Util.nullCheck(exprConstraints);
		
		// Setting flux constraints
		this.fluxConstraints = 
			new UnmodifiableMap<String, Interval>(fluxConstraints);
		
		// Setting expression constraints
		this.exprConstraints = 
			new UnmodifiableMap<MathExpr<String>, Interval>(exprConstraints);
	}

    /**
     * Constructs a set of FBA constraints from the provided argument.
     * 
     * @param fluxConstraints reactions names mapped to their 
     * corresponding flux constraint.
     */
    public FluxConstraints(Map<String, Interval> fluxConstraints) {

        // Checking input
        Util.nullCheck(fluxConstraints);
        
        // Setting flux constraints
        this.fluxConstraints = 
            new UnmodifiableMap<String, Interval>(fluxConstraints);
        
        // Creating empty map
        Map<MathExpr<String>, Interval> empty = Collections.emptyMap();
        
        // Setting expression constraints
        this.exprConstraints = 
            new UnmodifiableMap<MathExpr<String>, Interval>(empty);
    }
    
	/**
	 * Returns the flux constraints for single reactions.
	 * @return the flux constraints for single reactions.
	 */
	public UnmodifiableMap<String, Interval> getFluxConstraints() {
		return fluxConstraints;
	}
	
	/**
	 * Returns the constraints for mathematical expressions of fluxes.
	 * @return the constraints for mathematical expressions of fluxes.
	 */
	public UnmodifiableMap<MathExpr<String>, Interval> getExprConstraints() {
		return exprConstraints;
	}
	
	/**
	 * Returns the flux constraint for the provided reaction.
	 * 
	 * @param rxnName the name of the reaction whose flux constraint is to be returned.
	 * 
	 * @return the flux constraint for the provided reaction, or 
	 * {@code null} if the reaction is not present in this set.
	 */
	public Interval getConstraint(String rxnName) {
		return fluxConstraints.get(rxnName);
	}
	
	/**
	 * Returns the constraint of the provided mathematical expression of
	 * reaction names.
	 * 
	 * @param rxnExpr the mathematical expression of reaction names whose
	 * constraint is to be returned.
	 * 
	 * @return the constraint of the provided mathematical expression of
	 * reaction names, or {@code null} if the expression is not
	 * present in this set.
	 */
	public Interval getConstraint(MathExpr<String> rxnExpr) {
		return exprConstraints.get(rxnExpr);
	}
	
	/**
	 * Returns the total number of elements contained in this set of constraints.
	 * @return the total number of elements contained in this set of constraints.
	 */
	public int total() {
		return fluxConstraints.size() + exprConstraints.size();
	}
}
