/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.math.algebra.*;

/**
 * This interface is used to represent fluxomes for use in Flux Balance
 * Analysis. A fluxome is a set of biochemical reactions whose rates (i.e.
 * fluxes) may be subject to constraint. A fluxome is composed of
 * 
 * <table width="100%">
 * <tr>
 * <td width="5%">
 * <td>&bull; chemical species</tr>
 * <tr>
 * <td width="5%">
 * <td>&bull; the chemical reactions in which those chemical species
 * participate</tr>
 * <tr>
 * <td width="5%">
 * <td>&bull; the constraints on the fluxes of those reactions, and</tr>
 * </table>
 * 
 * 
 * <P>
 * Two basic types of reactions exist in a fluxome: internal reactions and exchange
 * reactions. Each reaction has a unique identifier, that is, a name.
 * Each reaction also has an associated flux constraint, which is either assigned
 * explicitly or by default.
 * </P>
 * <P>
 * A flux constraint can be represented as &alpha; &le; V &le; &beta;, where
 * &alpha; and &beta; are the lower and upper flux bounds, respectively, and V
 * is the flux under consideration. For a particular type of fluxome, the
 * allowed flux constraints may have a defined range. That is to say, the
 * allowed values of &alpha; and &beta; may differ between fluxomes of different
 * types or even between reactions of different type within the same fluxome.
 * For example, a popular type of fluxome uses the following convention: 0 &le;
 * &alpha; &le; &infin;, &alpha; &le; &beta; &le; &infin; for internal reactions
 * and -&infin; &le; &alpha; &le; &infin, &alpha; &le; &beta; &le; &infin; for
 * exchange reactions. This interface provides a way to maintain the integrity
 * of these definitions or conventions using
 * {@linkplain #isValid(Interval, String, double)}.
 * 
 * @author This interface was written and documented by Jeremiah Wright while in
 *         the Wagner lab.
 */
public interface Fluxome {

	/**
	 * Returns the set of all chemical reactions contained in this fluxome.
	 * This set is the union of internal and exchange reactions.
	 * 
	 * @return reaction names mapped to their corresponding reaction.
	 */
	public Map<String, ChemRxn> getRxns();

    /**
     * Returns the chemical reactions with the provided names.
     * 
     * @param rxnNames the names of the reactions to be returned.
     * 
     * @return reaction names mapped to their corresponding reaction.
     */
    public Map<String, ChemRxn> getRxns(Collection<String> rxnNames);
    
	/**
	 * Returns the set of all internal chemical reactions contained
	 * in this fluxome.
	 * 
	 * @return reaction names mapped to their corresponding reaction.
	 */
	public Map<String, ? extends ChemRxn> getInternalRxns();

	/**
	 * Returns all of the exchange reactions contained in this fluxome.
	 * 
	 * @return reaction names mapped to their corresponding reaction.
	 */
	public Map<String, ? extends ExchangeRxn> getExchangeRxns();

	/**
	 * Returns the chemical reaction in this fluxome with the provided name.
	 * 
	 * @param reactionName the name of the chemical reaction to be returned.
	 * 
	 * @return the chemical reaction in this fluxome with the provided name.
	 */
	public ChemRxn getRxn(String reactionName);

	/**
	 * Returns the set of names of all chemical reactions 
	 * contained in this fluxome.
	 * 
	 * @return the set of names of all chemical reactions 
	 * contained in this fluxome.
	 */
	public Set<String> getRxnNames();
	
	/**
	 * Returns the set of names of all internal chemical reactions
	 * contained in this fluxome.
	 * 
	 * @return the set of names of all internal chemical reactions
	 * contained in this fluxome.
	 */
	public Set<String> getInternalRxnNames();

	/**
	 * Returns the set of names of all exchange reactions contained 
	 * in this fluxome.
	 * 
	 * @return the set of names of all exchange reactions contained 
	 * in this fluxome.
	 */
	public Set<String> getExchangeRxnNames();

	/**
	 * Returns the set of names of all chemical species participating
	 * in reactions contained in this fluxome.
	 * 
	 * @return the set of names of all chemical species participating
	 * in reactions contained in this fluxome.
	 */
	public Set<String> getSpecies();

	/**
	 * Returns the flux constraint for the reaction with the provided
	 * name.
	 * 
	 * @param reactionName the name of the reaction whose flux constraint is
	 * to be returned.
	 * 
	 * @return the flux constraint for the reaction with the provided
	 * name.
	 */
	public Interval getFluxConstraint(String reactionName);

	/**
	 * Returns the constraint corresponding to the provided
	 * mathematical expression of reaction names.
	 * 
	 * @param rxnNameExpr the mathematical expression of
	 * reaction names whose constraint is to be returned.
	 * 
	 * @return the constraint corresponding to the provided
	 * mathematical expression of reaction names.
	 */
	public Interval getRxnExprConstraint(MathExpr<String> rxnNameExpr);

	/**
	 * Returns the set of all constraints for fluxes and 
	 * mathematical expressions of fluxes in this fluxome.
	 * 
	 * @return the set of all constraints for fluxes and 
	 * mathematical expressions of fluxes in this fluxome.
	 */
	public FluxConstraints getConstraints();

	/**
	 * Returns the stoichiometry matrix of this fluxome.
	 * 
	 * @return the stoichiometry matrix of this fluxome.
	 * Row variables in this matrix are the names of chemical species,
	 * and column variables are reaction names.
	 */
	public SparseDoubleMatrix<String, String> getStoichMatrix();
    
	/**
	 * Indicates if the provided string is the name of a reaction
	 * contained in this fluxome.
	 * 
	 * @param s the string to test.
	 * 
	 * @return {@code true} if the string is a valid reaction name; 
	 * {@code false} otherwise.
	 */
	public boolean isRxnName(String s);

	/**
	 * Indicates if the provided string is the name of an internal reaction
	 * contained in this fluxome.
	 * 
	 * @param s the string to test.
	 * 
	 * @return {@code true} if the string is a valid internal reaction name; 
	 * {@code false} otherwise.
	 */
	public boolean isInternalRxnName(String s);

	/**
	 * Indicates if the provided string is the name of an exchange reaction
	 * contained in this fluxome.
	 * 
	 * @param s the string to test.
	 * 
	 * @return {@code true} if the string is a valid exchange reaction name; 
	 * {@code false} otherwise.
	 */
	public boolean isExchangeRxnName(String s);
	
	/**
	 * Indicates if every string in the provided collection is the name
	 * of a reaction contained in this fluxome. 
	 * 
	 * @param c a collection of strings.
	 * 
	 * @return {@code true} if every string is a valid reaction name; 
	 * {@code false} otherwise.
	 */
	public boolean areRxnNames(Collection<String> c);

	/**
	 * Indicates if the provided string is the name of a chemical species
	 * contained in this fluxome.
	 * 
	 * @param s the string to test.
	 * 
	 * @return {@code true} if the string is a valid chemical species; 
	 * {@code false} otherwise.
	 */
	public boolean isSpecies(String s);

	/**
	 * Indicates if the provided flux vector lies within the feasible space
	 * of this fluxome, given the provided tolerance value. The provided flux
	 * vector must satisfy both the steady state constraint and all flux 
	 * constraints.
	 * 
	 * @param fluxVector reaction names mapped to flux values.
	 * 
	 * @param tolerance the amount by which a computed value is allowed to
	 * violate a defined constraint. This should be small positive number, like
	 * 1E-9 for example.
	 * 
	 * @return {@code true} if the vector satisfies all constraints; 
	 * {@code false} otherwise.
	 */
	public boolean isFluxVector(Map<String, Double> fluxVector, 
								double tolerance);

	/**
	 * Indicates if the provided interval is a valid
	 * flux constraint for the reaction with the provided
	 * name.
	 * 
	 * @param constraint the flux constraint to test.
	 * @param rxnName the name of the reaction.
	 * @param tolerance the allowed amount by which the provided
	 * constraint can violate its maximum defined interval. This
	 * maximum defined interval is implementation-dependent.
	 * 
	 * @return {@code true} if the constraint is valid; 
	 * {@code false} otherwise.
	 * 
	 * @throws IllegalArgumentException if the provided reaction
	 * name does not exist in this fluxome.
	 */
	public boolean isValid(Interval constraint, String rxnName, double tolerance);

	/**
	 * Returns the total number of chemical reactions contained
	 * in this fluxome.
	 * 
	 * @return the total number of chemical reactions contained
	 * in this fluxome.
	 */
	public int rxns();

	/**
	 * Returns the total number of internal reactions contained in this fluxome.
	 * 
	 * @return the total number of internal reactions contained in this fluxome.
	 */
	public int internalRxns();

	/**
	 * Returns the total number of exchange reactions contained in this fluxome.
	 * 
	 * @return the total number of exchange reactions contained in this fluxome.
	 */
	public int exchangeRxns();

	/**
	 * Returns the total number of chemical species contained in this fluxome.
	 * @return the total number of chemical species contained in this fluxome.
	 */
	public int species();

	/**
	 * Returns the total number of constraints added to this fluxome.
	 * @return the total number of constraints added to this fluxome.
	 * 
	 * @see #replaceConstraints(FluxConstraints, double)
	 */
	public int addedConstraints();

	/**
	 * Returns a copy of this fluxome, but with all constraints
	 * replaced by those provided. If a flux constraint for a
	 * particular reaction was not provided, its default value
	 * will be used.
	 * 
	 * @param constraints the new set of flux constraints.
	 * 
	 * @param constraintTolerance the allowed amount by which 
	 * constraints can violate their maximum defined intervals. These
	 * maximum defined intervals are implementation-dependent.
	 *  
	 * @return a copy of this fluxome, but with all constraints
	 * replaced by those provided.
	 */
	public Fluxome replaceConstraints(FluxConstraints constraints,
									        double constraintTolerance);

}
