/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import sbrt.kernel.math.*;

/**
 * 
 * This class is used to represent irreversible chemical reactions. Reactions
 * that proceed nearly to completion may be approximated as irreversible.
 * Reversible reactions can also sometimes be treated as two irreversible
 * reactions, with one as the "forward" reaction and one as the "reverse"
 * reaction.
 * 
 * <P>
 * Instances of this class are immutable.
 * </P>
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner Lab.
 */
public final class IrrevRxn extends AbstractRxn {

	/**
	 * Constructs a new irreversible reaction from the provided reactants and
	 * products.
	 * 
	 * @param reactants
	 *            the linear combination of chemical species representing the
	 *            "left-hand side" of the reaction.
	 * 
	 * @param products
	 *            the linear combination of chemical species representing the
	 *            "right-hand side" of the reaction.
	 * 
	 * @throws NullPointerException
	 *             if the reactants or products are {@code null}, or if any
	 *             chemical species within them is {@code null}.
	 * 
	 * @throws IllegalArgumentException
	 *             if
	 * 
	 * <P>
	 * &bull; the reactants or products contain a chemical species in common,
	 * </P>
	 * <P>
	 * &bull; either linear combination contains a non-zero constant,
	 * </P>
	 * <P>
	 * &bull; a stoichiometric coefficient is NaN, infinite, or &le; 0.
	 * </P>
	 */
	public IrrevRxn(LinearComb<String> reactants, LinearComb<String> products) {
		super(reactants, products);
	}

	/**
	 * Returns {@code false}.
	 * @return {@code false}.
	 */
	public boolean isReversible() {
		return false;
	}
	
	/**
	 * Returns an appropriate string representation of this irreversible
	 * reaction.
	 * 
	 * @return an appropriate string representation of this irreversible
	 *         reaction.
	 */
	public String toString() {
		return getReactants() + " --> " + getProducts();
	}

}
