/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import java.util.Collection;
import java.util.Map;
import java.util.Set;

import sbrt.kernel.utilities.And;
import sbrt.kernel.utilities.Or;


/**
 * This interface is used to represent fluxomes that contain only irreversible
 * reactions. All internal reactions must be of the type {@linkplain IrrevRxn}, and
 * all exchange reactions must be of the types {@linkplain SourceRxn} and
 * {@linkplain SinkRxn}. This sort of representation offers a variety of advantages
 * over more commonly used models. It allows one to constrain all forward and
 * reverse reactions independently of each other, and it also means that the
 * flux space formed by this fluxome lies entirely in the positive orthant.
 * <P>
 * For example, an organism like yeast can consume ethanol as a food source or
 * produce it as a waste product, depending on a variety of factors. Breaking
 * apart the ethanol-associated exchange reactions into forward and reverse
 * (that is, source and sink) reactions allows one to model this phenomenon in
 * finer detail.
 * 
 * @author This interface was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 * 
 */
public interface IrrevRxnFluxome extends CatalyzedFluxome {

    /**
     * Ensures each element of the provided collection is a valid reaction name.
     * 
     * @param rxnNames a collection of strings.
     * 
     * @throws IllegalArgumentException if any element of this collection
     * is not a valid reaction name.
     * 
     * @throws NullPointerException if the collection is null.
     */
    public void checkRxnNames(Collection<String> rxnNames);

    /**
     * Ensures the provided string is a valid reaction name.
     * 
     * @param reactionName a potential reaction name.
     * 
     * @throws IllegalArgumentException if the provided string
     * is not a valid reaction name.
     */
    public void checkRxnName(String reactionName);

    /**
     * Returns all of the internal reactions in this fluxome.
     * @return reaction names mapped their reaction.
     */
    public Map<String, IrrevRxn> getInternalRxns();

    /**
     * Returns the internal reactions with the provided names.
     * @param rxnNames the names of the reactions to be returned.
     * @return reaction names mapped to their reaction.
     */
    public Map<String, IrrevRxn> getInternalRxns(Collection<String> rxnNames);

    /**
     * Returns the names of all source reactions in this fluxome.
     * @return the names of all source reactions in this fluxome.
     */
    public Set<String> getSourceRxnNames();

    /**
     * Returns the source reactions with the provided names.
     * @param rxnNames the names of the reactions to be returned.
     * @return reaction names mapped to their reaction.
     */
    public Map<String, SourceRxn> getSourceRxns(Collection<String> rxnNames);
    
    /**
     * Returns the names of all sink reactions in this fluxome.
     * @return the names of all sink reactions in this fluxome.
     */
    public Set<String> getSinkRxnNames();

    /**
     * Returns the sink reactions with the provided names.
     * @param rxnNames the names of the reactions to be returned.
     * @return reaction names mapped to their reaction.
     */
    public Map<String, SinkRxn> getSinkRxns(Collection<String> rxnNames);
    
    /**
     * Returns the set of all source reactions in this fluxome.
     * @return the names of all source reactions mapped to their corresponding
     * reaction.
     */
    public Map<String, SourceRxn> getSourceRxns();
    
    /**
     * Returns the set of all sink reactions in this fluxome.
     * @return the names of all sink reactions mapped to their corresponding
     * reaction.
     */
    public Map<String, SinkRxn> getSinkRxns();

    /**
     * Returns the internal reaction in this fluxome with the provided name.
     * @param rxnName the name of the internal reaction to return.
     * @return the internal reaction in this fluxome with the provided name.
     */
    public IrrevRxn getInternalRxn(String rxnName);

    /**
     * Returns the source reaction in this fluxome with the provided name.
     * @param rxnName the name of the source reaction to return.
     * @return the source reaction in this fluxome with the provided name.
     */
    public SourceRxn getSourceRxn(String rxnName);

    /**
     * Returns the sink reaction in this fluxome with the provided name.
     * @param rxnName the name of the sink reaction to return.
     * @return the sink reaction in this fluxome with the provided name.
     */
    public SinkRxn getSinkRxn(String rxnName);

	/**
	 * Indicates if the provided string is the name of a source reaction
	 * contained in this fluxome.
	 * 
	 * @param s the string to test.
	 * 
	 * @return {@code true} if the string is a valid source reaction name; 
	 * {@code false} otherwise.
	 */
    public boolean isSourceRxnName(String s);
    
	/**
	 * Indicates if the provided string is the name of a sink reaction
	 * contained in this fluxome.
	 * 
	 * @param s the string to test.
	 * 
	 * @return {@code true} if the string is a valid sink reaction name; 
	 * {@code false} otherwise.
	 */
    public boolean isSinkRxnName(String s);
    
    
	/**
	 * Returns the total number of source reactions contained 
	 * in this fluxome.
	 * 
	 * @return the total number of source reactions contained 
	 * in this fluxome.
	 */
    public int sourceRxns();

	/**
	 * Returns the total number of sink reactions contained 
	 * in this fluxome.
	 * 
	 * @return the total number of sink reactions contained 
	 * in this fluxome.
	 */
    public int sinkRxns();

    /**
     * Returns a copy of this fluxome, but with all constraints
     * replaced by those provided. If a flux constraint for a
     * particular reaction was not provided, its default value
     * will be used.
     * 
     * @param constraints the new set of flux constraints.
     * 
     * @param constraintTolerance the allowed amount by which 
     * constraints can violate their maximum defined intervals. These
     * maximum defined intervals are implementation-dependent.
     *  
     * @return a copy of this fluxome, but with all constraints
     * replaced by those provided.
     */
    public IrrevRxnFluxome replaceConstraints(FluxConstraints constraints,
                                              double constraintTolerance);
    
    /**
     * Returns the boolean logic statement for the catalysts of the reaction
     * with the provided name.
     * 
     * @param rxnName the name of the reaction for which the catalyst rules
     * are to be returned.
     * 
     * @return the boolean logic statement for the catalysts of the reaction
     * with the provided name. If the specified reaction is not catalyzed,
     * an empty (non-null) rule is returned.
     */
    public Or<And<String>> getCatalystRule(String rxnName);

    /**
     * Returns all of the boolean logic statements regarding reaction catalysis.
     * @return the names of catalyzed reactions mapped to their boolean logic
     * rules of their catalyst(s).
     */
    public Map<String, Or<And<String>>> getCatalystRules();
    
}
