/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import java.util.Collection;
import java.util.Map;

import sbrt.kernel.utilities.And;
import sbrt.kernel.utilities.Or;


/**
 * This interface is used to represent fluxomes that contain multiple types of
 * chemical reactions. These reaction types include: irreversible internal
 * reactions, reversible internal reactions, source reactions, sink reactions,
 * and reversible exchange reactions.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public interface MixedFluxome extends CatalyzedFluxome {

    /**
     * Ensures each element of the provided collection is a valid reaction name.
     * 
     * @param rxnNames a collection of strings.
     * 
     * @throws IllegalArgumentException if any element of this collection
     * is not a valid reaction name.
     * 
     * @throws NullPointerException if the collection is null.
     */
    public void checkRxnNames(Collection<String> rxnNames);

    /**
     * Ensures the provided string is a valid reaction name.
     * 
     * @param reactionName a potential reaction name.
     * 
     * @throws IllegalArgumentException if the provided string
     * is not a valid reaction name.
     */
    public void checkRxnName(String reactionName);

    /**
     * Returns all of the internal reactions in this fluxome.
     * @return reaction names mapped their reaction.
     */
    public Map<String, ChemRxn> getInternalRxns();

    /**
     * Returns the internal reactions with the provided names.
     * @param rxnNames the names of the reactions to be returned.
     * @return reaction names mapped to their reaction.
     */
    public Map<String, ChemRxn> getInternalRxns(Collection<String> rxnNames);


    /**
     * Returns the exchange reactions with the provided names.
     * @param rxnNames the names of the reactions to be returned.
     * @return reaction names mapped to their reaction.
     */
    public Map<String, ExchangeRxn> getExchangeRxns(Collection<String> rxnNames);
    
	/**
	 * Returns all of the exchange reactions contained in this fluxome.
	 * 
	 * @return reaction names mapped to their corresponding reaction.
	 */
    public Map<String, ExchangeRxn> getExchangeRxns();

    /**
	 * Returns the internal reaction in this fluxome with the provided name.
	 * 
	 * @param rxnName
	 *            the name of the internal reaction to return.
	 * @return the internal reaction in this fluxome with the provided name.
	 */
    public ChemRxn getInternalRxn(String rxnName);

    /**
	 * Returns the exchange reaction in this fluxome with the provided name.
	 * 
	 * @param rxnName
	 *            the name of the exchange reaction to return.
	 * @return the exchange reaction in this fluxome with the provided name.
	 */
    public ExchangeRxn getExchangeRxn(String rxnName);

    /**
	 * Returns a copy of this fluxome, but with all constraints replaced by
	 * those provided. If a flux constraint for a particular reaction was not
	 * provided, its default value will be used.
	 * 
	 * @param constraints
	 *            the new set of flux constraints.
	 * 
	 * @param constraintTolerance
	 *            the allowed amount by which constraints can violate their
	 *            maximum defined intervals. These maximum defined intervals are
	 *            implementation-dependent.
	 * 
	 * @return a copy of this fluxome, but with all constraints replaced by
	 *         those provided.
	 */
    public MixedFluxome replaceConstraints(FluxConstraints constraints,
                                              double constraintTolerance);
    
    /**
	 * Returns the boolean logic statement for the catalysts of the reaction
	 * with the provided name.
	 * 
	 * @param rxnName
	 *            the name of the reaction for which the catalyst rules are to
	 *            be returned.
	 * 
	 * @return the boolean logic statement for the catalysts of the reaction
	 *         with the provided name. If the specified reaction is not
	 *         catalyzed, an empty (non-null) rule is returned.
	 */
    public Or<And<String>> getCatalystRule(String rxnName);

    /**
	 * Returns all of the boolean logic statements regarding reaction catalysis.
	 * 
	 * @return the names of catalyzed reactions mapped to their boolean logic
	 *         rules of their catalyst(s).
	 */
    public Map<String, Or<And<String>>> getCatalystRules();
    
}
