/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import java.util.*;
import java.util.Map.*;

import sbrt.kernel.math.*;
import sbrt.kernel.math.algebra.*;
import sbrt.kernel.utilities.And;
import sbrt.kernel.utilities.Or;
import sbrt.kernel.utilities.Util;

/**
 * This class is a concrete implementation of {@code MixedFluxome}.
 * All internal reactions must be of the types {@linkplain IrrevRxn} and
 * {@linkplain RevRxn}, and all exchange reactions must be of the types
 * {@linkplain SourceRxn}, {@linkplain SinkRxn}, and {@linkplain RevExchRxn}.
 * 
 * <p>
 * Instances of this class are immutable.
 * </p>
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 * 
 */
public final class MixedFluxomeV1 implements MixedFluxome {

    private static final Or<And<String>> emptyRule;
    
    static {
        Set<And<String>> s = Collections.emptySet();
        emptyRule = new Or<And<String>>(s);
    }
    
    /**
	 * The constraint used by default for irreversible internal reactions 
	 * and sink reactions: [0, Infinity).
	 */
    public static final Interval IRREV =
        new Interval(0d, Double.POSITIVE_INFINITY);

    /**
     * The constraint used by default for reversible internal reactions: 
     * (-Infinity, Infinity).
     */
    public static final Interval REV =
        new Interval(Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY);

    /**
	 * The constraint used by default for source reactions and
	 * reversible exchange reactions: [0, 0]. 
	 */
    public static final Interval OFF = new Interval(0d, 0d);

    // The names of all internal irreversible reactions mapped to 
    // their corresponding reaction.
    private final Map<String, IrrevRxn> irrevRxns;
    
    // The names of all internal reversible reactions mapped to 
    // their corresponding reaction.
    private final Map<String, RevRxn> revRxns;
    
    // The names of all source reactions mapped to 
    // their corresponding reaction.
    private final Map<String, SourceRxn> sourceRxns;
    
    // The names of all sink reactions mapped to 
    // their corresponding reaction.
    private final Map<String, SinkRxn> sinkRxns;
    
    // The names of all reversible exchange reactions mapped to 
    // their corresponding reaction.
    private final Map<String, RevExchRxn> revExchRxns;
    
    // The names of catalyzed reactions mapped to the 
    // boolean rules for catalysts.
    private final Map<String, Or<And<String>>> catalysts;
    
    
    /** 
     * Storage for all flux and rxn expression constraints. 
     */
    private final FluxConstraints constraints;

    /**
     * Records the number of added constraints.
     */
    private int addedConstraints;

    private MixedFluxomeV1(MixedFluxomeV1 fluxome,
                           FluxConstraints fluxConstraints,
                           double constraintTolerance) {

        // Copying internal reactions
    	irrevRxns = new TreeMap<String, IrrevRxn>();
    	irrevRxns.putAll(fluxome.irrevRxns);
    	
    	revRxns = new TreeMap<String, RevRxn>();
    	revRxns.putAll(fluxome.revRxns);
    	
    	// Copying exchange reactions
    	sourceRxns = new TreeMap<String, SourceRxn>();
    	sourceRxns.putAll(fluxome.sourceRxns);
    	
    	sinkRxns = new TreeMap<String, SinkRxn>();
    	sinkRxns.putAll(fluxome.sinkRxns);
    	
    	revExchRxns = new TreeMap<String, RevExchRxn>();
    	revExchRxns.putAll(fluxome.revExchRxns);
        
        // Copying catalysts
        catalysts = new TreeMap<String, Or<And<String>>>();
        catalysts.putAll(fluxome.catalysts);

        // Adding the provided flux constraints
        constraints = makeConstraints(fluxConstraints, constraintTolerance);
    }


    /**
	 * Constructs a new mixed fluxome from the provided internal and exchange
	 * reactions and catalysts. The flux of each reaction is constrained to its
	 * default value.
	 * 
	 * @param rxns
	 *            reaction names mapped to the internal and exchange reactions
	 *            to be contained in this fluxome.
	 * 
	 * @param catalysts
	 *            the names of internal reactions mapped to the boolean logic
	 *            rules governing their catalysis. Only the names of catalyzed
	 *            reactions need to be present as keys in this map.
	 */
    public MixedFluxomeV1(Map<String, ? extends ChemRxn> rxns,
                          Map<String, Or<And<String>>> catalysts) {
        
        // Creating storage
    	irrevRxns = new TreeMap<String, IrrevRxn>();
    	revRxns = new TreeMap<String, RevRxn>();
    	sourceRxns = new TreeMap<String, SourceRxn>();
    	sinkRxns = new TreeMap<String, SinkRxn>();
    	revExchRxns = new TreeMap<String, RevExchRxn>();

        this.catalysts = new TreeMap<String, Or<And<String>>>();

        setRxns(rxns);
        setCatalysts(catalysts);
        
        // Setting default constraints
        constraints = makeDefaultConstraints();
    }

    /**
	 * Constructs a new mixed fluxome from the provided internal and exchange
	 * reactions. The flux of each reaction is constrained to its default value.
	 * 
	 * @param rxns
	 *            reaction names mapped to the internal and exchange reactions
	 *            to be contained in this fluxome.
	 */
    public MixedFluxomeV1(Map<String, ? extends ChemRxn> rxns) {
        
        // Creating storage
    	irrevRxns = new TreeMap<String, IrrevRxn>();
    	revRxns = new TreeMap<String, RevRxn>();
    	sourceRxns = new TreeMap<String, SourceRxn>();
    	sinkRxns = new TreeMap<String, SinkRxn>();
    	revExchRxns = new TreeMap<String, RevExchRxn>();
        catalysts = new TreeMap<String, Or<And<String>>>();

        setRxns(rxns);
        
        // Setting default constraints
        constraints = makeDefaultConstraints();
    }
    
    private void setCatalysts(Map<String, Or<And<String>>> catalysts) {
        
        for (Map.Entry<String, Or<And<String>>> e : catalysts.entrySet()) {
            
            String rxnName = e.getKey();
            Or<And<String>> catRules = e.getValue();
        
            Util.nullCheck(catRules);
            
            if (!irrevRxns.containsKey(rxnName) && !revRxns.containsKey(rxnName)) {
                String s = "An internal reaction named '" 
                			+ rxnName + "' does not exist.";
                throw new IllegalArgumentException(s);
            }
        
            this.catalysts.put(rxnName, catRules);
        }
                
    }
    
    private void setRxns(Map<String, ? extends ChemRxn> rxns) {
        
        // Sorting out the provided reactions
        for (Map.Entry<String, ? extends ChemRxn> e : rxns.entrySet()) {
            
            String name = e.getKey();
            ChemRxn rxn = e.getValue();
            
            Util.nullCheck(name);
            Util.nullCheck(rxn);
                        
            if (rxn instanceof IrrevRxn) {
            	irrevRxns.put(name, (IrrevRxn) rxn);
            }
            
            else if (rxn instanceof RevRxn) {
            	revRxns.put(name, (RevRxn) rxn);
            }
            
            else if (rxn instanceof SourceRxn) {
            	sourceRxns.put(name, (SourceRxn) rxn);
            }
            
            else if (rxn instanceof SinkRxn) {
            	sinkRxns.put(name, (SinkRxn) rxn);
            }
            
            else if (rxn instanceof RevExchRxn) {
            	revExchRxns.put(name, (RevExchRxn) rxn);
            }
            
            else {
                String s = "Invalid reaction type: " + rxn.getClass().getCanonicalName();
                throw new IllegalArgumentException(s);
            }
        }
        
    }
    
    
    private Map<String, Interval> getDefaultConstraints() {
    	
    	// Creating storage for flux constraints
    	Map<String, Interval> fluxConsts = new TreeMap<String, Interval>();
    	
        for (String rxnName : irrevRxns.keySet()) {
            fluxConsts.put(rxnName, IRREV);
        }
        
        for (String rxnName : revRxns.keySet()) {
            fluxConsts.put(rxnName, REV);
        }
        
        for (String rxnName : sourceRxns.keySet()) {
            fluxConsts.put(rxnName, OFF);
        }
        
        for (String rxnName : sinkRxns.keySet()) {
            fluxConsts.put(rxnName, IRREV);
        }
        
        for (String rxnName : revExchRxns.keySet()) {
            fluxConsts.put(rxnName, OFF);
        }

    	return fluxConsts;
    }
    
    private FluxConstraints makeDefaultConstraints() {
    	
    	// Getting default flux constraints
    	Map<String, Interval> fluxConsts = getDefaultConstraints();
    	
    	// By default, no reaction expressions exist
    	Map<MathExpr<String>, Interval> exprConsts = Collections.emptyMap();
    	
    	// No additional constaints were added
    	addedConstraints = 0;
    	
    	return new FluxConstraints(fluxConsts, exprConsts);
    }
    
    private FluxConstraints 
    makeConstraints(FluxConstraints constraints, double tolerance) {
        
        // Recording the number of added constraints
        addedConstraints = constraints.total();
    	
    	// Getting provided flux constraints
    	Map<String, Interval> prov = constraints.getFluxConstraints();
    	
    	// Checking provided reaction names
    	checkRxnNames(prov.keySet());
    	
    	// Getting default flux constraints
    	Map<String, Interval> fluxConsts = getDefaultConstraints();

    	// Iterating thru each provided flux constraint
    	for (Map.Entry<String, Interval> e : prov.entrySet()) {

    		// Getting the provided name-constraint pair
    		String rxnName = e.getKey();
    		Interval c = e.getValue();
    		
    		// Checking it
    		if (!isValid(c, rxnName, tolerance)) {
    			String s = "The flux constraint '" + c + 
    			           "' is not valid for: " + rxnName;
    			throw new IllegalArgumentException(s);
    		}

    		// Replacing the default with the one provided
    		fluxConsts.put(e.getKey(), c);
    	}

    	// Getting flux expression constriants
    	Map<MathExpr<String>, Interval> exprConsts = 
    		constraints.getExprConstraints();
    	
    	// Iterating thru each expression-constraint pair
    	for (Map.Entry<MathExpr<String>, Interval> e : exprConsts.entrySet()) {
    		
    		// Checking the rxn names
    		checkRxnNames(e.getKey().getVariables());
    		
    		// Checking flux expression constraint
    		if (e.getValue() == null) {
    			String s = "The constraint for " + e.getKey() +
    					   " is null.";
    			throw new NullPointerException(s);
    		}
    	}
    	
    	return new FluxConstraints(fluxConsts, exprConsts);
    }

    /**
     * Ensures each element of the provided collection is a valid reaction name.
     * 
     * @param rxnNames a collection of strings.
     * 
     * @throws IllegalArgumentException if any element of this collection
     * is not a valid reaction name.
     */
    public void checkRxnNames(Collection<String> rxnNames) {
    	
    	// Checking each name
    	for (String name : rxnNames) {
    		checkRxnName(name);
    	}
    	
    }

    /**
     * Ensures the provided string is a valid reaction name.
     * 
     * @param reactionName a potential reaction name.
     * 
     * @throws IllegalArgumentException if the provided string
     * is not a valid reaction name.
     */
    public void checkRxnName(String reactionName) {

    	// Ensuring name is valid
        if (!isRxnName(reactionName)) {
            String s = "A reaction named '" + reactionName + 
            		   "' does not exist.";
            throw new IllegalArgumentException(s);
        }

    }

    /**
     * Ensures the provided string is a valid catalyst.
     * 
     * @param catalyst a potential catalyst.
     * 
     * @throws IllegalArgumentException if the provided string
     * is not a valid catalyst.
     */
    public void checkCatalyst(String catalyst) {

        // Ensuring name is valid
        if (!isCatalyst(catalyst)) {
            String s = "A catalyst named '" + catalyst + 
                       "' does not exist.";
            throw new IllegalArgumentException(s);
        }

    }

	/**
	 * Returns a copy of this fluxome, but with all constraints replaced by
	 * those provided. The lower flux bounds for irreversible reactions,
	 * including source and sink reactions, must be {@code >= 0}. The remaining
	 * flux bounds can be any values except {@code NaN}, lower bounds cannot be
	 * {@code Double.POSITIVE_INFINITY}, and upper bounds cannot be
	 * {@code Double.NEGATIVE_INFINITY}.
	 * 
	 * @param fluxConstraints
	 *            the new set of constraints.
	 * 
	 * @param constraintTolerance
	 *            the allowed amount by which constraints can violate their
	 *            maximum defined intervals.
	 * 
	 * @return a copy of this fluxome, but with all constraints replaced by
	 *         those provided.
	 *         
	 * @see #isValid(Interval, String, double)
	 */
    public MixedFluxomeV1 
    	replaceConstraints(FluxConstraints fluxConstraints, 
    					   double constraintTolerance) {

        return new MixedFluxomeV1(this, 
        						  fluxConstraints,
        						  constraintTolerance);
    }

    /**
     * Returns all constraints in this fluxome.
     * @return all constraints in this fluxome.
     */
    public FluxConstraints getConstraints() {
    	return constraints;
    }
    
    /**
     * Returns the constraints on each flux in this fluxome.
     * @return each reaction name in this fluxome mapped to its
     * flux constraint.
     */
    public Map<String, Interval> getFluxConstraints() {
    	return constraints.getFluxConstraints();
    }
    
    /**
     * Returns the flux constraint for the provided reaction.
     * 
     * @param reactionName the name of the reaction whose
     * flux constraint will be returned.
     * 
     * @return the flux constraint for the provided reaction.
     */
    public Interval getFluxConstraint(String reactionName) {

    	// Checking input
        checkRxnName(reactionName);

        // Returning the constraint
        return constraints.getConstraint(reactionName);
    }

    /**
	 * Returns the constraint associated with the expression provided. If the
	 * expression provided does not exist in this fluxome, the constraint
	 * (-Infinity, Infinity) is returned.
	 * 
	 * @param rxnNameExpr the mathematical expression of reaction names
	 * whose constraint is to be returned.
	 * 
	 * @return the constraint on the provided mathematical expression
	 * of fluxes.
	 */
    public Interval getRxnExprConstraint(MathExpr<String> rxnNameExpr) {

    	// Checking reaction names
    	checkRxnNames(rxnNameExpr.getVariables());
    	
    	// Getting the provided constraint if it exists 
        Interval c = constraints.getConstraint(rxnNameExpr);

        // If it does exist, return it
        if (c != null) {
        	return c;
        }
        
        // Otherwise, return an unbounded constraint
        return REV;
    }

	/**
	 * Returns all of the exchange reactions contained in this fluxome.
     * 
	 * @return the names of all exchange reactions in this fluxome 
     * mapped to their corresponding reaction.
	 */
    public Map<String, ExchangeRxn> getExchangeRxns() {
    	
        Map<String, ExchangeRxn> m =
            new TreeMap<String, ExchangeRxn>();
        
        m.putAll(sourceRxns);
        m.putAll(sinkRxns);
        m.putAll(revExchRxns);
        
        return m;
    }

	/**
	 * Returns all of the internal reactions contained in this fluxome
	 * that are catalyzed by the provided catalyst.
	 * 
	 * @param catalyst the name of the catalyst whose reactions
	 * are to be returned.
	 * 
	 * @return the set of all internal reactions contained in this fluxome
	 * that are catalyzed by the provided catalyst.
	 * 
	 * @throws IllegalArgumentException if the provided catalyst
	 * does not exist in this fluxome.
	 */
    public Map<String, ChemRxn> getCatalyzedRxns(String catalyst) {

        // Creating storage
        Map<String, ChemRxn> m = new TreeMap<String, ChemRxn>();
        
        // Iterating thru all reactions that are catalyzed
        for (Map.Entry<String, Or<And<String>>> e : catalysts.entrySet()) {
            
            String rxnName = e.getKey();
            
            // Iterating thru each group of co-required catalysts
            for (And<String> and : e.getValue()) {
                
                if (and.contains(catalyst)) {
                    m.put(rxnName, getInternalRxn(rxnName));
                    break;
                }
                
            }
            
        }
        
        // If the provided catalyst was not encountered
        if (m.isEmpty()) {
            String s = "The catalyst '" + catalyst + 
                       "' does not exist in this fluxome.";
            throw new IllegalArgumentException(s);
        }

        return m;
    }

	/**
	 * Returns all chemical reactions contained in this fluxome.
	 * This is the union of all internal and exchange reactions.
	 * 
	 * @return the names of all reactions contained in this fluxome mapped
     * to their corresponding reaction.
	 * 
	 * @see #getInternalRxns()
	 * @see #getExchangeRxns() 
	 */
    public Map<String, ChemRxn> getRxns() {
    	
        Map<String, ChemRxn> m =
            new TreeMap<String, ChemRxn>();
        
        m.putAll(irrevRxns);
        m.putAll(revRxns);
        m.putAll(sourceRxns);
        m.putAll(sinkRxns);
        m.putAll(revExchRxns);

        return m;
    }

    /**
     * Returns the chemical reactions with the provided names.
     * 
     * @param rxnNames the names of the reactions to be returned.
     * 
     * @return the provided reaction names mapped to their corresponding reaction.
     * 
     * @throws IllegalArgumentException if any of the provided names are invalid.
     */
    public Map<String, ChemRxn> getRxns(Collection<String> rxnNames) {
        
        // Creating storage
        Map<String, ChemRxn> rxns = 
            new TreeMap<String, ChemRxn>();
        
        for (String rxnName : rxnNames) {
            rxns.put(rxnName, getRxn(rxnName));
        }

        return rxns;
    }
    
	/**
	 * Returns the chemical reaction in this fluxome with the provided name.
	 * 
	 * @param reactionName the name of the chemical reaction to be returned.
	 * 
	 * @return the chemical reaction in this fluxome with the provided name.
	 * 
	 * @see #getInternalRxn(String)
	 * @see #getExchangeRxn(String)
	 * 
	 * @throws IllegalArgumentException if the provided string is not
	 * a valid reaction name.
	 */
    public ChemRxn getRxn(String reactionName) {
        
        ChemRxn rxn = irrevRxns.get(reactionName);
        
        if (rxn != null) {
            return rxn;
        }
        
        rxn = revRxns.get(reactionName);
        
        if (rxn != null) {
            return rxn;
        }
        
        rxn = sourceRxns.get(reactionName);
        
        if (rxn != null) {
            return rxn;
        }
        
        rxn = sinkRxns.get(reactionName);
        
        if (rxn != null) {
            return rxn;
        }
        
        rxn = revExchRxns.get(reactionName);
        
        if (rxn != null) {
            return rxn;
        }
        
        // If the reaction name was not found
        String s = "A reaction named '" + reactionName + "' does not exist.";
        throw new IllegalArgumentException(s);
    }

	/**
	 * Returns the exchange reaction in this fluxome with the provided name.
	 * 
	 * @param reactionName the name of the exchange reaction to be returned.
	 * 
	 * @return the chemical reaction in this fluxome with the provided name.
	 * 
	 * @throws IllegalArgumentException if the provided string is not
	 * the name of an exchange reaction in this fluxome.
	 */
    public ExchangeRxn getExchangeRxn(String reactionName) {
        
        ExchangeRxn rxn = sourceRxns.get(reactionName);
        
        if (rxn != null) {
            return rxn;
        }
        
        rxn = sinkRxns.get(reactionName);
        
        if (rxn != null) {
            return rxn;
        }
        
        rxn = revExchRxns.get(reactionName);
        
        if (rxn != null) {
            return rxn;
        }

        // If the reaction name was not found
        String s = "A reaction named '" + reactionName + "' does not exist.";
        throw new IllegalArgumentException(s);
    }

	/**
	 * Returns the internal reaction in this fluxome with the provided name.
	 * 
	 * @param reactionName the name of the internal reaction to be returned.
	 * 
	 * @return the internal reaction in this fluxome with the provided name.
	 * 
	 * @throws IllegalArgumentException if the provided string is not
	 * the name of an internal reaction in this fluxome.
	 */
    public ChemRxn getInternalRxn(String reactionName) {
        
    	ChemRxn rxn = irrevRxns.get(reactionName);
        
        if (rxn != null) {
            return rxn;
        }
        
        rxn = revRxns.get(reactionName);
        
        if (rxn != null) {
            return rxn;
        }

        // If the reaction name was not found
        String s = "A reaction named '" + reactionName + "' does not exist.";
        throw new IllegalArgumentException(s);
    }
    
    /**
     * Returns the names of the reactions catalyzed by the provided catalyst.
     *
     * @param catalyst the name of a catalyst in this fluxome. 
     * 
     * @return a set of names of the reactions catalyzed by the provided 
     * catalyst.
	 * 
	 * @throws IllegalArgumentException if the provided string is not
	 * the name of a catalyst in this fluxome.
     */
    public Set<String> getCatalyzedRxnNames(String catalyst) {

        Set<String> names = new TreeSet<String>();
        
        // Iterating thru each reaction name -> catalyst pair
        for (Map.Entry<String, Or<And<String>>> e : catalysts.entrySet()) {
            
            // Iteratign thru each group of co-required catalysts
            for (And<String> and : e.getValue()) {
                
                if (and.contains(catalyst)) {
                    names.add(e.getKey());
                    break;
                }
                
            }
            
        }
        
        return names;
    }

	/**
	 * Returns the set of names of all chemical reactions contained in this
	 * fluxome. This set is the union of the names of internal and exchange
	 * reactions.
	 * 
	 * @return the set of names of all chemical reactions contained in this
	 *         fluxome.
	 * 
	 * @see #getInternalRxnNames()
	 * @see #getExchangeRxnNames()
	 */
    public Set<String> getRxnNames() {
 
    	// Creating storage
    	Set<String> s = new TreeSet<String>();

    	// Copying into storage
        s.addAll(irrevRxns.keySet());
        s.addAll(revRxns.keySet());
        s.addAll(sourceRxns.keySet());
        s.addAll(sinkRxns.keySet());
        s.addAll(revExchRxns.keySet());

        return s;
    }

    /**
     * Returns the set of names of all internal reactions in this fluxome.
     * @return the set of names of all internal reactions in this fluxome.
     */
    public Set<String> getInternalRxnNames() {
        
        Set<String> s = new TreeSet<String>();
        s.addAll(irrevRxns.keySet());
        s.addAll(revRxns.keySet());

        return s;
    }

    /**
     * Returns all of the internal reactions in this fluxome.
     * @return all of the internal reactions in this fluxome.
     */
    public Map<String, ChemRxn> getInternalRxns() {
        
        Map<String, ChemRxn> m = 
            new TreeMap<String, ChemRxn>();
        
        m.putAll(irrevRxns);
        m.putAll(revRxns);
        
        return m;
    }

    /**
     * Returns the set of names of all exchange reactions in this fluxome.
     * @return the set of names of all exchange reactions in this fluxome.
     */
    public Set<String> getExchangeRxnNames() {

    	// Creating storage
        Set<String> s = new TreeSet<String>();
        
        // Copying into storage
        s.addAll(sourceRxns.keySet());
        s.addAll(sinkRxns.keySet());
        s.addAll(revExchRxns.keySet());

        return s;
    }


    /**
	 * Returns the exchange reactions with the provided names.
	 * 
	 * @param rxnNames
	 *            the names of the reactions to be returned.
	 * @return reaction names mapped to their reaction.
	 * 
	 * @throws IllegalArgumentException if an invalid name is provided.
	 */
	public Map<String, ExchangeRxn> getExchangeRxns(Collection<String> rxnNames) {

        Map<String, ExchangeRxn> m = 
            new TreeMap<String, ExchangeRxn>();
        
        for (String name : rxnNames) {
        	ExchangeRxn rxn = getExchangeRxn(name);
        	m.put(name, rxn);
        }
        
        return m;
	}


    
	/**
	 * Returns the total number of chemical reactions contained
	 * in this fluxome.
	 * 
	 * @return the total number of chemical reactions contained
	 * in this fluxome.
	 */
    public int rxns() {
        return internalRxns() + exchangeRxns();
    }

	/**
	 * Returns the total number of internal reactions contained
	 * in this fluxome.
	 * 
	 * @return the total number of internal reactions contained
	 * in this fluxome.
	 */
    public int internalRxns() {
        return irrevRxns.size() + revRxns.size();
    }

	/**
	 * Returns the total number of exchange reactions contained 
	 * in this fluxome.
	 * 
	 * @return the total number of exchange reactions contained 
	 * in this fluxome.
	 */
    public int exchangeRxns() {
        return sourceRxns.size() + sinkRxns.size() + revExchRxns.size();
    }

	/**
	 * Returns the total number of constraints added to this fluxome.
	 * @return the total number of constraints added to this fluxome.
	 * 
	 * @see #replaceConstraints(FluxConstraints, double)
	 */
    public int addedConstraints() {
        return addedConstraints;
    }

	/**
	 * Returns the total number of catalysts contained in this fluxome.
	 * @return the total number of catalysts contained in this fluxome.
	 */
    public int catalysts() {
        return getCatalysts().size();
    }

	/**
	 * Returns the total number of chemical species contained in this fluxome.
	 * @return the total number of chemical species contained in this fluxome.
	 */
    public int species() {
        return getSpecies().size();
    }

	/**
	 * Returns the set of names of all catalysts contained in this fluxome. 
	 * 
	 * @return the set of names of all catalysts contained in this fluxome. 
	 */
    public Set<String> getCatalysts() {
        
        Set<String> cats = new TreeSet<String>();
        
        for (Or<And<String>> or : catalysts.values()) {
            
            for (And<String> and : or) {
                cats.addAll(and.getConditions());
            }
            
        }
        
        return cats;
    }

	/**
	 * Returns the set of names of all chemical species participating
	 * in reactions contained in this fluxome.
	 * 
	 * @return the set of names of all chemical species participating
	 * in reactions contained in this fluxome.
	 */
    public Set<String> getSpecies() {
        
    	// Creating storage
    	Set<String> s = new TreeSet<String>();
        
        s.addAll(getSpecies(irrevRxns.values()));
        s.addAll(getSpecies(revRxns.values()));
        s.addAll(getSpecies(sourceRxns.values()));
        s.addAll(getSpecies(sinkRxns.values()));
        s.addAll(getSpecies(revExchRxns.values()));
        
        return s;
    }

    private Set<String> getSpecies(Collection<? extends ChemRxn> rxns) {
        
        Set<String> s = new TreeSet<String>();
        
        for (ChemRxn r : rxns) {
            s.addAll(r.getReactants().getVariables());
            s.addAll(r.getProducts().getVariables());
        }
        
        return s;
    }
    
	/**
     * Indicates if the provided string is the name of a chemical species
     * contained in this fluxome.
     * 
     * @param s the string to test.
     * @return {@code true} if the string is the name of a chemical species
     * contained in this fluxome; {@code false} otherwise.
     */
    public boolean isSpecies(String s) {

        if (isSpecies(s, irrevRxns.values())) {
            return true;
        }
        
        if (isSpecies(s, revRxns.values())) {
            return true;
        }

        if (isSpecies(s, sourceRxns.values())) {
            return true;
        }
        
        if (isSpecies(s, sinkRxns.values())) {
            return true;
        }

        if (isSpecies(s, revExchRxns.values())) {
            return true;
        }
        
        return false;
    }

    private boolean isSpecies(String s, Collection<? extends ChemRxn> rxns) {
        

        for (ChemRxn rxn : rxns) {
            
            if (rxn.getReactants().hasVariable(s)) {
                return true;
            }
            
            if (rxn.getProducts().hasVariable(s)) {
                return true;
            }
        }
        
        return false;
    }
    
	/**
     * Indicates if the provided string is the name of an catalyst contained in
     * this fluxome.
     * 
     * @param s the string to test.
     * @return {@code true} if the string is the name of a catalyst contained in
     * this fluxome; {@code false} otherwise.
     */
    public boolean isCatalyst(String s) {

        for (Or<And<String>> or : catalysts.values()) {
            
            for (And<String> and : or) {
                
                if (and.contains(s)) {
                    return true;
                }
            }
            
        }

        return false;
    }

	/**
     * Indicates if the provided string is the name of a reaction contained in
     * this fluxome.
     * 
     * @param s the string to test.
     * @return {@code true} if the string is the name of a reaction contained in
     * this fluxome; {@code false} otherwise.
     */
    public boolean isRxnName(String s) {

        if (irrevRxns.containsKey(s)) {
            return true;
        }
        
        if (revRxns.containsKey(s)) {
            return true;
        }

        if (sourceRxns.containsKey(s)) {
            return true;
        }
        
        if (sinkRxns.containsKey(s)) {
            return true;
        }

        if (revExchRxns.containsKey(s)) {
            return true;
        }
        
        return false;
    }
    
	/**
	 * Indicates if the provided string is the name of an internal reaction
	 * contained in this fluxome.
	 * 
	 * @param s the string to test.
	 * 
	 * @return {@code true} if the string is a valid internal reaction name; 
	 * {@code false} otherwise.
	 */
    public boolean isInternalRxnName(String s) {
    	return irrevRxns.containsKey(s) || revRxns.containsKey(s);
    }

	/**
	 * Indicates if the provided string is the name of an exchange reaction
	 * contained in this fluxome.
	 * 
	 * @param s the string to test.
	 * 
	 * @return {@code true} if the string is a valid exchange reaction name; 
	 * {@code false} otherwise.
	 */
    public boolean isExchangeRxnName(String s) {
    	return sourceRxns.containsKey(s) ||
    			sinkRxns.containsKey(s) ||
    			revExchRxns.containsKey(s);
    }
    
	/**
	 * Indicates if every string in the provided collection is the name
	 * of a reaction contained in this fluxome. 
	 * 
	 * @param c a collection of strings.
	 * 
	 * @return {@code true} if every string is a valid reaction name; 
	 * {@code false} otherwise.
	 */
    public boolean areRxnNames(Collection<String> c) {
    	
    	// Iterating thru each string
    	for (String s : c) {
    		
    		// Checking the name
    		if (!isRxnName(s)) {
    			return false;
    		}
    	}
    	
        return true;
    }

	/**
	 * Indicates if the provided interval is a valid
	 * flux constraint for the reaction having the provided
	 * name. The lower flux bounds for irreversible reactions,
	 * including source and sink reactions, must be {@code >= 0}. The remaining
	 * flux bounds can be any values except {@code NaN}, lower bounds cannot be
	 * {@code Double.POSITIVE_INFINITY}, and upper bounds cannot be
	 * {@code Double.NEGATIVE_INFINITY}.
	 * 
	 * @param c the flux constraint to test.
	 * @param rxnName the name of the reaction.
	 * @param tolerance the allowed amount by which constraints can violate their
	 *            maximum defined intervals.
	 * 
	 * @return {@code true} if the provided constraint lies within it 
	 * maximum defined limits; {@code false} otherwise.
	 * 
	 * @throws IllegalArgumentException if the provided reaction
	 * name does not exist in this fluxome; or if the provided
	 * tolerance is < 0, NaN, or infinite.
	 * 
	 * @throws NullPointerException if the provided constraint 
	 * is {@code null}
	 */
    public boolean isValid(Interval c, String rxnName, double tolerance) {
    	
    	// Checking reaction name
    	checkRxnName(rxnName);
    	
    	// Checking tolerance
        checkTolerance(tolerance);

        // If the reaction is reversible
        if (revRxns.containsKey(rxnName) || 
        	revExchRxns.containsKey(rxnName)) {

        	return c.getLower() != Double.POSITIVE_INFINITY &&
        			c.getUpper() != Double.NEGATIVE_INFINITY;
        }

        // If the reaction is irreversible
        if (c.getLower() == Double.POSITIVE_INFINITY) {
        	return false;
        }
        
        // Ensuring the lower bound is >= "zero"
        return c.getLower() >= -1 * tolerance;
    }

    private void checkTolerance(double tolerance) {

    	// Ensuring tolerance is greater than zero and not
    	// infinite or NaN
        if (tolerance < 0 ||
            Double.isInfinite(tolerance) ||
            Double.isNaN(tolerance)) {

            String s = "Invalid tolerance: " + tolerance;
            throw new IllegalArgumentException(s);
        }
    }
    
	/**
     * Returns the stoichiometry matrix of this fluxome.
     * 
     * @return the stoichiometry matrix of this fluxome. Row indices in this
     * matrix are the names of chemical species, and column indices are reaction
     * names.
     */
    public SparseDoubleMatrix<String, String> getStoichMatrix() {

    	// Determining appropriate sizes
    	int rows = species() + 1;
    	int cols = rxns() + 1;
    	
    	// Creating storage
        SmdMatrix<String, String> s = 
        	new HashMatrix<String, String>(rows, cols, 1.0F);

        // Iterating thru each internal reaction
        addInternalRxns(s, irrevRxns);
        addInternalRxns(s, revRxns);

        // Iterating thru each exchange reaction
        addExchangeRxns(s, sourceRxns, 1d);
        addExchangeRxns(s, sinkRxns, -1d);
        addExchangeRxns(s, revExchRxns, -1d);

        // Returning the stoichiometry matrix
        return new SparseDoubleMatrixWrap<String, String>(s);
    }

    private void addInternalRxns(SmdMatrix<String, String> s, 
    							 Map<String, ? extends ChemRxn> rxns) {
    	
        // Iterating thru each internal reaction
        for (Entry<String, ? extends ChemRxn> e : rxns.entrySet()) {

            String rxnName = e.getKey();
            ChemRxn rxn = e.getValue();
            
        	// Iterating thru the reactants expression
        	for (LinearComb.Term<String> term : rxn.getReactants()) {

        		// Setting the matrix element
                s.setElement(term.getVariable(),           // row 
                			 rxnName,                      // column
                			 -1 * term.getCoefficient());  // element
            }

        	// Iterating thru the products expression
        	for (LinearComb.Term<String> term : rxn.getProducts()) {

        		// Setting the matrix element
                s.setElement(term.getVariable(),       // row
                			 rxnName,                  // column
                             term.getCoefficient());   // element
            }

        }
    	
    }
    
    
    private void addExchangeRxns(SmdMatrix<String, String> s,
    							 Map<String, ? extends ExchangeRxn> rxns,
    							 double stoichCoeff) {
    	
        // Iterating thru each exchange reaction
        for (Map.Entry<String, ? extends ExchangeRxn> e : rxns.entrySet()) {

            ExchangeRxn rxn = e.getValue();
            
        	// Setting the matrix element
        	s.setElement(rxn.getSpecies(), e.getKey(), stoichCoeff);
        }
    	
    }
    
    
    
    
	/**
	 * Indicates if the provided flux vector lies within the feasible space
	 * of this fluxome, given the provided tolerance value.
	 * 
	 * @param fluxValues reaction names mapped to flux values.
	 * @param tolerance the numerical tolerance.
	 * @return {@code true} if the vector satisfies all constraints; 
	 * {@code false} otherwise.
	 */
    public boolean 
    isFluxVector(Map<String, Double> fluxValues, double tolerance) {

    	// Checking input
        checkTolerance(tolerance);

        // Comparing size
        if (fluxValues.size() != rxns()) {
        	return false;
        }
        
        // If all reaction names are not present in the provided flux vector
        if (!fluxValues.keySet().containsAll(getRxnNames())) {
        	return false;
        }
        
        /* Checking flux constraints, since it's relatively fast */
        
        // Iterating thru each reaction name-flux value pair
        for (Map.Entry<String, Double> e : fluxValues.entrySet()) {
        	
        	// Getting the pair
        	String rxnName = e.getKey();
        	Double value = e.getValue();
        	
        	// Getting flux constraint
        	Interval c = getFluxConstraint(rxnName);
        	
        	// If flux does not lie within constrained range
        	if (!c.isWithin(value, tolerance)) {
        		return false;
        	}
        	
        }
        
        /* Second, checking flux expression constraints */
        
        // Iterating thru each expression-constraint pair
        for (Map.Entry<MathExpr<String>, Interval> e :
        	 constraints.getExprConstraints().entrySet()) {
        	
        	// Getting pair
        	MathExpr<String> expr = e.getKey();
        	Interval c = e.getValue();

        	// Ensuring the expression is linear
        	if (!(expr instanceof LinearComb)) {
        		String s = "Only linear constraints are understood"
        					+ " at this point.";
        		throw new UnsupportedOperationException(s);
        	}
        	
        	// Safely casting
        	LinearComb<String> lin = (LinearComb<String>) expr;
        	
        	// Plugging in provided flux values
        	double constant = lin.plugInNumbers(fluxValues).getConstant();
        	
        	// If constant does not lie within contrained range
        	if (!c.isWithin(constant, tolerance)) {
        		return false;
        	}
        }
        
        /* Finally, checking steady-state constraints */

        // Creating storage for column flux vector
        SmdMatrix<String, Integer> fluxColumn = new RowOrTreeMatrix<String, Integer>();
        	
        // Creating dummy variable for the column index
        Integer columnIndex = Integer.valueOf(0);
        
        // Copying flux values into the column vector
        for (Map.Entry<String, Double> e : fluxValues.entrySet()) {
            fluxColumn.setElement(e.getKey(), columnIndex, e.getValue());
        }

        MatrixMultiplier multiplier = new MatrixMultiplier();
        
        SparseDoubleMatrix<String, String> s = getStoichMatrix();
        
        SmdMatrix<String, Integer> solutionVector = 
        	new RowOrTreeMatrix<String, Integer>();
        
        // Multiplying the stoichiometry matrix by the column flux vector
        multiplier.multiply(s, fluxColumn, solutionVector);

        // Getting the row indices of any non-zero elements
        Set<String> rowIndices =
                solutionVector.getRowIndNzEl(columnIndex);

        // Iterating thru these indices
        for (String chemSpecies : rowIndices) {

        	// Getting the non-zero value
            double value = solutionVector.getElement(chemSpecies,
            										 columnIndex);

            // If steady-state is violated (i.e. if the computed 
            // value is not sufficiently close to zero)
            if (Math.abs(value) > tolerance) {
                return false;
            }
        }

        return true;
    }

    
    /**
	 * Returns a mixed fluxome composed of the reactions in the provided fluxome
	 * with the names in the provided collection. This method can be used to
	 * copy an entire fluxome or to create a sub-fluxome. The constraints in the
	 * returned fluxome are set to their default values, that is, they are not
	 * copied from the provided fluxome.
	 * 
	 * @param reactionNames
	 *            the names of reactions in the provided fluxome used to compose
	 *            the returned fluxome.
	 * 
	 * @param fluxome
	 *            the fluxome used to compose the returned fluxome.
	 * 
	 * @return a mixed fluxome composed of the reactions in the provided fluxome
	 *         with the names in the provided collection.
	 */
	public static MixedFluxomeV1 
	make(Collection<String> reactionNames, MixedFluxome fluxome) {

		// Checking provided reaction names
		fluxome.checkRxnNames(reactionNames);
        
        // Getting all reactions
        Map<String, ChemRxn> rxns = fluxome.getRxns(reactionNames);
		
        // Getting catalysts
        Map<String, Or<And<String>>> catalysts = 
            new TreeMap<String, Or<And<String>>>(fluxome.getCatalystRules());
        
        // Retaining only the relevant catalysts
        catalysts.keySet().retainAll(rxns.keySet());
        
        // Constraints are not maintained  *********************
        
		// Returning fluxome created from the provided reaction names
		return new MixedFluxomeV1(rxns, catalysts);
	}

    /**
     * Returns all of the boolean logic statements regarding reaction catalysis.
     * @return the names of catalyzed reactions mapped to their boolean logic
     * rules of their catalyst(s).
     */
	public Map<String, Or<And<String>>> getCatalystRules() {
     
        Map<String, Or<And<String>>> m =
            new TreeMap<String, Or<And<String>>>();
        
        m.putAll(catalysts);
        
        return m;
    }
    
    /**
     * Returns the boolean logic statement for the catalysts of the reaction
     * with the provided name.
     * 
     * @param rxnName the name of the reaction for which the catalyst rules
     * are to be returned.
     * 
     * @return the boolean logic statement for the catalysts of the reaction
     * with the provided name. If the specified reaction is not catalyzed,
     * an empty (non-null) rule is returned.
     */
    public Or<And<String>> getCatalystRule(String rxnName) {

        checkRxnName(rxnName);
        
        Or<And<String>> cats = catalysts.get(rxnName);
        
        if (cats == null) {
            return emptyRule;
        }
        
        return cats;
    }

    /**
     * Returns the internal reactions whose names appear in the provided
     * collection of reaction names. Every string in the provided collection
     * must be the name of a reaction in this fluxome, but the names of exchange
     * reactions are ignored.
     * 
     * @param rxnNames a collection of reaction names.
     * 
     * @return the names of internal reactions contained in {@code rxnNames}
     * mapped to their corresponding reaction.
     * 
     * @throws IllegalArgumentException if a string in {@code rxnNames} is not
     * the name of a reaction in this fluxome.
     */
    public Map<String, ChemRxn> getInternalRxns(Collection<String> rxnNames) {
        
        Map<String, ChemRxn> m = new TreeMap<String, ChemRxn>();
        
        for (String rxnName : rxnNames) {
        	ChemRxn rxn = getInternalRxn(rxnName);
        	m.put(rxnName, rxn);
        }

        return m;
    }

    /**
     * Indicates if the specified catalyst is required for the reaction with 
     * the specified name.
     *  
     * @param catalyst the catalyst.
     * @param rxnName the name of the reaction.
     * 
     * @return {@code true} if the provided catalyst is required by the reaction
     * with the provided name; {@code false} otherwise.
     * 
     * @throws IllegalArgumentException if {@code rxnName} is not a valid reaction name.
     */
    public boolean isReqCatalyst(String catalyst, String rxnName) {
        
        Or<And<String>> rule = catalysts.get(rxnName);
        
        if (rule == null) {
            checkRxnName(rxnName);
            return false;
        }
 
        for (And<String> and : rule) {
            
            if (!and.contains(catalyst)) {
                return false;
            }
            
        }
        
        return true;
    }

    /**
     * Indicates if the reaction with the provided name could function if
     * each of the catalysts in the provided collection were simultaneously
     * absent.
	 * 
	 * @param catalysts the collection of catalysts.
	 * @param rxnName the name of the reaction.
	 * 
	 * @return {@code true} if the provided catalysts are required for the 
	 * specified reaction to function; {@code false} otherwise.
	 * 
	 * @throws IllegalArgumentException
	 *             if {@code rxnName} is not a valid reaction name.
	 */
    public boolean areReqCatalysts(Collection<String> catalysts, String rxnName) {
        
        // Ensuring the collection uses a standard equality test
        Set<String> cats = new TreeSet<String>(catalysts);
        
        // Getting the rule for this reaction if one exists
        Or<And<String>> rule = this.catalysts.get(rxnName);
        
        if (rule == null) {
            checkRxnName(rxnName);
            return false;
        }
 
        // Iterating thru each group of co-required catalysts
        for (And<String> and : rule) {
            
            // If this group contains no elements in common with the
            // provided collection of catalysts.
            if (Collections.disjoint(and.getConditions(), cats)) {
                return false;
            }
            
        }
        
        return true;
    }
    
    
}
