/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import java.util.Collections;
import java.util.Set;

import sbrt.kernel.math.*;

/**
 * This class is used to represent reversible exchange reactions, that is, exchange
 * reactions that can both supply and consume chemical species to and from a 
 * chemical system.
 * <p>
 * Instances of this class are immutable.
 * </p>
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 */
public final class RevExchRxn extends AbstractExchangeRxn {

	private static final LinearComb<String> EMPTY;
	
	static {
		Set<LinearComb.Term<String>> s = Collections.emptySet();
		EMPTY = new HashLinearComb<String>(s, 0d, 0d);
	}
	
	private final LinearComb<String> reactants;
	private final LinearComb<String> products;
	
    /**
     * Constructs a new reversible exchange reaction for the 
     * provided chemical species.
     * 
     * @param chemSpecies the chemical species associated with this
     * exchange reaction.
     * 
     * @param reactant indicates if the chemical species should be considered
     * as a reactant or product, which affects the way in which an
     * exchange reaction is written and the sign of its flux. 
     */
    public RevExchRxn(String chemSpecies, boolean reactant) {
    	
    	super(chemSpecies);
    	
    	if (reactant) {
    		reactants = HashLinearComb.singleTerm(chemSpecies);
    		products = EMPTY;
    	}
    	
    	else {
    		reactants = EMPTY;
    		products = HashLinearComb.singleTerm(chemSpecies);
    	}
    	
    }
    
    /**
     * Returns a linear combination representing the reactants of this
     * exchange reaction. This linear combination will be either empty or have
     * a single-term, depending on whether or not the chemical species is 
     * considered a reactant or product. If a single-term is present, 
     * the variable of that term is the chemical species and the coefficient is 1.0.
     * 
     * @return a linear combination representing the reactants of this
     * exchange reaction.
     */
    public LinearComb<String> getReactants() {
        return reactants;
    }
    
    /**
     * Returns a linear combination representing the products of this
     * exchange reaction. This linear combination will be either empty or have
     * a single-term, depending on whether or not the chemical species is 
     * considered a reactant or product. If a single-term is present, 
     * the variable of that term is the chemical species and the coefficient is 1.0.
     * 
     * @return a linear combination representing the products of this
     * exchange reaction.
     */
    public LinearComb<String> getProducts() {
        return products;
    }

	/**
	 * Returns {@code true}.
	 * @return {@code true}.
	 */
	public boolean isReversible() {
		return true;
	}

    /**
     * Returns a string representation of this exchange reaction.
     * @return a string representation of this exchange reaction.
     */
    public String toString() {
    	
    	if (reactants == EMPTY) {
    		return "Surroundings <==> " + products; 
    	}
    	
        return reactants + " <==> Surroundings";
    }


}
