/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Map.Entry;

import sbrt.kernel.math.LinearComb;

/**
 * This class is used to compare chemical reactions. Read the documentation
 * for each of these methods carefully, because similarly named methods can perform 
 * very different operations.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public class RxnComparer {


	/**
	 * Returns the ratio of the coefficients in the provided expressions.
	 * 
	 * @param expr1
	 *            a stoichiometric expression.
	 * @param expr2
	 *            a stoichiometric expression.
	 * 
	 * @return the coefficients of {@code expr1} divided by the coefficients of
	 *         {@code expr2}; or {@code Double.NaN} if the coefficients of the
	 *         terms in the provided expressions are not factors of each other.
	 */
	private static double getRatio(LinearComb<?> expr1, LinearComb<?> expr2) {

		// Comparing sets of chemical species
		Set<?> species1 = expr1.getVariables();
		Set<?> species2 = expr2.getVariables();

		if (!species1.equals(species2)) {
			return Double.NaN;
		}

		// Initializing to an out-of-band value
		double ratio = Double.NaN;

		// Iterating thru each chemical species
		for (Object s : species1) {

			// Getting stoichiometric coefficients from both expressions
			double x1 = expr1.getCoefficient(s);
			double x2 = expr2.getCoefficient(s);
			
			// If this is the first pair
			if (Double.isNaN(ratio)) {
				ratio = x1 / x2;
			}

			// If the ratio has already been defined, and it does not equal
			// the ratio of these coefficients
			else if (Math.abs(x1 / x2 - ratio) > 1E-9) {
				return Double.NaN;
			}

		}

		return ratio;
	}

	private static boolean stoichEquiv(LinearComb<?> rxn1Reactants, 
									   LinearComb<?> rxn1Products,
									   LinearComb<?> rxn2Reactants,
									   LinearComb<?> rxn2Products) {

		double ratio1 = getRatio(rxn1Reactants, rxn2Reactants);
		double ratio2 = getRatio(rxn1Products, rxn2Products);

		if (Double.isNaN(ratio1) || Double.isNaN(ratio2)) {
			return false;
		}

		if (ratio1 != ratio2) {
			return false;
		}

		return true;
	}
	
	
	

    /**
	 * Indicates if the provided chemical reactions are stoichiometrically
	 * equivalent. For two reactions to be considered equivlent by this method,
	 * their respective reactant and product species must be identical, and
	 * their stoichiometric coefficients can only differ by a constant factor.
	 * For example, {@code A <==> B} and {@code (2) A --> (2) B} are considered
	 * stoichiometrically equivalent by this method, but {@code A <==> B} and
	 * {@code B <==> A} are not.
	 * 
	 * @param rxn1
	 *            a chemical reaction.
	 * @param rxn2
	 *            a chemical reaction.
	 * 
	 * @return {@code true} if the reactions are equivalent; {@code false}
	 *         otherwise.
	 */
	public static boolean equivChemRxns(ChemRxn rxn1, ChemRxn rxn2) {
		
		return stoichEquiv(rxn1.getReactants(), rxn1.getProducts(),
						   rxn2.getReactants(), rxn2.getProducts());
	}

    /**
	 * Indicates if the provided irreversible reactions are stoichiometrically
	 * equivalent. For two reactions to be considered equivlent by this method,
	 * their respective reactant and product species must be identical, and
	 * their stoichiometric coefficients can only differ by a constant factor.
	 * For example, {@code A --> B} and {@code (2) A --> (2) B} are considered
	 * stoichiometrically equivalent by this method.
	 * 
	 * @param rxn1
	 *            an irreversible reaction.
	 * @param rxn2
	 *            an irreversible reaction.
	 * 
	 * @return {@code true} if the irreversible reactions are equivalent;
	 *         {@code false} otherwise.
	 */
	public static boolean equivIrrevRxns(IrrevRxn rxn1, IrrevRxn rxn2) {
		return equivChemRxns(rxn1, rxn2);
	}

    /**
     * Indicates if the provided source reactions are stoichiometrically
     * equivalent. For two reactions to be considered equivlent by this method, 
     * their associated chemical species must be identical.
     * 
     * @param rxn1 a source reaction.
     * @param rxn2 a source reaction.
     * 
     * @return {@code true} if the source reactions are equivalent; {@code false}
     * otherwise.
     */
	public static boolean equivSourceRxns(SourceRxn rxn1, SourceRxn rxn2) {
		return rxn1.getSpecies().equals(rxn2.getSpecies());
	}

    /**
     * Indicates if the provided sink reactions are stoichiometrically
     * equivalent. For two reactions to be considered equivlent by this method, 
     * their associated chemical species must be identical.
     * 
     * @param rxn1 a sink reaction.
     * @param rxn2 a sink reaction.
     * 
     * @return {@code true} if the sink reactions are equivalent; {@code false}
     * otherwise.
     */
	public static boolean equivSinkRxns(SinkRxn rxn1, SinkRxn rxn2) {
		return rxn1.getSpecies().equals(rxn2.getSpecies());
	}

    /**
	 * Indicates if the provided reversible exchange reactions are
	 * stoichiometrically equivalent. For two reactions to be considered
	 * equivlent by this method, their associated chemical species must be
	 * identical.
	 * 
	 * @param rxn1
	 *            a reversible exchange reaction.
	 * @param rxn2
	 *            a reversible exchange reaction.
	 * 
	 * @return {@code true} if the reversible exchange reactions are equivalent;
	 *         {@code false} otherwise.
	 */
	public static boolean equivRevExchRxns(RevExchRxn rxn1, RevExchRxn rxn2) {
		return rxn1.getSpecies().equals(rxn2.getSpecies());
	}

    /**
	 * Indicates if the provided reversible reactions are
	 * stoichiometrically equivalent. For two reactions to be considered
	 * equivlent by this method, the reactants of {@code rxn1} must be
	 * equivalent to either the reactants or products of {@code rxn2}. 
	 * The products of {@code rxn1} must also be equivalent to either the
	 * reactants or products of {@code rxn2}, but not equivlent to the
	 * reactants of {@code rxn1}. For example, {@code (2) A <==> (2) B} and 
	 * {@code B <==> A} are considered stoichiometrically equivalent by this 
	 * method.
	 * 
	 * @param rxn1 a reversible exchange reaction.
	 * @param rxn2 a reversible exchange reaction.
	 * 
	 * @return {@code true} if the reversible exchange reactions are equivalent;
	 *         {@code false} otherwise.
	 */
	public static boolean equivRevRxns(RevRxn rxn1, RevRxn rxn2) {
		return stoichEquiv(rxn1.getReactants(), rxn1.getProducts(), 
				   		   rxn2.getReactants(), rxn2.getProducts()) || 
			   stoichEquiv(rxn1.getReactants(), rxn1.getProducts(), 
					   	   rxn2.getProducts(),  rxn2.getReactants());
	}
	
	
	
	
	
    /**
     * Indicates if the provided chemical reactions are the reverse 
     * of each other. For two reactions to be considered the reverse of each other
     * by this method, the reactants and products of {@code rxn1} must be 
     * stoichiometrically equivalent to the products and reactants, 
     * respectively, of {@code rxn2}. For example, {@code A <==> B} and 
     * {@code (2) B --> (2) A} are considered to be the reverse of each other
	 * by this method.
     * 
     * @param rxn1 a reaction.
     * @param rxn2 a reaction.
     * 
     * @return {@code true} if the provided reactions are the reverse of each 
     * other; {@code false} otherwise.
     */
	public static boolean reverseChemRxns(ChemRxn rxn1, ChemRxn rxn2) {
		
		// Checking for stoichiometric equivalence of swapped 
		// reactants and products
		return stoichEquiv(rxn1.getReactants(), rxn1.getProducts(),
						   rxn2.getProducts(), rxn2.getReactants());
	}

    /**
     * Indicates if the provided chemical reactions are the reverse 
     * of each other. For two reactions to be considered the reverse of each other
     * by this method, the reactants and products of {@code rxn1} must be 
     * stoichiometrically equivalent to the products and reactants, 
     * respectively, of {@code rxn2}. For example, {@code A --> B} and 
     * {@code (2) B --> (2) A} are considered to be the reverse of each other
	 * by this method.
     * 
     * @param rxn1 a reaction.
     * @param rxn2 a reaction.
     * 
     * @return {@code true} if the provided reactions are the reverse of each 
     * other; {@code false} otherwise.
     */
	public static boolean reverseIrrevRxns(IrrevRxn rxn1, IrrevRxn rxn2) {
		return stoichEquiv(rxn1.getReactants(), rxn1.getProducts(),
						   rxn2.getProducts(), rxn2.getReactants());
	}

	/**
	 * Indicates if the provided exchange reactions are the reverse of each other.
	 * 
	 * @param rxn1 a source reaction.
	 * @param rxn2 a sink reaction.
	 * 
	 * @return {@code true} if the exchange reactions are associated with the same
	 * chemical species; {@code false} otherwise.
	 */
	public static boolean reverseExchangeRxns(SourceRxn rxn1, SinkRxn rxn2) {
		return rxn1.getSpecies().equals(rxn2.getSpecies());
	}

	/**
	 * Indicates if the provided exchange reactions are the reverse of each other.
	 * 
	 * @param rxn1 a sink reaction.
	 * @param rxn2 a source reaction.
	 * 
	 * @return {@code true} if the exchange reactions are associated with the same
	 * chemical species; {@code false} otherwise.
	 */
	public static boolean reverseExchangeRxns(SinkRxn rxn1, SourceRxn rxn2) {
		return rxn1.getSpecies().equals(rxn2.getSpecies());
	}

    
    /**
     * Returns the reactions in the provided map that are 
     * stoichiometrically equivalent to the reaction in the map with
     * the provided map key. The reaction corresponding to the provided
     * key is not included in the returned map.
     * 
     * @param <T> a comparable type. 
     * 
     * @param rxnId the map key of the reaction whose 
     * stoichiometrically equivalent reactions will be returned.
     * 
     * @param rxns keys mapped to reactions.
     * 
     * @return the entries of the provided map that contain reactions that are
     * stoichiometrically equivalent to the specified reaction.
     * 
     * @see #equivIrrevRxns(IrrevRxn, IrrevRxn)
     */
    public static <T extends Comparable<? super T>> Map<T, IrrevRxn> 
    	getEquivIrrevRxns(T rxnId, Map<T, IrrevRxn> rxns) {
        
        // Creating storage
        Map<T, IrrevRxn> m = new TreeMap<T, IrrevRxn>();
        
        // Getting the corresponding reaction
        IrrevRxn rxn = rxns.get(rxnId);
        
        // Iterating thru each internal reaction
        for (Entry<T, IrrevRxn> e : rxns.entrySet()) {
            
            IrrevRxn r = e.getValue();
            
            if (equivIrrevRxns(r, rxn)) {
                m.put(e.getKey(), r);
            }
            
        }
        
        // Removing the specified reaction
        m.remove(rxnId);
        
        return m;
    }
    
    /**
     * Returns the reactions in the provided map that are 
     * stoichiometrically equivalent to the reaction in the map with
     * the provided map key. The reaction corresponding to the provided
     * key is not included in the returned map.
     * 
     * @param <T> a comparable type. 
     * 
     * @param rxnId the map key of the reaction whose 
     * stoichiometrically equivalent reactions will be returned.
     * 
     * @param rxns keys mapped to reactions.
     * 
     * @return the entries of the provided map that contain reactions that are
     * stoichiometrically equivalent to the specified reaction.
     * 
     * @see #equivRevRxns(RevRxn, RevRxn)
     */
    public static <T extends Comparable<? super T>> Map<T, RevRxn> 
    	getEquivRevRxns(T rxnId, Map<T, RevRxn> rxns) {
        
        // Creating storage
        Map<T, RevRxn> m = new TreeMap<T, RevRxn>();
        
        // Getting the corresponding reaction
        RevRxn rxn = rxns.get(rxnId);
        
        // Iterating thru each internal reaction
        for (Entry<T, RevRxn> e : rxns.entrySet()) {
            
            RevRxn r = e.getValue();
            
            if (equivRevRxns(r, rxn)) {
                m.put(e.getKey(), r);
            }
            
        }
        
        // Removing the specified reaction
        m.remove(rxnId);
        
        return m;
    }
    
    /**
	 * Returns the keys in the provided map that correspond to reactions that
	 * are stoichiometrically equivalent to the reaction in the map with the
	 * specified map key. The provided key is not included in the returned set.
	 * 
	 * @param <T>
	 *            a comparable type.
	 * 
	 * @param rxnId
	 *            the map key of the reaction whose stoichiometrically
	 *            equivalent reaction keys will be returned.
	 * 
	 * @param rxns
	 *            keys mapped to reactions.
	 * 
	 * @return the keys of the provided map that correspond to reactions that are
	 *         stoichiometrically equivalent to the specified reaction.
	 * 
	 * @see #equivIrrevRxns(IrrevRxn, IrrevRxn)
	 * @see #getEquivIrrevRxns(Comparable, Map)
	 */
    public static <T extends Comparable<? super T>> Set<T> 
    	getEquivIrrevRxnIds(T rxnId, Map<T, IrrevRxn> rxns) {

    	return getEquivIrrevRxns(rxnId, rxns).keySet();
    }
    
    /**
	 * Returns the keys in the provided map that correspond to reactions that
	 * are stoichiometrically equivalent to the reaction in the map with the
	 * specified map key. The provided key is not included in the returned set.
	 * 
	 * @param <T>
	 *            a comparable type.
	 * 
	 * @param rxnId
	 *            the map key of the reaction whose stoichiometrically
	 *            equivalent reaction keys will be returned.
	 * 
	 * @param rxns
	 *            keys mapped to reactions.
	 * 
	 * @return the keys of the provided map that correspond to reactions that are
	 *         stoichiometrically equivalent to the specified reaction.
	 * 
	 * @see #equivRevRxns(RevRxn, RevRxn)
	 * @see #getEquivRevRxns(Comparable, Map)
	 */
    public static <T extends Comparable<? super T>> Set<T> 
    	getEquivRevRxnIds(T rxnId, Map<T, RevRxn> rxns) {

    	return getEquivRevRxns(rxnId, rxns).keySet();
    }
    
    /**
     * Returns a collection of sets of the names of stoichiometrically 
     * equivalent reactions contained in the provided map.
     * 
     * @param rxns IDs mapped to their corresponding reactions. 
     * @param <T> a comparable type.
     * 
     * @return a collection of sets of the names of stoichiometrically 
     * equivalent reactions contained in the provided map.
     * 
     * @see #equivIrrevRxns(IrrevRxn, IrrevRxn)
     */
    public static <T extends Comparable<? super T>> Collection<Set<T>> 
    	getEquivIrrevRxnIds(Map<T, IrrevRxn> rxns) {
        
        // Creating storage for the sets of reaction IDs
        List<Set<T>> equivs = new LinkedList<Set<T>>();
        
        // Creating list of reactions
        List<Entry<T, IrrevRxn>> rxnEntries = 
            new ArrayList<Entry<T, IrrevRxn>>(rxns.entrySet());
        
        // Creating a primary iterator
        ListIterator<Entry<T, IrrevRxn>> i = rxnEntries.listIterator();
        
        // Iterating thru each reaction
        while (i.hasNext()) {

            // Getting the next ID-reaction pair
            Entry<T, IrrevRxn> e1 = i.next();
            
            T r1Name = e1.getKey();
            IrrevRxn r1 = e1.getValue();

            int index = i.nextIndex();
            
            // Getting an iterator over the remaining reaction entries
            ListIterator<Entry<T, IrrevRxn>> j = rxnEntries.listIterator(index);
            
            // Iterating thru the remaining reactions
            while (j.hasNext()) {
                
                Entry<T, IrrevRxn> e2 = j.next();
                
                T r2Name = e2.getKey();
                IrrevRxn r2 = e2.getValue();

                // If the reactions are not equivalent,
                // skip ahead to the next rxn
                if (!equivIrrevRxns(r2, r1)) {
                    continue;
                }

                // Recording the name of redundant reactions
                addEquivRxns(r2Name, r1Name, equivs);

                // Breaking out of the loop
                break;
            }

        }
        
        return equivs;
    }
    
    /**
     * Returns a collection of sets of the names of stoichiometrically 
     * equivalent reactions contained in the provided map.
     * 
     * @param rxns IDs mapped to their corresponding reactions. 
     * @param <T> a comparable type.
     * 
     * @return a collection of sets of the names of stoichiometrically 
     * equivalent reactions contained in the provided map.
     * 
     * @see #equivRevRxns(RevRxn, RevRxn)
     */
    public static <T extends Comparable<? super T>> Collection<Set<T>> 
    	getEquivRevRxnIds(Map<T, RevRxn> rxns) {
        
        // Creating storage for the sets of reaction IDs
        List<Set<T>> equivs = new LinkedList<Set<T>>();
        
        // Creating list of reactions
        List<Entry<T, RevRxn>> rxnEntries = 
            new ArrayList<Entry<T, RevRxn>>(rxns.entrySet());
        
        // Creating a primary iterator
        ListIterator<Entry<T, RevRxn>> i = rxnEntries.listIterator();
        
        // Iterating thru each reaction
        while (i.hasNext()) {

            // Getting the next ID-reaction pair
            Entry<T, RevRxn> e1 = i.next();
            
            T r1Name = e1.getKey();
            RevRxn r1 = e1.getValue();

            int index = i.nextIndex();
            
            // Getting an iterator over the remaining reaction entries
            ListIterator<Entry<T, RevRxn>> j = rxnEntries.listIterator(index);
            
            // Iterating thru the remaining reactions
            while (j.hasNext()) {
                
                Entry<T, RevRxn> e2 = j.next();
                
                T r2Name = e2.getKey();
                RevRxn r2 = e2.getValue();

                // If the reactions are not equivalent,
                // skip ahead to the next rxn
                if (!equivRevRxns(r2, r1)) {
                    continue;
                }

                // Recording the name of redundant reactions
                addEquivRxns(r2Name, r1Name, equivs);

                // Breaking out of the loop
                break;
            }

        }
        
        return equivs;
    }

    /**
     * Returns a collection of sets of the names of stoichiometrically 
     * equivalent reactions contained in the provided map.
     * 
     * @param rxns IDs mapped to their corresponding reactions. 
     * @param <T> a comparable type.
     * 
     * @return a collection of sets of the names of stoichiometrically 
     * equivalent reactions contained in the provided map.
     * 
     * @see #equivSourceRxns(SourceRxn, SourceRxn)
     */
    public static <T extends Comparable<? super T>> Collection<Set<T>> 
    	getEquivSourceRxnIds(Map<T, SourceRxn> rxns) {
        
        // Creating storage for the sets of reaction IDs
        List<Set<T>> equivs = new LinkedList<Set<T>>();
        
        // Creating list of reactions
        List<Entry<T, SourceRxn>> rxnEntries = 
            new ArrayList<Entry<T, SourceRxn>>(rxns.entrySet());
        
        // Creating a primary iterator
        ListIterator<Entry<T, SourceRxn>> i = rxnEntries.listIterator();
        
        // Iterating thru each reaction
        while (i.hasNext()) {

            // Getting the next ID-reaction pair
            Entry<T, SourceRxn> e1 = i.next();
            
            T r1Name = e1.getKey();
            SourceRxn r1 = e1.getValue();

            int index = i.nextIndex();
            
            // Getting an iterator over the remaining reaction entries
            ListIterator<Entry<T, SourceRxn>> j = rxnEntries.listIterator(index);
            
            // Iterating thru the remaining reactions
            while (j.hasNext()) {
                
                Entry<T, SourceRxn> e2 = j.next();
                
                T r2Name = e2.getKey();
                SourceRxn r2 = e2.getValue();

                // If the reactions are not equivalent,
                // skip ahead to the next rxn
                if (!equivSourceRxns(r2, r1)) {
                    continue;
                }

                // Recording the name of redundant reactions
                addEquivRxns(r2Name, r1Name, equivs);

                // Breaking out of the loop
                break;
            }

        }
        
        return equivs;
    }

    /**
     * Returns a collection of sets of the names of stoichiometrically 
     * equivalent reactions contained in the provided map.
     * 
     * @param rxns IDs mapped to their corresponding reactions. 
     * @param <T> a comparable type.
     * 
     * @return a collection of sets of the names of stoichiometrically 
     * equivalent reactions contained in the provided map.
     * 
     * @see #equivSinkRxns(SinkRxn, SinkRxn)
     */
    public static <T extends Comparable<? super T>> Collection<Set<T>> 
    	getEquivSinkRxnIds(Map<T, SinkRxn> rxns) {
        
        // Creating storage for the sets of reaction IDs
        List<Set<T>> equivs = new LinkedList<Set<T>>();
        
        // Creating list of reactions
        List<Entry<T, SinkRxn>> rxnEntries = 
            new ArrayList<Entry<T, SinkRxn>>(rxns.entrySet());
        
        // Creating a primary iterator
        ListIterator<Entry<T, SinkRxn>> i = rxnEntries.listIterator();
        
        // Iterating thru each reaction
        while (i.hasNext()) {

            // Getting the next ID-reaction pair
            Entry<T, SinkRxn> e1 = i.next();
            
            T r1Name = e1.getKey();
            SinkRxn r1 = e1.getValue();

            int index = i.nextIndex();
            
            // Getting an iterator over the remaining reaction entries
            ListIterator<Entry<T, SinkRxn>> j = rxnEntries.listIterator(index);
            
            // Iterating thru the remaining reactions
            while (j.hasNext()) {
                
                Entry<T, SinkRxn> e2 = j.next();
                
                T r2Name = e2.getKey();
                SinkRxn r2 = e2.getValue();

                // If the reactions are not equivalent,
                // skip ahead to the next rxn
                if (!equivSinkRxns(r2, r1)) {
                    continue;
                }

                // Recording the name of redundant reactions
                addEquivRxns(r2Name, r1Name, equivs);

                // Breaking out of the loop
                break;
            }

        }
        
        return equivs;
    }


    /**
     * Returns a collection of sets of the names of stoichiometrically 
     * equivalent reactions contained in the provided map.
     * 
     * @param rxns IDs mapped to their corresponding reactions. 
     * @param <T> a comparable type.
     * 
     * @return a collection of sets of the names of stoichiometrically 
     * equivalent reactions contained in the provided map.
     * 
     * @see #equivRevExchRxns(RevExchRxn, RevExchRxn)
     */
    public static <T extends Comparable<? super T>> Collection<Set<T>> 
    	getEquivRevExchRxnIds(Map<T, RevExchRxn> rxns) {
        
        // Creating storage for the sets of reaction IDs
        List<Set<T>> equivs = new LinkedList<Set<T>>();
        
        // Creating list of reactions
        List<Entry<T, RevExchRxn>> rxnEntries = 
            new ArrayList<Entry<T, RevExchRxn>>(rxns.entrySet());
        
        // Creating a primary iterator
        ListIterator<Entry<T, RevExchRxn>> i = rxnEntries.listIterator();
        
        // Iterating thru each reaction
        while (i.hasNext()) {

            // Getting the next ID-reaction pair
            Entry<T, RevExchRxn> e1 = i.next();
            
            T r1Name = e1.getKey();
            RevExchRxn r1 = e1.getValue();

            int index = i.nextIndex();
            
            // Getting an iterator over the remaining reaction entries
            ListIterator<Entry<T, RevExchRxn>> j = rxnEntries.listIterator(index);
            
            // Iterating thru the remaining reactions
            while (j.hasNext()) {
                
                Entry<T, RevExchRxn> e2 = j.next();
                
                T r2Name = e2.getKey();
                RevExchRxn r2 = e2.getValue();

                // If the reactions are not equivalent,
                // skip ahead to the next rxn
                if (!equivRevExchRxns(r2, r1)) {
                    continue;
                }

                // Recording the name of redundant reactions
                addEquivRxns(r2Name, r1Name, equivs);

                // Breaking out of the loop
                break;
            }

        }
        
        return equivs;
    }
    
    private static <T extends Comparable<? super T>> void 
    	addEquivRxns(T rxnName1, T rxnName2, Collection<Set<T>> equivalentRxns) {
        
        // Iterating thru each set of redundant reactions
        for (Set<T> rxns : equivalentRxns) {

            // If this set contains either reaction name
            if (rxns.contains(rxnName1) || rxns.contains(rxnName2)) {

                // Adding both rxns
                rxns.add(rxnName1);
                rxns.add(rxnName2);

                // Finished
                return;
            }
        }

        // Creating a new set of redundant reaction names
        Set<T> redundants = new TreeSet<T>();
        
        // Adding both reaction names to the set
        redundants.add(rxnName1);
        redundants.add(rxnName2);

        // Adding this new set to the rest
        equivalentRxns.add(redundants);
        
    }
	

    /**
     * Returns all of the keys in the provided map that correspond to reactions
     * that are the reverse of the reaction specified by the provided key.
     * 
     * @param <T> a comparable type. 
     * 
     * @param rxnId a key in the provided map that corresponds to the
     * reaction for which reverse reactions will be returned.
     * 
     * @param rxns keys mapped to irreversible reactions.
     * 
     * @return the entries of the provided map that contain the 
     * identified reverse reactions.
     * 
     * @see #reverseIrrevRxns(IrrevRxn, IrrevRxn)
     * @see #getReverseIrrevRxns(Comparable, Map)
     */
    public static <T extends Comparable<? super T>> Set<T> 
    getReverseIrrevRxnIds(T rxnId, Map<T, IrrevRxn> rxns) {
        return getReverseIrrevRxns(rxnId, rxns).keySet();
    }

    /**
     * Returns all of the keys in the provided map that correspond to the 
     * source reactions that are the reverse of the provided sink reaction.
     * 
     * @param <T> a comparable type. 
     * 
     * @param sink a sink reaction.
     * 
     * @param rxns keys mapped to source reactions.
     * 
     * @return the keys of the provided map that correspond to the
     * identified source reactions.
     * 
     * @see #reverseExchangeRxns(SinkRxn, SourceRxn)
     * @see #getReverseSourceRxns(SinkRxn, Map)
     */
    public static <T extends Comparable<? super T>> Set<T> 
    getReverseSourceRxnIds(SinkRxn sink, Map<T, SourceRxn> rxns) {
    	return getReverseSourceRxns(sink, rxns).keySet();
    }

    /**
     * Returns all of the keys in the provided map that correspond to the 
     * sink reactions that are the reverse of the provided source reaction.
     * 
     * @param <T> a comparable type. 
     * 
     * @param source a source reaction.
     * 
     * @param rxns keys mapped to sink reactions.
     * 
     * @return the keys of the provided map that correspond to the
     * identified sink reactions.
     * 
     * @see #reverseExchangeRxns(SourceRxn, SinkRxn)
     * @see #getReverseSinkRxns(SourceRxn, Map)
     */
    public static <T extends Comparable<? super T>> Set<T>
    getReverseSinkRxnIds(SourceRxn source, Map<T, SinkRxn> rxns) {
    	return getReverseSinkRxns(source, rxns).keySet();
    }
    
    /**
     * Returns all of the reactions in the provided map that are
     * the reverse of the reaction specified by the provided key.
     * 
     * @param <T> a comparable type. 
     * 
     * @param rxnId a key in the provided map that corresponds to the
     * reaction for which reverse reactions will be returned.
     * 
     * @param rxns keys mapped to irreversible reactions.
     * 
     * @return the entries of the provided map that contain the 
     * identified reverse reactions.
     * 
     * @see #reverseIrrevRxns(IrrevRxn, IrrevRxn)
     */
    public static <T extends Comparable<? super T>> Map<T, IrrevRxn> 
    getReverseIrrevRxns(T rxnId, Map<T, IrrevRxn> rxns) {

    	Map<T, IrrevRxn> reverseRxns = new TreeMap<T, IrrevRxn>();
    	
    	IrrevRxn r1 = rxns.get(rxnId);
    	
    	for (Map.Entry<T, IrrevRxn> e : rxns.entrySet()) {
    		
    		IrrevRxn r2 = e.getValue();
    		
    		if (reverseIrrevRxns(r1, r2)) {
    			reverseRxns.put(e.getKey(), r2);
    		}
    		
    	}
    	
    	return reverseRxns;
    }
    
    /**
     * Returns all of the source reactions in the provided map that are
     * the reverse of the provided sink reaction.
     * 
     * @param <T> a comparable type. 
     * 
     * @param sink a sink reaction.
     * 
     * @param rxns keys mapped to source reactions.
     * 
     * @return the entries of the provided map that contain the 
     * identified source reactions.
     * 
     * @see #reverseExchangeRxns(SinkRxn, SourceRxn)
     */
    public static <T extends Comparable<? super T>> Map<T, SourceRxn> 
    getReverseSourceRxns(SinkRxn sink, Map<T, SourceRxn> rxns) {

    	Map<T, SourceRxn> reverseRxns = new TreeMap<T, SourceRxn>();
    	
    	for (Map.Entry<T, SourceRxn> e : rxns.entrySet()) {
    		
    		SourceRxn source = e.getValue();
    		
    		if (reverseExchangeRxns(source, sink)) {
    			reverseRxns.put(e.getKey(), source);
    		}
    		
    	}
    	
    	return reverseRxns;
    }
    
    /**
     * Returns all of the sink reactions in the provided map that are
     * the reverse of the provided source reaction.
     * 
     * @param <T> a comparable type. 
     * 
     * @param source a source reaction.
     * 
     * @param rxns keys mapped to sink reactions.
     * 
     * @return the entries of the provided map that contain the 
     * identified sink reactions.
     * 
     * @see #reverseExchangeRxns(SourceRxn, SinkRxn)
     */
    public static <T extends Comparable<? super T>> Map<T, SinkRxn> 
    getReverseSinkRxns(SourceRxn source, Map<T, SinkRxn> rxns) {

    	Map<T, SinkRxn> reverseRxns = new TreeMap<T, SinkRxn>();
    	
    	for (Map.Entry<T, SinkRxn> e : rxns.entrySet()) {
    		
    		SinkRxn sink = e.getValue();
    		
    		if (reverseExchangeRxns(source, sink)) {
    			reverseRxns.put(e.getKey(), sink);
    		}
    		
    	}
    	
    	return reverseRxns;
    }
    
    
	
}
