/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba;

import sbrt.kernel.math.HashLinearComb;
import sbrt.kernel.math.LinearComb;

/**
 * This class is used to represent exchange reactions that transport chemicals from
 * a chemical system to its surroundings. Sink reactions are unidirectional -
 * they consume chemical species, but do not supply them.
 * <p>
 * Instances of this class are immutable.
 * </p>
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 */
public final class SinkRxn extends AbstractExchangeRxn {

	/**
	 * Constructs a new sink reaction for the provided chemical species.
	 * 
	 * @param chemSpecies
	 *            the chemical species this sink consumes.
	 */
    public SinkRxn(String chemSpecies) {
        super(chemSpecies);
    }

    /**
	 * Returns a single-term linear combination. The variable of this term is
	 * the consumed chemical species and the coefficient is 1.0.
	 * 
	 * @return a single-term linear combination. The variable of this term is
	 *         the consumed chemical species and the coefficient is 1.0.
	 */
    public LinearComb<String> getReactants() {
        return HashLinearComb.singleTerm(getSpecies(), 1d);
    }

	/**
	 * Returns {@code false}.
	 * @return {@code false}.
	 */
	public boolean isReversible() {
		return false;
	}

    /**
     * Returns a string representation of this sink reaction.
     * @return a string representation of this sink reaction.
     */
    public String toString() {
        return getReactants() + " --> Surroundings";
    }

}
