/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.analysis;

import java.util.*;

import sbrt.kernel.math.Interval;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to analyze flux intervals.
 * 
 * <p>The {@code zeroCutoff} arguments to each method must be:</p>
 * <ul>
 * <li>Greater than zero.</li>
 * <li>Not {@code Double.NaN}.</li>
 * <li>Not infinite.</li>
 * </ul>
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public class FluxIntervalAnalyzer {

	private void checkTolerance(double zeroCutoff) {

		if (Double.compare(zeroCutoff, 0d) < 0 
				|| Double.isNaN(zeroCutoff)
				|| Double.isInfinite(zeroCutoff)) {

			String s = "Invalid zeroCutoff: " + zeroCutoff;
			throw new IllegalArgumentException(s);
		}

	}

	private boolean equalsZero(double x, double zeroCutoff) {

		checkTolerance(zeroCutoff);

		return Util.equal(x, 0.0d, zeroCutoff);
	}

	private boolean greaterThanZero(double x, double zeroCutoff) {

		checkTolerance(zeroCutoff);

		return Double.compare(x, zeroCutoff) > 0;
	}

	/**
	 * Returns the always-active fluxes in the provided map.
	 * 
	 * @param fluxIntervals
	 *            reaction names mapped to their corresponding flux interval.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the keys in the provided map whose corresponding interval has a
	 *         lower bound greater than zero.
	 */
	public Set<String> getAlwaysActive(Map<String, Interval> fluxIntervals,
			double zeroCutoff) {

		checkTolerance(zeroCutoff);

		// Initializing storage
		Set<String> active = new TreeSet<String>();

		// Iterating thru each flux interval
		for (Map.Entry<String, Interval> e : fluxIntervals.entrySet()) {

			// Getting the lower bound
			double lower = e.getValue().getLower();

			// If the lower bound > 0
			if (greaterThanZero(lower, zeroCutoff)) {
				active.add(e.getKey());
			}

		}

		return active;
	}

	/**
	 * Returns the potentially-active fluxes in the provided map.
	 * 
	 * @param fluxIntervals
	 *            reaction names mapped to their corresponding flux interval.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the keys in the provided map whose corresponding interval has a
	 *         lower bound equal to zero and an upper bound greater than zero.
	 */
	public Set<String> getPotentiallyActive(
			Map<String, Interval> fluxIntervals, double zeroCutoff) {

		checkTolerance(zeroCutoff);

		// Initializing storage
		Set<String> rxns = new TreeSet<String>();

		// Iterating thru each flux interval
		for (Map.Entry<String, Interval> e : fluxIntervals.entrySet()) {

			// Getting the lower and upper bounds
			double lower = e.getValue().getLower();
			double upper = e.getValue().getUpper();

			// If the lower == 0 and the upper > 0
			if (equalsZero(lower, zeroCutoff)
					&& greaterThanZero(upper, zeroCutoff)) {
				rxns.add(e.getKey());
			}

		}

		return rxns;
	}

	/**
	 * Returns the never-active fluxes in the provided map.
	 * 
	 * @param fluxIntervals
	 *            reaction names mapped to their corresponding flux interval.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the keys in the provided map whose corresponding interval has
	 *         lower and upper bounds equal to zero.
	 */
	public Set<String> getNeverActive(Map<String, Interval> fluxIntervals,
			double zeroCutoff) {

		checkTolerance(zeroCutoff);

		// Initializing storage for fluxes
		Set<String> rxns = new TreeSet<String>();

		// Iterating thru each flux interval
		for (Map.Entry<String, Interval> e : fluxIntervals.entrySet()) {

			// Getting the bounds
			double lower = e.getValue().getLower();
			double upper = e.getValue().getUpper();

			// If both bounds == 0
			if (equalsZero(lower, zeroCutoff) && equalsZero(upper, zeroCutoff)) {
				rxns.add(e.getKey());
			}

		}

		return rxns;
	}

	/**
	 * Returns the number of provided intervals whose lower bound equals zero.
	 * 
	 * @param intervals
	 *            a map whose values are intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose lower bound equals zero.
	 */
	public int lowerBoundsEqualZero(Map<?, Interval> intervals,
			double zeroCutoff) {
		return lowerBoundsEqualZero(intervals.values(), zeroCutoff);
	}

	/**
	 * Returns the number of provided intervals whose lower bound equals zero.
	 * 
	 * @param intervals
	 *            a collection of intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose lower bound equals zero.
	 */
	public int lowerBoundsEqualZero(Collection<Interval> intervals,
			double zeroCutoff) {

		checkTolerance(zeroCutoff);

		// Initializing counter for fluxes with
		// zero as their lower bound
		int n = 0;

		// Iterating thru each flux interval
		for (Interval interval : intervals) {

			// Getting the lower bound
			double lower = interval.getLower();

			// If it equals zero
			if (equalsZero(lower, zeroCutoff)) {
				++n;
			}

		}

		return n;
	}

	/**
	 * Returns the number of provided intervals whose lower bounds are greater
	 * than zero.
	 * 
	 * @param intervals
	 *            a map whose values are intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose lower bounds are greater
	 * than zero.
	 */
	public int lowerBoundsGreaterThanZero(Map<?, Interval> intervals,
			double zeroCutoff) {
		return lowerBoundsGreaterThanZero(intervals.values(), zeroCutoff);
	}

	/**
	 * Returns the number of provided intervals whose lower bounds are greater
	 * than zero.
	 * 
	 * @param intervals
	 *            a collection of intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose lower bounds are greater
	 * than zero.
	 */
	public int lowerBoundsGreaterThanZero(Collection<Interval> intervals,
			double zeroCutoff) {

		checkTolerance(zeroCutoff);

		// Initializing counter for fluxes with
		// a lower bound greater than zero
		int n = 0;

		// Iterating thru each interval
		for (Interval i : intervals) {

			// Getting the lower bound
			double lower = i.getLower();

			// If it's greater than the provided cutoff
			if (greaterThanZero(lower, zeroCutoff)) {
				++n;
			}

		}

		return n;
	}

	/**
	 * Returns the number of provided intervals whose upper bounds equal zero.
	 * 
	 * @param intervals
	 *            a map whose values are intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose upper bounds equal zero.
	 */
	public int upperBoundsEqualZero(Map<?, Interval> intervals,
			double zeroCutoff) {
		return upperBoundsEqualZero(intervals.values(), zeroCutoff);
	}

	/**
	 * Returns the number of provided intervals whose upper bounds equal zero.
	 * 
	 * @param intervals
	 *            a collection of intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose upper bounds equal zero.
	 */
	public int upperBoundsEqualZero(Collection<Interval> intervals,
			double zeroCutoff) {

		checkTolerance(zeroCutoff);

		// Initializing counter for intervals with
		// zero as their upper bound
		int n = 0;

		// Iterating thru each interval
		for (Interval i : intervals) {

			// Getting the upper bound
			double upper = i.getUpper();

			// If it equals zero
			if (equalsZero(upper, zeroCutoff)) {
				++n;
			}

		}

		return n;
	}

	/**
	 * Returns the number of provided intervals whose upper bounds are greater
	 * than zero.
	 * 
	 * @param intervals
	 *            a map whose values are intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose upper bounds are greater
	 * than zero.
	 */
	public int upperBoundsGreaterThanZero(Map<?, Interval> intervals,
			double zeroCutoff) {
		return upperBoundsGreaterThanZero(intervals.values(), zeroCutoff);
	}

	/**
	 * Returns the number of provided intervals whose upper bounds are greater than zero.
	 * 
	 * @param intervals
	 *            a collection of intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose upper bounds are greater than zero.
	 */
	public int upperBoundsGreaterThanZero(Collection<Interval> intervals,
			double zeroCutoff) {

		checkTolerance(zeroCutoff);

		// Initializing counter for intervals with
		// an upper bound greater than zero
		int n = 0;

		// Iterating thru each interval
		for (Interval i : intervals) {

			// Getting the upper bound
			double upper = i.getUpper();

			// If it's greater than the provided cutoff
			if (greaterThanZero(upper, zeroCutoff)) {
				++n;
			}

		}

		return n;
	}

	/**
	 * Returns the number of provided intervals whose ranges equal zero.
	 * 
	 * @param intervals
	 *            a map whose values are intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose ranges equal zero.
	 */
	public int rangesEqualZero(Map<?, Interval> intervals, double zeroCutoff) {
		return rangesEqualZero(intervals.values(), zeroCutoff);
	}

	/**
	 * Returns the number of provided intervals whose ranges equal zero.
	 * 
	 * @param intervals
	 *            a collection of intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose ranges equal zero.
	 */
	public int rangesEqualZero(Collection<Interval> intervals, double zeroCutoff) {

		checkTolerance(zeroCutoff);

		// Initializing counter for intervals with
		// ranges equal to zero
		int n = 0;

		// Iterating thru each interval
		for (Interval interval : intervals) {

			// Substracting the lower from the upper
			double r = interval.getUpper() - interval.getLower();

			// If it equals zero
			if (equalsZero(r, zeroCutoff)) {
				++n;
			}

		}

		return n;
	}

	/**
	 * Returns the number of provided intervals whose ranges are greater than zero.
	 * 
	 * @param intervals
	 *            a map whose values are intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose ranges are greater than zero.
	 */
	public int rangesGreaterThanZero(Map<?, Interval> intervals,
			double zeroCutoff) {
		return rangesGreaterThanZero(intervals.values(), zeroCutoff);
	}

	/**
	 * Returns the number of provided intervals whose ranges are greater than zero.
	 * 
	 * @param intervals
	 *            a collection of intervals.
	 * @param zeroCutoff
	 *            the amount by which a value can differ from zero, but still be
	 *            considered equal to zero.
	 * 
	 * @return the number of provided intervals whose ranges are greater than zero.
	 */
	public int rangesGreaterThanZero(Collection<Interval> intervals,
			double zeroCutoff) {

		checkTolerance(zeroCutoff);

		// Initializing counter for intervals with
		// a range greater than zero
		int n = 0;

		// Iterating thru each interval
		for (Interval interval : intervals) {

			// Substracting the lower from the upper
			double range = interval.getUpper() - interval.getLower();

			// If it's greater than the provided cutoff
			if (greaterThanZero(range, zeroCutoff)) {
				++n;
			}

		}

		return n;
	}

}
