/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;

/**
 * This abstract class is a skeleton implementation of {@linkplain FluxCvComputer}. 
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * 
 * @param <F> a type of fluxome. 
 *
 */
public abstract class AbstractCvComputer<F extends Fluxome> 
	implements FluxCvComputer<F> {

    private final F fluxome;

    /**
     * Constructs a new constraint-variation computer.
     * 
     * @param fluxome the fluxome for which values will be computed.
     * 
     * @throws NullPointerException if the provided fluxome is null.
     */
    public AbstractCvComputer(F fluxome) {
    	
    	// Checking and setting
        Util.nullCheck(fluxome);
        this.fluxome = fluxome;
    }

    /**
     * Returns the fluxome for which values will be computed.
     * @return the fluxome for which values will be computed.
     */
    public F getFluxome() {
        return fluxome;
    }

    /**
     * Ensures the reaction name is valid for the fluxome provided.
     * 
     * @param reactionName String The reaction name to check.
     * 
     * @throws IllegalArgumentException if the reaction name is not
     * present in the fluxome.
     */
    protected void checkRxnName(String reactionName) {

        if (!fluxome.isRxnName(reactionName)) {
            String s = "The reaction '" + reactionName +
                       "' does not exist in the fluxome provided.";
            throw new IllegalArgumentException(s);
        }

    }

    /**
     * Ensures the reaction names are valid for the fluxome provided.
     * 
     * @param reactionNames the reaction names to check.
     * 
     * @throws IllegalArgumentException if a reaction name in the provided
     * collection is not present in the fluxome.
     * @throws NullPointerException if the provided Collection is 
     * {@code null}.
     */
    protected void checkRxnNames(Collection<String> reactionNames) {
        
    	// Iterating thru each provided reaction name
    	for (String rxnName : reactionNames) {
            checkRxnName(rxnName);
        }
    }

    /**
     * Ensures the expression is valid for the fluxome provided.
     * 
     * @param e the expression to check.
     * 
     * @throws IllegalArgumentException if the expression contains variables
     * that are not valid reaction names in the fluoxme.
     * @throws NullPointerException if the provided expression is 
     * {@code null}.
     */
    protected void checkRxnExpr(MathExpr<String> e) {
        checkRxnNames(e.getVariables());
    }

    /**
     * Ensures the flux constraint is valid for the fluxome provided.
     * 
     * @param c the constraint to check.
     * @param rxnName the name of the reaction whose flux this constraint
     * applies to.
     * @param tolerance the numerical tolerance to use during the comparison.
     * 
     * @throws IllegalArgumentException if the flux constraint is not
     * valid for the fluoxme provided.
     * @throws NullPointerException if the provided constraint is 
     * {@code null}.
     */
    protected void 
    checkFluxConstraint(Interval c, String rxnName, double tolerance) {
    	
        if (!fluxome.isValid(c, rxnName, tolerance)) {
            String s = "The constraint '" + c +
                       "' is illegal for: " + rxnName;
            throw new IllegalArgumentException(s);
        }

    }

    /**
     * Ensures the expression constraint is valid for the fluxome provided.
     * 
     * @param c the constraint to check.
     * @throws NullPointerException if the provided constraint is 
     * {@code null}.
     */
    protected void checkRxnExprConstraint(Interval c) {
        Util.nullCheck(c);
    }

    /**
     * Ensures the provided number is not {@code NaN} or infinite.
     * 
     * @param number the number to check.
     * @throws IllegalArgumentException if the provided number is 
     * {@code NaN} or infinite.
     */
    protected void checkNumber(double number) {
    	
    	if (Double.isInfinite(number) || Double.isNaN(number)) {
    		String s = "Invalid number: " + number;
    		throw new IllegalArgumentException(s);
    	}
    }
    
    /**
     * Converts a reaction name to a single-term linear combination.
     * 
     * @param reactionName the reaction name to convert.
     * 
     * @return a linear combination containing a single term, with the
     * provided reaction name as the variable and coefficient equal to 1.
     */
    protected LinearComb<String> getRxnExpr(String reactionName) {
    	
    	// Checking input
        checkRxnName(reactionName);

        // Returning linear combination with a single term
        return HashLinearComb.singleTerm(reactionName);
    }

    /**
     * Converts a collection of reaction names to a linear combination.
     * 
     * @param reactionNames the reaction names to appear as variables.
     * 
     * @return a linear combination of the provided reaction names.
     */
    protected LinearComb<String> getRxnExpr(Collection<String> reactionNames) {

        // Removing any duplicates that may exist
        Set<String> s = new LinkedHashSet<String>(reactionNames);

        // Creating storage for linear combination terms
        List<LinearComb.Term<String>> terms = 
        	new LinkedList<LinearComb.Term<String>>();

        // Iterating thru each reaction name
        for (String rxnName : s) {

        	// Ensuring the name is valid
            checkRxnName(rxnName);

            // Creating new term
            LinearComb.Term<String> term =
            	new HashLinearComb.Term<String>(rxnName);
            
            // Adding term to the rest
            terms.add(term);
        }

        // Returning linear combination
        return new HashLinearComb<String>(terms, 0d, 0d);
    }

    /**
     * Changes the constraints on more than one mathematical expression 
     * of fluxes.
     * 
     * @param constraints mathematical expressions of reaction names mapped
     * to their respective constraints.
     */
    public void setRxnExprConstraints(
    		Map<? extends MathExpr<String>, Interval> constraints) {

    	for (Map.Entry<? extends MathExpr<String>, Interval> e :
    		constraints.entrySet()) {
    		
            setRxnExprConstraint(e.getKey(), e.getValue());
        }
    }

    /**
     * Changes the flux constraints on more than one reaction.
     * 
     * @param fluxConstraints reaction names mapped to their respective flux
     * constraints.
     */
    public void setFluxConstraints(Map<String, Interval> fluxConstraints) {

    	// Iterating thru each reaction name-constraint pair
    	for (Map.Entry<String, Interval> e : fluxConstraints.entrySet()) {
            setFluxConstraint(e.getKey(), e.getValue());
        }
    }

    /**
     * Restores the constraints on the provided expressions to the values
     * contained in the fluxome.
     * 
     * @param rxnNameExps the reaction name expressions whose contraints
     * are to be restored.
     */
    public void restoreRxnExprConstraints(
    		Collection<? extends MathExpr<String>> rxnNameExps) {

    	// Iterating thru each expression
    	for (MathExpr<String> e : rxnNameExps) {
            restoreRxnExprConstraint(e);
        }
    }

    /**
     * Restores the flux constraints for the provided reaction names to the
     * values contained in the fluxome.
     * 
     * @param reactionNames the reaction names whose flux constraints are to
     * be restored.
     */
    public void restoreFluxConstraints(Collection<String> reactionNames) {

    	// Iterating thru each reaction name
    	for (String rxnName : reactionNames) {
            restoreFluxConstraint(rxnName);
        }
    }

    /**
     * Restores the flux constraints for all reactions in the fluxome to those 
     * values contained in the fluxome.
     */
    public void restoreFluxConstraints() {
        restoreFluxConstraints(fluxome.getRxnNames());
    }

}
