/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.io.WarningHandlerFactory;
import sbrt.kernel.math.*;
import sbrt.kernel.math.optimization.*;
import sbrt.kernel.utilities.*;

/**
 * This abstract class is a skeleton implemenation of 
 * {@linkplain FluxIntervalComputer}.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type.
 */
public abstract class AbstractFluxIntervalComputer<F extends Fluxome> 
	extends AbstractCvComputer<F>
		implements FluxIntervalComputer<F> {

	/**
	 * The default zero cutoff.
	 */
	public static final double ZERO_CUTOFF = 1E-9;

	private final FbaOptimizer<F> computer;
	private final double zeroCutoff;

	/**
	 * Constructs a new flux interval computer with the default zero cutoff.
	 * 
	 * @param computer the computer with which all optimizations will be
	 * performed.
	 * 
	 * @see #ZERO_CUTOFF
	 */
	public AbstractFluxIntervalComputer(FbaOptimizer<F> computer) {
		this(computer, ZERO_CUTOFF);
	}

	/**
	 * Constructs a new flux interval computer with the provided zero cutoff.
	 * 
	 * @param computer the computer with which all optimizations will be
	 * performed.
	 * @param zeroCutoff the zero cutoff to be used by this computer.
	 */
	public AbstractFluxIntervalComputer(FbaOptimizer<F> computer,
										double zeroCutoff) {

		// Calling constructor of super class
		super(computer.getFluxome());

		// Checking inputs
		Util.nullCheck(computer);
		checkCutoff(zeroCutoff);

		// Setting
		this.computer = computer;
		this.zeroCutoff = zeroCutoff;
	}

	private void checkCutoff(double cutoff) {

		// Ensuring range cutoff is greater than zero
		// and not infinite or NaN
		if (Double.isInfinite(zeroCutoff) || Double.isNaN(zeroCutoff)
				|| zeroCutoff < 0) {

			String s = "Invalid range cutoff: " + zeroCutoff;
			throw new IllegalArgumentException(s);
		}

	}

	/**
	 * Returns the constraint tolerance used by this flux interval computer.
	 * @return the constraint tolerance used by this flux interval computer.
	 */
	public double getTolerance() {
		return computer.getTolerance();
	}

	/**
	 * Returns the zero cutoff used by this flux interval computer.
	 * @return the zero cutoff used by this flux interval computer.
	 */
	public double getZeroCutoff() {
		return zeroCutoff;
	}

	/**
	 * Computes the maximum flux of the reaction with the provided name.
	 * 
	 * @param reactionName the name of the reaction whose maximum flux will be 
	 * computed.
	 * 
	 * @return the maximum flux of the reaction with the provided name, or
	 * {@code Double.POSITIVE_INFINITY} if the flux is unbounded.
	 * 
	 * @throws FbaOptException if an optimal solution was not found.
	 * 
	 * @throws IllegalArgumentException if the provided string
	 * is not a reaction name of the fluxome used by this flux interval 
	 * computer.
	 */
	protected double getMaximum(String reactionName) {

		// Converting to linear combination
		MathExpr<String> expr = computer.getRxnExpr(reactionName);

		// Returning maximum flux thru this reaction
		return getMaximum(expr);
	}

	/**
	 * Computes the maximum value of the provided objective function.
	 * 
	 * @param rxnNameExpr a mathematical expression
	 * whose variables are reaction names.
	 * 
	 * @return the maximum value of the provided objective function, or
	 * {@code Double.POSITIVE_INFINITY} if the problem is unbounded.
	 * 
	 * @throws FbaOptException if an optimal solution was not found.
	 * 
	 * @throws IllegalArgumentException if the provided expression
	 * contains a variable that is not a reaction name.
	 * 
	 * @throws NullPointerException if the provided expression is {@code null}.
	 */
	protected double getMaximum(MathExpr<String> rxnNameExpr) {

		try {

			// Computing the maximum flux
			return computer.getMaximum(rxnNameExpr);

			// If something went wrong
		} catch (FbaOptException e) {

			// Getting the program solver status
			ProgramSolverStatus status = e.getStatus();

			switch (status) {

                case INFEASIBLE_OR_UNBOUNDED:  // Assuming unbounded

                    String s = "The program solver status for the last computed " +
                               "optimization is '" + status +
                               "'. The SBRT is assuming the problem was unbounded.";
                    
                    WarningHandlerFactory.getDefault().warn(s);

                    return Double.POSITIVE_INFINITY;
                    
                    // If the flux was unbounded, return positive infinity
			    case UNBOUNDED:
                    return Double.POSITIVE_INFINITY;

                    // Otherwise something else happened
                default:
                    throw e;
                        
            }

		}

	}

	/**
	 * Computes the minimum flux of the reaction with the provided name.
	 * 
	 * @param reactionName the name of the reaction whose minimum flux will be 
	 * computed.
	 * 
	 * @return the minimum flux of the reaction with the provided name, or
	 * {@code Double.NEGATIVE_INFINITY} if the flux is unbounded.
	 * 
	 * @throws FbaOptException if an optimal solution was not found.
	 * 
	 * @throws IllegalArgumentException if the provided string
	 * is not a reaction name of the fluxome used by this flux interval 
	 * computer.
	 */
	protected double getMinimum(String reactionName) {

		// Converting to linear combination
		MathExpr<String> exp = computer.getRxnExpr(reactionName);

		// Returning minimum flux thru this reaction
		return getMinimum(exp);
	}

	/**
	 * Computes the minimum value of the provided objective function.
	 * 
	 * @param rxnNameExpr a mathematical expression
	 * whose variables are reaction names.
	 * 
	 * @return the minimum value of the provided objective function, or
	 * {@code Double.NEGATIVE_INFINITY} if the problem is unbounded.
	 * 
	 * @throws FbaOptException if an optimal solution was not found.
	 * 
	 * @throws IllegalArgumentException if the provided expression
	 * contains a variable that is not a reaction name.
	 * 
	 * @throws NullPointerException if the provided expression is {@code null}.
	 */
	protected double getMinimum(MathExpr<String> rxnNameExpr) {

		try {

			// Computing the minimum flux
			return computer.getMinimum(rxnNameExpr);

			// If something went wrong
		} catch (FbaOptException e) {

			// Getting the program solver status
			ProgramSolverStatus status = e.getStatus();

			// If the flux was unbounded, return negative infinity
			if (status == ProgramSolverStatus.UNBOUNDED) {
				return Double.NEGATIVE_INFINITY;
			}

			// Otherwise something else happened
			throw e;
		}

	}

	/**
	 * Computes the minimum value of the provided objective function.
	 * 
	 * @param rxnNameExpr a mathematical expression of reaction names.
	 * 
	 * @return the minimum value of the provided objective function.
	 * 
	 * @throws UnboundedVariableException if the optimization problem
	 * proves to be unbounded.
	 */
	protected double getBoundedMin(MathExpr<String> rxnNameExpr) {

		double min = getMinimum(rxnNameExpr);

		if (Double.isInfinite(min)) {
			throw new UnboundedVariableException(rxnNameExpr);
		}

		return min;
	}

	/**
	 * Computes the maximum value of the provided objective function.
	 * 
	 * @param rxnNameExpr a mathematical expression of reaction names.
	 * 
	 * @return the maximum value of the provided objective function.
	 * 
	 * @throws UnboundedVariableException if the optimization problem
	 * proves to be unbounded.
	 */
	protected double getBoundedMax(MathExpr<String> rxnNameExpr) {

		double max = getMaximum(rxnNameExpr);

		if (Double.isInfinite(max)) {
			throw new UnboundedVariableException(rxnNameExpr);
		}

		return max;
	}

	/**
	 * Computes the minimum flux of the reaction with the provided name.
	 * 
	 * @param reactionName the name of a reaction in the fluxome used by this
	 * computer.
	 * 
	 * @return the minimum flux of the reaction with the provided name.
	 * 
	 * @throws UnboundedVariableException if the optimization problem
	 * proves to be unbounded.
	 */
	protected double getBoundedMin(String reactionName) {

		double min = getMinimum(reactionName);

		if (Double.isInfinite(min)) {
			throw new UnboundedVariableException(reactionName);
		}

		return min;
	}

	/**
	 * Computes the maximum flux of the reaction with the provided name.
	 * 
	 * @param reactionName the name of a reaction in the fluxome used by this
	 * computer.
	 * 
	 * @return the maximum flux of the reaction with the provided name.
	 * 
	 * @throws UnboundedVariableException if the optimization problem
	 * proves to be unbounded.
	 */
	protected double getBoundedMax(String reactionName) {

		// Getting the maximum flux
		double max = getMaximum(reactionName);

		// If the flux is infinite
		if (Double.isInfinite(max)) {
			throw new UnboundedVariableException(reactionName);
		}

		return max;
	}

	/**
	 * Computes the maximum flux of the reaction with the provided name while
	 * constraining the fluxes of the provided reactions to their minimum.
	 * 
	 * @param reactionName the name of a reaction in the fluxome used by this
	 * computer.
	 * 
	 * @param rxnNames a set of reaction names.
	 * 
	 * @return the maximum flux of the reaction with the provided name.
	 * 
	 * @throws UnboundedVariableException if the optimization problem
	 * proves to be unbounded.
	 */
	protected double 
	getBoundedMaxWithRxnsMinimized(String reactionName, 
							       Set<String> rxnNames) {
		
		double max = getMaxWithRxnsMinimized(reactionName, rxnNames);

		// If the flux is infinite, throw an error
		if (Double.isInfinite(max)) {
			throw new UnboundedVariableException(reactionName);
		}

		return max;
	}

	/**
	 * Computes the maximum flux of the reaction with the provided name while
	 * constraining the fluxes of the provided reactions to their minimum.
	 * 
	 * @param reactionName the name of a reaction in the fluxome used by this
	 * computer.
	 * 
	 * @param rxnNames a set of reaction names.
	 * 
	 * @return the maximum flux of the reaction with the provided name, or
	 * {@code Double.POSITIVE_INFINITY} if the problem is unbounded.
	 */
	protected double 
	getMaxWithRxnsMinimized(String reactionName, 
							Set<String> rxnNames) {

		// If no reverse reactions exist, just return the normal max
		if (rxnNames.isEmpty()) {
			return getMaximum(reactionName);
		}

		// Getting an expression of reverse reactions
		MathExpr<String> exp = computer.getRxnExpr(rxnNames);

		// Saving the current constraint for this expression if it exists
		Interval orig = getRxnExprConstraint(exp);

		// Computing the minimum value of this expression
		double minFlux = getBoundedMin(exp);

		// Constraining the reverse rxn expression to its minimum
		Interval min = getInterval(reactionName, orig, minFlux, minFlux);
		setRxnExprConstraint(exp, min);

		// Computing the maximum flux thru this reaction
		double max = getMaximum(reactionName);

		// Restoring original constraint for this expression
		setRxnExprConstraint(exp, orig);

		// Returning maximum flux thru the provided reaction with all reverse
		// reactions constrained to their minimum value
		return max;
	}

	/**
	 * Returns the provided minimum and maximum fluxes as an interval. This
	 * method performs several error-checking procedures to ensure the provided
	 * values are consistent with each other and with the defined constraint
	 * tolerance and zero cutoff. It also corrects for small numerical errors in
	 * the provided computed values.
	 * 
	 * @param rxn
	 *            the name of the reaction whose lower and upper flux bounds
	 *            have been computed.
	 * 
	 * @param fluxConstraint
	 *            the current constraint on the flux whose lower and upper
	 *            bounds have been computed.
	 * 
	 * @param computedMin
	 *            the computed lower flux bound.
	 * @param computedMax
	 *            the computed upper flux bound.
	 * 
	 * @return the provided minimum and maximum fluxes as an interval.
	 * 
	 * @throws ConstraintViolationException
	 *             if either of the computed fluxes lie outside the provided
	 *             flux constraint by more than the defined constraint
	 *             tolerance.
	 * 
	 * @throws IllegalBoundsException
	 *             if the provided minimum is greater than the provided maximum
	 *             by more than the defined zero cutoff.
	 */
	protected Interval getInterval(String rxn, 
								   Interval fluxConstraint,
								   double computedMin,
								   double computedMax) {

		// If the minimum is greater than the maximum by more than the
		// defined zero cutoff, complain
		if (computedMin - zeroCutoff > computedMax) {
			throw new IllegalBoundsException(rxn, computedMin, computedMax);
		}

		// If the minimum is greater than the maximum by less than the
		// defined zero cutoff, assume its just from small numerical
		// error and swap the values
		if (Double.compare(computedMin, computedMax) > 0) {
			double temp = computedMin;
			computedMin = computedMax;
			computedMax = temp;
		}

		// Ensuring the computed minimum and maximum fall within
		// the defined flux constraint (within the defined constraint
		// tolerance)
		checkFluxValue(rxn, fluxConstraint, computedMin);
		checkFluxValue(rxn, fluxConstraint, computedMax);

		// Getting the lower and upper bounds for this flux
		double lower = fluxConstraint.getLower();
		double upper = fluxConstraint.getUpper();

		// If the computed minimum is slighty less than the
		// defined lower bound, change it to the defined value
		if (Double.compare(computedMin, lower) < 0) {
			computedMin = lower;
		}

		// If the computed minimum is slighty greater than the
		// defined upper bound, change it to the defined value
		if (Double.compare(computedMin, upper) > 0) {
			computedMin = upper;
		}

		// If the computed maximum is slightly less than the
		// defined lower bound, change it to the defined value
		if (Double.compare(computedMax, lower) < 0) {
			computedMax = lower;
		}

		// If the computed maximum is slightly greater than the
		// defined upper bound, change it to the defined value
		if (Double.compare(computedMax, upper) > 0) {
			computedMax = upper;
		}

		// Returning a constraint with proper bounds
		return new Interval(computedMin, computedMax);
	}

	private void checkFluxValue(String rxnName, Interval bounds, double flux) {

		// If the provided flux lies between the provided bounds
		// (given the defined constraint tolerance), do nothing
		if (bounds.isWithin(flux, getTolerance())) {
			return;
		}

		// Otherwise complain
		throw new ConstraintViolationException(rxnName, bounds, flux);
	}

	/**
	 * Returns the provided minima and maxima as intervals. This method
	 * performs several error-checking procedures to ensure the provided
	 * minima and maxima are consistent with each other and with the defined
	 * constraint tolerance and zero cutoff. It also corrects for small
	 * numerical errors in the provided computed values.
	 * 
	 * @param fluxMinima the computed flux minima.
	 * @param fluxMaxima the computed flux maxima.
	 * 
	 * @return reaction names mapped to the interval dictated by the provided
	 * minimum and maximum values.
	 */
	@SuppressWarnings("boxing")
	protected Map<String, Interval> 
	getIntervals(Map<String, Double> fluxMinima, 
				 Map<String, Double> fluxMaxima) {

		// Comparing keys
		if (!fluxMinima.keySet().equals(fluxMaxima.keySet())) {
			String s = "Keys don't match: " + fluxMinima.keySet() + " != "
					+ fluxMaxima.keySet();
			throw new IllegalArgumentException(s);
		}

		// Determining appropriate size
		int size = fluxMinima.size() + 1;

		// Creating storage for flux ranges
		Map<String, Interval> fluxIntervals = 
			new LinkedHashMap<String, Interval>(size, 1.0F);

		// Iterating thru each
		for (Map.Entry<String, Double> e : fluxMinima.entrySet()) {

			// Getting reaction name and minimum flux
			String name = e.getKey();
			Double minFlux = e.getValue();

			// Getting maximum flux
			Double maxFlux = fluxMaxima.get(name);

			// Getting current flux constraint
			Interval current = getFluxConstraint(name);

			// Creating the flux interval
			Interval interval = getInterval(name, current, minFlux, maxFlux);

			// Storing flux range
			fluxIntervals.put(name, interval);
		}

		return fluxIntervals;
	}
	
    /**
     * Changes the flux constraint of the reaction with the provided name.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be changed.
     * 
     * @param c the new constraint.
     */
	public void setFluxConstraint(String reactionName, Interval c) {
		computer.setFluxConstraint(reactionName, c);
	}

    /**
     * Changes the flux constraints on more than one reaction.
     * 
     * @param fluxConstraints reaction names mapped to their respective flux
     * constraints.
     */
	public void setFluxConstraints(Map<String, Interval> fluxConstraints) {
		computer.setFluxConstraints(fluxConstraints);
	}

    /**
     * Changes the constraint on the provided mathematical expression 
     * of fluxes.
     * 
     * @param e a matheamtical expression of reaction names.
     * @param c the new constraint.
     */
	public void setRxnExprConstraint(MathExpr<String> e, Interval c) {
		computer.setRxnExprConstraint(e, c);
	}

    /**
     * Changes the constraints on more than one mathematical expression 
     * of fluxes.
     * 
     * @param constraints mathematical expressions of reaction names mapped
     * to their new respective constraints.
     */
	public void setRxnExprConstraints(
			Map<? extends MathExpr<String>, Interval> constraints) {

		computer.setRxnExprConstraints(constraints);
	}

    /**
     * Restores the flux constraint for the reaction with the provided 
     * name to the value contained in the fluxome.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be restored.
     */
	public void restoreFluxConstraint(String reactionName) {
		computer.restoreFluxConstraint(reactionName);
	}

    /**
     * Restores the flux constraints for the provided reaction names to the
     * values contained in the fluxome.
     * 
     * @param reactionNames the reaction names whose flux constraints are to
     * be restored.
     */
	public void restoreFluxConstraints(Collection<String> reactionNames) {
		computer.restoreFluxConstraints(reactionNames);
	}

    /**
     * Restores the flux constraints for all reactions in the fluxome to those 
     * values contained in the fluxome.
     */
	public void restoreFluxConstraints() {
		computer.restoreFluxConstraints();
	}

    /**
     * Restores the constraint on the provided flux expression to the value
     * contained in the fluxome.
     * 
     * @param rxnNameExp the expression of reaction names whose contraint
     * is to be restored.
     */
	public void restoreRxnExprConstraint(MathExpr<String> rxnNameExp) {
		computer.restoreRxnExprConstraint(rxnNameExp);
	}

    /**
     * Restores the constraints on the provided expressions to the values
     * contained in the fluxome.
     * 
     * @param rxnNameExps the reaction name expressions whose contraints
     * are to be restored.
     */
	public void restoreRxnExprConstraints(
			Collection<? extends MathExpr<String>> rxnNameExps) {
		computer.restoreRxnExprConstraints(rxnNameExps);
	}

    /**
     * Restores the constraints on all flux expressions to the values
     * contained in the fluxome.
     */
	public void restoreRxnExprConstraints() {
		computer.restoreRxnExprConstraints();
	}

    /**
     * Returns all current flux expression constraints.
     * @return mathematical expressions of reaction names mapped to their
     * respective constraint.
     */
	public Map<MathExpr<String>, Interval> getRxnExprConstraints() {
		return computer.getRxnExprConstraints();
	}

    /**
     * Returns all current flux constraints.
     * @return reaction names mapped to their respective constraint.
     */
	public Map<String, Interval> getFluxConstraints() {
		return computer.getFluxConstraints();
	}

    /**
     * Returns the current constraint on the provided expression.
     * 
     * @param rxnNameExp the mathematical expression of reaction names
     * whose constraint is to be returned.
     * 
     * @return the current constraint on the provided expression.
     */
	public Interval getRxnExprConstraint(MathExpr<String> rxnNameExp) {
		return computer.getRxnExprConstraint(rxnNameExp);
	}

    /**
     * Returns the current flux constraint on the reaction with the provided
     * name.
     * 
     * @param rxnName the name of the reaction whose flux constraint
     * is to be returned.
     * 
     * @return the current flux constraint on the reaction with the provided
     * name. 
     */
	public Interval getFluxConstraint(String rxnName) {
		return computer.getFluxConstraint(rxnName);
	}

}
