/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.math.*;
import sbrt.kernel.math.algebra.*;
import sbrt.kernel.math.geom.*;
import sbrt.kernel.math.geom.comp.CdHarComp;
import sbrt.kernel.utilities.*;

/**
 * This class is used to compute randomly distributed vectors within flux space.
 * It uses a coordinate-direction hit-and-run algorithm.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * @param <F> the fluxome type.
 * 
 */
public class CdHarFluxComp<F extends Fluxome> 
	extends AbstractCvComputer<F> {

    /** The default constraint tolerance. */
    public static final double DEFAULT_TOLERANCE = 1E-9;

    /** The default minimum chord length. */
    public static final double DEFAULT_MIN_CHORD = 5E-9;

    /** The default zero cutoff for coefficients in linear combinations. */
    public static final double DEFAULT_COEFF_CUTOFF = 1E-9;
    
    // The fluxome solution
    private final LinearSystemSolution<String> originalSolution;

    // The simplified solution
    private LinearSystemSolution<Object> modifiedSolution;

    // The initial flux vector
    private final Map<String, Double> startPoint;

    // The random flux vector generator
    private CdHarComp<Object> pointGenerator;

    // The psuedorandom number generator
    private Random random;

    // The constraint tolerance
    private double tolerance;
    
    // The points per points
    private int pointsPerPoint;

    // The minimum chord length
    private double minChordLength;

    // Storage for all flux constaints
    private Map<String, Interval> fluxConstraints;

    // Storage for all expression constraints
    private Map<LinearComb<String>, Interval> exprConstraints;

    /**
     * Constructs a new random flux vector computer for the provided fluxome.
     * 
     * @param fluxome the fluxome for which random flux vectors will be
     * computed.
     * 
     * @param systemSolution a solution to the system of linear equations formed
     * by the provided fluxome.
     */
    public CdHarFluxComp(F fluxome,
            			 LinearSystemSolution<String> systemSolution) {

        super(fluxome);

        // Checking variables in system solution
        checkRxnNames(systemSolution.getVariables());

        // Setting system solution
        this.originalSolution = systemSolution;

        // Determining appropriate size
        int size = fluxome.rxns() + 1;

        // Creating storage for flux constraints
        fluxConstraints = new LinkedHashMap<String, Interval>(size, 1.0F);

        // Creating storage for rxn expression constraints
        exprConstraints = new LinkedHashMap<LinearComb<String>, Interval>();

        // Filling constraints
        restoreFluxConstraints();
        restoreRxnExprConstraints();

        // Creating storage for the start point
        this.startPoint = new LinkedHashMap<String, Double>(size, 1.0F);

        // Initializing other fields
        tolerance = DEFAULT_TOLERANCE;
        minChordLength = DEFAULT_MIN_CHORD;
    }

    /**
     * Sets the random number generator.
     * 
     * @param random the random number generator with which all psuedorandom
     * numbers are to be produced.
     */
    public void setRandom(Random random) {
        Util.nullCheck(random);
        this.random = random;
    }

    /**
     * Sets the number of points to generate for each point added to the 
     * specified results manager.
     * 
     * @param pointsPerPoint the number of points to generate for each point added to the 
     * specified results manager.
     */
    public void setPointsPerPoint(int pointsPerPoint) {

    	if (pointsPerPoint <= 0) {
    		String s = "Invalid points-per-point: " + pointsPerPoint;
    		throw new IllegalArgumentException(s);
    	}
    	
    	this.pointsPerPoint = pointsPerPoint;
    }

    /**
     * Sets the constraint tolerance. This is the amount by which constraints
     * can be violated.
     * 
     * @param tolerance the amount by which constaints can be violated. This
     * should be a small positive number, like 1E-9 for example.
     */
    public void setTolerance(double tolerance) {

        // Checking input
        if (Double.isInfinite(tolerance)
                || Double.isNaN(tolerance)
                || tolerance < 0d) {

            String s = "Invalid constraint tolerance: " + tolerance;
            throw new IllegalArgumentException(s);
        }

        this.tolerance = tolerance;
    }


    /**
     * Sets the minimum chord length. If a computed chord is shorter than the
     * provided amount, the move is considered wasted. This can happen if the
     * random walker gets stuck in a corner.
     * 
     * @param minChordLength the minimum length a chord must have to be
     * considered valid. This should be a small positive number, like 1E-9 for
     * example.
     */
    public void setMinChordLength(double minChordLength) {

        // Checking input
        if (Double.isInfinite(minChordLength)
                || Double.isNaN(minChordLength)
                || minChordLength < 0d) {

            String s = "Invalid minimum chord length: " + minChordLength;
            throw new IllegalArgumentException(s);
        }

        this.minChordLength = minChordLength;
    }

    /**
     * Sets the initial flux vector to the one provided.
     * 
     * @param vector reaction names mapped to the flux of their corresponding
     * reaction.
     */
    public void setInitialFluxVector(Map<String, Double> vector) {

        // Ensuring the provided vector is a valid solution vector
        if (!originalSolution.isSolutionVector(vector, tolerance)) {
            String s = "The provided initial flux vector is " +
                        "incompatible with the system solution " +
                        "with the constraint tolerance " + tolerance;
            throw new InvalidFluxVectorException(vector, tolerance, s);
        }

        // Copying
        startPoint.clear();
        startPoint.putAll(vector);
    }
    
    /**
     * Returns the constraint tolerance. This is the amount by which constraints
     * can be violated.
     * 
     * @return the constraint tolerance.
     * 
     * @see #setTolerance(double)
     */
    public double getTolerance() {
        return tolerance;
    }
    
    /**
     * Returns the minimum length a computed chord must have to be considered
     * successful. Chords with a length smaller than the returned value are
     * considered wasted moves.
     * 
     * @return the minimum chord length.
     * 
     * @see #setMinChordLength(double)
     */
    public double getMinChordLength() {
        return minChordLength;
    }

    private void checkState() {

        // Ensuring an initial flux vector was provided
        if (startPoint.isEmpty()) {
            String s = "An initial flux vector was never provided.";
            throw new IllegalStateException(s);
        }
        
        if (random == null) {
        	String s = "A random number generator was never provided.";
        	throw new IllegalStateException(s);
        }

    }

    /**
     * Computes the next random flux vector from within the interior of the
     * convex polytope defined by the current set of constraints.
     * 
     * @return the results of computing the next random flux vector.
     */
    public Results run() {

        checkState();

        // If the point generator doesn't exist, create it
        if (pointGenerator == null) {

            // Setting the global variable
            pointGenerator = makePointGenerator();

            // Setting start point
            setStartPoint();
        }

        // Computing CD Hit and Run results
        CdHarComp<Object>.Results results = pointGenerator.next(pointsPerPoint);

        // Wrapping and returning
        return new Results(results);
    }

    private CdHarComp<Object> makePointGenerator() {

        // Creating the convex polytope
        ConvexPolytope<Object> polytope = makePolytope();

        // Creating a new Coordinate Direction Hit-and-Run point generator
        CdHarComp<Object> c = new CdHarComp<Object>(polytope);
        c.setRandom(random);
        c.setMinChordLength(minChordLength);
        c.setConstraintTolerance(tolerance);

        return c;
    }

    private void setStartPoint() {

        try {

            // Copying the initial flux vector into a new map
            Map<String, Double> m = new TreeMap<String, Double>();
            m.putAll(startPoint);

            // Getting the free variables
            Set<Object> vars = pointGenerator.getPolytope().getVariables();

            // Retaining only the free variables
            m.keySet().retainAll(vars);

            // Trying to set the start point
            pointGenerator.setPoint(m);

        } catch (InvalidPointException e) {

            // If something is wrong with the initial point
            String s = "The initial flux vector does not lie within " +
                       "the convex polytope defined by the current " +
                       "constraints with the constraint tolerance " +
                       tolerance;

            throw new InvalidFluxVectorException(startPoint, tolerance, s, e);
        }
    }

    private ConvexPolytope<Object> makePolytope() {

        // Determining appropriate size
        int size = originalSolution.variables() + exprConstraints.size() + 1;

        // Creating storage for a new linear system solution
        Map<Object, LinearComb<String>> solution =
                new LinkedHashMap<Object, LinearComb<String>>(size, 1.0F);

        // Adding all solutions from the original fluxome solution
        solution.putAll(originalSolution.toMap());

        // Iterating thru each expression constraint
        for (LinearComb<String> expr : exprConstraints.keySet()) {

            // Converting it to a linear combination of free variables
            LinearComb<String> sol = expr.plugIn(solution);

            // Storing it, using the original expression as the "variable"
            // and the simplified expression as the solution
            solution.put(expr, sol);
        }

        // Creating a new solution that contains each expression "variable"
        LinearSystemSolution<Object> fullSolution =
                new HashLinearSystemSolution<Object>(solution);

        // Determining appropriate size
        size = fluxConstraints.size() + exprConstraints.size() + 1;

        // Creating storage for all variables and their constraints
        Map<Object, Interval> constraints =
                new LinkedHashMap<Object, Interval>(size, 1.0F);

        // Adding all constraints
        constraints.putAll(fluxConstraints);
        constraints.putAll(exprConstraints);

        // Getting the variables whose range is zero
        Map<Object, Double> values = getFixedVariables(constraints);

        // Creating a simplified solution
        modifiedSolution = fullSolution.simplify(values, tolerance);

        // Creating the convex polyope formed by the current flux
        // and expression constraints
        return HashConvexPolytope
                .make(modifiedSolution, constraints, tolerance);
    }

    private static <V> Map<V, Double> getFixedVariables(Map<? extends V, Interval> variableBounds) {

        // Determining appropriate size
        int size = variableBounds.size() + 1;

        // Creating storage
        Map<V, Double> boundVars = new LinkedHashMap<V, Double>(size, 1.0F);

        // Iterating thru each variable-bound pair
        for (Map.Entry<? extends V, Interval> e : variableBounds.entrySet()) {

            // Getting the bounds
            Interval c = e.getValue();
            double lower = c.getLower();
            double upper = c.getUpper();

            // If the lower and upper bounds are equal
            if (upper == lower) {

                // Converting to Double
                Double x = Double.valueOf(lower);

                // Storing
                boundVars.put(e.getKey(), x);
            }

        }

        return boundVars;
    }

    /**
     * Changes the flux constraint of the reaction with the provided name.
     * 
     * @param reactionName the name of the reaction whose flux constraint is to
     * be changed.
     * 
     * @param c the new constraint.
     */
    public void setFluxConstraint(String reactionName, Interval c) {

        // Checking the inputs
        checkRxnName(reactionName);
        checkFluxConstraint(c, reactionName, tolerance);

        // Replacing the current flux constraint
        fluxConstraints.put(reactionName, c);

        // Destroying the current flux vector generator
        pointGenerator = null;
    }

    /**
     * Restores the flux constraint for the reaction with the provided name to
     * the value contained in the fluxome.
     * 
     * @param reactionName the name of the reaction whose flux constraint is to
     * be restored.
     */
    public void restoreFluxConstraint(String reactionName) {

        // Checking the input
        checkRxnName(reactionName);

        // Getting the original flux constraint
        Interval c = getFluxome().getFluxConstraint(reactionName);

        // Replacing the current flux constraint
        fluxConstraints.put(reactionName, c);

        // Destroying the current flux vector generator
        pointGenerator = null;
    }

    /**
     * Changes the constraint on the provided mathematical expression of fluxes.
     * 
     * @param rxnExpr a {@linkplain LinearComb} of reaction names.
     * @param c the new constraint.
     */
    public void setRxnExprConstraint(MathExpr<String> rxnExpr, Interval c) {

        // *UNSAFELY* casting
        LinearComb<String> lin = (LinearComb<String>) rxnExpr;

        // Setting the expression constraint
        setRxnExprConstraint(lin, c);
    }

    /**
     * Changes the constraint on the provided linear combination of fluxes.
     * 
     * @param rxnExpr a linear combination of reaction names.
     * @param c the new constraint.
     */
    public void setRxnExprConstraint(LinearComb<String> rxnExpr, Interval c) {

        // Checking the inputs
        checkRxnExpr(rxnExpr);
        checkRxnExprConstraint(c);

        // Replacing the current expression constraint
        exprConstraints.put(rxnExpr, c);

        // Destroying the current flux vector generator
        pointGenerator = null;
    }

    /**
     * Restores the constraint on the provided flux expression to the value
     * contained in the fluxome.
     * 
     * @param expr the {@linkplain LinearComb} of reaction names whose contraint is
     * to be restored.
     */
    public void restoreRxnExprConstraint(MathExpr<String> expr) {

        // *UNSAFELY* casting
        LinearComb<String> lin = (LinearComb<String>) expr;

        // Restoring constraint
        restoreRxnExprConstraint(lin);
    }

    /**
     * Restores the constraint on the provided flux expression to the value
     * contained in the fluxome.
     * 
     * @param expr the linear combination of reaction names whose contraint is
     * to be restored.
     */
    public void restoreRxnExprConstraint(LinearComb<String> expr) {

        // Checking the input
        checkRxnExpr(expr);

        // Getting the original expression constraints
        Map<MathExpr<String>, Interval> exprConsts =
                getFluxome().getConstraints().getExprConstraints();

        // Getting the original constraint, if it exists
        Interval c = exprConsts.get(expr);

        // If it doesn't
        if (c == null) {

            // Removing the expression
            exprConstraints.remove(expr);
        }

        // If it does
        else {

            // Replacing the current constraint
            exprConstraints.put(expr, c);
        }

        // Destroying the current flux vector generator
        pointGenerator = null;
    }

    /**
     * Restores the constraints on all flux expressions to the values contained
     * in the fluxome.
     */
    public void restoreRxnExprConstraints() {

        // Clearing all current expression constraints
        exprConstraints.clear();

        // Getting the original expression constraints
        Map<MathExpr<String>, Interval> expr =
                getFluxome().getConstraints().getExprConstraints();

        // Iterating thru each expression
        for (Map.Entry<MathExpr<String>, Interval> e : expr.entrySet()) {

            // UNSAFELY casting
            LinearComb<String> lin = (LinearComb<String>) e.getKey();

            // Setting constraint
            exprConstraints.put(lin, e.getValue());
        }

        // Destroying the current flux vector generator
        pointGenerator = null;
    }

    /**
     * Returns all current flux constraints.
     * 
     * @return reaction names mapped to their respective constraint.
     */
    public Map<String, Interval> getFluxConstraints() {

        // Determining appropriate size
        int size = fluxConstraints.size() + 1;

        // Creating storage
        Map<String, Interval> m =
                new LinkedHashMap<String, Interval>(size, 1.0F);

        // Copying
        m.putAll(fluxConstraints);

        return m;
    }

    /**
     * Returns all current flux expression constraints.
     * 
     * @return mathematical expressions of reaction names mapped to their
     * respective constraint.
     */
    public Map<LinearComb<String>, Interval> getRxnExprConstraints() {

        // Determining appropriate size
        int size = exprConstraints.size() + 1;

        // Creating storage
        Map<LinearComb<String>, Interval> m =
                new LinkedHashMap<LinearComb<String>, Interval>(size, 1.0F);

        // Copying
        m.putAll(exprConstraints);

        return m;
    }

    /**
     * Returns the current constraint on the provided expression.
     * 
     * @param rxnNameExp the mathematical expression of reaction names whose
     * constraint is to be returned.
     * 
     * @return the current constraint on the provided expression.
     */
    public Interval getRxnExprConstraint(MathExpr<String> rxnNameExp) {

        // Checking the input
        checkRxnExpr(rxnNameExp);

        // Getting the stored constraint if it exists
        Interval c = exprConstraints.get(rxnNameExp);

        // If it does, return it
        if (c != null) {
            return c;
        }

        // If not, return its default value from the fluxome
        return getFluxome().getRxnExprConstraint(rxnNameExp);
    }

    /**
     * Returns the current flux constraint on the reaction with the provided
     * name.
     * 
     * @param rxnName the name of the reaction whose flux constraint is to be
     * returned.
     * 
     * @return the current flux constraint on the reaction with the provided
     * name.
     */
    public Interval getFluxConstraint(String rxnName) {

        // Checking input
        checkRxnName(rxnName);

        return fluxConstraints.get(rxnName);
    }

    /**
     * This class is used to represent the results of a coordinate-direction
     * hit-and-run flux vector computation.
     * 
     * @author This class was written and documented by Jeremiah Wright while in
     * the Wagner lab.
     * 
     */
    public class Results {

        private final CdHarComp<Object>.Results results;

        private Results(CdHarComp<Object>.Results results) {
            this.results = results;
        }

        /**
         * Returns the random interior flux vector that was computed.
         * 
         * @return the random interior flux vector that was computed.
         */
        public Map<String, Double> getFluxVector() {

            // Getting the random interior point
            UnmodifiableMap<Object, Double> p = results.getPoint();

            // Getting the solution vector
            Map<Object, Double> solVector =
                    modifiedSolution.getSolutionVector(p);

            // Creating the flux vector
            Map<String, Double> fluxVector = new TreeMap<String, Double>();

            for (Map.Entry<Object, Double> e : solVector.entrySet()) {

                // If the variable is a reaction name
                if (e.getKey() instanceof String) {
                    fluxVector.put((String) e.getKey(), e.getValue());
                }

            }

            // Computing a full flux vector from this point
            return fluxVector;
        }

        /**
         * Returns the number of free variables whose values are the same as
         * they were in the initial flux vector. Ideally, this number should
         * equal 0.
         * 
         * @return the number of free variables whose values are the same as
         * they were in the initial flux vector.
         */
        public int unchangedFreeVariables() {
            return results.unchangedVariables();
        }
    }

}
