/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.math.algebra.SparseDoubleMatrix;
import sbrt.kernel.utilities.MutableProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;

/**
 * This interface is used to represent extreme current computers. An extreme
 * current is very similar to an extreme pathway. For a fluxome containing only
 * irreversible reactions (including exchange reactions), no extreme pathways
 * exists, only extreme currents.
 * 
 * @author This interface was written and documented by Jeremiah Wright while in
 * the Wagner lab.
 * 
 */
public interface ExtremeCurrentComputer {

    /**
     * Returns the extreme current matrix for the provided fluxome.
     * 
     * @param fluxome the fluxome for which all extreme currents will be
     * computed.
     * 
     * @return the extreme current matrix for the provided fluxome.
     */
    public SparseDoubleMatrix<?, String> getMatrix(IrrevRxnFluxome fluxome);

    /**
     * Returns a progress monitor for this extreme current computer.
     * 
     * @return a progress monitor for this extreme current computer.
     */
    public ProgressMonitor getProgress();

    /**
     * Returns the total number of "operations" required to compute the extreme
     * currents of the provided fluxome. The number returned by this method is
     * meant to be used with progress monitors.
     * 
     * @param fluxome the fluxome for which the number of operations will be
     * returned.
     * 
     * @return the total number of operations required to compute the extreme
     * currents of the provided fluxome.
     */
    public int getTotalOps(IrrevRxnFluxome fluxome);

    /**
     * Returns the extreme current matrix for the provided fluxome.
     * 
     * @param fluxome the fluxome for which all extreme currents will be
     * computed.
     * @param pm the progress monitor to be used during computation of extreme
     * currents.
     * 
     * @return the extreme current matrix for the provided fluxome.
     */
    public SparseDoubleMatrix<?, String> getMatrix(IrrevRxnFluxome fluxome,
                                                   MutableProgressMonitor pm);
}
