/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import sbrt.kernel.math.*;
import sbrt.kernel.math.optimization.*;
import sbrt.kernel.utilities.*;

/**
 * This class is used to indicate that an optimal solution could not be found
 * or does not exist.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 */
public class FbaOptException extends RuntimeException {

	private static final long serialVersionUID = 1L;
	
	private final MathExpr<?> objectiveFunction;
    private final OptSense sense;
    private final ProgramSolverStatus status;

    /**
     * Constructs a new FBA optimization exception.
     * 
     * @param objectiveFunction the objective function whose optimal value
     * could not be found.
     * @param sense the sense, or direction, of the attempted optimization.
     * @param status the status of the attempted optimization.
     */
    public FbaOptException(MathExpr<?> objectiveFunction,
                                    OptSense sense,
                                    ProgramSolverStatus status) {
        this(objectiveFunction, sense, status, null);
    }

    /**
     * Constructs a new FBA optimization exception.
     * 
     * @param objectiveFunction the objective function whose optimal value
     * could not be found.
     * @param sense the sense, or direction, of the attempted optimization.
     * @param status the status of the attempted optimization.
     * @param message a message useful for debugging.
     */
    public FbaOptException(MathExpr<?> objectiveFunction,
                                    OptSense sense,
                                    ProgramSolverStatus status,
                                    String message) {
        super(getMessage(objectiveFunction, sense, status, message));

        // Checking inputs
        Util.nullCheck(objectiveFunction);
        Util.nullCheck(sense);
        Util.nullCheck(status);

        // Setting
        this.objectiveFunction = objectiveFunction;
        this.sense = sense;
        this.status = status;
    }

    /**
     * Returns the objective function whose optimal value could not be found.
     * @return the objective function whose optimal value could not be found.
     */
    public MathExpr<?> getObjFunction() {
        return objectiveFunction;
    }

    /**
     * Returns the sense, or direction, of the attempted optimization.
     * @return the sense, or direction, of the attempted optimization.
     */
    public OptSense getSense() {
        return sense;
    }

    /**
     * Returns the status of the attempted optimization.
     * @return the status of the attempted optimization.
     */
    public ProgramSolverStatus getStatus() {
        return status;
    }

    private static String getMessage(MathExpr<?> objectiveFunction,
                                     OptSense sense,
                                     ProgramSolverStatus status,
                                     String message) {

        String s = "The program solver was unable to find an optimal " +
                   "solution for the problem '" + sense + ": " +
                   objectiveFunction + "'. Its status is '" + status + "'.";

        if (message != null) {
            s += " The following was also reported: " + message;
        }

        return s;
    }

}
