/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.math.*;
import sbrt.kernel.math.algebra.*;
import sbrt.kernel.math.optimization.*;
import sbrt.kernel.utilities.*;

/**
 * This class is the workhorse of all Flux Balance Analysis optimizations. 
 * It provides an interface for communication between FBA processes
 * and program solvers.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type. 
 *
 */
public final class FbaOptimizer<F extends Fluxome> implements FluxCvComputer<F> {

	/** The default saftey level. */
	public static final int DEFAULT_SAFETY_LEVEL = 1;

	/** The default constraint tolerance. */
	public static final double DEFAULT_CONST_TOL = 0.0d;
	
	private static final Interval unbounded = 
		new Interval(Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY);

	private static final Interval off = new Interval(0, 0);

	// The fluxome
	private final F fluxome;

	// The program solver used for all optimizations
	private final ProgramSolver<Object, Object> programSolver;

	// The current objective function
	private MathExpr<?> objectiveFunction;

	// The current optimization sense
	private OptSense sense;

	// Storage for all row constraints
	private Map<Object, Interval> rowConstraints;

	// Storage for all column constraints
	private Map<Object, Interval> columnConstraints;

	// Storage for all column-variable expression constraints
	private Map<MathExpr<?>, Interval> exprConstraints;

	// Tolerance for numerical error of column constraints
	private double tolerance;

	// Used to indicate whether computed values need to be checked
	// for numerical error
	private int safetyLevel;

	/**
	 * Constructs a new FBA optimization computer from the 
	 * provided fluxome and program solver.
	 * 
	 * @param fluxome the fluxome for which all values will be computed.
	 * @param solver the program solver used to compute all values.
	 */
	public FbaOptimizer(F fluxome, ProgramSolver<Object, Object> solver) {
		this(fluxome, solver, DEFAULT_CONST_TOL);
	}

	/**
	 * Constructs a new FBA optimization computer from the 
	 * provided fluxome, program solver, and constraint tolerance.
	 * 
	 * @param fluxome the fluxome for which all values will be computed.
	 * @param solver the program solver used to compute all values.
	 * @param constraintTolerance the allowed amount by which
	 * computed values can violate their defined constraints.
	 */
	public FbaOptimizer(F fluxome,
			            ProgramSolver<Object, Object> solver, 
			            double constraintTolerance) {

		// Checking and setting fluxome
		checkFluxome(fluxome);
		this.fluxome = fluxome;
		
		// Setting tolerance
		setTolerance(constraintTolerance);

		// Setting safety level to default value
		setSafetyLevel(DEFAULT_SAFETY_LEVEL);

		// Checking and setting program solver
		Util.nullCheck(solver);
		programSolver = solver;

		// Filling row constraints
		setRowConstraints();

		// Filling column constraints
		setColumnConstraints();
		
		// Filling column variable expression constraints
		setColExprConstraints();

		// Defining initial program matrix and constraints
		defineProgramSolver();
	}
	
	private void checkFluxome(Fluxome fluxome) {
		
		// Ensuring fluxome is not null
		if (fluxome == null) {
			String s = "The provided fluxome is null.";
			throw new NullPointerException(s);
		}
	}
	
	private void setRowConstraints() {

		// Determining appropriate size
		int size = fluxome.species() + 1;

		// Creating storage for row constraints
		rowConstraints = new LinkedHashMap<Object, Interval>(size, 1.0F);

		// Iterating thru each species (i.e. row variable)
		for (String name : fluxome.getSpecies()) {

			// Setting (steady-state) row constraint
			rowConstraints.put(name, off);
		}

	}

	private void setColumnConstraints() {
		
		// Getting all reaction names in the provided fluxome
		Set<String> rxnNames = fluxome.getRxnNames();
		
		// Determining appropriate size
		int size = rxnNames.size() + 1;
		
		// Creating storage
		columnConstraints = 
			new LinkedHashMap<Object, Interval>(size, 1.0F);
		
		// Iterating thru each name, setting its constraint
		for (String rxn : rxnNames) {
			
			// Getting its constraint
			Interval c = fluxome.getFluxConstraint(rxn);

			// Saving it
			columnConstraints.put(rxn, c);
		}

	}
	
	private void setColExprConstraints() {
		
		// Getting all constraints in the provided fluxome
		FluxConstraints constraints = fluxome.getConstraints();
		
		// Getting all reaction name expressions and their constraints
		Map<MathExpr<String>, Interval> exprConsts = 
			constraints.getExprConstraints();
		
		// Determining appropriate size
		int size = exprConsts.size() + 1;
		
		// Creating storage
		exprConstraints = 
			new LinkedHashMap<MathExpr<?>, Interval>(size, 1.0F);
		
		// Copying
		exprConstraints.putAll(exprConsts);
		
	}
	
	/**
	 * Sets the safety level. Safety levels are used to ensure
	 * that program solvers do not violate the defined constraints
	 * of the system. A 
	 * safety level of 0 means no error checking will be done. A
	 * safety level of 1 means the solution vector will be checked 
	 * for column constraint violations after each optimization.
	 * 
	 * @param level the saftey level.
	 * 
	 * @throws IllegalArgumentException if the provided value does not 
	 * equal 0 or 1.
	 * 
	 * @see #DEFAULT_SAFETY_LEVEL
	 * @see #setTolerance(double)
	 */
	public void setSafetyLevel(int level) {

		if (level != 0 && level != 1) {
			String s = "Invalid safety level: " + level;
			throw new IllegalArgumentException(s);
		}

		safetyLevel = level;
	}

	/**
	 * Sets the constraint tolerance.
	 *
	 * @param constraintTolerance the amount by which computed values can 
	 * violate their defined constraints.
	 * 
	 * @throws IllegalArgumentException if the provided value is NaN, 
	 * infinite, or less than 0.
	 * 
	 * @see #setSafetyLevel(int)
	 */
	public void setTolerance(double constraintTolerance) {

		// Ensuring the tolerance is neither infinite nor NaN
		checkNumber(constraintTolerance);

		// Ensuring it's positive
		if (constraintTolerance < 0) {
			String s = "Invalid tolerance: " + constraintTolerance;
			throw new IllegalArgumentException(s);
		}

		// Setting
		this.tolerance = constraintTolerance;
	}

	/**
	 * Returns the allowable amount by which column variables
	 * may violate their defined constraints.
	 * @return the allowable amount by which column variables
	 * may violate their defined constraints.
	 */
	public double getTolerance() {
		return tolerance;
	}

	/**
	 * Returns the {@code Class} of the program solver used by this 
	 * optimization computer.
	 * @return the {@code Class} of the program solver used by this 
	 * optimization computer.
	 */
	public Class<?> getProgramSolverType() {
		return programSolver.getClass();
	}

	/**
	 * Returns the fluxome used by this optimization computer.
	 * @return the fluxome used by this optimization computer.
	 */
	public F getFluxome() {
		return fluxome;
	}

	/**
	 * Changes the current objective function to the one provided.
	 * 
	 * @throws IllegalArgumentException if the provided objective
	 * function contains a variable that is not a valid column variable
	 * of the program matrix.
	 */
	private final void setObjectiveFunction(MathExpr<?> objectiveFunction) {

		// Checking input
		checkColExpr(objectiveFunction);

		// Setting locally
		this.objectiveFunction = objectiveFunction;

		// Setting in the program solver
		programSolver.setObjectiveFunction(objectiveFunction);
	}

	/**
	 * Changes the optimiztion sense to the one provided.
	 */
	private final void setOptimizationSense(OptSense sense) {

		// Ensuring the provided sense is not null
		if (sense == null) {
			String s = "The provided optimization sense is null.";
			throw new NullPointerException(s);
		}

		// Setting locally
		this.sense = sense;

		// Setting in the program solver
		programSolver.setOptimizationSense(sense);
	}

	/**
	 * Sets the specified element of the program matrix to the one provided.
	 * If either variable does not yet exist in this computer, its  
	 * constraint is set to (-Infinity, Infinity).
	 * 
	 * @param rowVar the row variable. 
	 * @param columnVar the column variable.
	 * @param value the new value.
	 * 
	 * @throws IllegalArgumentException if {@code value} is infinite or NaN.
	 * @throws NullPointerException if either variable is {@code null}. 
	 */
	public final void 
	setMatrixElement(Object rowVar, Object columnVar, double value) {
		
		// Checking row index
		if (rowVar == null) {
			String s = "The provided row index is null.";
			throw new NullPointerException(s);
		}
		
		// Checking column index
		if (columnVar == null) {
			String s = "The provided column index is null.";
			throw new NullPointerException(s);
		}

		// Checking value
		checkNumber(value);

		// Changing the matrix element in the program solver
		programSolver.setMatrixElement(rowVar, columnVar, value);

		// If this column variable does not yet exist,
		// set its constraint to be unbounded
		if (!columnConstraints.containsKey(columnVar)) {
			
			// Setting locally and in the program solver
			columnConstraints.put(columnVar, unbounded);
			programSolver.setColumnConstraint(columnVar, unbounded);
		}

		// If this row variable does not yet exist,
		// set its constraint to be unbounded
		if (!rowConstraints.containsKey(rowVar)) {

			// Setting locally and in the program solver
			rowConstraints.put(rowVar, unbounded);
			programSolver.setRowConstraint(rowVar, unbounded);
		}
	}

	/**
	 * Sets the specified elements of the program matrix to those provided.
	 * If any provided variable does not yet exist in this computer, its  
	 * constraint is set to (-Infinity, Infinity).
	 * 
	 * @param elements a matrix containing the elements to be set.
	 * 
	 * @throws IllegalArgumentException if {@code value} is infinite or NaN.
	 * @throws NullPointerException if either variable is {@code null}. 
	 */
	public final void setMatrixElements(DoubleMatrix<?, ?> elements) {

		// Iterating thru each column index
		for (Object columnIndex : elements.getColumnIndices()) {

			// Iterating thru each row index
			for (Object rowIndex : elements.getRowIndices()) {

				// Getting the value at these indices from the provided matrix
				double value = elements.getElement(rowIndex, columnIndex);

				// Changing the current value to the one provided
				setMatrixElement(rowIndex, columnIndex, value);
			}
		}

	}

	/**
	 * Restores the program matrix to the stoichiometry matrix of the fluxome
	 * used by this computer.
	 */
	public final void restoreMatrixElements() {

		// Getting all the original species names
		Set<String> species = fluxome.getSpecies();

		// Getting all the original reaction names
		Set<String> rxnNames = fluxome.getRxnNames();

		// Getting all the original reaction name expressions
		Set<MathExpr<String>> expressions = 
			fluxome.getConstraints().getExprConstraints().keySet();

		// Retaining only the original species and
		// their current constraints
        
        // ClassCastException is thrown with the following:
		// rowConstraints.keySet().retainAll(species);
        
        // So, the following must be done first:
        for (Iterator<?> i = rowConstraints.keySet().iterator(); i.hasNext(); ) {
            
            if (!(i.next() instanceof String)) {
                i.remove();
            }
            
        }

        rowConstraints.keySet().retainAll(species);
        
		// Retaining only the original reaction names and
		// their current constraints....with more Java nonsense
        for (Iterator<?> i = columnConstraints.keySet().iterator(); i.hasNext(); ) {
            
            if (!(i.next() instanceof String)) {
                i.remove();
            }
            
        }
        
		columnConstraints.keySet().retainAll(rxnNames);

		// Retaining only the original reaction name expressions and
		// their current constraints
		exprConstraints.keySet().retainAll(expressions);

		// Redefining the program solver
		defineProgramSolver();
	}

    /**
     * Changes the constraint of the provided column variable.
     * 
     * @param columnVar the column variable in the program matrix whose
     * constraint is to be changed.
     * @param c the new constraint.
     * 
     * @see #setFluxConstraint(String, Interval)
     * 
     * @throws IllegalArgumentException if {@code columnVar} is not a column
     * variable in the program matrix.
     * @throws NullPointerException if the provided constraint is {@code null}.
     */
	public final void setColumnConstraint(Object columnVar, Interval c) {

		// Ensuring column index is valid
		checkColumnVariable(columnVar);

		// If it's a reaction name, check the 
		// constraint as a flux constraint
		if (isRxnName(columnVar)) {
			checkFluxConstraint(c, (String) columnVar, tolerance);
		}

		// Otherwise, just make sure it's not null
		else {
			Util.nullCheck(c);
		}

		// Setting locally
		columnConstraints.put(columnVar, c);

		// Setting in the program solver
		programSolver.setColumnConstraint(columnVar, c);
	}

    /**
     * Changes the flux constraint of the reaction with the provided name.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be changed.
     * @param c the new constraint.
     * 
     * @see #setColumnConstraint(Object, Interval)
     * 
     * @throws IllegalArgumentException if {@code reactionName} is not the name
     * of a reaction in the fluxome used by this computer.
     * @throws NullPointerException if the provided constraint is {@code null}.
     */
	public final void setFluxConstraint(String reactionName, Interval c) {
		
		// Ensuring the reaction name is valid
		checkRxnName(reactionName);
		
		// Setting column constraint
		setColumnConstraint(reactionName, c);
	}

    /**
     * Changes the constraint on the provided mathematical expression 
     * of fluxes.
     * 
     * @param expr a matheamtical expression of reaction names.
     * @param c the new constraint.
     * 
     * @throws IllegalArgumentException if the variables in the provided
     * expressions are not the names of reactions in the fluxome used by this
     * computer.
     * @throws NullPointerException if either argument is {@code null}.
     */
	public final void setRxnExprConstraint(MathExpr<String> expr, Interval c) {
		
		// Ensuring all variables are valid reaction names
		checkRxnExpr(expr);
		
		// Setting constraints
		setColExprConstraint(expr, c);
	}

    /**
     * Changes the constraint on the provided mathematical expression 
     * of column variables.
     * 
     * @param expr a mathematical expression of column variables from the
     * program matrix.
     * @param c the new constraint.
     * 
     * @throws IllegalArgumentException if the variables in the provided
     * expressions are not column variables of the program matrix.
     * @throws NullPointerException if either argument is {@code null}.
     */
	public final void setColExprConstraint(MathExpr<?> expr, Interval c) {

		// Checking inputs
		checkColExpr(expr);
		Util.nullCheck(c);

		// Setting locally
		exprConstraints.put(expr, c);

		// Setting in program solver
		programSolver.setConstraint(expr, c);
	}

    /**
     * Changes the constraint of the provided row variable.
     * 
     * @param rowVariable the row variable in the program matrix whose
     * constraint is to be changed.
     * @param c the new constraint.
     * 
     * @throws IllegalArgumentException if {@code rowVariable} is not a row
     * variable of the program matrix.
     * @throws NullPointerException if the provided constraint is {@code null}. 
     */
	public final void setRowConstraint(Object rowVariable, Interval c) {

		// Checking inputs
		checkRowVariable(rowVariable);
		Util.nullCheck(c);

		// Setting locally
		rowConstraints.put(rowVariable, c);

		// Setting in the program solver
		programSolver.setRowConstraint(rowVariable, c);
	}

    /**
     * Changes the constraints on more than one column variable.
     * 
     * @param columnConstraints column variables mapped to their respective 
     * constraints.
     */
	public final void setColumnConstraints(Map<?, Interval> columnConstraints) {

		// Iterating thru each column variable-constraint pair
		for (Map.Entry<?, Interval> e : columnConstraints.entrySet()) {
			setColumnConstraint(e.getKey(), e.getValue());
		}
	}

    /**
     * Changes the flux constraints on more than one reaction.
     * 
     * @param fluxConstraints reaction names mapped to their respective flux
     * constraints.
     */
	public final void setFluxConstraints(Map<String, Interval> fluxConstraints) {

		// Iterating thru each reaction name-constraint pair
		for (Map.Entry<String, Interval> e : fluxConstraints.entrySet()) {
			setFluxConstraint(e.getKey(), e.getValue());
		}
	}

    /**
     * Changes the constraints on more than one mathematical expression 
     * of fluxes.
     * 
     * @param exprConstraints mathematical expressions of reaction names
     * mapped to their new respective constraints.
     */
	public final void setRxnExprConstraints(
			Map<? extends MathExpr<String>, Interval> exprConstraints) {

		// Iterating thru each expression-constraint pair
		for (Map.Entry<? extends MathExpr<String>, Interval> e : 
			exprConstraints.entrySet()) {

			// Setting rxn expression constraint
			setRxnExprConstraint(e.getKey(), e.getValue());
		}
	}

    /**
     * Changes the constraints on more than one mathematical expression 
     * of column variables.
     * 
     * @param exprConstraints mathematical expressions of column variables
     * mapped to their new respective constraints.
     */
	public final void setColExprConstraints(
			Map<? extends MathExpr<?>, Interval> exprConstraints) {

		// Iterating thru each expression-constraint pair
		for (Map.Entry<? extends MathExpr<?>, Interval> e : 
			exprConstraints.entrySet()) {

			// Setting column expression constraint
			setColExprConstraint(e.getKey(), e.getValue());
		}
	}

    /**
     * Changes the constraints on more than one row variable.
     * 
     * @param rowConstraints row variables mapped to their respective 
     * constraints.
     */
	public final void setRowConstraints(Map<?, Interval> rowConstraints) {

		// Iterating thru each row variable-constraint pair
		for (Map.Entry<?, Interval> e : rowConstraints.entrySet()) {
			setRowConstraint(e.getKey(), e.getValue());
		}
	}

    /**
     * Restores the constraint for the provided column variable 
     * to the value contained in the fluxome. If this variable is not 
     * present in the fluxome, its constraint is set to (-Infinity, Infinity).
     * 
     * @param columnVariable the column variable whose constraint
     * is to be restored.
     */
	public final void restoreColumnConstraint(Object columnVariable) {

		// Checking input
		checkColumnVariable(columnVariable);

		// Creating field for storage
		Interval original = null;

		// If the provided column variable is a reaciton name
		if (isRxnName(columnVariable)) {

			// Safely casting
			String rxnName = (String) columnVariable;
			
			// Use its constraint from the fluxome
			original = fluxome.getFluxConstraint(rxnName);
		}

		else {

			// Otherwise consider its original state as unbounded
			original = unbounded;
		}

		// Setting locally
		columnConstraints.put(columnVariable, original);

		// Setting in the program solver
		programSolver.setColumnConstraint(columnVariable, original);
	}

    /**
     * Restores the flux constraint for the reaction with the provided 
     * name to the value contained in the fluxome.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be restored.
     */
	public final void restoreFluxConstraint(String reactionName) {
		
		// Ensuring the reaction name is valid
		checkRxnName(reactionName);
		
		// Restoring the original flux constraint
		restoreColumnConstraint(reactionName);
	}

    /**
     * Restores the constraint on the provided flux expression to the value
     * contained in the fluxome.
     * 
     * @param rxnNameExpr the expression of reaction names whose contraint
     * is to be restored.
     */
	public final void restoreRxnExprConstraint(MathExpr<String> rxnNameExpr) {

		// Checking input for errors
		checkRxnExpr(rxnNameExpr);
		
		// Restoring constraints
		restoreColExprConstraint(rxnNameExpr);
	}

    /**
     * Restores the constraint on the provided expression of column variables
     * to the value contained in the fluxome. If this expression is not 
     * present in the fluxome, its constraint is set to (-Infinity, Infinity).
     * 
     * @param columnVarExpr the expression of column variables whose contraint
     * is to be restored.
     */
	@SuppressWarnings("unchecked")
	public final void restoreColExprConstraint(MathExpr<?> columnVarExpr) {

		// Checking input
		checkColExpr(columnVarExpr);

		// Creating field for storage
		Interval original = null;

		// If it's a reaction name expression
		if (isRxnNameExpr(columnVarExpr)) {

			// Safely casting
			MathExpr<String> rxnNameExpr = (MathExpr<String>) columnVarExpr;
			
			// Use its constraint from the fluxome
			original = fluxome.getRxnExprConstraint(rxnNameExpr);
		}

		// Otherwise consider its original state as unbounded
		else {
			original = unbounded;
		}

		// Setting locally
		exprConstraints.put(columnVarExpr, original);

		// Setting in the program solver
		programSolver.setConstraint(columnVarExpr, original);
	}

	/**
	 * Restores the constraint on the provided row variable. If the provided
	 * variable is a chemical species, its constraint is set to [0, 0]. If the
	 * provided variable is not a chemical species, its constraint is set to 
	 * (-Infinity, Infinity).
	 * 
	 * @param rowVariable a row variable from the program matrix.
	 */
	public final void restoreRowConstraint(Object rowVariable) {

		// Checking input
		checkRowVariable(rowVariable);

		// Creating field for storage
		Interval original = null;

		// If it's a metabolie name
		if (isChemSpecies(rowVariable)) {

			// Constrain to steady-state
			original = off;
		}

		// Otherwise
		else {

			// Consider its original state as unbounded
			original = unbounded;
		}

		// Setting locally
		rowConstraints.put(rowVariable, original);

		// Setting in the program solver
		programSolver.setRowConstraint(rowVariable, original);
	}

	/**
	 * Restores the constraints for the provided column variables to those in
	 * the provided fluxome.
	 * 
	 * @param columnVariables a collection of column variables.
	 * 
	 * @see #restoreColumnConstraint(Object)
	 */
	public final void restoreColumnConstraints(Collection<?> columnVariables) {

		// Iterating thru each provided column variable
		for (Object col : columnVariables) {
			restoreColumnConstraint(col);
		}
	}

	/**
	 * Restores the constraints for the provided reaction names to those in the
	 * provided fluxome.
	 * 
	 * @param reactionNames a collection of reaction names.
	 * 
	 * @see #restoreFluxConstraint(String)
	 */
	public final void restoreFluxConstraints(Collection<String> reactionNames) {

		// Iterating thru each provided reaction name
		for (String name : reactionNames) {
			restoreFluxConstraint(name);
		}
	}

	/**
	 * Restores the constraints for the provided reaction name expressions to
	 * those in the provided fluxome.
	 * 
	 * @param rxnNameExprs a collection of mathematical expressions of reaction
	 * names.
	 * 
	 * @see #restoreRxnExprConstraint(MathExpr)
	 */
	public final void restoreRxnExprConstraints(
			Collection<? extends MathExpr<String>> rxnNameExprs) {

		// Iterating thru each provided reaction name expression
		for (MathExpr<String> e : rxnNameExprs) {
			restoreRxnExprConstraint(e);
		}

	}

	/**
	 * Restores the constraints for the provided column variable expressions to
	 * those in the provided fluxome.
	 * 
	 * @param colVarExprs a collection of mathematical expressions of column
	 * variables.
	 * 
	 * @see #restoreColExprConstraint(MathExpr)
	 */
	public final void restoreColExprConstraints(
			Collection<? extends MathExpr<?>> colVarExprs) {

		// Iterating thru each provided reaction name expression
		for (MathExpr<?> e : colVarExprs) {
			restoreColExprConstraint(e);
		}

	}

	/**
	 * Restores the constraints for the provided row variables to those in the
	 * provided fluxome.
	 * 
	 * @param rowVariables a collection of row variables.
	 * 
	 * @see #restoreRowConstraint(Object)
	 */
	public final void restoreRowConstraints(Collection<?> rowVariables) {

		// Iterating thru each provided row variable
		for (Object rowVar : rowVariables) {
			restoreRowConstraint(rowVar);
		}
	}

	/**
	 * Restores the constraints on all column variables to those contained 
	 * in the fluxome used by this optimization computer.
	 * 
	 * @see #restoreColumnConstraint(Object)
	 */
	public final void restoreColumnConstraints() {

		// Getting all column variables
		Set<?> columnVariables = getColumnVariables();

		// Restoring the constraints on all column variables
		restoreColumnConstraints(columnVariables);
	}

	/**
	 * Restores the flux constraints for all reactions to those contained 
	 * in the fluxome used by this optimization computer.
	 * 
	 * @see #restoreFluxConstraint(String)
	 */
	public final void restoreFluxConstraints() {

		// Getting all reaction names
		Set<String> rxnNames = fluxome.getRxnNames();

		// Restoring the constraints on all reaction names
		restoreFluxConstraints(rxnNames);
	}

	/**
	 * Restores the constraints on all mathematical expressions of reaction
	 * names.
	 * 
	 * @see #restoreRxnExprConstraint(MathExpr)
	 */
	public final void restoreRxnExprConstraints() {

		// Getting all expressions
		Set<MathExpr<String>> expressions = getRxnExprs();

		// Restoring the constraints on all column variable expressions
		restoreRxnExprConstraints(expressions);
	}

	/**
	 * Restores the constraints on all mathematical expressions of column
	 * variables.
	 * 
	 * @see #restoreColExprConstraint(MathExpr)
	 */ 
	public final void restoreColExprConstraints() {

		// Getting all expressions
		Set<MathExpr<?>> expressions = getColExprs();

		// Restoring the constraints on all column variable expressions
		restoreColExprConstraints(expressions);
	}

	/**
	 * Restores the constraints on all row variables to those contained in 
	 * the fluxome used by this optimization computer.
	 * 
	 * @see #restoreRowConstraint(Object)
	 */
	public final void restoreRowConstraints() {

		// Getting all row variables
		Set<?> rowVariables = getRowVariables();

		// Restoring the constraints on all row variables
		restoreRowConstraints(rowVariables);
	}

	/**
	 * Returns the set of all column variables in the program matrix.
	 * @return the set of all column variables in the program matrix.
	 */
	public Set<?> getColumnVariables() {

		// Determining appropriate size
		int size = columnConstraints.size() + 1;

		// Creating storage
		Set<Object> s = new LinkedHashSet<Object>(size, 1.0F);

		// Copying
		s.addAll(columnConstraints.keySet());

		return s;
	}

	/**
	 * Returns the set of all mathematical expressions of reaction names.
	 * @return the set of all mathematical expressions of reaction names.
	 */
	@SuppressWarnings("unchecked")
	public Set<MathExpr<String>> getRxnExprs() {

		// Determining appropriate size
		int size = exprConstraints.size() + 1;

		// Creating storage
		Set<MathExpr<String>> s = 
			new LinkedHashSet<MathExpr<String>>(size, 1.0F);

		// Iterating thru all column variable expressions
		for (MathExpr<?> expr : exprConstraints.keySet()) {
			
			// If it's a reaction name expression
			if (isRxnNameExpr(expr)) {
				
				// Safely casting
				MathExpr<String> rxnExpr = (MathExpr<String>) expr;
				 
				// Adding to the rest
				s.add(rxnExpr);
			}
			
		}

		return s;
	}

	/**
	 * Returns the set of all mathematical expressions of column variables.
	 * @return the set of all mathematical expressions of column variables.
	 */
	public Set<MathExpr<?>> getColExprs() {

		// Determining appropriate size
		int size = exprConstraints.size() + 1;

		// Creating storage
		Set<MathExpr<?>> s = 
			new LinkedHashSet<MathExpr<?>>(size, 1.0F);

		// Copying
		s.addAll(exprConstraints.keySet());

		return s;
	}

	/**
	 * Returns the set of all row variables in the program matrix.
	 * @return the set of all row variables in the program matrix.
	 */
	public Set<?> getRowVariables() {

		// Determining appropriate size
		int size = rowConstraints.size() + 1;

		// Creating storage
		Set<Object> s = new LinkedHashSet<Object>(size, 1.0F);

		// Copying
		s.addAll(rowConstraints.keySet());

		return s;
	}

    /**
     * Returns the current constraint on the provided column variable.
     * 
     * @param columnVariable the column variable whose constraint
     * is to be returned.
     * 
     * @return the current constraint on the provided column variable.
     */
	public Interval getColumnConstraint(Object columnVariable) {

		// Checking input
		checkColumnVariable(columnVariable);

		return columnConstraints.get(columnVariable);
	}

    /**
     * Returns the current flux constraint on the reaction with the provided
     * name.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be returned.
     * 
     * @return the current flux constraint on the reaction with the provided
     * name. 
     */
	public Interval getFluxConstraint(String reactionName) {

		// Checking input
		checkRxnName(reactionName);

		return columnConstraints.get(reactionName);
	}

    /**
     * Returns the current constraint on the provided expression.
     * 
     * @param rxnNameExpr the mathematical expression of reaction names
     * whose constraint is to be returned.
     * 
     * @return the current constraint on the provided expression.
     */
	public Interval getRxnExprConstraint(MathExpr<String> rxnNameExpr) {

		// Checking input
		checkRxnExpr(rxnNameExpr);

		// Getting the stored constraint, if it exists
		Interval c = exprConstraints.get(rxnNameExpr);

		return (c == null ? unbounded : c);
	}

    /**
     * Returns the current constraint on the provided expression.
     * 
     * @param columnVariableExpr the mathematical expression of column 
     * variables whose constraint is to be returned.
     * 
     * @return the current constraint on the provided expression.
     */
	public Interval getColExprConstraint(MathExpr<?> columnVariableExpr) {

		// Checking input
		checkColExpr(columnVariableExpr);

		// Getting the stored constraint, if it exists
		Interval c = exprConstraints.get(columnVariableExpr);

		return (c == null ? unbounded : c);
	}

    /**
     * Returns the current constraint on the provided row variable.
     * 
     * @param rowVariable the row variable whose constraint
     * is to be returned.
     * 
     * @return the current constraint on the provided row variable.
     */
	public Interval getRowConstraint(Object rowVariable) {

		// Checking input
		checkRowVariable(rowVariable);

		return rowConstraints.get(rowVariable);
	}

	/**
	 * Returns the current constraints for each column variable in
	 * the program matrix.
	 * 
	 * @return each column variable in the program matrix mapped to
	 * its current constraint.
	 */
	public Map<?, Interval> getColumnConstraints() {

		// Determining appropriate size
		int size = columnConstraints.size() + 1;

		// Creating storage
		Map<Object, Interval> m = 
			new LinkedHashMap<Object, Interval>(size, 1.0F);

		// Copying
		m.putAll(columnConstraints);

		return m;
	}

	/**
	 * Returns all current flux constraints for each reaction
	 * in the fluxome used by this optimization computer.
	 * 
	 * @return the names of all reactions mapped to their current flux
	 * constraint.
	 */
	@SuppressWarnings("unchecked")
	public Map<String, Interval> getFluxConstraints() {

		// Getting all column constraints
		Map<?, Interval> m = getColumnConstraints();

		// Retaining only reaction names
		m.keySet().retainAll(fluxome.getRxnNames());

		// Casting safely
		return (Map<String, Interval>) m;
	}

	/**
	 * Returns all current constraints on mathematical expressions
	 * of column variables.
	 * 
	 * @return all mathematical expressions of column variables contained in
	 * this computer mapped to their current constaint.
	 */
	public Map<MathExpr<?>, Interval> getColExprConstraints() {

		// Determining appropriate size
		int size = exprConstraints.size() + 1;

		// Creating storage for a defensive copy
		Map<MathExpr<?>, Interval> m = 
			new LinkedHashMap<MathExpr<?>, Interval>(size, 1.0F);

		// Copying
		m.putAll(exprConstraints);

		return m;
	}

	/**
	 * Returns all current constraints on mathematical expressions
	 * of reaction names.
	 * 
	 * @return all mathematical expressions of reaction names contained in 
	 * this computer mapped to their current constraint. 
	 */
	@SuppressWarnings("unchecked")
	public Map<MathExpr<String>, Interval> getRxnExprConstraints() {
    	
		// Determining maximum size
		int size = exprConstraints.size() + 1;
		
		// Creating storage
		Map<MathExpr<String>, Interval> m =
			new LinkedHashMap<MathExpr<String>, Interval>(size, 1.0F);
		
		// Iterating thru each column variable expression
		for (Map.Entry<MathExpr<?>, Interval> e : exprConstraints.entrySet()) {
			
			// Getting the expression
			MathExpr<?> expr = e.getKey();
			
			// If it's a reaction name expression
			if (isRxnNameExpr(expr)) {
				
				// Safely casting
				MathExpr<String> rxnExpr = (MathExpr<String>) expr;
				
				// Safely casting and adding
				m.put(rxnExpr, e.getValue());
			}
		}
		
		return m;
    }

	/**
	 * Returns all the current constraints on row variables.
	 * 
	 * @return all row variables mapped to their current constraint.
	 */
	public Map<?, Interval> getRowConstraints() {

		// Determining appropriate size
		int size = rowConstraints.size() + 1;

		// Creating storage
		Map<Object, Interval> m = 
			new LinkedHashMap<Object, Interval>(size, 1.0F);

		// Copying
		m.putAll(rowConstraints);

		return m;
	}

	/**
	 * Returns the optimal value of the current objective function using the
	 * current optimization sense.
	 */
	private final Results getOptimal() {

		// Ensuring objective function & optimization sense were set
		checkState();

		// Computing the objective value
		if (!programSolver.optimize()) {
			throw new FbaOptException(objectiveFunction, sense,
					programSolver.getStatus());
		}

		// Setting global variables
		double objectiveValue = programSolver.getObjectiveValue();
		Map<Object, Double> solutionVector = programSolver.getSolutionVector();

		// Ensuring solution does not violate constraints
		switch (safetyLevel) {
		case 0:
			break; // do nothing
		case 1:
			checkSolutionVector(solutionVector);
			checkExprValues(solutionVector);
			break;
		default:
		    break; // do nothing
		}

		// Returning computed objective value
		return new Results(objectiveValue, solutionVector);
	}

	/**
	 * Returns the results of computing the optimal flux of the reaction 
	 * with the provided name.
	 * 
	 * @param reactionName the name of the reaction whose flux is to be
	 * optimized.
	 * @param sense the sense, or direction, of optimization.
	 * 
	 * @return the results of computing the optimal flux of the reaction 
	 * with the provided name.
	 */
	public final Results 
	getOptimal(String reactionName, OptSense sense) {

		// Converting to a reaction expression
		MathExpr<String> obj = getRxnExpr(reactionName);

		return getOptimal(obj, sense);
	}

	/**
	 * Returns the results of computing the optimal value of the 
	 * provided objective function.
	 * 
	 * @param objective a mathematical expression of column variables.
	 * @param sense the sense, or direction, of optimization.
	 * 
	 * @return the results of computing the optimal value of the 
	 * provided objective function.
	 */
	public final Results 
	getOptimal(MathExpr<?> objective, OptSense sense) {

		setObjectiveFunction(objective);
		setOptimizationSense(sense);

		return getOptimal();
	}

	/**
	 * Returns the results of computing the maximum value of the 
	 * provided objective function.
	 * 
	 * @param objective a mathematical expression of column variables.
	 * 
	 * @return the results of computing the maximum value of the 
	 * provided objective function.
	 */
	public final Results getMax(MathExpr<?> objective) {
		return getOptimal(objective, OptSense.MAX);
	}

	/**
	 * Returns the results of computing the minimum value of the 
	 * provided objective function.
	 * 
	 * @param objective a mathematical expression of column variables.
	 * 
	 * @return the results of computing the minimum value of the 
	 * provided objective function.
	 */
	public final Results getMin(MathExpr<?> objective) {
		return getOptimal(objective, OptSense.MIN);
	}

	/**
	 * Computes the maximum value of the provided objective function.
	 * 
	 * @param objective a mathematical expression
	 * whose variables are column variables in the program matrix.
	 * 
	 * @return the maximum value the provided mathematical
	 * expression can assume.
	 * 
	 * @throws NullPointerException if the provided expression is null.
	 * 
	 * @throws FbaOptException if an optimal solution cannot be found.
	 * 
	 * @throws IllegalArgumentException if the provided expression
	 * contains a variable that is not a column variable
	 * in the program matrix.
	 */
	public double getMaximum(MathExpr<?> objective) {

		OptSense sense = OptSense.MAX;
		Results r = getOptimal(objective, sense);
		
		return r.getObjectiveValue();
	}

	/**
	 * Computes the maximum flux of the reaction with the provided name.
	 * 
	 * @param reactionName the name of the reaction whose
	 * maximum flux will be computed.
	 * 
	 * @return the maximum flux of the reaction with the provided name.
	 * 
	 * @throws IllegalArgumentException if the provided string
	 * is not a reaction name of the fluxome used by this optimization 
	 * computer.
	 */
	public double getFluxMax(String reactionName) {

		// Converting to a reaction expression
		MathExpr<String> obj = getRxnExpr(reactionName);

		return getMaximum(obj);
	}

	/**
	 * Computes the maximum value of the provided column variable.
	 * 
	 * @param columnVariable a column variable from the program
	 * matrix whose maximum value is to be computed.
	 * 
	 * @return the maximum value the provided column variable
	 * can assume.
	 * 
	 * @throws IllegalArgumentException if the provided object
	 * is not a column variable in the program matrix.
	 */
	public double getColumnMax(Object columnVariable) {
		
		// Converting to column variable expression
		MathExpr<?> obj = getExpr(columnVariable);

		return getMaximum(obj);
	}

	/**
	 * Computes the maximum value of each provided objective function.
	 * 
	 * @param <V> the variable type.
	 * @param <E> the mathematical expression type.
	 * 
	 * @param objectives a collection of mathematical
	 * expressions whose variables are column variables in the program
	 * matrix.
	 * 
	 * @return the provided mathematical expressions mapped to their
	 * computed maximum value.
	 * 
	 * @throws NullPointerException if the provided collection, or any
	 * of its elements, is {@code null}.
	 * 
	 * @throws IllegalArgumentException if one of the provided
	 * mathematical expressions contains a variable that is not a 
	 * column variable. 
	 */
	public <V, E extends MathExpr<V>> Map<E, Double>
	getMaxima(Collection<E> objectives) {

		// Determining appropriate size
		int size = objectives.size() + 1;

		// Creating storage
		Map<E, Double> maxima = new LinkedHashMap<E, Double>(size, 1.0F);

		// Iterating thru each expression
		for (E expr : objectives) {

			// Computing the maximum
			double max = getMaximum(expr);

			// Storing
			maxima.put(expr, max);
		}

		return maxima;
	}

	/**
	 * Computes the maximum flux of each reaction with the provided names.
	 * 
	 * @param reactionNames a collection of reaction names from
	 * the fluxome used by this optimization computer.
	 * 
	 * @return the provided reaction names mapped to their
	 * computed maximum flux.
	 * 
	 * @throws NullPointerException if the provided collection
	 * is {@code null}.
	 * 
	 * @throws IllegalArgumentException if an element in the provided
	 * collection is not a reaction name.
	 */
	public Map<String, Double> getFluxMaxima(Collection<String> reactionNames) {

		// Determining appropriate size
		int size = reactionNames.size() + 1;

		// Creating storage
		Map<String, Double> maxima = 
			new LinkedHashMap<String, Double>(size, 1.0F);

		// Iterating thru each rxn name
		for (String rxnName : reactionNames) {

			// Computing the maximum
			double max = getFluxMax(rxnName);

			// Storing
			maxima.put(rxnName, max);
		}

		return maxima;
	}

	/**
	 * Computes the maximum values of each provided column variable.
	 * 
	 * @param <T> the column variable type.
	 * 
	 * @param columnVariables a collection of column variables in
	 * the program matrix.
	 * 
	 * @return the provided column variables mapped to their 
	 * computed maximum value.
	 * 
	 * @throws IllegalArgumentException if an element in the provided
	 * collection is not a legitimate column variable.
	 */
	public <T> Map<T, Double> getColumnMaxima(Collection<T> columnVariables) {

		// Determining appropriate size
		int size = columnVariables.size() + 1;

		// Creating storage
		Map<T, Double> maxima = new LinkedHashMap<T, Double>(size, 1.0F);

		// Iterating thru each provided column variable
		for (T col : columnVariables) {

			// Computing maximum
			double max = getColumnMax(col);

			// Storing
			maxima.put(col, max);
		}

		return maxima;
	}

	/**
	 * Computes the minimum value of the provided objective function.
	 * 
	 * @param objective a mathematical expression
	 * whose variables are column variables in the program matrix.
	 * 
	 * @return the minimum value of the provided objective function.
	 * 
	 * @throws IllegalArgumentException if a variable in the
	 * provided expression is not a column variable
	 * in the program matrix.
	 */
	public double getMinimum(MathExpr<?> objective) {

		OptSense sense = OptSense.MIN;
		Results r = getOptimal(objective, sense);

		return r.getObjectiveValue();
	}

	/**
	 * Computes the minimum flux of the reaction with the provided name.
	 * 
	 * @param reactionName a reaction name from the fluxome
	 * used by this optimization computer.
	 * 
	 * @return the minimum flux of the reaction with the provided name.
	 * 
	 * @throws IllegalArgumentException if the provided string
	 * is not a reaction name.
	 */
	public double getFluxMin(String reactionName) {

		// Converting to a reaction expression
		MathExpr<String> obj = getRxnExpr(reactionName);

		return getMinimum(obj);
	}

	/**
	 * Computes the minimum value of the provided column variable.
	 * 
	 * @param columnVariable a column variable from the program
	 * matrix.
	 * 
	 * @return the computed minimum value of the provided column 
	 * variable.
	 * 
	 * @throws IllegalArgumentException if the provided object
	 * is not a column variable in the program matrix.
	 */
	public double getColumnMin(Object columnVariable) {

		// Converting to a column variable expression
		MathExpr<?> obj = getExpr(columnVariable);
		
		return getMinimum(obj);
	}

	/**
	 * Computes the minimum value of each provided objective function.
	 * 
	 * @param <V> the variable type.
	 * @param <E> the mathematical expression type.
	 * 
	 * @param objectives a collection of mathematical
	 * expressions whose variables are column variables in the program
	 * matrix.
	 * 
	 * @return the provided mathematical expressions mapped to their
	 * computed minimum value.
	 * 
	 * @throws NullPointerException if the provided collection, or any
	 * of its elements, is {@code null}.
	 * 
	 * @throws IllegalArgumentException if one of the provided
	 * mathematical expressions contains a variable that is not a 
	 * column variable. 
	 */
	public <V, E extends MathExpr<V>> Map<E, Double> 
	getMinima(Collection<E> objectives) {

		// Determining appropriate size
		int size = objectives.size() + 1;

		// Creating storage
		Map<E, Double> minima = new LinkedHashMap<E, Double>(size, 1.0F);

		// Iterating thru each column variable expression
		for (E expr : objectives) {

			// Computing the minimum
			double max = getMinimum(expr);

			// Storing
			minima.put(expr, max);
		}

		return minima;
	}

	/**
	 * Computes the minimum flux of each reaction with the provided names.
	 * 
	 * @param reactionNames a collection of reaction names from
	 * the fluxome used by this optimization computer.
	 * 
	 * @return the provided reaction names mapped to their
	 * computed minimum flux.
	 * 
	 * @throws NullPointerException if the provided collection
	 * is {@code null}.
	 * 
	 * @throws IllegalArgumentException if an element in the provided
	 * collection is not a reaction name.
	 */
	public Map<String, Double> getFluxMinima(Collection<String> reactionNames) {

		// Determining appropriate size
		int size = reactionNames.size() + 1;

		// Creating storage
		Map<String, Double> minima = 
			new LinkedHashMap<String, Double>(size, 1.0F);

		// Iterating thru each reaction name
		for (String rxnName : reactionNames) {

			// Computing the minimum
			double min = getFluxMin(rxnName);

			// Storing
			minima.put(rxnName, min);
		}

		return minima;
	}

	/**
	 * Computes the minimum values of each provided column variable.
	 * 
	 * @param <T> the column variable type.
	 * 
	 * @param columnVariables a collection of column variables in
	 * the program matrix.
	 * 
	 * @return the provided column variables mapped to their 
	 * computed minimum value.
	 * 
	 * @throws IllegalArgumentException if an element in the provided
	 * collection is not a legitimate column variable.
	 */
	public <T> Map<T, Double> getColumnMinima(Collection<T> columnVariables) {

		// Determining appropriate size
		int size = columnVariables.size() + 1;

		// Creating storage
		Map<T, Double> minima = new LinkedHashMap<T, Double>(size, 1.0F);

		// Iterating thru each column variable
		for (T col : columnVariables) {

			// Computing minimum
			double min = getColumnMin(col);

			// Storing
			minima.put(col, min);
		}

		return minima;
	}

	/**
	 * Initializes the values in the program solver.
	 */
	private final void defineProgramSolver() {

		// Clearing any existing state the program solver may have
		programSolver.clear();

		// Setting program matrix
		setStoichMatrix();
		
		// Setting row constraints
		programSolver.setRowConstraints(rowConstraints);

		// Setting column constraints
		programSolver.setColumnConstraints(columnConstraints);

		// Setting additional column variable expression constraints
		programSolver.setConstraints(exprConstraints);

	}

	/**
	 * Sets the program matrix equal to the stoichiometry matrix
	 * of the fluxome.
	 */
	private void setStoichMatrix() {

		// Getting S (i.e. the stiochiometry matrix)
		SparseDoubleMatrix<String, String> s = fluxome.getStoichMatrix();

		//*** Setting program matrix ***

		// Iterating thru each reaction name
		for (String rxn : s.getColumnIndices()) {

			// Getting row variables (i.e. chemical species) with
			// nonzero elements in this column
			Set<String> rowIndices = s.getRowIndNzEl(rxn);

			// Iterating thru these indices
			for (String met : rowIndices) {

				// Getting element from S
				double value = s.getElement(met, rxn);

				// Setting element in program matrix
				programSolver.setMatrixElement(met, rxn, value);
			}

		}

	}

	/**
	 * Ensures that an objective function and optimization sense have been
	 * provided.
	 */
	private void checkState() {

		// Checking objective function
		if (objectiveFunction == null) {
			String s = "No objective function was provided.";
			throw new IllegalStateException(s);
		}

		// Checking optimization sense
		if (sense == null) {
			String s = "No optimization sense was provided.";
			throw new IllegalStateException(s);
		}

	}

	/**
	 * Indicates if the provided row variable is a valid chemical species.
	 */
	private boolean isChemSpecies(Object rowVariable) {

		if (!(rowVariable instanceof String)) {
			return false;
		}

		return fluxome.isSpecies((String) rowVariable);
	}

	/**
	 * Indicates if the provided column variable is a valid reaction name.
	 */
	private boolean isRxnName(Object columnVariable) {

		if (!(columnVariable instanceof String)) {
			return false;
		}

		return fluxome.isRxnName((String) columnVariable);
	}

	/**
	 * Indicates if all the variables in the provided expression are valid
	 * reaction names.
	 */
	private boolean isRxnNameExpr(MathExpr<?> expression) {

		// Iterating thru each variable
		for (Object var : expression.getVariables()) {

			if (!isRxnName(var)) {
				return false;
			}
		}

		return true;
	}

	/**
	 * Ensures the provided object is row variable.
	 * @param obj the object to check.
	 * @throws IllegalArgumentException if the provided object is not a row
	 * variable in the program matrix.
	 */
	protected void checkRowVariable(Object obj) {

		// If the row variable is present
		if (rowConstraints.containsKey(obj)) {
			return;
		}

		// Throwing IllegalArgumentException if necessary
		String s = "The row index '" + obj
				+ "' does not exist in the program matrix.";
		throw new IllegalArgumentException(s);
	}

	/**
	 * Ensures the provided object is column variable.
	 * @param obj the object to check.
	 * @throws IllegalArgumentException if the provided object is not a column
	 * variable in the program matrix.
	 */
	protected void checkColumnVariable(Object obj) {

		// If the column variable is present
		if (columnConstraints.containsKey(obj)) {
			return;
		}

		// Throwing IllegalArgumentException if necessary
		String s = "The column index '" + obj
				+ "' does not exist in the program matrix.";
		throw new IllegalArgumentException(s);
	}

	/**
	 * Ensures each element of the provided collection is column variable.
	 * @param c the collection of objects to check.
	 * @throws IllegalArgumentException if an element in the provided 
	 * collection is not a column variable in the program matrix.
	 */
	protected void checkColumnVariables(Collection<?> c) {

		// Iterating thru provided column variables
		for (Object col : c) {
			checkColumnVariable(col);
		}
	}

	/**
	 * Ensures all the variables in the provided expression are valid column
	 * variables.
	 * @param e the expression to check.
	 * @throws IllegalArgumentException if a variable in the provided 
	 * expression is not a column variable in the program matrix.
	 */
	protected void checkColExpr(MathExpr<?> e) {
		checkColumnVariables(e.getVariables());
	}

	/**
	 * Ensures the provided string is the name of a reaction in the fluxome 
	 * used by this optimization computer.
	 * @param string the string to check.
	 * @throws IllegalArgumentException if the provided string
	 * is not the name of a reaction in the fluxome used by this optimization
	 * computer.
	 */
	protected void checkRxnName(String string) {

		// If it's present in the fluxome
		if (fluxome.isRxnName(string)) {
			return;
		}
		
		// Throwing IllegalArgumentException if necesssary
		String s = "The reaction '" + string
					+ "' does not exist in the provided fluxome.";
		throw new IllegalArgumentException(s);
	}

	/**
	 * Ensures each element of the provided collection is reaction name.
	 * @param c the collection of strings to check.
	 * @throws IllegalArgumentException if a string in the provided 
	 * collection is not the name of a reaction in the fluxome used by this
	 * optimization computer.
	 */
	protected void checkRxnNames(Collection<String> c) {

		// Iterating thru each potential reaction name
		for (String rxn: c) {
			checkRxnName(rxn);
		}
	}

	/**
	 * Ensures all the variables in the provided expression are reaction names.
	 * @param e the expression to check.
	 * @throws IllegalArgumentException if a variable in the provided 
	 * expression is not the name of a reaction in the fluxome used by this
	 * optimization computer.
	 */
	protected void checkRxnExpr(MathExpr<String> e) {
		checkRxnNames(e.getVariables());
	}

	/**
	 * Ensures the provided number is finite and not NaN.
	 * @param number the number to check.
	 * @throws IllegalArgumentException if the provided number is infinite
	 * or NaN.
	 */
	protected void checkNumber(double number) {

		if (Double.isInfinite(number) || Double.isNaN(number)) {
			String s = "Invalid number: " + number;
			throw new IllegalArgumentException(s);
		}

	}

	/**
	 * Ensures the provided string is the name of a chemical species in the fluxome 
	 * used by this optimization computer.
	 * @param string the string to check.
	 * @throws IllegalArgumentException if the provided string
	 * is not the name of a chemical species in the fluxome used by this optimization
	 * computer.
	 */
	protected void checkChemSpecies(String string) {

		// If the species is valid
		if (fluxome.isSpecies(string)) {
			return;
		}
		
		// Throwing IllegalArgumentException if necesssary
		String s = "The chemical species '" + string
					+ "' does not exist in the fluxome provided.";
		throw new IllegalArgumentException(s);
	}

	/**
	 * Ensures the provided constraint is valid for reaction name-expressions 
	 * for the fluxome used by this optimization computer.
	 * 
	 * @param c the constraint to check.
	 * 
	 * @throws NullPointerException if the provided constraint is {@code null}.
	 */
	protected void checkRxnExprConstraint(Interval c) {

		// Throwing NullPointerException if necesssary
		Util.nullCheck(c);
	}

	/**
	 * Ensures the provided constraint is a valid flux constraint
	 * in the fluxome used by this optimization computer.
	 * 
	 * @param c the constraint to check.
	 * @param rxnName the name of the reaction for which this constraint
	 * applies.
	 * @param tolerance the allowable amount by which the bounds of the
	 * provided constraint can violate the maximum bounds defined the
	 * fluxome used by this optimization computer.
	 * 
	 * @throws IllegalArgumentException if the provided 
	 * constraint is not a valid flux constraint.
	 */
	protected void 
	checkFluxConstraint(Interval c, String rxnName, double tolerance) {

		// If the flux constraint is valid
		if (fluxome.isValid(c, rxnName, tolerance)) {
			return;
		}
		
		// Throwing IllegalArgumentException if necessary
		String s = "The constraint '" + c + "' is illegal for the reaction: " 
					+ rxnName + ".";
		throw new IllegalArgumentException(s);
	}


	/**
	 * Converts a column variable to a single-term linear combination.
	 * 
	 * @param <V> the column variable type.
	 * 
	 * @param columnVariable a column variable from the 
	 * program matrix.
	 * 
	 * @return a single-term linear combination containing the 
	 * provided column variable. The coefficient of this term 
	 * equals 1, and the constant equals 0.
	 * 
	 * @throws IllegalArgumentException if the provided object
	 * is not a column variable in the program matrix.
	 */
	protected <V> LinearComb<V> getExpr(V columnVariable) {

		// Checking input
		checkColumnVariable(columnVariable);

		// Returning single-term linear combination
		return HashLinearComb.singleTerm(columnVariable);
	}

	/**
	 * Converts a reaction name to a single-term linear combination.
	 * 
	 * @param reactionName a reaction name from the fluxome
	 * used to create this optimization computer.
	 * 
	 * @return a single-term linear combination containing the 
	 * provided reaction name. The coefficient of this term 
	 * equals 1, and the constant equals 0.
	 * 
	 * @throws IllegalArgumentException if the provided string
	 * is not a reaction name in the fluxome used by
	 * this optimization computer.
	 */
	protected LinearComb<String> getRxnExpr(String reactionName) {

		// Checking input
		checkRxnName(reactionName);

		// Returning single-term linear combination
		return HashLinearComb.singleTerm(reactionName);
	}

	/**
	 * Converts a collection of reaction names to a linear combination.
	 * 
	 * @param rxnNames a collection of reaction names from the fluxome
	 * used by this optimization computer.
	 * 
	 * @return a linear combination containing the provided reaction
	 * names. The coefficient of each term in the expression equals 1,
	 * and the constant equals 0.
	 * 
	 * @throws IllegalArgumentException if an element in the provided
	 * collection is not a reaction name in the fluxome used by
	 * this optimization computer.
	 */
	protected LinearComb<String> getRxnExpr(Collection<String> rxnNames) {
		
		// Checking reaction names
		checkRxnNames(rxnNames);
		
		return getExpr(rxnNames);
	}
	
	/**
	 * Converts a collection of column variables to a linear combination.
	 * 
	 * @param <V> the column variable type.
	 * 
	 * @param columnVariables a collection of column variables contained
	 * in the program matrix.
	 * 
	 * @return a linear combination containing the provided column
	 * variables. The coefficient of each term in the expression equals 1,
	 * and the constant equals 0.
	 * 
	 * @throws IllegalArgumentException if an element in the provided
	 * collection is not a column variable in the program matrix.
	 */
	protected <V> LinearComb<V> getExpr(Collection<V> columnVariables) {

		// Creating storage for terms
		List<LinearComb.Term<V>> terms = new LinkedList<LinearComb.Term<V>>();

		// Iterating thru each variable
		for (V col : columnVariables) {

			// Checking variable
			checkColumnVariable(col);

			// Creating term and adding to the rest
			terms.add(new HashLinearComb.Term<V>(col));

		}

		return new HashLinearComb<V>(terms, 0, 0.0);
	}

	/**
	 * Ensures the provided solution vector does not violate the defined
	 * constraints of the system.
	 * 
	 * @param solutionVector column variables mapped to their respective
	 * numerical values.
	 * 
	 * @throws ConstraintViolationException if a value in the
	 * provided vector violates its defined constraints 
	 * by more than the defined constraint tolerance.
	 * 
	 * @throws IllegalArgumentException if a key in the provided
	 * map is not a legitimate column variable in the program matrix.
	 * 
	 * @throws NullPointerException if a value in the provided
	 * map is {@code null}.
	 */
	protected void checkSolutionVector(Map<?, Double> solutionVector) {

		// Iterating thru each column variable-value pair
		for (Map.Entry<?, Double> e : solutionVector.entrySet()) {
			checkColumnValue(e.getKey(), e.getValue());
		}

	}

	/**
	 * Ensures the provided numerical value does not violate the defined 
	 * constraint of the provided column variable.
	 * 
	 * @param columnVariable the column variable whose value and 
	 * constraint will be compared.
	 * 
	 * @param value the numerical value of the provided column
	 * variable.
	 * 
	 * @throws ConstraintViolationException if the provided value
	 * violates the defined constraint for the provided column variable
	 * by more than the defined constraint tolerance.
	 * 
	 * @throws IllegalArgumentException if the provided
	 * column variable does not exist in the program matrix.
	 * 
	 * @throws NullPointerException if the provided value is {@code null}.
	 */
	protected void checkColumnValue(Object columnVariable, Double value) {
		checkColumnValue(columnVariable, value.doubleValue());
	}

	/**
	 * Ensures the provided numerical value does not violate the defined 
	 * constraint of the provided column variable.
	 * 
	 * @param columnVariable the column variable whose value and 
	 * constraint will be compared.
	 * 
	 * @param value the numerical value of the provided column
	 * variable.
	 * 
	 * @throws ConstraintViolationException if the provided value
	 * violates the defined constraint for the provided column variable
	 * by more than the defined constraint tolerance.
	 * 
	 * @throws IllegalArgumentException if the provided
	 * column variable does not exist in the program matrix.
	 */
	protected void checkColumnValue(Object columnVariable, double value) {
		Interval c = getColumnConstraint(columnVariable);
		checkValue(columnVariable, value, c);
	}

	/**
	 * Ensures the provided numerical value does not violate the provided 
	 * constraint of the provided column variable.
	 */
	private void checkValue(Object variable, double value, Interval bounds) {

		// Checking for large numerical error
		if (!bounds.isWithin(value, tolerance)) {
			throw new ConstraintViolationException(variable, bounds, value);
		}

	}

	/**
	 * Ensures the defined constraints on expressions are not violated.
	 */
	private void checkExprValues(Map<?, Double> solutionVector) {

		// Iterating thru each expression-constraint pair
		for (Map.Entry<MathExpr<?>, Interval> e : exprConstraints.entrySet()) {

			// Getting the pair
			MathExpr<?> expr = e.getKey();
			Interval bounds = e.getValue();

			// Plugging in flux vector
			MathExpr<?> pluggedIn = expr.plugInNumbers(solutionVector);

			// Checking constant
			checkValue(expr, pluggedIn.getConstant(), bounds);
		}

	}

	/**
	 * This class is used to store the results of an FBA optimization
	 * problem.
	 * 
	 * @author This class was written by Jeremiah Wright while in the Wagner lab.
	 *
	 */
	public class Results {

		private final double objectiveValue;

		private final Map<Object, Double> solutionVector;

		private Results(double objectiveValue, Map<?, Double> solutionVector) {

			// Setting objective value
			this.objectiveValue = objectiveValue;

			// Creating storage for a defensive copy
			int size = solutionVector.size() + 1;
			this.solutionVector = new LinkedHashMap<Object, Double>(size, 1.0F);

			// Setting solution vector
			this.solutionVector.putAll(solutionVector);
		}

		/**
		 * Returns the computed objective value.
		 * @return the computed objective value.
		 */
		public double getObjectiveValue() {
			return objectiveValue;
		}

		/**
		 * Returns the solution vector computed during optimization.
		 * @return the solution vector computed during optimization.
		 */
		public Map<?, Double> getSolutionVector() {

			// Determining appropriate size
			int size = solutionVector.size() + 1;

			// Creating storage
			Map<Object, Double> m = 
				new LinkedHashMap<Object, Double>(size, 1.0F);

			// Copying
			m.putAll(solutionVector);

			return m;
		}

		/**
		 * Returns the flux vector computed during optimization.
		 * @return the flux vector computed during optimization.
		 */
		@SuppressWarnings("unchecked")
		public Map<String, Double> getFluxVector() {

			// Getting solution vector
			Map<?, Double> m = getSolutionVector();

			// Retaining only fluxes
			m.keySet().retainAll(fluxome.getRxnNames());

			// Casting safely
			return (Map<String, Double>) m;
		}

	}

}
