/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.algebra.DoubleMatrix;
import sbrt.kernel.math.algebra.LinearSystemSolution;
import sbrt.kernel.math.algebra.LinearSystemSolver;

/**
 * This class is used to solve the system of linear equations <i>Sv = 0</i>.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 */
public class FbaSolver {

    /**
     * Solves <i>Sv = 0<i>, where <i>S</i> is the stoichiometry matrix of the provided
     * fluxome and <i>v</i> is the vector of fluxes.
     * 
     * @param fluxome the fluxome used to define Sv = 0.
     * @param solver the linear system solver to be used.
     * 
     * @return a solution to the system of linear equations represented as Sv = 0.
     */
    public static LinearSystemSolution<String> 
        solve(Fluxome fluxome, LinearSystemSolver<String> solver) {

        // Getting the stoichiometry matrix of the provided fluxome
        DoubleMatrix<String, String> s = fluxome.getStoichMatrix();
        
        // Converting it to a list of linear combinations
        List<LinearComb<String>> expressions = s.toLinearCombs();
        
        // Determing an appropriate size
        int size = expressions.size() + 1;
        
        // Creating storage for the system of equations
        Map<LinearComb<String>, Double> equations = 
            new HashMap<LinearComb<String>, Double>(size, 1f);

        // Creating a single zero object
        Double zero = Double.valueOf(0d);
        
        // Setting each linear combination of reaction names (fluxes) 
        // equal to zero, i.e. Sv = 0.
        for (LinearComb<String> expr : expressions) {
            equations.put(expr, zero);
        }
    
        // Computing and returning the solution
        return solver.solve(equations);
    }

}
