/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import sbrt.kernel.fba.IrrevRxn;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.RxnComparer;
import sbrt.kernel.math.HashLinearComb;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.algebra.DoubleMatrix;
import sbrt.kernel.math.comb.comp.SingleElementUnionsV2;
import sbrt.kernel.utilities.PercentProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;

/**
 * This class is used to compute <i>flux caps</i> for use in Mahadevan-Schilling
 * flux interval computations. A flux cap is a linear combination of fluxes
 * that can be used to cap, or bound, an unbound flux.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 */
public class FluxCapComputer {

    private final PercentProgressMonitor progress;
    private final IrrevRxnFluxome fluxome;
    private final Map<String, IrrevRxn> internalRxns;
    
    private final Collection<LinearComb<String>> cycles;
    private final Collection<Set<String>> equivRxns;
    private final Set<String> secondaryRxns;

    
    /**
     * Constructs a new flux cap computer from the cycles contained
     * in the provided fluxome.
     * 
     * @param fluxome the fluxome for which flux caps will be computed.
     * @param cycleMatrix the collection of all cycles in the 
     * provided fluxome.
     */
    public FluxCapComputer(IrrevRxnFluxome fluxome, 
    					   DoubleMatrix<?, String> cycleMatrix) {
        this(fluxome, cycleMatrix.toLinearCombs());
    }
    
    /**
     * Constructs a new flux cap computer from the cycles contained
     * in the provided fluxome.
     * 
     * @param fluxome the fluxome for which flux caps will be computed.
     * @param allCycles the collection of all cycles in the 
     * provided fluxome as linear combinations.
     */
    public FluxCapComputer(IrrevRxnFluxome fluxome, 
    		               Collection<LinearComb<String>> allCycles) {

        // Setting fluxome
        internalRxns = fluxome.getInternalRxns();
        this.fluxome = fluxome;

        // Getting all equivalent reactions in this fluxome
        equivRxns = RxnComparer.getEquivIrrevRxnIds(fluxome.getInternalRxns());

        // Creating storage for "secondary" reactions
        secondaryRxns = new TreeSet<String>();

        // Gathering secondary reactions
        for (Set<String> rxnNames : equivRxns) {

            // Getting an iterator for this set
            Iterator<String> j = rxnNames.iterator();

            // Ignoring the first reaction name
            j.next();

            // Adding all the rest
            while (j.hasNext()) {
                secondaryRxns.add(j.next());
            }
        }

        // Setting cycles
        int size = allCycles.size() + 1;
        cycles = new LinkedHashSet<LinearComb<String>>(size, 1.0F);
        cycles.addAll(allCycles);

        // Initializing progress monitor
        progress = new PercentProgressMonitor();
    }

    /**
     * Returns a progress monitor for this flux cap computer. 
     * @return a progress monitor for this flux cap computer. 
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    /**
     * Returns the fluxome for which flux caps are computed.
     * @return the fluxome for which flux caps are computed.
     */
    public IrrevRxnFluxome getFluxome() {
        return fluxome;
    }

    /**
     * Computes and returns all flux caps.
     * @return reaction names mappes to their corresponding set of flux caps.
     */
    public Map<String, Set<LinearComb<String>>> getFluxCaps() {

        // Starting timer
        progress.start();
        progress.resetCount();

        // Creating caps from cycles
        Map<String, Set<LinearComb<String>>> fluxCaps = makeFluxCaps();

        // Stopping timer and setting progress to 100%
        progress.complete();
        progress.stop();

        return fluxCaps;
    }

    
    
    
    private Map<String, Set<LinearComb<String>>> makeFluxCaps() {

        // Collecting all the cycles a reaction participates in
        Map<String, List<LinearComb<String>>> participation = 
        	getCycleParticipation(cycles);

        // Setting total for progress monitor
        progress.setTotal(participation.size());

        // Creating storage for flux caps
        Map<String, Set<LinearComb<String>>> fluxCaps = 
        	new TreeMap<String, Set<LinearComb<String>>>();

        // Iterating thru each reaction
        for (Map.Entry<String, List<LinearComb<String>>> e :
        	participation.entrySet()) {

            // Getting the next rxn and all cycles it participates in
            String rxn = e.getKey();
            Collection<LinearComb<String>> cycles = e.getValue();

            // Getting rxn name sets from cycles
            List<Set<String>> rxnNameSets = getRxnNameSets(rxn, cycles);

            // Getting rxn name unions from rxn name sets
            List<Set<String>> unions = getReactionNameUnions(rxnNameSets);

            // Converting the unions to flux caps
            Set<LinearComb<String>> caps = 
            	convertUnionsToCaps(unions, cycles);

            // Saving reaction with its flux caps
            fluxCaps.put(rxn, caps);

            // Incrementing progress
            progress.increment();
        }

        // Returning all reactions that participate in cycles
        // with their respective linear combination "flux caps"
        return fluxCaps;
    }

    private Set<LinearComb<String>> 
    convertUnionsToCaps(List<Set<String>> unions, 
    					Collection<LinearComb<String>> cycles) {

    	// Determining appropriate size
    	int size = unions.size() + 1;
    	
        // Creating storage for "flux caps"
        Set<LinearComb<String>> caps = 
        	new LinkedHashSet<LinearComb<String>>(size, 1.0F);

        // Iterating thru each union
        for (Set<String> rxnNames : unions) {

            // If this set contains no more than one reaction from each
            // cycle, keep it. The set is redundant otherwise.
            if (containsOnlyOneRxnFromEachCycle(rxnNames, cycles)) {

                // Converting to a linear combination "flux cap"
                LinearComb<String> min = convertToExpression(rxnNames);

                // Adding to this reaction's "flux cap"
                caps.add(min);

            }
        }

        // Returning all flux caps. Each cap contains no more than one
        // reaction from each provided cycle.
        return caps;
    }

    private boolean
    containsOnlyOneRxnFromEachCycle(Set<String> reactionNameSet,
    								Collection<LinearComb<String>> cycles) {

        // Iterating thru each cycle
    	for (LinearComb<String> cycle : cycles) {

            // Copying set
            Set<String> s = new TreeSet<String>(reactionNameSet);

            // Retaining all reactions in the cycle
            s.retainAll(cycle.getVariables());

            // If more than one reaction remains
            if (s.size() > 1) {
                return false;
            }
        }

        return true;
    }

    private Map<String, List<LinearComb<String>>> 
    	getCycleParticipation(Collection<LinearComb<String>> allCycles) {

        // Creating storage
        Map<String, List<LinearComb<String>>> participation = 
        	new TreeMap<String, List<LinearComb<String>>>();

        // Iterating thru all cycles
        for (LinearComb<String> cycle : allCycles) {

            // Iterating thru each of these reaction
            for (String rxn : cycle.getVariables()) {

                // Getting all cycles this reaction participates in
                List<LinearComb<String>> cycles = 
                	participation.get(rxn);

                // If this is the first, create a new list
                if (cycles == null) {
                    cycles = new LinkedList<LinearComb<String>>();
                    participation.put(rxn, cycles);
                }

                // Adding the cycle to the list for this reaction
                cycles.add(cycle);
            }
        }

        return participation;
    }

    private List<Set<String>> getRxnNameSets(String rxnId,
                                Collection<LinearComb<String>> cycles) {

        // Creating storage for reaction name sets
        List<Set<String>> reactionNameSets = new LinkedList<Set<String>>();
        
        // Iterating thru each cycle in which the provided reaction
        // participates
        for (LinearComb<String> cycle : cycles) {

            // Getting the set of reaction names contained in this cycle
            Set<String> rxnNames = new TreeSet<String>(cycle.getVariables());

            // Removing the provided reaction name from
            // the cycle reaction names
            rxnNames.remove(rxnId);

            // Also removing equivalent reactions if they exist
            Collection<String> equivs = 
            	RxnComparer.getEquivIrrevRxnIds(rxnId, internalRxns);

            rxnNames.removeAll(equivs);

            // Adding to the rest
            reactionNameSets.add(rxnNames);
        }

        return reactionNameSets;
    }

    private List<Set<String>> removeSetsWithSecondaryRxns(Collection<Set<String>> sets) {

        // Creating storage for remaining sets
        List<Set<String>> c = new LinkedList<Set<String>>();

        // Iterating thru each set of reaction names
        for (Set<String> orig : sets) {

            // Getting the size
            int origSize = orig.size();

            // Copying the set and removing all secondary reactions
            Set<String> s = new TreeSet<String>(orig);
            s.removeAll(secondaryRxns);

            // If the size hasn't changed, save it
            if (origSize == s.size()) {
                c.add(orig);
            }
        }

        // Returning the new list of reaction name sets
        return c;
    }

    private List<Set<String>> getReactionNameUnions(List<Set<String>> reactionNameSets) {

        // Removing all reaction name sets that contain secondary reactions
        List<Set<String>> c = removeSetsWithSecondaryRxns(reactionNameSets);

        // Creating union computer
        SingleElementUnionsV2 seu = new SingleElementUnionsV2();

        // Computing single-element unions of reaction name sets
        Collection<Set<String>> unions = seu.getUnions(c);

        // Creating storage for reaction names sets with secondary
        // reaction included
        List<Set<String>> fullList = new LinkedList<Set<String>>();

        // Iterating thru each union
        for (Set<String> union : unions) {

            // Getting a new set with secondary reactions included
            union = addSecondaryReactions(union);

            // Adding to the rest
            fullList.add(union);
        }

        return fullList;
    }

    private Set<String> addSecondaryReactions(Set<String> reactions) {

        // Creating new set
        Set<String> s = new TreeSet<String>(reactions);

        // Iterating thru the provided reactions
        for (String rxnName : reactions) {

            // Getting all equivalent reactions
            Collection<String> equivs = 
            	RxnComparer.getEquivIrrevRxnIds(rxnName, internalRxns);

            // Adding them to the set
            s.addAll(equivs);
        }

        // Returning the set with secondary reactions included
        return s;
    }

    /**
     * Converts a set of reaction names to a linear combination.
     * @param reactionName String
     * @return Expression
     */
    private LinearComb<String> 
    	convertToExpression(Collection<String> reactionNames) {

        // Removing any duplicates that may exist
        Set<String> s = new TreeSet<String>(reactionNames);

        // Creating storage for linear combination terms
        List<LinearComb.Term<String>> terms = 
        	new LinkedList<LinearComb.Term<String>>();

        // Iterating thru each reaction name
        for (String rxnName : s) {

            // Creating new term
            terms.add(new HashLinearComb.Term<String>(rxnName));
        }

        // Returning linear combination
        return new HashLinearComb<String>(terms, 0d, 0d);
    }

}
