/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;

/**
 * This class is used to indicate that a flux cap is invalid.  A flux cap is
 * invalid if it does not cap the flux of the intended reaction.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 */
public class FluxCapException extends RuntimeException {

	private static final long serialVersionUID = 1L;
	private final MathExpr<String> fluxCap;
    private final String reactionName;

    /**
     * Constructs a new flux cap exception.
     * 
     * @param fluxCap the invalid flux cap.
     * @param reactionName the name of the reaction whose flux was intended to
     * be capped.
     * @param cause the cause.
     */
    public FluxCapException(MathExpr<String> fluxCap,
                            String reactionName,
                            Throwable cause) {

        super("The expression '" + fluxCap +
              "' does not actually cap the flux of the reaction '" +
              reactionName + "'.", cause);

        Util.nullCheck(fluxCap);
        Util.nullCheck(reactionName);

        this.fluxCap = fluxCap;
        this.reactionName = reactionName;
    }

    /**
     * Constructs a new flux cap exception.
     * 
     * @param fluxCap the invalid flux cap.
     * @param reactionName the name of the reaction whose flux was intended to
     * be capped.
     */
    public FluxCapException(MathExpr<String> fluxCap, String reactionName) {
        this(fluxCap, reactionName, null);
    }

    /**
     * Returns the invalid flux cap.
     * @return the invalid flux cap.
     */
    public MathExpr<String> getFluxCap() {
        return fluxCap;
    }

    /**
     * Returns the name of the reaction whose flux was intended to be capped.
     * @return the name of the reaction whose flux was intended to be capped.
     */
    public String getReactioName() {
        return reactionName;
    }

}
