/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.Map;
import java.util.Collection;

import sbrt.kernel.fba.*;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.MathExpr;

/**
 * This interface is used to represent a flux constraint variation computer. A
 * constraint-variation computer uses a particular fluxome, and it allows one
 * to compute the effect of varying the constraints within that fluxome.
 * 
 * @author This interface was written and documented by Jeremiah Wright while in
 *         the Wagner lab.
 * 
 * @param <F> a type of fluxome.
 * 
 */
public interface FluxCvComputer<F extends Fluxome> {

	/**
	 * Returns the fluxome that this flux constraint-variation
	 * computer uses.
	 * 
	 * @return the fluxome that this flux constraint-variation
	 * computer uses.
	 */
    public F getFluxome();

    /**
     * Returns all current flux expression constraints.
     * @return mathematical expressions of reaction names mapped to their
     * respective constraint.
     */
    public Map<? extends MathExpr<String>, Interval> getRxnExprConstraints();

    /**
     * Returns all current flux constraints.
     * @return reaction names mapped to their respective flux constraint.
     */
    public Map<String, Interval> getFluxConstraints();

    /**
     * Returns the current constraint on the provided expression.
     * 
     * @param rxnNameExp the mathematical expression of reaction names
     * whose constraint is to be returned.
     * 
     * @return the current constraint on the provided expression.
     */
    public Interval getRxnExprConstraint(MathExpr<String> rxnNameExp);

    /**
     * Returns the current flux constraint for the reaction with the provided
     * name.
     * 
     * @param rxnName the name of the reaction whose flux constraint
     * is to be returned.
     * 
     * @return the current flux constraint for the reaction with the provided
     * name. 
     */
    public Interval getFluxConstraint(String rxnName);

    /**
     * Changes the flux constraint of the reaction with the provided name.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be changed.
     * 
     * @param c the new constraint.
     */
    public void setFluxConstraint(String reactionName, Interval c);

    /**
     * Changes the flux constraints on more than one reaction.
     * 
     * @param fluxConstraints reaction names mapped to their respective flux
     * constraints.
     */
    public void setFluxConstraints(Map<String, Interval> fluxConstraints);

    /**
     * Changes the constraint on the provided mathematical expression 
     * of fluxes.
     * 
     * @param e a matheamtical expression of reaction names.
     * @param c the new constraint.
     */
    public void setRxnExprConstraint(MathExpr<String> e, Interval c);

    /**
     * Changes the constraints on more than one mathematical expression 
     * of fluxes.
     * 
     * @param constraints mathematical expressions of reaction names mapped
     * to their new respective constraints.
     */
    public void setRxnExprConstraints(
    		Map<? extends MathExpr<String>, Interval> constraints);

    /**
     * Restores the flux constraints for all reactions in the fluxome to those 
     * values contained in the fluxome.
     */
    public void restoreFluxConstraints();

    /**
     * Restores the flux constraint for the reaction with the provided 
     * name to the value contained in the fluxome.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be restored.
     */
    public void restoreFluxConstraint(String reactionName);

    /**
     * Restores the flux constraints for the provided reaction names to the
     * values contained in the fluxome.
     * 
     * @param reactionNames the reaction names whose flux constraints are to
     * be restored.
     */
    public void restoreFluxConstraints(Collection<String> reactionNames);

    /**
     * Restores the constraints on all flux expressions to the values
     * contained in the fluxome.
     */
    public void restoreRxnExprConstraints();

    /**
     * Restores the constraint on the provided flux expression to the value
     * contained in the fluxome.
     * 
     * @param rxnNameExp the expression of reaction names whose contraint
     * is to be restored.
     */
    public void restoreRxnExprConstraint(MathExpr<String> rxnNameExp);

    /**
     * Restores the constraints on the provided expressions to the values
     * contained in the fluxome.
     * 
     * @param rxnNameExps the reaction name expressions whose contraints
     * are to be restored.
     */
    public void restoreRxnExprConstraints(
    		Collection<? extends MathExpr<String>> rxnNameExps);

}
