/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.*;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;

/**
 * This interface is used to represent flux interval computers. Every flux in a
 * fluxome can assume minimum and maximum values, which define an interval.
 * Implementation of this interface must have the ability to compute these 
 * flux intervals.
 * 
 * @author This interface was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type.
 *
 */
public interface FluxIntervalComputer<F extends Fluxome> extends FluxCvComputer<F> {

    /**
     * Computes and returns the intervals of every flux in the fluxome
     * used by this flux interval computer.
     * 
     * @return reaction names mapped to the computed interval of their 
     * corresponding flux.
     */
    public Map<String, Interval> getIntervals();

    /**
     * Computes and returns the flux intervals of the reactions with the
     * provided names.
     * 
     * @param rxnNames a collection of reaction names.
     * 
     * @return reaction names mapped to the computed interval of their
     * corresponding flux.
     */
    public Map<String, Interval> getIntervals(Collection<String> rxnNames);

    /**
	 * Returns the current zero cutoff. The zero cutoff is a value used to
	 * decide if numerical values are close enough to zero to be considered equal 
	 * to zero.
	 * 
	 * @return the zero cutoff.
	 */
    public double getZeroCutoff();

    /**
     * Returns the amount by which constraints can be violated.
     * @return the amount by which constraints can be violated.
     */
    public double getTolerance();

    /**
     * Returns a progress monitor for this flux interval computer. 
     * @return a progress monitor for this flux interval computer. 
     */
    public ProgressMonitor getProgress();

}
