/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

/**
 * This class is used to indicate that the range of a flux interval has exceeded
 * a defined cutoff value.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 */
public class FluxRangeException extends RuntimeException {

	private static final long serialVersionUID = 1L;

	private final double rangeCutoff;
	private final Object offender;
	
	/**
	 * Constructs a new flux range exception.
	 * 
	 * @param rangeCutoff the defined range cutoff value.
	 * @param offender the object whose range has exceeded the defined cutoff.
	 * @param message a message useful for debugging.
	 */
	public FluxRangeException(double rangeCutoff, 
							  Object offender, 
							  String message) {
		super(message);
		
		this.rangeCutoff = rangeCutoff;
		this.offender = offender;
	}
	
	/**
	 * Constructs a new flux range exception.
	 * 
	 * @param rangeCutoff the defined range cutoff value.
	 * @param offender the object whose range has exceeded the defined cutoff.
	 */
	public FluxRangeException(double rangeCutoff, Object offender) {
		this(rangeCutoff, offender, "");
	}

	/**
	 * Returns the defined range cutoff.
	 * @return the defined range cutoff.
	 */
	public double getRangeCutoff() {
		return rangeCutoff;
	}
	
	/**
	 * Returns the object whose range has exceeded the defined cutoff.
	 * @return the object whose range has exceeded the defined cutoff.
	 */
	public Object getOffender() {
		return offender;
	}
}
