/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.*;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;

/**
 * This class is used to compute initial points for coordinate-direction
 * hit-and-run random flux vector algorithms.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * @param <F> the fluxome type.
 *
 */
public class InitialPointComputer<F extends Fluxome> implements FluxCvComputer<F> {

	/**
	 * The default zero cutoff.
	 */
	public final double DEFAULT_ZERO_CUTOFF = 1E-9;
	
    // The computer used for all optimizations.
    private final FbaOptimizer<F> computer;

    // A mapping of reaction names to unique positioning variables.
    private final Map<String, Object> pVars;

    // The value of the positioning variable at the edge of flux space
    private double pIntercept;
    
    private double zeroCutoff;
    
    /**
     * Constructs a new initial point, or flux vector, computer.
     * @param computer the computer with which optimizations will be performed.
     */
    public InitialPointComputer(FbaOptimizer<F> computer) {

        // Setting optimization computer
        Util.nullCheck(computer);
        this.computer = computer;

        // Getting all reaction names
        Set<String> rxnNames = computer.getFluxome().getRxnNames();

        // Determining appropriate size
        int size = rxnNames.size() + 1;
        
        // Creating storage for positioning variables
        pVars = new LinkedHashMap<String, Object>(size, 1.0F);

        // Creating a positioning variable for each reaction
        for (String rxn : rxnNames) {
            pVars.put(rxn, new Object());
        }

        pIntercept = 1E15;
        zeroCutoff = DEFAULT_ZERO_CUTOFF;
    }

    /**
     * Returns the fluxome used by this computer.
     * @return the fluxome used by this computer.
     */
    public F getFluxome() {
        return computer.getFluxome();
    }
    
    /**
     * Returns the value of each positioning variable at the edges
     * of the convex polytope.
     * 
     * @return the value of each positioning variable at the edges
     * of the convex body.
     */
    public double getPIntercept() {
        return pIntercept;
    }
    
    /**
     * Returns the zero cutoff.
     * @return the zero cutoff.
     */
    public double getZeroCutoff() {
    	return zeroCutoff;
    }
    
    /**
     * Sets the value of the positioning variables at the edges of the 
     * convex polytope. With a typical barrier function, this value is
     * infinite, but with this linear "barrier function", it must be finite.
     * 
     * @param pIntercept a finite value >= 10
     */
    public void setPIntercept(double pIntercept) {
        
        if (pIntercept < 10 || 
            Double.isInfinite(pIntercept) || 
            Double.isNaN(pIntercept)) {
            
            String s = "Invalid pIntercept: " + pIntercept;
            throw new IllegalArgumentException(s);
        }
        
        this.pIntercept = pIntercept;
    }
    
    /**
     * Sets the zero cutoff. If the range of a flux interval is less than
     * the zero cutoff, it is assumed to be zero. Zero cutoffs should be
     * small positive numbers, such as {@code 1E-9}.
     * 
     * @param zeroCutoff the zero cutoff.
     */
    public void setZeroCutoff(double zeroCutoff) {
        
        if (zeroCutoff < 0d || 
            Double.isInfinite(pIntercept) || 
            Double.isNaN(pIntercept)) {
            
            String s = "Invalid zeroCutoff: " + zeroCutoff;
            throw new IllegalArgumentException(s);
        }
        
        this.zeroCutoff = zeroCutoff;
    }

    /**
     * Computes an initial point, or flux vector, in the flux space of the
     * fluxome used by this computer.
     * 
     * @return the results of computing an initial point in the flux space of
     * the fluxome used by this computer.
     */
    public Results run() {

        // Setting the bounding planes that form the position space(s)
        LinearComb<Object> objective = setPositioningSpace();

        // Computing objective value (i.e. P[min])
        FbaOptimizer<F>.Results results = computer.getMin(objective);

        // Computing P[min]
        double pMin = results.getObjectiveValue();

        // Getting solution vector
        Map<?, Double> solutionVector = results.getSolutionVector();

        // Returning results
        return new Results(objective.terms(), pMin, solutionVector);
    }

    private LinearComb<Object> setPositioningSpace() {

        // Restoring the program matrix to its original (pristine) state
        computer.restoreMatrixElements();

        // Getting all reaction names
        Set<String> allRxnNames = getFluxome().getRxnNames();

        // Iterating thru each reaction name
        for (String rxn : allRxnNames) {

            // Getting the current flux constraint for this reaction
            Interval fluxBounds = getFluxConstraint(rxn);

            // Adding the appropriate bounding planes
            addEightPlanes(rxn, fluxBounds);
        }

        // Creating storage for objective function terms
        List<LinearComb.Term<Object>> terms = 
        	new LinkedList<LinearComb.Term<Object>>();

        // Getting the positioning variables that were used
        Set<?> usedPVars = computer.getColumnVariables();
        usedPVars.removeAll(allRxnNames);

        // Adding each positioning variable to the objective terms
        for (Object pVar : usedPVars) {
            terms.add(new HashLinearComb.Term<Object>(pVar));
        }

        // Returng the objective function
        return new HashLinearComb<Object>(terms, 0, 0.0);
    }

    /**
     * This method adds eight (8) bounding hyperplanes (i.e. halfspaces)
     * to define the "positioning space" of the flux corresponding to the
     * provided reaction name. 
     *
     */
    private void addEightPlanes(String reactionName, Interval bounds) {

        // Getting lower and upper flux bounds
        double lowerFlux = bounds.getLower();
        double upperFlux = bounds.getUpper();

        // Unbounded fluxes are not allowed
        if (Double.isInfinite(lowerFlux) || Double.isInfinite(upperFlux)) {
            throw new UnboundedVariableException(reactionName);
        }

        // Computing the flux range
        double fluxRange = upperFlux - lowerFlux;

        // If the flux has no range, do nothing
        if (fluxRange < zeroCutoff) {
            return;
        }

        // Getting the positioning variable for this reaction name
        Object pVar = pVars.get(reactionName);

        // Creating new row variables for each bounding hyperplane
        Object plane1 = new Object();
        Object plane2 = new Object();
        Object plane3 = new Object();
        Object plane4 = new Object();
        Object plane5 = new Object();
        Object plane6 = new Object();
        Object plane7 = new Object();
        Object plane8 = new Object();

        // Position along the y (positioning) axis corresponding to the
        // extreme points of the bounding hyperplanes
        double y0 = pIntercept;
        double y1 = 1.0;
        double y2 = 0.5;
        double y3 = 0.1;
        double y4 = 0.0;
        double y5 = 0.1;
        double y6 = 0.5;
        double y7 = 1.0;
        double y8 = pIntercept;

        // Computing "rise" for each plane
        double p1Rise = y1 - y0;
        double p2Rise = y2 - y1;
        double p3Rise = y3 - y2;
        double p4Rise = y4 - y3;
        double p5Rise = y5 - y4;
        double p6Rise = y6 - y5;
        double p7Rise = y7 - y6;
        double p8Rise = y8 - y7;

        // Fractional position along the x (flux) axis corresponding to the
        // extreme points of the bounding hyperplanes
        double x0 = 0.000;
        double x1 = 0.001;
        double x2 = 0.010;
        double x3 = 0.050;
        double x4 = 0.500;
        double x5 = 0.950;
        double x6 = 0.990;
        double x7 = 0.999;
        double x8 = 1.000;

        // Computing "run" for each plane
        double p1Run = (x1 - x0) * fluxRange;
        double p2Run = (x2 - x1) * fluxRange;
        double p3Run = (x3 - x2) * fluxRange;
        double p4Run = (x4 - x3) * fluxRange;
        double p5Run = (x5 - x4) * fluxRange;
        double p6Run = (x6 - x5) * fluxRange;
        double p7Run = (x7 - x6) * fluxRange;
        double p8Run = (x8 - x7) * fluxRange;

        // Computing slopes for each bounding plane
        double p1Slope = p1Rise / p1Run;
        double p2Slope = p2Rise / p2Run;
        double p3Slope = p3Rise / p3Run;
        double p4Slope = p4Rise / p4Run;
        double p5Slope = p5Rise / p5Run;
        double p6Slope = p6Rise / p6Run;
        double p7Slope = p7Rise / p7Run;
        double p8Slope = p8Rise / p8Run;

        // Each hyperplane is of the form y = m * x + b, or equivalently
        // (after solving for b) y - m * x = b. The corresponding matrix
        // representation is:
        //
        // [ y |  x ]
        // ----------
        // [ 1 | -m ]

        // Setting matrix elements for each plane.
        computer.setMatrixElement(plane1, reactionName, -1 * p1Slope);
        computer.setMatrixElement(plane2, reactionName, -1 * p2Slope);
        computer.setMatrixElement(plane3, reactionName, -1 * p3Slope);
        computer.setMatrixElement(plane4, reactionName, -1 * p4Slope);
        computer.setMatrixElement(plane5, reactionName, -1 * p5Slope);
        computer.setMatrixElement(plane6, reactionName, -1 * p6Slope);
        computer.setMatrixElement(plane7, reactionName, -1 * p7Slope);
        computer.setMatrixElement(plane8, reactionName, -1 * p8Slope);

        computer.setMatrixElement(plane1, pVar, 1);
        computer.setMatrixElement(plane2, pVar, 1);
        computer.setMatrixElement(plane3, pVar, 1);
        computer.setMatrixElement(plane4, pVar, 1);
        computer.setMatrixElement(plane5, pVar, 1);
        computer.setMatrixElement(plane6, pVar, 1);
        computer.setMatrixElement(plane7, pVar, 1);
        computer.setMatrixElement(plane8, pVar, 1);

        // Getting row maxima, (i.e. intercepts)
        //      b  = y  -    m    * (            x             )
        double p1b = y1 - p1Slope * (lowerFlux + x1 * fluxRange);
        double p2b = y2 - p2Slope * (lowerFlux + x2 * fluxRange);
        double p3b = y3 - p3Slope * (lowerFlux + x3 * fluxRange);
        double p4b = y4 - p4Slope * (lowerFlux + x4 * fluxRange);
        double p5b = y5 - p5Slope * (lowerFlux + x5 * fluxRange);
        double p6b = y6 - p6Slope * (lowerFlux + x6 * fluxRange);
        double p7b = y7 - p7Slope * (lowerFlux + x7 * fluxRange);
        double p8b = y8 - p8Slope * (lowerFlux + x8 * fluxRange);

        // The rows have an unbounded maximum (i.e. their maximum is infinity)
        double posInf = Double.POSITIVE_INFINITY;

        // Setting row constraints
        computer.setRowConstraint(plane1, new Interval(p1b, posInf));
        computer.setRowConstraint(plane2, new Interval(p2b, posInf));
        computer.setRowConstraint(plane3, new Interval(p3b, posInf));
        computer.setRowConstraint(plane4, new Interval(p4b, posInf));
        computer.setRowConstraint(plane5, new Interval(p5b, posInf));
        computer.setRowConstraint(plane6, new Interval(p6b, posInf));
        computer.setRowConstraint(plane7, new Interval(p7b, posInf));
        computer.setRowConstraint(plane8, new Interval(p8b, posInf));

    }

    /**
     * Changes the flux constraint of the reaction with the provided name.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be changed.
     * 
     * @param c the new constraint.
     */
    public void setFluxConstraint(String reactionName, Interval c) {
        computer.setFluxConstraint(reactionName, c);
    }

    /**
     * Changes the flux constraints on more than one reaction.
     * 
     * @param fluxConstraints reaction names mapped to their respective flux
     * constraints.
     */
    public void setFluxConstraints(Map<String, Interval> fluxConstraints) {
        computer.setFluxConstraints(fluxConstraints);
    }

    /**
     * Changes the constraint on the provided mathematical expression 
     * of fluxes.
     * 
     * @param e a matheamtical expression of reaction names.
     * @param c the new constraint.
     */
    public void setRxnExprConstraint(MathExpr<String> e, Interval c) {
        computer.setRxnExprConstraint(e, c);
    }

    /**
     * Changes the constraints on more than one mathematical expression 
     * of fluxes.
     * 
     * @param constraints mathematical expressions of reaction names mapped
     * to their new respective constraints.
     */
    public void 
    setRxnExprConstraints(Map<? extends MathExpr<String>, Interval> 
    constraints) {
        computer.setRxnExprConstraints(constraints);
    }

    /**
     * Restores the flux constraint for the reaction with the provided 
     * name to the value contained in the fluxome.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be restored.
     */
    public void restoreFluxConstraint(String reactionName) {
        computer.restoreFluxConstraint(reactionName);
    }

    /**
     * Restores the flux constraints for the provided reaction names to the
     * values contained in the fluxome.
     * 
     * @param reactionNames the reaction names whose flux constraints are to
     * be restored.
     */
    public void restoreFluxConstraints(Collection<String> reactionNames) {
        computer.restoreFluxConstraints(reactionNames);
    }

    /**
     * Restores the flux constraints for all reactions in the fluxome to those 
     * values contained in the fluxome.
     */
    public void restoreFluxConstraints() {
        computer.restoreFluxConstraints();
    }

    /**
     * Restores the constraint on the provided flux expression to the value
     * contained in the fluxome.
     * 
     * @param expr the expression of reaction names whose contraint
     * is to be restored.
     */
    public void restoreRxnExprConstraint(MathExpr<String> expr) {
        computer.restoreRxnExprConstraint(expr);
    }

    /**
     * Restores the constraints on the provided expressions to the values
     * contained in the fluxome.
     * 
     * @param rxnNameExps the reaction name expressions whose contraints
     * are to be restored.
     */
    public void 
    restoreRxnExprConstraints(Collection<? extends MathExpr<String>>
    rxnNameExps) {
        computer.restoreRxnExprConstraints(rxnNameExps);
    }

    /**
     * Restores the constraints on all flux expressions to the values
     * contained in the fluxome.
     */
    public void restoreRxnExprConstraints() {
        computer.restoreRxnExprConstraints();
    }

    /**
     * Returns all current flux expression constraints.
     * @return mathematical expressions of reaction names mapped to their
     * respective constraint.
     */
    public Map<MathExpr<String>, Interval> getRxnExprConstraints() {
        return computer.getRxnExprConstraints();
    }

    /**
     * Returns all current flux constraints.
     * @return reaction names mapped to their respective constraint.
     */
    public Map<String, Interval> getFluxConstraints() {
        return computer.getFluxConstraints();
    }

    /**
     * Returns the current constraint on the provided expression.
     * 
     * @param rxnNameExp the mathematical expression of reaction names
     * whose constraint is to be returned.
     * 
     * @return the current constraint on the provided expression.
     */
    public Interval getRxnExprConstraint(MathExpr<String> rxnNameExp) {
        return computer.getRxnExprConstraint(rxnNameExp);
    }

    /**
     * Returns the current flux constraint on the reaction with the provided
     * name.
     * 
     * @param rxnName the name of the reaction whose flux constraint
     * is to be returned.
     * 
     * @return the current flux constraint on the reaction with the provided
     * name. 
     */
    public Interval getFluxConstraint(String rxnName) {
        return computer.getFluxConstraint(rxnName);
    }

    /**
     * This class is used to represent the results of the initial point
     * computation.
     *
     */
    public class Results {

    	private final int n;
    	private final double pMin;
        private final Map<?, Double> solutionVector;

        private Results(int n, double pMin, Map<?, Double> solutionVector) {
        	this.n = n;
        	this.pMin = pMin;
            this.solutionVector = solutionVector;
        }

        /**
         * Returns the minimum value of the linear combination of all created
         * positioning variables.
         * 
         * @return the minimum value of the linear combination of all created
         * positioning variables.
         */
        public double getPMin() {
            return pMin;
        }

        /**
         * Returns the average value of all created positioning variables.
         * @return the average value of all created positioning variables.
         */
        public double getPAvg() {
            return pMin / n;
        }
        
        /**
         * Returns the number of created positioning variables.
         * @return the number of created positioning variables.
         */
        public int pVariables() {
        	return n;
        }

        /**
         * Returns the computed flux vector.
         * @return the computed flux vector.
         */
        public Map<String, Double> getFluxVector() {

        	// Creating storage
        	Map<String, Double> m = new TreeMap<String, Double>();

        	// Iterating thru each variable-value pair
            for (Map.Entry<?, Double> e : solutionVector.entrySet()) {

            	// Getting the key
            	Object key = e.getKey();
            	
                // If the key is a String, it must be a reaction name
                if (key instanceof String) {

                	// Safely casting
                	String rxn = (String) key;
                	
                	// Saving
                	m.put(rxn, e.getValue());
                }

            }

            return m;
        }

        /**
         * Returns the vector of values of all positioning variables that 
         * were created. Note that a positioning variable is created only
         * if the range (max-min) of a flux interval is greater than zero.
         * 
         * @return reaction names mapped to the value of their associated
         * positioning variable.
         */
        public Map<String, Double> getPVector() {

            // Creating storage for positioning variable values
            Map<String, Double> pVector = new TreeMap<String, Double>();

            // Iterating thru each reaction-positioning variable pair
            for (Map.Entry<String, Object> e : pVars.entrySet()) {

                // Getting the next reaction name-positioning variable pair
                String rxnName = e.getKey();
                Object pVar = e.getValue();

                Double pValue = solutionVector.get(pVar);

                // If a value was computed for this positioning variable
                if (pValue != null) {
                    pVector.put(rxnName, pValue);
                }
            }

            // Returning a Map of reaction names mapped to their
            // corresponding positioning value
            return pVector;
        }
    }
    
    
}
