/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.*;

import sbrt.kernel.utilities.*;

/**
 * This class is used to indicate that a flux vector is invalid.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 */
public class InvalidFluxVectorException extends RuntimeException {

	private static final long serialVersionUID = 1L;
	
	private final UnmodifiableMap<String, Double> fluxVector;
	private final double tolerance;
	
	/**
	 * Constructs a new exception for the provided flux vector.
	 * @param fluxVector the invalid flux vector.
	 * @param tolerance the numerical tolerance used to check the 
	 * validity of the flux vector. 
	 */
    public InvalidFluxVectorException(Map<String, Double> fluxVector,
    								  double tolerance) {
        super();
        this.fluxVector = getVector(fluxVector);
        this.tolerance = tolerance;
    }

	/**
	 * Constructs a new exception for the provided flux vector.
	 * @param fluxVector the invalid flux vector.
	 * @param tolerance the numerical tolerance used to check the 
	 * validity of the flux vector. 
	 * @param message a message useful for debugging.
	 */
    public InvalidFluxVectorException(Map<String, Double> fluxVector,
    								  double tolerance,
    								  String message) {
        super(message);
        this.fluxVector = getVector(fluxVector);
        this.tolerance = tolerance;
    }

	/**
	 * Constructs a new exception for the provided flux vector.
	 * @param fluxVector the invalid flux vector.
	 * @param tolerance the numerical tolerance used to check the 
	 * validity of the flux vector. 
	 * @param cause the exception used to initially indicate that the flux
	 * vector is invalid. 
	 */
    public InvalidFluxVectorException(Map<String, Double> fluxVector, 
    								  double tolerance,
    								  Throwable cause) {
        super(cause);
        this.fluxVector = getVector(fluxVector);
        this.tolerance = tolerance;
    }

	/**
	 * Constructs a new exception for the provided flux vector.
	 * @param fluxVector the invalid flux vector.
	 * @param tolerance the numerical tolerance used to check the 
	 * validity of the flux vector. 
	 * @param message a message useful for debugging.
	 * @param cause the exception used to initially indicate that the flux
	 * vector is invalid. 
	 */
    public InvalidFluxVectorException(Map<String, Double> fluxVector, 
    								  double tolerance,
    								  String message,
                                      Throwable cause) {
        super(message, cause);
        this.fluxVector = getVector(fluxVector);
        this.tolerance = tolerance;
    }

    /**
     * Returns the invalid flux vector;
     * @return the invalid flux vector;
     */
    public UnmodifiableMap<String, Double> getFluxVector() {
        return fluxVector;
    }

    /**
     * Returns the numerical tolerance used to check the 
	 * validity of the flux vector. 
     * @return the numerical tolerance used to check the 
	 * validity of the flux vector. 
     */
    public double getTolerance() {
    	return tolerance;
    }
    
    private UnmodifiableMap<String, Double> 
    getVector(Map<String, Double> fluxVector) {

        if (fluxVector instanceof UnmodifiableMap) {
            return (UnmodifiableMap<String, Double>) fluxVector;
        }

        Map<String, Double> vector = 
        	new LinkedHashMap<String, Double>(fluxVector.size() + 1, 1.0F);
        vector.putAll(fluxVector);

        return new UnmodifiableMap<String, Double>(vector);
    }

}
