/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.optimization.ProgramSolverStatus;
import sbrt.kernel.utilities.PercentProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;

/**
 * This class is used to compute Mahadevan-Schilling flux intervals. The 
 * first description of these intervals can be found in the paper:
 * R. Mahadevan, C. H. Schilling, "Effects of Alternate Optimal Solutions in 
 * Constraint-based Genome Scale Metabolic Models", Metabolic Engineering, 2003,
 * 5:264-276.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type.
 *
 */
public class MsFluxIntervalComputer<F extends IrrevRxnFluxome> extends
        AbstractFluxIntervalComputer<F> {

    private final PercentProgressMonitor progress;

    // Reaction names mapped to their flux caps
    private final Map<String, Set<LinearComb<String>>> allFluxCaps;

    /**
     * Constructs a new Mahadevan-Schilling flux interval computer.
     * 
     * @param computer the computer with which all optimizations will be
     * computed.
     * @param fluxCaps the flux caps used to constrain all unbounded fluxes.
     */
    public MsFluxIntervalComputer(FbaOptimizer<F> computer,
    		Map<String, Set<LinearComb<String>>> fluxCaps) {
    	
        this(computer, fluxCaps, ZERO_CUTOFF);
    }

    /**
     * Constructs a new Mahadevan-Schilling flux interval computer.
     * 
     * @param computer the computer with which all optimizations will be
     * computed.
     * @param fluxCaps the flux caps used to constrain all unbounded fluxes.
     * @param zeroCutoff the zero cutoff used by this flux interval computer.
     * 
     */
    public MsFluxIntervalComputer(
    		FbaOptimizer<F> computer,
    		Map<String, Set<LinearComb<String>>> fluxCaps,
    		double zeroCutoff) {
    	
        super(computer, zeroCutoff);

        // Determining appropriate size
        int size = fluxCaps.size() + 1;
        
        // Creating storage
        allFluxCaps = 
        	new LinkedHashMap<String, Set<LinearComb<String>>>(size, 1.0F);

        // Copying flux caps
        allFluxCaps.putAll(fluxCaps);

        // Initializing progress monitor
        progress = new PercentProgressMonitor();
    }

    /**
     * Returns a progress monitor for this flux interval computer. 
     * @return a progress monitor for this flux interval computer. 
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    /**
     * Computes and returns the intervals of every flux in the fluxome
     * used to construct this flux interval computer.
     * 
     * @return reaction names mapped to the computed interval of their 
     * corresponding flux.
     */
    public Map<String, Interval> getIntervals() {

        // Getting all reaction names
        Set<String> allRxns = getFluxome().getRxnNames();

        // Returning flux ranges
        return getIntervals(allRxns);
    }

    /**
     * Computes and returns the flux intervals of the reactions with the
     * provided names.
     * 
     * @param rxnNames a collection of reaction names.
     * 
     * @return reaction names mapped to the computed interval of their
     * corresponding flux.
     */
    public Map<String, Interval> getIntervals(Collection<String> rxnNames) {

        try {

        	// Starting progress monitor
            progress.start();
            progress.resetCount();

            // Getting and setting total
            int total = allFluxCaps.size() + rxnNames.size() * 2 + 1;
            progress.setTotal(total);

            // Computing flux intervals
            Map<String, Interval> intervals = computeIntervals(rxnNames);

            // Stopping progress monitor
            progress.complete();
            progress.stop();

            // Returning flux ranges
            return intervals;

        } catch (RuntimeException e) {
            
            if (!progress.stopped()) {
                progress.stop();
            }
            
            throw e;
        }

    }

    private Map<String, Interval> 
    	computeIntervals(Collection<String> rxnNames) {

        // Getting current flux constraints
        Map<String, Interval> originalFluxConstraints = 
        	getFluxConstraints();

        // Getting maxima for cycle-participators. This is
        // necessary, regardless of the reaction names provided, because
        // these upper bounds are used to define the flux space. This means
        // that the bounds on other fluxes could be indirectly affected.
        Map<String, Double> largestMaxima = getMaximaWithFluxCaps();

        // Setting upper bounds to their computed maxima
        setUpperBounds(largestMaxima);

        // Getting the minima for the reactions provided
        Map<String, Double> minima = getMinimumFluxes(rxnNames);

        // Getting the maxima for the reactions provided
        Map<String, Double> maxima = getMaximumFluxes(rxnNames);

        // Getting flux intervals from the computed minima and maxima
        Map<String, Interval> intervals = getIntervals(minima, maxima);

        // Restoring original constraints
        setFluxConstraints(originalFluxConstraints);

        // Returning the flux intervals for the provided reaction names
        return intervals;
    }

    private Map<String, Double> getMinimumFluxes(Collection<String> rxnNames) {

    	// Determining appropriate size
    	int size = rxnNames.size() + 1;

        // Creating storage
        Map<String, Double> minima = 
        	new LinkedHashMap<String, Double>(size, 1.0F);

        // Iterating thru each reaction name
        for (String rxn : rxnNames) {

            // Computing the minimum flux
            double min = getBoundedMin(rxn);

            // Storing it
            minima.put(rxn, min);

            // Incrementing progress
            progress.increment();
        }

        // Returning the minimum fluxes thru each reaction provided
        return minima;
    }

    private Map<String, Double> getMaximumFluxes(Collection<String> rxnNames) {

    	// Determining appropriate size
        int size = rxnNames.size() + 1;

        // Creating storage
        Map<String, Double> maxima = 
        	new LinkedHashMap<String, Double>(size, 1.0F);

        Set<String> exchRxnNames = getFluxome().getExchangeRxnNames();
        
        // Iterating thru each internal reaction name
        for (String rxn : rxnNames) {

            double max = Double.NaN;
            
            // If it's an exchange reaction
            if (exchRxnNames.contains(rxn)) {
                max = getMaximum(rxn);
            }
            
            // If it's an internal reaction
            else {
                max = getBoundedMax(rxn);
            }
            
            // Storing it
            maxima.put(rxn, max);

            // Incrementing progress
            progress.increment();
        }

        // Returning the maximum fluxes thru each reaction provided
        return maxima;
    }

    private void setUpperBounds(Map<String, Double> computedMaxima) {

    	// Iterating thru each upper bound
    	for (Map.Entry<String, Double> e : computedMaxima.entrySet()) {

    		// Getting the next reaction-max pair
            String rxn = e.getKey();
            Double max = e.getValue();

            // Getting the current constraint
            Interval current = getFluxConstraint(rxn);

            // Creating new constraint
            double min = current.getLower();
            Interval bound = getInterval(rxn, current, min, max);

            // Changing the upper bound
            setFluxConstraint(rxn, bound);
        }

    }

    private Map<String, Double> getMaximaWithFluxCaps() {

        // Determining appropriate size
    	int size = allFluxCaps.size() + 1;

        // Creating storage
        Map<String, Double> maxima = 
        	new LinkedHashMap<String, Double>(size, 1.0F);

        // Iterating thru all rxns that have flux caps
        for (String rxn : allFluxCaps.keySet()) {

            // Getting current constraint
            Interval current = getFluxConstraint(rxn);

            // If the upper bound is finite, finding the largest maximum
            // isn't necessary.
            if (!Double.isInfinite(current.getUpper())) {
                // Incrementing progress
                progress.increment();
                continue;
            }

            // Computing the largest finite maximum flux while constraining
            // the provided flux caps to equal their minimum
            double max = getLargestFlux(rxn);

            // Saving
            maxima.put(rxn, max);

            // Incrementing progress
            progress.increment();
        }

        return maxima;
    }

    /**
     * Returns the largest possible finite flux for the provided reaction.
     * @param reactionName String
     * @param futileCycles Collection
     * @return double
     */
    private double getLargestFlux(String reactionName) {

        // First trying to determine if the flux has a finite upper bound.
        // If one exists, then it's automatically the largest possible,
        // so it's returned. If one doesn't exist, it will have to be
        // computed by constraining the flux caps.
        double max = getMaximum(reactionName);
        if (!Double.isInfinite(max)) {
            return max;
        }

        // Getting the "flux caps" for this reaction
        Set<? extends LinearComb<String>> fluxCaps = 
        	allFluxCaps.get(reactionName);

        // Setting an initial value for largest finite maximum flux
        double largestMaximum = Double.NEGATIVE_INFINITY;

        // Iterating thru each "flux cap"
        for (LinearComb<String> cap : fluxCaps) {

            // Getting current constraint for this flux cap if it exists
            Interval orig = getRxnExprConstraint(cap);

            // Computing its minimum value
            double min = getMinimum(cap);

            // Creating the bounds for the cap
            Interval c = new Interval(min, min);

            // Maximizing the flux thru the reaction under consideration
            // while holding the flux cap equal to its minimum value
            max = getMaximumWithFluxCap(reactionName, cap, c);

            // If flux is still unbounded the flux cap is invalid
            if (Double.isInfinite(max)) {
                throw new FluxCapException(cap, reactionName);
            }

            // If this maximum flux is greater than the largest computed 
            // so far, save it
            if (max > largestMaximum) {
                largestMaximum = max;
            }

            // Restoring the original constraint on the flux cap
            setRxnExprConstraint(cap, orig);
        }

        // Returning the largest maximum flux for the provided reaction
        return largestMaximum;
    }
    
    private double getMaximumWithFluxCap(String rxn, 
    		LinearComb<String> fluxCap, Interval fluxCapBounds) {

    	// Getting the range for the flux cap, which is ideally 0.0
    	double bounds = fluxCapBounds.getUpper() - fluxCapBounds.getLower();
    	
    	// Getting the range cutoff
    	double cutoff = getZeroCutoff();
    	
    	// Checking the bounds, which are only changed by calling
    	// this method recursively
    	if (Double.compare(bounds, cutoff) > 0) {
    		String s = "The range of " + fluxCapBounds + " for the flux cap " +
    					fluxCap + " exceeds the zero cutoff " + cutoff;
    		throw new FluxRangeException(cutoff, fluxCap, s);
    	}

        // Creating a very small value to account for 
        // possible numerical error from the program solver
        double tolerance = 2.5E-10;

        // Constraining the cap to its minimum value
        setRxnExprConstraint(fluxCap, fluxCapBounds);

        // Creating field for storing maximum flux
        double max = Double.NaN;
        
        try {
        	
            // Trying to maximize the flux thru the reaction 
        	// under consideration
        	max = getMaximum(rxn);
        	
        	// Catching any optimization exceptions thrown
        } catch (FbaOptException e) {

        	// If no feasible solution exists
        	if (e.getStatus() == ProgramSolverStatus.INFEASIBLE) {
        		
        		// Creating a new "loosened" constraint
        		double lower = fluxCapBounds.getLower() - tolerance;
        		double upper = fluxCapBounds.getUpper() + tolerance;
        		Interval c = new Interval(lower, upper);
        		
        		// Calling this method recursively with the 
        		// loosened constraint
        		return getMaximumWithFluxCap(rxn, fluxCap, c);
        	}
        	
        	// Otherwise just throw the exception right thru
        	throw e;
        	
        }

        return max;
    }

}
