/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.IrrevRxn;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.IrrevRxnFluxomeV1;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.MixedFluxomeV1;
import sbrt.kernel.fba.RevExchRxn;
import sbrt.kernel.fba.RevRxn;
import sbrt.kernel.fba.RxnComparer;
import sbrt.kernel.fba.SinkRxn;
import sbrt.kernel.fba.SourceRxn;

/**
 * This class is used to remove "redundant" reactions from fluxomes. If a fluxome
 * contains two stoichiometrically equivalent reactions, one of them is considered
 * redundant.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 *
 */
public class RedundRxnRemover {


	/**
	 * Returns a fluxome for which all "redundant" reactions have been
	 * eliminated.
	 * 
	 * @param fluxome a fluxome.
	 * 
	 * @return  a fluxome for which all "redundant" reactions have been
	 * eliminated.
	 * 
	 * @see #getEquivRxns(Fluxome)
	 */
	public static MixedFluxome remove(MixedFluxome fluxome) {

        // Getting the names of all non-redundant reactions
        Set<String> rxnNames = getRxns(fluxome);

		// Returning the new fluxome
		return MixedFluxomeV1.make(rxnNames, fluxome);
	}

	
	/**
	 * Returns a fluxome for which all "redundant" reactions have been
	 * eliminated.
	 * 
	 * @param fluxome a fluxome.
	 * 
	 * @return  a fluxome for which all "redundant" reactions have been
	 * eliminated.
	 * 
	 * @see #getEquivRxns(Fluxome)
	 */
	public static IrrevRxnFluxome remove(IrrevRxnFluxome fluxome) {
		
        // Getting the names of all non-redundant reactions
        Set<String> rxnNames = getRxns(fluxome);

		// Returning the new fluxome
		return IrrevRxnFluxomeV1.make(rxnNames, fluxome);
	}

	private static Set<String> getRxns(Fluxome fluxome) {
		
        // Getting the names of all reactions
        Set<String> rxnNames = fluxome.getRxnNames();
        
        // Getting the names of all equivalent reactions
        Collection<Set<String>> equivRxns = getEquivRxns(fluxome);

		// Iterating thru each set of equivalent reactions
		for (Set<String> equivs : equivRxns) {

			// Getting an iterator over the set
			Iterator<String> i = equivs.iterator();

			// Ignoring the first reaction name
			i.next();

			// Iterating thru each remaining reaction
			while (i.hasNext()) {

				// Removing it from the rest
				rxnNames.remove(i.next());
			}
		}

		return rxnNames;
	}
	
	/**
	 * Returns the equivalent reactions contained in the provided fluxome.
	 * @param fluxome a fluxome.
	 * @return the equivalent reactions contained in the provided fluxome.
	 */
	public static Collection<Set<String>> getEquivRxns(Fluxome fluxome) {
		
		Map<String, ? extends ChemRxn> rxns = fluxome.getRxns();
		
		Map<String, IrrevRxn> irrevRxns = new TreeMap<String, IrrevRxn>();
		Map<String, RevRxn> revRxns = new TreeMap<String, RevRxn>();
		Map<String, SourceRxn> sourceRxns = new TreeMap<String, SourceRxn>();
		Map<String, SinkRxn> sinkRxns = new TreeMap<String, SinkRxn>();
		Map<String, RevExchRxn> revExchRxns = new TreeMap<String, RevExchRxn>();
		
		List<Set<String>> equivRxns = new LinkedList<Set<String>>();
		
		for (Map.Entry<String, ? extends ChemRxn> e : rxns.entrySet()) {
			
			String rxnName = e.getKey();
			ChemRxn r = e.getValue();
			
			if (r instanceof IrrevRxn) {
				irrevRxns.put(rxnName, (IrrevRxn) r);
			}
			
			else if (r instanceof RevRxn) {
				revRxns.put(rxnName, (RevRxn) r);
			}
			
			else if (r instanceof SourceRxn) {
				sourceRxns.put(rxnName, (SourceRxn) r);
			}
			
			else if (r instanceof SinkRxn) {
				sinkRxns.put(rxnName, (SinkRxn) r);
			}
			
			else if (r instanceof RevExchRxn) {
				revExchRxns.put(rxnName, (RevExchRxn) r);
			}
			
			else {
				String s = "Invalid reaction type: " + 
							r.getClass().getCanonicalName();
				throw new UnsupportedOperationException(s);
			}
			
		}
		
		equivRxns.addAll(RxnComparer.getEquivIrrevRxnIds(irrevRxns));
		equivRxns.addAll(RxnComparer.getEquivRevRxnIds(revRxns));
		equivRxns.addAll(RxnComparer.getEquivSourceRxnIds(sourceRxns));
		equivRxns.addAll(RxnComparer.getEquivSinkRxnIds(sinkRxns));
		equivRxns.addAll(RxnComparer.getEquivRevExchRxnIds(revExchRxns));
		
		return equivRxns;
	}
	
	
}
