/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.Collection;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import sbrt.kernel.fba.IrrevRxn;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.RxnComparer;
import sbrt.kernel.fba.SinkRxn;
import sbrt.kernel.fba.SourceRxn;
import sbrt.kernel.math.IllegalBoundsException;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.MathExpr;
import sbrt.kernel.utilities.MutableProgressMonitor;
import sbrt.kernel.utilities.PercentProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;

/**
 * This class is used to compute flux intervals while constraining the fluxes of
 * reverse reactions to their minimum. An interval is computed by minimizing and
 * maximizing the flux of a given reaction. If a flux is unbounded in a 
 * particular direction, as they
 * sometimes are, {@code Double.NEGATIVE_INFINITY} or
 * {@code Double.POSITIVE_INFINITY} are used to denote the flux bounds.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * @param <F> the fluxome type.
 * 
 */
public class ReverseRxnIntervalComp<F extends IrrevRxnFluxome> extends
        AbstractFluxIntervalComputer<F> {

    private final PercentProgressMonitor progress;

    /**
     * Constructs a new flux interval computer.
     * 
     * @param computer the computer with which all optimizations will be
     * computed.
     * 
     * @param zeroCutoff the amount by which a computed lower bound can
     * <em>exceed</em> a computed upper bound.
     */
    public ReverseRxnIntervalComp(FbaOptimizer<F> computer,
            							  double zeroCutoff) {          
        super(computer, zeroCutoff);
        progress = new PercentProgressMonitor();
    }

    /**
     * Returns a progress monitor for this flux interval computer.
     * 
     * @return a progress monitor for this flux interval computer.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    /**
     * Computes and returns the intervals of every flux in the fluxome used to
     * construct this flux interval computer.
     * 
     * @return reaction names mapped to the computed interval of their
     * corresponding flux.
     */
    public Map<String, Interval> getIntervals() {
        return getIntervals(getFluxome().getRxnNames());
    }

    /**
     * Computes and returns the flux intervals of the reactions with the
     * provided names.
     * 
     * @param rxnNames a collection of reaction names.
     * 
     * @return reaction names mapped to the computed interval of their
     * corresponding flux.
     */
    public Map<String, Interval> getIntervals(Collection<String> rxnNames) {

        try {

            int total = getFluxome().rxns() + rxnNames.size() * 2 + 1;

            // Starting progress monitor
            progress.resetCount();
            progress.setTotal(total);
            progress.start();

            // Getting all flux constraints
            Map<String, Interval> fluxConstraints = getFluxConstraints();
            
            Map<MathExpr<String>, Interval> exprConstraints =
                    getRxnExprConstraints();

            // Determining and setting upper bounds
            setUpperBounds(fluxConstraints);

            // Computing flux minima and maxima
            Map<String, Double> minima = computeMinima(rxnNames);
            Map<String, Double> maxima = computeMaxima(rxnNames);

            // Converting to intervals
            Map<String, Interval> intervals = getIntervals(minima, maxima);

            // Restoring constraints to their previous values
            setFluxConstraints(fluxConstraints);
            setRxnExprConstraints(exprConstraints);

            // Stopping progress monitor
            progress.complete();
            progress.stop();

            return intervals;

        } catch (RuntimeException e) {

            if (!progress.stopped()) {
                progress.stop();
            }

            throw e;
        }
    }

    private Map<String, Double> computeMinima(Collection<String> reactionNames) {
        return getFluxMinima(reactionNames, progress);
    }

    private Map<String, Double> computeMaxima(Collection<String> reactionNames) {

        // Creating storage
        Map<String, Double> maxima = new TreeMap<String, Double>();

        // Iterating thru each reaction name
        for (String rxn : reactionNames) {

            // Computing maximum flux with all reverse rxns
            // constrained to their minima
            double max = getMaximum(rxn);

            // Storing it
            maxima.put(rxn, max);

            // Incrementing progress
            progress.increment();
        }

        return maxima;
    }

    /**
     * Computes and returns the minimum fluxes of the reactions with the
     * provided names. The provided progress monitor is incremented after each
     * minimum flux is computed.
     * 
     * @param rxnNames a collection of reaction names.
     * @param progress the progress monitor to be incremented.
     * 
     * @return reaction names mapped to their corresponding minimum flux.
     */
    protected Map<String, Double> getFluxMinima(Collection<String> rxnNames,
            MutableProgressMonitor progress) {

        // Creating storage
        Map<String, Double> minima = new TreeMap<String, Double>();

        // Iterating thru each reaction name
        for (String rxn : rxnNames) {

            // Computing minimum flux
            double min = getMinimum(rxn);

            // Storing it
            minima.put(rxn, min);

            // Incrementing progress
            progress.increment();
        }

        return minima;
    }

    /**
     * Computes and returns the maximum fluxes of the reactions with the
     * provided names. Each flux is maximized while the fluxes of its reverse
     * reactions are constrained to their minima. The provided progress monitor
     * is incremented after each maximum flux is computed.
     * 
     * @param rxnNames a collection of reaction names.
     * @param progress the progress monitor to be incremented.
     * 
     * @return reaction names mapped to their corresponding maximum flux.
     */
    protected Map<String, Double> getFluxMaxima(Collection<String> rxnNames,
            									MutableProgressMonitor progress) {

        // Creating storage
        Map<String, Double> maxima = new TreeMap<String, Double>();

        // Iterating thru each reaction name
        for (String rxn : rxnNames) {

        	// Getting reverse reactions
        	Set<String> reverse = getReverseRxnNames(rxn);
        	
            // Computing maximum flux with all reverse rxns
            // constrained to their minima
            double max = getMaxWithRxnsMinimized(rxn, reverse);

            // Storing it
            maxima.put(rxn, max);

            // Incrementing progress
            progress.increment();
        }

        return maxima;
    }

    private Set<String> getReverseRxnNames(String rxnName) {
    	
    	IrrevRxnFluxome fluxome = getFluxome();
    	
    	if (fluxome.isInternalRxnName(rxnName)) {
    		
    		Map<String, IrrevRxn> rxns = fluxome.getInternalRxns();
    		
    		return RxnComparer.getReverseIrrevRxnIds(rxnName, rxns);
    	}
    	
    	if (fluxome.isSourceRxnName(rxnName)) {
    		
    		SourceRxn source = fluxome.getSourceRxn(rxnName);
    		Map<String, SinkRxn> rxns = fluxome.getSinkRxns();
    		
    		return RxnComparer.getReverseSinkRxnIds(source, rxns);
    	}
    	
    	if (fluxome.isSinkRxnName(rxnName)) {
    		
    		SinkRxn source = fluxome.getSinkRxn(rxnName);
    		Map<String, SourceRxn> rxns = fluxome.getSourceRxns();
    		
    		return RxnComparer.getReverseSourceRxnIds(source, rxns);
    	}
    	
    	String s = "Invalid reaction name: " + rxnName;
    	throw new IllegalArgumentException(s);
    }
     
    private void setUpperBounds(Map<String, Interval> currentConstraints) {

        // Computing upper bounds for every flux in the network
        Map<String, Double> bounds = getUpperBounds();

        // Using these upper bounds to define flux constraints
        for (Map.Entry<String, Double> e : bounds.entrySet()) {

            String rxnName = e.getKey();

            Interval c = currentConstraints.get(rxnName);

            double lower = c.getLower();
            double upper = e.getValue();

            // Creating new constraint
            if (upper >= lower) {
                c = new Interval(lower, upper);
            }

            // If lower bound > upper bound
            else {

                if (lower - upper > getZeroCutoff()) {
                    throw new IllegalBoundsException(rxnName, lower, upper);
                }

                // Assuming lower bound is more "accurate" than computed upper
                // bound, since the lower bound may have been defined by the
                // user.
                c = new Interval(lower, lower);
            }

            setFluxConstraint(rxnName, c);
        }

    }

    private Map<String, Double> getUpperBounds() {

        // Creating storage
        Map<String, Double> maxima = new TreeMap<String, Double>();

        // Iterating thru each reaction name
        for (String rxn : getFluxome().getRxnNames()) {

        	// Getting reverse reactions
        	Set<String> reverse = getReverseRxnNames(rxn);

            // Computing maximum flux with all reverse rxns
            // constrained to their minima
            double max = getMaxWithRxnsMinimized(rxn, reverse);

            // Storing it
            maxima.put(rxn, max);

            // Incrementing progress
            progress.increment();
        }

        return maxima;
    }

}
