/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.IrrevRxnFluxomeV1;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.MathExpr;
import sbrt.kernel.math.optimization.OptSense;

/**
 * This class is used to compute the effect of deleting reactions in a
 * fluxome.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * 
 * @param <F> a type of fluxome. 
 *
 */
public class RxnDeletionComputer<F extends Fluxome> extends AbstractCvComputer<F> {

    private final FbaOptimizer<F> computer;
    private FbaOptimizer<F>.Results wildtype;
    
    private double zeroCutoff;
    private OptSense sense;
    private MathExpr<String> objective;
    
    /**
     * Constructs a new reaction deletion computer.
     * 
     * @param computer the FBA optimization computer to use for all
     * optimizations.
     */
    public RxnDeletionComputer(FbaOptimizer<F> computer) {
        super(computer.getFluxome());

        // Setting computer
        this.computer = computer;

        // Setting cutoff to the most conservative possible value
        zeroCutoff = 0.0d;
    }

    /**
     * Sets the zero cutoff to the provided value. The zero cutoff is the 
     * value used to decide which computed values are close enough to zero
     * to be considered equal to zero.  This value is an upper bound and 
     * should itself be close to zero, like 1E-6 for example.
     * 
     * @param zeroCutoff the zero cutoff.
     */
    public void setZeroCutoff(double zeroCutoff) {

        // Checking input
        if (Double.isInfinite(zeroCutoff) || 
        	Double.isNaN(zeroCutoff) ||
            zeroCutoff < 0d) {

        	String s = "Invalid zero cutoff: " + zeroCutoff;
            throw new IllegalArgumentException(s);
        }

        // Setting
        this.zeroCutoff = zeroCutoff;
    }
    

	/**
	 * Sets the constraint tolerance.
	 *
	 * @param tolerance the amount by which computed values can 
	 * violate their defined constraints.
	 * 
	 * @throws IllegalArgumentException if the provided value is NaN, 
	 * infinite, or less than 0.
	 */
    public void setTolerance(double tolerance) {
    	computer.setTolerance(tolerance);
    }

	/**
	 * Sets the optimization sense to the provided value.
	 * @param sense an optimization sense.
	 */
    public void setOptimizationSense(OptSense sense) {
    	
    	if (this.sense != sense) {
    		this.sense = sense;
    		wildtype = null;
    	}
    	
    }

	/**
	 * Sets the objective function to the provided mathematical expression.
	 * @param objective  a mathematical expression of reaction names.
	 */
    public void setObjectiveFunction(MathExpr<String> objective) {
    	this.objective = objective;
    	wildtype = null;
    }

    /**
     * Returns the current objective function.
     * @return the current objective function.
     */
    public MathExpr<String> getObjectiveFunction() {
		return objective;
    }

    /**
     * Returns the current optimization sense.
     * @return the current optimization sense.
     */
    public OptSense getOptimizationSense() {
		return sense;
    }

    /**
     * Returns the current zero cutoff.
     * @return the current zero cutoff.
     * @see #setZeroCutoff(double)
     */
    public double getZeroCutoff() {
        return zeroCutoff;
    }

    /**
     * Returns the constraint tolerance.
     * @return the constraint tolerance.
     */
    public double getTolerance() {
    	return computer.getTolerance();
    }
    
    /**
     * Ensures all required information has been supplied to this
     * computer.
     */
    private void checkState() {

        if (sense == null) {
        	String s = "An optimization sense was never set.";
        	throw new IllegalStateException(s);
        }

        if (objective == null) {
            String s = "An objective function was never set.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Computes and returns the results of deleting the reactions whose names
     * appear in the provided set.
     * 
     * @param rxnNames the names of reactions whose effect upon deletion
     * will be computed.
     * 
     * @return the results of deleting the reactions whose names
     * occur in the provided set.
     * 
     * @throws IllegalStateException if no objective function or optimization
     * sense were provided.
     */
    public FbaOptimizer<F>.Results delete(Set<String> rxnNames) {

    	checkState();
    	
    	// If no wild-type flux vector exists, compute one
    	if (wildtype == null) {
    		wildtype = computer.getOptimal(objective, sense);
    	}
    	
        // Return previous results if computing the effects of
        // these deletions isn't necessary
        if (!necessaryDeletions(rxnNames)) {
            return wildtype;
        }

        // Getting current flux constraints
        Map<String, Interval> original = getFluxConstraints();

        // Getting deletion constraints
        Map<String, Interval> deletions = getDeletionConstraints(rxnNames);

        // Setting deletion constraints
        computer.setFluxConstraints(deletions);

        // Computing optimal solution
        FbaOptimizer<F>.Results results = computer.getOptimal(objective, sense);

        // Restoring original flux constraints
        computer.setFluxConstraints(original);

        // Returing computed results
        return results;
    }

    private boolean necessaryDeletions(Set<String> rxnNames) {

        // If no previous results exist new results must be computed
        if (wildtype == null) {
            return true;
        }

        // Getting the previous flux vector
        Map<String, Double> fluxVector = wildtype.getFluxVector();

        // Iterating thru each reaction name
        for (String rxnName : rxnNames) {

            // Getting the previously computed flux
            Double flux = fluxVector.get(rxnName);

            // Ensuring the provided reaction name is valid
            if (flux == null) {
                String s = "The reaction '" + rxnName + "' doesn't exist.";
                throw new IllegalArgumentException(s);
            }

            // If the absolute value of this flux is greater than the zero 
            // cutoff, the effects of reaction deletions must be computed
            if (Math.abs(flux) > zeroCutoff) {
                return true;
            }
        }

        // If all previously computed fluxes equal zero, their deletion
        // will have no effect (i.e. they are unnecessary)
        return false;
    }

    private final Map<String, Interval> 
    	getDeletionConstraints(Collection<String> reactionNames) {

    	// Determining appropriate size
        int size = reactionNames.size() + 1;

        // Creating storage for deletion constraints
        Map<String, Interval> fluxConstraints = 
        	new LinkedHashMap<String, Interval>(size, 1.0F);

        // Iterating thru each reaction name
        for (String rxn : reactionNames) {

            // Storing the deletion constraint
            fluxConstraints.put(rxn, IrrevRxnFluxomeV1.OFF);
        }

        // Returning all deletion constraints
        return fluxConstraints;
    }

    /**
     * Returns the set of all constraints on mathematical expressions of
     * fluxes.
     * 
     * @return mathematical expressions of reaction names mapped to their
     * corresponding constaint.
     */
    public Map<MathExpr<String>, Interval> getRxnExprConstraints() {
        return computer.getRxnExprConstraints();
    }

    /**
     * Returns the set of all constraints on single fluxes.
     * 
     * @return the names of reaction in the fluxome mapped to their
     * corresponding flux constaint.
     */
    public Map<String, Interval> getFluxConstraints() {
        return computer.getFluxConstraints();
    }

    /**
     * Returns the constraint on the provided mathematical expression of
     * fluxes.
     * 
     * @param rxnNameExp a mathematical expression of reaction names.
     * 
     * @return the constraint on the provided mathematical expression of
     * fluxes.
     */
    public Interval getRxnExprConstraint(MathExpr<String> rxnNameExp) {
        return computer.getRxnExprConstraint(rxnNameExp);
    }

    /**
     * Returns the flux constraint of the reaction with the provided name.
     * 
     * @param rxnName the name of the reaction whose flux constraint is to
     * be returned.
     * 
     * @return the flux constaint of the reaction with the provided name.
     */
    public Interval getFluxConstraint(String rxnName) {
        return computer.getFluxConstraint(rxnName);
    }

    /**
     * Changes the flux constraint of the reaction with the provided name.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be changed.
     * 
     * @param c the new constraint.
     */
    public void setFluxConstraint(String reactionName, Interval c) {

        // Destroying previous results
        wildtype = null;

        computer.setFluxConstraint(reactionName, c);
    }

    /**
     * Changes the constraint on the provided mathematical expression 
     * of fluxes.
     * 
     * @param e a matheamtical expression of reaction names.
     * @param c the new constraint.
     */
    public void setRxnExprConstraint(MathExpr<String> e, Interval c) {

        // Destroying previous results
        wildtype = null;

        computer.setRxnExprConstraint(e, c);
    }

    /**
     * Restores the constraints on all flux expressions to the values
     * contained in the fluxome.
     */
    public void restoreRxnExprConstraints() {

        // Destroying previous results
        wildtype = null;

        computer.restoreRxnExprConstraints();
    }

    /**
     * Restores the flux constraint for the reaction with the provided 
     * name to the value contained in the fluxome.
     * 
     * @param reactionName the name of the reaction whose flux constraint
     * is to be restored.
     */
    public void restoreFluxConstraint(String reactionName) {

        // Destroying previous results
        wildtype = null;

        computer.restoreFluxConstraint(reactionName);
    }

    /**
     * Restores the constraint on the provided flux expression to the value
     * contained in the fluxome.
     * 
     * @param rxnNameExp the expression of reaction names whose contraint
     * is to be restored.
     */
    public void restoreRxnExprConstraint(MathExpr<String> rxnNameExp) {

        // Destroying previous results
        wildtype = null;

        computer.restoreRxnExprConstraint(rxnNameExp);
    }

}
