/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.*;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;

/**
 * This class is used to compute flux intervals in the simplest possible way. An
 * interval is computed by minimizing and maximizing the flux of a given
 * reaction. If a flux is unbounded in a particular direction, as they sometimes are,
 * {@code Double.NEGATIVE_INFINITY} or 
 * {@code Double.POSITIVE_INFINITY} are used to denote the flux bounds.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * @param <F> the fluxome type.
 * 
 */
public class SimpleFluxIntervalComputer<F extends Fluxome> 
	extends AbstractFluxIntervalComputer<F> {

    private final PercentProgressMonitor progress;

    /**
     * Constructs a new flux interval computer.
     * 
     * @param computer the computer with which all optimizations will be
     * computed.
     * @param zeroCutoff the amount by which a computed lower bound can
     * <em>exceed</em> a computed upper bound.
     */
    public SimpleFluxIntervalComputer(FbaOptimizer<F> computer,
                                      double zeroCutoff) {

        super(computer, zeroCutoff);
        progress = new PercentProgressMonitor();

    }

    /**
     * Returns a progress monitor for this flux interval computer.
     * 
     * @return a progress monitor for this flux interval computer.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    /**
     * Computes and returns the intervals of every flux in the fluxome used to
     * construct this flux interval computer.
     * 
     * @return reaction names mapped to the computed interval of their
     * corresponding flux.
     */
    public Map<String, Interval> getIntervals() {
        return getIntervals(getFluxome().getRxnNames());
    }

    /**
     * Computes and returns the flux intervals of the reactions with the
     * provided names.
     * 
     * @param rxnNames a collection of reaction names.
     * 
     * @return reaction names mapped to the computed interval of their
     * corresponding flux.
     */
    public Map<String, Interval> getIntervals(Collection<String> rxnNames) {

        try {

            // Starting progress monitor
            progress.resetCount();
            progress.setTotal(rxnNames.size() * 2 + 1);
            progress.start();

            // Getting flux minima and maxima
            Map<String, Double> minima = getMinimumFluxes(rxnNames);
            Map<String, Double> maxima = getMaximumFluxes(rxnNames);

            // Converting to intervals
            Map<String, Interval> intervals = getIntervals(minima, maxima);

            // Stopping progress monitor
            progress.complete();
            progress.stop();

            return intervals;

        } catch (RuntimeException e) {
            
            if (!progress.stopped()) {
                progress.stop();
            }
            
            throw e;
            
        }

    }

    /**
     * Computes and returns the minimum fluxes of the reactions with the
     * provided names.
     * 
     * @param rxnNames a collection of reaction names.
     * 
     * @return reaction names mapped to their corresponding minimum flux.
     */
    private Map<String, Double> getMinimumFluxes(Collection<String> rxnNames) {

        // Creating storage
        Map<String, Double> minima =
                new LinkedHashMap<String, Double>(rxnNames.size() + 1, 1.0F);

        // Iterating thru each reaction name
        for (String rxn : rxnNames) {

            // Computing minimum flux
            double min = getMinimum(rxn);

            // Storing it
            minima.put(rxn, min);

            // Incrementing progress
            progress.increment();
        }

        return minima;
    }

    /**
     * Computes and returns the maximum fluxes of the reactions with the
     * provided names.
     * 
     * @param rxnNames a collection of reaction names.
     * 
     * @return reaction names mapped to their corresponding maximum flux.
     */
    private Map<String, Double> getMaximumFluxes(Collection<String> rxnNames) {
        return getMaximumFluxes(rxnNames, progress);
    }

    /**
     * Computes and returns the maximum fluxes of the reactions with the
     * provided names. The provided progress monitor is incremented after
     * each maximum flux is computed.
     * 
     * @param rxnNames a collection of reaction names.
     * @param progress the progress monitor to be incremented.
     * 
     * @return reaction names mapped to their corresponding maximum flux.
     */
    public Map<String, Double> 
    getMaximumFluxes(Collection<String> rxnNames, MutableProgressMonitor progress) {

        // Creating storage
        Map<String, Double> maxima =
                new LinkedHashMap<String, Double>(rxnNames.size() + 1, 1.0F);

        // Iterating thru each reaction name
        for (String rxn : rxnNames) {

            // Computing minimum flux
            double max = getMaximum(rxn);

            // Storing it
            maxima.put(rxn, max);

            // Incrementing progress
            progress.increment();
        }

        return maxima;
    }

}
