/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.Set;

import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.IrrevRxnFluxomeV1;
import sbrt.kernel.math.algebra.SparseDoubleMatrix;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to identify the steady state cycles in stoichiometric
 * networks. This is accomplished by removing all exchange reactions from 
 * a network and then identifying the extreme currents of the sub-network.
 * Those extreme currents are necessarily cycles, and are also the set of
 * all Type III extreme pathways of the sub-network. 
 * 
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public class SlpCycleComputer implements CycleComputer<IrrevRxnFluxome> {

	private final ExtremeCurrentComputer computer;

	/**
	 * Constructs a new steady state cycle computer from the given extreme
	 * current computer.
	 * 
	 * @param computer
	 *            the computer with which all extreme currents will be identified.
	 */
	public SlpCycleComputer(ExtremeCurrentComputer computer) {

		// Setting extreme pathway computer
		Util.nullCheck(computer);
		this.computer = computer;

	}

	/**
	 * Returns a progress monitor for this steady state cycle computer.
	 * 
	 * @return a progress monitor for this steady state cycle computer.
	 */
	public ProgressMonitor getProgress() {
		return computer.getProgress();
	}

	/**
	 * Returns the set of all steady state cycles contained in the provided fluxome.
	 * 
	 * @param fluxome
	 *            the fluxome for which all cycles will be identified.
	 * 
	 * @return the extreme current matrix. The column variables are the names
	 * of all internal reactions, and each row corresponds to a steady state cycle.
	 */
	public SparseDoubleMatrix<?, String> getCycles(IrrevRxnFluxome fluxome) {

		// If the fluxome contains exchange reactions
		if (fluxome.exchangeRxns() > 0) {

			// Getting only the names of internal reactions
			Set<String> rxnNames = fluxome.getInternalRxnNames();

			// Creating a fluxome with no exchange reactions
			fluxome = IrrevRxnFluxomeV1.make(rxnNames, fluxome);

		}

		// Computing and returning the cycles
		return computer.getMatrix(fluxome);
	}

}
