/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.io.WarningHandlerFactory;
import sbrt.kernel.math.algebra.MatrixTransposer;
import sbrt.kernel.math.algebra.RowOrTreeMatrix;
import sbrt.kernel.math.algebra.SmdMatrix;
import sbrt.kernel.math.algebra.SparseDoubleMatrix;
import sbrt.kernel.utilities.MutableProgressMonitor;
import sbrt.kernel.utilities.PercentProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is an implementation of an algorithm for identifying extreme
 * currents in stoichiometric networks. This algorithm was described in the
 * paper: C. Schilling, D. Letscher and B. O. Palsson, Theory for systemic
 * definition of metabolic pathways and their use in intrepreting metabolic
 * function from a pathway-oriented perspective. <i>Journal of Theoretical
 * Biology</i><b> 203 </b>(2000), pp. 285-301. This algorithm was intended for
 * computing <i>extreme pathways</i>, but it has been modified here to compute
 * <i>extreme currents</i> instead. A brief mention of extreme currents can be
 * found in the paper cited above.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public class SlpExCurrComputerV2 implements ExtremeCurrentComputer {

    private static final double zeroCutoff = 1E-90;
    
    private final PercentProgressMonitor progress;
    private int indexCount;

    // Temporary storage
    private SmdMatrix<Integer, String> tableau;
    private SmdMatrix<Integer, String> nextTableau;
    private Set<String> allColumnIndices;

    /**
     * Constructs a new extreme current computer.
     */
    public SlpExCurrComputerV2() {
        progress = new PercentProgressMonitor();
    }

    /**
     * Returns a progress monitor for this extreme current computer. 
     * @return a progress monitor for this extreme current computer. 
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }
    
    /**
     * Returns the total number of "operations" required to compute the extreme
     * currents of the provided fluxome. The number returned by this method is meant
     * to be used with progress monitors.
     * 
     * @param fluxome the fluxome for which the number of operations will be returned.
     * 
     * @return the total number of operations required to compute the extreme
     * currents of the provided fluxome. 
     */
    public int getTotalOps(IrrevRxnFluxome fluxome) {
        return fluxome.species();
    }

    /**
     * Returns the extreme current matrix for the provided fluxome.
     * 
     * @param fluxome the fluxome for which all extreme currents will be 
     * computed.
     * 
     * @return the extreme current matrix for the provided fluxome.
     */
    public SparseDoubleMatrix<?, String> getMatrix(IrrevRxnFluxome fluxome) {
        
        // Initializing progress monitor
        progress.resetCount();
        progress.setTotal(fluxome.species());
        progress.start();
        
        // Computing the extreme current matrix
        SparseDoubleMatrix<?, String> matrix = getMatrix(fluxome, progress);

        // Stopping timer and setting progress to 100%
        progress.complete();
        progress.stop();
        
        return matrix;
    }

    /**
	 * Returns the extreme current matrix for the provided fluxome.
	 * 
	 * @param fluxome
	 *            the fluxome for which all extreme currents will be computed.
	 * @param pm
	 *            the progress monitor to be used during computation of extreme
	 *            currents.
	 * 
	 * @return the extreme current matrix for the provided fluxome. Column
	 *         variables are reaction names, and each row corresponds to an extreme
	 *         current.
	 */
    public SparseDoubleMatrix<?, String> getMatrix(IrrevRxnFluxome fluxome, 
                                                   MutableProgressMonitor pm) {

        boolean debug = false;
        
        if (debug) {
            printInfo(fluxome);
        }
        
        // Initializing row index counter
        indexCount = Integer.MIN_VALUE;

        // Getting the initial tableau
        tableau = getInitialTableau(fluxome);
        
        // Getting all column indices
        allColumnIndices = tableau.getColumnIndices();
        
        // Getting an ordered list of chemical species (i.e. column indices)
        List<String> species = 
            getOrderedSpecies(fluxome.getSpecies(), tableau);

        // Iterating thru each chemical species
        for (String m : species) {

            // Creating storage space for next tableau
            nextTableau = new RowOrTreeMatrix<Integer, String>();
            
            // Creating a list of all row indices of the current tableau
            List<Integer> rowIndices = new ArrayList<Integer>(tableau.getRowIndices());
            
            step2(m, rowIndices);
            step3(m, rowIndices);
            
            // The current tableau and its row indices are no longer needed
            tableau = null;
            rowIndices = null;
            
            step4();

            // Setting the "next" tableau to be the "current" tableau
            tableau = nextTableau;

            // Incrementing progress
            pm.increment();
        }

        // Freeing up temporary storage for garbage collection
        SmdMatrix<Integer, String> temp = tableau;
        tableau = null;
        nextTableau = null;
        allColumnIndices = null;

        // Removing all chemical species columns
        for (String m : species) {
        	temp.removeColumn(m);
        }
        
        // Returning extreme pathway matrix
        return temp;
    }

    private Integer getNewIndex() {
        
        if (indexCount == Integer.MAX_VALUE) {
            throw new IllegalStateException("arithmetic overflow has occurred");
        }
        
        Integer index = Integer.valueOf(indexCount ++);
        
        return index;
    }
    
    private void printInfo(IrrevRxnFluxome fluxome) {

        int nonZeroEntries = 0;
        
        SparseDoubleMatrix<String, String> s = fluxome.getStoichMatrix();
        
        for (String rowIndex : s.getRowIndices()) {
            nonZeroEntries += s.getColIndNzEl(rowIndex).size();
        }
        
        Util.debugPrintln("Unbounded Fluxome Data");
        Util.debugPrintln("Chemical Species = " + fluxome.species());
        Util.debugPrintln("Reactions = " + fluxome.rxns());
        Util.debugPrintln("Non-Zero Entries = " + nonZeroEntries);
        
    }

    private List<String> 
    getOrderedSpecies(Set<String> species,
    				  final SparseDoubleMatrix<?, ?> initTableau) {

        // Sorting chemical species according to the number of non-zero
        // elements in the corresponding tableau column (i.e. their
        // degree). Columns with fewer non-zero entries will appear
        // first. From my own personal experience, this seems to greatly
        // reduce execution time. - J. Wright

    	// Creating list of all species
        List<String> list = new ArrayList<String>(species);

        // First sorting them alphabetically.
        // This ensures reproducible behavior regardless of the implementation
        // of the provided species Set. 
        Collections.sort(list);
        
        // Now sorting them by degree, preserving the order of equal elements.
        Collections.sort(list, new Comparator<String>() {

            public int compare(String m1, String m2) {

            	// Getting the number of non-zero elements for each species
                int nz1 = initTableau.getRowIndNzEl(m1).size();
                int nz2 = initTableau.getRowIndNzEl(m2).size();

                return nz1 - nz2;
            }

        });

        // Returing the sorted list
        return list;
    }

    private SmdMatrix<Integer, String> getInitialTableau(IrrevRxnFluxome fluxome) {

        // Creating storage for tableau
        SmdMatrix<String, String> tableau = new RowOrTreeMatrix<String, String>();
        
        // Filling the identity matrix section
        for (String rxnName : fluxome.getRxnNames()) {
            tableau.setElement(rxnName, rxnName, Double.valueOf(1d));
        }

        // Getting the stoichiometry matrix S of the fluxome
        SparseDoubleMatrix<String, String> s = fluxome.getStoichMatrix();

        // Creating storage for the transpose of S
        SmdMatrix<String, String> sTrans = new RowOrTreeMatrix<String, String>();

        // Getting the transpose of S
        new MatrixTransposer().transpose(s, sTrans);
        
        // Filling the S transpose section
        tableau.setElements(sTrans);

        // Returning the tableau after replacing row indices with Integers
        return replaceRowIndices(tableau);
    }

    private SmdMatrix<Integer, String> replaceRowIndices(SmdMatrix<String, String> m) {
        
        // Creating storage
        SmdMatrix<Integer, String> matrix = new RowOrTreeMatrix<Integer, String>();
        
        // Getting row and column indices
        Set<String> rowIndices = m.getRowIndices();
        Set<String> colIndices = m.getColumnIndices();
        
        // Iterating thru each row index
        for (String rowIndex : rowIndices) {
            
            // Creating a unique integer index
            Integer newIndex = getNewIndex();
            
            // Iterating thru each column index
            for (String columnIndex : colIndices) {
                
                // Getting element
                Double value = m.getElement(rowIndex, columnIndex);
                
                // Setting element
                matrix.setElement(newIndex, columnIndex, value);
                
            }
            
        }
        
        return matrix;
    }
    
    private void step2(String species, 
                       List<Integer> allRowIndices) {
        
        // Iterating thru each row index in the current tableau
        for (Iterator<Integer> i = allRowIndices.iterator(); i.hasNext(); ) {
            
            Integer rowIndex = i.next();
            
            // Getting the element at this row and column
            Double value = tableau.getElement(rowIndex, species);
            
            // If it equals zero
            if (Math.abs(value) < zeroCutoff) {

                // Copying the row into the next tableau
                copyRow(tableau, rowIndex, nextTableau);

                // Removing the row from the current tableau
                tableau.removeRow(rowIndex);
                i.remove();
            }
            
        }
        
    }

    private void step3(String species, List<Integer> rowIndices) {

        // Iterating thru each row index
        for (int i = 0, size = rowIndices.size(); i < size; i++) {

            // Getting next row index
            Integer r1Index = rowIndices.get(i);

            // Getting the element at this row and column in the current tableau
            Double r1Element = tableau.getElement(r1Index, species);

            // Iterating thru each succeeding row index
            for (int j = i + 1; j < size; j++) {

                // Getting next row index
                Integer r2Index = rowIndices.get(j);

                // Getting the element at this row and column in the current tableau
                Double r2Element = tableau.getElement(r2Index, species);

                // If the elements do not have opposite signs, move on to the next
                if (!Util.oppositeSign(r1Element, r2Element)) {
                    continue;
                }

                // Copying both rows into the next tableau for temporary storage
                copyRow(tableau, r1Index, nextTableau);
                copyRow(tableau, r2Index, nextTableau);

                // Multiplying both rows by constants
                multiplyRow(nextTableau, r1Index, Math.abs(r2Element));
                multiplyRow(nextTableau, r2Index, Math.abs(r1Element));

                // Adding the rows together to form a new row r'
                Integer rPrimeIndex = getNewIndex();
                addRows(nextTableau, r1Index, r2Index, rPrimeIndex);

                // Removing the original two rows
                nextTableau.removeRow(r1Index);
                nextTableau.removeRow(r2Index);
            }
        }

    }

    private void step4() {

    	// Getting all row indices in the next tableau
        List<Integer> rowIndices = 
        	new ArrayList<Integer>(nextTableau.getRowIndices());

        // Iterating thru each row in the next tableau
        for (int i = 0, size = rowIndices.size(); i < size; i++) {

            // Getting the next row index r1
            Integer r1Index = rowIndices.get(i);

            // If it's no longer an index, skip it
            if (!nextTableau.isRowIndex(r1Index)) {
                continue;
            }

            // Getting the column indices for which the elements
            // in row 1 are not equal to zero
            Set<String> row1ColumnIndices =
                nextTableau.getColIndNzEl(r1Index);

            // Iterating thru each succeeding row index
            for (int j = 0; j < size; j++) {

                // Not comparing the row to itself
                if (j == i) {
                    continue;
                }

                // Getting the next row index r2
                Integer r2Index = rowIndices.get(j);

                // If it's no longer an index, skip it
                if (!nextTableau.isRowIndex(r2Index)) {
                    continue;
                }

                // Getting the column indices for which the elements
                // in row 2 are not equal to zero
                Set<String> row2ColumnIndices =
                    nextTableau.getColIndNzEl(r2Index);

                // If row 1 has more *non-zero* elements than row 2,
                // it can't possibly be a non-negative linear combination
                // of other rows.
                int columns = nextTableau.columns();
                int row1ZeroElements = columns - row1ColumnIndices.size();
                int row2ZeroElements = columns - row2ColumnIndices.size();

                // If row 2 has more *zero* elements than row 1,
                // the corresponding column indices in row 2 cannot possibly
                // be a subset of the corresponding column indices in row 1.
                // So row 2 will not be removed.
                if (row2ZeroElements > row1ZeroElements) {
                    continue;
                }

                // If the column indices of *non-zero* elements in row 1
                // is a subset of the column indices of *non-zero* elements
                // in row 2, then the column indices of *zero* elements in
                // row 2 is a subset of the column indices of *zero* elements
                // in row 1.  So row 2 gets removed.
                if (row2ColumnIndices.containsAll(row1ColumnIndices)) {
                    nextTableau.removeRow(r2Index);
                }
            }
        }

    }

    private void addRows(SmdMatrix<Integer, String> m,
                         Integer rowIndex1,
                         Integer rowIndex2,
                         Integer newRowIndex) {

        // Iterating thru each column index
        for (String columnIndex : allColumnIndices) {
            
            // Getting the values in each row
            double value1 = m.getElement(rowIndex1, columnIndex);
            double value2 = m.getElement(rowIndex2, columnIndex);
            
            // Computing the sum
            double sum = value1 + value2;
            
            // Attempting to eliminate numerical error if it exists
            sum = removeError(sum);

            // Setting the value in the matrix
            m.setElement(newRowIndex, columnIndex, sum);
        }

    }

    private void multiplyRow(SmdMatrix<Integer, String> m,
                             Integer rowIndex, 
                             double constant) {

        // Iterating thru each column index
        for (String columnIndex : allColumnIndices) {
            
            // Multiplying the value at this row and column by the constant
            double value = m.getElement(rowIndex, columnIndex) * constant;
            
            // Attempting to eliminate numerical error if it exists
            value = removeError(value);

            // Setting the value in the matrix
            m.setElement(rowIndex, columnIndex, value);

        }

    }

    private void copyRow(SparseDoubleMatrix<Integer, String> tableau,
                         Integer rowIndex,
                         SmdMatrix<Integer, String> nextTableau) {
        
        // Iterating thru each column index
        for (String columnIndex : allColumnIndices) {
            
            // Getting the value at this row and column
            Double value = tableau.getElement(rowIndex, columnIndex);
            
            // Setting the value
            nextTableau.setElement(rowIndex, columnIndex, value);
        }
        
    }
    
    private double removeError(double value) {
        
        if (Math.abs(value) < zeroCutoff) {

            if (value != 0d) {
                String s = value + " was considered equal to zero."; 
                WarningHandlerFactory.getDefault().warn(s);
            }
            
            return 0d;
        }
        
        return value;
    }
    
}
