/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.comp;

import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.IrrevRxnFluxomeV1;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.MathExpr;
import sbrt.kernel.math.optimization.ProgramSolver;
import sbrt.kernel.utilities.PercentProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to reduce the size of fluxomes for the purpose of
 * identifying the steady state cycles they contain. It is an implementation of
 * the algorithm described by Wright and Wagner (WW).
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public class WwFluxomeReducer implements FluxomeReducer<IrrevRxnFluxome> {


    private PercentProgressMonitor progress;

    private final ProgramSolver<Object, Object> programSolver;

    private IrrevRxnFluxome fluxome;
    
    /**
     * Constructs a fluxome reducer from the provided linear
     * program solver.
     * 
     * @param solver the linear program solver with which all optimizations will
     * be computed.
     */
    public WwFluxomeReducer(ProgramSolver<Object, Object> solver) {

        // Checking and setting program solver
        Util.nullCheck(solver);
        programSolver = solver;

        // Initializing progress monitors
        progress = new PercentProgressMonitor();
    }

    /**
     * Returns a progress monitor for this fluxome reducer.
     * @return a progress monitor for this fluxome reducer.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }


    /**
     * Returns the sub-fluxome resulting from network reduction.
     * @param fluxome the fluxome for which a sub-fluxome will be returned.
     * @return the sub-fluxome resulting from network reduction.
     */
    public IrrevRxnFluxome makeFluxome(IrrevRxnFluxome fluxome) {

    	// Storing temporarily
    	this.fluxome = fluxome;
    	
    	// Initializing the progress monitor
        progress.start();
        progress.resetCount();
        progress.setTotal(fluxome.internalRxns() + 1);
        
        // Determining the unbounded fluxes
        Map<String, Double> maxima = getFluxMaxima();

        // Creating a sub-fluxome of the reactions with unbounded flux
        IrrevRxnFluxome sub = getUnboundedFluxome(maxima);
        
        // Stopping the progress monitor
        progress.stop();
        progress.complete();
        
        // Allowing garbage collection
        this.fluxome = null;
        
        return sub;
    }
    
    
    private Map<String, Double> getFluxMaxima() {

        // Creating optimization computer
        FbaOptimizer<IrrevRxnFluxome> optComputer =
                new FbaOptimizer<IrrevRxnFluxome>(fluxome, programSolver);

        // Creating reverse reaction flux range computer
        ReverseRxnIntervalComp<IrrevRxnFluxome> rangeComputer =
           new ReverseRxnIntervalComp<IrrevRxnFluxome>(optComputer, 1E-5d);

        // Setting appropriate constraints
        setConstraints(rangeComputer);

        // Getting all internal reaction names
        Set<String> rxns = fluxome.getInternalRxnNames();

        // Getting maximum fluxes for each internal reaction
        return rangeComputer.getFluxMaxima(rxns, progress);
    }
    

    private void setConstraints(FluxIntervalComputer<IrrevRxnFluxome> computer) {

        // Creating constraint for internal reactions
        Interval on = new Interval(0d, Double.POSITIVE_INFINITY);

        // Getting internal reaction names
        Set<String> internalRxns = computer.getFluxome().getInternalRxnNames();

        // Changing constraints for all internal reactions
        for (String rxnName : internalRxns) {
            computer.setFluxConstraint(rxnName, on);
        }

        // Creating constraint for all reaction name expressions
        on = new Interval(Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY);

        // Getting all existing reaction name expressions
        Set<? extends MathExpr<String>> expressions =
                computer.getRxnExprConstraints().keySet();

        // Changing constraints for all reaction name expressions
        for (MathExpr<String> expr : expressions) {
            computer.setRxnExprConstraint(expr, on);
        }

        // Creating constraint for exchange reactions
        Interval off = new Interval(0d, 0d);

        // Getting exchange reaction names
        Set<String> exchangeRxns = computer.getFluxome().getExchangeRxnNames();

        // Changing constraints for all exchange reactions
        for (String rxnName : exchangeRxns) {
            computer.setFluxConstraint(rxnName, off);
        }

    }

    private IrrevRxnFluxome getUnboundedFluxome(Map<String, Double> maxima) {

        // Creating storage for reactions with unbounded fluxes
        Set<String> rxns = new TreeSet<String>();

        // Iterating thru all reactions and their maximum fluxes
        for (Map.Entry<String, Double> e : maxima.entrySet()) {

            // If the flux is unbounded
            if (e.getValue().isInfinite()) {
                rxns.add(e.getKey());
            }

        }

        // Returning a new fluxome created from the unbounded fluxes
        return IrrevRxnFluxomeV1.make(rxns, fluxome);
    }



    
}
