/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.fba.comp.*;
import sbrt.kernel.math.*;

/**
 * This class is used to represent constraint-variation Flux Balance Analysis
 * processes.  It offers two static methods to aid in this sort of process.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type.
 * 
 * @param <R> the results type. 
 *
 */
public abstract class AbstractCvFbaOptProcess<F extends Fluxome, R> 
	extends FbaProcess<F, R> {

	/**
	 * Sets the provided constraints in the provided constraint-variation 
	 * computer.
	 * 
	 * @param constraints the set of flux constraints.
	 * @param computer the computer in which to set the provided constraints.
	 */
    public static void 
    setConstraints(FluxConstraints constraints, FluxCvComputer<?> computer) {

    	// Setting flux constraints
    	computer.setFluxConstraints(constraints.getFluxConstraints());
    	
    	// Setting expression constraints
    	computer.setRxnExprConstraints(constraints.getExprConstraints());

    }

	/**
	 * Restores the provided constraints in the provided constraint-variation 
	 * computer.
	 * 
	 * @param constraints the set of flux constraints to be restored.
	 * @param computer the computer in which to restore the provided 
	 * constraints.
	 */
    public static void 
    restoreConstraints(FluxConstraints constraints, FluxCvComputer<?> computer) {

    	// Getting reaction names
    	Set<String> rxnNames = constraints.getFluxConstraints().keySet();
    	
    	// Restoring flux constraints
        computer.restoreFluxConstraints(rxnNames);
    	
    	// Getting expressions
    	Set<MathExpr<String>> exprs = 
    		constraints.getExprConstraints().keySet();
    	
    	// Restoring expression constraints
        computer.restoreRxnExprConstraints(exprs);

//      computer.restoreFluxConstraints();
//      computer.restoreRxnExprConstraints();
    	
    }

}
