/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.fba.CatalyzedFluxome;
import sbrt.kernel.fba.comp.FbaOptimizer;
import sbrt.kernel.fba.comp.RxnDeletionComputer;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.MathExpr;
import sbrt.kernel.math.optimization.OptSense;
import sbrt.kernel.utilities.DiscreteProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to perform catalyst deletion processes.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * @param <F> the fluxome type.
 *
 */
public final class CatalystDeletionProcess<F extends CatalyzedFluxome> 
	extends FbaProcess<F, FbaOptimizer<F>.Results> 
        implements MonitoredProcess<FbaOptimizer<F>.Results>  {

    private final RxnDeletionComputer<F> computer;
    private final DiscreteProgressMonitor progress;
    
    private Iterable<Set<String>> deletionSource;
    
    /**
     * Constructs a new catalyst deletion process.
     * 
     * @param computer the computer with which all results will be computed.
     */
    public CatalystDeletionProcess(RxnDeletionComputer<F> computer) {

        // Checking and setting computer
        Util.nullCheck(computer);
        this.computer = computer;

        // Initializing progress monitor
        progress = new DiscreteProgressMonitor();
    }

	/**
	 * Returns the fluxome used by the reaction deletion computer.
	 * @return the fluxome used by the reaction deletion computer.
	 */
    public F getFluxome() {
        return computer.getFluxome();
    }

    /**
     * Returns the current zero cutoff used by the reaction deletion computer.
     * @return the current zero cutoff used by the reaction deletion computer.
     */
    public double getZeroCutoff() {
        return computer.getZeroCutoff();
    }

    /**
     * Returns the current objective function.
     * @return the current objective function.
     */
    public MathExpr<String> getObjectiveFunction() {
		return computer.getObjectiveFunction();
    }

    /**
     * Returns the current optimization sense.
     * @return the current optimization sense.
     */
    public OptSense getOptimizationSense() {
		return computer.getOptimizationSense();
    }

	/**
     * Sets the objective function to the provided mathematical expression.
     * 
     * @param objective a mathematical expression of reaction names.
     */
	public void setObjectiveFunction(MathExpr<String> objective) {
		computer.setObjectiveFunction(objective);
	}

    /**
     * Sets the optimization sense to the provided value.
     * 
     * @param sense an optimization sense.
     */
	public void setOptimizationSense(OptSense sense) {
		computer.setOptimizationSense(sense);
	}

    /**
     * Sets the catalyst deletion input source.
     * 
     * @param deletionSource a source of catalyst names sets whose effect upon
     * deletion is to be computed.
     */
    public void setCatalystDelSource(Iterable<Set<String>> deletionSource) {
        this.deletionSource = deletionSource;
    }

    /**
     * Sets the zero cutoff to the provided value. The zero cutoff is the value
     * used to decide which computed values are close enough to zero to be
     * considered equal to zero. This value is an upper bound and should itself
     * be close to zero, like 1E-6 for example.
     * 
     * @param zeroCutoff the zero cutoff.
     */
    public void setZeroCutoff(double zeroCutoff) {
        computer.setZeroCutoff(zeroCutoff);
    }

    private void checkState() {

        if (deletionSource == null) {
            String s = "An catalyst deletion source was never set.";
            throw new IllegalStateException(s);
        }
        
    }

    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

	/**
	 * Runs this process and adds the generated results to the provided results
	 * manager via its {@linkplain ResultsManager#addData(Object)} method.
	 * <p>
	 * A set of catalyst names is retrieved from the deletion source, their
	 * effect upon deletion is computed, and the results are added to the
	 * provided results manager. This is repeated until the end of the deletion
	 * source is reached.
	 * 
	 * @param resultsManager
	 *            the results manager to which the results will be added.
	 */
    public void run(ResultsManager<FbaOptimizer<F>.Results> resultsManager) {
    	
        checkState();

        try {
        	
        	// Initialized progress monitor
            progress.resetCount();
            progress.start();

            // Running the process
            runProcess(resultsManager);

            // Finalizing the progress monitor
            progress.stop();
            
        } catch (RuntimeException e) {

        	// Finalizing the progress monitor
            if (!progress.stopped()) {
                progress.stop();
            }
            
            // Throwing the original exception
            throw e;
        }
    }

    private void runProcess(ResultsManager<FbaOptimizer<F>.Results> rm) {

        // Iterating thru each set of catalysts to be deleted
    	for (Set<String> catalysts : deletionSource) {

            // Getting the associated reaction names
            Set<String> names = getRxnNames(catalysts, computer.getFluxome());

            // Computing the effects of deleting these reactions
            FbaOptimizer<F>.Results results = computer.delete(names);

            // Adding to the results manager
            rm.addData(results);

            // Incrementing progress
            progress.increment();
        }

    }

    /**
     * Returns the set of names of reactions whose activity depends on the
     * provided set of catalysts.
     * 
     * @param catalysts a set of catalyst names.
     * @param fluxome the fluxome containing the reaction-catalyst association.
     * 
     * @return the set of names of reactions whose activity depends on the
     * provided set of catalysts.
     */
    public static Set<String> 
    getRxnNames(Set<String> catalysts, CatalyzedFluxome fluxome) {

    	// Creating storage for the names of reactions to be deleted
        Set<String> allRxnNames = new TreeSet<String>();
        
        // Iterating thru each reaction name
        for (String rxnName : fluxome.getRxnNames()) {
            
            if (fluxome.areReqCatalysts(catalysts, rxnName)) {
                allRxnNames.add(rxnName);
            }
            
        }

        return allRxnNames;
    }

}
