/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.CdHarFluxComp;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.ExceededLimitException;
import sbrt.kernel.utilities.PercentProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to perform coordinate direction hit-and-run flux vector
 * processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in
 * the Wagner lab.
 * @param <F> the fluxome type.
 *
 */
public final class CdHarFluxVectorProcess<F extends Fluxome>
	extends FbaProcess<F, CdHarFluxComp<F>.Results> 
	implements MonitoredProcess<CdHarFluxComp<F>.Results> {

    private final CdHarFluxComp<F> computer;
    private final PercentProgressMonitor progress;

    private int iterations;
    private int maxWastedMoves;
    private int pointsPerPoint;

    /**
     * Constructs a new coordinate direction hit-and-run flux vector
     * process.
     * 
     * @param computer the computer with which all flux vectors will be 
     * generated.
     */
    public CdHarFluxVectorProcess(CdHarFluxComp<F> computer) {

        // Setting the Coordinate Direction Hit-and-Run computer
        Util.nullCheck(computer);
        this.computer = computer;

        // Initializing other fields
        progress = new PercentProgressMonitor();
        iterations = -1;
        maxWastedMoves = -1;
    }

    /**
     * Sets the total number of flux vectors to be generated.
     * @param iterations the total number of flux vector to be generated.
     */
    public void setIterations(int iterations) {

        // Checking input
        if (iterations < 0) {
            String s = "iterations is less than 0: " + iterations;
            throw new IllegalArgumentException(s);
        }

        // Setting
        this.iterations = iterations;
    }

    /**
     * Sets the maximum number of wasted moves to allow.
     * @param maxWastedMoves the number of wasted moves to allow.
     */
    public void setMaxWastedMoves(int maxWastedMoves) {

        // Checking input
        if (maxWastedMoves < 0) {
            String s = "maxWastedMoves is less than 0: " + maxWastedMoves;
            throw new IllegalArgumentException(s);
        }

        // Setting
        this.maxWastedMoves = maxWastedMoves;
    }

    /**
     * Sets the number of points to generate for each point added to the 
     * specified results manager.
     * 
     * @param pointsPerPoint the number of points to generate for each point added to the 
     * specified results manager.
     */
    public void setPointsPerPoint(int pointsPerPoint) {

        // Checking input
        if (pointsPerPoint < 0) {
            String s = "pointsPerPoint is less than 0: " + pointsPerPoint;
            throw new IllegalArgumentException(s);
        }

        // Setting
        this.pointsPerPoint = pointsPerPoint;
    }
    
    /**
     * Returns the total number of flux vectors to be generated.
     * @return the total number of flux vectors to be generated.
     */
    public int getIterations() {

        // Checking state
        if (iterations < 0) {
        	return 0;
        }

        return iterations;
    }

    /**
     * Returns the allowed number of wasted moves.
     * @return the allowed number of wasted moves.
     */
    public int getMaxWastedMoves() {

        // Checking state
        if (maxWastedMoves < 0) {
        	return 0;
        }

        return maxWastedMoves;
    }

    /**
     * Returns the number of points generated per points returned.
     * @return the number of points generated per points returned.
     */
    public int getPointsPerPoint() {

        // Checking state
        if (pointsPerPoint < 0) {
            return 0;
        }

        return pointsPerPoint;
    }

    /**
     * Returns the fluxome for which flux vectors are to be generated.
     * @return the fluxome for which flux vectors are to be generated.
     */
    public F getFluxome() {
        return computer.getFluxome();
    }

    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    private void checkState() {
    	
        if (iterations < 0) {
            String s = "The number of iterations was never set.";
            throw new IllegalStateException(s);
        }
        
        if (maxWastedMoves < 0) {
            String s = "The maximum number of wasted moves was never set.";
            throw new IllegalStateException(s);
        }
        
        if (pointsPerPoint < 0) {
            String s = "The number of points per point was never set.";
            throw new IllegalStateException(s);
        }
    }

    /**
     * Generates the specified number of random flux vectors and adds each to
     * the provided results manager.
     * 
     * @param resultsManager the results manager to which all computed vectors
     * will be added.
     */
    public void run(ResultsManager<CdHarFluxComp<F>.Results> resultsManager) {
    	
        checkState();

        try {

            // Initializing the progress monitor
            progress.resetCount();
            progress.setTotal(iterations);
            progress.start();

            // Running the process
            runProcess(resultsManager);
            
            // Finalizing the progress monitor
            progress.complete();
            progress.stop();

        } catch (RuntimeException e) {

            // Stopping the timer
            if (!progress.stopped()) {
                progress.stop();
            }

            // Throwing the original exception
            throw e;
        }
    }

    private final void runProcess(ResultsManager<CdHarFluxComp<F>.Results> rm) {

        // Initializing wasted moves counter
        int totalWastedMoves = 0;

        // Setting points per point
        computer.setPointsPerPoint(pointsPerPoint);
        
        // Generating the specified number of random flux vectors
        for (int i = 0; i < iterations; i++) {

            // Computing the next flux vector
            CdHarFluxComp<F>.Results results = computer.run();

            // Incrementing the number of wasted moves
            totalWastedMoves += results.unchangedFreeVariables();

            // If the total exceeds the user-specified maximum
            if (totalWastedMoves > maxWastedMoves) {
                throw new ExceededLimitException();
            }

            // Adding the results to the manager
            rm.addData(results);

            // Incrementing the progress
            progress.increment();
        }

    }
}
