/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.fba.FluxConstraints;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.FbaOptimizer;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.MathExpr;
import sbrt.kernel.math.optimization.OptSense;
import sbrt.kernel.utilities.DiscreteProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to perform constraint-variation FBA optimization
 * processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * @param <F> the fluxome type.
 * 
 */
public final class CvFbaOptProcess<F extends Fluxome> 
	extends FbaProcess<F, FbaOptimizer<F>.Results> 
	implements MonitoredProcess<FbaOptimizer<F>.Results> {

	// The computer with which all optimizations are performed.
	private final FbaOptimizer<F> computer;
    private final DiscreteProgressMonitor progress;
    
	private MathExpr<String> objFunction;
	private OptSense sense;
    private Iterable<FluxConstraints> constraintSource;

	/**
	 * Constructs a new constraint-variation FBA optimization process.
	 * 
	 * @param computer
	 *            the computer with which all results will be computed.
	 */
    public CvFbaOptProcess(FbaOptimizer<F> computer) {

		// Checking and setting computer
		Util.nullCheck(computer);
		this.computer = computer;
		
        progress = new DiscreteProgressMonitor();
    }

	/**
	 * Returns the fluxome used in this process.
	 * 
	 * @return the fluxome used in this process.
	 */
	public F getFluxome() {
		return computer.getFluxome();
	}

	/**
	 * Returns the current objective function.
	 * 
	 * @return the current objective function.
	 */
	public MathExpr<String> getObjectiveFunction() {
		return objFunction;
	}

	/**
	 * Returns the current optimization sense.
	 * 
	 * @return the current optimization sense.
	 */
	public OptSense getOptimizationSense() {
		return sense;
	}

	/**
	 * Sets the objective function to the provided mathematical expression.
	 * 
	 * @param objective
	 *            a mathematical expression of reaction names.
	 */
	public void setObjectiveFunction(MathExpr<String> objective) {
		this.objFunction = objective;
	}

	/**
	 * Sets the optimization sense to the provided value.
	 * 
	 * @param sense
	 *            an optimization sense.
	 */
	public void setOptimizationSense(OptSense sense) {
		this.sense = sense;
	}

	/**
	 * Sets the constraint variation input source.
	 * 
	 * @param constraintSource
	 *            a source of constraint sets whose effect upon application is
	 *            to be computed.
	 */
    public void 
    setConstraintSource(Iterable<FluxConstraints> constraintSource) {
        this.constraintSource = constraintSource;
    }

	/**
	 * Returns a progress monitor for this process.
	 * 
	 * @return a progress monitor for this process.
	 */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    private void checkState() {

        if (sense == null) {
        	String s = "An optimization sense was never set.";
        	throw new IllegalStateException(s);
        }

        if (objFunction == null) {
            String s = "An objective function was never set.";
            throw new IllegalStateException(s);
        }

        if (constraintSource == null) {
            String s = "A constraints source was never set.";
            throw new IllegalStateException(s);
        }
    }

	/**
	 * Runs this process and adds the generated results to the provided results
	 * manager via its {@linkplain ResultsManager#addData(Object)} method.
	 * <p>
	 * A set of flux constraints is retrieved from the constraint source, their
	 * effect is computed, and the results are added to the provided results
	 * manager. This is repeated until the end of the deletion source is
	 * reached.
	 * 
	 * @param resultsManager
	 *            the results manager to which the results will be added.
	 */
    public void run(ResultsManager<FbaOptimizer<F>.Results> resultsManager) {
    	
        checkState();

        try {
        	
        	// Initialized progress monitor
            progress.resetCount();
            progress.start();

            // Running the process
            runProcess(resultsManager);

            // Finalizing the progress monitor
            progress.stop();
            
        } catch (RuntimeException e) {

        	// Finalizing the progress monitor
            if (!progress.stopped()) {
                progress.stop();
            }
            
            // Throwing the original exception
            throw e;
        }
    }

    private void runProcess(ResultsManager<FbaOptimizer<F>.Results> rm) {

        // Iterating thru each set of flux constraints
    	for (FluxConstraints constraints : constraintSource) {

            // Setting these constraints
            setConstraints(constraints);

            // Computing an optimal solution
            FbaOptimizer<F>.Results results = 
            	computer.getOptimal(objFunction, sense);

            // Adding these results to the manager
            rm.addData(results);

            // Restoring original constraints
            restoreConstraints(constraints);

            // Incrementing progress
            progress.increment();
        }

    }

    private void setConstraints(FluxConstraints fc) {
        AbstractCvFbaOptProcess.setConstraints(fc, computer);
    }

    private void restoreConstraints(FluxConstraints fc) {
        AbstractCvFbaOptProcess.restoreConstraints(fc, computer);
    }

}
