/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.*;

import sbrt.kernel.*;
import sbrt.kernel.fba.*;
import sbrt.kernel.fba.comp.*;
import sbrt.kernel.io.*;
import sbrt.kernel.math.Interval;
import sbrt.kernel.utilities.*;

/**
 * This class is used to perform constraint variation flux interval processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 *         
 * @param <F> the fluxome type.
 * 
 */
public final class CvFluxIntervalProcess<F extends Fluxome> extends
		AbstractCvFbaOptProcess<F, Map<String, Interval>> implements
		MonitoredProcess<Map<String, Interval>> {

	private final FluxIntervalComputer<F> computer;

	private final DiscreteProgressMonitor progress;

	private Set<String> reactionNames;

	private Iterable<FluxConstraints> constraintSource;

	/**
	 * Constructs a new constraint variation flux interval process.
	 * 
	 * @param computer
	 *            the computer with which all flux intervals will be computed.
	 */
	public CvFluxIntervalProcess(FluxIntervalComputer<F> computer) {

		// Checking and setting
		Util.nullCheck(computer);
		this.computer = computer;

		// Initializing progress monitor
		progress = new DiscreteProgressMonitor();
	}

	/**
	 * Returns the fluxome used in this process.
	 * 
	 * @return the fluxome used in this process.
	 */
	public F getFluxome() {
		return computer.getFluxome();
	}

	/**
	 * Returns a progress monitor for this process.
	 * 
	 * @return a progress monitor for this process.
	 */
	public ProgressMonitor getProgress() {
		return progress.asUnmodifiable();
	}

	/**
	 * Sets the constraint variation input source.
	 * 
	 * @param constraintSource
	 *            a source of constraint sets whose effect upon application is
	 *            to be computed.
	 */
	public void setConstraintSource(Iterable<FluxConstraints> constraintSource) {

		// Checking and setting
		Util.nullCheck(constraintSource);
		this.constraintSource = constraintSource;
	}

	/**
	 * Sets the names of reactions for which flux intervals are to be computed.
	 * 
	 * @param reactionNames
	 *            the names of reactions for which flux intervals are to be
	 *            computed.
	 */
	public void setReactionNames(Collection<String> reactionNames) {

		// Creating storage
		this.reactionNames = new TreeSet<String>();

		// Iterating thru provided names
		for (String rxn : reactionNames) {

			// Checking name
			if (!computer.getFluxome().isRxnName(rxn)) {
				String s = "The reaction '" + rxn
						+ "' does not exist in the fluxome provided.";
				throw new IllegalArgumentException(s);
			}

			// Adding to the rest
			this.reactionNames.add(rxn);
		}

	}

	private void checkState() {

		if (reactionNames == null) {
			String s = "No reaction names were provided.";
			throw new IllegalStateException(s);
		}

		if (constraintSource == null) {
			String s = "A constraint source was never set.";
			throw new IllegalStateException(s);
		}

	}

	/**
	 * Runs this process and adds the generated results to the provided results
	 * manager via its {@linkplain ResultsManager#addData(Object)} method.
	 * <p>
	 * First, a set of flux constraints is retrieved from the constraint source
	 * and applied. Then the flux intervals of the reactions with the current names
	 * are computed and added to the provided results manager. Finally, the 
	 * flux constraints are restored to their original values. This process
	 * is repeated until the end of the constraint source is reached.
	 * 
	 * @param resultsManager
	 *            the results manager to which the results will be added.
	 */
	public void run(ResultsManager<Map<String, Interval>> resultsManager) {

		checkState();

		try {

			// Initialized progress monitor
			progress.resetCount();
			progress.start();

			// Running the process
			runProcess(resultsManager);

			// Finalizing the progress monitor
			progress.stop();

		} catch (RuntimeException e) {

			// Finalizing the progress monitor
            if (!progress.stopped()) {
                progress.stop();
            }

			// Throwing the original exception
			throw e;
		}
	}

	private void 
	runProcess(ResultsManager<Map<String, Interval>> rm) {

		// Iterating thru each set of flux constraints
		for (FluxConstraints constraints : constraintSource) {

			// Setting constraints
			setConstraints(constraints, computer);

			// Computing flux intervals
			Map<String, Interval> fluxIntervals = 
				computer.getIntervals(reactionNames);

			// Adding to the results manager
			rm.addData(fluxIntervals);

			// Restoring original constraints
			restoreConstraints(constraints, computer);

			// Incrementing progress
			progress.increment();
		}

	}

}
