/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.fba.FluxConstraints;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.FbaOptimizer;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.MathExpr;
import sbrt.kernel.math.optimization.OptSense;
import sbrt.kernel.utilities.DiscreteProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to perform constraint-variation objective function
 * processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * @param <F> the fluxome type.
 * 
 */
public final class CvObjectiveFunctionProcess<F extends Fluxome> 
	extends FbaProcess<F, FbaOptimizer<F>.Results> 
	implements MonitoredProcess<FbaOptimizer<F>.Results> {

	// The computer with which all optimizations are performed.
	private final FbaOptimizer<F> computer;
    private final DiscreteProgressMonitor progress;
    
    private Iterable<FluxConstraints> constraintSource;
    private Iterable<? extends MathExpr<String>> objectiveSource;
    private OptSense sense;

	/**
	 * Constructs a new constraint-variation objective function process.
	 * 
	 * @param computer
	 *            the computer with which all results will be computed.
	 */
    public CvObjectiveFunctionProcess(FbaOptimizer<F> computer) {

		// Checking and setting computer
		Util.nullCheck(computer);
		this.computer = computer;
		
        progress = new DiscreteProgressMonitor();
    }

	/**
	 * Returns the fluxome used in this process.
	 * 
	 * @return the fluxome used in this process.
	 */
	public F getFluxome() {
		return computer.getFluxome();
	}

	/**
	 * Returns the current optimization sense.
	 * 
	 * @return the current optimization sense.
	 */
	public OptSense getOptimizationSense() {
		return sense;
	}

	/**
	 * Sets the optimization sense to the provided value.
	 * 
	 * @param sense
	 *            an optimization sense.
	 */
	public void setOptimizationSense(OptSense sense) {
		this.sense = sense;
	}

    /**
     * Sets the objective function input source.
     * 
     * @param objectiveSource a source of objective functions whose 
     * optimal values are to be computed.
     */
    public void setObjectiveFunctionSource(
    		Iterable<? extends MathExpr<String>> objectiveSource) {
        this.objectiveSource = objectiveSource;
    }

	/**
	 * Sets the constraint variation input source.
	 * 
	 * @param constraintSource
	 *            a source of constraint sets whose effect upon application is
	 *            to be computed.
	 */
    public void setConstraintSource(
    		Iterable<FluxConstraints> constraintSource) {
        this.constraintSource = constraintSource;
    }

	/**
	 * Returns a progress monitor for this process.
	 * 
	 * @return a progress monitor for this process.
	 */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    private void checkState() {

        if (sense == null) {
        	String s = "An optimization sense was never set.";
        	throw new IllegalStateException(s);
        }

        if (objectiveSource == null) {
            String s = "An objective function source was never set.";
            throw new IllegalStateException(s);
        }

        if (constraintSource == null) {
            String s = "A constraints source was never set.";
            throw new IllegalStateException(s);
        }
    }

	/**
	 * Runs this process and adds the generated results to the provided results
	 * manager via its {@linkplain ResultsManager#addData(Object)} method.
	 * <p>
	 * First, a set of flux constraints is retrieved from the constraint source
	 * and applied. Then an objective function is retrieved from the objective
	 * source. The optimal value is computed and the results are added to the
	 * results manager. Each objective function is iterated through in the same
	 * fashion. When the end of the objective source is reached, the flux
	 * constraints are restored to their original value and
	 * {@code resultsManager.breakPoint()} is called. At this point, the next
	 * set of flux constraints is retrieved from the constraint source. The
	 * entire process is repeated until the end of the constraint source is
	 * reached.
	 * 
	 * @param resultsManager
	 *            the results manager to which the results will be added.
	 */
    public void run(ResultsManager<FbaOptimizer<F>.Results> resultsManager) {
    	
        checkState();

        try {
        	
        	// Initialized progress monitor
            progress.resetCount();
            progress.start();

            // Running the process
            runProcess(resultsManager);

            // Finalizing the progress monitor
            progress.stop();
            
        } catch (RuntimeException e) {

        	// Finalizing the progress monitor
            if (!progress.stopped()) {
                progress.stop();
            }
            
            // Throwing the original exception
            throw e;
        }
    }

    private void runProcess(ResultsManager<FbaOptimizer<F>.Results> resultsManager) {

        // Iterating thru each set of constraints
    	for (FluxConstraints constraints : constraintSource) {

            // Setting those constraints
            AbstractCvFbaOptProcess.setConstraints(constraints, computer);

            // Iterating thru each objective function
            for (MathExpr<String> objFunction : objectiveSource) {

                // Computing an optimal solution
                FbaOptimizer<F>.Results results = 
                	computer.getOptimal(objFunction, sense);

                // Adding those results to the manager
                resultsManager.addData(results);

                // Incrementing progress
                progress.increment();
            }

            // Restoring original constraints
            AbstractCvFbaOptProcess.restoreConstraints(constraints, computer);

            // Setting break point
            resultsManager.breakPoint();
        }

    }

}
