/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.List;
import java.util.Set;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.fba.FluxConstraints;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.FbaOptimizer;
import sbrt.kernel.fba.comp.RxnDeletionComputer;
import sbrt.kernel.io.ConflictingInputException;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.MathExpr;
import sbrt.kernel.math.optimization.OptSense;
import sbrt.kernel.utilities.DiscreteProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to perform constraint-variation reaction deletion
 * processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 *         
 * @param <F> the fluxome type.
 * 
 */
public final class CvRxnDeletionProcess<F extends Fluxome> 
	extends FbaProcess<F, FbaOptimizer<F>.Results>
	implements MonitoredProcess<FbaOptimizer<F>.Results> {

    private final RxnDeletionComputer<F> computer;
	private final DiscreteProgressMonitor progress;

	private Iterable<FluxConstraints> constraintSource;
    private Iterable<Set<String>> deletionSource;

	/**
	 * Constructs a new constraint-variation reaction deletion process.
	 * 
	 * @param computer
	 *            the computer with which all results will be computed.
	 */
	public CvRxnDeletionProcess(RxnDeletionComputer<F> computer) {
		
        // Checking and setting computer
        Util.nullCheck(computer);
        this.computer = computer;

		progress = new DiscreteProgressMonitor();
	}

	/**
	 * Returns the fluxome used by the reaction deletion computer.
	 * @return the fluxome used by the reaction deletion computer.
	 */
    public F getFluxome() {
        return computer.getFluxome();
    }

    /**
     * Returns the current zero cutoff used by the reaction deletion computer.
     * @return the current zero cutoff used by the reaction deletion computer.
     */
    public double getZeroCutoff() {
        return computer.getZeroCutoff();
    }

    /**
     * Returns the current objective function.
     * @return the current objective function.
     */
    public MathExpr<String> getObjectiveFunction() {
		return computer.getObjectiveFunction();
    }

    /**
     * Returns the current optimization sense.
     * @return the current optimization sense.
     */
    public OptSense getOptimizationSense() {
		return computer.getOptimizationSense();
    }

	/**
	 * Sets the objective function to the provided mathematical expression.
	 * 
	 * @param objective
	 *            a mathematical expression of reaction names.
	 */
	public void setObjectiveFunction(MathExpr<String> objective) {
		computer.setObjectiveFunction(objective);
	}

	/**
	 * Sets the optimization sense to the provided value.
	 * 
	 * @param sense
	 *            an optimization sense.
	 */
	public void setOptimizationSense(OptSense sense) {
		computer.setOptimizationSense(sense);
	}

    /**
     * Sets the reaction deletion input source.
     * 
     * @param deletionSource a source of reaction names sets whose effect upon
     * deletion is to be computed.
     */
    public void 
    setRxnDeletionSource(Iterable<Set<String>> deletionSource) {
        this.deletionSource = deletionSource;
    }

	/**
	 * Sets the constraint variation input source.
	 * 
	 * @param constraintSource
	 *            a source of constraint sets whose effect upon application is
	 *            to be computed.
	 */
	public void 
	setConstraintSource(Iterable<FluxConstraints> constraintSource) {
		this.constraintSource = constraintSource;
	}

    /**
     * Sets the zero cutoff to the provided value. The zero cutoff is the 
     * value used to decide which computed values are close enough to zero
     * to be considered equal to zero.  This value is an upper bound and 
     * should itself be close to zero, like 1E-6 for example.
     * 
     * @param zeroCutoff the zero cutoff.
     */
    public void setZeroCutoff(double zeroCutoff) {
        computer.setZeroCutoff(zeroCutoff);
    }

	/**
	 * Ensures all required information has been supplied to this process.
	 */
	private void checkState() {

		if (constraintSource == null) {
			String s = "A constraint source was never set.";
			throw new IllegalStateException(s);
		}

		if (deletionSource == null) {
            String s = "A reaction deletion source was never set.";
            throw new IllegalStateException(s);
        }

	}

	/**
	 * Returns a progress monitor for this process.
	 * @return a progress monitor for this process.
	 */
	public ProgressMonitor getProgress() {
		return progress.asUnmodifiable();
	}

	/**
	 * Runs this process and adds the generated results to the provided results
	 * manager via its {@linkplain ResultsManager#addData(Object)} method.
	 * <p>
	 * First, a set of flux constraints is retrieved from the constraint source
	 * and applied. Then a set of reaction names in the deletion source is
	 * retrieved. The results are computed and added to the results manager.
	 * Each set of reaction names is iterated through in the same fashion. When
	 * the end of the deletion source is reached, the flux constraints are
	 * restored to their original value and {@code resultsManager.breakPoint()}
	 * is called. At this point, the next set of flux constraints is retrieved
	 * from the constraint source. The entire process is repeated until the end
	 * of the constraint source is reached.
	 * 
	 * @param resultsManager
	 *            the results manager to which the results will be added.
	 */
	public void run(ResultsManager<FbaOptimizer<F>.Results> resultsManager) {

		checkState();

		try {

			// Initialized progress monitor
			progress.resetCount();
			progress.start();

			// Running the process
			runProcess(resultsManager);

			// Finalizing the progress monitor
			progress.stop();

		} catch (RuntimeException e) {

			// Finalizing the progress monitor
            if (!progress.stopped()) {
                progress.stop();
            }

			// Throwing the original exception
			throw e;
		}
	}

	private void runProcess(ResultsManager<FbaOptimizer<F>.Results> resultsManager) {

		// Iterating thru each set of constraints
		for (FluxConstraints constraints : constraintSource) {

			// Setting these constraints
			AbstractCvFbaOptProcess.setConstraints(constraints, computer);

			// Iterating thru each set of reaction names
			for (Set<String> names : deletionSource) {

				// Checking for conflicts
				checkForConflict(constraints, names);

				// Computing the effects of deleting these reactions
				FbaOptimizer<F>.Results results = computer.delete(names);

				// Adding results to the manager
				resultsManager.addData(results);

				// Incrementing progress
				progress.increment();
			}

			// Restoring constraints
			AbstractCvFbaOptProcess.restoreConstraints(constraints, computer);

			// Setting break point
			resultsManager.breakPoint();
		}

	}

	/**
	 * Ensures that the provided flux constraints and set of reaction names do
	 * not conflict.
	 * 
	 * @param fc a set of flux constraints.
	 * @param deletionNames a set of reaction names.
	 * 
	 * @throws ConflictingInputException if reaction names occur in both
	 * the provided flux constraints and set of reaction names. A list
	 * of common reaction names can be retrieved with 
	 * {@code ConflictingInputException#getConflicts()}.
	 */
	public static void 
	checkForConflict(FluxConstraints fc, Set<String> deletionNames) {

		// Getting reaction names
		Set<String> rxnNames = fc.getFluxConstraints().keySet();

		// Getting reaction names and expressions contained in both sets
		List<String> commonElements = Util.getCommonElements(rxnNames,
				deletionNames);

		// Throw an exception if common reactions exist
		if (!commonElements.isEmpty()) {
			throw new ConflictingInputException(commonElements);
		}
	}

}
