/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.CycleComputer;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.algebra.DoubleMatrix;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to execute steady state cycle identification processes.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type. 
 *
 */
public final class CycleIdProcess<F extends Fluxome>
	extends FbaProcess<F, DoubleMatrix<?, String>>
        implements MonitoredProcess<DoubleMatrix<?, String>> {

    private final F fluxome;
    private final CycleComputer<F> computer;
    
    /**
     * Constructs a new steady state cycle identification process.
     * 
     * @param fluxome the fluxome whose cycles are to be computed.
     * @param computer the computer with which all results will be computed.
     */
    public CycleIdProcess(F fluxome,
                          CycleComputer<F> computer) {

        // Setting fluxome
        Util.nullCheck(fluxome);
        this.fluxome = fluxome;

        // Setting computer
        Util.nullCheck(computer);
        this.computer = computer;
    }

	/**
	 * Returns the fluxome used in this process.
	 * @return the fluxome used in this process.
	 */
    public F getFluxome() {
        return fluxome;
    }

    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return computer.getProgress();
    }

	/**
     * Runs this process and adds the generated results to the provided results
     * manager via its {@linkplain ResultsManager#addData(Object)} method.
     * 
     * @param resultsManager the results manager to which the results will be
     * added.
     */
    public void run(ResultsManager<DoubleMatrix<?, String>> resultsManager) {

    	// Computing the cycles
        DoubleMatrix<?, String> cycles = computer.getCycles(fluxome);
        
        // Adding to the results manager
        resultsManager.addData(cycles);

    }

}
