/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.FbaOptimizer;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.optimization.OptSense;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to perform a basic Flux Balance Analysis optimization.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 *         
 * @param <F> the fluxome type. 
 * 
 */
public final class FbaOptProcess<F extends Fluxome>
	extends FbaProcess<F, FbaOptimizer<F>.Results> {

	// The computer with which all optimizations are performed.
	private final FbaOptimizer<F> computer;

	private LinearComb<String> objFunction;
	private OptSense sense;
	
	/**
	 * Constructs a new FBA optimization process.
	 * 
	 * @param computer
	 *            the computer with which all optimizations will be performed.
	 */
	public FbaOptProcess(FbaOptimizer<F> computer) {

		// Checking and setting computer
		Util.nullCheck(computer);
		this.computer = computer;
	}

	/**
	 * Returns the fluxome used in this process.
	 * 
	 * @return the fluxome used in this process.
	 */
	public F getFluxome() {
		return computer.getFluxome();
	}

	/**
	 * Returns the current objective function.
	 * 
	 * @return the current objective function.
	 */
	public LinearComb<String> getObjectiveFunction() {
		return objFunction;
	}

	/**
	 * Returns the current optimization sense.
	 * 
	 * @return the current optimization sense.
	 */
	public OptSense getOptimizationSense() {
		return sense;
	}

	/**
	 * Sets the objective function to the provided mathematical expression.
	 * 
	 * @param objective
	 *            a mathematical expression of reaction names.
	 */
	public void setObjectiveFunction(LinearComb<String> objective) {
		this.objFunction = objective;
	}

	/**
	 * Sets the optimization sense to the provided value.
	 * 
	 * @param sense
	 *            an optimization sense.
	 */
	public void setOptimizationSense(OptSense sense) {
		this.sense = sense;
	}

	/**
	 * Ensures all required information has been supplied to this process.
	 */
	private void checkState() {

        if (sense == null) {
        	String s = "An optimization sense was never set.";
        	throw new IllegalStateException(s);
        }

        if (objFunction == null) {
            String s = "An objective function was never set.";
            throw new IllegalStateException(s);
        }
	}

	/**
	 * Computes the optimal value of the current objective function and adds the
	 * results to the provided results manager via its
	 * {@linkplain ResultsManager#addData(Object)} method.
	 * 
	 * @param resultsManager
	 *            the results manager to which the computed results will be
	 *            added.
	 */
	public void run(ResultsManager<FbaOptimizer<F>.Results> resultsManager) {

		checkState();
		
		// Computing optimal soution
		FbaOptimizer<F>.Results results = 
			computer.getOptimal(objFunction, sense);

		// Adding to results manager
		resultsManager.addData(results);
	}

}
