/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.FbaSolver;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.algebra.LinearSystemSolution;
import sbrt.kernel.math.algebra.LinearSystemSolver;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to solve the equation Sv = 0.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <F> the fluxome type.
 * 
 * @see FbaSolver
 */
public final class FbaSolverProcess<F extends Fluxome>
	extends FbaProcess<F, LinearSystemSolution<String>> {

    private final F fluxome;
    private LinearSystemSolver<String> solver;

    /**
     * Constructs a new FBA solver process.
     * 
     * @param fluxome the fluxome with for which Sv = 0 is solved.
     */
    public FbaSolverProcess(F fluxome) {
        Util.nullCheck(fluxome);
        this.fluxome = fluxome;
    }

    /**
     * Returns the fluxome for which Sv = 0 is solved.
     * @return the fluxome for which Sv = 0 is solved.
     */
    public F getFluxome() {
        return fluxome;
    }

    /**
     * Sets the linear system solver to be used.
     * @param solver the linear system solver to be used.
     */
    public void setSolver(LinearSystemSolver<String> solver) {
        Util.nullCheck(solver);
        this.solver = solver;
    }
    
    private void checkState() {
        
        if (solver == null) {
            String s = "No linear system solver was provided.";
            throw new IllegalStateException(s);
        }
        
    }
    
    /**
     * Solves Sv = 0 and adds the result to the provided manager.
     * 
     * @param resultsManager the results manager to which the solution to Sv = 0
     * will be added.
     */
    public void run(ResultsManager<LinearSystemSolution<String>> resultsManager) {
        
        checkState();
        
        LinearSystemSolution<String> sol = FbaSolver.solve(fluxome, solver);
        
        resultsManager.addData(sol);
    }


}
