/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.*;

import sbrt.kernel.*;
import sbrt.kernel.fba.*;
import sbrt.kernel.io.*;
import sbrt.kernel.utilities.*;

/**
 * This class is used to analyze the activity of fluxes.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type.
 * 
 */
public final class FluxActivityProcess<F extends Fluxome> extends
		FbaProcess<F, FluxActivityProcess.Results> implements
		MonitoredProcess<FluxActivityProcess.Results> {

	private static final Integer ZERO = Integer.valueOf(0);

	private final F fluxome;

	private final DiscreteProgressMonitor progress;

	private final Map<String, Integer> fluxActivities;

	private Iterable<Map<String, Double>> vectorSource;

	private double zeroCutoff;

	/**
	 * Constructs a new flux activity process.
	 * 
	 * @param fluxome
	 *            the fluxome for which flux activities will be determined.
	 */
	public FluxActivityProcess(F fluxome) {

		// Checking and setting
		Util.nullCheck(fluxome);
		this.fluxome = fluxome;

		// Getting the total number of reactions in the provided fluxome
		int n = fluxome.rxns();

		// Getting each reaction name in the provided fluxome
		Set<String> rxns = fluxome.getRxnNames();

		// Creating storage for flux activity counts
		fluxActivities = new LinkedHashMap<String, Integer>(n + 1, 1.0F);

		// Filling activity counts
		for (String rxnName : rxns) {
			fluxActivities.put(rxnName, ZERO);
		}

		// Creating progress monitor
		progress = new DiscreteProgressMonitor();

		// Initialzing tolerance to an out-of-band value
		zeroCutoff = Double.POSITIVE_INFINITY;
	}

	/**
	 * Returns the fluxome used in this flux activity process.
	 * 
	 * @return the fluxome used in this flux activity process.
	 */
	public F getFluxome() {
		return fluxome;
	}

	/**
	 * Returns a progress monitor for this process.
	 * 
	 * @return a progress monitor for this process.
	 */
	public ProgressMonitor getProgress() {
		return progress;
	}

	/**
	 * Sets the flux vector source. These are the flux vectors for which flux
	 * activities will be measured.
	 * 
	 * @param vectorSource a source of reaction names mapped to their 
	 * corresponding flux values.
	 */
	public void setFluxVectorSource(Iterable<Map<String, Double>> vectorSource) {

		// Checking and setting
		Util.nullCheck(vectorSource);
		this.vectorSource = vectorSource;
	}

	/**
	 * Sets the zero cutoff to the provided value. The zero cutoff is the value
	 * used to decide which computed values are close enough to zero to be
	 * considered equal to zero. This value is an upper bound and should itself
	 * be close to zero, like 1E-6 for example.
	 * 
	 * @param zeroCutoff
	 *            the zero cutoff.
	 */
	public void setZeroCutoff(double zeroCutoff) {

		// Checking input
		if (Double.isInfinite(zeroCutoff) || Double.isNaN(zeroCutoff)
				|| zeroCutoff < 0d) {

			String s = "Invalid zero cutoff: " + zeroCutoff;
			throw new IllegalArgumentException(s);
		}

		// Setting
		this.zeroCutoff = zeroCutoff;
	}

	private void checkState() {

		// Checking flux vector source
		if (vectorSource == null) {
			String s = "No flux vector source was provided.";
			throw new IllegalStateException(s);
		}

		// Checking numerical tolerance
		if (Double.isInfinite(zeroCutoff)) {
			String s = "No numerical tolerance was provided.";
			throw new IllegalStateException(s);
		}

	}

	/**
	 * Iterates thru the flux vector source set via 
	 * {@code setFluxVectorSource(Iterable)}, counts the number of active and
	 * inactive fluxes in each vector, and adds the results of each to the
	 * provided results manager.
	 * 
	 * @param resultsManager the results manager to which each set of results
	 * is added.
	 */
	public void run(ResultsManager<Results> resultsManager) {

		checkState();

		try {

			// Initialized progress monitor
			progress.resetCount();
			progress.start();

			// Running the process
			runProcess(resultsManager);

			// Finalizing the progress monitor
			progress.stop();

		} catch (RuntimeException e) {

			// Finalizing the progress monitor
            if (!progress.stopped()) {
                progress.stop();
            }

			// Throwing the original exception
			throw e;
		}
	}

	private void runProcess(ResultsManager<Results> resultsManager) {

		// Initialzing flux activity counts
		initializeCounts();

		// Getting an iterator over the provided flux vector
		Iterator<Map<String, Double>> i = vectorSource.iterator();

		// Iterating thru each of the provided flux vectors
		while (i.hasNext()) {

			// Getting the next flux vector
			Map<String, Double> vector = i.next();

			// Determining the active and inactive fluxes
			Results results = computeResults(vector);

			// Adding to the results manager
			resultsManager.addData(results);

			// Incrementing progress
			progress.increment();
		}

	}

	private void initializeCounts() {

		// Iterating thru the activity vector
		for (Map.Entry<String, Integer> e : fluxActivities.entrySet()) {

			// Setting the activity count to zero
			e.setValue(ZERO);
		}

	}

	private Results computeResults(Map<String, Double> vector) {

		// Initializing counters for flux activity
		int inactive = 0;
		int active = 0;

		// Iterating thru each flux
		for (Double flux : vector.values()) {

			// Is the flux < -zero cutoff ?
			boolean lessThanZero = Double.compare(flux, -1 * zeroCutoff) < 0;

			// Is the flux > zero cutoff ?
			boolean greaterThanZero = Double.compare(flux, zeroCutoff) > 0;

			// If the flux lies in the interval [-zero cutoff, zero cutoff]
			// it's inactive
			if (!lessThanZero && !greaterThanZero) {
				++inactive; // Incrementing inactive count
			}

			// Otherwise, it's active
			else {
				// Incrementing active count
				++active;
			}
		}

		// Returning results
		return new Results(active, inactive);
	}

	/**
	 * This class is used to represent the results of a flux activity
	 * process.
	 * 
	 * @author This class was written and documented by Jeremiah Wright while in
	 *         the Wagner lab.
	 * 
	 */
	public static class Results {

		private final int active;
		private final int inactive;

		private Results(int active, int inactive) {
			this.active = active;
			this.inactive = inactive;
		}

		/**
		 * Returns the number of active fluxes.
		 * 
		 * @return the number of active fluxes.
		 */
		public int active() {
			return active;
		}

		/**
		 * Returns the number of inactive fluxes.
		 * 
		 * @return the number of inactive fluxes.
		 */
		public int inactive() {
			return inactive;
		}

	}
}
