/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.Map;
import java.util.Set;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.comp.FluxCapComputer;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to perform a flux cap identification. It simply adds the
 * results of a FluxCapComputer to a ResultsManager.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 */
public final class FluxCapProcess
	extends FbaProcess<IrrevRxnFluxome, Map<String, Set<LinearComb<String>>>>
        implements MonitoredProcess<Map<String, Set<LinearComb<String>>>> {

    private final FluxCapComputer computer;

	/**
	 * Constructs a new flux cap process.
	 * 
	 * @param computer
	 *            the computer with which all flux caps will be computed.
	 */
    public FluxCapProcess(FluxCapComputer computer) {
        // Setting computer
        Util.nullCheck(computer);
        this.computer = computer;
    }

    /**
     * Returns the fluxome for which flux caps are computed.
     * @return the fluxome for which flux caps are computed.
     */
    public IrrevRxnFluxome getFluxome() {
        return computer.getFluxome();
    }

    /**
     * Computes the flux caps and adds the result to the provided manager.
     * 
     * @param resultsManager the results manager to which the flux cap-results
     * will be added.
     */
    public void 
    run(ResultsManager<Map<String, Set<LinearComb<String>>>> resultsManager) {

        // Computing flux caps
    	Map<String, Set<LinearComb<String>>> fluxCaps = computer.getFluxCaps();
        
    	// Adding to the results manager
        resultsManager.addData(fluxCaps);

    }

	/**
	 * Returns a progress monitor for this process.
	 * 
	 * @return a progress monitor for this process.
	 */
    public ProgressMonitor getProgress() {
        return computer.getProgress();
    }

}
