/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.*;

import sbrt.kernel.*;
import sbrt.kernel.fba.*;
import sbrt.kernel.fba.comp.*;
import sbrt.kernel.io.*;
import sbrt.kernel.math.Interval;
import sbrt.kernel.utilities.*;

/**
 * This class is used to perform flux interval computations.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * @param <F> the fluxome type.
 *
 */
public final class FluxIntervalProcess<F extends Fluxome> 
	extends FbaProcess<F, Map<String, Interval>> 
	implements MonitoredProcess<Map<String, Interval>> {

    private final FluxIntervalComputer<F> computer;
    private Set<String> reactionNames;
    
    /**
     * Constructs a new flux interval process.
     * 
     * @param computer the computer with which all flux intervals will be computed.
     */
    public FluxIntervalProcess(FluxIntervalComputer<F> computer) {
    	
    	// Checking and setting
        Util.nullCheck(computer);
        this.computer = computer;
    }

    /**
     * Sets the names of reactions for which flux ranges are to be computed.
     * @param reactionNames the names of reactions for which flux intervals are 
     * to be computed.
     */
    public void setReactionNames(Set<String> reactionNames) {

    	// Creating storage
        this.reactionNames = new TreeSet<String>();

        // Iterating thru each rxn name
        for (String rxn : reactionNames) {

        	// Checking the name
            if (!computer.getFluxome().isRxnName(rxn)) {
                String s = "The reaction '" + rxn +
                           "' does not exist in the fluxome provided.";
                throw new IllegalArgumentException(s);
            }

            // Adding to the rest
            this.reactionNames.add(rxn);
        }

    }

	/**
	 * Returns the fluxome used in this proces.
	 * @return the fluxome used in this process.
	 */
    public F getFluxome() {
        return computer.getFluxome();
    }

    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return computer.getProgress();
    }

    private void checkState() {
    	
    	if (reactionNames == null) {
    		String s = "No reaction names were provided.";
    		throw new IllegalStateException(s);
    	}
    	
    }
    
    /**
     * Computes the flux intervals of the reactions with the current names and
     * adds them to the provided results manager.
     * 
     * @param resultsManager the results manager to which the computed
     * flux intervals will be added.
     */
    public void run(ResultsManager<Map<String, Interval>> resultsManager) {

    	checkState();
    	
    	// Computing flux ranges
        Map<String, Interval> fluxIntervals = 
        	computer.getIntervals(reactionNames);

        // Adding to the results manager
        resultsManager.addData(fluxIntervals);

    }

}
