/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.*;

import sbrt.kernel.*;
import sbrt.kernel.fba.*;
import sbrt.kernel.fba.analysis.FluxIntervalAnalyzer;
import sbrt.kernel.io.*;
import sbrt.kernel.math.Interval;
import sbrt.kernel.utilities.*;

/**
 * This class is used to analyze the plasticity of fluxes.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type.
 * 
 */
public final class FluxPlasticityProcess<F extends Fluxome> 
	extends FbaProcess<F, FluxPlasticityProcess.Results> 
	implements MonitoredProcess<FluxPlasticityProcess.Results> {

    private final F fluxome;
    private final DiscreteProgressMonitor progress;

    private Iterable<Map<String, Interval>> intervalSource;
    private double zeroCutoff;

	/**
	 * Constructs a new flux plasticity process.
	 * 
	 * @param fluxome
	 *            the fluxome for which flux plasticities will be determined.
	 */
    public FluxPlasticityProcess(F fluxome) {

        // Checking and setting
        Util.nullCheck(fluxome);
        this.fluxome = fluxome;

        // Creating progress monitor
        progress = new DiscreteProgressMonitor();

        // Initialzing tolerance to an out-of-band value
        zeroCutoff = Double.POSITIVE_INFINITY;
    }

	/**
	 * Returns the fluxome used in this flux plasticity process.
	 * 
	 * @return the fluxome used in this flux plasticity process.
	 */
    public F getFluxome() {
        return fluxome;
    }

	/**
	 * Returns a progress monitor for this process.
	 * 
	 * @return a progress monitor for this process.
	 */
    public ProgressMonitor getProgress() {
        return progress;
    }

	/**
	 * Sets the flux interval source. These are the flux intervals for which 
	 * flux plasticities will be measured.
	 * 
	 * @param intervalSource a source of flux ranges, or constraints.
	 */
    public void setFluxIntervalSource(Iterable<Map<String, Interval>> intervalSource) {

        // Checking and setting
        Util.nullCheck(intervalSource);
        this.intervalSource = intervalSource;
    }

	/**
	 * Sets the zero cutoff to the provided value. The zero cutoff is the value
	 * used to decide which computed values are close enough to zero to be
	 * considered equal to zero. This value is an upper bound and should itself
	 * be close to zero, like 1E-6 for example.
	 * 
	 * @param zeroCutoff
	 *            the zero cutoff.
	 */
	public void setZeroCutoff(double zeroCutoff) {

		// Checking input
		if (Double.isInfinite(zeroCutoff) || Double.isNaN(zeroCutoff)
				|| zeroCutoff < 0d) {

			String s = "Invalid zero cutoff: " + zeroCutoff;
			throw new IllegalArgumentException(s);
		}

		// Setting
		this.zeroCutoff = zeroCutoff;
	}

    private void checkState() {

        // Checking flux vector source
        if (intervalSource == null) {
            String s = "No flux vector source was provided.";
            throw new IllegalStateException(s);
        }

        // Checking numerical tolerance
        if (Double.isInfinite(zeroCutoff)) {
            String s = "No numerical tolerance was provided.";
            throw new IllegalStateException(s);
        }

    }

	/**
	 * Iterates thru the flux interval source set via
	 * {@code setFluxRangeSource(Iterable)}, computes the flux plasticity
	 * measures, and adds the results of each to the provided results manager.
	 * 
	 * @param resultsManager
	 *            the results manager to which each set of results is added.
	 */
    public void run(ResultsManager<Results> resultsManager) {

    	checkState();
    	
        // Checking input
        Util.nullCheck(resultsManager);

        // Initializing progress monitor
        progress.start();
        progress.resetCount();

        // Getting an iterator over the provided flux range vector
        Iterator<Map<String, Interval>> i = intervalSource.iterator();

        // Iterating thru each of the provided flux vectors
        while (i.hasNext()) {

            // Getting the next flux interval vector
            Map<String, Interval> vector = i.next();

            // Determining the active and inactive fluxes
            Results results = computeResults(vector);

            // Adding to the results manager
            resultsManager.addData(results);

            // Incrementing progress
            progress.increment();
        }

        // Stopping timer
        progress.stop();
    }

    private Results computeResults(Map<String, Interval> vector) {

      	// Creating flux interval analyzer
      	FluxIntervalAnalyzer fia = new FluxIntervalAnalyzer();
      	
    	// Getting the number of plastic fluxes
    	int plastic = fia.rangesGreaterThanZero(vector, zeroCutoff);
    	
    	// Getting the number of rigid fluxes
    	int rigid = fia.rangesEqualZero(vector, zeroCutoff);

    	// Getting the always active reactions
    	Set<String> alwaysActive =
    		fia.getAlwaysActive(vector, zeroCutoff);
    	
    	// Getting the conditionally active reactions
    	Set<String> conditionallyActive =
    		fia.getPotentiallyActive(vector, zeroCutoff);
    	
    	// Getting the never-active fluxes
    	Set<String> neverActive =
    		fia.getNeverActive(vector, zeroCutoff);
    	
        // Returning results
        return new Results(plastic, 
        				   rigid, 
        				   alwaysActive, 
        				   conditionallyActive, 
        				   neverActive); 
    }

	/**
	 * This class is used to represent the results of a flux plasticity
	 * process.
	 * 
	 * @author This class was written and documented by Jeremiah Wright while in
	 *         the Wagner lab.
	 * 
	 */
    public static class Results {

        private final int plastic;
        private final int rigid;
        private final UnmodifiableSet<String> alwaysActive;
        private final UnmodifiableSet<String> potentiallyActive;
        private final UnmodifiableSet<String> neverActive;
        
        private Results(int plastic, 
        				int rigid, 
        				Set<String> alwaysActive, 
        				Set<String> conditionallyActive,
        				Set<String> neverActive) {

            this.plastic = plastic;
            this.rigid = rigid;
            this.alwaysActive = new UnmodifiableSet<String>(alwaysActive);
            this.potentiallyActive = new UnmodifiableSet<String>(conditionallyActive);
            this.neverActive = new UnmodifiableSet<String>(neverActive);
        }

        /**
         * Returns the number of plastic fluxes.
         * @return the number of plastic fluxes.
         */
        public int plastic() {
            return plastic;
        }

        /**
         * Returns the number of rigid fluxes.
         * @return the number of rigid fluxes.
         */
        public int rigid() {
            return rigid;
        }

        /**
         * Returns the number of fluxes that are always active. Always-active
         * fluxes are those whose lower bound is greater than zero.
         * 
         * @return the number of fluxes that are always active.
         */
        public int alwaysActive() {
        	return alwaysActive.size();
        }

        /**
         * Returns the number of fluxes that are potentially active. 
         * Potentially-active fluxes are those whose lower bound is zero and
         * whose upper bound is greater than zero.
         * 
         * @return the number of fluxes that are potentially active.
         */
        public int potentiallyActive() {
        	return potentiallyActive.size();
        }

        /**
         * Returns the number of fluxes that are never active. Never-active
         * fluxes are those whose upper bound is zero.
         * 
         * @return the number of fluxes that are never active.
         */
        public int neverActive() {
        	return neverActive.size();
        }

        /**
         * Returns the set of fluxes that are always active. Always-active
         * fluxes are those whose lower bound is greater than zero.
         * 
         * @return the set of fluxes that are always active.
         */
        public UnmodifiableSet<String> getAlwaysActive() {
        	return alwaysActive;
        }

        /**
         * Returns the set of fluxes that are potentially active. 
         * Potentially-active fluxes are those whose lower bound is zero and
         * whose upper bound is greater than zero.
         * 
         * @return the set of fluxes that are potentially active.
         */
        public UnmodifiableSet<String> getPotActive() {
        	return potentiallyActive;
        }

        /**
         * Returns the set of fluxes that are never active. Never-active
         * fluxes are those whose upper bound is zero.
         * 
         * @return the set of fluxes that are never active.
         */
        public UnmodifiableSet<String> getNeverActive() {
        	return neverActive;
        }

    }
}
