/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.InitialPointComputer;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to execute initial point generators. Initial points are
 * required for random flux vector processes.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type.
 *
 * @see CdHarFluxVectorProcess
 */
public final class InitialPointGenerator<F extends Fluxome> 
	extends FbaProcess<F, InitialPointComputer<F>.Results> {

    private final InitialPointComputer<F> computer;

    /**
     * Constructs a new initial point process.
     * 
     * @param computer the computer with which initial points will be computed.
     */
    public InitialPointGenerator(InitialPointComputer<F> computer) {
    	
        // Checking and setting computer
        Util.nullCheck(computer);
        this.computer = computer;
    }

    /**
     * Returns the fluxome used in this initial point process.
     * @return the fluxome used in this initial point process.
     */
    public F getFluxome() {
        return computer.getFluxome();
    }

    /**
     * Generates an initial point and adds it to the provided results manager.
     * @param resultsManager the results manager to which the initial point
     * will be added.
     */
    public void run(ResultsManager<InitialPointComputer<F>.Results> resultsManager) {

    	// Computing results
        InitialPointComputer<F>.Results results = computer.run();
        
        // Adding to the results manager
        resultsManager.addData(results);

    }

}
