/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.Collection;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import sbrt.kernel.fba.CatalyzedFluxome;
import sbrt.kernel.fba.comp.RedundRxnRemover;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.algebra.SparseDoubleMatrix;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to gather basic information about a fluxome.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type.
 */
public final class NetworkInfoProcess<F extends CatalyzedFluxome> 
	extends FbaProcess<F, NetworkInfoProcess.Results<F>> {

	private final F fluxome;

	/**
	 * Constructs a new network information gatherer process.
	 * 
	 * @param fluxome
	 *            the fluxome for which information will be gathered.
	 */
	public NetworkInfoProcess(F fluxome) {
		
		// Checking and setting
		Util.nullCheck(fluxome);
		this.fluxome = fluxome;
	}

	/**
	 * Returns the fluxome for which information is to be gathered.
	 * 
	 * @return the fluxome for which information is to be gathered.
	 */
	public F getFluxome() {
		return fluxome;
	}

	/**
	 * Adds the information gathered about the fluxome to the provided
	 * results manager.
	 * 
	 * @param resultsManager the results manager to which the gathered
	 * information will be added.
	 */
	public void run(ResultsManager<Results<F>> resultsManager) {
		
		// Creating results
		Results<F> r = new Results<F>(fluxome);
		
		// Adding to the manager
		resultsManager.addData(r);
	}

	/**
	 * This class is used to represent the information gathered about a 
	 * fluxome.
	 * 
	 * @author This class was written and documented by Jeremiah Wright 
	 * while in the Wagner lab.
	 * 
	 * @param <F> the fluxome type. 
	 */
	public static class Results<F extends CatalyzedFluxome> {
		
		private final F fluxome;
		
		private Results(F fluxome) {
			this.fluxome = fluxome;
		}
		
		/**
		 * Returns the metabolites contained in the fluxome.
		 * @return the metabolites contained in the fluxome.
		 */
		public Set<String> getChemSpecies() {
			return fluxome.getSpecies();
		}
		
		/**
		 * Returns the names of all reactions contained in the fluxome.
		 * @return the names of all reactions contained in the fluxome.
		 */
		public Set<String> getRxnNames() {
			return fluxome.getRxnNames();
		}
		
		/**
		 * Returns the catalysts contained in the fluxome.
		 * @return the catalysts contained in the fluxome.
		 */
		public Set<String> getCatalysts() {
			return fluxome.getCatalysts();
		}
		
		/**
		 * Returns the equivalent reactions contained in the fluxome.
		 * @return the equivalent reactions contained in the fluxome.
		 */
		public Collection<Set<String>> getEquivRxns() {
			return RedundRxnRemover.getEquivRxns(fluxome);
		}
		
		/**
		 * Returns the catalyst-reaction associations contained in the fluxome.
		 * @return catalysts mapped to the names of the reactions they catalyze.
		 */
		public Map<String, Set<String>> getCatalystRxnAssoc() {
			
			// Creating storage
			Map<String, Set<String>> m = 
				new TreeMap<String, Set<String>>();
			
			// Iterating thru each catalyst
			for (String catalyst : fluxome.getCatalysts()) {

				// Getting the reactions catalyzed by this catalyst
				Set<String> rxns = fluxome.getCatalyzedRxnNames(catalyst);

				// Adding to the file
				m.put(catalyst, rxns);
			}
			
			return m;
		}
		
		/**
		 * Returns the stoichiometry matrix of the fluxome.
		 * @return the stoichiometry matrix of the fluxome.
		 */
		public SparseDoubleMatrix<String, String> getStoichMatrix() {
			return fluxome.getStoichMatrix();
		}
		
		
	}
	
}
