/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.*;

import sbrt.kernel.*;
import sbrt.kernel.fba.*;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;

/**
 * This class is used to generate random flux constraints.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * 
 * @param <F> the fluxome type.
 *
 */
public final class RandConstGeneratorV1<F extends Fluxome> 
	extends FbaProcess<F, Map<String, Interval>> 
        implements MonitoredProcess<Map<String, Interval>> {

    private final F fluxome;
    private final PercentProgressMonitor progress;
    
    private ArrayList<String> reactionNames;

    private int iterations;
    private int activeFluxes;
    private Interval irrevConst;
    private Interval revConst;
    private Interval inactive;
    private Random random;

    /**
     * Constructs a new random constraint generator.
     * 
     * @param fluxome the fluxome for which random flux constraints are to
     * be generated.
     */
    public RandConstGeneratorV1(F fluxome) {
        
    	// Checking and setting fluxome
    	Util.nullCheck(fluxome);
        this.fluxome = fluxome;

        // Initializing progress monitor
        progress = new PercentProgressMonitor();

        // Initializing to out-of-band values
        iterations = -1;
        activeFluxes = -1;
    }

    /**
     * Returns the fluxome for which random constraints are to be
     * generated.
     * @return the fluxome for which random constraints are to be
     * generated.
     */
    public F getFluxome() {
        return fluxome;
    }

    /**
     * Returns the number of sets of random constraints to be generated.
     * @return the number of sets of random constraints to be generated.
     */
    public int getIterations() {
        return iterations;
    }

    /**
     * Sets the number of sets of random flux constraints to be produced.
     * 
     * @param iterations the number of random flux constraints to be 
     * produced.
     */
    public void setIterations(int iterations) {
    	
        if (iterations < 0) {
            String s = "Invalid iterations: " + iterations;
            throw new IllegalArgumentException(s);
        }
        
        this.iterations = iterations;
    }

    /**
     * Returns the number of active fluxes in each set of randomly generated
     * flux constraints.
     * @return the number of active fluxes in each set of randomly generated
     * flux constraints.
     */
    public int getActiveFluxes() {
        return activeFluxes;
    }

    /**
     * Sets the total number of active fluxes to be contained in each randomly
     * generated set of flux constraints.
     * 
     * @param activeFluxes the total number of active fluxes to be contained 
     * in each randomly generated set of flux constraints.
     */
    public void setActiveFluxes(int activeFluxes) {

        if (activeFluxes < 0) {
            String s = "Invalid number of active fluxes: " + activeFluxes +
            ". Total fluxes = " + fluxome.rxns();
            throw new IllegalArgumentException(s);
        }

        this.activeFluxes = activeFluxes;
    }

    /**
     * Sets the reaction names for which random constraints are to be generated.
     * @param reactionNames the reaction names for which random constraints 
     * are to be generated. 
     */
    public void setReactions(Set<String> reactionNames) {
    	Util.nullCheck(reactionNames);
    	this.reactionNames = new ArrayList<String>(reactionNames);
    }

    /**
     * Sets the random number generator to the one provided.
     * @param random the random number generator to use.
     */
    public void setRandom(Random random) {
        Util.nullCheck(random);
        this.random = random;
    }

    /**
     * Returns the constraint used for active irreversible reactions.
     * @return the constraint used for active irreversible reactions.
     */
    public Interval getIrrevConstraint() {
        return irrevConst;
    }

    /**
     * Returns the constraint used for active reversible reactions.
     * @return the constraint used for active reversible reactions.
     */
    public Interval getRevConstraint() {
        return revConst;
    }

    /**
     * Returns the constraint used for inactive reactions.
     * @return the constraint used for inactive reactions.
     */
    public Interval getInactive() {
        return inactive;
    }

    /**
     * Sets the constraint used for active irreversible reactions.
     * @param constraint the constraint used for active irreversible reactions.
     */
    public void setIrrevConstraint(Interval constraint) {
        Util.nullCheck(constraint);
        this.irrevConst = constraint;
    }
    
    /**
     * Sets the constraint used for active reversible reactions.
     * @param constraint the constraint used for active reversible reactions.
     */
    public void setRevConstraint(Interval constraint) {
        Util.nullCheck(constraint);
        this.revConst = constraint;
    }

    /**
     * Sets the constraint used for inactive reactions.
     * @param inactive the constraint used for inactive reactions.
     */
    public void setInactive(Interval inactive) {
        Util.nullCheck(inactive);
        this.inactive = inactive;
    }

    private void checkState() {

    	if (reactionNames == null) {
    		String s = "No reactions were provided.";
    		throw new IllegalStateException(s);
    	}
    	
        if (iterations < 0) {
            String s = "Iterations was never set.";
            throw new IllegalStateException(s);
        }

        if (activeFluxes < 0) {
            String s = "Active flux number was never set.";
            throw new IllegalStateException(s);
        }

        if (random == null) {
            String s = "Random number generator was never set.";
            throw new IllegalStateException(s);
        }

        if (irrevConst == null) {
            String s = "'activeIrrev' constraint was never set.";
            throw new IllegalStateException(s);
        }

        if (revConst == null) {
            String s = "'activeRev' constraint was never set.";
            throw new IllegalStateException(s);
        }

        if (inactive == null) {
            String s = "'inactive' constraint was never set.";
            throw new IllegalStateException(s);
        }
        
        if (activeFluxes > reactionNames.size()) {
        	String s = "Specified active fluxes (" + activeFluxes +
        			   ") is greater than the number of provided reactions (" +
        			   + reactionNames.size() + ")";
        	throw new IllegalStateException(s);
        }
        
    }

    /**
     * Generates sets of random flux constraints and adds them to the provided
     * results manager. Each set of random constraints contains a constant
     * number of active and inactive constraints, and it is the assignment of these
     * constraints to reaction names that is random.
     * 
     *  @param resultsManager the results manager to which reaction names
     *  mapped to their randomly assigned constraint is added.
     */
    public void run(ResultsManager<Map<String, Interval>> resultsManager) {
    	
        checkState();

        // Starting progress monitor
        progress.start();
        progress.resetCount();
        progress.setTotal(iterations);

        // Getting number of names
        int size = reactionNames.size();

        for (int i = 0; i < iterations; i ++) {

        	// Randomizing the order of names
            Collections.shuffle(reactionNames, random);
            
            // Creating storage
            Map<String, Interval> fluxConstraints = 
            	new LinkedHashMap<String, Interval>(size + 1, 1.0F);

            // Adding active reactions
            for (int j = 0; j < activeFluxes; j ++) {
            	
            	String rxnName = reactionNames.get(j);
            	
            	ChemRxn rxn = fluxome.getRxn(rxnName);
            	
            	if (rxn.isReversible()) {
            		fluxConstraints.put(reactionNames.get(j), revConst);
            	} 
            	
            	else {
            		fluxConstraints.put(reactionNames.get(j), irrevConst);
            	}
                
            }

            // Adding inactive reactions
            for (int j = activeFluxes; j < size; j ++) {
                fluxConstraints.put(reactionNames.get(j), inactive);
            }

            // Adding to the results manager
            resultsManager.addData(fluxConstraints);

            // Incrementing progress
            progress.increment();
        }

        // Stopping progress monitor
        progress.complete();
        progress.stop();
    }

    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }


}
