/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.*;

import sbrt.kernel.*;
import sbrt.kernel.fba.*;
import sbrt.kernel.io.*;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;

/**
 * This class is used to generate random objective functions.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 * @param <F> the fluxome type.
 *
 */
public final class RandomObjectiveGenerator<F extends Fluxome> 
	extends FbaProcess<F, LinearComb<String>>
        implements MonitoredProcess<LinearComb<String>> {

    private final F fluxome;
    private final PercentProgressMonitor progress;
    private Random random;
    private int iterations;

    /**
     * Constructs a new random objective function generator.
     * 
     * @param fluxome the fluxome for which random objective functions are to
     * be generated.
     */
    public RandomObjectiveGenerator(F fluxome) {

        // Setting fluxome
        Util.nullCheck(fluxome);
        this.fluxome = fluxome;

        // Creating progress monitor
        progress = new PercentProgressMonitor();

        // Initializing with an out-of-band value
        iterations = -1;
    }

    /**
     * Returns the fluxome for which random objective functions are to be
     * generated.
     * @return the fluxome for which random objective functions are to be
     * generated.
     */
    public F getFluxome() {
        return fluxome;
    }

    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress;
    }

    /**
     * Sets the random number generator to the one provided.
     * @param random the random number generator to use.
     */
    public void setRandom(Random random) {

        // Setting random number generator
        Util.nullCheck(random);
        this.random = random;
    }

    /**
     * Sets the number of random objective functions to be produced.
     * 
     * @param iterations the number of random objective functions to be 
     * produced.
     */
    public void setIterations(int iterations) {

        // Checking
        if (iterations < 0) {
            String s = "Invalid number of iterations: " + iterations;
            throw new IllegalArgumentException(s);
        }

        // Setting
        this.iterations = iterations;
    }

    private void checkState() {

        // Checking random number generator
        if (random == null) {
            String s = "No random number generator was provided.";
            throw new IllegalStateException(s);
        }

        // Checking iterations
        if (iterations < 0) {
            String s = "Number of iterations was never set.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Generates random objective functions and adds them to the provided
     * results manager.
     * 
     * @param resultsManager the results manager to which the random objective
     * functions will be added.
     */
    public void run(ResultsManager<LinearComb<String>> resultsManager) {

        // Ensuring all required values were provided
        checkState();

        // Initializing progress monitor
        progress.start();
        progress.resetCount();
        progress.setTotal(iterations);

        // Getting reaction names
        Set<String> rxnNames = fluxome.getRxnNames();

        for (int i = 0; i < iterations; i ++) {

            // Getting random objective function
            LinearComb<String> objective = getObjective(rxnNames);

            // Adding to the results manager
            resultsManager.addData(objective);

            // Incrementing progress
            progress.increment();
        }

        // Stoping timer
        progress.stop();
    }

    private LinearComb<String> getObjective(Set<String> rxnNames) {

        // Creating storage for all terms
        List<LinearComb.Term<String>> terms = 
        	new ArrayList<LinearComb.Term<String>>(rxnNames.size());

        // Iterating thru each reaction name
        for (String rxn : rxnNames) {

            // Getting random coefficient between -1 and 1
            double coeff = random.nextDouble() * 2 - 1;

            // Adding term to the rest
            terms.add(new HashLinearComb.Term<String>(rxn, coeff));
        }

        // Returning random objective function
        return new HashLinearComb<String>(terms, 0.0, 0.0);
    }
}
