/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.comp.RedundRxnRemover;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to remove redundant reactions from a fluxome.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 * @see RedundRxnRemover
 */
public final class RedundRxnRemoverProcess extends FbaProcess<Fluxome, Fluxome> {

	private final Fluxome fluxome;

	private final boolean mixed;

	/**
	 * Constructs a new redundant reaction remover process.
	 * 
	 * @param fluxome
	 *            the fluxome for which redundant reactions will be removed.
	 */
	public RedundRxnRemoverProcess(MixedFluxome fluxome) {

		// Checking and setting
		Util.nullCheck(fluxome);
		this.fluxome = fluxome;

		mixed = true;
	}

	/**
	 * Constructs a new redundant reaction remover process.
	 * 
	 * @param fluxome
	 *            the fluxome for which redundant reactions will be removed.
	 */
	public RedundRxnRemoverProcess(IrrevRxnFluxome fluxome) {

		// Checking and setting
		Util.nullCheck(fluxome);
		this.fluxome = fluxome;

		mixed = false;
	}

	/**
	 * Returns the fluxome for which redundant reactions are to be removed.
	 * 
	 * @return the fluxome for which redundant reactions are to be removed.
	 */
	public Fluxome getFluxome() {
		return fluxome;
	}

	/**
	 * Adds the information gathered about the fluxome to the provided results
	 * manager.
	 * 
	 * @param resultsManager
	 *            the results manager to which the gathered information will be
	 *            added.
	 */
	public void run(ResultsManager<Fluxome> resultsManager) {

		if (mixed) {

			Fluxome f = RedundRxnRemover.remove((MixedFluxome) fluxome);

			resultsManager.addData(f);
		}

		else {

			Fluxome f = RedundRxnRemover.remove((IrrevRxnFluxome) fluxome);

			resultsManager.addData(f);
		}

	}

}
