/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.TreeMap;

import sbrt.kernel.KernelProcess;
import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.ExchangeRxn;
import sbrt.kernel.fba.IrrevRxn;
import sbrt.kernel.fba.RevExchRxn;
import sbrt.kernel.fba.RevRxn;
import sbrt.kernel.fba.SinkRxn;
import sbrt.kernel.fba.SourceRxn;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to break reversible reactions into pairs of irreversible
 * reactions.  Reactions of type {@code RevRxn} are broken apart into a pair of
 * "forward" and "reverse" reactions of type {@code IrrevRxn}. Reactions of 
 * type {@code RevExchRxn} are broken apart in a pair of "source" and "sink"
 * reactions of types {@code SourceRxn} and {@code SinkRxn}.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 */
public final class RevRxnBreakerProcess
	implements KernelProcess<Map<String, ChemRxn>> {

	/**
	 * The default suffix added to the names or IDs of forward reactions.
	 */
	public static final String FORWARD_SUFFIX = "_Forward";

	/**
	 * The default suffix added to the names or IDs of reverse reactions.
	 */
	public static final String REVERSE_SUFFIX = "_Reverse";

	/**
	 * The default suffix added to the names or IDs of source reactions.
	 */
	public static final String SOURCE_SUFFIX = "_Source";
	
	/**
	 * The default suffix added to the names or IDs of sink reactions.
	 */
	public static final String SINK_SUFFIX = "_Sink";
	
	private String forwardSuffix;
	private String reverseSuffix;
	
	private String sourceSuffix;
	private String sinkSuffix;
	
	private Map<String, ChemRxn> rxns; 
	
	/**
	 * Constructs a new reversible reaction breaker process.
	 */
    public RevRxnBreakerProcess() {
    	forwardSuffix = FORWARD_SUFFIX;
    	reverseSuffix = REVERSE_SUFFIX;
    	sourceSuffix = SOURCE_SUFFIX;
    	sinkSuffix = SINK_SUFFIX;
    }

    /**
     * Sets the forward suffix if a non-default value is desired.
     * @param suffix the string to be appended to the names of forward reactions.
     */
    public void setForwardSuffix(String suffix) {
    	Util.nullCheck(suffix);
    	forwardSuffix = suffix;
    }

    /**
     * Sets the reverse suffix if a non-default value is desired.
     * @param suffix the string to be appended to the names of reverse reactions.
     */
    public void setReverseSuffix(String suffix) {
    	Util.nullCheck(suffix);
    	reverseSuffix = suffix;
    }

    /**
     * Sets the source suffix if a non-default value is desired.
     * @param suffix the string to be appended to the names of source reactions.
     */
    public void setSourceSuffix(String suffix) {
    	Util.nullCheck(suffix);
    	sourceSuffix = suffix;
    }

    /**
     * Sets the sink suffix if a non-default value is desired.
     * @param suffix the string to be appended to the names of sink reactions.
     */
    public void setSinkSuffix(String suffix) {
    	Util.nullCheck(suffix);
    	sinkSuffix = suffix;
    }
    
    /**
     * Sets the reactions for which reversible reactions will be broken.
     * @param rxns reactions names mapped to their corresponding reaction.
     */
    public void setRxns(Map<String, ? extends ChemRxn> rxns) {
    	
    	// Using a map that preserves the order of keys
    	this.rxns = new LinkedHashMap<String, ChemRxn>();
    	this.rxns.putAll(rxns);
    }

    private void checkState() {
    	
    	if (rxns == null) {
    		String s = "No reactions were provided.";
    		throw new IllegalStateException(s);
    	}
    	
    }
    
    /**
     * Breaks the reversible reactions supplied to this process and adds
     * the result to the provided manager. Irreversible reactions are
     * unaffected, and are also supplied to the results manager.
     * 
     * @param resultsManager the results manager to which all reactions
     * will be added.
     */
    public void run(ResultsManager<Map<String, ChemRxn>> resultsManager) {

    	checkState();
    	
        resultsManager.addData(breakRxns(rxns));
    }

    /**
     * Breaks reversible reactions into pairs of irreversible reactions.
     * Irreversible reactions are unaffected, and are also included in
     * the returned result.
     * 
     * @param rxns reaction names mapped to their corresponding reaction.
     * 
     * @return reaction names mapped to their corresponding irreversible
     * reaction.
     */
    public Map<String, ChemRxn> breakRxns(Map<String, ChemRxn> rxns) {
    	
    	// Using a map that preserves the order of keys
    	Map<String, ChemRxn> brokenRxns = 
    		new LinkedHashMap<String, ChemRxn>();
    	
    	for (Map.Entry<String, ChemRxn> e : rxns.entrySet()) {
    		
    		String id = e.getKey();
    		ChemRxn rxn = e.getValue();
    		
    		// Preserving irreversible reactions as they are
    		if (rxn instanceof IrrevRxn || 
    		    rxn instanceof SourceRxn ||
    		    rxn instanceof SinkRxn) {
    			
    			brokenRxns.put(id, rxn);
    		}
    		
    		// Breaking reversible reactions
    		else if (rxn instanceof RevRxn) {
    			brokenRxns.putAll(breakRxn(id, (RevRxn) rxn));
    		}
    		
    		else if (rxn instanceof RevExchRxn) {
    			brokenRxns.putAll(breakRxn(id, (RevExchRxn) rxn));
    		}
    		
    		else {
    			String s = "Unknown reaction type: " + rxn.getClass();
    			throw new IllegalArgumentException(s);
    		}
    	}
    	
    	return brokenRxns;
    }

    /**
     * Breaks a reversible reaction into a pair of irreversible reactions.
     * 
     * @param rxnId the reaction name or ID.
     * @param rxn the reversible reaction.
     * 
     * @return reaction names mapped to their corresponding irreversible
     * reaction.
     */
    public Map<String, IrrevRxn> breakRxn(String rxnId, RevRxn rxn) {
    	
    	Map<String, IrrevRxn> m = new TreeMap<String, IrrevRxn>();

    	String forwardId = rxnId + forwardSuffix;
    	String reverseId = rxnId + reverseSuffix;
    	
    	IrrevRxn forward = new IrrevRxn(rxn.getReactants(), rxn.getProducts());
    	IrrevRxn reverse = new IrrevRxn(rxn.getProducts(), rxn.getReactants());
    	
    	m.put(forwardId, forward);
    	m.put(reverseId, reverse);
    	
    	return m;
    }
    

    /**
     * Breaks a reversible exchange reaction into a pair of irreversible 
     * source and sink exchange reactions.
     * 
     * @param rxnId the reaction name or ID.
     * @param rxn the reversible exchange reaction.
     * 
     * @return reaction names mapped to their corresponding irreversible
     * exchange reaction.
     */
    public Map<String, ExchangeRxn> breakRxn(String rxnId, RevExchRxn rxn) {
    	
    	Map<String, ExchangeRxn> m = 
    		new LinkedHashMap<String, ExchangeRxn>(3, 1f);

    	String sourceId = rxnId + sourceSuffix;
    	String sinkId = rxnId + sinkSuffix;
    	
    	String species = rxn.getSpecies();
    	
    	SourceRxn source = new SourceRxn(species);
    	SinkRxn sink = new SinkRxn(species);
    	
    	m.put(sourceId, source);
    	m.put(sinkId, sink);
    	
    	return m;
    }
    
}
