/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import java.util.Iterator;
import java.util.Map;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.fba.FluxConstraints;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.CdHarFluxComp;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.utilities.DiscreteProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to perform space variation coordinate direction 
 * hit-and-run flux vector processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in
 * the Wagner lab.
 * @param <F> the fluxome type.
 *
 */
public final class SvCdHarFluxVectorProcess<F extends Fluxome> 
	extends AbstractCvFbaOptProcess<F, CdHarFluxComp<F>.Results>
	implements MonitoredProcess<CdHarFluxComp<F>.Results> {

    private final CdHarFluxComp<F> computer;
    private final DiscreteProgressMonitor progress;

    private int iterations;
    private Iterable<FluxConstraints> intervalSource;
    private Iterable<Map<String, Double>> fluxVectorSource;

    /**
     * Constructs a new flux space variation coordinate direction hit-and-run 
     * flux vector process.
     * 
     * @param computer the computer with which all flux vectors will be 
     * generated.
     */
    public SvCdHarFluxVectorProcess(CdHarFluxComp<F> computer) {

    	// Checking and saving input
        Util.nullCheck(computer);
        this.computer = computer;

        progress = new DiscreteProgressMonitor();
        iterations = -1;
    }

    /**
     * Sets the total number of flux vectors to be generated.
     * @param iterations the total number of flux vector to be generated.
     */
    public void setIterations(int iterations) {
    	
        if (iterations < 0) {
            String s = "iterations is less than 0: " + iterations;
            throw new IllegalArgumentException(s);
        }

        this.iterations = iterations;
    }

    /**
     * Returns the fluxome for which flux vectors are to be generated.
     * @return the fluxome for which flux vectors are to be generated.
     */
    public F getFluxome() {
        return computer.getFluxome();
    }

    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

	/**
	 * Sets the flux interval input source.
	 * 
	 * @param intervalSource a source of flux intervals.
	 */
    public void 
    setIntervalSource(Iterable<FluxConstraints> intervalSource) {
        
    	// Checking and setting
    	Util.nullCheck(intervalSource);
        this.intervalSource = intervalSource;
    }

    /**
     * Sets the source of flux vectors to serve as initial points.
     * @param fluxVectorSource a source of initial points.
     */
    public void setFluxVectorSource(
    		Iterable<Map<String, Double>> fluxVectorSource) {
        
    	// Checking and setting
    	Util.nullCheck(fluxVectorSource);
        this.fluxVectorSource = fluxVectorSource;
    }

    private void checkState() {
    	
    	if (intervalSource == null) {
    		String s = "No flux interval source was provided.";
    		throw new IllegalStateException(s);
    	}
    	
    	if (fluxVectorSource == null) {
    		String s = "No flux vector source was provided.";
    		throw new IllegalStateException(s);
    	}

        if (iterations < 0) {
            String s = "The number of iterations was never set.";
            throw new IllegalStateException(s);
        }
    }

	/**
	 * Runs this process and adds the generated results to the provided results
	 * manager via its {@linkplain ResultsManager#addData(Object)} method.
	 * <p>
	 * First, a set of flux intervals is retrieved from the interval source and
	 * applied as constraints. Then an initial point is retrieved from the flux
	 * vector source and set. The random flux vectors are then computed and
	 * added to the provided results manager. Finally, the flux constraints are
	 * restored to their original values. This process is repeated until the
	 * ends of the constraint and flux vector sources are reached.
	 * 
	 * @param resultsManager
	 *            the results manager to which the results will be added.
	 */
    public void run(ResultsManager<CdHarFluxComp<F>.Results> resultsManager) {
    	
        checkState();

        try {

            // Initializing the progress monitor
            progress.resetCount();
            progress.start();

            // Running the process
            runProcess(resultsManager);
            
            // Finalizing the progress monitor
            progress.stop();

        } catch (RuntimeException e) {

            // Stopping the timer
            if (!progress.stopped()) {
                progress.stop();
            }

            // Throwing the original exception
            throw e;
        }
    }

    private void 
    runProcess(ResultsManager<CdHarFluxComp<F>.Results> resultsManager) {

    	// Getting iterators over both sources
        Iterator<FluxConstraints> i = intervalSource.iterator();
        Iterator<Map<String, Double>> j = fluxVectorSource.iterator();

        // Iterating over both
        while (i.hasNext() && j.hasNext()) {

        	// Getting and setting flux constraints
            FluxConstraints constraints = i.next();
            setConstraints(constraints, computer);

            // Getting and setting flux vector
            Map<String, Double> initialFluxVector = j.next();
            computer.setInitialFluxVector(initialFluxVector);

            for (int k = 0; k < iterations; k++) {

            	// Computing a random flux vector
            	CdHarFluxComp<F>.Results results = computer.run();

            	// Adding it to the results manager
            	resultsManager.addData(results);
            }

            // Restoring original flux constraints
            restoreConstraints(constraints, computer);
            
            // Incrementing progress
            progress.increment();
            
            // Setting break point
            resultsManager.breakPoint();
        }

    }


}
