/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.fba.proc;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.fba.FluxConstraints;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.InitialPointComputer;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.utilities.DiscreteProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to perform flux space variation initial point processes.
 * Initial points are required for random flux vector processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * @param <F> the fluxome type.
 * 
 * @see CdHarFluxVectorProcess
 */
public final class SvInitialPointProcess<F extends Fluxome> 
	extends AbstractCvFbaOptProcess<F, InitialPointComputer<F>.Results>
        implements MonitoredProcess<InitialPointComputer<F>.Results> {

    private final DiscreteProgressMonitor progress;

    private final InitialPointComputer<F> computer;

    private Iterable<FluxConstraints> constraintSource;

    /**
     * Constructs a new constraint variation initial point process.
     * 
     * @param computer the computer with which initial points will be computed.
     */
    public SvInitialPointProcess(InitialPointComputer<F> computer) {

        // Setting computer
        Util.nullCheck(computer);
        this.computer = computer;

        // Creating progress monitor
        progress = new DiscreteProgressMonitor();
    }

    /**
     * Returns the fluxome used in this initial point process.
     * 
     * @return the fluxome used in this initial point process.
     */
    public F getFluxome() {
        return computer.getFluxome();
    }

    /**
     * Sets the flux interval input source.
     * 
     * @param intervalSource a source of flux intervals.
     */
    public void setIntervalSource(Iterable<FluxConstraints> intervalSource) {

        // Checking and setting
        Util.nullCheck(intervalSource);
        this.constraintSource = intervalSource;
    }

    /**
     * Returns a progress monitor for this process.
     * 
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    private void checkState() {

        if (constraintSource == null) {
            String s = "No constraint source was provided.";
            throw new IllegalArgumentException(s);
        }

    }

    /**
	 * Runs this process and adds the generated results to the provided results
	 * manager via its {@linkplain ResultsManager#addData(Object)} method.
	 * <p>
	 * First, a set of flux intervals is retrieved from the interval source and
	 * applied as constraints. Then the initial point is computed and added to
	 * the provided results manager. Finally, the flux constraints are restored
	 * to their original values. This process is repeated until the end of the
	 * constraint source is reached.
	 * 
	 * @param resultsManager
	 *            the results manager to which the results will be added.
	 */
    public void run(ResultsManager<InitialPointComputer<F>.Results> resultsManager) {

        checkState();

        try {

            // Initialized progress monitor
            progress.resetCount();
            progress.start();

            // Running the process
            runProcess(resultsManager);

            // Finalizing the progress monitor
            progress.stop();

        } catch (RuntimeException e) {

            // Finalizing the progress monitor
            if (!progress.stopped()) {
                progress.stop();
            }

            // Throwing the original exception
            throw e;
        }
    }

    private void runProcess(ResultsManager<InitialPointComputer<F>.Results> resultsManager) {

        // Iterating thru eac set of constraints
        for (FluxConstraints constraints : constraintSource) {

            // Setting constraints
            setConstraints(constraints, computer);

            // Computing results
            InitialPointComputer<F>.Results results = computer.run();

            // Adding results to the manager
            resultsManager.addData(results);

            // Restoring original constraints
            restoreConstraints(constraints, computer);

            // Incrementing progress
            progress.increment();
        }

    }

}
