/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.io;

import java.util.*;

import sbrt.kernel.utilities.UnmodifiableIterator;

/**
 * This class is used to store a list of results that can be easily retrieved.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 * @param <R> the result type.
 */
public class ListResultsManager<R> 
	extends AbstractResultsManager<R>
	implements Iterable<R> {

    private final List<R> results;

    /**
     * Constructs a new list-based results manager.
     */
    public ListResultsManager() {
    	results = new LinkedList<R>();
    }
    
    /**
     * Returns an iterator over the list of results.
     * @return an iterator over the list of results.
     */
    public UnmodifiableIterator<R> iterator() {
    	return new UnmodifiableIterator<R>(results.iterator());
    }
    
    /**
     * Returns a copy of the list of results added to this results manager.
     * @return a copy of the list of results added to this results manager.
     */
    public List<R> getResults() {
        return new LinkedList<R>(results);
    }

    /**
     * Adds the provided results to this results manager.
     * 
     * @param <S> a type extending the primary result type.
     * @param results the results to be stored.
     */
    public <S extends R> void addData(S results) {
    	
        if (isClosed()) {
            throw new IllegalStateException("already closed");
        }
        
        this.results.add(results);
    }

}
