/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.io;

/**
 * This interace is used to represent a results-manager.
 * A results-manager is supplied with some form of results,
 * or data, and then it processes those results in an appropriate way. This may 
 * involve writing all, some, or none of the data to a
 * file or another form of data repository. Other 
 * possibilities exist as well.
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab. 
 *
 * @param <T> the type of data an instance of this class
 * can manage.
 */
public interface ResultsManager<T> {

	/**
	 * Adds data to this results manager.
	 * 
	 * @param <S> a type extending the primary result type.
	 * @param results the results, or data, to be added
	 * to this manager.
	 * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
	 */
    public <S extends T> void addData(S results);

    /**
     * Signals to this manager that a point of 
     * discontinuity or change has been reached. 
     * The specific actions or response to a call of
     * this method is left to implementors of this 
     * interface.
     * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
     */
    public void breakPoint();

    /**
     * Closes this results manager so that no further
     * results can be added.
     */
    public void close();

    /**
     * Indicates if this results manager has already 
     * been closed.
     * 
     * @return {@code true} if results can no longer be
     * added to this manager; {@code false} otherwise. 
     * 
     * @see #close()
     */
    public boolean isClosed();

    /**
     * Returns the name of this results manager.
     * 
     * @return the name of this results manager, or
     * {@code null} if no name exists.
     */
    public String getName();
}
