/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import java.util.*;

import sbrt.kernel.utilities.Util;

/**
 * This abstract class is a skeleton implementation of {@linkplain LinearComb}.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 * @param <V> the variable type.
 */
public abstract class AbstractLinearComb<V> extends AbstractPolynomial<V>
		implements LinearComb<V> {

	private final double zeroCutoff;
	
	/**
	 * Constructs a new linear combination with the provided constant and
	 * zero cutoff.
	 * 
	 * @param constant the constant of this linear combination.
	 * @param zeroCutoff the value used to decide if coefficients or constants
	 * are close enough to zero to be considered equal to zero. 
	 * 
	 * @see #getZeroCutoff()
     * 
     * @throws IllegalArgumentException if a provided value is either
     * infinite or NaN; or if the zero cutoff is negative.
	 */
	public AbstractLinearComb(double constant, double zeroCutoff) {
		super(constant);
		
		// Checking and setting
		checkConstant(zeroCutoff);
		
		if (zeroCutoff < 0d) {
			String s = "Negative zero cutoff: " + zeroCutoff;
			throw new IllegalArgumentException(s);
		}
		
		this.zeroCutoff = zeroCutoff;
	}
	
	/**
	 * Constructs a new linear combination with the provided constant. The 
	 * zero cutoff is set to 0d.
	 * 
	 * @param constant the constant of this linear combination.
	 */
	public AbstractLinearComb(double constant) {
		this(constant, 0.0d);
	}

    /**
	 * Returns the current zero cutoff.
	 * <p>
	 * The zero cutoff is a value used to decide if coefficients are close
	 * enough to zero to be considered equal to zero. In other words, the
	 * absolute value of the coefficient of every term of this linear
	 * combination will be greater than the zero cutoff.
	 * 
	 * @return the zero cutoff.
	 */
	public double getZeroCutoff() {
		return zeroCutoff;
	}
    
    /**
     * Returns the terms of this linear combination as list.
     * @return the terms of this linear combination as list.
     */
	public List<LinearComb.Term<V>> getTermList() {
		
		// Determining appropriate size
		int size = terms();
		
		// Creating storage
		ArrayList<LinearComb.Term<V>> terms =
			new ArrayList<LinearComb.Term<V>>(size);
		
		// Copying
		for (LinearComb.Term<V> t : getTermArray()) {
			terms.add(t);
		}
		
		// Trimming
		terms.trimToSize();
		
		return terms;
	}
    
    /**
     * Returns an iterator over the terms of this linear combination.
     * @return an iterator over the terms of this linear combination.
     */
	public Iterator<LinearComb.Term<V>> iterator() {
		
		final LinearComb.Term<V>[] terms = getTermArray();
		
		return new Iterator<LinearComb.Term<V>>() {

			private int currentIndex = 0;
			
			public boolean hasNext() {
				return currentIndex < terms.length;
			}

			public LinearComb.Term<V> next() {
				
				if (!hasNext()) {
					throw new NoSuchElementException();
				}
				
				return terms[currentIndex ++];
			}

			/**
			 * Throws UnsupportedOperationException.
			 */
			public void remove() {
				throw new UnsupportedOperationException();
			}
						
		};
		
	}

    /**
     * Indicates if the provided object is equal to this
     * linear combination.
     * <P>
     * For two linear combinations to be equal, they must contain
     * the same numerical constant and terms.
     * 
     * @param obj the object to be tested for equality with
     * this linear combination.
     * 
     * @return {@code true} if this linear combination equals the provided 
     * object; {@code false} otherwise.
     * 
     * @see Term#equals(Object)
     */
	public boolean equals(Object obj) {

		if (this == obj) {
			return true;
		}

		// Comparing types
		if (!(obj instanceof LinearComb)) {
			return false;
		}

		// Safely casting
		LinearComb<?> e = (LinearComb) obj;

		// Comparing constants
		if (Double.compare(getConstant(), e.getConstant()) != 0) {
			return false;
		}

		// Comparing variables
		Set<V> variables = getVariables();
		if (!e.getVariables().equals(variables)) {
			return false;
		}

		// Iterating thru each variable
		for (V var : variables) {

			// Getting the coefficient for this variable
			double coeff = getCoefficient(var);

			// Comparing coefficients
			if (Double.compare(e.getCoefficient(var), coeff) != 0) {
				return false;
			}
		}

		return true;
	}

	/**
	 * This class is a skeleton implementation of {@linkplain LinearComb.Term}.
	 * 
	 * @author This class was written and documented by Jeremiah Wright while in
	 * the Wagner lab.
	 * 
	 * @param <V> the variable type.
	 */
	public abstract static class Term<V> implements LinearComb.Term<V> {

		private final V variable;
		private final double coefficient;

		/**
		 * Constructs a new linear combination term with the provided
		 * coefficient and variable.
		 * 
		 * @param coefficient the coefficient of this term.
		 * @param variable the variable of this term.
		 */
		public Term(double coefficient, V variable) {

			// Checking inputs
			Util.nullCheck(variable);
			checkConstant(coefficient);

			// Setting
			this.coefficient = coefficient;
			this.variable = variable;
		}

		/**
		 * Constructs a new linear combination term with the provided
		 * coefficient and variable.
		 * 
		 * @param variable the variable of this term.
		 * @param coefficient the coefficient of this term.
		 */
		public Term(V variable, double coefficient) {
			this(coefficient, variable);
		}

		/**
		 * Constructs a new linear combination term with the provided
		 * variable. The coefficient equals 1.
		 * 
		 * @param variable the variable of this term.
		 */
		public Term(V variable) {
			this(1.0d, variable);
		}

    	/**
    	 * Returns the variable contained in this term.
    	 * @return the variable contained in this term.
    	 */
		public V getVariable() {
			return variable;
		}

		/**
		 * Returns the coefficient of this term.
		 * @return the coefficient of this term.
		 */
		public double getCoefficient() {
			return coefficient;
		}

	    /**
	     * Indicates if the provided object is equal to this
	     * term.
	     * <P>
	     * Two terms are considered equal if they contain
	     * the same coefficient and variable.
	     * 
	     * @param obj the object to be tested for equality with
	     * this term.
	     * 
	     * @return {@code true} if this term is equal to the provided object;
	     * {@code false} otherwise.
	     */
		public boolean equals(Object obj) {
			
			if (this == obj) {
				return true;
			}
			if (!(obj instanceof LinearComb.Term)) {
				return false;
			}
			
			LinearComb.Term<?> t = (LinearComb.Term) obj;

			if (Double.compare(coefficient, t.getCoefficient()) != 0) {
				return false;
			}
			if (!variable.equals(t.getVariable())) {
				return false;
			}

			return true;
		}

	    /**
	     * Returns a content-based hash code for this term.
	     * @return a content-based hash code for this term.
	     */
		public int hashCode() {
			
			// Computing content-based hash code
			long x = Double.doubleToLongBits(coefficient);

			int result = 17;
			result = 37 * result + (int) (x ^ (x >>> 32));
			result = 37 * result + variable.hashCode();

			return result;
		}

	    /**
	     * Returns an appropriate string representation of this
	     * term.
	     * 
	     * @return an appropriate string representation of this
	     * term.
	     */
		public String toString() {
			
			// Creating storage
			StringBuilder s = new StringBuilder(100);

			// If the coefficient does not equal 1, append it
			if (Double.compare(coefficient, 1.0) != 0) {
				s.append("(" + coefficient + ") ");
			}
			
			// Appending the variable
			s.append(variable);
			
			return s.toString();
		}
		
		/**
		 * Returns an iterator over the variables in this term.
		 * @return an iterator over the variables in this term.
		 */
		public Iterator<V> iterator() {
			
			return new Iterator<V>() {

				private boolean hasNext = true;
				
				public boolean hasNext() {
					return hasNext;
				}

				public V next() {
					
					// Checking state
					if (!hasNext) {
						throw new NoSuchElementException();
					}
					
					// Changing state
					hasNext = false;

					// Returning the only variable
					return variable;
				}

				public void remove() {
					throw new UnsupportedOperationException();
				}
				
			};
		}

		/**
		 * Returns the power of the provided variable in this
		 * term.
		 * 
		 * @param variable the variable whose power is to be
		 * returned.
		 * 
		 * @return the power of the provided variable in this
		 * term. If the provided object is not a variable in
		 * this term, 0 is returned.
		 */
		public int power(Object variable) {
			return this.variable.equals(variable) ? 1 : 0;
		}

		/**
		 * Returns the set of variables contained in this term.
		 * @return the set of variables contained in this term.
		 */
		public Set<V> getVariables() {
			return Collections.singleton(variable);
		}

	    /**
	     * Indicates if this term contains all
	     * of the variables in the provided collection.
	     * 
	     * @param variables the collection of variables whose
	     * presence in this term is to be tested.
	     * 
	     * @return {@code true} if each element of the provided
	     * collection is a variable in this term; 
	     * {@code false} otherwise.
	     */
		public boolean hasAllVariables(Collection<?> variables) {

			// Iterating over each element in the collection
			for (Object var : variables) {
				
				if (!this.variable.equals(var)) {
					return false;
				}
				
			}

			return true;
		}
	    
	    /**
	     * Indicates if the provided object is a variable contained
	     * in this term.
	     * 
	     * @param variable the variable whose presence in this
	     * term is to be tested.
	     * 
	     * @return {@code true} if this term contains the
	     * provided variable; {@code false} otherwise. 
	     */
		public boolean hasVariable(Object variable) {
			return this.variable.equals(variable);
		}

		/**
		 * Returns {@code true}.
		 * @return {@code true}.
		 */
		public boolean hasVariables() {
			return true;
		}

		/**
		 * Returns {@code 1}.
		 * @return {@code 1}.
		 */
		public int variables() {
			return 1;
		}

	}
}
