/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

/**
 * This abstract class is a skeleton implementation of {@linkplain Polynomial}.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab. 
 *
 * @param <V> the variable type.
 */
public abstract class AbstractPolynomial<V> implements Polynomial<V> {

    private final double constant;

    /**
     * Constructs a new polynomial with the provided constant.
     * @param constant the constant of this polynomial.
     * 
     * @throws IllegalArgumentException if the provided constant is either
     * infinite or NaN.
     */
    public AbstractPolynomial(double constant) {
        
    	// Checking and saving
    	checkConstant(constant);
        this.constant = constant;
    }

    /**
     * Returns the constant of this polynomial.
     * @return the constant of this polynomial.
     */
    public double getConstant() {
        return constant;
    }

    /**
     * Returns a content-based hash code of this polynomial.
     * @return a content-based hash code of this polynomial.
     */
    public int hashCode() {
    	
    	// Computing content-based hash code
        int result = 17;
        result *= 37 * Double.doubleToLongBits(constant);
        result *= 17 * getTermArray().hashCode();
        
        return result;
    }

    /**
     * Returns a string representation of this polynomial.
     * @return a string representation of this polynomial.
     */
    public String toString() {
    	
    	// Getting terms
    	Term<V>[] terms = getTermArray();

    	// If no terms exist, just return the constant
    	if (terms.length == 0) {
    		return "(" + constant + ")";
    	}
    	
    	// Guessing appropriate size
    	int size = terms.length * 100;
    	
    	// Creating storage
        StringBuilder s = new StringBuilder(size);
        
        // Iterating thru each term
        for (int i = 0; i < terms.length; i ++) {
        	
        	// Appending term string
        	s.append(terms[i]);
        	
        	// Adding plus sign if necessary
        	if (i < terms.length - 1) {
        		s.append(" + ");
        	}
        }

        // If the constant does not equal zero, append it
        if (Double.compare(constant, 0) != 0) {
            s.append(" + ");
            s.append("(" + constant + ")");
        }
        
        return s.toString();
    }

    /**
     * Ensures the provided constant is not infinite and not NaN.
     * 
     * @param constant the constant to check.
     * 
     * @throws IllegalArgumentException if the provided constant is either
     * infinite or NaN.
     */
    protected static void checkConstant(double constant) {
    	
        if (Double.isInfinite(constant) || Double.isNaN(constant)) {
            String s = "The constant '" + constant + "' is invalid.";
            throw new IllegalArgumentException(s);
        }
        
    }

}
