/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import sbrt.kernel.utilities.*;

/**
 * This class is used to indicate a constraint violation.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 */
public class ConstraintViolationException extends RuntimeException {

	private static final long serialVersionUID = 1L;
	private final Object variable;
    private final Interval bounds;
    private final Number violatingValue;

    /**
     * Constructs a new constraint violation exception.
     * 
     * @param variable the variable whose constraints have been violated.
     * @param bounds the constraints that have been violated.
     * @param violatingValue the value that violates its constraints.
     */
    public ConstraintViolationException(Object variable,
                                        Interval bounds,
                                        double violatingValue) {
    	
        this(variable, bounds, Double.valueOf(violatingValue));
    }

    /**
     * Constructs a new constraint violation exception.
     * 
     * @param variable the variable whose constraints have been violated.
     * @param bounds the constraints that have been violated.
     * @param violatingValue the value that violates its constraints.
     */
    public ConstraintViolationException(Object variable,
                                        Interval bounds,
                                        Number violatingValue) {
        super("Variable: " + variable +
              ", Bounds: " + bounds +
              ", Value: " + violatingValue);

        Util.nullCheck(variable);
        Util.nullCheck(bounds);
        Util.nullCheck(violatingValue);

        this.variable = variable;
        this.bounds = bounds;
        this.violatingValue = violatingValue;
    }

    /**
     * Returns the variable that has violated its constraints.
     * @return the variable that has violated its constraints.
     */
    public Object getVariable() {
        return variable;
    }

    /**
     * Returns the constraints that have been violated.
     * @return the constraints that have been violated.
     */
    public Interval getBounds() {
        return bounds;
    }

    /**
     * Returns the value that has violated its constraints.
     * @return the value that has violated its constraints.
     */
    public Number getValue() {
        return violatingValue;
    }
}
