/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

/**
 * This class is used to compare double precision numbers for equivalence.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public class DoubleEquiv implements EquivTester<Double> {

    private final double tolerance;

    /**
     * Creates a new double precision number equivalence tester with the
     * provided tolerance.
     * 
     * @param tolerance the maximum amount by which two doubles can differ, but
     * still be considered equivalent.
     */
    public DoubleEquiv(double tolerance) {

        if (Double.isInfinite(tolerance)
                || Double.isNaN(tolerance)
                || tolerance < 0d) {

            throw new IllegalArgumentException("Invalid tolerance: "
                    + tolerance);
        }

        this.tolerance = tolerance;
    }

    /**
     * Returns the tolerance used by this equivalence tester.
     * 
     * @return the maximum amount by which two doubles can differ, but
     * still be considered equivalent.
     */
    public double getTolerance() {
        return tolerance;
    }

    /**
     * Indicates if the provided values are equivalent.
     * @param value1 a value.
     * @param value2 a value.
     * @return {@code true} if the provided values are equivalent, 
     * {@code false} otherwise.
     */
    public boolean areEquiv(Double value1, Double value2) {
        return areEquiv(value1.doubleValue(), value2.doubleValue());
    }

    /**
     * Indicates if the provided values are equivalent.
     * @param value1 a value.
     * @param value2 a value.
     * @return {@code true} if the provided values are equivalent, 
     * {@code false} otherwise.
     */
    public boolean areEquiv(double value1, double value2) {

        // For NaN and infinity
        if (value1 == value2) {
            return true;
        }

        return Math.abs(value1 - value2) <= tolerance;
    }

}
