/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import java.util.*;

import sbrt.kernel.utilities.*;

/**
 * This class is a concrete implemenation of {@linkplain LinearComb}. 
 * The hash codes of the provided variables are used extenstively in this 
 * implementation to provide constant-time lookup for internally stored
 * information.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 * @param <V> the variable type.
 */
public final class HashLinearComb<V> extends AbstractLinearComb<V> {

	/** Variables -> their corresponding coefficient. */
	private final HashMap<V, Double> terms;

	/**
	 * Constructs a new linear combination from the provided terms and
	 * constant. If the absolute value of coefficient of a term in the 
	 * provided array is less than the provided zero cutoff, that term 
	 * will not be included in the new linear combination. 
	 * 
	 * @param terms an array containing the terms used to create this linear 
	 * combination.
	 * @param constant the constant of this linear combination.
	 * @param zeroCutoff the zero cutoff.
	 */
	public HashLinearComb(LinearComb.Term<? extends V>[] terms,
						  double constant, 
						  double zeroCutoff) {
		
		this(Arrays.asList(terms), constant, zeroCutoff);
	}

	/**
	 * Constructs a new linear combination from the provided terms and
	 * constant. If the absolute value of coefficient of a term in the 
	 * provided collection is less than the provided zero cutoff, that term 
	 * will not be included in the new linear combination. 
	 * 
	 * @param terms a collection of terms used to create this linear 
	 * combination.
	 * @param constant the constant of this linear combination.
	 * @param zeroCutoff the zero cutoff.
	 * 
	 * @throws IllegalArgumentException if a variable occurs in more than
	 * one provided term.
     * 
     * @throws IllegalArgumentException if a provided value is either
     * infinite or NaN; or if the zero cutoff is negative.
	 */
	public HashLinearComb(
			Collection<? extends LinearComb.Term<? extends V>> terms, 
			double constant,
			double zeroCutoff) {

		super(constant, zeroCutoff);
        
        // Determining appropriate size
        int size = terms.size() + 1;

		// Creating storage for all terms
		this.terms = new LinkedHashMap<V, Double>(size, 1f);

		// Iterating thru each term
		for (LinearComb.Term<? extends V> term : terms) {

			// Getting the variable-coefficient pair
			V variable = term.getVariable();
			double coefficient = term.getCoefficient();

			// Checking for duplicate variables
			if (this.terms.containsKey(variable)) {
				String s = "The variable '" + variable
						+ "' occurs more than once in the list of terms: "
						+ terms;
				throw new IllegalArgumentException(s);
			}

			// If the absolute value of the coefficient 
			// is greater than the zero cutoff
			if (Math.abs(coefficient) > zeroCutoff) {
				this.terms.put(variable, coefficient);
			}
		}

	}

	private HashLinearComb(HashMap<V, Double> terms,
	                       double constant,
	                       double zeroCutoff) {
		
        super(constant, zeroCutoff);

		// Setting terms without copying
		Util.nullCheck(terms);
		this.terms = terms;
		
		// Getting entry set
		Set<Map.Entry<V, Double>> entries = terms.entrySet();
		
		// Getting an iterator over the entries
		Iterator<Map.Entry<V, Double>> i = entries.iterator();
		
		// Iterating over each term
		while (i.hasNext()) {
			
			// Getting the next entry
			Map.Entry<V, Double> e = i.next();
			
			// Getting the absolute value of the coefficient
			double abs = Math.abs(e.getValue());
			
			// If that value is less than the zero cutoff, remove it
			if (Double.compare(abs, zeroCutoff) <= 0) {
				i.remove();
			}
			
		}

	}
    
    /**
     * Returns the terms of this linear combination as an array.
     * @return the terms of this linear combination as an array.
     */
	@SuppressWarnings("unchecked")
	public HashLinearComb.Term<V>[] getTermArray() {

		// Creating storage
		HashLinearComb.Term<V>[] array = new Term[terms.size()];

		// Initializing index
		int index = 0;

		// Iterating over terms
		for (Map.Entry<V, Double> e : terms.entrySet()) {
			array[index++] = new Term<V>(e.getKey(), e.getValue());
		}

		return array;
	}
    
    /**
     * Returns the number of terms contained in this linear combination.
     * The numerical constant is not considered a term.
     * 
     * @return the number of terms contained in this linear combination.
     */
	public int terms() {
		return terms.size();
	}
	
	/**
	 * Plugs in the provided linear combinations to produce a new
	 * linear combination.
	 * <P>
	 * If some of the keys in the provided map are not variables
	 * in this expression, their provided value is simply ignored.
	 * 
	 * @param variableValues objects mapped to linear combinations.
	 * 
	 * @return the linear combination obtained by "plugging in"
	 * the provided linear combinations. 
	 */
	public HashLinearComb<V> 
	plugIn(Map<?, ? extends LinearComb<? extends V>> variableValues) {
		
		// Initializing temporary field
		HashLinearComb<V> temp = this;
		
		// Iterating thru each variable-linear combination pair
		for (Map.Entry<?, ? extends LinearComb<? extends V>> e
				: variableValues.entrySet()) {

			// Plugging in the provided linear combination into 
			// the temporary field
			temp = temp.plugIn(e.getKey(), e.getValue());
		}

		// Returning the resulting linear combination
		return temp;		
		
	}
	
	/**
	 * Plugs in the provided numerical values and linear combinations to
	 * produce a new linear combination.
	 * <P>
	 * If some of the keys in either of the provided maps are not variables
	 * in this expression, their provided values are simply ignored.
	 * 
	 * @param numbers objects mapped to numerical values.
	 * @param linearCombs objects mapped to linear combinations.
	 * 
	 * @return the linear combination obtained by "plugging in"
	 * the provided numerical values and linear combinations.
	 * 
	 *  @see #plugIn(Map)
	 *  @see #plugInNumbers(Map)
	 */
	public HashLinearComb<V> plugIn(Map<?, Double> numbers,
		Map<?, ? extends LinearComb<? extends V>> linearCombs) {

		// Plugging in all numbers
		HashLinearComb<V> temp = plugInNumbers(numbers);

		// Iterating thru each variable-linear combination pair
		for (Map.Entry<?, ? extends LinearComb<? extends V>> e
				: linearCombs.entrySet()) {

			// Getting the next variable-value pair
			Object variable = e.getKey();
			LinearComb<? extends V> exp = e.getValue();

			// Plugging in number values into the provided linear combination
			LinearComb<? extends V> simplified= exp.plugInNumbers(numbers);

			// Plugging in the provided linear combination in this one
			temp = temp.plugIn(variable, simplified);
		}

		// Returning the resulting linear combination
		return temp;
	}
	
	/**
	 * Plugs in the provided numerical values to produce a new
	 * linear combination.
	 * <P>
	 * If some of the keys in the provided map are not variables
	 * in this expression, their provided value is simply ignored. 
	 * If a value is provided for each variable in this linear 
	 * combination, the resulting expression will not contain any
	 * variables of its own, only a constant.
	 * 
	 * @param variableValues objects mapped to numerical values.
	 * 
	 * @return the linear combination obtained by "plugging in"
	 * the provided numerical values. 
	 */
	public HashLinearComb<V> plugInNumbers(Map<?, Double> variableValues) {
		
		// Creating storage for new terms
		List<LinearComb.Term<V>> newTerms =
			new LinkedList<LinearComb.Term<V>>();
		
		// Getting the constant in this linear combination
		double constant = getConstant();

		// Iterating thru each term in this linear combination
		for (Map.Entry<V, Double> e : terms.entrySet()) {
		
			// Getting the next variable-coefficient pair
			V variable = e.getKey();
			Double coeff = e.getValue();
			
			// Getting the provided value for this 
			// variable if it exists
			Double value = variableValues.get(variable);
			
			// If no value was provided for this variable
			if (value == null) {
				
				// Copying the term as it is
				newTerms.add(new HashLinearComb.Term<V>(variable, coeff));
			}
			
			// If a value was provided
			else {

				// Checking the provided value
				checkConstant(value);

				// Updating constant
				constant += coeff * value;
			}
			
		}
		
		// Returning the resulting linear combination
		return new HashLinearComb<V>(newTerms, constant, getZeroCutoff());
	}
	
	/**
	 * Plugs the provided numerical value into the provided variable to 
	 * produce a new linear combination.
	 * 
	 * @param variable the variable whose value is to be assigned.
	 * @param value the value to be assigned to the provided variable in this
	 * linear combination.
	 * 
	 * @return the linear combination obtained by "plugging in"
	 * the provided numerical value; or this linear combination if the 
	 * provided object is not a variable.
	 */
	public HashLinearComb<V> plugIn(Object variable, double value) {

		// Checking the provided value
		checkConstant(value);

		// Getting the corresponding coefficient for the provided variable
		double coeff = getCoefficient(variable);

		// If the coefficient is zero (i.e. the variable does not exist),
		// just return this linear combination
		if (Double.compare(coeff, 0) == 0) {
			return this;
		}

		// Computing the new constant
		double constant = getConstant() + coeff * value;

		// Determining appropriate size
		int size = terms.size();

		// Creating storage
		HashMap<V, Double> newTerms = new LinkedHashMap<V, Double>(size);

		// Copying the terms in this linear combination into a new HashMap
		newTerms.putAll(terms);

		// Removing the variable
		newTerms.remove(variable);

		// Returning the resulting linear combination
		return new HashLinearComb<V>(newTerms, constant, getZeroCutoff());
	}
	
	/**
	 * Plugs the provided linear combination into the provided variable to 
	 * produce a new linear combination.
	 * 
	 * @param variable the variable whose value is to be assigned.
	 * @param expr the value to be assigned to the provided variable in this
	 * linear combination.
	 * 
	 * @return the linear combination obtained by "plugging in"
	 * the provided linear combination; or this linear combination if the 
	 * provided object is not a variable.
	 */
	public HashLinearComb<V> plugIn(Object variable, 
								    LinearComb<? extends V> expr) {

		// Getting the coefficient for this variable
		double coeff = getCoefficient(variable);
		
		// If the provided variables is not present in this 
		// linear combination, just return this one
		if (Double.compare(coeff, 0.0d) == 0) {
			return this;
		}

		// Multiplying the provided expression by the coefficient
		expr = expr.multiply(coeff);

		// Removing the provided variable
		HashLinearComb<V> result = plugIn(variable, 0);
		
		// Adding them together
		result = result.add(expr);

		return result;
	}

    /**
     * Returns the coefficient of the term containing the provided variable.
     * If the provided variable does not exist in this linear combination,
     * it's coefficient equals 0.
     * 
     * @param variable the variable whose corresponding coefficient is to be
     * returned.
     * 
     * @return the coefficient of the term containing the provided variable.
     */
	public double getCoefficient(Object variable) {

		// Getting the coefficient assosciated with the provided variable
		Double coeff = terms.get(variable);

		// If no such term exists, return 0.
		// Otherwise, return the corresponding coefficient.
		return coeff == null ? 0.0d : coeff;
	}
    
    /**
     * Equates this linear combination to the provided constant and solves for
     * the provided variable.
     * 
     * @param variable the variable in this linear combination for which to 
     * solve.
     * @param rightHandSide the constant with which this linear combination is
     * equated.
     * 
     * @return a new linear combination representing the value of the provided
     * variable.
     */
	public HashLinearComb<V> solveFor(Object variable, double rightHandSide) {

		// Checking the provided right-hand-side
		checkConstant(rightHandSide);

		// Getting the coefficient for the provided variable
		double coeff = getCoefficient(variable);

		// If the variable does not exist, throw an exception
		if (Double.compare(coeff, 0.0d) == 0) {
			String s = "The variable '" + variable + 
						"' does not exist in: " + this;
			throw new IllegalArgumentException(s);
		}

		// Determining appropriate size
		int size = terms.size() + 1;
		
		// Creating storage for new terms
		HashMap<V, Double> newTerms =
			new LinkedHashMap<V, Double>(size, 1.0F);
		
		// Copying original terms into new terms
		newTerms.putAll(terms);
		
		// Removing the variable under consideration
		newTerms.remove(variable);

		// Creating a "multiplier"
		double multiplier = -1.0 / coeff;
		
		// Iterating thru each "term"
		for (Map.Entry<V, Double> e : newTerms.entrySet()) {
			
			// Multiplying each coefficient by the multiplier
			double newCoeff = multiplier * e.getValue();
			
			// Replacing original coefficient
			e.setValue(newCoeff);
		}
		
		// Calculating new constant
		double constant = getConstant() * multiplier + 
						  rightHandSide / coeff;

		// Returning the resulting linear combination
		return new HashLinearComb<V>(newTerms, constant, getZeroCutoff());
	}

    /**
     * Adds, in the mathematical sense, the provided constant to this 
     * linear combination.
     * 
     * @param constant the constant with which to add.
     * 
     * @return the result of adding this linear combination with the provided
     * constant.
     */
	public HashLinearComb<V> add(double constant) {

		// Checking input
		checkConstant(constant);

		// Calculating new constant
		double c = getConstant() + constant;

		// Determining appropriate size
		int size = terms.size() + 1;

		// Creating storage
		HashMap<V, Double> map = new LinkedHashMap<V, Double>(size, 1.0F);

		// Copying terms into storage
		map.putAll(terms);

		// Returning resulting linear combination
		return new HashLinearComb<V>(map, c, getZeroCutoff());
	}

    /**
     * Adds, in the mathematical sense, the provided linear combination to this 
     * linear combination.
     * 
     * @param expression the linear combination with which to add.
     * 
     * @return the result of adding this linear combination with the one 
     * provided.
     */
	public HashLinearComb<V> add(LinearComb<? extends V> expression) {

		// Determining maximum size
		int size = terms() + expression.terms() + 1;

		// Creating storage for new terms
		HashMap<V, Double> newTerms =
			new LinkedHashMap<V, Double>(size, 1.0F);
		
		// Copying original terms from this linear combination
		// into the storage for new terms
		newTerms.putAll(terms);
		
		// Iterating thru the provided linear combination
		for (LinearComb.Term<? extends V> t : expression) {
			
			// Getting the variable-coefficient pair
			V var = t.getVariable();
			double coeff = t.getCoefficient();
			
			// Getting the corresponding coefficient from
			// this linear combination
			double thisCoeff = getCoefficient(var);
			
			// Setting the coefficient in the new terms
			// to equal the sum of both
			newTerms.put(var, coeff + thisCoeff);
			
		}

		// Calculating the new constant
		double constant = getConstant() + expression.getConstant();

		// Returing the resulting linear combination
		return new HashLinearComb<V>(newTerms, constant, getZeroCutoff());
	}
    /**
     * Multiplies this linear combination by the provided constant.
     * 
     * @param constant the constant with which to multiply this linear 
     * combination.
     * 
     * @return the result of multiplying this linear combination by the 
     * provided constant.
     */
	public HashLinearComb<V> multiply(double constant) {

		// Checking input
		checkConstant(constant);

		// If the provided constant == 1
		if (Double.compare(constant, 1) == 0) {
			
			// Returing this linear combination
			return this;
		}

		// If the provided constant == 0 
		if (Double.compare(constant, 0) == 0) {

			// Creating an empty list of terms
			List<LinearComb.Term<V>> empty = Collections.emptyList();
			
			// Returning an empty linear combination
			return new HashLinearComb<V>(empty, 0, getZeroCutoff());
		}
		
		// Determining appropriate size for new terms
		int size = terms.size() + 1;

		// Creating storage
		HashMap<V, Double> newTerms = 
			new LinkedHashMap<V, Double>(size, 1.0F);

		// Iterating thru each term
		for (Map.Entry<V, Double> entry : terms.entrySet()) {

			// Calculating new coefficient
			double coeff = entry.getValue() * constant;

			// Adding term to the rest
			newTerms.put(entry.getKey(), coeff);
		}

		// Calculating new constant
		double c = getConstant() * constant;

		// Returning resulting linear combination
		return new HashLinearComb<V>(newTerms, c, getZeroCutoff());
	}
	
	/**
	 * Returns the set of variables contained in this
	 * mathematical expression.
	 * 
	 * @return the set of variables contained in this
	 * mathematical expression.
	 */
	public Set<V> getVariables() {

		// Determining appropriate size
		int size = terms.size() + 1;

		// Creating storage
		Set<V> s = new LinkedHashSet<V>(size, 1.0F);

		// Copying into storage
		s.addAll(terms.keySet());

		return s;
	}

    /**
     * Indicates if this mathematical expression contains any
     * variables.
     * 
     * @return {@code true} if this expression contains at least
     * one variable; {@code false} otherwise.
     */
	public boolean hasVariables() {
		return !terms.isEmpty();
	}

    /**
     * Indicates if this mathematical expression contains all
     * of the variables in the provided collection.
     * 
     * @param variables the collection of variables whose
     * presence in this expression is to be tested.
     * 
     * @return {@code true} if each element of the provided
     * collection is a variable in this expression; 
     * {@code false} otherwise.
     */
	public boolean hasAllVariables(Collection<?> variables) {
		return terms.keySet().containsAll(variables);
	}
    
    /**
     * Indicates if the provided object is a variable contained
     * in this mathematical expression.
     * 
     * @param variable the variable whose presence in this
     * mathematical expression is to be tested.
     * 
     * @return {@code true} if this expression contains the
     * provided variable; {@code false} otherwise. 
     */
	public boolean hasVariable(Object variable) {
		return terms.get(variable) != null;
	}

    /**
     * Returns the number of variables in this mathematical
     * expression.
     * 
     * @return the number of variables in this mathematical
     * expression.
     */
	public int variables() {
		return terms.size();
	}
    
    /**
     * Returns an iterator over the terms of this linear combination.
     * @return an iterator over the terms of this linear combination.
     */
	public Iterator<LinearComb.Term<V>> iterator() {

		final Iterator<Map.Entry<V, Double>> i = 
			terms.entrySet().iterator();

		return new Iterator<LinearComb.Term<V>>() {

			public LinearComb.Term<V> next() {
				Map.Entry<V, Double> e = i.next();
				return new Term<V>(e.getKey(), e.getValue());
			}

			public boolean hasNext() {
				return i.hasNext();
			}

			public void remove() {
				throw new UnsupportedOperationException();
			}
		};
	}

	/**
	 * Returns a single-term linear combination with the provided variable and
	 * coefficient. The zero cutoff is set to 0.
	 * 
	 * @param <V> the variable type.
	 * 
	 * @param variable the variable. 
	 * @param coefficient the coefficient.
	 * 
	 * @return a linear combination containing a single term with the provided
	 * variable and constant.
	 */
	public static <V> HashLinearComb<V> 
		singleTerm(V variable, double coefficient) {
		return singleTerm(variable, coefficient, 0.0d);
	}

	/**
	 * Returns a single-term linear combination with the provided variable,
	 * coefficient, and zero cutoff.
	 * 
	 * @param <V> the variable type.
	 * 
	 * @param variable the variable. 
	 * @param coefficient the coefficient.
	 * @param zeroCutoff the zero cutoff.
	 * 
	 * @return a linear combination containing a single term with the provided
	 * variable, constant, and zero cutoff.
	 */
	public static <V> HashLinearComb<V> 
		singleTerm(V variable, double coefficient, double zeroCutoff) {

		// Creating term
		Term<V> t = new Term<V>(variable, coefficient);

		// Creating singleton list of terms
		List<Term<V>> terms = Collections.singletonList(t);

		// Returning the single-term linear combination
		return new HashLinearComb<V>(terms, 0, zeroCutoff);
	}

	/**
	 * Returns a single-term linear combination with the provided variable and
	 * coefficient. The zero cutoff is set to 0.
	 * 
	 * @param <V> the variable type.
	 * 
	 * @param coefficient the coefficient.
	 * @param variable the variable. 
	 * 
	 * @return a linear combination containing a single term with the provided
	 * variable and constant.
	 */
	public static <V> HashLinearComb<V> singleTerm(double coefficient, 
                                                   V variable) {
		return singleTerm(variable, coefficient);
	}

	/**
	 * Returns a single-term linear combination with the provided variable.
	 * The coefficient of the term containing this variable is 1, and the 
	 * zero cutoff is 0. 
	 * 
	 * @param <V> the variable type.
	 * @param variable the variable. 
	 * 
	 * @return a linear combination containing a single term with the provided
	 * variable.
	 */
	public static <V> HashLinearComb<V> singleTerm(V variable) {
		return singleTerm(variable, 1.0);
	}

    /**
     * Returns a single-term linear combination with the provided term.
     * The zero cutoff is set to 0.
     * 
     * @param <V> the variable type.
     * @param term the term. 
     * 
     * @return a linear combination containing a single term with the provided
     * variable.
     */
    public static <V> HashLinearComb<V> singleTerm(LinearComb.Term<V> term) {
        return singleTerm(term.getCoefficient(), term.getVariable());
    }
    
	/**
	 * This class is a concrete implementation of {@linkplain LinearComb.Term}.
	 * 
	 * @author This class was written and documented by Jeremiah Wright 
	 * while in the Wagner lab.
	 *
	 * @param <V> the variable type.
	 */
	public static class Term<V> extends AbstractLinearComb.Term<V> {

		/**
		 * Constructs a new linear combination term with the provided
		 * coefficient and variable.
		 * 
		 * @param coefficient the coefficient of this term.
		 * @param variable the variable of this term.
		 */
		public Term(double coefficient, V variable) {
			super(coefficient, variable);
		}

		/**
		 * Constructs a new linear combination term with the provided
		 * coefficient and variable.
		 * 
		 * @param variable the variable of this term.
		 * @param coefficient the coefficient of this term.
		 */
		public Term(V variable, double coefficient) {
			super(variable, coefficient);
		}

		/**
		 * Constructs a new linear combination term with the provided
		 * variable. The coefficient equals 1.
		 * 
		 * @param variable the variable of this term.
		 */
		public Term(V variable) {
			super(variable);
		}
		
	}

	
	
	/**
	 * Used for testing purposes.
	 * @param args none required.
	 */
	@SuppressWarnings("unchecked")
	public static void main(String args[]) {
		
		Term[] termsA = { 
				
				new Term<Object>("x1", 1.0),
				new Term<Object>("x2", 10.0),
				new Term<Object>("x3", 0.0),
				new Term<Object>("x4", 5.0),
				new Term<Object>("x5", 3.3),
				new Term<Object>("x6", 50.2),
				
		};
		
		LinearComb<Object> linA = new HashLinearComb<Object>(termsA, 9, 0.0d);
		
		System.out.println(linA);
		
		Map<Object, Double> numbers = new TreeMap<Object, Double>();
		
		numbers.put("x1", 0.0);
		numbers.put("x2", 0.0);
		numbers.put("x3", 0.0);
		numbers.put("x4", 0.0);
		numbers.put("x5", 0.0);
		numbers.put("x6", 0.0);
		
		System.out.println(linA.plugInNumbers(numbers));

		System.out.println("x1 = " + linA.solveFor("x1", 0.0));
		System.out.println("x2 = " + linA.solveFor("x2", 0.0));
		System.out.println("x4 = " + linA.solveFor("x4", 0.0));
		System.out.println("x5 = " + linA.solveFor("x5", 0.0));
		System.out.println("x6 = " + linA.solveFor("x6", 0.0));
		
		
		Term[] termsB = { 
				
				new Term<Object>("x2", 1.0),
				new Term<Object>("x3", 1.0),
				new Term<Object>("x4", 1.0),
				new Term<Object>("x5", 1.1),
				new Term<Object>("x6", 1.0),
				
		};
		
		LinearComb<Object> linB = new HashLinearComb<Object>(termsB, -1, 0.0d);
		
		System.out.println(linB);
		
		Map<String, LinearComb<Object>> m = Collections.singletonMap("x1", linB);
		
		System.out.println(linA.plugIn(m));
	}
}
