/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import sbrt.kernel.utilities.Util;

/**
 * This class is used to indicate the existence of illegal bounds. Bounds
 * are considered illegal if the upper bound is less than the lower bound.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 */
public class IllegalBoundsException extends RuntimeException {

	private static final long serialVersionUID = 1L;
	private final Object obj;
    private final Number lower;
    private final Number upper;

    /**
     * Constructs a new illegal bounds exception from the provided lower
     * and upper bounds. The lower bound is assumed to be greater than the
     * upper bound, hence the need for this exception.
     * 
     * @param lower the lower bound.
     * @param upper the upper bound.
     */
	public IllegalBoundsException(Number lower, Number upper) {
        this(null, lower, upper);
    }

    /**
     * Constructs a new illegal bounds exception from the provided lower
     * and upper bounds. The lower bound is assumed to be greater than the
     * upper bound, hence the need for this exception.
     * 
     * @param lower the lower bound.
     * @param upper the upper bound.
     */
    public IllegalBoundsException(double lower, double upper) {
        this(null, lower, upper);
    }

    /**
     * Constructs a new illegal bounds exception from the provided lower
     * and upper bounds. The lower bound is assumed to be greater than the
     * upper bound, hence the need for this exception.
     * 
     * @param obj an object with which these illegal bounds are associated.
     * @param lower the lower bound.
     * @param upper the upper bound.
     */
    public IllegalBoundsException(Object obj, double lower, double upper) {
        this(obj, Double.valueOf(lower), Double.valueOf(upper));
    }

    /**
     * Constructs a new illegal bounds exception from the provided lower
     * and upper bounds. The lower bound is assumed to be greater than the
     * upper bound, hence the need for this exception.
     * 
     * @param obj an object with which these illegal bounds are associated.
     * @param lower the lower bound.
     * @param upper the upper bound.
     */
    public IllegalBoundsException(Object obj, Number lower, Number upper) {
        super("Lower Bound: " + lower +
              ", Upper Bound: " + upper + 
              (obj == null ? "" : ", Associated object: " + obj));

        Util.nullCheck(lower);
        Util.nullCheck(upper);

        this.obj = obj;
        this.lower = lower;
        this.upper = upper;
    }

    /**
     * Returns the object with which these illegal bounds are associated. 
     * @return the object with which these illegal bounds are associated, or
     * {@code null} if no such object exists.
     */
    public Object getVariable() {
        return obj;
    }

    /**
     * Returns the lower bound used to create this exception.
     * @return the lower bound used to create this exception.
     */
    public Number getLower() {
        return lower;
    }

    /**
     * Returns the upper bound used to create this exception.
     * @return the upper bound used to create this exception.
     */
    public Number getUpper() {
        return upper;
    }
}
