/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import sbrt.kernel.utilities.*;

/**
 * This class is used to indicate that no feasible solution exists.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 */
public class InfeasibleSolutionException extends RuntimeException {

	private static final long serialVersionUID = 1L;
	private final Object direction;

	/**
	 * Constructs a new infeasible solution exception.
	 * @param direction the variable, or direction in hyperspace, in which
	 * infeasibility was first detected. 
	 */
    public InfeasibleSolutionException(Object direction) {
        super(direction.toString());
        this.direction = direction;
    }

	/**
	 * Constructs a new infeasible solution exception.
	 * @param direction the variable, or direction in hyperspace, in which
	 * infeasibility was first detected.
	 * @param message an error message useful for debugging.
	 */
    public InfeasibleSolutionException(Object direction, String message) {
        super(message);
        Util.nullCheck(direction);
        this.direction = direction;
    }

	/**
	 * Constructs a new infeasible solution exception.
	 * @param direction the variable, or direction in hyperspace, in which
	 * infeasibility was first detected. 
	 * @param cause the exception thrown to indicate an infeasible solution.
	 */
    public InfeasibleSolutionException(Object direction, Throwable cause) {
        super(cause);
        Util.nullCheck(direction);
        this.direction = direction;
    }

	/**
	 * Constructs a new infeasible solution exception.
	 * @param direction the variable, or direction in hyperspace, in which
	 * infeasibility was first detected.
	 * @param message an error message useful for debugging.
	 * @param cause the exception thrown to indicate an infeasible solution.
	 */
    public InfeasibleSolutionException(Object direction,
                                      String message,
                                      Throwable cause) {
        super(message, cause);
        Util.nullCheck(direction);
        this.direction = direction;
    }

    /**
     * Returns the variable, or direction in hyperspace, in which infeasibility
     * was first detected.
     * 
     * @return the variable, or direction in hyperspace, in which infeasibility
     * was first detected.
     */
    public Object getDirection() {
        return direction;
    }
}
