/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import java.io.*;

/**
 * This class is used to represent closed intervals
 * of the form [&alpha;, &beta], where &alpha; &le; &beta;.
 * <P>
 * Instances of this class are immutable.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 */
public final class Interval implements Serializable {

    private static final long serialVersionUID = 1L;

    // The lower and upper bounds defining this interval
    private final double lowerBound;
    private final double upperBound;

    /**
     * Returns the lower bound (or endpoint) of this interval.
     * @return the lower bound (or endpoint) of this interval.
     */
    public double getLower() {
        return lowerBound;
    }

    /**
     * Returns the upper bound (or endpoint) of this interval.
     * @return the upper bound (or endpoint) of this interval.
     */
    public double getUpper() {
        return upperBound;
    }

    /**
     * Returns {@code getUpper() - getLower()}.
     * @return {@code getUpper() - getLower()}.
     */
    public double getRange() {
    	return upperBound - lowerBound;
    }
    
    /**
     * Constructs a new interval with the provided lower and upper bounds.
     * 
     * @param lowerBound the lower bound of the interval, &alpha;
     * @param upperBound the upper bound of the interval, &beta;
     * 
     * @throws IllegalBoundsException if lower bound > upper bound
     * 
     * @throws IllegalArgumentException if either value is 
     * {@code Double.NaN}, {@code lowerBound == Double.POSITIVE_INFINITY},
     * or {@code upperBound == Double.NEGATIVE_INFINITY}.
     */
    public Interval(double lowerBound, double upperBound) {

    	if (Double.isNaN(lowerBound) || lowerBound == Double.POSITIVE_INFINITY) {
            String s = "Invalid lower bound: " + lowerBound;
            throw new IllegalArgumentException(s);
        }

    	if (Double.isNaN(upperBound) || upperBound == Double.NEGATIVE_INFINITY) {
            String s = "Invalid upper bound: " + upperBound;
            throw new IllegalArgumentException(s);
        }

    	if (lowerBound > upperBound) {
            throw new IllegalBoundsException(lowerBound, upperBound);
        }

    	this.lowerBound = lowerBound;
        this.upperBound = upperBound;
    }

    /**
     * Indicates if the provided value lies within this
     * interval.
     * This can be represented mathematically as: 
     * &alpha; &le; {@code value} &le; &beta;.
     * This is determined by applying the Java 
     * language numerical comparison operators
     * {@code <} and {@code >} to the primitive 
     * {@code double} values. This is equivalent to calling
     * {@linkplain #isWithin(double, double)} with a tolerance
     * of {@code 0.0d}.
     * 
     * @param value the numerical value to test.
     * 
     * @return {@code true} if the provided values lies 
     * within this interval; {@code false} otherwise. 
     */
    public boolean isWithin(double value) {
        return isWithin(value, 0.0d);
    }

    /**
     * Indicates if the provided value lies within this
     * interval, given the provided numerical tolerance.
     * This can be represented mathematically as: 
     * &alpha; - {@code tolerance} &le; 
     * {@code value} &le; 
     * &beta; + {@code tolerance}.
     * This is determined by applying the Java 
     * language numerical comparison operators
     * {@code <} and {@code >} to the primitive 
     * {@code double} values.

     * @param value the numerical value to test.
     * @param tolerance the numerical tolerance.
     * 
     * @return {@code true} if the provided values lies 
     * within this interval; {@code false} otherwise. 
     */
    public boolean isWithin(double value, double tolerance) {

        if (value < lowerBound - tolerance) {
            return false;
        }

        if (value > upperBound + tolerance) {
            return false;
        }

        return true;
    }

    /**
     * Indicates if the provided object is equal to this
     * interval. Two intervals A and B are equal if 
     * &alpha;<sub>A</sub> = &alpha;<sub>B</sub> and
     * &beta;<sub>A</sub> = &beta;<sub>B</sub>. This 
     * equality is determined by applying the Java 
     * language numerical comparison operator
     * {@code ==} to the primitive {@code double} values.
     * 
     * @param o the object to test for equality.
     * 
     * @return {@code true} if the provided object equals 
     * this interval; {@code false} otherwise.
     */
    public boolean equals(Object o) {
    	
        if (this == o) {
            return true;
        }
        
        if (!(o instanceof Interval)) {
            return false;
        }
        
        Interval c = (Interval) o;
        
        if (Double.compare(lowerBound, c.getLower()) != 0) {
            return false;
        }
        
        if (Double.compare(upperBound, c.getUpper()) != 0) {
            return false;
        }
        
        return true;
    }

    /**
     * Returns a content-based hash code.
     * @return a content-based hash code.
     */
    public int hashCode() {
    	return (int) lowerBound ^ (int) upperBound;
    }
    
    /**
     * Returns a string representation of this interval.
     * @return a string representation of this interval.
     */
    public String toString() {
    	
        StringBuffer s = new StringBuffer(100);

        // Appending the opening brace
        if (Double.isInfinite(lowerBound)) {
            s.append('(');
        } else {
            s.append('[');
        }
        
        // Appending the values lower and upper bounds
        s.append(toString(lowerBound) + ", " + toString(upperBound));
        
        // Appending the closing brace
        if (Double.isInfinite(upperBound)) {
            s.append(')');
        } else {
            s.append(']');
        }
        
        return s.toString();
    }

    /**
     * Returns a string representation of the provided bound.
     */
    private static String toString(double bound) {
    	
        if (bound == Double.NEGATIVE_INFINITY) {
            return "-Infinity";
        }
        
        if (bound == Double.POSITIVE_INFINITY) {
            return "Infinity";
        }
        
        return Double.toString(bound);
    }

}
