/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import java.util.*;

/**
 * This class is used to compare vectors of intervals.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public class IntervalComparer {

	/**
	 * Returns the set of differences between the provided interval vectors.
	 * Intervals are compared for equivalence for each variable that occurs in 
	 * both provided vectors.
	 * 
	 * @param <V> the variable type.
	 * @param m1 variables mapped to intervals.
	 * @param m2 variables mapped to intervals.
	 * @param tolerance the amount by which a bound of two intervals can differ
	 * but still be considered equivalent.
	 * 
	 * @return the set of differences between the provided interval vectors.
	 * 
	 * @see IntervalEquiv#areEquiv(Interval, Interval)
	 */
    public <V> Set<Diff<V>> 
    getDiffs(Map<? extends V, Interval> m1,
    		 Map<? extends V, Interval> m2,
    		 double tolerance) {

        IntervalEquiv eqTester = new IntervalEquiv(tolerance);

        // Storage for variables with unequal intervals
        Set<Diff<V>> vars = new LinkedHashSet<Diff<V>>();

        for (Map.Entry<? extends V, Interval> e : m1.entrySet()) {

            V variable = e.getKey();

            Interval a = e.getValue();
            Interval b = m2.get(variable);

            if (b == null) {
                continue;
            }

            if (!eqTester.areEquiv(a, b)) {
                Diff<V> d = new Diff<V>(variable, a, b);
                vars.add(d);
            }

        }

        return vars;
    }

    /**
     * Instances of this class are used to hold a variable and a pair of
     * corresponding intervals. The pair of intervals are considered to be
     * non-equivalent, or different.
     * 
     * @author This class was written and documented by Jeremiah Wright while in
     * the Wagner lab.
     * 
     * @param <V> the variable type.
     */
    public static class Diff<V> {

        private V variable;

        private Interval i1;

        private Interval i2;

        private Diff(V variable, Interval i1, Interval i2) {
            this.variable = variable;
            this.i1 = i1;
            this.i2 = i2;
        }

        /**
         * Returns the variable for which the intervals correspond.
         * @return the variable for which the intervals correspond.
         */
        public V getVariable() {
            return variable;
        }

        /**
         * Returns the "first" interval of the pair of intervals.
         * @return the "first" interval of the pair of intervals.
         */
        public Interval getInterval1() {
            return i1;
        }

        /**
         * Returns the "second" interval of the pair of intervals.
         * @return the "second" interval of the pair of intervals.
         */
        public Interval getInterval2() {
            return i2;
        }
    }

}
