/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;


/**
 * This class is used to compare intervals for equivalence. Two intervals
 * <i>A</i> and <i>B</i> are equivalent if their lower and upper bounds
 * are equivalent, respectively.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 */
public class IntervalEquiv implements EquivTester<Interval> {

    private final DoubleEquiv eqTester;

    /**
     * Creates a new interval equivalence tester with the
     * provided tolerance.
     * 
     * @param tolerance the maximum amount by which two bounds can differ, but
     * still be considered equivalent.
     */
    public IntervalEquiv(double tolerance) {
        eqTester = new DoubleEquiv(tolerance);
    }

    /**
     * Returns the tolerance used by this equivalence tester.
     * 
     * @return the maximum amount by which two doubles can differ, but
     * still be considered equivalent.
     */
    public double getTolerance() {
        return eqTester.getTolerance();
    }

    /**
     * Returns the double equivalence tester used by this interval equivalence tester.
     * 
     * @return the double equivalence tester used by this interval equivalence tester.
     */
    public DoubleEquiv getDoubleTester() {
        return eqTester;
    }

    /**
     * Indicates if the provided intervals are equivalent.
     * @param value1 an interval.
     * @param value2 an interval.
     * @return {@code true} if the provided intervals are equivalent, 
     * {@code false} otherwise.
     */
    public boolean areEquiv(Interval value1, Interval value2) {
        
        return eqTester.areEquiv(value1.getLower(), value2.getLower()) &&
                eqTester.areEquiv(value1.getUpper(), value2.getUpper());
    }

}
