/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import java.util.*;

import sbrt.kernel.utilities.*;

/**
 * This class is used to indicate that a point is invalid.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 */
public class InvalidPointException extends RuntimeException {

	private static final long serialVersionUID = 1L;
	
	private final UnmodifiableMap<?, ?> point;

	/**
	 * Constructs an exception to indicate the provided point is invalid.
	 * @param point the invalid point.
	 */
    public InvalidPointException(Map<?, ?> point) {
        super(point.toString());
        this.point = getPoint(point);
    }

	/**
	 * Constructs an exception to indicate the provided point is invalid.
	 * @param point the invalid point.
	 * @param message a message useful for debugging.
	 */
    public InvalidPointException(Map<?, ?> point, String message) {
        super(point.toString() + ". " + message);
        this.point = getPoint(point);
    }

    /**
     * Returns the invalid point.
     * @return the invalid point.
     */
    public UnmodifiableMap<?, ?> getPoint() {
        return point;
    }

    private UnmodifiableMap<?, ?> getPoint(Map<?, ?> point) {

    	// If the map is already unmodifiable, return it
        if (point instanceof UnmodifiableMap) {
            return (UnmodifiableMap<?, ?>) point;
        }

        // Determining appropriate size
        int size = point.size() + 1;
        
        // Creating storage
        Map<Object, Object> p = new LinkedHashMap<Object, Object>(size, 1.0F);

        // Copying
        p.putAll(point);

        return new UnmodifiableMap<Object, Object>(p);
    }

}
