/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import java.util.*;

/**
 * This interface is used to represent linear combinations. A linear 
 * combination is a type of polynomial.
 * 
 * @author This interface was written and documented by 
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <V> the variable type. 
 */
public interface LinearComb<V> 
	extends Polynomial<V>, Iterable<LinearComb.Term<V>> {

    /**
	 * Returns the current zero cutoff.
	 * <p>
	 * The zero cutoff is a value used to decide if coefficients are close
	 * enough to zero to be considered equal to zero. In other words, the
	 * absolute value of the coefficient of every term of this linear
	 * combination will always be greater than the zero cutoff.
	 * 
	 * @return the zero cutoff.
	 */
	public double getZeroCutoff();

    /**
     * Returns the coefficient of the term containing the provided variable.
     * 
     * @param variable the variable whose corresponding coefficient is to be
     * returned.
     * 
     * @return the coefficient of the term containing the provided variable.
     */
    public double getCoefficient(Object variable);
    
    // Methods for addition

    /**
     * Adds, in the mathematical sense, the provided linear combination to this 
     * linear combination.
     * 
     * @param linComb the linear combination with which to add.
     * 
     * @return the result of adding this linear combination with the one 
     * provided.
     */
    public LinearComb<V> add(LinearComb<? extends V> linComb);

    /**
     * Adds, in the mathematical sense, the provided constant to this 
     * linear combination.
     * 
     * @param constant the constant with which to add.
     * 
     * @return the result of adding this linear combination with the provided
     * constant.
     */
    public LinearComb<V> add(double constant);

    
    // Methods for multiplying
    /**
     * Multiplies this linear combination by the provided constant.
     * 
     * @param constant the constant with which to multiply this linear 
     * combination.
     * 
     * @return the result of multiplying this linear combination by the 
     * provided constant.
     */
    public LinearComb<V> multiply(double constant);
    
    
    // Methods for solving for variables
    
    /**
     * Equates this linear combination to the provided constant and solves for
     * the provided variable.
     * 
     * @param variable the variable in this linear combination for which to 
     * solve.
     * @param constant the constant with which this linear combination is
     * equated.
     * 
     * @return a new linear combination representing the value of the provided
     * variable.
     */
    public LinearComb<V> solveFor(Object variable, double constant);


    // Methods for plugging in values for variables
    
    /**
     * Assigns the value of the provided variable to the provided linear 
     * combination.
     * 
     * @param variable the variable in this linear combination whose value
     * will be assigned the provided linear combination.
     * @param expr the linear combination to be "plugged in".
     * 
     * @return the result of "plugging in" the provided linear combination
     * for the provided variable.
     */
    public LinearComb<V> plugIn(Object variable, 
    							LinearComb<? extends V> expr);
    
    /**
     * Assigns the value of the provided variable to the provided constant.
     * 
     * @param variable the variable in this linear combination whose value
     * will be assigned the provided constant.
     * @param constant the value to be "plugged in".
     * 
     * @return the result of "plugging in" the provided constant
     * for the provided variable.
     */
    public LinearComb<V> plugIn(Object variable, double constant);
	
	/**
	 * Plugs in the provided numerical values to produce a new
	 * linear combination.
	 * <P>
	 * If some of the keys in the provided map are not variables
	 * in this expression, their provided value is simply ignored. 
	 * If a value is provided for each variable in this linear 
	 * combination, the resulting expression will not contain any
	 * variables of its own, only a constant.
	 * 
	 * @param numbers objects mapped to numerical values.
	 * 
	 * @return the linear combination obtained by "plugging in"
	 * the provided numerical values. 
	 */
    public LinearComb<V> plugInNumbers(Map<?, Double> numbers);
	
	/**
	 * Plugs in the provided linear combinations to produce a new
	 * linear combination.
	 * <P>
	 * If some of the keys in the provided map are not variables
	 * in this expression, their provided value is simply ignored.
	 * 
	 * @param linearCombs objects mapped to linear combinations.
	 * 
	 * @return the linear combination obtained by "plugging in"
	 * the provided linear combinations. 
	 */
    public LinearComb<V> 
    plugIn(Map<?, ? extends LinearComb<? extends V>> linearCombs);
	
	/**
	 * Plugs in the provided numerical values and linear combinations to
	 * produce a new linear combination.
	 * <P>
	 * If some of the keys in either of the provided maps are not variables
	 * in this expression, their provided values are simply ignored.
	 * 
	 * @param numbers objects mapped to numerical values.
	 * @param linearCombs objects mapped to linear combinations.
	 * 
	 * @return the linear combination obtained by "plugging in"
	 * the provided numerical values and linear combinations.
	 * 
	 *  @see #plugIn(Map)
	 *  @see #plugInNumbers(Map)
	 */
    public LinearComb<V> plugIn(Map<?, Double> numbers,
    		Map<?, ? extends LinearComb<? extends V>> linearCombs);
    
    
    /**
     * Returns the terms of this linear combination as list.
     * @return the terms of this linear combination as list.
     */
    public List<LinearComb.Term<V>> getTermList();
    
    /**
     * Returns the terms of this linear combination as an array.
     * @return the terms of this linear combination as an array.
     */
    public LinearComb.Term<V>[] getTermArray();
    
    /**
     * Returns an iterator over the terms of this linear combination.
     * @return an iterator over the terms of this linear combination.
     */
    public Iterator<LinearComb.Term<V>> iterator();

    /**
     * Returns a content-based hash code for this linear combination.
     * 
     * @return a content-based hash code for this linear combination.
     */
    public int hashCode();

    /**
     * Indicates if the provided object is equal to this
     * linear combination.
     * <P>
     * For two linear combinations to be equal, they must contain
     * the same numerical constant and terms.
     * 
     * @param obj the object to be tested for equality with
     * this linear combination.
     * 
     * @return {@code true} if this linear combination equals the provided 
     * object; {@code false} otherwise.
     * 
     * @see Term#equals(Object)
     */
    public boolean equals(Object obj);

    /**
     * Returns an appropriate string representation of this
     * linear combination.
     * 
     * @return an appropriate string representation of this
     * linear combination.
     */
    public String toString();

    /**
     * This interface is used to represent the terms of linear combinations.
     * A term contains a variable and a coefficient.
     * 
     * @author This interface was written and documented by Jeremiah Wright
     * while in the Wagner lab.
     *
     * @param <V> the variable type.
     */
    public static interface Term<V> extends Polynomial.Term<V> {

    	/**
    	 * Returns the variable contained in this term.
    	 * @return the variable contained in this term.
    	 */
        public V getVariable();

    }
}
