/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import java.util.*;

/**
 * This interface is used to represent mathematical expressions. 
 * <P>
 * A mathematical expression contains a numerical constant and
 * a set of variables to which numerical values may be assigned. 
 * This interface captures these basic facts about mathematical 
 * expressions.
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <V> the variable type.
 */
public interface MathExpr<V> {

	/**
	 * Returns the numerical constant of this mathematical expression.
	 * @return the numerical constant of this mathematical expression.
	 */
	public double getConstant();
	
	/**
	 * Plugs in the provided numerical values to produce a new
	 * mathematical expression.
	 * <P>
	 * If some of the keys in the provided map are not variables
	 * in this expression, their provided value is simply ignored. 
	 * If a value is provided for each variable in this mathematical
	 * expression, the resulting expression will not contain any
	 * variables of its own, only a constant.
	 * 
	 * @param variableValues objects mapped to numerical values.
	 * 
	 * @return the mathematical expression obtained by "plugging in"
	 * the provided numerical values. 
	 */
	public MathExpr<V> plugInNumbers(Map<?, Double> variableValues);
	
	/**
	 * Returns the set of variables contained in this
	 * mathematical expression.
	 * 
	 * @return the set of variables contained in this
	 * mathematical expression.
	 */
    public Set<V> getVariables();

    /**
     * Returns the number of variables in this mathematical
     * expression.
     * 
     * @return the number of variables in this mathematical
     * expression.
     */
    public int variables();
    
    /**
     * Indicates if the provided object is a variable contained
     * in this mathematical expression.
     * 
     * @param variable the variable whose presence in this
     * mathematical expression is to be tested.
     * 
     * @return {@code true} if this expression contains the
     * provided variable; {@code false} otherwise. 
     */
    public boolean hasVariable(Object variable);

    /**
     * Indicates if this mathematical expression contains any
     * variables.
     * 
     * @return {@code true} if this expression contains at least
     * one variable; {@code false} otherwise.
     */
    public boolean hasVariables();

    /**
     * Indicates if this mathematical expression contains all
     * of the variables in the provided collection.
     * 
     * @param variables the collection of variables whose
     * presence in this expression is to be tested.
     * 
     * @return {@code true} if each element of the provided
     * collection is a variable in this expression; 
     * {@code false} otherwise.
     */
    public boolean hasAllVariables(Collection<?> variables);

    /**
     * Returns a content-based hash code for this mathematical
     * expression.
     * 
     * @return a content-based hash code for this mathematical
     * expression.
     */
    public int hashCode();

    /**
     * Indicates if the provided object is equal to this
     * mathematical expression.
     * <P>
     * At minimum, for two expressions to be considered equal,
     * they must both contain the same variables and constant.
     * A stricter definition of equality is left to the 
     * extenders and implementors of this interface.
     * 
     * @param obj the object to be tested for equality with
     * this mathematical expression.
     * 
     * @return {@code true} if this mathematical expression is equal to the 
     * provided object; {@code false} otherwise.
     */
    public boolean equals(Object obj);

    /**
     * Returns an appropriate string representation of this
     * mathematical expression.
     * 
     * @return an appropriate string representation of this
     * mathematical expression.
     */
    public String toString();

}
