/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import java.util.*;

/**
 * This interface is used to represent polynomials. 
 * <P>
 * A polynomial is a type of mathematical expression. It is the
 * sum of a numerical constant and a set of terms. Each term 
 * contains a numerical
 * coefficient and at least one variable. Each variable in a term
 * can be raised to a non-negative whole number power. These basic 
 * facts about polynomials have been captured in this interface.
 * 
 * @author This interface was written and documented by 
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <V> the variable type.
 */
public interface Polynomial<V> extends MathExpr<V> {
	
	/**
	 * Plugs in the provided numerical values to produce a new
	 * polynomial.
	 * <P>
	 * If some of the keys in the provided map are not variables
	 * in this expression, their provided value is simply ignored. 
	 * If a value is provided for each variable in this polynomial,
	 * the resulting polynomial will not contain any
	 * terms of its own, only a constant.
	 * 
	 * @param variableValues objects mapped to numerical values.
	 * 
	 * @return the polynomial obtained by "plugging in"
	 * the provided numerical values in the standard mathematical
	 * sense. 
	 */
	public Polynomial<V> plugInNumbers(Map<?, Double> variableValues);

	/**
	 * Returns the numerical constant of this polynomial.
	 * @return the numerical constant of this polynomial.
	 */
    public double getConstant();

    /**
     * Returns an array containing all of the terms of this
     * polynomial.
     * 
     * @return an array containing all of the terms of this
     * polynomial.
     */
    public Term<V>[] getTermArray();
    
    /**
     * Returns the number of terms contained in this polynomial.
     * The numerical constant is not considered a term.
     * 
     * @return the number of terms contained in this polynomial.
     */
    public int terms();

    /**
     * Returns a content-based hash code for this polynomial.
     * 
     * @return a content-based hash code for this polynomial
     */
    public int hashCode();

    /**
     * Indicates if the provided object is equal to this
     * polynomial.
     * <P>
     * For two polynomials to be equal, they must contain
     * the same numerical constant and terms.
     * 
     * @param obj the object to be tested for equality with
     * this mathematical expression.
     * 
     * @return {@code true} if this polynomial is equal to the provided object;
     * {@code false} otherwise.
     * 
     * @see Term#equals(Object)
     */
    public boolean equals(Object obj);

    /**
     * Returns an appropriate string representation of this
     * polynomial.
     * 
     * @return an appropriate string representation of this
     * polynomial.
     */
    public String toString();

    /**
     * This interface is used to represent the terms of a 
     * polynomial. A term consists of a numerical coefficient
     * and a set of variables, with each variable raised to 
     * a non-negative whole number power.
     * 
     * @author This interface was written and documented by
     * Jeremiah Wright while in the Wagner lab.
     * 
     * @param <V> the variable type.
     */
	public static interface Term<V> extends Iterable<V> {
	    
		/**
		 * Returns the coefficient of this term.
		 * @return the coefficient of this term.
		 */
		public double getCoefficient();
		
		/**
		 * Returns an iterator over the variables in this term.
		 * @return an iterator over the variables in this term.
		 */
		public Iterator<V> iterator();

		/**
		 * Returns the power of the provided variable in this
		 * term.
		 * 
		 * @param variable the variable whose power is to be
		 * returned.
		 * 
		 * @return the power of the provided variable in this
		 * term. If the provided object is not a variable in
		 * this term, 0 is returned.
		 */
		public int power(Object variable);

		/**
		 * Returns the set of variables contained in this term.
		 * @return the set of variables contained in this term.
		 */
	    public Set<V> getVariables();

	    /**
	     * Returns the number of variables contained in this term.
	     * @return the number of variables contained in this term.
	     */
	    public int variables();
	    
	    /**
	     * Indicates if the provided object is a variable contained
	     * in this term.
	     * 
	     * @param variable the variable whose presence in this
	     * term is to be tested.
	     * 
	     * @return {@code true} if this term contains the
	     * provided variable; {@code false} otherwise. 
	     */
	    public boolean hasVariable(Object variable);

	    /**
	     * Indicates if this term contains all
	     * of the variables in the provided collection.
	     * 
	     * @param variables the collection of variables whose
	     * presence in this term is to be tested.
	     * 
	     * @return {@code true} if each element of the provided
	     * collection is a variable in this term; 
	     * {@code false} otherwise.
	     */
	    public boolean hasAllVariables(Collection<?> variables);

	    /**
	     * Returns a content-based hash code for this term.
	     * 
	     * @return a content-based hash code for this term.
	     */
	    public int hashCode();

	    /**
	     * Indicates if the provided object is equal to this
	     * term.
	     * <P>
	     * Two terms are considered equal if they 1) contain
	     * the same coefficient, 2) contain the same variables,
	     * and 3) the powers of each variable are the same in
	     * both terms.
	     * 
	     * @param obj the object to be tested for equality with
	     * this term.
	     * 
	     * @return {@code true} if this term is equal to the provided object;
	     * {@code false} otherwise.
	     */
	    public boolean equals(Object obj);

	    /**
	     * Returns an appropriate string representation of this
	     * term.
	     * 
	     * @return an appropriate string representation of this
	     * term.
	     */
	    public String toString();

	}
	
}
