/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math;

import java.util.*;

/**
 * This class is used to compare vectors of double precision numbers.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public class VectorComparer {

	/**
	 * Indicates if the provided numerical vectors are equivalent.
	 * 
	 * @param m1 variables mapped to a numerical value.
	 * @param m2 variables mapped to a numerical value.
	 * @param tolerance the amount by which two numerical values can differ
	 * but still be considered equivalent.
	 * 
	 * @return {@code true} if the provided vectors are equivalent; 
	 * {@code false} otherwise.
	 */
    public boolean areSame(Map<?, Double> m1, 
                           Map<?, Double> m2, 
                           double tolerance) {
        
        if (!m1.keySet().equals(m2.keySet())) {
            return false;
        }
        
        DoubleEquiv eqTester = new DoubleEquiv(tolerance);

        for (Map.Entry<?, Double> e : m1.entrySet()) {
            
            Object key = e.getKey();
            
            Double a = e.getValue();
            Double b = m2.get(key);
        
            if (!eqTester.areEquiv(a, b)) {
                return false;
            }
        }
        
        return true;
    }

	/**
	 * Returns the set of differences between the provided numerical vectors.
	 * Values are compared for equivalence for each variable that occurs in 
	 * both provided vectors.
	 * 
	 * @param <V> the variable type.
	 * @param m1 variables mapped to numerical values.
	 * @param m2 variables mapped to numerical values.
	 * @param tolerance the amount by which a two numerical values can differ
	 * but still be considered equivalent.
	 * 
	 * @return the set of differences between the provided vectors.
	 * 
	 * @see DoubleEquiv#areEquiv(double, double)
	 */
    public <V> Set<Diff<V>> getDiffs(Map<? extends V, Double> m1,
                                     Map<? extends V, Double> m2,
                                     double tolerance) {

        DoubleEquiv eqTester = new DoubleEquiv(tolerance);

        // Storage for variables with unequal intervals
        Set<Diff<V>> vars = new LinkedHashSet<Diff<V>>();

        for (Map.Entry<? extends V, Double> e : m1.entrySet()) {

            V variable = e.getKey();

            Double a = e.getValue();
            Double b = m2.get(variable);

            if (b == null) {
                continue;
            }

            if (eqTester.areEquiv(a, b)) {
                Diff<V> d = new Diff<V>(variable, a, b);
                vars.add(d);
            }

        }

        return vars;
    }

    /**
     * Instances of this class are used to hold a variable and a pair of
     * corresponding numerical values.
     * 
     * @author This class was written and documented by Jeremiah Wright while in
     * the Wagner lab.
     * 
     * @param <V> the variable type.
     */
    public static class Diff<V> {

        private V variable;

        private Double v1;

        private Double v2;

        private Diff(V variable, Double v1, Double v2) {
            this.variable = variable;
            this.v1 = v1;
            this.v2 = v2;
        }

        /**
         * Returns the variable for which the numerical values correspond.
         * @return the variable for which the numerical values correspond.
         */
        public V getVariable() {
            return variable;
        }

        /**
         * Returns the "first" value of the pair of values.
         * @return the "first" value of the pair of values.
         */
        public Double getValue1() {
            return v1;
        }

        /**
         * Returns the "second" value of the pair of values.
         * @return the "second" value of the pair of values.
         */
        public Double getValue2() {
            return v2;
        }
        
    }

}
