/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

/**
 * This abstract class is a skeleton implementation of
 * {@linkplain DoubleMatrixBuilder}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <R> the row type.
 * @param <C> the column type.
 */
public abstract class AbstractMdMatrix<R, C> implements DoubleMatrixBuilder<R, C> {

    /**
     * Indicates if this matrix is empty.
     * 
     * @return {@code rows() == 0 && columns() == 0}.
     */
    public boolean isEmpty() {
        return rows() == 0 && columns() == 0;
    }

    /**
     * Returns a string representation of this matrix.
     * @return a string representation of this matrix.
     */
    public String toString() {

    	// Getting system-dependent line separator string
        String newLine = System.getProperty("line.separator");
        
        // Creating string buffer
        StringBuffer s = new StringBuffer();

        // Iterating thru each column index
        for (C columnIndex : getColumnIndices()) {

        	// Appending to string buffer
            s.append("\t" + columnIndex);
        }

        // Appending line separator
        s.append(newLine);

        // Iterating thru each row index
        for (R rowIndex : getRowIndices()) {

        	// Appending row index
            s.append(rowIndex);
            
            // Iterating thru each column index
            for (C columnIndex : getColumnIndices()) {

            	// Appending element
                s.append("\t" + getElement(rowIndex, columnIndex));
            }
            
            // Appending line separator
            s.append(newLine);
        }

        // Returning string
        return s.toString();
    }

    /**
     * Indicates if the provided object is equal to this matrix.
     * Two matrices are considered equal if they share the same
     * sets of column and row indices and the elements in each
     * matrix for each column-row index pair are the same.
     * 
     * @param obj the object with which to compare.
     * 
     * @return {@code true} if this object equals the provided
     * object; {@code false} otherwise.
     */
    public boolean equals(Object obj) {

    	if (this == obj) {
            return true;
        }
        if (!(obj instanceof Matrix)) {
            return false;
        }
        
        // Casting safely
        Matrix<?, ?, ?> m = (Matrix) obj;
        
        // If the dimensions are not the same
        if (m.columns() != columns() || m.rows() != rows()) {
            return false;
        }
        
        // If column indices are not the same
        if (!m.getColumnIndices().equals(getColumnIndices())) {
            return false;
        }

        // If row indices are not the same
        if (!m.getRowIndices().equals(getRowIndices())) {
            return false;
        }

        // Iterating thru each row index
        for (Object rowIndex : m.getRowIndices()) {

        	// Iterating thru each column index
        	for (Object columnIndex : m.getColumnIndices()) {

        		// If the element at this row and column 
        		// is the same in both matrices
                if (m.getElement(rowIndex, columnIndex) !=
                    getElement(rowIndex, columnIndex)) {
                    return false;
                }
            }
        }

        return true;
    }

    /**
     * Returns a content-based hash code.
	 * @return a content-based hash code.
     */
    public int hashCode() {
    	return rows() ^ columns();
    }
    
    /**
     * Sets the corresponding elements of this matrix to 
     * those of the provided matrix.
     * 
     * @param m the matrix whose elements are to be copied 
     * into this matrix.
     */
    public void setElements(DoubleMatrix<? extends R, ? extends C> m) {

        // Iterating thru each column index
        for (C columnIndex : m.getColumnIndices()) {

        	// Iterating thru each row index
        	for (R rowIndex : m.getRowIndices()) {

        		// Getting element
                double value = m.getElement(rowIndex, columnIndex);

                // Setting element
                setElement(rowIndex, columnIndex, value);
            }
        }
    }

	
}
