/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import java.util.*;

import sbrt.kernel.math.HashLinearComb;
import sbrt.kernel.math.LinearComb;

/**
 * This abstract class is a skeleton implementation of
 * {@linkplain SmdMatrix}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <R> the row type.
 * @param <C> the column type.
 */
public abstract class AbstractSmdMatrix<R, C> extends
		AbstractMdMatrix<R, C> implements SmdMatrix<R, C> {

    /**
     * Sets the element at the provided row and column indices
     * to the provided double precision value.
     * 
     * @param <X> a type extending the primary row type.
     * @param <Y> a type extending the primary column type.
     * 
     * @param rowIndex the row index.
     * @param columnIndex the column index.
     * @param value the double precision element value.
     */
	public <X extends R, Y extends C> void 
	setElement(X rowIndex, Y columnIndex, double value) {

		// Convert it to a Double using Double.valueOf(double)
		setElement(rowIndex, columnIndex, Double.valueOf(value));
	}

    /**
     * Sets the corresponding elements of this matrix to 
     * those of the provided matrix.
     * 
     * @param m the matrix whose elements are to be copied 
     * into this matrix.
     */
	public void 
	setElements(Matrix<? extends R, ? extends C, ? extends Double> m) {

		// Iterating thru each column index
		for (C columnIndex : m.getColumnIndices()) {

			// Iterating thru each row index
			for (R rowIndex : m.getRowIndices()) {

				// Getting element at this row and column
				Double value = m.getElement(rowIndex, columnIndex);

				// Setting the element
				setElement(rowIndex, columnIndex, value);
			}
		}
	}

    /**
     * Sets the corresponding elements of this matrix to 
     * those of the provided matrix.
     * 
     * @param m the matrix whose elements are to be copied 
     * into this matrix.
     */
	public void setElements(SparseDoubleMatrix<? extends R, ? extends C> m) {

		// Iterating thru each column index
		for (C columnIndex : m.getColumnIndices()) {

			// Iterating thru each row index
			for (R rowIndex : m.getRowIndices()) {

				// Getting element at this row and column
				Double value = m.getElement(rowIndex, columnIndex);

				// Setting the element
				setElement(rowIndex, columnIndex, value);
			}
		}
	}

    /**
     * Sets the corresponding elements of this matrix to 
     * those provided.
     * 
     * @param elements the elements to be copied into this matrix.
     */
    public void setElements(Map<IndexPair<R, C>, Double> elements) {
        
        // Iterating thru each matrix element
        for(Map.Entry<IndexPair<R, C>, Double> e : elements.entrySet()) {
            
            IndexPair<R, C> ip = e.getKey();
            Double value = e.getValue();
            
            setElement(ip.getRowIndex(), ip.getColumnIndex(), value);
        }
    }

    /**
     * Indicates if the provided object is equal to this matrix.
     * Two matrices are considered equal if they share the same
     * sets of column and row indices and the elements in each
     * matrix for each column-row index pair are the same.
     * 
     * @param obj the object with which to compare.
     * 
     * @return {@code true} if this object equals the provided
     * object; {@code false} otherwise.
     */
	public boolean equals(Object obj) {

		if (this == obj) {
			return true;
		}
		if (!(obj instanceof SparseDoubleMatrix)) {
			return super.equals(obj);
		}

		// Casting safely
		SparseDoubleMatrix<?, ?> m = (SparseDoubleMatrix) obj;

		// If dimensions are not the same
		if (m.columns() != columns() || m.rows() != rows()) {
			return false;
		}

		// If column indices are not the same
		if (!m.getColumnIndices().equals(getColumnIndices())) {
			return false;
		}

		// If row indices are not the same
		if (!m.getRowIndices().equals(getRowIndices())) {
			return false;
		}

		// Iterating thru each row index
		for (Object rowIndex : getRowIndices()) {

			// Getting column indices of non-zero elements in the current row
			Set<?> columnIndices = getColIndNzEl(rowIndex);
			Set<?> mColumnIndices = m.getColIndNzEl(rowIndex);

			// If these sets are not the same
			if (!columnIndices.equals(mColumnIndices)) {
				return false;
			}

			// Iterating thru each column index
			for (Object columnIndex : columnIndices) {

				// If the elements are not the same
				if (m.getElement(rowIndex, columnIndex) != getElement(rowIndex,
						columnIndex)) {
					return false;
				}
			}
		}

		return true;
	}

    /**
     * Returns a content-based hash code.
	 * @return a content-based hash code.
     */
	public int hashCode() {
		return rows() ^ columns();
	}

    /**
     * Returns this matrix as a list of linear combinations. In other 
     * words, this method returns the product of this matrix with a 
     * column vector of the column indices of this matrix.
     * 
     * @return this matrix as a list of linear combinations.
     */
	@SuppressWarnings("unchecked")
	public List<LinearComb<C>> toLinearCombs() {

		// Creating storage
		List<LinearComb<C>> list = 
			new ArrayList<LinearComb<C>>(rows());

		// Iterating thru each row index
		for (R rowIndex : getRowIndices()) {

			// Getting the column indices of non-zero elements in this row
			Set<C> columnIndices = getColIndNzEl(rowIndex);

			// Determining appropriate size
			int size = columnIndices.size();

			// Creating storage for terms
			LinearComb.Term<C>[] terms = new LinearComb.Term[size];

			// Initializing the index
			int index = 0;

			// Iterating thru each column index
			for (C columnIndex : columnIndices) {

				// Getting the element
				double value = getElement(rowIndex, columnIndex);

				// Creating and storing the term
				terms[index++] = new HashLinearComb.Term<C>(columnIndex,
						value);
			}

			// Adding the linear combination to the rest
			list.add(new HashLinearComb<C>(terms, 0, 0));
		}

		return list;
	}

}
