/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import java.util.List;

import sbrt.kernel.math.LinearComb;

/**
 * This class is used to wrap mutable matrices as an immutable
 * {@code DoubleMatrix}. This type of wrapping eliminates the possibility of
 * casting a mutable matrix back to its mutable type. This ensures that
 * a mutable matrix can be viewed by others in a safe way, without
 * fear of modification. 
 * 
 * <p>Instances of this class are immutable.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <R> the row type.
 * @param <C> the column type.
 */
public class DoubleMatrixWrap<R, C> 
	extends MatrixWrap<R, C, Double> implements DoubleMatrix<R, C> {

	// The matrix to wrap
    private final DoubleMatrix<R, C> m;

    /**
     * Constructs a new wrapper around the provided matrix.
     * 
     * @param m the matrix to be wrapped by this wrapper object.
     * 
     * @throws NullPointerException if the provided matrix is
     * {@code null}.
     */
    public DoubleMatrixWrap(DoubleMatrix<R, C> m) {
        super(m);
        this.m = m;
    }

    /**
     * Returns the element of this matrix at the provided
     * row and column indices.
     * 
     * @param rowIndex the row index.
     * @param columnIndex the column index.
     * 
     * @return the element of this matrix at the provided
     * row and column indices.
     */
    public Double getElement(Object rowIndex, Object columnIndex) {
        return m.getElement(rowIndex, columnIndex);
    }

    /**
     * Returns the row corresponding to the provided row index.
     * A row is itself a 1 x n matrix, where n is the number of
     * columns in this matrix. 
     * 
     * @param rowIndex the index of the row to be returned.
     * 
     * @return the row corresponding to the provided row index.
     */
    public DoubleMatrix<R, C> getRow(Object rowIndex) {
        return m.getRow(rowIndex);
    }

    /**
     * Returns the column corresponding to the provided column index.
     * A column is itself an m x 1 matrix, where m is the number of
     * rows in this matrix. 
     * 
     * @param columnIndex the index of the column to be returned.
     * 
     * @return the column corresponding to the provided column index.
     */
    public DoubleMatrix<R, C> getColumn(Object columnIndex) {
        return m.getColumn(columnIndex);
    }

    /**
     * Returns this matrix as a list of linear combinations. In other 
     * words, this method returns the product of this matrix with a 
     * column vector of the column indices of this matrix.
     * 
     * @return this matrix as a list of linear combinations.
     */
    public List<LinearComb<C>> toLinearCombs() {
        return m.toLinearCombs();
    }

}
