/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;

/**
 * This class is a concrete implemenation of {@linkplain LinearSystemSolution}. The
 * hash codes of the provided variables are used extenstively in this
 * implementation to provide constant-time lookup for internally stored
 * information.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <V> the variable type.
 */
public class HashLinearSystemSolution<V> implements LinearSystemSolution<V> {

    /**
     * Returns a simplified version of the provided linear system solution
     * given the variable bounds. 
     * 
     * @param <V> the variable type.
     * 
     * @param solution a solution to a system of linear equations.
     * @param variableBounds the bounds on each variable in that solution.
     * @param tolerance the tolerance used to simplify the provided linear
     * system solution. See {@linkplain LinearSystemSolution#simplify(Map, double)}.
     * 
     * @return a simplified version of the provided linear system solution
     * given the variable bounds.
     */
    public static <V> LinearSystemSolution<V> 
    simplify(LinearSystemSolution<V> solution,
    		 Map<? extends V, Interval> variableBounds,
    		 double tolerance) {

        // Comparing the variables from both inputs
        if (!variableBounds.keySet().equals(solution.getVariables())) {
            String s = "The sets of variables don't match.";
            throw new IllegalArgumentException(s);
        }

        // Getting the values of variables whose constraints make them fixed
        Map<V, Double> variableValues = 
        	getFixedVariables(variableBounds);

        // Simplifying the linear system solution
        return solution.simplify(variableValues, tolerance);
    }

    private static <V> Map<V, Double> 
    getFixedVariables(Map<? extends V, Interval> variableBounds) {

        // Determining appropriate size
        int size = variableBounds.size() + 1;

        // Creating storage
        Map<V, Double> boundVars = new LinkedHashMap<V, Double>(size, 1.0F);

        // Iterating thru each variable-bound pair
        for (Map.Entry<? extends V, Interval> e : variableBounds.entrySet()) {

            // Getting the bounds
            Interval bounds = e.getValue();
            double upper = bounds.getUpper();
            double lower = bounds.getLower();
            
            // If the lower and upper bounds are equal
            if (upper == lower) {

                // Storing the value
                boundVars.put(e.getKey(), lower);
            }

        }

        return boundVars;
    }

    /**
     * All free variables in the system.
     */
    private final Set<V> freeVariables;

    /**
     * Fixed variables that equal a constant.
     */
    private final Map<V, Double> constants;

    /**
     * Fixed variables that equal a linear combination of free variables.
     */
    private final Map<V, LinearComb<V>> fixedVariables;

    private HashLinearSystemSolution(Set<V> freeVariables,
            						 Map<V, Double> constants, 
            						 Map<V, LinearComb<V>> fixedVariables) {

        // Copying free variables
        int size1 = freeVariables.size() + 1;
        this.freeVariables = new LinkedHashSet<V>(size1, 1.0F);
        this.freeVariables.addAll(freeVariables);

        // Copying fixed variables equal to a constant
        int size2 = constants.size() + 1;
        this.constants = new LinkedHashMap<V, Double>(size2, 1.0F);
        this.constants.putAll(constants);

        // Copying fixed variables equal to a linear
        // combination of free variables
        int size3 = fixedVariables.size() + 1;
        this.fixedVariables = new LinkedHashMap<V, LinearComb<V>>(size3, 1.0F);
        this.fixedVariables.putAll(fixedVariables);

    }

    /**
     * Constructs a new hash code-based linear system solution.
     * 
     * @param solutions variables mapped to the linear combination of free
     * variables to which they equal. The set of variables in the provided
     * linear combinations must be a subset of the keys in the provided map.
     * 
     * @throws MissingVariableException if at least one free variable does not
     * appear as a key in the provided map.
     */
    public HashLinearSystemSolution(
            Map<? extends V, ? extends LinearComb<? extends V>> solutions) {

        // Checking input
        checkSolution(solutions);

        // Determining free variables
        freeVariables = determineFreeVariables(solutions);

        // Determining variables equal to a constant
        constants = determineConstants(solutions);

        // Determining variables equal to a linear
        // combination of free variables
        fixedVariables = determineFixedVariables(solutions, freeVariables);
    }

    /**
     * Returns the free variables.
     */
    private Set<V> determineFreeVariables(Map<? extends V, ? extends LinearComb<? extends V>> solutions) {

        // Determining maximum possible size
        int size = solutions.size() + 1;

        // Creating storage
        Set<V> freeVars = new LinkedHashSet<V>(size, 1.0F);

        // Iterating thru each solution
        for (LinearComb<? extends V> sol : solutions.values()) {

            // Adding all variables in this solution
            // (i.e. adding all free variables)
            freeVars.addAll(sol.getVariables());
        }

        // Getting total number of free variables
        int n = freeVars.size();

        // Creating smallest possible storage
        Set<V> s = new LinkedHashSet<V>(n + 1, 1.0F);

        // Copying into smaller storage
        s.addAll(freeVars);

        return s;
    }

    /**
     * Returns the fixed variables that equal a constant.
     */
    private Map<V, Double> determineConstants(Map<? extends V, ? extends LinearComb<? extends V>> solutions) {

        // Determining maximum possible size
        int size = solutions.size() + 1;

        // Creating storage
        Map<V, Double> constants = new LinkedHashMap<V, Double>(size, 1.0F);

        // Iterating thru each variable-solution pair
        for (Map.Entry<? extends V, ? extends LinearComb<? extends V>> e : solutions
                .entrySet()) {

            // Getting the linear combination solution
            LinearComb<? extends V> sol = e.getValue();

            // If the solution does not have variables
            if (!sol.hasVariables()) {

                // Storing variable with constant
                constants.put(e.getKey(), sol.getConstant());
            }
        }

        // Getting total number of variables equal to a constant
        int n = constants.size();

        // Creating smallest possible storage
        Map<V, Double> m = new LinkedHashMap<V, Double>(n + 1, 1.0F);

        // Copying into smaller storage
        m.putAll(constants);

        return m;
    }

    /**
     * Returns the fixed variables that equal a linear combination of free
     * variables.
     */
    @SuppressWarnings("unchecked")
    private Map<V, LinearComb<V>> 
    determineFixedVariables(Map<? extends V, ? extends LinearComb<? extends V>> solutions,
    						Set<V> freeVariables) {

        // Determining maximum possible size
        int size = solutions.size() + 1;

        // Creating storage
        Map<V, LinearComb<V>> fixedVars =
                new LinkedHashMap<V, LinearComb<V>>(size, 1.0F);

        // Iterating thru each variable-solution pair
        for (Map.Entry<? extends V, ? extends LinearComb<? extends V>> e : solutions
                .entrySet()) {

            // Getting the variable-solution pair
            V variable = e.getKey();
            LinearComb<V> sol = (LinearComb<V>) e.getValue();

            // If the solution has no variables, move on to the next
            if (!sol.hasVariables()) {
                continue;
            }

            // If the variable is a free variable, move on to the next
            if (freeVariables.contains(variable)) {
                continue;
            }

            // Saving variable-solution pair
            fixedVars.put(variable, sol);
        }

        // Getting total number of fixed variables
        // equal to a linear combination of free variables
        int n = fixedVars.size();

        // Creating smallest possible storage
        Map<V, LinearComb<V>> m =
                new LinkedHashMap<V, LinearComb<V>>(n + 1, 1.0F);

        // Copying into smaller storage
        m.putAll(fixedVars);

        return m;
    }

    /**
     * Checks the integrity of the provided system solution.
     */
    private void checkSolution(Map<? extends V, ? extends LinearComb<? extends V>> solutions) {

        // Getting the variables on the "left hand side"
        Set<? extends V> leftSideVars = solutions.keySet();

        // Checking that no variables are null, just in
        // case the Map allows null keys
        for (V variable : leftSideVars) {

            if (variable == null) {
                String s = "null variables are not allowed.";
                throw new NullPointerException(s);
            }
        }

        // Creating storage
        Set<V> rightSideVars =
                new LinkedHashSet<V>(leftSideVars.size() + 1, 1.0F);

        // Gathering all variables on the "right-hand side"
        for (LinearComb<? extends V> sol : solutions.values()) {
            rightSideVars.addAll(sol.getVariables());
        }

        // Checking that all rhs variables are truly free variables
        // i.e. they equal themselves in the system solution
        for (V var : rightSideVars) {

            // Getting solution
            LinearComb<? extends V> sol = solutions.get(var);

            if (sol == null) {
                String s = "The variable " + var + 
                           " does not appear as a key in the provided map.";
                throw new IllegalArgumentException(s);
            }
            
            // If a) the solution has only one variable,
            // b) the solution has the free variable,
            // c) the coefficient for the free variable is 1.0,
            // & d) the constant is 0.0
            if (sol.variables() != 1 || 
            	!sol.hasVariable(var) ||
            	sol.getCoefficient(var) != 1d || 
            	sol.getConstant() != 0d) {

                throw new IllegalFreeVariableException(var);
            }
        }

        // Checking that all free variables appear on the left side too
        rightSideVars.removeAll(leftSideVars);

        if (!rightSideVars.isEmpty()) {
            V missingVariable = rightSideVars.iterator().next();
            throw new MissingVariableException(missingVariable);
        }

    }

    /**
     * Returns the total number of variables in this solution. This equals the
     * sum of free and fixed variables.
     * 
     * @return the total number of variables in this solution.
     * 
     * @see #freeVariables()
     * @see #fixedVariables()
     */
    public int variables() {
        return freeVariables() + fixedVariables();
    }

    /**
     * Returns the number of free variables in this solution.
     * 
     * @return the number of free variables in this solution.
     */
    public int freeVariables() {
        return freeVariables.size();
    }

    /**
     * Returns the number of fixed variables in this solution.
     * 
     * @return the number of fixed variables in this solution.
     */
    public int fixedVariables() {
        return constants.size() + fixedVariables.size();
    }

    /**
     * Returns the set of variables contained in this solution.
     * 
     * @return the set of variables contained in this solution.
     */
    public Set<V> getVariables() {

        // Determining appropriate size
        int size = variables() + 1;

        // Creating storage
        Set<V> variables = new LinkedHashSet<V>(size, 1.0F);

        // Adding all variables
        variables.addAll(constants.keySet());
        variables.addAll(fixedVariables.keySet());
        variables.addAll(freeVariables);

        return variables;
    }

    /**
     * Returns the set of free variables contained in this solution.
     * 
     * @return the set of free variables contained in this solution.
     */
    public Set<V> getFreeVariables() {

        // Determining appropriate size
        int size = freeVariables() + 1;

        // Creating storage
        Set<V> freeVars = new LinkedHashSet<V>(size, 1.0F);

        // Copying
        freeVars.addAll(freeVariables);

        return freeVars;
    }

    /**
     * Converts a constant to a linear combination.
     */
    private LinearComb<V> convertConstant(double constant) {

        // Creating empty list of terms
        List<LinearComb.Term<V>> terms = Collections.emptyList();

        return new HashLinearComb<V>(terms, constant, 0d);
    }

    /**
     * Converts a single variable to a linear combination.
     */
    private <T extends V> LinearComb<T> convertVar(T variable) {

        // Creating term
        LinearComb.Term<T> term = new HashLinearComb.Term<T>(variable);

        // Creating list of terms
        List<LinearComb.Term<T>> terms = Collections.singletonList(term);

        return new HashLinearComb<T>(terms, 0d, 0d);
    }

    /**
     * Returns the linear combination of free variables to which the provided
     * variable is equal.
     * 
     * @param variable a variable from this solution.
     * 
     * @return the linear combination of free variables to which the provided
     * variable is equal.
     */
    @SuppressWarnings("unchecked")
    public LinearComb<V> getSolution(Object variable) {

        // Looking for solution in fixedVariables first
        LinearComb<V> solution = fixedVariables.get(variable);

        // If it was found, return it
        if (solution != null) {
            return solution;
        }

        // Now looking in constants
        Double value = constants.get(variable);

        // If it was found, return it as a no-term linear combination
        if (value != null) {
            return convertConstant(value);
        }

        // Now looking in free variables
        if (freeVariables.contains(variable)) {

            // Casting safely
            V var = (V) variable;

            // Returning the free variable as a
            // single-term linear combination
            return convertVar(var);

        }

        // If it wasn't found anywhere
        String s = "The variable '" + variable + "' does not exist.";
        throw new IllegalArgumentException(s);
    }

    /**
     * Returns this solution as a map.
     * 
     * @return the variables of the original system of equations mapped to the
     * linear combination of free variables to which they equal.
     */
    public Map<V, LinearComb<V>> toMap() {

        // Determining appropriate size
        int size = variables() + 1;

        // Creating storage
        Map<V, LinearComb<V>> solutions =
                new LinkedHashMap<V, LinearComb<V>>(size, 1.0F);

        // Adding fixed variables
        solutions.putAll(fixedVariables);

        // Adding variables equal to a constant
        for (Map.Entry<V, Double> e : constants.entrySet()) {
            solutions.put(e.getKey(), convertConstant(e.getValue()));
        }

        // Adding free variables
        for (V variable : freeVariables) {
            solutions.put(variable, convertVar(variable));
        }

        return solutions;
    }

    /**
     * Computes and returns a solution vector by plugging in the provided values
     * for the free variables in this solution.
     * 
     * @param freeVariableValues free variables mapped to their corresponding
     * double precision value. Every key in the map must be a free variable in
     * this solution and all free variables must be present.
     * 
     * @return the solution vector resulting from the provided free variable
     * values.
     */
    @SuppressWarnings("unchecked")
    public Map<V, Double> getSolutionVector(Map<?, Double> freeVariableValues) {

        // Ensuring values for all free variables were provided
        if (!freeVariableValues.keySet().equals(freeVariables)) {
            String s = "The keys are invalid.";
            throw new IllegalArgumentException(s);
        }

        // Creating storage
        Map<V, Double> solutionVector =
                new LinkedHashMap<V, Double>(variables() + 1, 1.0F);

        // Iterating thru each free variable-value pair
        for (Map.Entry<?, Double> e : freeVariableValues.entrySet()) {

            // Getting the pair
            Object freeVar = e.getKey();
            Double value = e.getValue();

            // Checking the provided value
            checkValue(freeVar, value);

            // Adding it to the solution vector
            solutionVector.put((V) freeVar, value);
        }

        // Adding values for variables equal to a constant
        solutionVector.putAll(constants);

        // Adding values for fixed variables (not equal to a constant)
        for (Map.Entry<V, LinearComb<V>> e : fixedVariables.entrySet()) {

            // Getting the next pair
            V variable = e.getKey();
            LinearComb<? extends V> sol = e.getValue();

            // Plugging in the provided free variable values
            double value = sol.plugInNumbers(freeVariableValues).getConstant();

            // Saving
            solutionVector.put(variable, value);
        }

        // Returning the solution vector computed from the provided
        // free variable values
        return solutionVector;
    }

    /**
     * Computes and returns a simplified solution by plugging in the provided
     * values. Values can be provided for both free and fixed variables.
     * 
     * @param variableValues variables mapped to their corresponding double
     * precision value. Every key in the map must be a variable in this
     * solution.
     * 
     * @param tolerance the amount by which a provided or computed value is
     * allowed to differ from any previously existing value. This should be a
     * small positive number, like 1E-9 for example.
     * 
     * @return the solution resulting from plugging in the provided values.
     */
    @SuppressWarnings("unchecked")
    public LinearSystemSolution<V> simplify(Map<?, Double> variableValues,
            								double tolerance) {

        checkTolerance(tolerance);

        // Checking all variables and values provided
        checkValues(variableValues);

        // Checking for conflicts with constants
        compareToConstants(variableValues, tolerance);

        // Determining maximum possible size
        int size = variables() + 1;

        // Creating storage for constants in the simplified system
        Map<V, Double> newConstants = new LinkedHashMap<V, Double>(size, 1.0F);

        // Creating storage for free variables in the simplified system
        Set<V> newFreeVars = new LinkedHashSet<V>(size, 1.0F);

        // Creating storage for fixed variables in the simplified system
        Map<V, LinearComb<V>> newFixedVars =
                new LinkedHashMap<V, LinearComb<V>>(size, 1.0F);

        // Adding provided variable values to constants
        newConstants.putAll((Map<V, Double>) variableValues);

        // Preserving original values exactly
        newConstants.putAll(constants);

        // Plugging provided values into the fixed variables
        Map<V, LinearComb<V>> plugInResults =
                plugIntoFixedVariables(variableValues, tolerance);

        // Processing plug-in results
        for (Map.Entry<V, LinearComb<V>> e : plugInResults.entrySet()) {

            // Getting variable-solution pair
            V variable = e.getKey();
            LinearComb<V> solution = e.getValue();

            // If the solution still has variables,
            // it doesn't equal a constant
            if (solution.hasVariables()) {

                // Adding to the simplified fixed variables
                newFixedVars.put(variable, solution);

                // Checking for inconsistencies caused by numerical error
                if (newConstants.containsKey(variable)) {
                    String s =
                            "The value " + newConstants.get(variable)
                                    + " for "
                                    + variable
                                    + " was not plugged in properly. "
                                    + solution;
                    throw new IllegalStateException(s);
                }

            }

            // If it doesn't, the variable equals a constant
            else {
                newConstants.put(variable, solution.getConstant());
            }
        }

        // Gathering new free variables
        newFreeVars.addAll(freeVariables);
        newFreeVars.removeAll(newConstants.keySet());
        newFreeVars.removeAll(newFixedVars.keySet());

        // Returning the simplified linear system solution
        return new HashLinearSystemSolution<V>(newFreeVars, newConstants,
                newFixedVars);
    }

    private V getVarToSolveFor(LinearComb<V> expression) {

        // Creating field for the variable with coefficient
        // closest to 1 (or -1)
        V closestVariable = null;

        // Initializing to maximum possible value
        double closestToOne = Double.POSITIVE_INFINITY;

        // Iterating thru each term
        for (LinearComb.Term<V> t : expression) {

            // Getting the distance from 1
            double difference = Math.abs(t.getCoefficient()) - 1;

            // If this coefficient is closer to 1 than the closest
            // encountered so far, save it
            if (Math.abs(difference) < closestToOne) {
                closestToOne = difference;
                closestVariable = t.getVariable();
            }
        }

        // Returning the variable with coefficient whose
        // absolute value is closest to one
        return closestVariable;
    }

    @SuppressWarnings("unchecked")
    private Map<V, LinearComb<V>> plugIntoFixedVariables(Map<?, Double> variableValues,
            double tolerance) {

        // Determining appropriate size
        int size = fixedVariables.size() + 1;

        // Creating storage
        Map<V, LinearComb<V>> fixedVars =
                new LinkedHashMap<V, LinearComb<V>>(size, 1.0F);

        // Copying
        fixedVars.putAll(fixedVariables);

        // First plugging in numerical values. This can speed up the
        // next section a great deal.
        for (Map.Entry<V, LinearComb<V>> e : fixedVars.entrySet()) {

            // Getting the solution
            LinearComb<V> solution = e.getValue();

            // Plugging in variable values and saving
            e.setValue(solution.plugInNumbers(variableValues));
        }

        // Copying
        Map<V, Double> fixedVarValues =
                new LinkedHashMap<V, Double>((Map<V, Double>) variableValues);

        // Retaining values only for fixed variables
        fixedVarValues.keySet().retainAll(fixedVars.keySet());

        // Iterating thru values provided for the fixed variables
        for (Map.Entry<V, Double> e : fixedVarValues.entrySet()) {

            // Getting the variable-value pair
            V var = e.getKey();
            Double value = e.getValue();

            // Getting the solution for this variable
            LinearComb<V> solution = fixedVars.get(var);

            // If the solution no longer has variables
            if (!solution.hasVariables()) {

                // Checking the computing value
                compareConstants(var, solution.getConstant(), value, tolerance);

                // Move on to the next fixed variable
                continue;
            }

            // Getting an arbitrary variable to solve for
            V arbitraryVar = getVarToSolveFor(solution);

            // Solving for that variable
            solution = solution.solveFor(arbitraryVar, value);

            // Updating the solution for the arbitrary variable
            fixedVars.put(arbitraryVar, solution);

            // Propagating the solution for that arbitrary variable
            // throughout all solutions
            for (Map.Entry<V, LinearComb<V>> e2 : fixedVars.entrySet()) {

                // Getting old solution
                LinearComb<V> oldSol = e2.getValue();

                // Plugging in
                LinearComb<V> pluggedIn = oldSol.plugIn(arbitraryVar, solution);

                // Replacing with plug-in result
                e2.setValue(pluggedIn);
            }
        }

        return fixedVars;
    }

    private void compareToConstants(Map<?, Double> variableValues,
            double tolerance) {

        // Creating temporary fields
        Map<?, Double> biggest = null;
        Map<?, Double> smallest = null;

        // Determining which is bigger,
        // the provided values or the constants
        if (variableValues.size() > constants.size()) {
            biggest = variableValues;
            smallest = constants;
        } else {
            biggest = constants;
            smallest = variableValues;
        }

        // Iterating thru the smallest map to save time.
        // The values from both maps are checked, even though checking
        // the values from 'constants' isn't really necessary.
        for (Map.Entry<?, Double> e : smallest.entrySet()) {

            // Getting the variable
            Object variable = e.getKey();

            // Getting the value from this map and checking it
            Double thisValue = e.getValue();
            checkValue(variable, thisValue);

            // Getting the value for this variable from the other map
            Double otherValue = biggest.get(variable);

            // If no value exists, move on
            if (otherValue == null) {
                continue;
            }

            // Otherwise, check it
            checkValue(variable, otherValue);

            // Comparing values from both maps
            compareConstants(variable, thisValue, otherValue, tolerance);
        }

    }

    private void compareConstants(Object variable,
            double v1,
            double v2,
            double tolerance) {

        if (!Util.equal(v1, v2, tolerance)) {
            String s =
                    "The provided value for the variable '" + variable
                            + "' does not match it's current value: "
                            + v1
                            + " != "
                            + v2;
            throw new IllegalArgumentException(s);
        }

    }

    private void checkValues(Map<?, Double> variableValues) {

        // Checking the variables and values
        for (Map.Entry<?, Double> e : variableValues.entrySet()) {

            // Getting pair
            Object variable = e.getKey();
            Double value = e.getValue();

            // Checking each
            checkVariable(variable);
            checkValue(variable, value);
        }

    }

    /**
     * Indicates if the provided object is a variable in this solution.
     * 
     * @param obj the potential variable.
     * 
     * @return {@code true} if the provided object is a variable in this
     * solution; {@code false} otherwise.
     */
    public boolean isVariable(Object obj) {
        return freeVariables.contains(obj) || constants.containsKey(obj)
                || fixedVariables.containsKey(obj);
    }

    private void checkVariable(Object variable) {

        if (!isVariable(variable)) {
            String s = "The variable '" + variable + "' does not exist.";
            throw new IllegalArgumentException(s);
        }

    }

    private Double getValue(Object variable, Map<?, Double> vector) {

        Double value = vector.get(variable);
        checkValue(variable, value);

        return value;
    }

    private void checkTolerance(double x) {

        if (x < 0d || Double.isInfinite(x) || Double.isNaN(x)) {
            String s = "Invalid tolerance: " + x;
            throw new IllegalArgumentException(s);
        }

    }

    private void checkValue(Object variable, Double value) {

        // If value is null
        if (value == null) {
            String s = "No value was provided for the variable: " + variable;
            throw new NullPointerException(s);
        }

        // If value is infinite or NaN
        if (value.isInfinite() || value.isNaN()) {
            String s =
                    "The value '" + value
                            + "' for the variable '"
                            + variable
                            + "' is invalid.";
            throw new IllegalArgumentException(s);
        }

    }

    /**
     * Indicates if the provided solution vector is consistent with this
     * solution.
     * 
     * @param point the variables of this solution mapped to their values. Every
     * key in the map must be a variable in this solution and all variables must
     * be present.
     * 
     * @param tolerance the amount by which a provided value is allowed to
     * differ from any computed or previously existing value. This should be a
     * small positive number, like 1E-9 for example.
     * 
     * @return {@code true} if the provided vector is consistent with this
     * solution; {@code false} otherwise.
     */
    public boolean isSolutionVector(Map<?, Double> point, double tolerance) {

        checkTolerance(tolerance);

        // Checking that the number of variables in the point equals
        // the number of variables in the system
        if (point.size() != variables()) {
            String s =
                    "The point contains " + point.size()
                            + " variables "
                            + "but the system contains "
                            + variables()
                            + " variables.";
            throw new IllegalArgumentException(s);
        }

        // Checking every variable and value in the provided point
        for (Map.Entry<?, Double> e : point.entrySet()) {
            checkVariable(e.getKey());
            checkValue(e.getKey(), e.getValue());
        }

        // Iterating thru constants first since it's fast
        for (Map.Entry<V, Double> e : constants.entrySet()) {

            // Getting the variable that equals a constant
            V variable = e.getKey();

            // Getting the constant
            Double value = e.getValue();

            // Getting the value in the provided point
            Double pointValue = getValue(variable, point);

            // Comparing the two values
            if (!Util.equal(value, pointValue, tolerance)) {
//            	System.out.println(variable + ": " + value + " != " + pointValue);
                return false;
            }
        }

        // Iterating thru each fixed variable and its solution
        for (Map.Entry<V, LinearComb<V>> e : fixedVariables.entrySet()) {

            // Getting the variable
            V variable = e.getKey();

            // Getting the solution for that variable
            LinearComb<? extends V> solution = e.getValue();

            // Plugging the point into the solution
            double value = solution.plugInNumbers(point).getConstant();

            // Getting the value for this variable from the point
            Double pointValue = getValue(variable, point);

            // Comparing the two values
            if (!Util.equal(value, pointValue, tolerance)) {
                return false;
            }

        }

        return true;
    }

}
