/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import java.util.*;

import sbrt.kernel.utilities.*;

/**
 * This class is a {@code java.util.LinkedHashMap} based implementation of {@linkplain SmdMatrix}.
 * The hash codes of the provided variables are used extenstively in this 
 * implementation to provide constant-time lookup for internally stored
 * information.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 * @param <R> the row type.
 * @param <C> the column type.
 * 
 * @see HashMap
 */
public class HashMatrix<R, C> extends AbstractSmdMatrix<R, C> {

    private final Map<C, Map<R, Double>> columnToRow;
    private final Map<R, Map<C, Double>> rowToColumn;

    /**
     * Constructs an emtpy {@code HashMatrix} with the default
     * initial capacity (16) and the default load factor (0.75).
     */
    public HashMatrix() {
        columnToRow = new LinkedHashMap<C, Map<R, Double>>();
        rowToColumn = new LinkedHashMap<R, Map<C, Double>>();
    }

    /**
     * Constructs an empty {@code HashMatrix} with the provided 
     * initial capacities and load factor.
     * 
     * @param initialRowCapacity the initial row capacity.
     * @param initialColumnCapacity the initial column capacity.
     * @param loadFactor the load factor.
     */
    public HashMatrix(int initialRowCapacity, 
    				  int initialColumnCapacity,
                      float loadFactor) {
        columnToRow = 
        	new LinkedHashMap<C, Map<R, Double>>(initialColumnCapacity, 
        										 loadFactor);
        rowToColumn = 
        	new LinkedHashMap<R, Map<C, Double>>(initialRowCapacity, 
        										 loadFactor);
    }

    /**
     * Sets the element at the provided row and column indices
     * to the provided double precision value.
     * 
     * @param <X> a type extending the primary row type.
     * @param <Y> a type extending the primary column type.
     * @param <Z> Double.
     * 
     * @param rowIndex the row index.
     * @param columnIndex the column index.
     * @param value the double precision element value.
     */
    public <X extends R, Y extends C, Z extends Double> void 
    	setElement(X rowIndex, Y columnIndex, Z value) {
        
    	// Checking inputs
    	Util.nullCheck(rowIndex);
        Util.nullCheck(columnIndex);
        Util.nullCheck(value);

        Map<R, Double> rowIndices = columnToRow.get(columnIndex);
        if (rowIndices == null) {
            rowIndices = new LinkedHashMap<R, Double>();
            columnToRow.put(columnIndex, rowIndices);
        }

        Map<C, Double> columnIndices = rowToColumn.get(rowIndex);
        if (columnIndices == null) {
            columnIndices = new LinkedHashMap<C, Double>();
            rowToColumn.put(rowIndex, columnIndices);
        }

        // Ignoring values equal to 0
        if (Math.abs(value.doubleValue()) < Double.MIN_VALUE) {
            rowIndices.remove(rowIndex);
            columnIndices.remove(columnIndex);
        }

        else {
            rowIndices.put(rowIndex, value);
            columnIndices.put(columnIndex, value);
        }
    }

    /**
     * Indicates if the provided object is a row index in
     * this matrix.
     * 
     * @param obj the potential row index.
     * 
     * @return {@code true} if the provided object is a row
     * index in this matrix; {@code false} otherwise.
     */
    public boolean isRowIndex(Object obj) {
        return rowToColumn.containsKey(obj);
    }

    /**
     * Indicates if the provided object is a column index in
     * this matrix.
     * 
     * @param obj the potential column index.
     * 
     * @return {@code true} if the provided object is a column
     * index in this matrix; {@code false} otherwise.
     */
    public boolean isColumnIndex(Object obj) {
        return columnToRow.containsKey(obj);
    }

    /**
     * Returns the row corresponding to the provided row index.
     * A row is itself a 1 x n matrix, where n is the number of
     * columns in this matrix. 
     * 
     * @param rowIndex the index of the row to be returned.
     * 
     * @return the row corresponding to the provided row index.
     */
    @SuppressWarnings("unchecked")
    public HashMatrix<R, C> getRow(Object rowIndex) {

        // Ensuring the provided row index exists
        if (!isRowIndex(rowIndex)) {
            String s = "The row index '" + rowIndex + "' does not exist.";
            throw new NoSuchElementException(s);
        }

    	// Determining appropriate size
        int rows = 2;
        int columns = columns() + 1;

        // Creating storage
        HashMatrix<R, C> m = new HashMatrix<R, C>(rows, columns, 1.0F);

        // Creating a single zero object to save space
        Double zero = Double.valueOf(0d);
        
        // Iterating thru each column index, initially setting each value to zero.
        // This ensures that each column index appears the returned row matrix.
        for (C columnIndex : columnToRow.keySet()) {
            m.setElement( (R) rowIndex, columnIndex, zero);
        }
        
        // Setting the actual values.
        for (Map.Entry<C, Double> e : rowToColumn.get(rowIndex).entrySet()) {
            m.setElement( (R) rowIndex, e.getKey(), e.getValue());
        }
        
        return m;
    }

    /**
     * Returns the column corresponding to the provided column index.
     * A column is itself an m x 1 matrix, where m is the number of
     * rows in this matrix. 
     * 
     * @param columnIndex the index of the column to be returned.
     * 
     * @return the column corresponding to the provided column index.
     */
    @SuppressWarnings("unchecked")
    public HashMatrix<R, C> getColumn(Object columnIndex) {

        // Ensuring the provided column index exists
        if (!isColumnIndex(columnIndex)) {
            String s = "The column index '" + columnIndex + "' does not exist.";
            throw new NoSuchElementException(s);
        }

    	// Determining appropriate size
        int rows = rows() + 1;
        int columns = 2;

        // Creating storage
        HashMatrix<R, C> m = new HashMatrix<R, C>(rows, columns, 1.0F);

        // Creating a single zero object to save space
        Double zero = Double.valueOf(0d);
        
        // Iterating thru each row index, initially setting each value to zero.
        // This ensures that each row index appears the returned column matrix.
        for (R rowIndex : rowToColumn.keySet()) {
            m.setElement(rowIndex, (C) columnIndex, zero);
        }
        
        // Setting the actual values
        for (Map.Entry<R, Double> e : columnToRow.get(columnIndex).entrySet()) {
            m.setElement(e.getKey(), (C) columnIndex, e.getValue());
        }
        
        return m;
    }

    /**
     * Removes the row corresponding to the provided 
     * row index.
     * 
     * @param rowIndex the index of the row to be removed.
     */
    public void removeRow(Object rowIndex) {

        // Getting the column indices of non-zero elements in this row
        Set<C> columnIndices = rowToColumn.get(rowIndex).keySet();

        // Ensuring the provided row index exists
        if (columnIndices == null) {
            String s = "The row index '" + rowIndex + "' does not exist.";
            throw new NoSuchElementException(s);
        }

        // Iterating thru each column index of non-zero elements in this row
        for (C columnIndex : rowToColumn.get(rowIndex).keySet()) {
            
            // Getting the row indices of non-zero elements in this column
            Map<R, Double> rowIndices = columnToRow.get(columnIndex);
            
            // Removing the row index from these indices
            rowIndices.remove(rowIndex);
            
        }
        
        rowToColumn.remove(rowIndex);

    }

    /**
     * Removes the column corresponding to the provided 
     * column index.
     * 
     * @param columnIndex the index of the column to be removed.
     */
    public void removeColumn(Object columnIndex) {

        // Getting the row indices of non-zero elements in this column
        Set<R> rowIndices = columnToRow.get(columnIndex).keySet();
        
        // Ensuring the provided column index exists
        if (rowIndices == null) {
            String s = "The column index '" + columnIndex + "' does not exist.";
            throw new NoSuchElementException(s);
        }

        // Iterating thru each row index of non-zero elements in this column
        for (R rowIndex : rowIndices) {
            
            // Getting the column indices of non-zero elements in this row
            Map<C, Double> columnIndices = rowToColumn.get(rowIndex);
            
            // Removing the column index from these indices
            columnIndices.remove(columnIndex);
            
        }
        
        columnToRow.remove(columnIndex);

    }

    /**
     * Returns the element of this matrix at the provided
     * row and column indices.
     * 
     * @param rowIndex the row index.
     * @param columnIndex the column index.
     * 
     * @return the element of this matrix at the provided
     * row and column indices.
     */
    public Double getElement(Object rowIndex, Object columnIndex) {

        Map<R, Double> rowIndices = columnToRow.get(columnIndex);
        
        // Ensuring provided column index exists
        if (rowIndices == null) {
            String s = "The column index '" + columnIndex + "' does not exist.";
            throw new IllegalArgumentException(s);
        }

        // Ensuring provided row index exists
        if (!isRowIndex(rowIndex)) {
            String s = "The row index '" + rowIndex + "' does not exist.";
            throw new IllegalArgumentException(s);
        }

        // Getting the value (if it does not equal zero)
        Double value = rowIndices.get(rowIndex);

        return (value == null ? Double.valueOf(0.0d) : value);
    }
    
    /**
     * Returns the set of column indices contained in this
     * matrix.
     * 
     * @return the set of column indices contained in this
     * matrix.
     */
    public Set<C> getColumnIndices() {

    	// Creating storage
        Set<C> s = new LinkedHashSet<C>(columnToRow.size() + 1, 1.0F);
        
        // Copying
        s.addAll(columnToRow.keySet());

        return s;
    }
    
    /**
     * Returns the set of row indices contained in this
     * matrix.
     * 
     * @return the set of row indices contained in this
     * matrix.
     */
    public Set<R> getRowIndices() {

    	// Creating storage
        Set<R> s = new LinkedHashSet<R>(rowToColumn.size() + 1, 1.0F);
        
        // Copying
        s.addAll(rowToColumn.keySet());

        return s;
    }

	/**
	 * Returns the set of row indices whose elements are nonzero 
	 * in the column corresponding to the provided index.
	 * 
	 * @param columnIndex the column index.
	 * 
	 * @return the set of row indices whose elements are nonzero 
	 * in the column corresponding to the provided index.
	 */
    public Set<R> getRowIndNzEl(Object columnIndex) {

    	// Getting rows -> values
        Map<R, Double> rowIndices = columnToRow.get(columnIndex);

        // If none exist for the provided column
        if (rowIndices == null) {
            String s = "The column index '" + columnIndex + "' does not exist.";
            throw new NoSuchElementException(s);
        }

        // Creating storage
        Set<R> s = new LinkedHashSet<R>(rowIndices.size() + 1, 1.0F);
        
        // Copying
        s.addAll(rowIndices.keySet());

        return s;
    }

	/**
	 * Returns the set of column indices whose elements are nonzero 
	 * in the row corresponding to the provided index.
	 * 
	 * @param rowIndex the row index.
	 * 
	 * @return the set of column indices whose elements are nonzero 
	 * in the row corresponding to the provided index.
	 */
    public Set<C> getColIndNzEl(Object rowIndex) {

    	// Getting columns -> values
        Map<C, Double> columnIndices = rowToColumn.get(rowIndex);

        // If none exist for the provided row
        if (columnIndices == null) {
            String s = "The row index '" + rowIndex + "' does not exist.";
            throw new NoSuchElementException(s);
        }

        // Creating storage
        Set<C> s = new LinkedHashSet<C>(columnIndices.size() + 1, 1.0F);
        
        // Copying
        s.addAll(columnIndices.keySet());

        return s;
    }

    /**
     * Returns the number of rows in this matrix.
     * @return the number of rows in this matrix.
     */
    public int rows() {
        return rowToColumn.size();
    }

    /**
     * Returns the number of columns in this matrix.
     * @return the number of columns in this matrix.
     */
    public int columns() {
        return columnToRow.size();
    }

    /**
     * Returns a content-based hash code for this matrix.
     * @return a content-based hash code for this matrix.
     */
    public int hashCode() {
        return rowToColumn.hashCode() ^ columnToRow.hashCode();
    }

}
