/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import sbrt.kernel.utilities.Util;

/**
 * This class is used to represent pairs of matrix indices.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <R> the row index type.
 * @param <C> the column index type.
 */
public final class IndexPair<R, C> {

    private final R rowIndex;
    private final C columnIndex;
    
    /**
     * Constructs a new pair of matrix indices.
     * 
     * @param <S> a type extending the primary row type.
     * @param <T> a type extending the primary column type.
     * @param rowIndex the row index.
     * @param columnIndex the column index.
     */
    public <S extends R, T extends C> IndexPair(S rowIndex, T columnIndex) {
        
        Util.nullCheck(rowIndex);
        Util.nullCheck(columnIndex);
        
        this.rowIndex = rowIndex;
        this.columnIndex = columnIndex;
    }
    
    /**
     * Returns the row index.
     * @return the row index.
     */
    public R getRowIndex() {
        return rowIndex;
    }

    /**
     * Returns the column index.
     * @return the column index.
     */
    public C getColumnIndex() {
        return columnIndex;
    }
    
    /**
     * Returns a content-based hash code.
     * @return a content-based hash code.
     */
    public int hashCode() {
        return rowIndex.hashCode() ^ columnIndex.hashCode();
    }
    
    /**
     * Indicates if this index pair is equal to the provided object. Two 
     * index pairs are equal if their row and column indices are equal,
     * respectively.
     * 
     * @param obj the reference object with which to compare.  
     * 
     * @return {@code true} if this index pair equals the provided object;
     * {@code false} otherwise.
     */
    public boolean equals(Object obj) {
        
        if (obj == this) {
            return true;
        }
        
        if (obj == null || !(obj instanceof IndexPair)) {
            return false;
        }
        
        IndexPair<?, ?> ip = (IndexPair) obj;
        
        if (!rowIndex.equals(ip.rowIndex)) {
            return false;
        }
        
        if (!columnIndex.equals(ip.columnIndex)) {
            return false;
        }
        
        return true;
    }
}
