/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import java.util.*;

import sbrt.kernel.math.*;

/**
 * This interface is used to represent a general solution to a
 * system of linear equations. It is primarily intended for underdetermined
 * systems,  whose solutions consist of both free and fixed variables.
 * In this representation, the variables 
 * contained in the
 * solution are the same as those contained in the original system of equations.
 * So in the solution, each variable from the original system is equal to a 
 * linear combination
 * of those same variables. Take, for instance, the system
 * <table align=center><tr><td>x + y = 1</tr>
 * <tr><td>2x + 2y = 2.</tr></table>
 * One solution is
 * <table align=center><tr><td>x = x</tr>
 * <tr><td>y = 1 - x,</tr></table>
 * where x is a free variable and y is fixed. By assigning a value to x, the 
 * value of y is also determined. Another equivalent solution is
 * <table align=center><tr><td>x =  1 - y</tr>
 * <tr><td>y = y,</tr></table>
 * where this time y is a free variable and x is fixed. Note that free variables
 * are simply equal to themselves, and fixed variables are equal to linear
 * combinations of free variables.  Some fixed variables may even equal just
 * a constant.  This interface captures
 * these basic facts about solutions to systems of linear equations.
 *  
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <V> the variable type.
 */
public interface LinearSystemSolution<V> {

	/**
	 * Returns the set of variables contained in this solution.
	 * 
	 * @return the set of variables contained in this solution.
	 */
    public Set<V> getVariables();

	/**
	 * Returns the set of free variables contained in this
	 * solution.
	 * 
	 * @return the set of free variables contained in this
	 * solution.
	 */
    public Set<V> getFreeVariables();

    /**
     * Indicates if the provided object is a variable in this
     * solution.
     * 
     * @param obj the potential variable.
     * 
     * @return {@code true} if the provided object is a variable
     * in this solution; {@code false} otherwise.
     */
    public boolean isVariable(Object obj);

    /**
     * Returns the linear combination of free variables to which the provided
     * variable is equal.
     * 
     * @param variable a variable from this solution.
     * 
     * @return the linear combination of free variables to which the provided
     * variable is equal.
     */
    public LinearComb<V> getSolution(Object variable);

    /**
     * Returns this solution as a map.
     * @return the variables of the original system of equations mapped to
     * the linear combination of free variables to which they equal. 
     */
    public Map<V, LinearComb<V>> toMap();

    /**
     * Computes and returns a simplified solution by plugging in the provided
     * values. Values can be provided for both free and fixed variables.
     *  
     * @param variableValues variables mapped to their
     * corresponding double precision value. Every key in the map must be a 
     * variable in this solution.
     * 
     * @param tolerance the amount by which a provided or computed value is
     * allowed to differ from any previously existing value. This should be 
     * a small positive number, like 1E-9 for example.
     * 
     * @return the solution resulting from plugging in the provided values.
     */
    public LinearSystemSolution<V> 
    simplify(Map<?, Double> variableValues, double tolerance);

    /**
     * Computes and returns a solution vector by plugging in the 
     * provided values for the free variables in this solution.
     *
     * @param freeVariableValues free variables mapped to their
     * corresponding double precision value.  Every key in the map must be a free variable
     * in this solution and all free variables must be present.
     *
     * @return the solution vector resulting from the provided free variable
     * values.
     */
    public Map<V, Double> 
    getSolutionVector(Map<?, Double> freeVariableValues);

    /**
     * Indicates if the provided solution vector is consistent with 
     * this solution.
     * 
     * @param variableValues the variables of this solution mapped to their
     * values.  Every key in the map must be a variable
     * in this solution and all variables must be present.
     * 
     * @param tolerance the amount by which a provided value is
     * allowed to differ from any computed or previously existing value.
     * This should be a small positive number, like 1E-9 for example.
     * 
     * @return {@code true} if the provided vector is consistent with this
     * solution; {@code false} otherwise.
     */
    public boolean 
    isSolutionVector(Map<?, Double> variableValues, double tolerance);

    /**
     * Returns the total number of variables in this solution. This equals
     * the sum of free and fixed variables.
     * 
     * @return the total number of variables in this solution.
     * 
     * @see #freeVariables()
     * @see #fixedVariables()
     */
    public int variables();

    /**
     * Returns the number of free variables in this solution.
     * @return the number of free variables in this solution.
     */
    public int freeVariables();

    /**
     * Returns the number of fixed variables in this solution.
     * @return the number of fixed variables in this solution.
     */
    public int fixedVariables();

}
