/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import java.util.*;

/**
 * This interface is used to represent matrices. A matrix is 
 * composed of rows and columns, which are referred to by 
 * their corresponding indices. Typically these indices are 
 * integer values from 1 to <i>m</i> and <i>n</i>, where 
 * <i>m</i> is the number of rows and <i>n</i> is the number
 * of columns. In this matrix representation, however, indices
 * can be any type of {@code Object}. Indices are compared
 * via their {@linkplain Object#equals(Object)} method.
 *  
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <R> the row index type.
 * @param <C> the column index type.
 * @param <E> the element type.
 */
public interface Matrix <R, C, E> {

    /**
     * Returns the element of this matrix at the provided
     * row and column indices.
     * 
     * @param rowIndex the row index.
     * @param columnIndex the column index.
     * 
     * @return the element of this matrix at the provided
     * row and column indices.
     */
    public E getElement(Object rowIndex, Object columnIndex);
    
    /**
     * Returns the set of column indices contained in this
     * matrix.
     * 
     * @return the set of column indices contained in this
     * matrix.
     */
    public Set<C> getColumnIndices();
    
    /**
     * Returns the set of row indices contained in this
     * matrix.
     * 
     * @return the set of row indices contained in this
     * matrix.
     */
    public Set<R> getRowIndices();

    /**
     * Indicates if the provided object is a row index in
     * this matrix.
     * 
     * @param obj the potential row index.
     * 
     * @return {@code true} if the provided object is a row
     * index in this matrix; {@code false} otherwise.
     */
    public boolean isRowIndex(Object obj);

    /**
     * Indicates if the provided object is a column index in
     * this matrix.
     * 
     * @param obj the potential column index.
     * 
     * @return {@code true} if the provided object is a column
     * index in this matrix; {@code false} otherwise.
     */
    public boolean isColumnIndex(Object obj);

    /**
     * Returns the row corresponding to the provided row index.
     * A row is itself a 1 x n matrix, where n is the number of
     * columns in this matrix. 
     * 
     * @param rowIndex the index of the row to be returned.
     * 
     * @return the row corresponding to the provided row index.
     */
    public Matrix<R, C, E> getRow(Object rowIndex);

    /**
     * Returns the column corresponding to the provided column index.
     * A column is itself an m x 1 matrix, where m is the number of
     * rows in this matrix. 
     * 
     * @param columnIndex the index of the column to be returned.
     * 
     * @return the column corresponding to the provided column index.
     */
    public Matrix<R, C, E> getColumn(Object columnIndex);

    /**
     * Returns the number of rows in this matrix.
     * @return the number of rows in this matrix.
     */
    public int rows();

    /**
     * Returns the number of columns in this matrix.
     * @return the number of columns in this matrix.
     */
    public int columns();

    /**
     * Indicates if this matrix is empty.
     * 
     * @return {@code true} if at least one row and column exists
     * in this matrix; {@code false} otherwise.
     */
    public boolean isEmpty();

    /**
     * Returns a string representation of this matrix.
     * @return a string representation of this matrix.
     */
    public String toString();

    /**
     * Indicates if the provided object is equal to this matrix.
     * Two matrices are considered equal if they share the same
     * sets of column and row indices and the elements in each
     * matrix for each column-row index pair are the same.
     * 
     * @param obj the object with which to compare.
     * 
     * @return {@code true} if this object equals the provided
     * object; {@code false} otherwise.
     */
    public boolean equals(Object obj);

    /**
     * Returns a content-based hash code for this matrix.
     * @return a content-based hash code for this matrix.
     */
    public int hashCode();

}
