/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import sbrt.kernel.utilities.Util;

/**
 * This class is used to represent matrix elements.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <R> the row index type.
 * @param <C> the column index type.
 * @param <E> the matrix element type.
 */
public final class MatrixElement<R, C, E> {

    private final IndexPair<R, C> indices;
    private final E element;
    
    /**
     * Constructs a new matrix element.
     * @param <S> a type extending the primary row type.
     * @param <T> a type extending the primary column type.
     * @param <U> a type extending the primary element type.
     * @param rowIndex the row index.
     * @param columnIndex the column index.
     * @param element the matrix element at the specified row and column indices.
     */
    public <S extends R, T extends C, U extends E> 
        MatrixElement(S rowIndex, T columnIndex, U element) {
        
        indices = new IndexPair<R, C>(rowIndex, columnIndex);

        Util.nullCheck(element);
        this.element = element;
    }

    /**
     * Constructs a new matrix element.
     * @param <U> a type extending the primary element type.
     * @param indices the row and column index of this element.
     * @param element the matrix element at the specified row and column indices.
     */
    public <U extends E> MatrixElement(IndexPair<R, C> indices, U element) {
        
        Util.nullCheck(indices);
        Util.nullCheck(element);
        
        this.indices = indices;
        this.element = element;
    }
    
    /**
     * Returns the row index of this matrix element.
     * @return the row index of this matrix element.
     */
    public R getRowIndex() {
        return indices.getRowIndex();
    }

    /**
     * Returns the column index of this matrix element.
     * @return the column index of this matrix element.
     */
    public C getColumnIndex() {
        return indices.getColumnIndex();
    }

    /**
     * Returns the pair of indices.
     * @return the pair of indices.
     */
    public IndexPair<R, C> getIndices() {
        return indices;
    }
    
    /**
     * Returns the value of this matrix element.
     * @return the value of this matrix element.
     */
    public E getElement() {
        return element;
    }
    
    /**
     * Indicates if this matrix element is equal to the provided object. Two 
     * matrix elements are equal if their row index, column index, and element
     * are equal, respectively.
     * 
     * @param obj the reference object with which to compare.  
     * 
     * @return {@code true} if this matrix element equals the provided object;
     * {@code false} otherwise.
     */
    public boolean equals(Object obj) {
        
        if (obj == null || !(obj instanceof MatrixElement)) {
            return false;
        }
        
        MatrixElement<?, ?, ?> me = (MatrixElement) obj;
        
        if (!element.equals(me.element)) {
            return false;
        }
        
        if (!indices.equals(me.indices)) {
            return false;
        }
        
        return true;
    }
    
    /**
     * Returns a content-based hash code.
     * @return a content-based hash code.
     */
    public int hashCode() {
    	return 37 * indices.hashCode();
    }
}
