/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import java.util.Set;

/**
 * This class is used to multiply matrices.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 */
public class MatrixMultiplier {

    /**
     * Fills {@code product} with {@code matrixA} * {@code matrixB}.
     * @param <R> the row type of {@code matrixA}.
     * @param <C> the column type of {@code matrixA} and row type of {@code matrixB}.
     * @param <T> the column type of {@code matrixB}.
     * @param matrixA a matrix.
     * @param matrixB a matrix.
     * @param product an empty matrix.
     */
    public <R, C, T> void multiply(SparseDoubleMatrix<R, C> matrixA,
                                   SparseDoubleMatrix<C, T> matrixB,
                                   DoubleMatrixBuilder<R, T> product) {
        

        // Ensuring the provided product matrix is empty
        if (!product.isEmpty()) {
            throw new IllegalArgumentException("matrix must be empty");
        }

        // If the number of columns in matrix A does not
        // equal the number of rows in matrix B
        if (matrixA.columns() != matrixB.rows()) {
            String s = matrixA.rows() + "x" + matrixA.columns() + " * " + 
                       matrixB.rows() + "x" + matrixB.columns() + " is undefined";
            throw new IllegalArgumentException(s);
        }

        // If the column indices in matrix A do not match
        // the row indices in matrix B
        if (!matrixB.getRowIndices().equals(matrixA.getColumnIndices())) {
            String s = "row indices in matrix 'm' do not match the "
                    + "column indices in this matrix";
            throw new IllegalArgumentException(s);
        }

        // Getting column indices from the provided matrix
        Set<T> mColumnIndices = matrixB.getColumnIndices();

        // Iterating thru each row index of this matrix
        for (R rowIndex : matrixA.getRowIndices()) {

            // Getting the corresponding row
            SparseDoubleMatrix<R, C> row = matrixA.getRow(rowIndex);

            // Iterating thru each column index of the provided matrix
            for (T columnIndex : mColumnIndices) {

                // Getting column vector from provided matrix
                DoubleMatrix<C, T> column = matrixB.getColumn(columnIndex);

                // Computing dot product of row and column vectors
                double dotProd = getRowColumnProduct(row, column);

                // Setting element in the empty matrix
                product.setElement(rowIndex, columnIndex, dotProd);
            }

        }

        
    }
    

    /**
     * Returns the product of the provided row and column vectors.
     * 
     * @param <R> the row type of the row vector.
     * @param <C> the column type of the row vector.
     * @param <T> the column type of the column vector.
     * 
     * @param rowVector the row vector
     * @param columnVector the column vector.
     * 
     * @return the product of the provided row and column vectors.
     * 
     * @throws IllegalArgumentException 
     * if {@code rowVector} has more than one row; 
     * if {@code columnVector} has more than one column; 
     * if {@code rowVector.columns() != columnVector.rows()}
     * 
     * @throws NullPointerException if either argument is {@code null}.
     */
    private <R, C, T> double 
        getRowColumnProduct(SparseDoubleMatrix<R, C> rowVector,
                            DoubleMatrix<C, T> columnVector) {

        // Ensuring the provided row vector contains one (and only one) row
        if (rowVector.rows() != 1) {
            String s = "The provided row vector has more than one row. "
                    + "It's size is " + rowVector.rows() + " x "
                    + rowVector.columns();
            throw new IllegalArgumentException(s);
        }

        // Ensuring the provided column vector contains one (and only one) column
        if (columnVector.columns() != 1) {
            String s = "The provided column vector has more than one column. "
                    + "It's size is " + columnVector.rows() + " x "
                    + columnVector.columns();
            throw new IllegalArgumentException(s);
        }

        // Ensuring the number of columns in the row vector equals
        // the number of rows in the column vector
        if (rowVector.columns() != columnVector.rows()) {
            String s = rowVector.rows() + " x " + rowVector.columns() + " * " + 
                       columnVector.rows() + " x " + columnVector.columns() + 
                       " is undefined";
            throw new IllegalArgumentException(s);
        }

        // Getting the one and only row index in the row vector
        R rowIndex = rowVector.getRowIndices().iterator().next();

        // Getting the one and only column index in the column vector
        T cIndex = columnVector.getColumnIndices().iterator().next();

        // Getting the column indices of non-zero elements in the row vector
        Set<? extends C> cols = rowVector.getColIndNzEl(rowIndex);

        // Initializing sum
        double sum = 0;

        // Iterating thru the column indices of non-zero elements
        for (C columnIndex : cols) {

            // Getting the element from this matrix
            double element1 = rowVector.getElement(rowIndex, columnIndex);

            // Getting the element from the provided matrix
            double element2 = columnVector.getElement(columnIndex, cIndex);

            // Updating sum
            sum += element1 * element2;
        }

        return sum;
    }

    
}
