/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

/**
 * This class is used to transpose matrices.
 * 
 * @author This class was written and documented by Jeremiah Wright 
 * while in the Wagner lab.
 *
 */
public class MatrixTransposer {

    /**
     * Fills {@code emptyMatrix} with the transpose of {@code matrix}.
     * 
     * @param <R> the row index type. 
     * @param <C> the column index type.
     * @param <E> the element or value type.
     * 
     * @param matrix the matrix for which a transpose will be created. 
     * @param emptyMatrix the empty matrix where the tranpose will be stored.
     * 
     * @throws IllegalArgumentException if {@code emptyMatrix} is not empty.
     * @throws NullPointerException if either matrix is {@code null}.
     */
    public <R, C, E> void transpose(Matrix<R, C, E> matrix, 
                                    MatrixBuilder<? super C, ? super R, E> emptyMatrix) {

        // Ensuring the provided matris is empty
        if (!emptyMatrix.isEmpty()) {
            throw new IllegalArgumentException("matrix must be empty");
        }

        // Iterating thru each column index
        for (C columnIndex : matrix.getColumnIndices()) {

            // Iterating thru each row index
            for (R rowIndex : matrix.getRowIndices()) {

                // Getting element
                E value = matrix.getElement(rowIndex, columnIndex);

                // Transposing
                emptyMatrix.setElement(columnIndex, rowIndex, value);
            }
        }

        
    }
    
    
    
}
