/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import sbrt.kernel.utilities.*;

/**
 * This class is used to wrap mutable matrices as an immutable
 * matrix. This type of wrapping eliminates the possibility of
 * casting a mutable matrix back to its mutable type. This ensures that
 * a mutable matrix can be viewed by others in a safe way, without
 * fear of modification. 
 * 
 * <p>Instances of this class are immutable.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <R> the row type.
 * @param <C> the column type.
 * @param <E> the element type.
 */
public class MatrixWrap<R, C, E> implements Matrix<R, C, E> {

	// The wrapped matrix
    private final Matrix<R, C, E> m;

    /**
     * Constructs a new wrapper around the provided matrix.
     * 
     * @param m the matrix to be wrapped by this wrapper object.
     * 
     * @throws NullPointerException if the provided matrix is
     * {@code null}.
     */
    public MatrixWrap(Matrix<R, C, E> m) {
        Util.nullCheck(m);
        this.m = m;
    }

    /**
     * Returns the element of this matrix at the provided
     * row and column indices.
     * 
     * @param rowIndex the row index.
     * @param columnIndex the column index.
     * 
     * @return the element of this matrix at the provided
     * row and column indices.
     */
	public E getElement(Object rowIndex, Object columnIndex) {
		return m.getElement(rowIndex, columnIndex);
	}
    
    /**
     * Returns the set of column indices contained in this
     * matrix.
     * 
     * @return the set of column indices contained in this
     * matrix.
     */
    public UnmodifiableSet<C> getColumnIndices() {
        return new UnmodifiableSet<C>(m.getColumnIndices());
    }
    
    /**
     * Returns the set of row indices contained in this
     * matrix.
     * 
     * @return the set of row indices contained in this
     * matrix.
     */
    public UnmodifiableSet<R> getRowIndices() {
        return new UnmodifiableSet<R>(m.getRowIndices());
    }

    /**
     * Indicates if the provided object is a row index in
     * this matrix.
     * 
     * @param obj the potential row index.
     * 
     * @return {@code true} if the provided object is a row
     * index in this matrix; {@code false} otherwise.
     */
    public boolean isRowIndex(Object obj) {
        return m.isRowIndex(obj);
    }

    /**
     * Indicates if the provided object is a column index in
     * this matrix.
     * 
     * @param obj the potential column index.
     * 
     * @return {@code true} if the provided object is a column
     * index in this matrix; {@code false} otherwise.
     */
    public boolean isColumnIndex(Object obj) {
        return m.isColumnIndex(obj);
    }

    /**
     * Returns the row corresponding to the provided row index.
     * A row is itself a 1 x n matrix, where n is the number of
     * columns in this matrix. 
     * 
     * @param rowIndex the index of the row to be returned.
     * 
     * @return the row corresponding to the provided row index.
     */
    public Matrix<R, C, E> getRow(Object rowIndex) {
        return m.getRow(rowIndex);
    }

    /**
     * Returns the column corresponding to the provided column index.
     * A column is itself an m x 1 matrix, where m is the number of
     * rows in this matrix. 
     * 
     * @param columnIndex the index of the column to be returned.
     * 
     * @return the column corresponding to the provided column index.
     */
    public Matrix<R, C, E> getColumn(Object columnIndex) {
        return m.getColumn(columnIndex);
    }

    /**
     * Returns the number of rows in this matrix.
     * @return the number of rows in this matrix.
     */
    public int rows() {
        return m.rows();
    }

    /**
     * Returns the number of columns in this matrix.
     * @return the number of columns in this matrix.
     */
    public int columns() {
        return m.columns();
    }

    /**
     * Indicates if this matrix is empty.
     * 
     * @return {@code true} if at least one row and column exists
     * in this matrix; {@code false} otherwise.
     */
    public boolean isEmpty() {
        return m.isEmpty();
    }

    /**
     * Returns a string representation of this matrix.
     * @return a string representation of this matrix.
     */
    public String toString() {
        return m.toString();
    }

    /**
     * Indicates if the provided object is equal to this matrix.
     * Two matrices are considered equal if they share the same
     * sets of column and row indices and the elements in each
     * matrix for each column-row index pair are the same.
     * 
     * @param obj the object with which to compare.
     * 
     * @return {@code true} if this object equals the provided
     * object; {@code false} otherwise.
     */
    public boolean equals(Object obj) {
        return m.equals(obj);
    }

    /**
     * Returns a content-based hash code for this matrix.
     * @return a content-based hash code for this matrix.
     */
    public int hashCode() {
        return m.hashCode();
    }

}
