/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import java.util.*;

import sbrt.kernel.utilities.*;

/**
 * This class is a {@code java.util.TreeMap}  based implementation of {@linkplain SmdMatrix}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <R> the row type.
 * @param <C> the column type.
 * 
 * @see TreeMatrix
 * @see HashMatrix
 */
public class RowOrTreeMatrix<R extends Comparable<? super R>, C extends Comparable<? super C>>
	extends AbstractSmdMatrix<R, C> {

	// Row indices -> { column indices -> values }
    private final Map<R, Map<C, Double>> rowToColumn;
    
    // All column indices of the matrix
    private final Set<C> columnIndices;
    
    /**
     * Constructs an emtpy {@code RowOrTreeMatrix}.
     */
    public RowOrTreeMatrix() {
        rowToColumn = new TreeMap<R, Map<C, Double>>();
        columnIndices = new TreeSet<C>();
    }

    /**
     * Sets the element at the provided row and column indices
     * to the provided double precision value.
     * 
     * @param <X> a type extending the primary row type.
     * @param <Y> a type extending the primary column type.
     * @param <Z> Double.
     * 
     * @param rowIndex the row index.
     * @param columnIndex the column index.
     * @param value the double precision element value.
     * 
     * @throws NullPointerException if any argument is {@code null}.
     */
    public <X extends R, Y extends C, Z extends Double> void 
    	setElement(X rowIndex, Y columnIndex, Z value) {
        
    	// Checking inputs
    	Util.nullCheck(rowIndex);
        Util.nullCheck(columnIndex);
        Util.nullCheck(value);

        Map<C, Double> columnValues = rowToColumn.get(rowIndex);
        
        if (columnValues == null) {
            columnValues = new TreeMap<C, Double>();
            rowToColumn.put(rowIndex, columnValues);
        }

        // Removing values equal to 0
        if (Math.abs(value) < Double.MIN_VALUE) {
            columnValues.remove(columnIndex);
        }

        else {
            columnValues.put(columnIndex, value);
        }
        
        columnIndices.add(columnIndex);
    }

    /**
     * Indicates if the provided object is a row index in
     * this matrix.
     * 
     * @param obj the potential row index.
     * 
     * @return {@code true} if the provided object is a row
     * index in this matrix; {@code false} otherwise.
     */
    public boolean isRowIndex(Object obj) {
        return rowToColumn.containsKey(obj);
    }

    /**
     * Indicates if the provided object is a column index in
     * this matrix.
     * 
     * @param obj the potential column index.
     * 
     * @return {@code true} if the provided object is a column
     * index in this matrix; {@code false} otherwise.
     */
    public boolean isColumnIndex(Object obj) {
    	return columnIndices.contains(obj);
    }

    /**
     * Returns the row corresponding to the provided row index.
     * A row is itself a 1 x n matrix, where n is the number of
     * columns in this matrix. 
     * 
     * @param rowIndex the index of the row to be returned.
     * 
     * @return the row corresponding to the provided row index.
     */
    @SuppressWarnings("unchecked")
    public RowOrTreeMatrix<R, C> getRow(Object rowIndex) {

        // Ensuring the provided row index exists
        if (!isRowIndex(rowIndex)) {
            String s = "The row index '" + rowIndex + "' does not exist.";
            throw new NoSuchElementException(s);
        }

        // Creating storage
        RowOrTreeMatrix<R, C> m = new RowOrTreeMatrix<R, C>();

		// Creating a single zero object to save space
		Double zero = Double.valueOf(0d);
		
		// Setting the values initially to zero for all column indices.
        // This ensures that each column index appears the returned row matrix.
		for (C columnIndex : columnIndices) {
			m.setElement( (R) rowIndex, columnIndex, zero);
		}

        // Setting the actual values
		for (Map.Entry<C, Double> e : rowToColumn.get(rowIndex).entrySet()) {
			m.setElement( (R) rowIndex, e.getKey(), e.getValue());
		}
        
        return m;
    }

    /**
     * Returns the column corresponding to the provided column index.
     * A column is itself an m x 1 matrix, where m is the number of
     * rows in this matrix. 
     * 
     * @param columnIndex the index of the column to be returned.
     * 
     * @return the column corresponding to the provided column index.
     */
    @SuppressWarnings("unchecked")
    public RowOrTreeMatrix<R, C> getColumn(Object columnIndex) {

        // Ensuring the provided column index exists
        if (!isColumnIndex(columnIndex)) {
            String s = "The column index '" + columnIndex + "' does not exist.";
            throw new NoSuchElementException(s);
        }

        // Creating storage
        RowOrTreeMatrix<R, C> m = new RowOrTreeMatrix<R, C>();

		// Creating a single zero object to save space
		Double zero = Double.valueOf(0d);
		
        // Iterating thru each row index
		for (Map.Entry<R, Map<C, Double>> e : rowToColumn.entrySet()) {
			
            // Getting the next row index
			R rowIndex = e.getKey();
            
            // Getting the next value at this column and row (if it's not zero)
			Double value = e.getValue().get(columnIndex);
			
            // If no value was found, it is zero
			if (value == null) {
				m.setElement(rowIndex, (C) columnIndex, zero);
			}
            
            // If a non-zero value exists
            else {
				m.setElement(rowIndex, (C) columnIndex, value);
			}
			
		}

        return m;
    }

    /**
     * Removes the row corresponding to the provided 
     * row index.
     * 
     * @param rowIndex the index of the row to be removed.
     */
    public void removeRow(Object rowIndex) {
        rowToColumn.remove(rowIndex);
    }

    /**
     * Removes the column corresponding to the provided 
     * column index.
     * 
     * @param columnIndex the index of the column to be removed.
     */
    public void removeColumn(Object columnIndex) {
    	
    	for (Map<C, Double> columnValues : rowToColumn.values()) {
    		columnValues.remove(columnIndex);
    	}
    	
    	columnIndices.remove(columnIndex);
    }

    /**
     * Returns the element of this matrix at the provided
     * row and column indices.
     * 
     * @param rowIndex the row index.
     * @param columnIndex the column index.
     * 
     * @return the element of this matrix at the provided
     * row and column indices.
     */
    public Double getElement(Object rowIndex, Object columnIndex) {

        Map<C, Double> columnValues = rowToColumn.get(rowIndex);
        
    	// Ensuring provided row index exists
        if (columnValues == null) {
            String s = "The row index '" + rowIndex + "' does not exist.";
            throw new IllegalArgumentException(s);
        }

        // Ensuring provided column index exists
        if (!isColumnIndex(columnIndex)) {
            String s = "The column index '" + columnIndex + "' does not exist.";
            throw new IllegalArgumentException(s);
        }

        // Getting the value (if it does not equal zero)
        Double value = columnValues.get(columnIndex);

        return (value == null ? Double.valueOf(0.0d) : value);
    }
    
    /**
     * Returns the set of column indices contained in this
     * matrix.
     * 
     * @return the set of column indices contained in this
     * matrix.
     */
    public TreeSet<C> getColumnIndices() {
    	return new TreeSet<C>(columnIndices);
    }
    
    /**
     * Returns the set of row indices contained in this
     * matrix.
     * 
     * @return the set of row indices contained in this
     * matrix.
     */
    public TreeSet<R> getRowIndices() {
    	return new TreeSet<R>(rowToColumn.keySet());
    }

	/**
	 * Returns the set of row indices whose elements are nonzero 
	 * in the column corresponding to the provided index.
	 * 
	 * @param columnIndex the column index.
	 * 
	 * @return the set of row indices whose elements are nonzero 
	 * in the column corresponding to the provided index.
	 */
    public TreeSet<R> getRowIndNzEl(Object columnIndex) {

    	// Checking provided object
    	if (!columnIndices.contains(columnIndex)) {
            String s = "The column index '" + columnIndex + "' does not exist.";
            throw new NoSuchElementException(s);
    	}
        
        // Creating storage for the row indices
        TreeSet<R> rowIndices = new TreeSet<R>();
    	
    	// Iterating thru each row -> (column -> value) pair
    	for (Map.Entry<R, Map<C, Double>> e : rowToColumn.entrySet()) {
    		
            // If the column index exist in this map
    		if (e.getValue().containsKey(columnIndex)) {
    			rowIndices.add(e.getKey());
    		}
    		
    	}

        return rowIndices;
    }

	/**
	 * Returns the set of column indices whose elements are nonzero 
	 * in the row corresponding to the provided index.
	 * 
	 * @param rowIndex the row index.
	 * 
	 * @return the set of column indices whose elements are nonzero 
	 * in the row corresponding to the provided index.
	 */
    public TreeSet<C> getColIndNzEl(Object rowIndex) {

    	// Getting columns -> values
        Map<C, Double> columnIndices = rowToColumn.get(rowIndex);

        // If none exist for the provided row
        if (columnIndices == null) {
            String s = "The row index '" + rowIndex + "' does not exist.";
            throw new NoSuchElementException(s);
        }

        // Creating storage
        TreeSet<C> s = new TreeSet<C>();
        
        // Copying
        s.addAll(columnIndices.keySet());

        return s;
    }

    /**
     * Returns the number of rows in this matrix.
     * @return the number of rows in this matrix.
     */
    public int rows() {
        return rowToColumn.size();
    }

    /**
     * Returns the number of columns in this matrix.
     * @return the number of columns in this matrix.
     */
    public int columns() {
        return columnIndices.size();
    }

    /**
     * Returns a content-based hash code for this matrix.
     * @return a content-based hash code for this matrix.
     */
    public int hashCode() {
        return rowToColumn.hashCode();
    }

}
