/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra;

import java.util.*;

/**
 * This interface is used to represent sparse matrices whose
 * elements are strictly double precision numbers.
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <R> the row type.
 * @param <C> the column type.
 */
public interface SparseDoubleMatrix<R, C> extends DoubleMatrix<R, C> {

	/**
	 * Returns the set of row indices whose elements are nonzero 
	 * in the column corresponding to the provided index.
	 * 
	 * @param columnIndex the column index.
	 * 
	 * @return the set of row indices whose elements are nonzero 
	 * in the column corresponding to the provided index.
	 */
    public Set<R> getRowIndNzEl(Object columnIndex);

	/**
	 * Returns the set of column indices whose elements are nonzero 
	 * in the row corresponding to the provided index.
	 * 
	 * @param rowIndex the row index.
	 * 
	 * @return the set of column indices whose elements are nonzero 
	 * in the row corresponding to the provided index.
	 */
    public Set<C> getColIndNzEl(Object rowIndex);

    /**
     * Returns the row corresponding to the provided row index.
     * A row is itself a 1 x n matrix, where n is the number of
     * columns in this matrix. 
     * 
     * @param rowIndex the index of the row to be returned.
     * 
     * @return the row corresponding to the provided row index.
     */
    public SparseDoubleMatrix<R, C> getRow(Object rowIndex);

    /**
     * Returns the column corresponding to the provided column index.
     * A column is itself an m x 1 matrix, where m is the number of
     * rows in this matrix. 
     * 
     * @param columnIndex the index of the column to be returned.
     * 
     * @return the column corresponding to the provided column index.
     */
    public SparseDoubleMatrix<R, C> getColumn(Object columnIndex);

}
