/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra.proc;

import java.util.Map;

import sbrt.kernel.KernelProcess;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.algebra.LinearSystemSolution;
import sbrt.kernel.math.algebra.LinearSystemSolver;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to solve systems of linear equations.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <V> the variable type.
 */
public class LinearSolverProcess<V> implements KernelProcess<LinearSystemSolution<V>> {

    private LinearSystemSolver<V> solver;

    private Map<LinearComb<V>, Double> equations;
    
    /**
     * Sets the system of linear equations to be solver.
     * 
     * @param equations linear combinations mapped to a double. Each key-value
     * pair represents the left and right hand sides, respectively, of an equation.
     */
    public void setEquations(Map<LinearComb<V>, Double> equations) {
        Util.nullCheck(equations);
        this.equations = equations;
    }

    /**
     * Sets the linear system solver.
     * @param solver the linear system solver.
     */
    public void setSolver(LinearSystemSolver<V> solver) {
        Util.nullCheck(solver);
        this.solver = solver;
    }

    private void checkState() {

        if (solver == null) {
            String s = "The linear system solver was never provided.";
            throw new IllegalStateException(s);
        }

        if (equations == null) {
            String s = "The linear equations were never provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Solves the system of linear equations and adds the solution 
     * to the provided results manager.
     * 
     * @param resultsManager the results manager to which the solution is added.
     */
    public void run(ResultsManager<LinearSystemSolution<V>> resultsManager) {

        checkState();

        LinearSystemSolution<V> solution = solver.solve(equations);
        
        resultsManager.addData(solution);
    }

}
