/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.algebra.proc;

import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import sbrt.kernel.io.ResultsManager;

/**
 * This class is used to intersect vectors whose variables (that is, map keys) are
 * not identical.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <K> the key, or variable, type.
 * @param <V> the value type.
 * 
 */
public class VectorIntersection<K, V> implements
        sbrt.kernel.KernelProcess<Map<K, V>> {

    private Iterator<Map<K, V>> vectorSource;

    /**
     * Constructs a new vector intersector.
     */
    public VectorIntersection() { 
        // does nothing
    }

    /**
     * Sets the vector source.
     * 
     * @param vectorSource a source of variables mapped to their respective values.
     */
    public void setVectorSource(Iterator<Map<K, V>> vectorSource) {

        // Checking input
        if (vectorSource == null) {
            throw new NullPointerException("'vectors' cannot be null.");
        }

        // Saving
        this.vectorSource = vectorSource;
    }

    private void checkState() {

        if (vectorSource == null) {
            String s = "No vector source was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * 
     * and adds the results to the provided results manager.
     * 
     * @param resultsManager the results manager to which the computed results
     * are to be added.
     */
    public void run(ResultsManager<Map<K, V>> resultsManager) {
		
		// Ensuring all required inputs were provided
		checkState();

        List<Map<K, V>> allVectors = getIntersections();
        
        // Adding all vectors to the results manager one-by-one
		for (Map<K, V> vector : allVectors) {
            resultsManager.addData(vector);
        }
	}
    
    
    private List<Map<K, V>> getIntersections() {
        

        // Creating storage for all vectors
        List<Map<K, V>> allVectors = new LinkedList<Map<K, V>>();
        
        // Creating storage for the variables common to all vectors
        Set<K> sharedVariables = new LinkedHashSet<K>();
        
        // Iterating thru each provided vector
        while (vectorSource.hasNext()) {
            
            Map<K, V> vector = vectorSource.next();
            
            // Copying the map in case it is unmodifiable
            int size = vector.size() + 1;
            
            Map<K, V> copy = new LinkedHashMap<K, V>(size, 1f);

            boolean justifiableUpdate = false;
            
            // Iterating thru each of the relevant variables
            for (Iterator<K> i = sharedVariables.iterator(); i.hasNext(); ) {
                
                // Getting the next variable
                K var = i.next();
                
                // Getting the value for this variable in this vector
                V value = vector.get(var);
                
                // If this vector does not contain a value for this variable,
                // remove the variable from the set of relevant variables
                if (value == null) {
                    
                    i.remove();
                    
                    // Bailing out if the intersection is necessarily empty
                    if (sharedVariables.isEmpty()) {
                        return Collections.emptyList();
                    }

                    justifiableUpdate = true;
                }
                
                // Otherwise, copy this value into the new map
                else {
                    copy.put(var, value);
                }
            }
            
            // Removing irrelevant variables to (hopefully)
            // reduce the memory footprint
            if (justifiableUpdate) {
                
                for (Map<K, V> m : allVectors) {
                    m.entrySet().retainAll(sharedVariables);
                }
                
            }
            
            // Adding the copied vector to the rest
            allVectors.add(copy);
        }
        
        return allVectors;
    }

}
