/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.comb.comp;

import java.util.*;

/**
 * This class is used to compute the single-element unions of collections of 
 * sets. A single-element union of sets will contain (at least) one element
 * from each set. Take, for example, the following collection of sets:<br>
 * 
 * <table align=center cellspacing=20><tr>
 * <td>{a, b, c}
 * <td>{d}
 * <td>{e, f}.</tr></table><br>
 * The sets of all possible single-element unions from these sets are:<br>
 * 
 * <table align=center cellspacing=20><tr>
 * <td>{a, d, e}
 * <td>{a, d, f}</tr><tr>
 * <td>{b, d, e}
 * <td>{b, d, f}</tr><tr>
 * <td>{c, d, e}
 * <td>{c, d, f}.</tr></table><br>
 * 
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 */
public class SingleElementUnionsV2 {

	/**
	 * Returns a set of singleton sets created from the provided set. The
	 * singleton sets each contain a single element from the provided set.
	 * This can be thought of as the single-element union of the provided
	 * set and the empty set.
	 *  
	 * @param <E> the element type.
	 * @param set a single set of elements.
	 * 
	 * @return a set of singleton sets created from the provided set.
	 */
    public <E> Set<Set<E>> getSingletonSets(Set<E> set) {
        
    	// Ensuring set is not empty
    	emptyCheck(set);

        // Creating storage for element sets
        Set<Set<E>> elements = 
        	new LinkedHashSet<Set<E>>(set.size() + 1, 1.0F);

        // Iterating thru each element in the provided set
        for (E element : set) {

            // Creating a singleton set containing the next element
            Set<E> singleton = Collections.singleton(element);

            // Adding to the rest
            elements.add(singleton);
        }

        // Returning a collection of singleton sets containing the elements
        // of the provided set
        return elements;
    }

    private void emptyCheck(Set<?> set) {
        if (set.isEmpty()) {
            String s = "An empty set was provided.";
            throw new IllegalArgumentException(s);
        }
    }

    /**
     * Returns the set of single-element unions of the sets in the provided 
     * collection.
     * 
     * @param <E> the element type.
     * @param sets the collection of sets.
     * 
     * @return the set of single-element unions in the provided collection of
     * sets.
     */
    public <E extends Comparable<? super E>> Set<Set<E>> getUnions(Collection<Set<E>> sets) {
        return useStubMethod(sets);
    }

    private <E> Set<Set<E>> useStubMethod(Collection<Set<E>> sets) {

        // Getting number of sets in the provided colletion
        int n = sets.size();

        // If the collection is empty, return an empty set
        if (n == 0) {
            return Collections.emptySet();
        }

        // If only one set is present, return a collection of singleton sets
        if (n == 1) {
            return getSingletonSets(sets.iterator().next());
        }

        // Getting an iterator over the sets
        Iterator<Set<E>> i = sets.iterator();

        // Filling the initial stubs
        Set<E> s = i.next();
        Set<Set<E>> stubs = getSingletonSets(s);

        // Creating a variable for temporary storage
        Set<Set<E>> newStubs = null;

        // Iterating thru the remaining sets
        while (i.hasNext()) {

            // Getting the next set
            s = i.next();
            emptyCheck(s);
            
            // Creating storage for new stubs
            newStubs = new HashSet<Set<E>>();

            // Iterating thru each element in this set
            for (E element : s) {

                // Adding this element to the existing stubs
            	for (Set<E> stub : stubs) {
            		
                    // Creating a new stub by first copying the original stub
                    Set<E> newStub = new TreeSet<E>();
                    newStub.addAll(stub);

                    // Adding the element to the new stub
                    newStub.add(element);

                    // Adding this new stub to the rest
                    newStubs.add(newStub);
                }

            }

            // Replacing old stubs with new stubs
            stubs = newStubs;
        }

        return stubs;
    }


}
