/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.comb.comp;

import java.util.*;

/**
 * This class is used to compute the single-element unions of collections of
 * sets. A single-element union of sets will contain one, and only one, element
 * from each set. Take, for example, the following collection of sets:<br>
 * 
 * <table align=center cellspacing=20>
 * <tr>
 * <td>{a, b, c}
 * <td>{d}
 * <td>{e, f}.</tr>
 * </table><br>
 * The sets of all possible single-element unions from these sets are:<br>
 * 
 * <table align=center cellspacing=20>
 * <tr>
 * <td>{a, d, e}
 * <td>{a, d, f}</tr>
 * <tr>
 * <td>{b, d, e}
 * <td>{b, d, f}</tr>
 * <tr>
 * <td>{c, d, e}
 * <td>{c, d, f}.</tr>
 * </table><br>
 * 
 * Here's another, slightly more complicated example. Take the following
 * collection of sets:
 * 
 * <table align=center cellspacing=20>
 * <tr>
 * <td>{a, b, c}
 * <td>{c, d}
 * <td>{e, f}.</tr>
 * </table><br>
 * 
 * Note that the first and second sets contain the same element <i>c</i>. The
 * sets of all possible single-element unions from these sets are:<br>
 * 
 * <table align=center cellspacing=20>
 * <tr>
 * <td>{a, d, e}
 * <td>{a, d, f}</tr>
 * <tr>
 * <td>{b, d, e}
 * <td>{b, d, f}</tr>
 * <tr>
 * <td>{c, e}
 * <td>{c, f}.</tr>
 * </table><br>
 * 
 * Note that the set {c, d, f} (and other similar sets) is not present because
 * it contains more than one element from the set {c, d}.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public class StrictSingleElementUnions {

	private final SingleElementUnionsV1 seu;

	/**
	 * Constructs a new strict single-element union computer.
	 */
	public StrictSingleElementUnions() {
		seu = new SingleElementUnionsV1();
	}

	/**
	 * Returns a set of singleton sets created from the provided set. The
	 * singleton sets each contain a single element from the provided set. This
	 * can be thought of as the single-element union of the provided set and the
	 * empty set.
	 * 
	 * @param <E>
	 *            the element type.
	 * @param set
	 *            a single set of elements.
	 * 
	 * @return a set of singleton sets created from the provided set.
	 */
	public <E> Set<Set<E>> getSingletonSets(Set<E> set) {
		return seu.getSingletonSets(set);
	}

	/**
	 * Returns the set of strict single-element unions of the sets in the provided
	 * collection.
	 * 
	 * @param <E>
	 *            the element type.
	 * @param sets
	 *            the collection of sets.
	 * 
	 * @return the set of strict single-element unions in the provided collection of
	 *         sets.
	 */
	public <E> Set<Set<E>> getUnions(Collection<Set<E>> sets) {

		// Getting single-element unions
		Set<Set<E>> unions = seu.getUnions(sets);

		// Creating storage
		Set<Set<E>> strictUnions = new LinkedHashSet<Set<E>>();

		// Iterating thru each single-element union
		for (Set<E> u : unions) {

			// If the union contains only one element from each set
			if (containsOnlyOneElementFromEachSet(u, sets)) {
				strictUnions.add(u);
			}
		}

		return strictUnions;
	}

	private <E> boolean containsOnlyOneElementFromEachSet(Set<E> union,
			Collection<Set<E>> sets) {

		// Iterating thru each provided set
		for (Set<E> s : sets) {

			// Copying the union
			Set<E> copy = new LinkedHashSet<E>(union.size() + 1, 1F);
			copy.addAll(union);

			// Retaining all element from the provided set
			copy.retainAll(s);

			// If more than one element remains
			if (copy.size() > 1) {
				return false;
			}

		}

		return true;
	}

}
