/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.comb.proc;

import java.util.*;

import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.comb.comp.SingleElementUnionsV1;

/**
 * This class is used to wrap the single element union computer as a process.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 * @param <E> the element type.
 */
public class SingleElementUnionProc<E>
	implements sbrt.kernel.KernelProcess<Set<Set<E>>> {

	private final SingleElementUnionsV1 seu;
	private Collection<Set<E>> sets;
	
	/**
	 * Constructs a new single-element union process.
	 */
	public SingleElementUnionProc() {
		
		// Creating the single element union computer
		this.seu = new SingleElementUnionsV1();
	}
	
	/**
	 * Sets the collection of sets to that provided.
	 * @param sets a collection of sets of elements.
	 */
	public void setSets(Collection<Set<E>> sets) {
		
		// Checking input
		if (sets == null) {
			throw new NullPointerException("'sets' cannot be null.");
		}
		
		// Saving
		this.sets = sets;
	}

	
	private void checkState() {
		
		if (sets == null) {
			String s = "A collection of sets was not provided.";
			throw new IllegalStateException(s);
		}
		
	}
	
	/**
	 * Computes the single-element unions of the provided collection of sets,
	 * and adds the results to the provided results manager.
	 * 
	 * @param resultsManager the results manager to which the computed results
	 * are to be added.
	 */
	public void run(ResultsManager<Set<Set<E>>> resultsManager) {
		
		// Ensuring all required inputs were provided
		checkState();

		// Computing the single-element unions
		Set<Set<E>> singleElementUnions = seu.getUnions(sets);

		// Saving results
		resultsManager.addData(singleElementUnions);

	}


}
