/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.geom;

import java.util.*;

/**
 * This interface is used to represent convex polytopes that are represented
 * as the intersection of a finite set of half-spaces.
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <V> the variable type.
 */
public interface ConvexPolytope<V> {

	/**
	 * Returns the variables contained in the hyperplanes defining the bounding
	 * half-spaces.
	 *  
	 * @return the variables contained in the hyperplanes defining the bounding
	 * half-spaces.
	 */
    public Set<V> getVariables();

    /**
     * Indicates if the provided object is a variable in the bounding 
     * hyperplanes of this convex polytope.
     * 
     * @param variable the potential variable.
     * 
     * @return {@code true} if the provided object is a variable in the
     * bounding hyperplanes of this convex polytope; {@code false} otherwise.
     */
    public boolean isVariable(Object variable);

    /**
     * Returns the set of bounding hyperplanes that define this convex polytope.
     * @return the set of bounding hyperplanes that define this convex polytope.
     */
    public Set<ParallelHyperplanes<V>> getBoundingPlanes();

    /**
     * Indicates if the provided point lies within this convex polytope.
     * 
     * @param point variables mapped to numerical values. Every key of this
     * map must be a variable of this polytope and every variable must be
     * present.
     * 
     * @param tolerance the amount by which a point may lay outside this
     * polytope, but still considered within it. This should be a small 
     * positive numer, like 1E-9 for example.
     * 
     * @return {@code true} if the provided point lies within this polytope;
     * {@code false} otherwise.
     */
    public boolean isWithin(Map<? extends V, Double> point, double tolerance);

    /**
     * Returns the dimensions of the space in which this polytope lies.
     * @return the dimensions of the space in which this polytope lies.
     */
    public int dimensions();

}
