/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.geom;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.math.algebra.HashLinearSystemSolution;
import sbrt.kernel.math.algebra.LinearSystemSolution;
import sbrt.kernel.utilities.*;

/**
 * This class is a concrete implemenation of {@linkplain ConvexPolytope}.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <V> the variable type.
 */
public class HashConvexPolytope<V> implements ConvexPolytope<V> {

    /**
     * Constructs a new convex poltyope from the provided linear system solution
     * and variable bounds.
     * 
     * @param <V> the variable type.
     * 
     * @param systemSolution a solution to a system of linear equations.
     * @param variableBounds the bounds on each variable in that solution.
     * @param tolerance the tolerance used to simplify the provided linear
     * system solution. See {@linkplain LinearSystemSolution#simplify(Map, double)}.
     * 
     * @return a new convex poltyope from the provided linear system solution
     * and variable bounds.
     */
    public static <V> HashConvexPolytope<V> 
    make(LinearSystemSolution<V> systemSolution,
    	 Map<? extends V, Interval> variableBounds,
    	 double tolerance) {

        // Simplifying the linear system solution
        LinearSystemSolution<V> simplified =
            HashLinearSystemSolution.simplify(systemSolution,
            								  variableBounds, 
            								  tolerance);

        // Creating the set of bounding hyperplanes
        Set<ParallelHyperplanes<V>> planes =
                getBoundingPlanes(simplified, variableBounds);

        // Returning the convex polytope
        return new HashConvexPolytope<V>(planes);
    }

    private static <V> Set<ParallelHyperplanes<V>> 
    getBoundingPlanes(LinearSystemSolution<V> solution,
    				  Map<? extends V, Interval> variableBounds) {

        // Getting the solution as a map
        Map<V, LinearComb<V>> solutions = solution.toMap();

        // Determining an appropriate size
        int size = solutions.size() + 1;

        // Creatin storage
        Set<ParallelHyperplanes<V>> planes =
                new LinkedHashSet<ParallelHyperplanes<V>>(size, 1.0F);

        // Iterating thru each variable-solution pair
        for (Map.Entry<V, LinearComb<V>> e : solutions.entrySet()) {

            // Getting the pair
            V variable = e.getKey();
            LinearComb<V> sol = e.getValue();

            // If the solution contains free variables
            if (sol.hasVariables()) {

                // Getting the bounds on this variable
                Interval bounds = variableBounds.get(variable);

                // Creating the parallel hyperplanes
                ParallelHyperplanes<V> p = getPlane(sol, bounds);

                // Saving
                planes.add(p);
            }

        }

        return planes;
    }

    private static <V> ParallelHyperplanes<V> 
    getPlane(LinearComb<V> solution, Interval bounds) {

        // Getting the constant
        double constant = solution.getConstant();

        // If it's zero
        if (constant == 0d) {
            return new ParallelHyperplanes<V>(solution, bounds);
        }

        // Getting the bounds
        double lower = bounds.getLower();
        double upper = bounds.getUpper();

        // Computing new bounds
        bounds = new Interval(lower - constant, upper - constant);

        // Creating a new hyperplane
        solution = new HashLinearComb<V>(solution.getTermList(), 
        								 0d, 
        								 solution.getZeroCutoff());

        return new ParallelHyperplanes<V>(solution, bounds);
    }

    /**
     * The bounding parallel hyperplanes forming this convex polytope.
     */
    private final UnmodifiableSet<ParallelHyperplanes<V>> planes;

    /**
     * The variables used to form this convex polytope.
     */
    private final UnmodifiableSet<V> variables;

    /**
     * Constructs a new convex poltyope from the provided linear system solution
     * and variable bounds.
     * 
     * @param boundingPlanes the bounding hyperplanes.
     */
    public HashConvexPolytope(Set<ParallelHyperplanes<V>> boundingPlanes) {

        // Determining upper bound on size
        int size = boundingPlanes.size() + 1;

        // Creating storage for variables
        Set<V> tempVariables = new LinkedHashSet<V>(size, 1F);

        // Creating storage for planes
        Set<ParallelHyperplanes<V>> tempPlanes =
                new LinkedHashSet<ParallelHyperplanes<V>>(size, 1F);

        // Checking each pair of hyperplanes
        for (ParallelHyperplanes<V> planes : boundingPlanes) {

            // Getting the associated interval
            Interval interval = planes.getBounds();

            // Checking the interval
            if (rangeIsZero(interval)) {

                String s =
                        "The interval in '" + planes
                                + "' is not allowed. "
                                + "The polytope must be full-dimensional.";
                throw new IllegalArgumentException(s);

            }

            // Adding planes to the rest
            tempPlanes.add(planes);

            // Adding variables to the rest
            tempVariables.addAll(planes.getPlane().getVariables());

        }

        // Saving as unmodifiable sets
        variables = new UnmodifiableSet<V>(tempVariables);
        planes = new UnmodifiableSet<ParallelHyperplanes<V>>(tempPlanes);
    }

    private boolean rangeIsZero(Interval interval) {
        return interval.getUpper() - interval.getLower() < Double.MIN_VALUE;
    }

    /**
     * Returns the variables contained in the hyperplanes defining the bounding
     * half-spaces.
     * 
     * @return the variables contained in the hyperplanes defining the bounding
     * half-spaces.
     */
    public UnmodifiableSet<V> getVariables() {
        return variables;
    }

    /**
     * Indicates if the provided object is a variable in the bounding
     * hyperplanes of this convex polytope.
     * 
     * @param variable the potential variable.
     * 
     * @return {@code true} if the provided object is a variable in the bounding
     * hyperplanes of this convex polytope; {@code false} otherwise.
     */
    public boolean isVariable(Object variable) {
        return variables.contains(variable);
    }

    /**
     * Returns the set of bounding hyperplanes that define this convex polytope.
     * 
     * @return the set of bounding hyperplanes that define this convex polytope.
     */
    public UnmodifiableSet<ParallelHyperplanes<V>> getBoundingPlanes() {
        return planes;
    }

    /**
     * Returns the dimensions of the space in which this polytope lies.
     * 
     * @return the dimensions of the space in which this polytope lies.
     */
    public int dimensions() {
        return variables.size();
    }

    /**
     * Indicates if the provided point lies within this convex polytope.
     * 
     * @param point variables mapped to numerical values. Every key of this map
     * must be a variable of this polytope and every variable must be present.
     * 
     * @param tolerance the amount by which a point may lay outside this
     * polytope, but still be considered within it. This should be a small
     * positive numer, like 1E-9 for example.
     * 
     * @return {@code true} if the provided point lies within this polytope;
     * {@code false} otherwise.
     */
    public boolean isWithin(Map<? extends V, Double> point, double tolerance) {

        // Iterating thru each pair of bounding hyperplanes
        for (ParallelHyperplanes<V> p : planes) {

            // Plugging in the provided point
            LinearComb<V> plane = p.getPlane().plugInNumbers(point);

            // Checking for remaining variables
            if (plane.hasVariables()) {

                String s = "The provided point does not contain values for all " + 
                			"the variables in this convex polytope.";
                throw new IllegalArgumentException(s);

            }

            // Getting the bounds on these planes
            Interval bounds = p.getBounds();

            // Is the computed value between the bounds?
            if (!bounds.isWithin(plane.getConstant(), tolerance)) {
                return false;
            }

        }

        return true;
    }
}
