/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.geom;

import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;

/**
 * This class is used to represent a pair of bounding parallel hyperplanes.
 * They have the general form:  lower bound &le; hyperplane &le; upper bound.
 * These are also called half-spaces.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <V> the variable type. 
 *
 */
public class ParallelHyperplanes<V> {

    private final LinearComb<V> plane;
    private final Interval bounds;

    /**
     * Constructs a new pair of bounding parallel hyperplanes.
     * 
     * @param hyperplane the hyperplane.
     * @param bounds the bounding values used to define the inequality.
     */
    public ParallelHyperplanes(LinearComb<V> hyperplane, Interval bounds) {

    	// Checking input
    	Util.nullCheck(hyperplane);
        Util.nullCheck(bounds);
        
        // Saving
        this.plane = hyperplane;
        this.bounds = bounds;
    }

    /**
     * Returns the hyperplane.
     * @return the hyperplane.
     */
    public LinearComb<V> getPlane() {
        return plane;
    }

    /**
     * Returns the bounding values used to define the inequality.
     * @return the bounding values used to define the inequality.
     */
    public Interval getBounds() {
        return bounds;
    }

    /**
     * Returns a content-based hash code.
     * @return a content-based hash code.
     */
    public int hashCode() {
        int result = 37;
        result *= bounds.hashCode();
        result *= plane.hashCode();
        return result;
    }

    /**
     * Returns a string representation of this pair of bounding parallel 
     * hyperplanes.
     * @return a string representation of this pair of bounding parallel 
     * hyperplanes.
     */
    public String toString() {
        return plane + " = " + bounds;
    }

    /**
     * Indicates if the provided object is equal to this pair of
     * bounding parallelt hyperplanes. Two pairs of bounding parallel
     * hyperplanes are equal if their bounds and hyperplanes are equal.
     * 
     * @param obj the object with which to compare.
     * 
     * @return {@code true} if this object is equal to the provided object;
     * {@code false} otherwise.
     */
    public boolean equals(Object obj) {
    	
        if (obj == this) {
            return true;
        }
        if (!(obj instanceof ParallelHyperplanes)) {
            return false;
        }
        
        ParallelHyperplanes<?> p = (ParallelHyperplanes) obj;
        
        if (!bounds.equals(p.getBounds())) {
            return false;
        }
        if (!p.getPlane().equals(plane)) {
            return false;
        }
        
        return true;
        
    }
}
