/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.geom.comp;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.math.geom.ConvexPolytope;
import sbrt.kernel.math.geom.DimensionException;
import sbrt.kernel.math.geom.ParallelHyperplanes;
import sbrt.kernel.utilities.*;

/**
 * This class is an implemenation of a coordinate-direction hit-and-run
 * algorithm.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <V> the variable type.
 * 
 */
public class CdHarComp<V> {

    // The convex polytope in which random points will be generated.
    private final ConvexPolytope<V> polytope;

    // The parallel hyperplanes that define the convex polytope.
    private final Set<ParallelHyperplanes<V>> hyperplanes;

    // The variables in the convex polytope.
    private final ArrayList<V> variables;

    // The current values for each variable.
    private final Map<V, Double> variableValues;

    // The psuedorandom number generator
    private Random random;

    // The constraint tolerance.
    private double tolerance;

    // The minimum chord length of valid moves.
    private double minChordLength;

    /**
     * Constructs a new random point generator for the provided convex polytope.
     * 
     * @param polytope the convex polytope for which random interior points are
     * to be generated.
     * 
     * @throws DimensionException if the dimension of the provided polytope is &lt; 1. 
     */
    public CdHarComp(ConvexPolytope<V> polytope) {

        // Checking polytope dimensions
        if (polytope.dimensions() < 1) {
            throw new DimensionException(polytope.dimensions());
        }
        
        // Setting the constraint tolerance to an out-of-band value
        tolerance = Double.NaN;

        // Setting the minimum chord (i.e. line segment) length to an
        // out-of-band value
        minChordLength = Double.NaN;

        // Getting the hyperplanes in this polyope
        hyperplanes = polytope.getBoundingPlanes();

        // Getting the variables in this polytope
        Set<V> vars = polytope.getVariables();

        // Filling array with variables for quicker access
        variables = new ArrayList<V>(vars);

        // Determining appropriate size
        int size = variables.size() + 1;

        // Creating storage for variable values
        variableValues = new LinkedHashMap<V, Double>(size, 1.0F);

        // Setting the polytope
        this.polytope = polytope;
    }

    /**
     * Sets the initial point. This method sets the initial interior point from
     * which the algorithm is started.
     * 
     * @param startPoint variables mapped to their numerical values.
     * 
     * @throws InvalidPointException if the provided point does not lie within
     * the polytope.
     */
    public void setPoint(Map<? extends V, Double> startPoint) {

        // Making sure the point lies within the polytope
        if (!polytope.isWithin(startPoint, tolerance)) {
            String s = "The provided point does not lie within the poltyope.";
            throw new InvalidPointException(startPoint, s);
        }

        // Iterating thru each free variable
        for (int i = 0, size = variables.size(); i < size; i++) {

            // Getting the next variable
            V var = variables.get(i);

            // Getting its corresponding value
            Double value = startPoint.get(var);

            // Saving
            variableValues.put(var, value);
        }

    }

    /**
     * Sets the random number generator.
     * 
     * @param random the random number generator with which all psuedorandom
     * numbers are to be produced.
     */
    public void setRandom(Random random) {
        Util.nullCheck(random);
        this.random = random;
    }

    /**
     * Sets the constraint tolerance. This is the amount by which constraints
     * can be violated.
     * 
     * @param tolerance the amount by which constaints can be violated. This
     * should be a small positive number, like 1E-9 for example.
     */
    public void setConstraintTolerance(double tolerance) {

        if (Double.isInfinite(tolerance) || Double.isNaN(tolerance)
                || tolerance < 0d) {

            String s = "Invalid constraint tolerance: " + tolerance;
            throw new IllegalArgumentException(s);
        }

        this.tolerance = tolerance;
    }

    /**
     * Sets the minimum chord length. If a computed chord is shorter than the
     * provided amount, the move is considered obstructed. This can happen if the
     * random walker gets stuck in a corner.
     * 
     * @param minChordLength the minimum length a chord must have to be
     * considered valid. This should be a small positive number, like 1E-9 for
     * example.
     */
    public void setMinChordLength(double minChordLength) {

        if (Double.isInfinite(minChordLength) || Double.isNaN(minChordLength)
                || minChordLength < 0d) {

            String s = "Invalid minimum chord length: " + minChordLength;
            throw new IllegalArgumentException(s);
        }

        this.minChordLength = minChordLength;
    }

    /**
     * Returns the constraint tolerance. This is the amount by which constraints
     * can be violated.
     * 
     * @return the constraint tolerance.
     * 
     * @see #setConstraintTolerance(double)
     */
    public double getConstraintTolerance() {

        // Checking state
        if (Double.isNaN(tolerance)) {
            String s = "The constraint tolerance was never set.";
            throw new IllegalStateException(s);
        }

        return tolerance;
    }

    /**
     * Returns the minimum length a computed chord must have to be considered
     * successful. Chords with a length smaller than the returned value are
     * considered obstructed moves.
     * 
     * @return the minimum chord length.
     * 
     * @see #setMinChordLength(double)
     */
    public double getMinChordLength() {

        // Checking state
        if (Double.isNaN(minChordLength)) {
            String s = "The minimum chord length was never set.";
            throw new IllegalStateException(s);
        }

        return minChordLength;
    }

    /**
     * Returns the convex polytope for which random points are generated.
     * 
     * @return the convex polytope for which random points are generated.
     */
    public ConvexPolytope<V> getPolytope() {
        return polytope;
    }

    private void checkState() {

        getConstraintTolerance();
        getMinChordLength();

        if (random == null) {
            String s = "The pseudo-random number generator was never set.";
            throw new IllegalStateException(s);
        }

        if (variableValues.isEmpty()) {
            String s = "The initial point was never set.";
            throw new IllegalStateException(s);
        }
    }

    /**
     * Computes and returns the next random interior point of the convex
     * polytope.
     * 
     * @return the next random interior point of the convex polytope.
     */
    public Results next() {
        return next(1);
    }

    /**
     * Computes the specified number of random interior points, and returns the
     * last one computed. The changed and unchanged variables in the returned
     * {@code Results} are the cumulative totals from all the random points
     * generated.
     * 
     * @param totalPoints the total number of random points to compute.
     * 
     * @return the last random interior point of the convex polytope computed by
     * this method.
     */
    public Results next(int totalPoints) {

        // Ensuring all required inputs were provided
        checkState();

        // Checking input
        if (totalPoints < 1) {
            String s = "totalPoints < 1: " + totalPoints;
            throw new IllegalArgumentException(s);
        }
        
        // Creating storage for the unchanged variables,
        // which hopefully remains empty
        Set<V> unchangedVariables = new LinkedHashSet<V>();

        // Generating the next random point
        makePoints(totalPoints, unchangedVariables);

        // Returning the computed results
        return new Results(unchangedVariables, variableValues);
    }

    private void makePoints(int n, Set<V> unchangedVariables) {

        for (int i = 0; i < n; i++) {
            makePoint(unchangedVariables);
        }

    }

    private void makePoint(Set<V> unchangedVariables) {

        // Randomizing the order of the variables
        Collections.shuffle(variables, random);

        // Iterating thru each variable
        for (int i = 0, size = variables.size(); i < size; i++) {

            // Getting the next variable
            V var = variables.get(i);

            // Getting the lower and upper bounds for the chord
            // (i.e. line segment) in the direction of this variable
            Interval lambdaBounds = getLambdaBounds(var);

            // Computing the length of this chord
            double range = lambdaBounds.getUpper() - lambdaBounds.getLower();

            // If this length is less than the specified minimum,
            // consider the variable unchanged
            if (range < minChordLength) {
                unchangedVariables.add(var);
            }

            // Getting a random (uniformly distributed) point
            // along this chord
            double lambda = getRandomLambda(lambdaBounds);

            // Moving along the chord by lambda
            move(var, lambda);
        }

    }

    /**
     * Computes and returns a random interior point of the convex polytope
     * starting from the provided point.
     * 
     * @param startPoint variables mapped to their numerical values.
     * 
     * @return a random interior point of the convex polytope.
     * 
     * @see #setPoint(Map)
     * 
     * @throws InvalidPointException if the provided point does not lie within
     * the polytope.
     */
    public Results run(Map<? extends V, Double> startPoint) {

        // Setting the start point
        setPoint(startPoint);

        // Computing random point
        return next();
    }

    private double getRandomLambda(Interval lambdaBounds) {

        // Computing the length of the chord
        double length = lambdaBounds.getUpper() - lambdaBounds.getLower();

        // Generating a random number
        double r = random.nextDouble();

        // Returning the random position on the chord
        return lambdaBounds.getLower() + r * length;
    }

    private Collection<Interval> getAllLambdaValues(V directionVariable) {

        // Creating storage for lower and upper lambda values
        Set<Interval> lambdaValues = new LinkedHashSet<Interval>();

        for (ParallelHyperplanes<V> p : hyperplanes) {

            // If this pair of hyperplanes does not contain the provided
            // variable, skip ahead to the next
            if (!p.getPlane().hasVariable(directionVariable)) {
                continue;
            }

            // Computing lower and upper bounds on lambda
            Interval range = getLambdaRange(directionVariable, p);

            // Saving the lambda range
            lambdaValues.add(range);
        }

        return lambdaValues;
    }

    private Interval getLambdaRange(V directionVariable,
            ParallelHyperplanes<V> p) {

        // Getting the plane
        LinearComb<V> plane = p.getPlane();

        // Getting the right-hand sides
        Interval rightHandSides = p.getBounds();
        double lowerRhs = rightHandSides.getLower() - tolerance;
        double upperRhs = rightHandSides.getUpper() + tolerance;

        // If these planes do not form edges of the polytope
        if (Double.isInfinite(lowerRhs) && Double.isInfinite(upperRhs)) {
            return rightHandSides;
        }

        // Plugging in values from current point
        double a = plane.plugInNumbers(variableValues).getConstant();

        // Getting coefficient of the term with the direction variable
        double b = plane.getCoefficient(directionVariable);

        // Computing both lambdas
        double lambda1 = (lowerRhs - a) / b;
        double lambda2 = (upperRhs - a) / b;

        // Distinguishing lower from upper
        double lowerLambda = Util.getSmallest(lambda1, lambda2);
        double upperLambda = Util.getLargest(lambda1, lambda2);

        // If lower lambda > 0
        if (Double.compare(lowerLambda, 0) > 0) {

            // If lower lamba is greater than the defined tolerance,
            // throw an exception
            if (lowerLambda > tolerance) {
                String s =
                        "In the direction '"
                                + directionVariable
                                + "', the lower lambda value is greater than 0: "
                                + lowerLambda;
                throw new IllegalStateException(s);
            }

            // Otherwise correct it
            lowerLambda = 0;
        }

        // If upper lambda < 0
        if (Double.compare(upperLambda, 0) < 0) {

            // If absolute value of the upper lamba is greater
            // than the defined tolerance, throw an exception
            if (Math.abs(upperLambda) > tolerance) {
                String s =
                        "In the direction '" + directionVariable
                                + "', the upper lambda value is less than 0: "
                                + lowerLambda;
                throw new IllegalStateException(s);
            }

            // Otherwise correct it
            upperLambda = 0;
        }

        // Returning the distances to the nearest facets in both
        // positive and negative directions
        return new Interval(lowerLambda, upperLambda);
    }

    private Interval getLambdaBounds(V directionVariable) {

        // Getting all lower and upper lambda values
        Collection<Interval> lambdaValues =
                getAllLambdaValues(directionVariable);

        // Initializing fields to out-of-band values
        double smallestUpper = Double.POSITIVE_INFINITY;
        double largestLower = Double.NEGATIVE_INFINITY;

        // Iterating thru each pair of lambda values
        for (Interval c : lambdaValues) {

            // Getting the smallest and largest values
            smallestUpper = Util.getSmallest(smallestUpper, c.getUpper());
            largestLower = Util.getLargest(largestLower, c.getLower());
        }

        // If either is infinite, then the polytope is unbounded
        // in the current direction
        if (Double.isInfinite(smallestUpper) || Double.isInfinite(largestLower)) {

            throw new UnboundedVariableException(directionVariable);
        }

        return asBounds(directionVariable, largestLower, smallestUpper);
    }

    private Interval asBounds(V directionVariable,
            double lowerValue,
            double upperValue) {

        // If the lower lambda value is greater than the upper lambda
        // value by more than the constraint tolerance, no feasible
        // solutions exist
        if (lowerValue - tolerance > upperValue) {
            String s =
                    "Lower lambda value = " + lowerValue
                            + ", Upper lambda value = " + upperValue;
            throw new InfeasibleSolutionException(directionVariable, s);
        }

        // Correcting for small numerical error
        if (lowerValue > upperValue) {
            lowerValue = 0.0;
            upperValue = 0.0;
        }

        return new Interval(lowerValue, upperValue);
    }

    private void move(V directionVariable, double lambda) {

        // Getting the current value for the direction variable
        Double oldValue = variableValues.get(directionVariable);

        // Computing the new value
        double newValue = oldValue.doubleValue() + lambda;

        // Converting to Double
        Double x = Double.valueOf(newValue);

        // Setting the new value
        variableValues.put(directionVariable, x);
    }

    /**
     * This class is used to represent the results of a random point
     * computation.
     * 
     * @author This class was written and documented by Jeremiah Wright while in
     * the Wagner lab.
     * 
     * 
     */
    public class Results {

        private final UnmodifiableMap<V, Double> point;

        private final UnmodifiableSet<V> unchangedVariables;

        private Results(Set<V> unchangedVariables, Map<V, Double> point) {

            this.unchangedVariables = copySet(unchangedVariables);
            this.point = copyMap(point);

        }

        private UnmodifiableMap<V, Double> copyMap(Map<V, Double> map) {

            // Determining an appropriate size
            int size = map.size() + 1;

            // Copying
            Map<V, Double> temp = new LinkedHashMap<V, Double>(size, 1.0F);
            temp.putAll(map);

            // Returning as unmodifiable
            return new UnmodifiableMap<V, Double>(temp);
        }

        private UnmodifiableSet<V> copySet(Set<V> set) {

            // Determining an appropriate size
            int size = set.size() + 1;

            // Copying
            Set<V> temp = new LinkedHashSet<V>(size, 1.0F);
            temp.addAll(set);

            // Returning as unmodifiable
            return new UnmodifiableSet<V>(temp);
        }

        /**
         * Returns the random interior point that was computed.
         * 
         * @return the random interior point that was computed.
         */
        public UnmodifiableMap<V, Double> getPoint() {
            return point;
        }

        /**
         * Returns the variables whose values are different than they were in
         * the previous point. Ideally, all of the variables of the polytope are
         * present in this set.
         * 
         * @return the variables whose values are different than they were in
         * the previous point.
         */
        public Set<V> getChangedVariables() {

            // Creating storage
            Set<V> s = new LinkedHashSet<V>(point.size() + 1, 1.0F);

            // Filling
            s.addAll(variables);
            s.removeAll(unchangedVariables);

            return s;
        }

        /**
         * Returns the variables whose values are the same as they were in the
         * previous point. Ideally, this set should be empty. If it isn't, it
         * could mean the random walker is stuck in a corner.
         * 
         * @return the variables whose values are the same as they were in the
         * previous point.
         */
        public UnmodifiableSet<V> getUnchangedVariables() {
            return unchangedVariables;
        }

        /**
         * Returns the number of variables whose values are different than they
         * were in the previous point. Ideally, this number should equal the
         * total number of variables in the convex polytope.
         * 
         * @return the number of variables whose values are different than they
         * were in the previous point.
         */
        public int changedVariables() {
            return variables.size() - unchangedVariables.size();
        }

        /**
         * Returns the number of variables whose values are the same as they
         * were in the previous point. Ideally, this number should equal 0.
         * 
         * @return the number of variables whose values are the same as they
         * were in the previous point.
         */
        public int unchangedVariables() {
            return unchangedVariables.size();
        }

    }

}
