/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.geom.comp;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

/**
 * This class is used to compute Hamming distances - the number of differences
 * between two "strings" of equal length.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 *
 */
public final class Hamming {

	/**
	 * Computes and returns the Hamming distance between the provided strings.
	 * 
	 * @param x a string.
	 * @param y another string.
	 * @return the Hamming distance between the provided strings.
	 * 
	 * @throws IllegalArgumentException if the lengths of the provided strings 
	 * are not equal.
	 */
	public int getDist(String x, String y) {
		
		check(x, y);
		
		int diffs = 0;
		
		int size = x.length();
		
		for (int i = 0; i < size; i ++) {
			
			if (x.charAt(i) != y.charAt(i)) {
				++ diffs;
			}
			
		}
		
		return diffs;
	}
	
	private void check(String x, String y) {
		
		if (x.length() != y.length()) {
			String s = "Provided strings have different lengths: " +
						x.length() + " != " + y.length();
			throw new IllegalArgumentException(s);
		}
		
	}

	/**
	 * Computes and returns the Hamming distance between the provided arrays.
	 * Objects are compared via their {@code equals()} method.
	 * 
	 * @param x an array of objects.
	 * @param y another array of objects.
	 * @return the Hamming distance between the provided arrays.
	 * 
	 * @throws IllegalArgumentException if the lengths of the provided arrays 
	 * are not equal.
	 */
	public int getDist(Object[] x, Object[] y) {
		
		check(x, y);
		
		return getDist(Arrays.asList(x), Arrays.asList(y));
	}
	
	private void check(Object[] x, Object[] y) {
		
		if (x.length != y.length) {
			String s = "Provided arrays have different lengths: " +
			x.length + " != " + y.length;
			throw new IllegalArgumentException(s);
		}
		
	}

	/**
	 * Computes and returns the Hamming distance between the provided collections.
	 * Objects are compared via their {@code equals()} method upon iteration
	 * over each collection.
	 * 
	 * @param x a collection of objects.
	 * @param y another collection of objects.
	 * @return the Hamming distance between the provided collections.
	 * 
	 * @throws IllegalArgumentException if the lengths of the provided collections 
	 * are not equal.
	 */
	public int getDist(Collection<?> x, Collection<?> y) {
		
		check(x, y);
		
		int diffs = 0;
		
		Iterator<?> i = x.iterator();
		Iterator<?> j = y.iterator();
		
		while (i.hasNext()) {
			
			if (!i.next().equals(j.next())) {
				++ diffs;
			}
			
		}
		
		return diffs;
	}
	
	private void check(Collection<?> x, Collection<?> y) {
		
		if (x.size() != y.size()) {
			String s = "Provided collections have different lengths: " +
			x.size() + " != " + y.size();
			throw new IllegalArgumentException(s);
		}
		
	}

	/**
	 * Computes and returns the Hamming distance between the provided maps.
	 * The sets of keys from each map must be equal. Map values are compared 
	 * via their {@code equals()} method.
	 * 
	 * @param x a map.
	 * @param y another map.
	 * @return the Hamming distance between the provided maps.
	 * 
	 * @throws IllegalArgumentException if the lengths of the provided maps 
	 * or the key sets are not equal.
	 */
	public int getDist(Map<?, ?> x, Map<?, ?> y) {
		
		check(x, y);
		
		int diffs = 0;
		
		for (Map.Entry<?, ?> e : x.entrySet()) {
			
			Object key = e.getKey();
			
			Object xValue = e.getValue();
			Object yValue = y.get(key);
			
			if (!xValue.equals(yValue)) {
				++ diffs;
			}
		}
		
		return diffs;
	}
	
	private void check(Map<?, ?> x, Map<?, ?> y) {
		
		if (x.size() != y.size()) {
			String s = "Provided maps have different lengths: " +
			x.size() + " != " + y.size();
			throw new IllegalArgumentException(s);
		}
		
		if (!x.keySet().equals(y.keySet())) {
			String s = "Provided maps have different keys: " +
				x.keySet() + " != " + y.keySet();
			throw new IllegalArgumentException(s);
		}

	}


}
