/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.geom.comp;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

/**
 * This class is used to compute Taxicab distances. See Wikipedia for more 
 * information. All provided numbers are converted to double precision via
 * the method {@linkplain Number#doubleValue()}. 
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 *
 */
public final class Taxicab {


	/**
	 * Computes and returns the Taxicab distance between the provided arrays.
	 * 
	 * @param x an array of numbers.
	 * @param y another array of numbers.
	 * @return the Taxicab distance between the provided arrays.
	 * 
	 * @throws IllegalArgumentException if the lengths of the provided arrays 
	 * are not equal.
	 */
	public double getDist(Number[] x, Number[] y) {
		
		check(x, y);
		
		return getDist(Arrays.asList(x), Arrays.asList(y));
	}
	
	private void check(Object[] x, Object[] y) {
		
		if (x.length != y.length) {
			String s = "Provided arrays have different lengths: " +
			x.length + " != " + y.length;
			throw new IllegalArgumentException(s);
		}
		
	}

	/**
	 * Computes and returns the Taxicab distance between the provided collections.
	 * Numbers are compared upon iteration over each collection.
	 * 
	 * @param x a collection of numbers.
	 * @param y another collection of numbers.
	 * @return the Taxicab distance between the provided collections.
	 * 
	 * @throws IllegalArgumentException if the lengths of the provided collections 
	 * are not equal.
	 */
	public double getDist(Collection<? extends Number> x, 
					   	  Collection<? extends Number> y) {
		
		check(x, y);
		
		double sum = 0d;
		
		Iterator<? extends Number> i = x.iterator();
		Iterator<? extends Number> j = y.iterator();
		
		while (i.hasNext()) {
			
			double xNum = i.next().doubleValue();
			double yNum = j.next().doubleValue();
			
			sum += Math.abs( xNum - yNum );
			
		}
		
		return sum;
	}
	
	private void check(Collection<?> x, Collection<?> y) {
		
		if (x.size() != y.size()) {
			String s = "Provided collections have different lengths: " +
			x.size() + " != " + y.size();
			throw new IllegalArgumentException(s);
		}
		
	}

	/**
	 * Computes and returns the Taxicab distance between the provided maps.
	 * The sets of keys from each map must be equal.
	 * 
	 * @param x a map.
	 * @param y another map.
	 * @return the Taxicab distance between the provided maps.
	 * 
	 * @throws IllegalArgumentException if the lengths of the provided maps 
	 * or the key sets are not equal.
	 */
	public double getDist(Map<?, ? extends Number> x, 
						  Map<?, ? extends Number> y) {
		
		check(x, y);
		
		double sum = 0d;
		
		for (Map.Entry<?, ? extends Number> e : x.entrySet()) {
			
			Object key = e.getKey();
			
			double xNum = e.getValue().doubleValue();
			double yNum = y.get(key).doubleValue();
			
			sum += Math.abs( xNum - yNum );

		}
		
		return sum;
	}
	
	private void check(Map<?, ?> x, Map<?, ?> y) {
		
		if (x.size() != y.size()) {
			String s = "Provided maps have different lengths: " +
			x.size() + " != " + y.size();
			throw new IllegalArgumentException(s);
		}
		
		if (!x.keySet().equals(y.keySet())) {
			String s = "Provided maps have different keys: " +
				x.keySet() + " != " + y.keySet();
			throw new IllegalArgumentException(s);
		}

	}


}
