/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.geom.proc;

import java.util.Map;
import java.util.Random;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.ExceededLimitException;
import sbrt.kernel.math.InvalidPointException;
import sbrt.kernel.math.geom.ConvexPolytope;
import sbrt.kernel.math.geom.comp.CdHarComp;
import sbrt.kernel.utilities.PercentProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;

/**
 * This class is used to generate random, uniformly distributed points from the
 * interior of a convex polytope.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <V> the variable type.
 */
public final class CdHarProcess<V> implements MonitoredProcess<CdHarComp<V>.Results> {

    private final PercentProgressMonitor progress;

    private final CdHarComp<V> computer;

    private int iterations;

    private int pointsPerPoint;

    private int maxObstructedMoves;

    /**
     * Constructs a new random point generator process for the provided polytope.
     * 
     * @param polytope the polytope from which random points will be generated.
     */
    public CdHarProcess(ConvexPolytope<V> polytope) {
        
        computer = new CdHarComp<V>(polytope);
        progress = new PercentProgressMonitor();

        iterations = -1;
        pointsPerPoint = -1;
        maxObstructedMoves = -1;
    }

    /**
     * Sets the initial point. This method sets the initial interior point from
     * which the algorithm is started.
     * 
     * @param startPoint variables mapped to their numerical values.
     * 
     * @throws InvalidPointException if the provided point does not lie within
     * the polytope.
     */
    public void setPoint(Map<? extends V, Double> startPoint) {
        computer.setPoint(startPoint);
    }

    /**
     * Sets the random number generator.
     * 
     * @param random the random number generator with which all psuedorandom
     * numbers are to be produced.
     */
    public void setRandom(Random random) {
        computer.setRandom(random);
    }

    /**
     * Sets the constraint tolerance. This is the amount by which constraints
     * can be violated.
     * 
     * @param tolerance the amount by which constaints can be violated. This
     * should be a small positive number, like 1E-9 for example.
     */
    public void setConstraintTolerance(double tolerance) {
        computer.setConstraintTolerance(tolerance);
    }

    /**
     * Sets the minimum chord length. If a computed chord is shorter than the
     * provided amount, the move is considered obstructed. This can happen if the
     * random walker gets stuck in a corner.
     * 
     * @param minChordLength the minimum length a chord must have to be
     * considered valid. This should be a small positive number, like 1E-9 for
     * example.
     */
    public void setMinChordLength(double minChordLength) {
        computer.setMinChordLength(minChordLength);
    }
    
    /**
     * Sets the total number of random points to be added to the provided
     * results manager.
     * 
     * @param iterations the total number of random points to be added to the
     * provided results manager.
     * 
     * @throws IllegalArgumentException if {@code maxObstructedMoves} &lt; 0
     */
    public void setIterations(int iterations) {

        // Checking input
        if (iterations < 0) {
            String s = "iterations is less than 0: " + iterations;
            throw new IllegalArgumentException(s);
        }

        // Setting
        this.iterations = iterations;
    }

    /**
     * Sets the total number of points to generate for each point added to the
     * provided results manager.
     * 
     * @param pointsPerPoint the total number of points to generate for each
     * point added to the provided results manager.
     * 
     * @throws IllegalArgumentException if {@code pointsPerPoint} &lt; 1
     */
    public void setPointsPerPoint(int pointsPerPoint) {

        // Checking input
        if (pointsPerPoint < 1) {
            String s = "pointsPerPoint is less than 1: " + pointsPerPoint;
            throw new IllegalArgumentException(s);
        }

        // Setting
        this.pointsPerPoint = pointsPerPoint;
    }

    /**
     * Sets the maximum number of obstructed moves to allow.
     * 
     * @param maxObstructedMoves the number of obstructed moves to allow.
     * 
     * @throws IllegalArgumentException if {@code maxObstructedMoves} &lt; 0
     */
    public void setMaxObstructedMoves(int maxObstructedMoves) {

        // Checking input
        if (maxObstructedMoves < 0) {
            String s =
                    "maxObstructedMoves is less than 0: " + maxObstructedMoves;
            throw new IllegalArgumentException(s);
        }

        // Setting
        this.maxObstructedMoves = maxObstructedMoves;
    }

    private void checkState() {

        if (iterations < 0) {
            String s = "The number of iterations was never set.";
            throw new IllegalStateException(s);
        }

        if (maxObstructedMoves < 0) {
            String s = "The maximum number of obstructed moves was never set.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Returns a progress monitor for this process.
     * 
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    /**
     * Adds random points to the provided results manager.
     * 
     * @param resultsManager the results manager to which the random points are
     * added.
     * 
     * @throws IllegalStateException if any required information was not provided.
     */
    public void run(ResultsManager<CdHarComp<V>.Results> resultsManager) {

        checkState();

        try {

            // Initializing progress monitor
            progress.resetCount();
            progress.setTotal(iterations);
            progress.start();

            // Running the process
            runProcess(resultsManager);

            // Finalizing the progress monitor
            progress.stop();

        } catch (RuntimeException e) {

            // Finalizing the progress monitor
            if (!progress.stopped()) {
                progress.stop();
            }

            // Throwing the original exception
            throw e;
        }

    }

    private final void runProcess(ResultsManager<CdHarComp<V>.Results> rm) {

        // Initializing obstructed moves counter
        int totalObstructedMoves = 0;

        // Generating the specified number of random points
        for (int i = 0; i < iterations; i++) {

            // Computing the next random point
        	CdHarComp<V>.Results results = computer.next(pointsPerPoint);

            // Incrementing the number of obstructed moves
            totalObstructedMoves += results.unchangedVariables();

            // If the total exceeds the user-specified maximum
            if (totalObstructedMoves > maxObstructedMoves) {
                throw new ExceededLimitException();
            }

            // Adding the results to the manager
            rm.addData(results);

            // Incrementing the progress
            progress.increment();
        }

    }

}
