/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.graph_theory;

import java.util.*;

import sbrt.kernel.utilities.*;

/**
 * This class is used to represent directed graphs.
 * <P>
 * A directed graph is composed of nodes that are connected
 * by unidirectional edges.
 * 
 * @author This class was written and documented by 
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <N> the type of node.
 */
public final class DirectedGraph<N> {

    /** Nodes -> their successors. */
    private final Map<N, Set<N>> successors;

    /** Nodes -> their predecessors. */
    private final Map<N, Set<N>> predecessors;

    /** The total number of edges in the graph. */
    private int edges;

    /**
     * Constructs a new directed graph from the provided
     * collection of edges.
     * 
     * @param edges the collection of all edges contained
     * in this graph.
     */
    public DirectedGraph(Set<Edge<N>> edges) {

        // Creating storage of minimum size
        successors = new LinkedHashMap<N, Set<N>>(edges.size());
        predecessors = new LinkedHashMap<N, Set<N>>(edges.size());

        // Initializing edges
        this.edges = 0;

        // Adding each provided edge
        for (Edge<N> e : edges) {
            addEdge(e);
        }
    }

    private void addEdge(Edge<N> e) {

        // Getting source and sink
        N sourceNode = e.getSource();
        N sinkNode = e.getSink();

        // Setting up fields for temporary storage
        Set<N> sourceSuccessors = null;
        Set<N> sinkSuccessors = null;

        // Retrieving the successors of the source node
        sourceSuccessors = successors.get(sourceNode);
  
        // If no successors exist yet for the source node
        if (sourceSuccessors == null) {
            sourceSuccessors = new LinkedHashSet<N>();
            successors.put(sourceNode, sourceSuccessors);
        }

        // Retrieving the successors of the sink node
        sinkSuccessors = successors.get(sinkNode);

        // If no successors exist yet for the sink node
        if (sinkSuccessors == null) {
            sinkSuccessors = new LinkedHashSet<N>();
            successors.put(sinkNode, sinkSuccessors);
        }

        // Retrieving the predecessors of the source node
        Set<N> sourcePredecessors = predecessors.get(sourceNode);

        // If no predecessors exist yet for the source node
        if (sourcePredecessors == null) {
            sourcePredecessors = new LinkedHashSet<N>();
            predecessors.put(sourceNode, sourcePredecessors);
        }

        // Retrieving the predecessors of the sink node
        Set<N> sinkPredecessors = predecessors.get(sinkNode);

        // If no predecessors exist yet for the sink node
        if (sinkPredecessors == null) {
            sinkPredecessors = new LinkedHashSet<N>();
            predecessors.put(sinkNode, sinkPredecessors);
        }

        // Adding the sink node to the source node's successors
        if (sourceSuccessors.add(sinkNode)) {
            ++edges;
        }

        // Adding the source node to the sink node's predecessors
        sinkPredecessors.add(sourceNode);
    }

    /**
     * Indicates if the provided object is a node in this graph.
     * 
     * @param potentialNode the reference object with which to 
     * compare. 
     * 
     * @return {@code true} if the provided object is a node in
     * this graph; {@code false} otherwise.
     */
    public boolean isNode(Object potentialNode) {
        return successors.containsKey(potentialNode);
    }

    /**
     * Indicates if an edge exists between the provided nodes.
     * 
     * @param sourceNode the source node.
     * @param sinkNode the sink node.
     * @return {@code true} if an edge exists between the provided
     * source and sink nodes; {@code false} otherwise.
     */
    public boolean isEdge(Object sourceNode, Object sinkNode) {
    	
    	// Checking to see if nodes exist
    	if (!isNode(sourceNode) || ! isNode(sinkNode)) {
    		return false;
    	}
    	
    	// Getting all successors of the provided source node
    	Set<N> succ = successors.get(sourceNode);
    	
    	// Checking to see if the provided sink node is in this set 
    	return succ.contains(sinkNode);
    }
    
    /**
     * Returns the set of all nodes contained in this graph.
     * @return the set of all nodes contained in this graph.
     */
    public Set<N> getNodes() {

    	// Determining appropriate size
        Set<N> s = new LinkedHashSet<N>(successors.size() + 1, 1.0F);
        
        // Copying
        s.addAll(successors.keySet());

        return s;
    }

    /**
     * Returns the set of all edges contained in this graph.
     * @return the set of all edges contained in this graph.
     */
    public Set<Edge<N>> getEdges() {

        // Creating storage of appropriate size
        Set<Edge<N>> s = new LinkedHashSet<Edge<N>>(edges + 1, 1.0F);

        // Iterating thru each node
        for (Map.Entry<N, Set<N>> entry : successors.entrySet()) {

            // Getting the source node and its successors
            N source = entry.getKey();
            Set<N> sinks = entry.getValue();

            // Iterating thru each sink node
            for (N sink : sinks) {

                // Adding this edge to the rest
                s.add(new Edge<N>(source, sink));
            }

        }

        return s;
    }

    /**
     * Returns the number of nodes in this graph.
     * @return the number of nodes in this graph.
     */
    public int nodes() {
    	return successors.size();
    }
    
    /**
     * Returns the number of edges in this graph.
     * @return the number of edges in this graph.
     */
    public int edges() {
    	return edges;
    }
    
    /**
     * Returns the number of nodes neighboring the provided node.
     * 
     * @param node the node whose number of neighbors will be returned.
     * 
     * @return the number neighboring nodes of the provided node.
     * 
     * @throws IllegalArgumentException if the provided node
     * does not exist in this graph.
     */
    public int neighbors(Object node) {

        // Getting the provided node's successors and predecessors
        Set<N> succ = successors.get(node);
        Set<N> pred = predecessors.get(node);

        // If the provided node does not exist in this graph
        if (succ == null) {
            throw new IllegalArgumentException("No such node: " + node);
        }

        return succ.size() + pred.size();
    }
    
    /**
     * Returns the neighboring nodes of the provided node.
     * 
     * @param node the node whose neighbors will be returned.
     * 
     * @return the neighboring nodes of the provided node.
     * 
     * @throws IllegalArgumentException if the provided node
     * does not exist in this graph.
     */
    public Collection<N> getNeighbors(Object node) {

        // Getting the provided node's successors and predecessors
        Set<N> succ = successors.get(node);
        Set<N> pred = predecessors.get(node);

        // If the provided node does not exist in this graph
        if (succ == null) {
            throw new IllegalArgumentException("No such node: " + node);
        }

        // Combining them
        Set<N> n = new LinkedHashSet<N>(succ.size() + pred.size() + 1, 1.0F);
        n.addAll(succ);
        n.addAll(pred);

        // Returning all nodes neighboring the provided node
        return n;
    }

    /**
     * Returns the number of successors of the provided node.
     * This is also sometimes called the out-degree.
     * 
     * @param node the node whose number of successors will be
     * returned.
     * 
     * @return the number of successors of the provided node.
     * 
     * @throws IllegalArgumentException if the provided node
     * does not exist in this graph.
     */
    public int successors(Object node) {

        // Getting the provided node's successors
        Set<N> succ = successors.get(node);

        // If the provided node does not exist in this graph
        if (succ == null) {
            throw new IllegalArgumentException("No such node: " + node);
        }

        return succ.size();
    }

    /**
     * Returns the successors of the provided node.
     * 
     * @param node the node whose successors will be
     * returned.
     * 
     * @return the successors of the provided node.
     * 
     * @throws IllegalArgumentException if the provided node
     * does not exist in this graph.
     */
    public Set<N> getSuccessors(Object node) {

        // Getting the provided node's successors
        Set<N> succ = successors.get(node);

        // If the provided node does not exist in this graph
        if (succ == null) {
            throw new IllegalArgumentException("No such node: " + node);
        }

        // Copying into a new Set
        Set<N> s = new LinkedHashSet<N>(succ.size() + 1, 1.0F);
        s.addAll(succ);

        // Returing all successors of the provided node
        return s;
    }

    /**
     * Returns the number of predecessors of the provided node.
     * This is also sometimes called the in-degree.
     * 
     * @param node the node whose number of predecessors will be
     * returned.
     * 
     * @return the number of predecessors of the provided node.
     * 
     * @throws IllegalArgumentException if the provided node
     * does not exist in this graph.
     */
    public int predecessors(Object node) {

        // Getting the provided node's predecessors
        Set<N> pred = predecessors.get(node);

        // If the provided node does not exist in this graph
        if (pred == null) {
            throw new IllegalArgumentException("No such node: " + node);
        }

        return pred.size();
    }

    /**
     * Returns the predecessors of the provided node.
     * 
     * @param node the node whose predecessors will be
     * returned.
     * 
     * @return the predecessors of the provided node.
     * 
     * @throws IllegalArgumentException if the provided node
     * does not exist in this graph.
     */
    public Set<N> getPredecessors(Object node) {

        // Getting the provided node's predecessors
        Set<N> pred = predecessors.get(node);

        // If the provided node does not exist in this graph
        if (pred == null) {
            throw new IllegalArgumentException("No such node: " + node);
        }

        // Copying into a new Set
        Set<N> s = new LinkedHashSet<N>(pred.size() + 1, 1.0F);
        s.addAll(pred);

        // Returing all predecessors of the provided node
        return s;
    }
    

    /**
     * This class is used to represent the unidirectional
     * edges of a directed graph. An edge is directed from a
     * source node to a sink node.
     * 
     * @author This class was written and documented by
     * Jeremiah Wright while in the Wagner lab.
     * 
     * @param <N> the type of node.
     */
    public static final class Edge<N> {

        private final N source, sink;

        /**
         * Constructs a new directed edge from the provided
         * source and sink nodes.
         * 
         * @param <T> a type that extends N.
         * @param <S> a type that extends N.
         * @param source the source node.
         * @param sink the sink node.
         */
        public <T extends N, S extends N> Edge(T source, S sink) {

            // Checking inputs
            Util.nullCheck(source);
            Util.nullCheck(sink);

            // Saving
            this.source = source;
            this.sink = sink;
        }

        /**
         * Returns the source node of this directed edge.
         * @return the source node of this directed edge.
         */
        public N getSource() {
            return source;
        }

        /**
         * Returns the sink node of this directed edge.
         * @return the sink node of this directed edge.
         */
        public N getSink() {
            return sink;
        }

        /**
         * Indicates if the provided object is a node of this directed edge.
         * @param node the potential node.
         * @return {@code true} if the provided object equals the source or
         * sink node; {@code false} otherwise.
         */
        public boolean hasNode(Object node) {
        	return source.equals(node) || sink.equals(node);
        }
        
        /**
         * Indicates if the provided object is equal to this
         * directed edge. An object is equal to this edge if
         * 1) it is an instance of {@code Edge}, 2)
         * its source node is the same as this edge's, and
         * 3) its sink node is the same as this edge's.
         * 
         * @param o the reference object with which to compare.
         * 
         * @return {@code true} if this object is equal to the
         * provided object; {@code false} otherwise. 
         */
        public boolean equals(Object o) {

            if (o == this) {
                return true;
            }

            if (!(o instanceof DirectedGraph.Edge)) {
                return false;
            }

            Edge<?> e = (Edge) o;

            return source.equals(e.source) && sink.equals(e.sink);
        }

        /**
         * Returns a content-based hash code for this directed edge.
         * @return a content-based hash code for this directed edge.
         */
        public int hashCode() {
            return source.hashCode() ^ sink.hashCode();
        }
        
        /**
         * Returns an appropriate string representation of this edge.
         * @return an appropriate string representation of this edge.
         */
        public String toString() {
        	return source + "->" + sink;
        }
    }

}
