/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.graph_theory;

import java.util.*;

/**
 * <p>
 * This class can be used for representing paths in a graph
 * or network.  A path is a sequence
 * {x<sub>0</sub>, x<sub>1</sub>, ..., x<sub>n-1</sub>}
 * such that (x<sub>0</sub>, x<sub>1</sub>), (x<sub>1</sub>, x<sub>2</sub>),
 *  ..., (x<sub>n-2</sub>, x<sub>n-1</sub>) are
 * edges of a graph, where n is the total number of nodes in the path.
 * </p>
 *
 * <p>Instances of this class are immutable.</p>
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <N> the node type.
 */

public class Path<N> {

    // Storage for the path
    private final N[] path;

    /**
     * Constructs a path containing the nodes from the provided 
     * array, in the order they occur starting from index 0.
     * 
     * @param <T> a type extending the primary node type.
     *
     * @param path the array whose elements will represent nodes
     * in this path.
     * 
     * @throws NullPointerException if the provided array is 
     * null or if a node contained within that array is {@code null}.
     */
    @SuppressWarnings("unchecked")
	public <T extends N> Path(T[] path) {

        // Creating storage
        this.path = (N[]) new Object[path.length];

        // Copying the provided path
        for (int i = 0; i < path.length; i++) {

            // Checking node
            if (path[i] == null) {
                String s = "null node found at position: " + i;
                throw new NullPointerException(s);
            }

            // Saving it
            this.path[i] = path[i];
        }
    }

    /**
     * Constructs a path containing the nodes from the provided 
     * collection, in the order they are returned from the 
     * collection's iterator.
     *
     * @param path the collection whose elements will represent
     * nodes in this path.
     * 
     * @throws NullPointerException if the provided collection 
     * is {@code null} or if a node contained within that 
     * collection is {@code null}.
     */
    @SuppressWarnings("unchecked")
	public Path(Collection<? extends N> path) {

        // Creating storage
        this.path = (N[]) new Object[path.size()];

        // Initializing index
        int index = 0;

        // Copying the provided path
        for (N node : path) {

            // Checking node
            if (node == null) {
                String s = "null node found at position: " + index;
                throw new NullPointerException(s);
            }

            // Saving it
            this.path[index++] = node;
        }
    }

    /**
     * Returns the number of nodes in this path.
     *
     * @return the number of nodes in this path.
     */
    public int nodes() {
        return path.length;
    }

    /**
     * Returns the node at the specified position in this path.
     *
     * @param index index of node to return.
     * @return the node at the specified position in this path.
     * @throws IndexOutOfBoundsException if index is out of range
     * (index < 0 || index >= nodes()).
     */
    public N get(int index) {

        // Checking input
        if (index < 0 || index >= path.length) {
            String s = "out of bounds position: " + index;
            throw new IllegalArgumentException(s);
        }

        // Returning the node at the specified position in the path
        return path[index];
    }
    
    /**
     * Returns the index in this path of the first occurrence of the 
     * specified node, or -1 if this path does not contain that node.
     *
     * @param node node to search for.
     * 
     * @return the index in this path of the first occurrence of the 
     * specified node, or -1 if this path does not contain that node.
     */
    public int indexOf(Object node) {

        for (int i = 0; i < path.length; i++) {
            if (path[i].equals(node)) {
                return i;
            }
        }

        return -1;
    }
    
    /**
     * Returns true if this path contains the specified node.
     *
     * @param node node whose presence in this Path is to be 
     * tested.
     * 
     * @return {@code true} if the specified node is present; 
     * {@code false} otherwise.
     */
    public boolean contains(Object node) {
    	return indexOf(node) >= 0;
    }

    /**
     * Returns an iterator over the nodes in this path in 
     * proper sequence.
     *
     * @return an iterator over the nodes in this path in 
     * proper sequence.
     */
    public Iterator<N> iterator() {

        return new Iterator<N>() {

            // Initializing index
            private int index = 0;

            public N next() {

                if (!hasNext()) {
                    throw new NoSuchElementException();
                }

                return path[index++];
            }

            public boolean hasNext() {
                return index == path.length;
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    /**
     * Returns an array representation of this path.  Nodes in 
     * this path become elements in the returned array.  The 
     * nodes occur in the same order as the do in this path, 
     * starting from index 0.
     *
     * @return an array representation of this path.
     */
    @SuppressWarnings("unchecked")
	public N[] toArray() {

        // Creating new array
        N[] array = (N[]) new Object[path.length];

        // Copying
        for (int i = 0; i < path.length; i++) {
            array[i] = path[i];
        }

        // Returning an array representation of the path
        return array;
    }

    /**
     * Returns a list representation of this path.  Nodes in this 
     * path become elements in the returned list.  The nodes occur 
     * in the same order as they do in the path.
     *
     * @return a list representation of this path.
     */
    public List<N> toList() {

        // Creating random access storage
        List<N> list = new ArrayList<N>(path.length);

        // Copying
        for (int i = 0; i < path.length; i++) {
            list.add(path[i]);
        }

        // Returning a list representation of the path
        return list;
    }

    /**
     * Compares the specified object with this path for equality.
     * Returns true if and only if the specified object is also a 
     * path, both paths have the same size, and all corresponding 
     * pairs of nodes in the two paths are equal. In other words, 
     * two paths are defined to be equal if they contain the same 
     * nodes in the same order.
     * 
     * @param obj the object with which to compare.
     *
     * @return true if the specified object is equal to this path.
     */
    public boolean equals(Object obj) {

        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Path)) {
            return false;
        }

        Path<?> p = (Path) obj;

        return Arrays.equals(p.path, path);
    }

    /**
     * Returns a hash code based on the nodes contained in this path.
     *
     * @return a content-based hash code
     */
    public int hashCode() {
        return Arrays.hashCode(path);
    }

    /**
     * Returns a string representation of this path.
     * @return a string representation of this path.
     */
    public String toString() {
        return Arrays.deepToString(path);
    }
}
