/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.graph_theory.comp;

import java.util.*;

import sbrt.kernel.math.graph_theory.DirectedGraph;
import sbrt.kernel.math.graph_theory.Path;
import sbrt.kernel.utilities.*;

/**
 * This class is an implemenation of an algorithm for finding all pathways in a
 * directed graph.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 * @param <N> the node type.
 */
public class PathFinder<N> {

	private final DirectedGraph<N> graph;

	/**
	 * Constructs a new path finder for the provided graph.
	 * 
	 * @param graph the directed graph in which pathways will be computed.
	 */
	public PathFinder(DirectedGraph<N> graph) {
		// Checking and setting
		Util.nullCheck(graph);
		this.graph = graph;
	}

	/**
	 * Returns the graph in which pathways will be computed.
	 * @return the graph in which pathways will be computed.
	 */
	public DirectedGraph<N> getGraph() {
		return graph;
	}

	/**
	 * Returns a list of all paths.
	 * @return a list of all paths.
	 */
	public List<Path<N>> getPaths() {

		// Getting all nodes in the graph
		Set<N> nodes = graph.getNodes();

		// Returning all paths in the graph
		return getPaths(nodes, nodes);
	}

	/**
	 * Returns a list of all paths between the provided source and sink nodes.
	 * 
	 * @param sources the collection of nodes at which all returned paths will
	 * begin.
	 * @param sinks the collection of nodes at which all returned paths will end.
	 * 
	 * @return a list of all paths between the provided source and sink nodes.
	 */
	public List<Path<N>> 
	getPaths(Collection<? extends N> sources, Collection<? extends N> sinks) {

		// Creating storage for all paths
		List<Path<N>> paths = new LinkedList<Path<N>>();

		// Iterating thru each source node
		for (N source : sources) {

			// Iterating thru each sink node
			for (N sink : sinks) {

				// Getting the paths between these source and sink nodes
				List<Path<N>> p = getPaths(source, sink);

				// Adding these paths to the rest
				paths.addAll(p);
			}
		}

		// Returning all the paths between the provided source and sink nodes
		return paths;
	}

	/**
	 * Returns a list of all paths between the provided source and sink nodes.
	 * Each path in the returned list will begin at the source node and end at
	 * the sink node.
	 * 
	 * @param source
	 *            the node at which all returned paths will begin.
	 * @param sink
	 *            the node at which all returned paths will end.
	 * @param <T>
	 *            a type extending the primary node type.
	 * @param <S>
	 *            a type extending the primary node type.
	 * 
	 * @return a list of all paths between the provided source and sink nodes.
	 */
	public <T extends N, S extends N> List<Path<N>> getPaths(T source, S sink) {

		// Checking source node
		if (!graph.isNode(source)) {
			String s = "Node does not exist: " + source;
			throw new IllegalArgumentException(s);
		}

		// Checking sink node
		if (!graph.isNode(sink)) {
			String s = "Node does not exist: " + sink;
			throw new IllegalArgumentException(s);
		}

		// Creating storage for all paths
		List<Path<N>> paths = new LinkedList<Path<N>>();

		// Starting at Step 1
		int step = 1;

		// While the path-finding algorithm still has work to do
		while (step != 0) {

			switch (step) {

			// Doing Step 1
			case 1:
				step = doStep1(source, sink);
				break;

			// Doing Step 2
			case 2:
				step = doStep2();
				break;

			// Doing Step 3
			case 3:
				step = doStep3(paths);
				break;
			}
		}

		// Returning all paths between the provided source and sink nodes
		return paths;
	}

	/**
	 * Returns the number of paths between the provided source and sink nodes.
	 * 
	 * @param source
	 *            the node at which all computed paths will begin.
	 * @param sink
	 *            the node at which all computed paths will end.
	 * @param <T>
	 *            a type extending the primary node type.
	 * @param <S>
	 *            a type extending the primary node type.
	 * 
	 * @return the number of paths between the provided source and sink nodes.
	 */
	public <T extends N, S extends N> int paths(T source, S sink) {

		// Getting all the paths between the source and sink nodes
		Collection<Path<N>> paths = getPaths(source, sink);

		// Returning the number of computed paths
		return paths.size();
	}

	/**
	 * Returns the total number of paths.
	 * 
	 * @return the total number of paths.
	 */
	public int paths() {

		// Getting all nodes in the graph
		Set<N> nodes = graph.getNodes();

		// Returning the total number of paths in the graph
		return paths(nodes, nodes);
	}

	/**
	 * Returns the number of paths between the provided source and sink nodes.
	 * 
	 * @param sources
	 *            the collection of nodes at which all computed paths will
	 *            begin.
	 * 
	 * @param sinks
	 *            the collection of nodes at which all computed paths will end.
	 * 
	 * @return the number of paths between the provided source and sink nodes.
	 */
	public int paths(Collection<? extends N> sources,
			Collection<? extends N> sinks) {

		// Initializing path counter
		int p = 0;

		// Iterating thru each source node
		for (N source : sources) {

			// Iterating thru each sink node
			for (N sink : sinks) {

				// Getting the number of paths between the two nodes
				// and updating the count
				p += paths(source, sink);
			}
		}

		// Returning the number of computed paths
		return p;
	}

	
	// Note: The following section should be cleaned-up considerably
	
	// Initialization of variables required for the path-finding algorithm
	private N v;

	private N q;

	private Set<N> c;

	private Set<N> d;

	private LinkedList<N> w;

	private LinkedList<N> T;

	private Map<N, LinkedList<N>> S;

	private int doStep1(N source, N sink) {

		// This is the Step 1 of the path finding algorithm.

		// Creating storage
		S = new HashMap<N, LinkedList<N>>();
		T = new LinkedList<N>();
//		c = new LinkedList<N>();
//		d = new LinkedList<N>();
		c = new HashSet<N>();
		d = new HashSet<N>();
		w = new LinkedList<N>();

		// Initialzing values
		v = source;
		q = sink;

		w.add(v);

		// Go to Step 2 from here
		return 2;
	}

	private int doStep2() {

		// STEP 2.1
		T.clear();
		c.clear();
		d.clear();

		c.addAll(w);

		// STEP 2.2
		T.add(q);
		c.add(q);
		d.add(q);

		// STEP 2.3
		while (!T.isEmpty()) { // STEP 2.4

			N r = T.removeLast();

			Set<N> predecessors = graph.getPredecessors(r);

			for (N pred : predecessors) {

				if (!c.contains(pred)) {
					T.add(pred);
					c.add(pred);
					d.add(pred);
				}
			}
		}

		// STEP 2.5
		LinkedList<N> successors = new LinkedList<N>();

		for (N succ : graph.getSuccessors(v)) {

			if (d.contains(succ)) {
				successors.add(succ);
			}
		}

		S.put(v, successors);

		// Go to Step 3 from here
		return 3;
	}

	private int doStep3(List<Path<N>> paths) {

		LinkedList<N> vSuccessors = S.get(v);

		if (vSuccessors.isEmpty()) {

			// STEP 5
			w.removeLast();

			if (!w.isEmpty()) {

				v = w.getLast();

				// Go to Step 3 from here
				return 3;
			}

			// All paths have been found
			return 0;

			// END STEP 5
		}

		N k = vSuccessors.removeLast();

		// STEP 4

		// If the current node is the path terminus, the path is found
		if (k.equals(q)) {

			// Copying the list of nodes
			List<N> nodes = new ArrayList<N>(w.size() + 1);
			nodes.addAll(w);
			nodes.add(k);
			
			// Creating a Path
			Path<N> p = new Path<N>(nodes);
			
			// Saving the path
			paths.add(p);
			
			// Go to Step 3 from here
			return 3;

		} 

		w.add(k);
		v = k;

		// Go to Step 2 from here
		return 2;

		// END STEP 4
	}

}
