/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.graph_theory.proc;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.graph_theory.DirectedGraph;
import sbrt.kernel.math.graph_theory.Path;
import sbrt.kernel.math.graph_theory.DirectedGraph.Edge;
import sbrt.kernel.math.graph_theory.comp.PathFinder;
import sbrt.kernel.utilities.MutableProgressMonitor;
import sbrt.kernel.utilities.PercentProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;

/**
 * This class is used to identify all the simple cycles in a directed graph.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 * @param <N> the node type.
 */
public class CycleIdProcess<N> implements MonitoredProcess<Path<N>> {

	private final PercentProgressMonitor progress;
    private DirectedGraph<N> graph;

    /**
     * Constructs a new cycle-identification process.
     */
    public CycleIdProcess() {
    	progress = new PercentProgressMonitor();
    }
    
    /**
     * Sets the graph for which all cycles will be identified.
     * @param graph the graph for which all paths will be identified. 
     */
    public void setGraph(DirectedGraph<N> graph) {
    	Util.nullCheck(graph);
    	this.graph = graph;
    }
    
    private void checkState() {
    	
    	if (graph == null) {
    		String s = "No graph was provided.";
    		throw new IllegalStateException(s);
    	}
    	
    }

    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    /**
     * Adds to the provided results manager each identified cycle in the current
     * graph.
     * 
     * @param resultsManager the results manager to which the cycles are added.
     */
    public void run(ResultsManager<Path<N>> resultsManager) {
    	
        checkState();

        try {
        	
        	// Initializing progress monitor
            progress.resetCount();
            progress.start();

            // Running the process
            runProcess(resultsManager);

            // Finalizing the progress monitor
            progress.stop();
            
        } catch (RuntimeException e) {

        	// Finalizing the progress monitor
            if (!progress.stopped()) {
                progress.stop();
            }
            
            // Throwing the original exception
            throw e;
        }
        
    }
    
    private void runProcess(ResultsManager<Path<N>> resultsManager) {

    	// Getting the number of nodes
    	int n = graph.nodes();
    	
    	// Setting total for progress monitor equalt to # nodes + 1
    	progress.setTotal(n + 1);
    	
    	// Identifying all cycles
    	computeResults(progress, resultsManager);

    	// Signalling completion
    	progress.complete();
    	
    }
	
	private void computeResults(MutableProgressMonitor pm,
								ResultsManager<Path<N>> resultsManager) {
		
		// Initializing a temporary graph
		DirectedGraph<N> tempGraph = graph;

		// Iterating thru each node
		for (N n : graph.getNodes()) {
			
			// Getting all cycles in which this node participates
			List<Path<N>> cycles = getCycles(n, tempGraph);

			// Adding these cycles to the rest
			recordResults(cycles, resultsManager);

			// Allowing garbage collection
			cycles = null;
			
			// Removing this node from the temporary graph
			tempGraph = removeNode(n, tempGraph);

			pm.increment();
		}
		
	}
	
	
	private void recordResults(List<Path<N>> cycles, ResultsManager<Path<N>> rm) {
		
		for (Path<N> p : cycles) {
			rm.addData(p);
		}
		
	}
	
	
	private List<Path<N>> getCycles(N node, DirectedGraph<N> graph) {
		
		// If the node no longer exists in the graph
		// (because all of its edges have been removed)
		if (!graph.isNode(node)) {
			return Collections.emptyList();
		}
		
		PathFinder<N> pathFinder = new PathFinder<N>(graph);
		
		return pathFinder.getPaths(node, node);
	}
	
	
	private DirectedGraph<N> removeNode(N node, DirectedGraph<N> dg) {
		
		// Getting the original set of edges
		Set<Edge<N>> edges = dg.getEdges();
		
		// Creating storage for a new set of edges
		int size = edges.size() + 1;
		
		Set<Edge<N>> newEdges = new LinkedHashSet<Edge<N>>(size, 1f);
		
		for (Edge<N> e : edges) {
			
			// Omitting edges that contain the provided node
			if (!e.hasNode(node)) {
				newEdges.add(e);
			}
			
		}

		return new DirectedGraph<N>(newEdges);
	}
	
}
