/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.graph_theory.proc;

import java.util.*;

import sbrt.kernel.math.graph_theory.DirectedGraph;
import sbrt.kernel.math.graph_theory.Path;
import sbrt.kernel.math.graph_theory.comp.PathFinder;
import sbrt.kernel.utilities.PercentProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;
import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.io.ResultsManager;

/**
 * This class is used to identify all the simple paths in a directed graph.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 * @param <N> the node type.
 */
public class PathIdProcess<N> implements MonitoredProcess<Path<N>> {

	private final PercentProgressMonitor progress;
    private DirectedGraph<N> graph;

    /**
     * Constructs a new path identification process.
     */
    public PathIdProcess() {
    	progress = new PercentProgressMonitor();
    }
    
    /**
     * Sets the graph for which all paths will be identified.
     * @param graph the graph for which all paths will be identified. 
     */
    public void setGraph(DirectedGraph<N> graph) {
    	Util.nullCheck(graph);
    	this.graph = graph;
    }
    
    private void checkState() {
    	
    	if (graph == null) {
    		String s = "The graph is null.";
    		throw new IllegalStateException(s);
    	}
    	
    }

    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

    /**
     * Adds to the provided results manager each computed path in the current
     * graph.
     * 
     * @param resultsManager the results manager to which the paths are added.
     */
    public void run(ResultsManager<Path<N>> resultsManager) {
    	
        checkState();

        try {
        	
        	// Initialized progress monitor
            progress.resetCount();
            progress.start();

            // Running the process
            runProcess(resultsManager);

            // Finalizing the progress monitor
            progress.stop();
            
        } catch (RuntimeException e) {

        	// Finalizing the progress monitor
            if (!progress.stopped()) {
                progress.stop();
            }
            
            // Throwing the original exception
            throw e;
        }
        
    }
    
    private void runProcess(ResultsManager<Path<N>> resultsManager) {

    	// Getting the number of nodes
    	int n = graph.nodes();
    	
    	// Setting total for progress monitor
    	progress.setTotal(n * n + 1);
    	
    	// Creating path finder
    	PathFinder<N> pf = new PathFinder<N>(graph);
    	
    	// Getting all nodes in the graph as a list
    	ArrayList<N> nodes = new ArrayList<N>(graph.getNodes());
    	nodes.trimToSize();
    	
    	// Iterating thru each node
    	for (int i = 0; i < n; i ++) {
    		
    		// Getting the source node
    		N source = nodes.get(i);
    		
    		for (int j = 0; j < n; j ++) {
    			
    			// Ignoring cycles
    			if (i == j) {
    				continue;
    			}
    			
    			// Getting the sink node
    			N sink = nodes.get(j);

    			// Identifying all paths between them 
    			computeResults(source, sink, pf, resultsManager);
    		}
    	}

    	// Signalling completion
    	progress.complete();
    	
    }
    
    private void computeResults(N source, 
    							N sink, 
    							PathFinder<N> pathFinder, 
    							ResultsManager<Path<N>> resultsManager) {
		
		// Getting all paths between the source and sink
		List<Path<N>> paths = pathFinder.getPaths(source, sink);
    	
        // Adding to the results manager
    	for (Path<N> p : paths) {
            resultsManager.addData(p);
    	}
    	
		// Incrementing progress
		progress.increment();
    }

}
