/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.optimization;

import java.util.*;

import sbrt.kernel.math.*;
//import sbrt.kernel.math.lin_alg.*;

/**
 * This interface is used to represent program solvers, as in linear
 * or quadratic program solvers.
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <R> the row variable type.
 * @param <C> the column variable type.
 */
public interface ProgramSolver<R, C> {

	/**
	 * Clears all information contained in this program solver. Calling 
	 * this method restores this program solver to its original, pristine 
	 * state.
	 */
    public void clear();

    // Methods for defining the system

    /**
     * Sets the element of the program matrix at the provided row and 
     * column variables to the one provided.
     * 
     * @param <X> a type extending the primary row variable type.
     * @param <Y> a type extending the primary column variable type.
     * 
     * @param row the row variable.
     * @param column the column variable.
     * @param value the double precision element value.
     */
    public <X extends R, Y extends C> void 
    	setMatrixElement(X row, Y column, double value);
 
    /**
     * Sets the constraint on the provided column variable.
     * 
     * @param <Y> a type extending the primary column variable type.
     * @param column the column variable for which the constraint is set.
     * @param c the constraint.
     */
    public <Y extends C> void setColumnConstraint(Y column, Interval c);

    /**
     * Sets the constraints for the provided column variables.
     * 
     * @param columnConstraints column variables mapped to their corresponding
     * constraint.
     */
    public void
    	setColumnConstraints(Map<? extends C, Interval> columnConstraints);

    /**
     * Sets the constraint for the provided row variable.
     * 
     * @param <X> a type extending the primary row variable type.
     * @param row the row variable for which the constraint is set.
     * @param c the constraint.
     */
    public <X extends R> void setRowConstraint(X row, Interval c);

    /**
     * Sets the constraints for the provided row variables.
     * 
     * @param rowConstraints row variables mapped to their corresponding
     * constraint.
     */
    public void
    	setRowConstraints(Map<? extends R, Interval> rowConstraints);

    /**
     * Sets a constraint on the provided mathematical expression of
     * column variables.
     * 
     * @param expression a mathematical expression of column variables for
     * which the constraint is set.
     * @param c the constraint.
     */
    public void 
    	setConstraint(MathExpr<? extends C> expression, Interval c);

    /**
     * Sets constraints on the provided mathematical expressions of
     * column variables.
     * 
     * @param expressionConstraints mathematical expressions of column variables
     * mapped to their corresponding constraint.
     */
    public void setConstraints(Map<? extends MathExpr<? extends C>, 
    							Interval> expressionConstraints);

    /**
     * Sets the objective function and optimization sense to the provided
     * values.
     * 
     * @param expression a mathematical expression of column variables from
     * this program solver.
     * @param sense the optimization sense.
     * 
     * @see #setObjectiveFunction(MathExpr)
     * @see #setOptimizationSense(OptSense)
     */
    public void setObjectiveFunction(MathExpr<? extends C> expression,
    						 		 OptSense sense);

    /**
     * Sets the objective function to the provided mathematical expression of
     * column variables.
     * 
     * @param expression a mathematical expression of column variables from
     * this program solver.
     */
    public void setObjectiveFunction(MathExpr<? extends C> expression);

    /**
     * Sets the optimization sense.
     * @param sense the optimization sense.
     */
    public void setOptimizationSense(OptSense sense);

    /**
     * Computes the optimal value of the current objective function in the 
     * direction of the current optimization sense.
     * 
     * @return {@code true} if an optimal solution was found; 
     * {@code false} otherwise.
     * 
     * @see #getObjectiveValue()
     * @see #getSolutionVector()
     */
    public boolean optimize();
    
    /**
     * Returns the most recently computed objective value.
     * @return the most recently computed objective value.
     * @see #optimize()
     */
    public double getObjectiveValue();
    
    /**
     * Indicates if the provided object is a column variable in the matrix of
     * this program solver.
     * 
     * @param obj the potential column variable.
     * 
     * @return {@code true} if the provided object is a column variable in the
     * matrix of this program solver; {@code false} otherwise.
     */
    public boolean isColumnVar(Object obj);
    
    /**
     * Indicates if the provided object is a row variable in the matrix of
     * this program solver.
     * 
     * @param obj the potential row variable.
     * 
     * @return {@code true} if the provided object is a row variable in the
     * matrix of this program solver; {@code false} otherwise.
     */
    public boolean isRowVar(Object obj);
    
    /**
     * Returns the most recently computed solution vector.
     * @return column variables mapped to their corresponding numerical values.
     * @see #optimize()
     */
    public Map<C, Double> getSolutionVector();

    /**
     * Returns the status of the most recently attempted optimization.
     * @return the status of the most recently attempted optimization.
     * @see #optimize()
     */
    public ProgramSolverStatus getStatus();
    
    /**
     * Reads the program solver parameter file with the provided name.
     * 
     * @param fileName the name of the file containing parameter values.
     * 
     * @throws UnsupportedOperationException if this program solver does
     * offer this functionality.
     */
    public void setParamFile(String fileName);

    /**
     * Returns the current program matrix.
     * @return the current program matrix.
     */
//    public DoubleMatrix<R, C> getMatrix();
}
