/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.stat;

/**
 * This class is used to represent normal distributions.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 * 
 * 
 */
class NormalDist {

//    public static final double DEFAULT_CUTOFF = 1E-32;

    private final double average;
    private final double stdev;

    /**
     * Constructs a new normal distribution with the provided average
     * and standard deviation.
     * 
     * @param average the mean of the distribution.
     * @param stdev the standard deviation of the distribution.
     */
    public NormalDist(double average, double stdev) {

        if (!isValid(average)) {
            String s = "Invalid average: " + average;
            throw new IllegalArgumentException(s);
        }

        if (!isValid(stdev)) {
            String s = "Invalid stdev: " + stdev;
            throw new IllegalArgumentException(s);
        }

        this.average = average;
        this.stdev = stdev;
    }

    private boolean isValid(double x) {
        return !Double.isInfinite(x) && !Double.isNaN(x);
    }

    /**
     * Returns the mean, or average, of this distribution.
     * @return the mean, or average, of this distribution.
     */
    public double getMean() {
        return average;
    }

    /**
     * Returns the standard deviation of this distribution.
     * @return the standard deviation of this distribution.
     */
    public double getStdev() {
        return stdev;
    }

    /**
     * Returns the variance of this distribution.
     * @return the variance of this distribution.
     */
    public double getVariance() {
        return stdev * stdev;
    }

    
    /*
    public double getAreaBetween(double xLower, double xUpper) {
        return getAreaBetween(xLower, xUpper, DEFAULT_CUTOFF);
    }

    public double getAreaBetween(double xLower, double xUpper, double cutoff) {

        if (xLower > xUpper) {
            String s = "xLower > xUpper: " + xLower + " > " + xUpper;
            throw new IllegalArgumentException(s);
        }

        return getLeftTailArea(xUpper, cutoff / 2) -
                getLeftTailArea(xLower, cutoff / 2);
    }

    public double getRightTailArea(double x) {
        return getRightTailArea(x, DEFAULT_CUTOFF);
    }

    public double getRightTailArea(double x, double cutoff) {
        return 1 - getLeftTailArea(x, cutoff);
    }

    public double getLeftTailArea(double x) {
        return getLeftTailArea(x, DEFAULT_CUTOFF);
    }

    public double getLeftTailArea(double x, double cutoff) {

        if (!isValid(x)) {
            String s = "Invalid x: " + x;
            throw new IllegalArgumentException(s);
        }

        double erf = getErf(x, cutoff);
        double p = 0.5 * (1.0 + erf);

        System.out.println("erf = " + erf);
//        System.out.println("p = " + p);
        System.out.println();

        return p;
    }

    private double getErf(double x) {
        return getErf(x, DEFAULT_CUTOFF);
    }

    private double getErf(double x, double cutoff) {

//        System.out.println("x = " + x);

//        double z = (x - average) / (stdev * Math.sqrt(2));

        double diff = Double.POSITIVE_INFINITY;
        double sum = 0.0;

        for (int n = 0; Math.abs(diff) > cutoff; n++) {

            diff = Math.pow(-1, n);
            diff = diff * Math.pow(x, 2 * n + 1);
//            diff = diff / Util.factorial(n).longValue();
            diff = diff / (2 * n + 1);

            sum += diff;

            System.out.println("diff = " + diff);
            System.out.println(n + "! = " + Util.factorial(n));
            System.out.println("sum = " + sum);
            System.out.println();

        }

        System.out.println("diff = " + diff);

        return (2 / Math.sqrt(Math.PI)) * sum;
    }

    public static void main(String[] args) {

        int errors = 0;

        NormalDist nd = new NormalDist(0, 1);

        for (double x = 0; x < 3; x += 0.1) {
            System.out.println("x = " + x + ", erf = " + nd.getErf(x));
        }

        System.exit(0);

        double p = nd.getLeftTailArea(0);
        double trueP = 0.5;
        if (!Util.equal(p, trueP, Double.MIN_VALUE)) {
            errors++;
            System.err.println("Computed P = " + p + ", True P = " + trueP);
        }

        p = nd.getAreaBetween(0, 0);
        trueP = 0;
        if (!Util.equal(p, trueP, Double.MIN_VALUE)) {
            errors++;
            System.err.println("Computed P = " + p + ", True P = " + trueP);
        }

        p = nd.getAreaBetween( -1.0, 1.0);
        trueP = .6826894921371;
        if (!Util.equal(p, trueP, 1E-13)) {
            errors++;
            System.err.println("Computed P = " + p + ", True P = " + trueP);
        }

        p = nd.getAreaBetween( -2, 2);
        trueP = .9544997361036;
        if (!Util.equal(p, trueP, 1E-13)) {
            errors++;
            System.err.println("Computed P = " + p + ", True P = " + trueP);
        }

        p = nd.getAreaBetween( -3, 3);
        trueP = .6826894921371;
        if (!Util.equal(p, trueP, 1E-13)) {
            errors++;
            System.err.println("Computed P = " + p + ", True P = " + trueP);
        }

        p = nd.getAreaBetween( -4, 4);
        trueP = .9999366575163;
        if (!Util.equal(p, trueP, 1E-13)) {
            errors++;
            System.err.println("Computed P = " + p + ", True P = " + trueP);
        }

        p = nd.getAreaBetween( -5, 5);
        trueP = .9999994266969;
        if (!Util.equal(p, trueP, 1E-13)) {
            errors++;
            System.err.println("Computed P = " + p + ", True P = " + trueP);
        }

        p = nd.getAreaBetween( -6, 6);
        trueP = .9999999980268;
        if (!Util.equal(p, trueP, 1E-13)) {
            errors++;
            System.err.println("Computed P = " + p + ", True P = " + trueP);
        }

        p = nd.getAreaBetween( -7, 7);
        trueP = .9999999999974;
        if (!Util.equal(p, trueP, 1E-13)) {
            errors++;
            System.err.println("Computed P = " + p + ", True P = " + trueP);
        }

        if (errors > 0) {
            System.err.println("Errors = " + errors);
            System.exit(1);
        }
    }

    */
}
