/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.stat.comp;

import java.util.*;

import sbrt.kernel.utilities.*;

/**
 * This abstract class is a skeleton implementation of a Kendall's tau
 * correlation statistic computer. It is implemented using algorithms from the
 * book <i>Rank Correlation Methods, 5th Edition</i> by Kendall & Gibbons.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <V> the variable type.
 */
public abstract class KendallTau<V> implements Correlator<V, TauResult> {

    /**
     * Computes and returns the correlation statistic for the provided values.
     * 
     * @param xValues an ordered array of values.
     * @param yValues an ordered array of values.
     * 
     * @return the correlation statistic for the provided values.
     */
    public abstract double getTau(double[] xValues, double[] yValues);

    /**
     * Computes and returns the correlation statistic for the provided values.
     * The key sets of the provided maps need not be identical, but only the
     * intersecting values will be considered.
     * 
     * @param xValues variables mapped to their value.
     * @param yValues variables mapped to their value.
     * 
     * @return the correlation statistic for the provided values.
     */
    public double getTau(Map<? extends V, Double> xValues,
            Map<? extends V, Double> yValues) {

        double[][] values = Util.getArrays(xValues, yValues);

        return getTau(values[0], values[1]);
    }

    /**
     * Computes and returns the correlation statistic for the provided values.
     * 
     * @param xValues an ordered collection of values.
     * @param yValues an ordered collection of values.
     * 
     * @return the correlation statistic for the provided values.
     */
    public double getTau(Collection<Double> xValues, Collection<Double> yValues) {
        return getTau(Util.getArray(xValues), Util.getArray(yValues));
    }

    /**
     * Computes and returns the correlation results for the provided values. The
     * key sets of the provided maps need not be identical, but only the
     * intersecting values will be considered.
     * 
     * @param xValues variables mapped to their value.
     * @param yValues variables mapped to their value.
     * 
     * @return the correlation results for the provided values.
     */
    public TauResult corr(Map<? extends V, Double> xValues,
            Map<? extends V, Double> yValues) {

        double[][] values = Util.getArrays(xValues, yValues);

        return corr(values[0], values[1]);
    }

    /**
     * Computes and returns the correlation results for the provided values.
     * 
     * @param xValues an ordered collection of values.
     * @param yValues an ordered collection of values.
     * 
     * @return the correlation results for the provided values.
     */
    public TauResult corr(Collection<Double> xValues, Collection<Double> yValues) {
        return corr(Util.getArray(xValues), Util.getArray(yValues));
    }

    /**
     * Computes and returns the correlation results for the provided values.
     * 
     * @param xValues an ordered array of values.
     * @param yValues an ordered array of values.
     * 
     * @return the correlation results for the provided values.
     */
    public TauResult corr(Double[] xValues, Double[] yValues) {
        return corr(Arrays.asList(xValues), Arrays.asList(yValues));
    }

    /**
     * Computes and returns the correlation results for the provided values.
     * 
     * @param xValues an ordered array of values.
     * @param yValues an ordered array of values.
     * 
     * @return the correlation results for the provided values.
     */
    public TauResult corr(double[] xValues, double[] yValues) {

        double tau = getTau(xValues, yValues);
        double z = getApproxZ(xValues, yValues);
        int n = xValues.length;

        return new TauResult(tau, z, n);
    }

    /**
     * Computes and returns the S value of the provided values. The key sets of
     * the provided maps need not be identical, but only the intersecting values
     * will be considered.
     * 
     * @param xValues variables mapped to their value.
     * @param yValues variables mapped to their value.
     * 
     * @return the S value of the provided values.
     */
    public int getS(Map<? extends V, Double> xValues,
            Map<? extends V, Double> yValues) {

        double[][] values = Util.getArrays(xValues, yValues);

        return getS(values[0], values[1]);
    }

    /**
     * Computes and returns the S value of the provided values.
     * 
     * @param xValues an ordered collection of values.
     * @param yValues an ordered collection of values.
     * 
     * @return the S value of the provided values.
     */
    public int getS(Collection<Double> xValues, Collection<Double> yValues) {
        return getS(Util.getArray(xValues), Util.getArray(yValues));
    }

    /**
     * Computes and returns the S value of the provided values.
     * 
     * @param xValues an ordered array of values.
     * @param yValues an ordered array of values.
     * 
     * @return the S value of the provided values.
     */
    public int getS(double[] xValues, double[] yValues) {

        int n = xValues.length;

        if (n != yValues.length) {
            String s = "Arrays not the same size.";
            throw new IllegalArgumentException(s);
        }

        int s = 0;

        // Iterating through each value
        for (int i = 0; i < n; i++) {

            double currentX = xValues[i];
            double currentY = yValues[i];

            // Iterating through all succeeding values
            for (int j = i + 1; j < n; j++) {

                // Comparing values at i with those at j
                int xComparison = Double.compare(currentX, xValues[j]);
                int yComparison = Double.compare(currentY, yValues[j]);

                // Computing rank order score
                int rankScore = xComparison * yComparison;

                // If rank order for both pairs is the same
                if (rankScore > 0) {
                    s += 1;
                }

                // If rank order for both pairs is different
                if (rankScore < 0) {
                    s -= 1;
                }
            }
        }

        return s;
    }

    /**
     * Computes and returns the tie score of the provided values.
     * 
     * @param values variables mapped to their value.
     * 
     * @return the tie score of the provided values.
     */
    public double getTieScore(Map<? extends V, Double> values) {
        return getTieScore(values.values());
    }

    /**
     * Computes and returns the tie score of the provided values.
     * 
     * @param values a collection of values.
     * 
     * @return the tie score of the provided values.
     */
    public double getTieScore(Collection<Double> values) {
        return getTieScore(Util.getArray(values));
    }

    /**
     * Computes and returns the tie score of the provided values.
     * 
     * @param values an array of values.
     * 
     * @return the tie score of the provided values.
     */
    public double getTieScore(double[] values) {

        int n = values.length;

        // Getting new array of sorted values
        double[] sortedValues = Util.getSorted(values);

        double score = 0.0;

        for (int i = 0; i < n; i++) {

            double value = sortedValues[i];
            int tieCount = 1;

            while (i + 1 < n && Double.compare(sortedValues[i + 1], value) == 0) {
                tieCount++;
                i++;
            }

            score += tieCount * (tieCount - 1);
        }

        return score / 2;
    }

    /**
     * Computes and returns a Z approximation for the provided values. The key
     * sets of the provided maps need not be identical, but only the
     * intersecting values will be considered.
     * 
     * @param xValues variables mapped to their value.
     * @param yValues variables mapped to their value.
     * 
     * @return a Z approximation for the provided values.
     */
    public double getApproxZ(Map<? extends V, Double> xValues,
            Map<? extends V, Double> yValues) {

        double[][] values = Util.getArrays(xValues, yValues);

        return getApproxZ(values[0], values[1]);
    }

    /**
     * Computes and returns a Z approximation for the provided values.
     * 
     * @param xValues an ordered collection of values.
     * @param yValues an ordered collection of values.
     * 
     * @return a Z approximation for the provided values.
     */
    public double getApproxZ(Collection<Double> xValues,
            Collection<Double> yValues) {
        return getApproxZ(Util.getArray(xValues), Util.getArray(yValues));
    }

    /**
     * Computes and returns a Z approximation for the provided values.
     * 
     * @param xValues an ordered array of values.
     * @param yValues an ordered array of values.
     * 
     * @return a Z approximation for the provided values.
     */
    public double getApproxZ(double[] xValues, double[] yValues) {

        int n = xValues.length;
        if (n != yValues.length) {
            String s = "Arrays not the same size.";
            throw new IllegalArgumentException(s);
        }

        // Getting S
        int s = getS(xValues, yValues);

        // Applying continuity correction
        if (s > 0) {
            s = s - 1;
        } else if (s < 0) {
            s = s + 1;
        }

        // Getting variance of S
        double varS = getVarS(xValues, yValues);

        // Returning Z
        return s / Math.sqrt(varS);
    }

    /**
     * Computes and returns the variance of S for the provided values. The key
     * sets of the provided maps need not be identical, but only the
     * intersecting values will be considered.
     * 
     * @param xValues variables mapped to their value.
     * @param yValues variables mapped to their value.
     * 
     * @return the variance of S for the provided values.
     */
    public double getVarS(Map<? extends V, Double> xValues,
            Map<? extends V, Double> yValues) {

        double[][] values = Util.getArrays(xValues, yValues);

        return getVarS(values[0], values[1]);
    }

    /**
     * Computes and returns the variance of S for the provided values.
     * 
     * @param xValues an ordered collection of values.
     * @param yValues an ordered collection of values.
     * 
     * @return the variance of S for the provided values.
     */
    public double getVarS(Collection<Double> xValues, Collection<Double> yValues) {
        return getVarS(Util.getArray(xValues), Util.getArray(yValues));
    }

    /**
     * Computes and returns the variance of S for the provided values.
     * 
     * @param xValues an ordered array of values.
     * @param yValues an ordered array of values.
     * 
     * @return the variance of S for the provided values.
     */
    public double getVarS(double[] xValues, double[] yValues) {

        int n = xValues.length;
        if (n != yValues.length) {
            String s = "Arrays not the same size.";
            throw new IllegalArgumentException(s);
        }

        double[] xSorted = Util.getSorted(xValues);
        double[] ySorted = Util.getSorted(yValues);

        double firstTerm = n * (n - 1) * (2 * n + 5);
        firstTerm = firstTerm - get2ndPartOf1stTerm(xSorted);
        firstTerm = firstTerm - get2ndPartOf1stTerm(ySorted);
        firstTerm = (1 / 18.0) * firstTerm;

        double secondTerm = 1.0 / (9 * n * (n - 1) * (n - 2));
        secondTerm = secondTerm * get2ndPartOf2ndTerm(xSorted);
        secondTerm = secondTerm * get2ndPartOf2ndTerm(ySorted);

        double thirdTerm = 1.0 / (2 * n * (n - 1));
        thirdTerm = thirdTerm * get2ndPartOf3rdTerm(xSorted);
        thirdTerm = thirdTerm * get2ndPartOf3rdTerm(ySorted);

        return firstTerm + secondTerm + thirdTerm;
    }

    private double get2ndPartOf1stTerm(double[] sortedValues) {

        int n = sortedValues.length;

        double sum = 0.0;

        for (int i = 0; i < n; i++) {

            double value = sortedValues[i];
            int tieCount = 1;

            while (i + 1 < n && Double.compare(sortedValues[i + 1], value) == 0) {
                tieCount++;
                i++;
            }

            sum += tieCount * (tieCount - 1) * (2 * tieCount + 5);
        }

        return sum;
    }

    private double get2ndPartOf2ndTerm(double[] sortedValues) {

        int n = sortedValues.length;

        double sum = 0.0;

        for (int i = 0; i < n; i++) {

            double value = sortedValues[i];
            int tieCount = 1;

            while (i + 1 < n && Double.compare(sortedValues[i + 1], value) == 0) {
                tieCount++;
                i++;
            }

            sum += tieCount * (tieCount - 1) * (tieCount - 2);
        }

        return sum;
    }

    private double get2ndPartOf3rdTerm(double[] sortedValues) {

        int n = sortedValues.length;

        double sum = 0.0;

        for (int i = 0; i < n; i++) {

            double value = sortedValues[i];
            int tieCount = 1;

            while (i + 1 < n && Double.compare(sortedValues[i + 1], value) == 0) {
                tieCount++;
                i++;
            }

            sum += tieCount * (tieCount - 1);
        }

        return sum;
    }

    /**
     * Tests the Kendall Tau A and B computers.
     * 
     * @param args not required.
     */
    public static void main(String[] args) {

        double tolerance = 1E-9;
        int errors = 0;

        KendallTauA<Object> kTauA = new KendallTauA<Object>();
        KendallTauB<Object> kTauB = new KendallTauB<Object>();

        // Examples from Rank Correlation Methods, 5th Edition
        // by Kendall & Gibbons
        // pg. 3-4
        Map<String, Double> mathScores = new TreeMap<String, Double>();
        mathScores.put("A", Double.valueOf(7));
        mathScores.put("B", Double.valueOf(4));
        mathScores.put("C", Double.valueOf(3));
        mathScores.put("D", Double.valueOf(10));
        mathScores.put("E", Double.valueOf(6));
        mathScores.put("F", Double.valueOf(2));
        mathScores.put("G", Double.valueOf(9));
        mathScores.put("H", Double.valueOf(8));
        mathScores.put("I", Double.valueOf(1));
        mathScores.put("J", Double.valueOf(5));

        Map<String, Double> musicScores = new TreeMap<String, Double>();
        musicScores.put("A", Double.valueOf(5));
        musicScores.put("B", Double.valueOf(7));
        musicScores.put("C", Double.valueOf(3));
        musicScores.put("D", Double.valueOf(10));
        musicScores.put("E", Double.valueOf(1));
        musicScores.put("F", Double.valueOf(9));
        musicScores.put("G", Double.valueOf(6));
        musicScores.put("H", Double.valueOf(2));
        musicScores.put("I", Double.valueOf(8));
        musicScores.put("J", Double.valueOf(4));

        double tauA = kTauA.getTau(mathScores, musicScores);
        double tauB = kTauB.getTau(mathScores, musicScores);

        double trueTauA = -3.0 / 45.0;
        double trueTauB = -3.0 / 45.0;

        if (!Util.equal(tauA, trueTauA, tolerance)) {
            errors++;
            System.err.println("True Tau A = "
                    + trueTauA
                    + ", Computed Tau A = "
                    + tauA);
            System.err.println("Computed S = "
                    + kTauA.getS(mathScores, musicScores));
            System.err.println("U = " + kTauA.getTieScore(mathScores));
            System.err.println("V = " + kTauA.getTieScore(musicScores));
        }

        if (!Util.equal(tauB, trueTauB, tolerance)) {
            errors++;
            System.err.println("True Tau B = "
                    + trueTauB
                    + ", Computed Tau B = "
                    + tauB);
            System.err.println("Computed S = "
                    + kTauB.getS(mathScores, musicScores));
            System.err.println("U = " + kTauB.getTieScore(mathScores));
            System.err.println("V = " + kTauB.getTieScore(musicScores));
        }

        // pg. 41-42
        Collection<Double> x = new LinkedList<Double>();
        x.add(Double.valueOf(1));
        x.add(Double.valueOf(2.5));
        x.add(Double.valueOf(2.5));
        x.add(Double.valueOf(4.5));
        x.add(Double.valueOf(4.5));
        x.add(Double.valueOf(6.5));
        x.add(Double.valueOf(6.5));
        x.add(Double.valueOf(8));
        x.add(Double.valueOf(9.5));
        x.add(Double.valueOf(9.5));

        Collection<Double> y = new LinkedList<Double>();
        y.add(Double.valueOf(1));
        y.add(Double.valueOf(2));
        y.add(Double.valueOf(4.5));
        y.add(Double.valueOf(4.5));
        y.add(Double.valueOf(4.5));
        y.add(Double.valueOf(4.5));
        y.add(Double.valueOf(8));
        y.add(Double.valueOf(8));
        y.add(Double.valueOf(8));
        y.add(Double.valueOf(10));

        tauA = kTauA.getTau(x, y);
        tauB = kTauB.getTau(x, y);

        trueTauA = 33.0 / 45.0;
        trueTauB = 33.0 / Math.sqrt(41 * 36);

        if (!Util.equal(tauA, trueTauA, tolerance)) {
            errors++;
            System.err.println("True Tau A = "
                    + trueTauA
                    + ", Computed Tau A = "
                    + tauA);
            System.err.println("Computed S = " + kTauA.getS(x, y));
            System.err.println("U = " + kTauA.getTieScore(x));
            System.err.println("V = " + kTauA.getTieScore(y));
        }

        if (!Util.equal(tauB, trueTauB, tolerance)) {
            errors++;
            System.err.println("True Tau B = "
                    + trueTauB
                    + ", Computed Tau B = "
                    + tauB);
            System.err.println("Computed S = " + kTauB.getS(x, y));
            System.err.println("U = " + kTauB.getTieScore(x));
            System.err.println("V = " + kTauB.getTieScore(y));
        }

        // pg. 66 - 67
        x.clear();
        x.add(Double.valueOf(1.5));
        x.add(Double.valueOf(1.5));
        x.add(Double.valueOf(3));
        x.add(Double.valueOf(4));
        x.add(Double.valueOf(6));
        x.add(Double.valueOf(6));
        x.add(Double.valueOf(6));
        x.add(Double.valueOf(8));
        x.add(Double.valueOf(9.5));
        x.add(Double.valueOf(9.5));
        x.add(Double.valueOf(11));
        x.add(Double.valueOf(12));

        y.clear();
        y.add(Double.valueOf(2.5));
        y.add(Double.valueOf(2.5));
        y.add(Double.valueOf(7));
        y.add(Double.valueOf(4.5));
        y.add(Double.valueOf(1));
        y.add(Double.valueOf(4.5));
        y.add(Double.valueOf(6));
        y.add(Double.valueOf(11.5));
        y.add(Double.valueOf(11.5));
        y.add(Double.valueOf(8.5));
        y.add(Double.valueOf(8.5));
        y.add(Double.valueOf(10));

        int s = kTauA.getS(x, y);
        if (s != 34) {
            errors++;
            System.err.println("True S = 34, Computed S = " + s);
        }

        double varS = kTauA.getVarS(x, y);
        double trueVarS =
                (1. / 18)
                        * (12 * 11 * 29 - 6 * 2 * 9 - 3 * 2 * 11)
                        + (1. / (2 * 12 * 11))
                        * (2 * 2 + 3 * 2)
                        * (4 * 2);
        if (!Util.equal(varS, trueVarS, 1E-9)) {
            errors++;
            System.err.println("True Var S = "
                    + trueVarS
                    + ", Computed Var S = "
                    + varS);
        }

        double z = kTauA.getApproxZ(x, y);
        double trueZ = 33 / Math.sqrt(trueVarS);

        if (!Util.equal(z, trueZ, 1E-9)) {
            errors++;
            System.err.println("True Z = " + trueZ + ", Computed Z = " + z);
        }

        // Example from Biometry, 3rd Edition by Sokal & Rohlf
        // pg. 594-597
        Map<Integer, Double> y1 = new TreeMap<Integer, Double>();
        y1.put(Integer.valueOf(1), Double.valueOf(8.7));
        y1.put(Integer.valueOf(2), Double.valueOf(8.5));
        y1.put(Integer.valueOf(3), Double.valueOf(9.4));
        y1.put(Integer.valueOf(4), Double.valueOf(10.0));
        y1.put(Integer.valueOf(5), Double.valueOf(6.3));
        y1.put(Integer.valueOf(6), Double.valueOf(7.8));
        y1.put(Integer.valueOf(7), Double.valueOf(11.9));
        y1.put(Integer.valueOf(8), Double.valueOf(6.5));
        y1.put(Integer.valueOf(9), Double.valueOf(6.6));
        y1.put(Integer.valueOf(10), Double.valueOf(10.6));
        y1.put(Integer.valueOf(11), Double.valueOf(10.2));
        y1.put(Integer.valueOf(12), Double.valueOf(7.2));
        y1.put(Integer.valueOf(13), Double.valueOf(8.6));
        y1.put(Integer.valueOf(14), Double.valueOf(11.1));
        y1.put(Integer.valueOf(15), Double.valueOf(11.6));

        Map<Integer, Double> y2 = new TreeMap<Integer, Double>();
        y2.put(Integer.valueOf(1), Double.valueOf(5.95));
        y2.put(Integer.valueOf(2), Double.valueOf(5.65));
        y2.put(Integer.valueOf(3), Double.valueOf(6.00));
        y2.put(Integer.valueOf(4), Double.valueOf(5.70));
        y2.put(Integer.valueOf(5), Double.valueOf(4.70));
        y2.put(Integer.valueOf(6), Double.valueOf(5.53));
        y2.put(Integer.valueOf(7), Double.valueOf(6.40));
        y2.put(Integer.valueOf(8), Double.valueOf(4.18));
        y2.put(Integer.valueOf(9), Double.valueOf(6.15));
        y2.put(Integer.valueOf(10), Double.valueOf(5.93));
        y2.put(Integer.valueOf(11), Double.valueOf(5.70));
        y2.put(Integer.valueOf(12), Double.valueOf(5.68));
        y2.put(Integer.valueOf(13), Double.valueOf(6.13));
        y2.put(Integer.valueOf(14), Double.valueOf(6.30));
        y2.put(Integer.valueOf(15), Double.valueOf(6.03));

        tauA = kTauA.getTau(y1, y2);
        tauB = kTauB.getTau(y1, y2);

        trueTauA = 104.0 / (15.0 * 14.0);
        trueTauB = 104.0 / Math.sqrt(15 * 14 * (15 * 14 - 2));

        if (!Util.equal(tauA, trueTauA, tolerance)) {
            errors++;
            System.err.println("True Tau A = "
                    + trueTauA
                    + ", Computed Tau A = "
                    + tauA);
            System.err.println("Computed S = " + kTauA.getS(x, y));
            System.err.println("U = " + kTauA.getTieScore(x));
            System.err.println("V = " + kTauA.getTieScore(y));
        }

        if (!Util.equal(tauB, trueTauB, tolerance)) {
            errors++;
            System.err.println("True Tau B = "
                    + trueTauB
                    + ", Computed Tau B = "
                    + tauB);
            System.err.println("Computed S = " + kTauB.getS(x, y));
            System.err.println("U = " + kTauB.getTieScore(x));
            System.err.println("V = " + kTauB.getTieScore(y));
        }

        if (errors > 0) {
            System.err.println("Errors: " + errors);
            System.exit(1);
        }

    }

}
