/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.stat.comp;

/**
 * This class is used to store the results of a Mann-Whitney U test.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 *
 */
public class MannWhitneyResults {

    private final double u;
    private final double tApprox;
    private final int n1;
    private final int n2;

    /**
     * Constructs a new set of results for a Mann-Whitney U test.
     * 
     * @param u the computed U statistic.
     * @param tApprox the approximated t value.
     * @param n1 the size of the larger sample.
     * @param n2 the size of the smaller sample.
     */
    public MannWhitneyResults(double u, double tApprox, int n1, int n2) {
        this.u = u;
        this.tApprox = tApprox ;
        this.n1 = n1;
        this.n2 = n2;
    }

    /**
     * Constructs a new set of results for a Mann-Whitney U test.
     * The approximate t value is considered to be {@code Double.NaN};
     * 
     * @param u the computed U statistic.
     * @param n1 the size of the larger sample.
     * @param n2 the size of the smaller sample.
     */
    public MannWhitneyResults(double u, int n1, int n2) {
    	this(u, Double.NaN, n1, n2);
    }
    
    /**
     * Returns the computed U statistic.
     * @return the computed U statistic.
     */
    public double getU() {
        return u;
    }

    /**
     * Returns the t value approximated from the computed U statistic.
     * @return the t value approximated from the computed U statistic.
     */
    public double getApproxT() {
        return tApprox;
    }

    /**
     * Returns the size of the larger sample.
     * @return the size of the larger sample.
     */
    public int getN1() {
        return n1;
    }

    /**
     * Returns the size of the smaller sample.
     * @return the size of the smaller sample.
     */
    public int getN2() {
        return n2;
    }

    /**
     * Returns a string represenation of these results.
     * @return a string represenation of these results.
     */
    public String toString() {
    	
    	if (Double.isNaN(tApprox)) {
    		return "U=" + u + ",n1=" + n1 + ",n2=" + n2;
    	}
    	
        return "U=" + u+ ",t=" + tApprox + ",n1=" + n1 + ",n2=" + n2;
    }

    /**
     * Indicates if the provided object is equal to this results object.
     * 
     * @param obj the object with which to compare.
     * 
     * @return {@code true} if {@code obj} is an instance of 
     * {@code MannWhitneyResults} and its U, t value, and sample sizes
     * are the same as this object's; {@code false} otherwise. 
     */
    public boolean equals(Object obj) {

        if (this == obj) {
            return true;
        }

        if (!(obj instanceof MannWhitneyResults)) {
            return false;
        }

        MannWhitneyResults mwr = (MannWhitneyResults) obj;

        if (n1 != mwr.getN1()) {
            return false;
        }

        if (n2 != mwr.getN2()) {
            return false;
        }

        if (Double.compare(u, mwr.getU()) != 0) {
            return false;
        }

        if (Double.compare(tApprox, mwr.getApproxT()) != 0) {
            return false;
        }

        return true;
    }
    
    /**
     * Returns a content-based hash code.
     * @return a content-based hash code.
     */
    public int hashCode() {
    	return n1 ^ n2;
    }
}
