/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.stat.comp;

import java.util.List;

import sbrt.kernel.utilities.Util;

/**
 * This class is used to compute Mann-Whitney U statistics. The implementation is 
 * based on the algorithm and equations provided on pages 428-430 in
 * <cite>
 * Sokal, Robert and Rohlf, James (1995). Biometry, Third Edition. NY, W. H. 
 * Freeman and Company; ISBN: 0-7167-2411-1.
 * </cite>
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public class MannWhitneyU {
	
	/**
	 * Computes and returns the Mann-Whitney U statistic for the provided samples. 
	 *  
	 * @param sampleA a sample of numbers.
	 * @param sampleB a sample of numbers.
	 * 
	 * @return the Mann-Whitney U statistic for the provided samples.
	 */
	public double getU(List<? extends Number> sampleA, List<? extends Number> sampleB) {

    	// Getting sorted arrays
    	double[] x = Util.getSortedArray(sampleA);
    	double[] y = Util.getSortedArray(sampleB);

    	return getU(x, y);
	}
	
	/**
	 * Computes and returns the Mann-Whitney U statistic for the provided samples. 
	 *  
	 * @param sampleA a sample of integers.
	 * @param sampleB a sample of integers.
	 * 
	 * @return the Mann-Whitney U statistic for the provided samples.
	 */
    public double getU(int[] sampleA, int[] sampleB) {

    	// Getting sorted arrays
    	int[] x = Util.getSorted(sampleA);
    	int[] y = Util.getSorted(sampleB);
    	
    	// Initializing sum
    	double c = 0d;
    	
    	// Iterating thru an array
    	for (int i = 0; i < x.length; i ++) {
    		
    		// Iterating thru the other array
    		for (int j = 0; j < y.length; j ++) {
    			
    			// If the value in y < the value in x
    			if (y[j] < x[i]) {
    				c += 1d;
    			}
    			
    			// If the two values are equal
    			else if (y[j] == x[i]) {
    				c += 0.5d;
    			}
    			
    			// If the value in y has become > than the value in x
    			// break from the loop
    			else {
    				j = y.length;
    			}
    			
    		}
    		
    	}
    	
    	// Computing the quantity n1 * n2 - C
    	double q = x.length * y.length - c;
    	
    	// Returning the greater of the two quanties c and q
    	return c > q ? c : q;
    }
	
	/**
	 * Computes and returns the Mann-Whitney U statistic for the provided samples. 
	 *  
	 * @param sampleA a sample of doubles.
	 * @param sampleB a sample of doubles.
	 * 
	 * @return the Mann-Whitney U statistic for the provided samples.
	 */
    public double getU(double[] sampleA, double[] sampleB) {

    	// Getting sorted arrays
    	double[] x = Util.getSorted(sampleA);
    	double[] y = Util.getSorted(sampleB);
    	
    	// Initializing sum
    	double c = 0d;
    	
    	// Iterating thru an array
    	for (int i = 0; i < x.length; i ++) {
    		
    		// Iterating thru the other array
    		for (int j = 0; j < y.length; j ++) {
    			
    			// If the value in y < the value in x
    			if (y[j] < x[i]) {
    				c += 1d;
    			}
    			
    			// If the two values are equal
    			else if (y[j] == x[i]) {
    				c += 0.5d;
    			}
    			
    			// If the value in y has become > than the value in x
    			// break from the loop
    			else {
    				j = y.length;
    			}
    			
    		}
    		
    	}
    	
    	// Computing the quantity n1 * n2 - C
    	double q = x.length * y.length - c;
    	
    	// Returning the greater of the two quanties c and q
    	return c > q ? c : q;
    }

    /**
     * Computes and returns an approximate t value for the provided U value.
     * This method uses the equation at the bottom of page 429 in Biometry.
     * No correction for ties is made here.
     * 
     * @param uValue the computed U value.
     * @param n1 the size of the larger sample
     * @param n2 the size of the smaller sample
     * 
     * @return an approximate t value for the provided U value. 
     * 
     * @throws IllegalArgumentException if n1 &lt; n2, or if n1 &le; 20
     */
    public double getT(double uValue, int n1, int n2) {

    	if (n2 > n1) {
    		String s = "n1 must be >= n2";
    		throw new IllegalArgumentException(s);
    	}

    	if (n1 <= 20) {
    		String s = "n1 is only " + n1 + 
    		". The Statistical Table U should be consulted instead.";
    		throw new IllegalArgumentException(s);
    	}

    	// Computing the numerator
        double num = uValue - (n1 * n2 / 2d);
        
        // Computing the denominator
        double den = Math.sqrt(n1 * n2 * (n1 + n2 + 1) / 12d);

        // Returning the approximate t value
        return num / den;
    }

    /**
     * Returns the largest provided sample size. 
     * @param size1 a sample size.
     * @param size2 a sample size.
     * @return the largest provided sample size.
     */
    public int getN1(int size1, int size2) {
    	return Util.getLargest(size1, size2);
    }

    /**
     * Returns the smallest provided sample size. 
     * @param size1 a sample size.
     * @param size2 a sample size.
     * @return the smallest provided sample size.
     */

    public int getN2(int size1, int size2) {
    	return Util.getSmallest(size1, size2);
    }
    
    /**
     * Used for testing.
     * @param args none required.
     */
    public static void main(String[] args) {
    	
    	// Taken from page 428 in Biometry 
    	int[] sampleA = {
    			104,
    			109,
    			112,
    			114,
    			116,
    			118,
    			118,
    			119,
    			121,
    			123,
    			125,
    			126,
    			126,
    			128,
    			128,
    			128
    	};
    	
    	int[] sampleB = {
    			100,
    			105,
    			107,
    			107,
    			108,
    			111,
    			116,
    			120,
    			121,
    			123
    	};
    	
    	MannWhitneyU uTest = new MannWhitneyU();
    	
    	double u1 = uTest.getU(sampleA, sampleB);
    	double u2 = uTest.getU(sampleB, sampleA);
    	
    	System.out.println("U = " + u1);
    	System.out.println("U = " + u2);

    }
}
