/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.stat.comp;

/**
 * This class is used to store the results of a correlation computation.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 *
 */
public class TauResult implements CorrResult {

    private final double statistic;
    private final double zApprox;
    private final int n;

    /**
     * Constructs a new correlation results.
     * 
     * @param statistic the computed correlation statistic.
     * @param zApproximation the Z value approximated from the computed 
     * correlation statistic.
     * @param n the size of the sample.
     */
    public TauResult(double statistic, double zApproximation, int n) {
        this.statistic = statistic;
        this.zApprox = zApproximation;
        this.n = n;
    }

    /**
     * Returns the computed correlation statistic.
     * @return the computed correlation statistic.
     */
    public double getCoeff() {
        return statistic;
    }

    /**
     * Returns the Z value approximated from the computed correlation statistic.
     * @return the Z value approximated from the computed correlation statistic.
     */
    public double getApproxZ() {
        return zApprox;
    }

    /**
     * Returns the sample size.
     * @return the sample size.
     */
    public int getN() {
        return n;
    }

    /**
     * Returns a string represenation of these correlation results.
     * @return a string represenation of these correlation results.
     */
    public String toString() {
        return "statistic=" + statistic + ",Z=" + zApprox + ",n=" + n;
    }

    /**
     * Indicates if the provided object is equal to this correlation results
     * object.
     * 
     * @param obj the object with which to compare.
     * 
     * @return {@code true} if {@code obj} is an instance of 
     * {@code CorrelationResults} and its statistic, Z value, and sample size
     * are the same as this object's; {@code false} otherwise. 
     */
    public boolean equals(Object obj) {

        if (this == obj) {
            return true;
        }

        if (!(obj instanceof TauResult)) {
            return false;
        }

        TauResult cr = (TauResult) obj;

        if (n != cr.getN()) {
            return false;
        }

        if (Double.compare(statistic, cr.getCoeff()) != 0) {
            return false;
        }

        if (Double.compare(zApprox, cr.getApproxZ()) != 0) {
            return false;
        }

        return true;
    }

    /**
     * Returns a content-based hash code.
     * @return a content-based hash code.
     */
    public int hashCode() {
    	return n * (int) statistic ^ (int) zApprox;
    }
}
