/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.stat.proc;

import java.util.*;

import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.stat.comp.*;
import sbrt.kernel.utilities.DiscreteProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;

/**
 * This class is used to perform multiple correlation analyses.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <V> the variable type.
 * @param <T> the correlation result type.
 */
public class MultipleCorrelationAnalysis<V, T extends CorrResult> implements
        MonitoredProcess<T> {

    private final Correlator<V, T> correlator;

    private Map<? extends V, Double> xValues;

    private Iterable<Map<V, Double>> yValues;

    private DiscreteProgressMonitor progress;

    /**
     * Constructs a new correlation process.
     * 
     * @param correlator the correlator with which all computations will be
     * performed.
     */
    public MultipleCorrelationAnalysis(Correlator<V, T> correlator) {

        // Checking input
        if (correlator == null) {
            String s = "'correlator' cannot be null.";
            throw new NullPointerException(s);
        }

        // Saving
        this.correlator = correlator;

        progress = new DiscreteProgressMonitor();
    }

    /**
     * Sets the x values.
     * 
     * @param xValues variables mapped to their respective values.
     */
    public void setXValues(Map<? extends V, Double> xValues) {

        // Checking input
        if (xValues == null) {
            throw new NullPointerException("'xValues' cannot be null.");
        }

        // Saving
        this.xValues = xValues;
    }

    /**
     * Sets the source of y values.
     * 
     * @param yValues a source of variables mapped to their respective values.
     */
    public void setYValues(Iterable<Map<V, Double>> yValues) {

        // Checking input
        if (yValues == null) {
            throw new NullPointerException("'yValues' cannot be null.");
        }

        // Saving
        this.yValues = yValues;
    }

    private void checkState() {

        if (xValues == null) {
            String s = "No X values were provided.";
            throw new IllegalStateException(s);
        }

        if (yValues == null) {
            String s = "No Y values were provided.";
            throw new IllegalStateException(s);
        }
    }

    /**
     * Computes the correlation statistic between the x and y values, and adds
     * the results to the provided results manager.
     * 
     * @param resultsManager the results manager to which the computed results
     * are to be added.
     */
    public void run(ResultsManager<T> resultsManager) {

        // Ensuring all required inputs were provided
        checkState();

        try {
            
            progress.start();
            progress.resetCount();

            // Iterating thru each set of y values
            for (Map<? extends V, Double> y : yValues) {

                // Computing correlation results
                T r = correlator.corr(xValues, y);

                // Adding to results manager
                resultsManager.addData(r);

                progress.increment();
            }

            progress.stop();

        } catch (Throwable t) {
            progress.stop();
        }
    }

    /**
     * Returns a progress monitor for this process.
     * 
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress.asUnmodifiable();
    }

}
