/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.math.stat.proc;

import java.util.*;

import sbrt.kernel.io.ResultsManager;
import sbrt.kernel.math.stat.comp.*;

/**
 * This class is used to perform Mann-Whitney U tests.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 * @param <V> the variable type.
 */
public class MwUAnalysis<V> 
	implements sbrt.kernel.KernelProcess<MannWhitneyResults> {

	private final MannWhitneyU uTester;
	private List<? extends Number> sampleA;
	private List<? extends Number> sampleB;
	
	/**
	 * Constructs a new Mann-Whitney U analysis.
	 */
	public MwUAnalysis() {
		
		// Creating Mann-Whitney U computer
		this.uTester = new MannWhitneyU();
	}
	
	/**
	 * Sets sample A to the provided list of numbers.
	 * @param sampleA a list of values.
	 */
	public void setSampleA(List<? extends Number> sampleA) {
		
		// Checking input
		if (sampleA == null) {
			throw new NullPointerException("'sampleA' cannot be null.");
		}
		
		// Saving
		this.sampleA = sampleA;
	}
	
	/**
	 * Sets sample B to the provided list of numbers.
	 * @param sampleB a list of values.
	 */
	public void setSampleB(List<? extends Number> sampleB) {
		
		// Checking input
		if (sampleB == null) {
			throw new NullPointerException("'sampleB' cannot be null.");
		}
		
		// Saving
		this.sampleB = sampleB;
	}
	
	private void checkState() {
		
		if (sampleA == null) {
			String s = "Sample A was not provided.";
			throw new IllegalStateException(s);
		}		
		
		if (sampleB == null) {
			String s = "Sample B was not provided.";
			throw new IllegalStateException(s);
		}
	}
	
	/**
	 * Computes the Mann-Whitney U statistic of the samples A and B,
	 * and adds the results to the provided results manager.
	 * 
	 * @param resultsManager the results manager to which the computed results
	 * are to be added.
	 */
	public void run(ResultsManager<MannWhitneyResults> resultsManager) {
		
		// Ensuring all required inputs were provided
		checkState();

		// Computing the U statistic
		double u = uTester.getU(sampleA, sampleB);
		
		// Getting both sample sizes
		int sizeA = sampleA.size();
		int sizeB = sampleB.size();
		
		// Determining n1 and n2
		int n1 = uTester.getN1(sizeA, sizeB);
		int n2 = uTester.getN2(sizeA, sizeB);
		
		// If n1 is too small to compute an approximate t value
		if (n1 <= 20) {
			
			MannWhitneyResults r = new MannWhitneyResults(u, n1, n2);
			
			resultsManager.addData(r);
			
		}
		
		// If n1 is large enough to compute an approximate t value
		else {

			// Computing the approximate t value
			double t = uTester.getT(u, n1, n2);
			
			MannWhitneyResults r = new MannWhitneyResults(u, t, n1, n2);
			
			resultsManager.addData(r);
		}

	}


}
