/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities;

/**
 * This abstract class is a skeleton implementation of {@linkplain ProgressMonitor}.
 * This class provided methods for counting events and measuring elapsed times.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public abstract class AbstractProgressMonitor implements MutableProgressMonitor {

	// The number of completed operations
    private int counter;
    
    // The time at which this monitor was started
    private long startTime;
    
    // The last recorded duration
    private long duration;
    
    /**
     * Constructs a new progress monitor.
     */
    public AbstractProgressMonitor() {

    	// Initializing counter
        resetCount();

        // Initializing timer
        startTime = Long.MIN_VALUE;
        duration = Long.MIN_VALUE;
    }

    /**
     * Returns the number of times of {@code increment()} has been called.
     * @return the number of times of {@code increment()} has been called.
     */
    public synchronized int getCount() {
        return counter;
    }

    /**
     * Causes this progress monitor to increment an internal counter.
     * @see #getCount()
     */
    public synchronized void increment() {
        ++counter;
    }

    /**
     * Causes this progress monitor to reset it's internal counter to 0.
     * @see #increment()
     * @see #getCount()
     */
    public synchronized void resetCount() {
        counter = 0;
    }

    /**
     * Causes this progress monitor to start monitoring a duration of time. 
     * @throws IllegalStateException if {@code start()} was already called
     * without a subsequent call to {@code stop()}. 
     */
    public synchronized void start() {
    	
    	// If start() was called without a subsequent stop()
    	if (started()) {
    		String s = "start() was already called, but stop() wasn't.";
    		throw new IllegalStateException(s);
    	}

    	// Setting start time
    	startTime = System.currentTimeMillis();
    	
    	// Clearing the last recorded duration
    	duration = Long.MIN_VALUE;
    }

    /**
     * Causes this progress monitor to stop monitoring a duration of time.
     * 
     * @throws IllegalStateException if {@code start()} was never called, or
     * if {@code stop()} has already been called since the last call to
     * {@code start()}.
     * 
     */
    public synchronized void stop() {

    	// If start() was never called
        if (!started()) {
            String s = "start() was never called.";
            throw new IllegalStateException(s);
        }

        // If stop() was already called
        if (stopped()) {
        	String s = "stop() was already called.";
        	throw new IllegalStateException(s);
        }

        // Recording the duration
        duration = System.currentTimeMillis() - startTime;
        
        // Resetting start time
        startTime = Long.MIN_VALUE;
    }

    /**
     * Indicates if this progress monitor has already been started.
     * @return {@code true} if this progress monitor has already been started;
     * {@code false} otherwise.
     */
    public boolean started() {
        return startTime != Long.MIN_VALUE;
    }

    /**
     * Indicates if this progress monitor has already been stopped.
     * @return {@code true} if this progress monitor has already been stopped;
     * {@code false} otherwise.
     */
    public boolean stopped() {
        return duration != Long.MIN_VALUE;
    }
    
    /**
     * Returns the elapsed time between the calls of {@code start()} and
     * {@code stop()}. If {@code start()} was never called, {@code 0} is 
     * returned.
     * If {@code stop()} hasn't been called since the last call to 
     * {@code start()}, the elapsed time since the last call to {@code start()}
     * is returned. 
     * 
     * @return the elapsed time between the calls of {@code start()} and
     * {@code stop()}.
     */
    public synchronized long getDuration() {
    	
    	// If a duration has been recorded, return it
        if (stopped()) {
            return duration;
        }
        
        // If started but not yet stopped
        if (started() && !stopped()) {
            return System.currentTimeMillis() - startTime;
        }

        // If never started
        return 0;
    }

}
