/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities;

/**
 * This class is used to monitor the progress of a process that occurs in
 * discrete events. 
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public class DiscreteProgressMonitor extends AbstractProgressMonitor {

	/**
	 * Constructs a new discrete progress monitor.
	 */
    public DiscreteProgressMonitor() {
        super();
    }

    /**
     * Returns an integer value indicating a progress.
     * @return an integer value indicating a progress.
     */
    public synchronized Integer getProgress() {
        int i = getCount();
        return Integer.valueOf(i);
    }

    /**
     * Returns this progress monitor as an immutable object. This allows
     * the contents of this progress monitor to be queried without the 
     * possibility of modification.
     * 
     * @return this progress monitor as an immutable object.
     */
    public ProgressMonitor asUnmodifiable() {

    	final DiscreteProgressMonitor pm = this;
    	
        return new ProgressMonitor() {

        	/**
        	 * Returns the current duration of the process being monitored.
        	 * @return the current duration of the process in milliseconds.
        	 */
            public synchronized long getDuration() {
                return pm.getDuration();
            }

            /**
             * Returns an integer value indicating a progress.
             * @return an integer value indicating a progress.
             */
            public synchronized Integer getProgress() {
                return pm.getProgress();
            }

            /**
             * Indicates if this progress monitor has already been started.
             * @return {@code true} if this progress monitor has already been started;
             * {@code false} otherwise.
             */
            public boolean started() {
                return pm.started();
            }

            /**
             * Indicates if this progress monitor has already been stopped.
             * @return {@code true} if this progress monitor has already been stopped;
             * {@code false} otherwise.
             */
            public boolean stopped() {
                return pm.stopped();
            }

        };
    }

}
