/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.*;

/**
 * This class is used to represent logical relationships or operations. Logical
 * relationships have a set of <i>conditions</i> which usually evaluate to
 * either "true" or "false".
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 * @param <C>
 *            the condition type.
 */
public abstract class LogicalRelation<C> implements Iterable<C> {

	private UnmodifiableSet<C> elements;

	/**
	 * Constructs a new logical relationship from the provided
	 * conditions.
	 * 
	 * @param conditions the set of conditions.
	 */
	public LogicalRelation(Collection<C> conditions) {

		int size = conditions.size() + 1;

		Set<C> s = new LinkedHashSet<C>(size, 1f);
		s.addAll(conditions);

		this.elements = new UnmodifiableSet<C>(s);
	}

	/**
	 * Returns the set of conditions of this logical relationship.
	 * @return the set of conditions of this logical relationship.
	 */
	public UnmodifiableSet<C> getConditions() {
		return elements;
	}

	/**
	 * Returns the number of conditions of this logical relationship.
	 * @return the number of conditions of this logical relationship.
	 */
	public int elements() {
		return elements.size();
	}

	/**
	 * Returns an iterator over the conditions of this logical relationship.
	 * @return an iterator over the conditions of this logical relationship.
	 */
	public Iterator<C> iterator() {
		return elements.iterator();
	}

	/**
	 * Indicates if this logical relationship contains the provided condition.
	 * 
	 * @param condition the condition for which occurence is checked.
	 * 
	 * @return {@code true} if the provide condition is contained in this
	 * logical relationship; {@code false} otherwise.
	 */
	public boolean contains(Object condition) {
		return elements.contains(condition);
	}

	/**
	 * Returns a content-based hash code.
	 * @return a content-based hash code.
	 */
	public int hashCode() {
		return elements.hashCode();
	}

	/**
	 * Inidicates if this logical relationship is "equal to" the provided
	 * object.
	 * 
	 * @param obj the object for which comparison is made.
	 * 
	 * @return {@code true} if this logical relationship equals the provided
	 * object; {@code false} otherwise.
	 */
	public abstract boolean equals(Object obj);

}
