/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities;

/**
 * This class is used to monitor a process whose progress can be measured as
 * a percentage.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public class PercentProgressMonitor extends AbstractProgressMonitor {

    private static final Double ZERO = Double.valueOf(0d);
    private static final Double ONE_HUNDRED = Double.valueOf(100d);

    // The total number of events to expect
    private int total;
    
    // The number of completed events
    private int completed;

	/**
	 * Constructs a new percentage-based progress monitor.
	 */
    public PercentProgressMonitor() {
        super();
        
        // Initializing counters
        completed = 0;
        total = 0;
    }

	/**
	 * Constructs a new percentage-based progress monitor.
	 * 
	 * @param total the maximum number of calls allowed for {@code increment()}.
	 * @see #setTotal(int)
	 */
    public PercentProgressMonitor(int total) {
        super();
        setTotal(total);
    }

    /**
     * Sets the total number of times {@code increment()} can be called. This
     * is the value for which computed percentages are relative to.
     * 
     * @param total the total number of times {@code increment()} can be called.
     */
    public synchronized void setTotal(int total) {
    	
    	// Checking input
        if (total < 0) {
            String s = "Invalid total : " + total;
            throw new IllegalArgumentException(s);
        }
        
        this.total = total;
    }

    /**
     * Causes this progress monitor to reset it's internal counter to 0.
     * @see #increment()
     * @see #getCount()
     */
    public synchronized void resetCount() {
        completed = 0;
    }

    /**
     * Causes this progress monitor to increment an internal counter.
     * @see #getCount()
     * @throws IllegalStateException if this call causes the internal counter
     * to exceed the provided total.
     */
    public synchronized void increment() {
    	
    	// Checking counter
        if (completed >= total) {
            String s = "the total " + total + " was exceeded";
            throw new IllegalStateException(s);
        }
        
        ++ completed;
    }

    /**
     * Returns a double precision number indicating the completed percentage 
     * of the process being monitored.
     * 
     * @return a double precision number indicating the completed percentage 
     * of the process being monitored.
     */
    public synchronized Double getProgress() {
    	
    	// If total has not yet been set 
        if (total < 1) {
            return ZERO;
        }
        
        // If counter equals total
        if (completed == total) {
            return ONE_HUNDRED;
        }
        
        // Calculating the percentage
        double fraction = (double) completed / (double) total * 100d;

        // Returning the percentage as a Double
        return Double.valueOf(fraction);
    }

    /**
     * Sets the internal counter equal to the provided total. Subsequent calls
     * to {@code getProgress()} will return 100.
     */
    public synchronized void complete() {
        total = completed;
    }

    /**
     * Returns this progress monitor as an immutable object. This allows
     * the contents of this progress monitor to be queried without the 
     * possibility of modification.
     * 
     * @return this progress monitor as an immutable object.
     */
    public ProgressMonitor asUnmodifiable() {

    	final PercentProgressMonitor pm = this;
    	
        return new ProgressMonitor() {

        	/**
        	 * Returns the current duration of the process being monitored.
        	 * @return the current duration of the process in milliseconds.
        	 */
            public synchronized long getDuration() {
                return pm.getDuration();
            }

            /**
             * Returns a double precision number indicating the completed
             * percentage of the process being monitored.
             * 
             * @return a double precision number indicating the completed 
             * percentage of the process being monitored.
             */
            public synchronized Double getProgress() {
                return pm.getProgress();
            }

            /**
             * Indicates if this progress monitor has already been started.
             * @return {@code true} if this progress monitor has already been started;
             * {@code false} otherwise.
             */
            public boolean started() {
                return pm.started();
            }

            /**
             * Indicates if this progress monitor has already been stopped.
             * @return {@code true} if this progress monitor has already been stopped;
             * {@code false} otherwise.
             */
            public boolean stopped() {
                return pm.stopped();
            }
        };
    }


}
