/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities;

import java.util.*;

/**
 * This class is used to create an unmodifiable view of collections. It
 * is intended to mimic the behavior of 
 * {@code java.util.Collections.unmodifiableCollection(java.util.Collection)}, 
 * but in a more explicit way. If a method returns an 
 * {@code UnmodifiableCollection}, it's clear that the collection is 
 * unmodifiable and that {@code UnsupportedOperationException} will be thrown
 * if modification is attempted.
 * 
 * @author This class was written and documented by Jeremiah Wright while 
 * in the Wagner lab.
 *
 * @param <E> the element type.
 */
public class UnmodifiableCollection<E> implements Collection<E> {

    private final Collection<E> c;

    /**
     * Constructs a new unmodifiable view of the provided collection.
     * @param c the collection for which an unmodifiable view is created.
     */
    public UnmodifiableCollection(Collection<E> c) {
    	// Checking and setting
        Util.nullCheck(c);
        this.c = Collections.unmodifiableCollection(c);
    }

    /**
     * See {@linkplain Collection#add(Object)}.
     */
	public boolean add(E o) {
		return c.add(o);
	}

    /**
     * See {@linkplain Collection#addAll(Collection)}.
     */
	public boolean addAll(Collection<? extends E> c) {
		return this.c.addAll(c);
	}

    /**
     * See {@linkplain Collection#clear()}.
     */
	public void clear() {
		c.clear();
	}

    /**
     * See {@linkplain Collection#contains(Object)}.
     */
	public boolean contains(Object o) {
		return c.contains(o);
	}

    /**
     * See {@linkplain Collection#containsAll(Collection)}.
     */
	public boolean containsAll(Collection<?> c) {
		return this.c.containsAll(c);
	}

    /**
     * See {@linkplain Collection#isEmpty()}.
     */
	public boolean isEmpty() {
		return c.isEmpty();
	}

    /**
     * See {@linkplain Collection#iterator()}.
     */
	public Iterator<E> iterator() {
		return c.iterator();
	}

    /**
     * See {@linkplain Collection#remove(Object)}.
     */
	public boolean remove(Object o) {
		return c.remove(o);
	}

    /**
     * See {@linkplain Collection#removeAll(Collection)}.
     */
	public boolean removeAll(Collection<?> c) {
		return this.c.removeAll(c);
	}

    /**
     * See {@linkplain Collection#retainAll(Collection)}.
     */
	public boolean retainAll(Collection<?> c) {
		return this.c.retainAll(c);
	}

    /**
     * See {@linkplain Collection#size()}.
     */
	public int size() {
		return c.size();
	}

    /**
     * See {@linkplain Collection#toArray()}.
     */
	public Object[] toArray() {
		return c.toArray();
	}

    /**
     * See {@linkplain Collection#toArray(Object[])}.
     */
	public <T> T[] toArray(T[] a) {
		return c.toArray(a);
	}

    /**
     * Returns {@code toString()} on the underlying collection.
     * @return {@code toString()} on the underlying collection.
     */
	public String toString() {
		return c.toString();
	}

	/**
	 * See {@linkplain Collection#equals(Object)}.
	 */
	public boolean equals(Object obj) {
		return c.equals(obj);
	}

	/**
	 * See {@linkplain Collection#hashCode()}.
	 */
	public int hashCode() {
		return c.hashCode();
	}
}
