/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities;

import java.util.*;

/**
 * This class is used to create an iterator that cannot modify its 
 * underlying collection. 
 * If a method returns an {@code UnmodifiableIterator}, it's clear that the
 * underlying collection cannot be modified with that iterator and that 
 * {@code UnsupportedOperationException} will be thrown
 * if {@code remove()} is called.
 * 
 * @author This class was written and documented by Jeremiah Wright while 
 * in the Wagner lab.
 *
 * @param <E> the element type.
 */
public class UnmodifiableIterator<E> implements Iterator<E> {

    private final Iterator<E> i;

    /**
     * Constructs a new unmodifiable iterator from the provided iterator.
     * @param i the iterator for which an unmodifiable version is created.
     */
    public UnmodifiableIterator(Iterator<E> i) {

    	// Checking and setting
    	Util.nullCheck(i);
        this.i = i;
    }

    /**
     * See {@linkplain Iterator#next()}.
     */
    public E next() {
    	return i.next();
    }
    
    /**
     * See {@linkplain Iterator#hasNext()}.
     */
    public boolean hasNext() {
    	return i.hasNext();
    }
    
    /**
     * Throws {@code UnsupportedOperationException}.
     * @throws UnsupportedOperationException
     */
    public void remove() {
    	String s = "Attempted to remove an element.";
    	throw new UnsupportedOperationException(s);
    }
}
