/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities;

import java.util.*;

/**
 * This class is used to create an unmodifiable view of lists. It
 * is intended to mimic the behavior of 
 * {@code java.util.Collections.unmodifiableList(List)},
 * but in a more explicit way. If a method returns an 
 * {@code UnmodifiableList}, it's clear that the list is 
 * unmodifiable and that {@code UnsupportedOperationException} will be thrown
 * if modification is attempted.
 * 
 * @author This class was written and documented by Jeremiah Wright while 
 * in the Wagner lab.
 *
 * @param <E> the element type.
 */
public class UnmodifiableList<E> implements List<E> {

    private final List<E> list;

    /**
     * Constructs a new unmodifiable view of the provided list.
     * @param list the list for which an unmodifiable view is created.
     */
    public UnmodifiableList(List<E> list) {

    	// Checking and setting
    	Util.nullCheck(list);
        this.list = Collections.unmodifiableList(list);
    }

    /**
     * See {@linkplain List#add(Object)}.
     */
	public boolean add(E o) {
		return list.add(o);
	}

    /**
     * See {@linkplain List#add(int, Object)}.
     */
	public void add(int index, E element) {
		list.add(index, element);
	}

    /**
     * See {@linkplain List#addAll(Collection)}.
     */
	public boolean addAll(Collection<? extends E> c) {
		return list.addAll(c);
	}

    /**
     * See {@linkplain List#addAll(int, Collection)}.
     */
	public boolean addAll(int index, Collection<? extends E> c) {
		return list.addAll(index, c);
	}

    /**
     * See {@linkplain List#clear()}.
     */
	public void clear() {
		list.clear();
	}

    /**
     * See {@linkplain List#contains(Object)}.
     */
	public boolean contains(Object o) {
		return list.contains(o);
	}

    /**
     * See {@linkplain List#containsAll(Collection)}.
     */
	public boolean containsAll(Collection<?> c) {
		return list.containsAll(c);
	}

    /**
     * See {@linkplain List#get(int)}.
     */
	public E get(int index) {
		return list.get(index);
	}

    /**
     * See {@linkplain List#indexOf(Object)}.
     */
	public int indexOf(Object o) {
		return list.indexOf(o);
	}

    /**
     * See {@linkplain List#isEmpty()}.
     */
	public boolean isEmpty() {
		return list.isEmpty();
	}

    /**
     * See {@linkplain List#iterator()}.
     */
	public Iterator<E> iterator() {
		return list.iterator();
	}

    /**
     * See {@linkplain List#lastIndexOf(Object)}.
     */
	public int lastIndexOf(Object o) {
		return list.lastIndexOf(o);
	}

    /**
     * See {@linkplain List#listIterator()}.
     */
	public ListIterator<E> listIterator() {
		return list.listIterator();
	}

    /**
     * See {@linkplain List#listIterator(int)}.
     */
	public ListIterator<E> listIterator(int index) {
		return list.listIterator(index);
	}

    /**
     * See {@linkplain List#remove(Object)}.
     */
	public boolean remove(Object o) {
		return list.remove(o);
	}

    /**
     * See {@linkplain List#remove(int)}.
     */
	public E remove(int index) {
		return list.remove(index);
	}

    /**
     * See {@linkplain List#removeAll(Collection)}.
     */
	public boolean removeAll(Collection<?> c) {
		return list.removeAll(c);
	}

    /**
     * See {@linkplain List#retainAll(Collection)}.
     */
	public boolean retainAll(Collection<?> c) {
		return list.retainAll(c);
	}

    /**
     * See {@linkplain List#set(int, Object)}.
     */
	public E set(int index, E element) {
		return list.set(index, element);
	}

    /**
     * See {@linkplain List#size()}.
     */
	public int size() {
		return list.size();
	}

    /**
     * See {@linkplain List#subList(int, int)}.
     */
	public List<E> subList(int fromIndex, int toIndex) {
		return list.subList(fromIndex, toIndex);
	}

    /**
     * See {@linkplain List#toArray()}.
     */
	public Object[] toArray() {
		return list.toArray();
	}

    /**
     * See {@linkplain List#toArray(Object[])}.
     */
	public <T> T[] toArray(T[] a) {
		return list.toArray(a);
	}

    /**
     * Returns {@code toString()} on the underlying list.
     * @return {@code toString()} on the underlying list.
     */
	public String toString() {
		return list.toString();
	}

	/**
	 * See {@linkplain List#equals(Object)}.
	 */
	public boolean equals(Object obj) {
		return list.equals(obj);
	}

	/**
	 * See {@linkplain List#hashCode()}.
	 */
	public int hashCode() {
		return list.hashCode();
	}
}
