/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities;

import java.util.*;

/**
 * This class is used to create an unmodifiable view of maps. It
 * is intended to mimic the behavior of 
 * {@code java.util.Collections.unmodifiableMap(Map)}, 
 * but in a more explicit way. If a method returns an 
 * {@code UnmodifiableMap}, it's clear that the map is 
 * unmodifiable and that {@code UnsupportedOperationException} will be thrown
 * if modification is attempted.
 * 
 * @author This class was written and documented by Jeremiah Wright while 
 * in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public class UnmodifiableMap<K, V> implements Map<K, V> {

    private final Map<K, V> m;

    /**
     * Constructs a new unmodifiable view of the provided map.
     * @param m the map for which an unmodifiable view is created.
     */
    public UnmodifiableMap(Map<? extends K, ? extends V> m) {

    	// Checking and setting
    	Util.nullCheck(m);
        this.m = Collections.unmodifiableMap(m);
    }

    /**
     * See {@linkplain Map#clear()}.
     */
	public void clear() {
		m.clear();
	}

    /**
     * See {@linkplain Map#containsKey(Object)}.
     */
	public boolean containsKey(Object key) {
		return m.containsKey(key);
	}

    /**
     * See {@linkplain Map#containsValue(Object)}.
     */
	public boolean containsValue(Object value) {
		return m.containsValue(value);
	}

    /**
     * See {@linkplain Map#entrySet()}.
     */
	public UnmodifiableSet<java.util.Map.Entry<K, V>> entrySet() {
		return new UnmodifiableSet<Map.Entry<K, V>>(m.entrySet());
	}

    /**
     * See {@linkplain Map#get(Object)}.
     */
	public V get(Object key) {
		return m.get(key);
	}

    /**
     * See {@linkplain Map#isEmpty()}.
     */
	public boolean isEmpty() {
		return m.isEmpty();
	}

    /**
     * See {@linkplain Map#keySet()}.
     */
	public UnmodifiableSet<K> keySet() {
		return new UnmodifiableSet<K>(m.keySet());
	}

    /**
     * See {@linkplain Map#put(Object, Object)}.
     */
	public V put(K key, V value) {
		return m.put(key, value);
	}

    /**
     * See {@linkplain Map#putAll(Map)}.
     */
	public void putAll(Map<? extends K, ? extends V> t) {
		m.putAll(t);
	}

    /**
     * See {@linkplain Map#remove(Object)}.
     */
	public V remove(Object key) {
		return m.remove(key);
	}

    /**
     * See {@linkplain Map#size()}.
     */
	public int size() {
		return m.size();
	}

    /**
     * See {@linkplain Map#values()}.
     */
	public UnmodifiableCollection<V> values() {
		return new UnmodifiableCollection<V>(m.values());
	}

    /**
     * Returns {@code toString()} on the underlying map.
     * @return {@code toString()} on the underlying map.
     */
	public String toString() {
		return m.toString();
	}

	/**
	 * See {@linkplain Map#equals(Object)}.
	 */
	public boolean equals(Object obj) {
		return m.equals(obj);
	}

	/**
	 * See {@linkplain Map#hashCode()}.
	 */
	public int hashCode() {
		return m.hashCode();
	}
}
