/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities;

import java.util.*;

/**
 * This class is used to create an unmodifiable view of sets. It
 * is intended to mimic the behavior of 
 * {@code java.util.Collections.unmodifiableSet(Set)},
 * but in a more explicit way. If a method returns an 
 * {@code UnmodifiableSet}, it's clear that the set is 
 * unmodifiable and that {@code UnsupportedOperationException} will be thrown
 * if modification is attempted.
 * 
 * @author This class was written and documented by Jeremiah Wright while 
 * in the Wagner lab.
 *
 * @param <E> the element type.
 */
public class UnmodifiableSet<E> implements Set<E> {

	private final Set<E> s;

    /**
     * Constructs a new unmodifiable view of the provided set.
     * @param set the set for which an unmodifiable view is created.
     */
    public UnmodifiableSet(Set<E> set) {

    	// Checking and setting
    	Util.nullCheck(set);
    	s = Collections.unmodifiableSet(set);
    }

    /**
     * See {@linkplain Set#add(Object)}.
     */
	public boolean add(E o) {
		return s.add(o);
	}

    /**
     * See {@linkplain Set#addAll(Collection)}.
     */
	public boolean addAll(Collection<? extends E> c) {
		return s.addAll(c);
	}

    /**
     * See {@linkplain Set#clear()}.
     */
	public void clear() {
		s.clear();
	}

    /**
     * See {@linkplain Set#contains(Object)}.
     */
	public boolean contains(Object o) {
		return s.contains(o);
	}

    /**
     * See {@linkplain Set#containsAll(Collection)}.
     */
	public boolean containsAll(Collection<?> c) {
		return s.containsAll(c);
	}

    /**
     * See {@linkplain Set#isEmpty()}.
     */
	public boolean isEmpty() {
		return s.isEmpty();
	}

    /**
     * See {@linkplain Set#iterator()}.
     */
	public Iterator<E> iterator() {
		return s.iterator();
	}

    /**
     * See {@linkplain Set#remove(Object)}.
     */
	public boolean remove(Object o) {
		return s.remove(o);
	}

    /**
     * See {@linkplain Set#removeAll(Collection)}.
     */
	public boolean removeAll(Collection<?> c) {
		return s.removeAll(c);
	}

    /**
     * See {@linkplain Set#retainAll(Collection)}.
     */
	public boolean retainAll(Collection<?> c) {
		return s.retainAll(c);
	}

    /**
     * See {@linkplain Set#size()}.
     */
	public int size() {
		return s.size();
	}

    /**
     * See {@linkplain Set#toArray()}.
     */
	public Object[] toArray() {
		return s.toArray();
	}

    /**
     * See {@linkplain Set#toArray(Object[])}.
     */
	public <T> T[] toArray(T[] a) {
		return s.toArray(a);
	}

    /**
     * Returns {@code toString()} on the underlying set.
     * @return {@code toString()} on the underlying set.
     */
	public String toString() {
		return s.toString();
	}

	/**
	 * See {@linkplain Set#equals(Object)}.
	 */
	public boolean equals(Object obj) {
		return s.equals(obj);
	}

	/**
	 * See {@linkplain Set#hashCode()}.
	 */
	public int hashCode() {
		return s.hashCode();
	}
}
