/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities;

import java.util.*;

/**
 * This class contains static methods that perform utility functions.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 */
public final class Util {

    private Util() {
    	// prevents invocation
    } 

    /**
     * Ensures the provided object is not null.
     * @param obj the object to check.
     * @throws NullPointerException if the provided argument is {@code null}.
     */
    public static void nullCheck(Object obj) {
        if (obj == null) {
            throw new NullPointerException();
        }
    }

    /**
     * Returns a new array containing the values from the provided array
     * in ascending numerical order.
     * The returned array is completely autonomous relative to the provided array. 
     * 
     * @param values the array whose values are to be sorted.
     * 
     * @return a new array containing the values from the provided array
     * in ascending numerical order.
     */
    public static int[] getSorted(int[] values) {

        int[] sorted = new int[values.length];

        for (int i = 0; i < values.length; i++) {
            sorted[i] = values[i];
        }

        Arrays.sort(sorted);
        return sorted;
    }

    /**
     * Returns a new array containing the values from the provided array
     * in ascending numerical order.
     * The returned array is completely autonomous relative to the provided array. 
     * 
     * @param values the array whose values are to be sorted.
     * 
     * @return a new array containing the values from the provided array
     * in ascending numerical order.
     */
    public static double[] getSorted(double[] values) {

        double[] sorted = new double[values.length];

        for (int i = 0; i < values.length; i++) {
            sorted[i] = values[i];
        }

        Arrays.sort(sorted);
        return sorted;
    }

    /**
     * Returns an array containing the values in the provided collection as doubles.
     * The order of appearance of values in the array is the same as that
     * returned by an iterator over the collection.
     * The returned array is completely autonomous relative to the provided collection. 
     * 
     * @param values a collection of double precision values.
     * 
     * @return an array containing the values in the provided collection.
     * 
     * @see Number#doubleValue()
     */
    public static double[] getArray(Collection<? extends Number> values) {

        double[] array = new double[values.size()];
        int index = 0;

        for (Number n : values) {
            array[index++] = n.doubleValue();
        }

        return array;
    }

    /**
     * Returns an array containing the values in the provided collection.
     * The order of appearance of values in the array is the same as that
     * returned by an iterator over the collection.
     * The returned array is completely autonomous relative to the provided collection. 
     * 
     * @param values a collection of double precision values.
     * 
     * @return an array containing the values in the provided collection.
     */
    public static int[] getArray(Collection<Integer> values) {

        int[] array = new int[values.size()];
        int index = 0;

        for (Integer i : values) {
            array[index++] = i.intValue();
        }

        return array;
    }

    /**
     * Returns a sorted array containing the values in the provided collection.
     * The returned array is completely autonomous relative to the provided collection. 
     * 
     * @param values a collection of double precision values.
     * 
     * @return an array containing the values in the provided collection.
     * 
     * @see Arrays#sort(double[])
     * @see Number#doubleValue()
     */
    public static double[] getSortedArray(Collection<? extends Number> values) {

        double[] array = getArray(values);
        Arrays.sort(array);

        return array;
    }

    /**
     * Returns a sorted array containing the values in the provided collection.
     * The returned array is completely autonomous relative to the provided collection. 
     * 
     * @param values a collection of double precision values.
     * 
     * @return an array containing the values in the provided collection.
     * 
     * @see Arrays#sort(int[])
     */
    public static int[] getSortedArray(Collection<Integer> values) {

        int[] array = getArray(values);
        Arrays.sort(array);

        return array;
    }

    /**
     * Returns a {@code 2 x n} array containing the values from the provided
     * maps. The array at index {@code 0} contains the x values, and the array
     * at index {@code 1} contains the y values. The order of appearance of
     * values in these arrays is the same as that returned by an iterator over
     * {@code xValues.entrySet()}. The key sets of the provided maps need not
     * be identical, but only the intersecting values will be considered when
     * filling the arrays. The lengths {@code n} of these arrays equals the size
     * of the intersection of the key sets of the provided maps. Therefore, if
     * the provided maps contain identical keys, then {@code n == xValues.size() &&
     * n == yValues.size()} evaluates to {@code true}.
     * 
     * @param xValues keys mapped to values.
     * @param yValues keys mapped to values.
     * 
     * @return a {@code 2 x n} array containing the values from the provided
     * maps. 
     */
    public static double[][] getArrays(Map<?, Double> xValues,
                                       Map<?, Double> yValues) {
            
        // Creating storage
        double[] xArray = new double[xValues.size()];
        double[] yArray = new double[xArray.length];
        
        int index = 0;

        for (Map.Entry<?, Double> e : xValues.entrySet()) {

            // Getting the key
            Object variable = e.getKey();

            // Getting the y value, if it exists
            Double y = yValues.get(variable);

            // Moving on to the next if it doesn't exist
            if (y == null) {
                continue;
            }

            // Getting the x value
            Double x = e.getValue();

            // Setting those values in the arrays
            xArray[index] = x.doubleValue();
            yArray[index] = y.doubleValue();
            
            ++index;
        }

        double[][] arrays = new double[2][];
        
        // Trimming to appropriate lengths
        arrays[0] = Arrays.copyOf(xArray, index);
        arrays[1] = Arrays.copyOf(yArray, index);
        
        return arrays;
    }
    
    /**
     * Indicates if the provided value equals 0.
     * @param d the value to test.
     * @return {@code Math.abs(d) < Double.MIN_VALUE;}.
     */
    public static boolean isZero(double d) {
        return Math.abs(d) < Double.MIN_VALUE;
    }

    /**
     * Indicates if the provided values are equal to each other, given the
     * provided tolerance.
     * 
     * @param value1 a value.
     * @param value2 a value.
     * @param tolerance the amount by which the provided values can differ
     * but still be considered equal.
     * 
     * @return {@code Math.abs(value1 - value2) < tolerance;}
     */
    public static boolean equal(double value1, double value2,
                                double tolerance) {
        
        if (value1 == value2) {
            return true;
        }
        
        return Math.abs(value1 - value2) < tolerance;
    }


    /**
     * Attemps to format the provided string such that system-dependent
     * newline characters appear no more than {@code maxWidth} characters 
     * apart. This method is not robust.
     *  
     * @param string the string to format.
     * @param maxWidth the ideal maximum distance between newline characters. 
     * 
     * @return the formatted version of the provided string.
     */
    public static String widthFormat(String string, int maxWidth) {
        
        if (maxWidth <= 0) {
            String s = "Invalid maxWidth: " + maxWidth;
            throw new IllegalArgumentException(s);
        }
        
        if (string.length() < maxWidth) {
            return string;
        }
        
        StringBuilder sb = new StringBuilder(string);
        
        int startIndex = 0;
        int currIndex = startIndex;
        
        while (currIndex < sb.length()) {
        
            int width = currIndex - startIndex;
            
            // If max width has been reached
            if (width != 0 && width % maxWidth == 0) {
                
                // Try to break on whitespace
                int newIndex = replaceWhiteWithNewLine(sb, currIndex, 5); 
                
                // If it worked
                if (newIndex > 0) {
                    currIndex = newIndex;
                    startIndex = newIndex;
                }
                
                // Insert break manually 
                else {
                    startIndex = insertBreak(sb, currIndex);
                    currIndex = startIndex;
                }
                
            } else {
                ++ currIndex; 
            }
        }
        
        return sb.toString();
    }
    
    private static int insertBreak(StringBuilder sb, int index) {
        
        sb.insert(index, "\\\n");
        
        return index + 2;
    }
    
    private static int replaceWhiteWithNewLine(StringBuilder sb,
                                               int index,
                                               int windowSize) {
        
        for (int i = 0; i < windowSize; i ++) {
            
            int currIndex = index - i;
            
            if (currIndex < 0) {
                return -1;
            }
            
            if (Character.isWhitespace(sb.charAt(currIndex))) {
                sb.replace(currIndex, currIndex, System.getProperty("line.separator"));
                return currIndex + 1;
            }
            
        }
        
        return -1;
    }
    
    /**
     * Prints the provided object to {@linkplain System#err}, preceded by a carriage
     * return character. 
     * @param o the object to print.
     */
    public static void debugPrintln(Object o) {
        System.err.println('\r');
        System.err.println(o);
    }

    /**
     * Prints the elements of the provided collection to {@linkplain System#out},
     * one per line.
     * 
     * @param c the collection to be printed.
     */
    public static void print(Collection<?> c) {
    	for (Object obj : c) {
            System.out.println(obj);
        }
    }

    /**
     * Returns the smallest of the provided numbers.
     * 
     * @param v1 a numerical value.
     * @param v2 a numerical value.
     * 
     * @return the smallest of the provided numbers.
     */
    public static double getSmallest(double v1, double v2) {
        if (v1 < v2) {
            return v1;
        }
        return v2;
    }

    /**
     * Returns the largest of the provided numbers.
     * 
     * @param v1 a numerical value.
     * @param v2 a numerical value.
     * 
     * @return the largest of the provided numbers.
     */
    public static double getLargest(double v1, double v2) {
        if (v1 > v2) {
            return v1;
        }
        return v2;
    }

    /**
     * Returns the smallest of the provided numbers.
     * 
     * @param v1 an integer value.
     * @param v2 an integer value.
     * 
     * @return the smallest of the provided numbers.
     */
    public static int getSmallest(int v1, int v2) {
        if (v1 < v2) {
            return v1;
        }
        return v2;
    }

    /**
     * Returns the largest of the provided numbers.
     * 
     * @param v1 an integer value.
     * @param v2 an integer value.
     * 
     * @return the largest of the provided numbers.
     */
    public static int getLargest(int v1, int v2) {
        if (v1 > v2) {
            return v1;
        }
        return v2;
    }

    /**
     * Returns a list of duplicate elements contained in the provided
     * collection.
     * 
     * @param <E> the element type.
     * @param c a collection of elements.
     * 
     * @return a list of duplicate elements contained in the provided
     * collection.
     */
    public static <E> List<E> getDuplicates(Collection<E> c) {
        
    	// Creating storage
        Set<E> temp = new HashSet<E>();
        List<E> duplicates = new LinkedList<E>();

        // Iterating thru the provided collection
        for (E e : c) {

        	// If the element was not added
            if (!temp.add(e)) {
                duplicates.add(e);
            }
        }

        return duplicates;
    }

    /**
     * Returns a list of elements common to both provided collections.
     * 
     * @param <E> the element type.
     * @param c1 a collection of elements.
     * @param c2 a collection of elements.
     * 
     * @return a list of elements common to both provided collections.
     */
    public static <E> List<E> getCommonElements(Collection<? extends E> c1, 
    											Collection<? extends E> c2) {

    	// Creating variables
        Collection<? extends E> smallest = null;
        Collection<? extends E> largest = null;

        // Determining the smallest collection
        if (c1.size() < c1.size()) {
            smallest = c1;
            largest = c2;
        } else {
            smallest = c2;
            largest = c1;
        }

        // Creating storage for common elements
        List<E> elements = new LinkedList<E>();

        // Iterating thru the smallest collection
        for (E e : smallest) {

        	// If the other set also contains e
            if (largest.contains(e)) {
                elements.add(e);
            }
        }

        return elements;
    }

    /**
     * Indicates if the provided numbers are of opposite sign.
     * 
     * @param x1 a numerical value.
     * @param x2 a numerical value.
     * 
     * @return {@code true} if one number is positive and the other negative;
     * {@code false} otherwise.
     */
    public static boolean oppositeSign(double x1, double x2) {
        return (x1 < 0 && x2 > 0) || (x1 > 0 && x2 < 0);
    }

    /**
     * Indicates if the provided numbers are of opposite sign.
     * 
     * @param x1 an integer value.
     * @param x2 an integer value.
     * 
     * @return {@code true} if one number is positive and the other negative;
     * {@code false} otherwise.
     */
    public static boolean oppositeSign(int x1, int x2) {
        return (x1 < 0 && x2 > 0) || (x1 > 0 && x2 < 0);
    }

    /**
     * Used for testing purposes.
     * @param args ignored.
     */
    public static void main(String[] args) {
        
        
        Map<String, Double> m1 = new TreeMap<String, Double>();
        m1.put("A", 1d);
        m1.put("B", 2d);
        m1.put("C", 3d);
        

        Map<String, Double> m2 = new TreeMap<String, Double>();
        m2.put("A", 7d);
        m2.put("B", 8d);
        m2.put("D", 3d);
        
        double[][] values = getArrays(m1, m2);
        System.out.println(Arrays.toString(values[0]));
        System.out.println(Arrays.toString(values[1]));
    }
}
