/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities.proc;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.math.IntervalComparer.Diff;
import sbrt.kernel.utilities.DiscreteProgressMonitor;
import sbrt.kernel.utilities.MutableProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;
import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.io.ResultsManager;

/**
 * This class is used to compare vectors of intervals.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 * @param <V> the variable type.
 */
public class IntervalComparison<V> implements MonitoredProcess<Set<Diff<V>>> {

    private final MutableProgressMonitor progress;
    private IntervalEquiv eqTester;
    
    private Iterable<Map<V, Interval>> source1;
    private Iterable<Map<V, Interval>> source2;
    
    /**
     * Constructs a new interval vector comparison process.
     */
    public IntervalComparison() {
        progress = new DiscreteProgressMonitor();
        eqTester = new IntervalEquiv(0d);
    }

    /**
     * Sets the equivalence tolerance.
     * @param tolerance the amount by which a bound of two intervals can
     * differ but still be considered equivalent.
     */
    public void setTolerance(double tolerance) {
        eqTester = new IntervalEquiv(tolerance);
    }
    
    /**
     * Returns the equivalence tolerance.
     * @return the equivalence tolerance.
     */
    public double getTolerance() {
        return eqTester.getTolerance();
    }
    
    /**
     * Sets of the sources of interval vectors.
     * @param source1 an interval vector source.
     * @param source2 an interval vector source.
     */
    public void setIntervalSources(Iterable<Map<V, Interval>> source1,
                                   Iterable<Map<V, Interval>> source2) {
        
        Util.nullCheck(source1);
        Util.nullCheck(source2);
        
        this.source1 = source1;
        this.source2 = source2;
    }
    
    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress;
    }
    
    private void checkState() {
    	
    	if (source1 == null) {
    		String s = "Interval sources were not provided.";
    		throw new IllegalStateException(s);
    	}
    	
    }
    
    /**
     * Adds to the provided results manager the variables (keys) whose intervals
     * (values) are not considered equivalent.
     * 
     * @param resultsManager the results manager to sets of variables are added.
     */
    public void run(ResultsManager<Set<Diff<V>>> resultsManager) {

    	checkState();
        
        progress.start();
        progress.resetCount();
        
        double tolerance = getTolerance();
        
        IntervalComparer comp = new IntervalComparer();
        
        Iterator<Map<V, Interval>> i = source1.iterator();
        Iterator<Map<V, Interval>> j = source2.iterator();
        
        while (i.hasNext() && j.hasNext()) {
        
            Map<V, Interval> m1 = i.next();
            Map<V, Interval> m2 = j.next();
            
            Set<Diff<V>> diffs = comp.getDiffs(m1, m2, tolerance);

            resultsManager.addData(diffs);
            
            progress.increment();
        }

        progress.stop();
    }
    
}
