/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities.proc;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.KernelProcess;
import sbrt.kernel.io.ResultsManager;

/**
 * This class is used to organize a collection of mathematical expressions
 * according to the variables they contain.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 * @param <E> the expression type. 
 * @param <V> the variable type.
 */
public class VariableParticipation<E extends MathExpr<V>, V> 
	implements KernelProcess<Map<V, List<E>>> {

    private Collection<E> expressions;

    /**
     * Sets the collection of mathematical expressions to organize.
     * @param expressions the collection of mathematical expressions to
     * organize.
     */
    public void 
    setExprs(Collection<? extends E> expressions) {

    	// Determining appropriate size
    	int size = expressions.size() + 1;
    	
    	// Creating storage
        this.expressions = 
        	new LinkedHashSet<E>(size, 1.0F);

        // Copying
        this.expressions.addAll(expressions);    	
    }
    
    private void checkState() {
    	
    	if (expressions == null) {
    		String s = "The collection of expressions is null.";
    		throw new IllegalStateException(s);
    	}
    	
    }
    
    /**
     * Adds to the provided results manager a map of variables mapped to a
     * list of mathematical expressions in which they participate.
     * 
     * @param resultsManager the results manager to which the map is added.
     */
    public void 
    run(ResultsManager<Map<V, List<E>>> resultsManager) {

    	checkState();
    	
        // Creating storage
        Map<V, List<E>> participation = 
        	new LinkedHashMap<V, List<E>>();

        // Iterating thru all expressions
        for (E expr : expressions) {

            // Iterating thru each variable in this expression
            for (V var : expr.getVariables()) {

                // Getting all expressions this variable participates in
                List<E> expressions = participation.get(var);

                // If this is the first, create a new list
                if (expressions == null) {
                    expressions = new LinkedList<E>();
                    participation.put(var, expressions);
                }

                // Adding the expression to the list for this variable
                expressions.add(expr);
            }
        }

        // Adding to the results manager
        resultsManager.addData(participation);

    }

}
