/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.kernel.utilities.proc;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.utilities.DiscreteProgressMonitor;
import sbrt.kernel.utilities.MutableProgressMonitor;
import sbrt.kernel.utilities.ProgressMonitor;
import sbrt.kernel.utilities.Util;
import sbrt.kernel.MonitoredProcess;
import sbrt.kernel.io.ResultsManager;

/**
 * This class is used to compare numerical vectors.
 * 
 * @author This class was written and documented by Jeremiah Wright
 * while in the Wagner lab.
 *
 * @param <V> the variable type.
 */
public class VectorComparison<V> implements MonitoredProcess<Map<V, Double>> {

    private final MutableProgressMonitor progress;
    private DoubleEquiv eqTester;
    
    private Iterable<Map<V, Double>> source;

    /**
     * Constructs a new numerical vector comparison process.
     */
    public VectorComparison() {
        progress = new DiscreteProgressMonitor();
        eqTester = new DoubleEquiv(0d);
    }

    /**
     * Sets the equivalence tolerance.
     * @param tolerance the amount by which two numerical values can
     * differ but still be considered equivalent.
     */
    public void setTolerance(double tolerance) {
        eqTester = new DoubleEquiv(tolerance);
    }

    /**
     * Returns the equivalence tolerance.
     * @return the equivalence tolerance.
     */
    public double getTolerance() {
        return eqTester.getTolerance();
    }

    /**
     * Sets of the source of numerical vectors.
     * @param source a numerical vector source.
     */
    public void setVectorSource(Iterable<Map<V, Double>> source) {
        Util.nullCheck(source);
        this.source = source;
    }
    
    /**
     * Returns a progress monitor for this process.
     * @return a progress monitor for this process.
     */
    public ProgressMonitor getProgress() {
        return progress;
    }
    
    private void checkState() {
    	
    	if (source == null) {
    		String s = "Vector sources were not provided.";
    		throw new IllegalStateException(s);
    	}
    	
    }
    
    /**
     * Adds to the provided results manager the variables (keys) whose intervals
     * (values) are unequal within the provided tolerance. 
     * 
     * @param resultsManager the results manager to sets of variables are added.
     */
    public void run(ResultsManager<Map<V, Double>> resultsManager) {

    	checkState();
        
        progress.start();
        progress.resetCount();
        
        double tolerance = getTolerance();
        
        VectorComparer comp = new VectorComparer();
        
        int currentM1 = 0;
        int currentM2 = 0;
        
        for (Map<V, Double> m1 : source) {
        
            ++ currentM1;
            
            for (Map<V, Double> m2 : source) {

                ++ currentM2;
                
                if (currentM2 <= currentM1) {
                    continue;
                }
                
                boolean same = comp.areSame(m1, m2, tolerance);
                
                progress.increment();

                if (same) {
                    
                    resultsManager.addData(m1);
                    resultsManager.addData(m2);
                    progress.stop();
                    
                    return;
                }
                
            }
            
            currentM2 = 0;
            
        }

        progress.stop();
    }
    
}
