/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

import sbrt.shell.io.RomiFile;
import sbrt.shell.mng.*;
import sbrt.shell.text.*;

/**
 * This class is used to create {@code ProcessManager} objects.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public class ProcessManagerFactory {

	private static final String DEFAULT_NAME = "sbrt_process_names.txt";
	
    /**
     * The keyword used to denote a process.
     */
    public static final String PROCESS = "Process";

    /**
     * The keyword used to denote a procedure name file.
     */
    public static final String PROCESS_NAME_FILE = "Process Name File";

    @SuppressWarnings("unchecked")
	private static 
	Class<ProcessManager<String, String>> getClass(String className) {

        try {

            return (Class<ProcessManager<String, String>>) 
            	Class.forName(className);

        } catch (ExceptionInInitializerError e) {

            Throwable cause = e.getCause();
            if (cause instanceof ApplicationException) {
                throw (ApplicationException) cause;
            }

            String s = "An unexpected exception has occurred in a static " +
                       "initializer in the class '" + className + "'.";
            throw new ApplicationException(s, cause);

        } catch (LinkageError e) {

            String s = "A linkage error has occurred for the class '" +
                       className + "'.";
            throw new ApplicationException(s, e);

        } catch (ClassNotFoundException e) {

            String s = "The class '" + className + "' was not found.";
            throw new ApplicationException(s, e);

        }

    }

    /**
     * Creates a formatted string representation of the provided input
     * and prints it to the provided print writer.
     * 
     * @param input the input to be formatted.
     * @param out the print writer to which the string will be printed.
     */
    public static void 
    printInput(ManagerInput<String, String> input, PrintWriter out) {

    	// Creating map formatter
        MapFormat<String, String> f = ColonMapFormatV1.getBasic();

        // Creating storage for string
        StringBuilder header = new StringBuilder(100);
        
        // Appending title
        header.append("#  Process Specifications  #");

        // Padding
        while (header.length() < 79) {
            header.insert(0, '#');
            
            if (header.length() < 79) {
                header.append('#');
            }
        }

        // Printing header
        out.println(header);

        // Iterating thru each input
        for (Map.Entry<String, String> e : input.toMap().entrySet()) {

        	// Printing formatted entry
            out.println(f.formatPair(e));
        }

        // Printing a closing string
        for (int i = 0; i < header.length() && i < 79; i++) {
            out.print('#');
        }

        out.println();
    }

    private static String getProcessNameFile(ManagerInput<String, String> input) {
    	
    	String fileName = input.getOptionalValue(PROCESS_NAME_FILE);
    	
    	if (fileName != null) {
    		File f = new File(WorkingDirectory.get(), fileName);
    		return f.getPath();
    	}
    	
    	File home = HomeDirectory.get();
    	
    	File file = new File(home, DEFAULT_NAME);
    	
    	if (!file.exists()) {
    		String s = "The default process name file '" + DEFAULT_NAME +
    				   "' could not be found. A process name file must, " +
    				   "therefore, be specified manually.";
    		throw new ApplicationException(s);
    	}

    	return file.getPath();
    }
    
    /**
     * Prints a formatted string representation of the provided input to the
     * provided print writer and returns its corresponding process manager.
     * 
     * @param input a manager input containing values for {@linkplain #PROCESS} and
     * {@linkplain #PROCESS_NAME_FILE} and all other values required to create
     * the process manager corresponding to those two values.
     * 
     * @param out the print writer to which the string will be printed.
     * 
     * @return the process manager corresponding to the provided input.
     * 
     * @throws ApplicationException if a process manager cannot be 
     * created from the provided manager input.
     */
    public static ProcessManager<String, String> 
    make(ManagerInput<String, String> input, PrintWriter out) {

    	// Printing the manager input
        printInput(input, out);
        out.println();

        // Getting the provided process name 
        String processName = input.getRequiredValue(PROCESS);

        // Getting the provided process name file
        String processNameFile = getProcessNameFile(input);
        
        // Creating file parser
        MapFormat<String, String> format = 
        	EqualsMapFormatV1.getBasic();
        
        // Opening the file
        RomiFile<String, String> file = 
        	new RomiFile<String, String>(processNameFile, format);
        
        // Retrieving the corresponding class name
        String className = file.getMap().get(processName);

        // If no such object exists
        if (className == null) {
            String s = "The process '" + processName +
                       "' does not exist in '" + file.getCanonicalName() + "'.";
            throw new ApplicationException(s);
        }

        // Retrieving the corresponding class
        Class<ProcessManager<String, String>> managerType = 
        	getClass(className);

        // Creating storage for final result
        ProcessManager<String, String> manager = null;

        try {

            // Getting a constructor
            Class<?>[] parameters = new Class<?>[0];
            Constructor<ProcessManager<String, String>> constructor =
            	managerType.getConstructor(parameters);

            // Getting a new ProcessManager object
            Object[] initArgs = new Object[0];
            manager = constructor.newInstance(initArgs);

            // Preparing input for the new manager
            Map<String, String> managerInputs = input.toMap();
            managerInputs.remove(PROCESS);
            managerInputs.remove(PROCESS_NAME_FILE);

            // Setting the new manager input
            input = new MapManagerInput<String, String>(managerInputs, 
            		                                    input.getName());
            
            manager.setInput(input, out);

        } catch (NoSuchMethodException e) {

            String s = "The class '" + managerType.getName() +
                       "' was not properly implemented.";
            throw new ApplicationException(s, e);

        } catch (InstantiationException e) {

            String s = "The class '" + managerType.getName() +
                       "' is abstact.";
            throw new ApplicationException(s, e);

        } catch (IllegalAccessException e) {

            String s = "The constructor for '" + managerType.getName() +
                       "' is inaccessible.";
            throw new ApplicationException(s, e);

        } catch (InvocationTargetException e) {

            Throwable cause = e.getCause();
            if (cause instanceof ApplicationException) {
                throw (ApplicationException) cause;
            }

            String s = "The constructor for '" + managerType.getName() +
                       "' threw an unknown exception during instantiation.";
            throw new ApplicationException(s, cause);

        }

        return manager;
    }

    private ProcessManagerFactory() { /* prevents invocation */}


}
