/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell;

import java.io.*;
import java.util.*;

import sbrt.kernel.math.optimization.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.mng.*;

/**
 * This class is used to create {@code ProgramSolver} objects.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public class ProgramSolverFactory {

    private ProgramSolverFactory() { /* prevents invocation */
    }

    /**
     * The keyword used to denote the name of a program solver.
     */
    public static final String SOLVER = "Program Solver";

    /**
     * The keyword used to denote a parameter file for a program solver.
     */
    public static final String PARAM_FILE = "Program Solver Parameter File";

    /**
     * The string used to denote the GLPK program solver from GNU.
     */
    public static final String GLPK = "GLPK";

    /**
     * The string used to denote the CPLEX program solver from ILOG.
     */
    public static final String CPLEX = "CPLEX";

    /**
     * The correspondence between the strings used to denote a particular
     * program solver and the name of the actual class it represents. The keys
     * of this map are the strings used to denote a program solver and the
     * values of this map are the names of the actual class it represents.
     */
    public static final UnmodifiableMap<String, String> nameClassMap;
    static {

        Map<String, String> m = new TreeMap<String, String>();

        m.put(GLPK,  "sbrt.ext.glpk.GlpkSolver");
        m.put(CPLEX, "sbrt.ext.cplex.CplexSolver");

        nameClassMap = new UnmodifiableMap<String, String>(m);
    }

    /**
     * Uses the information contained in the provided input to create a new
     * program solver object. The keyword {#link {@linkplain #SOLVER} is required to
     * be present in the input, and the keyword {@linkplain #PARAM_FILE} is optional.
     * If {@code PARAM_FILE} <i>is</i> present, the program solver indicated by
     * {@code SOLVER} must have a functioning
     * {@linkplain ProgramSolver#setParamFile(String)} method.
     * 
     * @param <R> the row type.
     * @param <C> the column type.
     * @param input the input from which information is used to create a new
     * program solver object.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return the program solver corresponding to the provided input.
     * 
     * @throws ApplicationException if a problem ocurrs while creating a program
     * solver object.
     * 
     * @throws UnsupportedOperationException if a call to
     * {@linkplain ProgramSolver#setParamFile(String)} fails.
     */
    public static <R, C> ProgramSolver<R, C> make(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting the name of the program solver
        String solverName = input.getRequiredValue(SOLVER);

        // Getting the name of the parameter file, if one was provided
        String paramFile = input.getOptionalValue(PARAM_FILE);

        // Printing an opening message
        out.print("Initializing '" + solverName + "' ... ");
        out.flush();

        // Creating field to store program solver
        ProgramSolver<R, C> ps = null;

        try {

            // Getting an instance of the program solver
            ps = getInstance(solverName);

            // Setting the parameter file if one was provided
            if (paramFile != null) {

                try {

                    ps.setParamFile(paramFile);

                } catch (UnsupportedOperationException e) {

                    String s =
                            "The program solver '"
                                    + solverName
                                    + "' does not support parameter files.";
                    throw new ApplicationException(s, e);
                }
            }

        } catch (ProgramSolverException e) {
            throw interpret(e);
        }

        // Printing a closing message
        out.println("DONE");

        return ps;
    }

    /**
     * Interprets a program solver exception and returns a corresponding
     * application exception.
     * 
     * @param e the program solver exception to be interpretted.
     * 
     * @return an application exception that is the intrpretted version of the
     * provided program solver exception.
     */
    public static ApplicationException interpret(ProgramSolverException e) {

        String s =
                "The program solver threw an exception. Its error message is '"
                        + e.getCause().getMessage().trim()
                        + "'.";

        return new ApplicationException(s, e);
    }

    @SuppressWarnings("unchecked")
    private static final <R, C> ProgramSolver<R, C> getInstance(String solverName) {

        // Getting class name
        String className = nameClassMap.get(solverName);

        // If provided solver name does not exist
        if (className == null) {
            String s =
                    "The program solver '" + solverName + "' does not exist.";
            throw new ApplicationException(s);
        }

        try {

            // Getting a class object
            Class<?> c = Class.forName(className);

            // Casting
            Class<ProgramSolver<R, C>> psClass = (Class<ProgramSolver<R, C>>) c;

            // Returning a new instance of the class
            return psClass.newInstance();

        } catch (NoClassDefFoundError e) {

            String s =
                    "A linkage error occurred for the class '"
                            + className
                            + "'. Please ensure that "
                            + "the proper files are located in the SBRT's"
                            + " 'lib' directory.";
            throw new ApplicationException(s, e);

        } catch (UnsatisfiedLinkError e) {

            String s =
                    "A linkage error occurred for the class '"
                            + className
                            + "'. Please ensure that "
                            + "the proper files are located in the SBRT's"
                            + " 'lib' directory.";
            throw new ApplicationException(s, e);

        } catch (LinkageError e) {

            String s =
                    "A linkage error occurred for the class '"
                            + className
                            + "'. Please ensure that "
                            + "the proper files are located in the SBRT's"
                            + " 'lib' directory.";
            throw new ApplicationException(s, e);

        } catch (ClassNotFoundException e) {

            String s = "The class '" + className + "' was not found.";
            throw new ApplicationException(s, e);

        } catch (InstantiationException e) {

            String s =
                    "The class '"
                            + className
                            + "' could not be instantiated. "
                            + "The implementation of '"
                            + solverName
                            + "' appears to be "
                            + "flawed.";
            throw new ApplicationException(s, e);

        } catch (IllegalAccessException e) {

            String s = "The class '" + className + "' could not be accessed.";
            throw new ApplicationException(s, e);

        }

    }

}
