/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Insets;
import java.io.OutputStream;
import java.io.PrintWriter;

import javax.swing.JTextPane;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.StyledDocument;

/**
 * This class is used to create the output pane of the SBRT GUI where text is written
 * by each process.
 *  
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 *
 */
public class SbrtOutputPane extends JTextPane {


	private static final long serialVersionUID = 1L;

	private static final String PLAIN = "plain";

    private static final String ERROR = "error";
    
    private static final String SECTION_BREAK = "section_break";
    
    
	private final PrintWriter plainWriter;
	private final PrintWriter errorWriter;
	private final PrintWriter sectionBreakWriter;
	
	/**
	 * Creates a new SBRT GUI output pane.
	 */
	public SbrtOutputPane() {
		
		plainWriter = makeWriter(PLAIN);
		errorWriter = makeWriter(ERROR);
		sectionBreakWriter = makeWriter(SECTION_BREAK);

        setMargin(new Insets(5, 5, 5, 5));
        setEditable(false);
        setPreferredSize(new Dimension(725, 750));

        StyledDocument doc = getStyledDocument();
        addStylesToDocument(doc);
        
	}

    private PrintWriter makeWriter(String style) {
        OutputStream out = new TextPaneOutputStream(this, style);
        return new PrintWriter(out, true);
    }
    

    private void addStylesToDocument(StyledDocument doc) {

    	StyleContext context = StyleContext.getDefaultStyleContext();
    	
        Style def = context.getStyle(StyleContext.DEFAULT_STYLE);

        Style plain = doc.addStyle(PLAIN, def);
        StyleConstants.setFontSize(plain, 14);
        StyleConstants.setFontFamily(plain, "Courier");

        Style error = doc.addStyle(ERROR, plain);
        StyleConstants.setForeground(error, Color.red);
        
        Style sectionBreak = doc.addStyle(SECTION_BREAK, plain);
        StyleConstants.setForeground(sectionBreak, Color.blue);
        
    }
    
    
    /**
     * Returns the print writer used for writing plain text to the
     * output pane.
     * 
     * @return the print writer used for writing plain text to the
     * output pane.
     */
	public PrintWriter getPlainWriter() {
		return plainWriter;
	}
    
    /**
     * Returns the print writer used for writing error messages to this
     * output pane.
     * 
     * @return the print writer used for writing error messages to this
     * output pane.
     */
	public PrintWriter getErrorWriter() {
		return errorWriter;
	}

	/**
	 * Writes the provided string to this output pane as plain text.
	 * @param s the string to be written.
	 */
	public void writePlain(String s) {
		plainWriter.write(s);
	}

	/**
	 * Writes the provided string to this output pane as an error message.
	 * @param s the string to be written.
	 */
	public void writeError(String s) {
		errorWriter.write(s);
	}
	
	/**
	 * Writes a section break to this output pane.
	 */
	public void writeSectionBreak() {
		
        for (int i = 0; i < 40; i++) {
            sectionBreakWriter.print("_ ");
        }
        
        sectionBreakWriter.println();
        
	}
	
}
