/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.gui;

import java.awt.BorderLayout;
import java.awt.Desktop;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URI;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import sbrt.FatalException;
import sbrt.ProcessRunner;
import sbrt.Sbrt;
import sbrt.shell.ApplicationException;
import sbrt.shell.HomeDirectory;
import sbrt.shell.WorkingDirectory;

/**
 * This class is used to create a GUI for the SBRT.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public class TextPaneGui extends JPanel implements ActionListener {

    private static final long serialVersionUID = -89245306682693526L;

    private SbrtOutputPane outputPane;

    private JButton workingDirButton;

    private JButton executeButton;

    private JButton helpLink;
    
    private JFileChooser fc;

    private JFileChooser dc;

    /**
     * Constructs a new text pane GUI for the SBRT.
     *
     */
    public TextPaneGui() {
        super(new BorderLayout());
        initGui();
    }

    private void initGui() {
    	
        // Creating the upper panel
        JPanel panel = new JPanel();
        addMainButtonPanel(panel);
        addHelpPanel(panel);

        // Creating output window
        outputPane = new SbrtOutputPane();

        // Making output scrollable
        JScrollPane outputScrollPane = new JScrollPane(outputPane);
        
        
        // Adding the panel and output to this TextPaneGui
        add(panel, BorderLayout.PAGE_START);
        add(outputScrollPane, BorderLayout.CENTER);
    }

    
    private void addMainButtonPanel(JPanel panel) {

        // Creating file choosers
        dc = getDirectoryChooser();
        fc = getFileChooser();

        // Creating buttons
        workingDirButton = getWorkingDirButton();
        executeButton = getExecuteButton();
        
        // Creating a separate panel for buttons
        JPanel p = new JPanel();
        p.add(workingDirButton);
        p.add(executeButton);
    	
        panel.add(p);
    }

    private void addHelpPanel(JPanel panel) {
    	
    	// Adding a help button only if it is supported
        if (!Desktop.isDesktopSupported()) {
        	return;
        }
        
        if (!Desktop.getDesktop().isSupported(Desktop.Action.BROWSE)) {
        	return;
        }
        
        helpLink = getHelpLink();
    	
    	JPanel p = new JPanel();
    	p.add(helpLink);
    	
    	panel.add(p);
    }
    
    
    private JFileChooser getDirectoryChooser() {

        JFileChooser fc = new JFileChooser(WorkingDirectory.get());
        fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);

        return fc;
    }

    private JFileChooser getFileChooser() {

        JFileChooser fc = new JFileChooser(WorkingDirectory.get());
        fc.setFileSelectionMode(JFileChooser.FILES_ONLY);

        return fc;
    }

    private JButton getExecuteButton() {

        JButton button = new JButton("Execute an SBRT process");
        button.addActionListener(this);

        return button;
    }

    private JButton getWorkingDirButton() {

        JButton button = new JButton("Select a working directory");
        button.addActionListener(this);

        return button;
    }

    private JButton getHelpLink() {
    	
    	JButton button = new JButton("<html><font color=blue>Help</font></html>");
    	button.addActionListener(this);
    
    	return button;
    }

    
    /**
     * Responds to the provided action event.
     * @param e the action event.
     */
    public void actionPerformed(ActionEvent e) {

    	try {
    		
			if (e.getSource() == workingDirButton) {
				handleDirectoryChooser();
				return;
			}

			if (e.getSource() == executeButton) {
				handleFileChooser();
				return;
			}

			if (e.getSource() == helpLink) {
				handleHelpLink();
				return;
			}

		} catch (RuntimeException exception) {
			FatalException fe = ProcessRunner.interpret(exception);
			writeErrorMessage(fe.getMessage());
		}
    }


    /**
	 * Creates and shows the SBRT's GUI.
	 * 
	 */
    public void createAndShowWindow() {

        // Creating the window
        String title = "Systems Biology Research Tool " + Sbrt.getVersion();
        JFrame frame = new JFrame(title);
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        // Creating the content pane
        setOpaque(true);
        frame.setContentPane(this);

        // Displaying the window
        frame.pack();
        frame.setVisible(true);
    }

    private void handleDirectoryChooser() {

        int returnVal = dc.showOpenDialog(this);

        if (returnVal == JFileChooser.APPROVE_OPTION) {

            File file = dc.getSelectedFile();

            WorkingDirectory.set(file);
            fc.setCurrentDirectory(file);
        }
        
    }

    private void handleFileChooser() {

        // Handling the open button action
        int returnVal = fc.showOpenDialog(this);

        if (returnVal == JFileChooser.APPROVE_OPTION) {

            File file = fc.getSelectedFile();

            // Starting a new thread to run the SBRT process
            Thread process = new ProcessThread(file);
            process.start();

        }

    }

    private void handleHelpLink() {
		
		File sbrtHome = HomeDirectory.get();
		
		String usersGuide = "doc" + File.separator +
							"users_guide" + File.separator + 
							"index.html";
		
		File f = new File(sbrtHome, usersGuide);
		
		if (!f.exists()) {
			
			String s = "The User's Guide could not be found at '" + f.getAbsolutePath() + "'.";
			writeErrorMessage(s);
			
			return;
		}
		
		if (!f.canRead()) {
			
			String s = "The User's Guide was found, but could not be opened.";
			writeErrorMessage(s);
			
			return;
		}
		
		
    	URI uri = f.toURI();
    	
    	try {
			Desktop.getDesktop().browse(uri);
		} catch (IOException e1) {
			
			String s = "The default browser of this computer could not be launched to open the User's Guide.";
			
			ApplicationException e = new ApplicationException(s, e1);
			
			writeErrorMessage(ProcessRunner.interpret(e).getMessage());
			
			return;
		}
    	
    }
    
    private class ProcessThread extends Thread {

        private final File processFile;

        /**
         * Constructs a new SBRT process thread.
         * @param processFile an SBRT process file.
         */
        protected ProcessThread(File processFile) {
            super();
            this.processFile = processFile;
        }

        /**
         * Runs the SBRT process.
         */
        public void run() {

            // Disabling buttons
            executeButton.setEnabled(false);
            workingDirButton.setEnabled(false);

            PrintWriter out = outputPane.getPlainWriter();

            out.println();

            // Running the SBRT process
            try {
                new ProcessRunner().run(processFile.getAbsolutePath(), out);
            } catch (Throwable t) {
            	writeErrorMessage(t.getMessage());
            }

            out.println();
            outputPane.writeSectionBreak();
            out.println();

            // Enabling buttons again
            workingDirButton.setEnabled(true);
            executeButton.setEnabled(true);
        }

        
    }
    
    
    private void writeErrorMessage(String s) {

    	PrintWriter err = outputPane.getErrorWriter();
    	
        err.println();
        err.println();
        err.println();

        err.println(s);

        err.println();
        err.flush();
    	
    }


}
