/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.gui;

import java.io.IOException;
import java.io.OutputStream;
import javax.swing.JTextPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.StyledDocument;

/**
 * An output stream that writes its output to a javax.swing.JTextPane. This class is intended
 * to work in conjuction with {@linkplain SbrtOutputPane}.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab, and was based on a class written by Ranganath Kini.
 */
public class TextPaneOutputStream extends OutputStream {
	
    private static final String newLine = "\n";
    private static final String carriageReturn = "\r";
    
    private JTextPane textPane;
    private StyledDocument doc;
    private String style;
    
    private int lastNewLinePosition = 0;
    private boolean carriageReturnSeen = false;
    
    /**
     * Creates a new instance of TextAreaOutputStream which writes
     * to the specified instance of javax.swing.JTextArea control.
     *
     * @param textPane   A reference to the javax.swing.JTextArea
     *                  control to which the output must be redirected
     *                  to.
     * @param style 
     */
    public TextPaneOutputStream(JTextPane textPane, String style) {
        doc = textPane.getStyledDocument();
        this.textPane = textPane;
        this.style = style;
    }
    
    /**
     * Writes the specified byte as a character to the
     * javax.swing.JTextArea.
     *
     * @param   b   The byte to be written as character to the
     *              JTextArea.
     * @throws IOException 
     */
    public void write(int b) throws IOException {
        
        // This method emulates the result of carriage returns '\r'
        // in terminal windows.
        
        String s = String.valueOf((char) b);

        if (s.equals(newLine)) {
            lastNewLinePosition = doc.getLength() + 1;
            carriageReturnSeen = false;
        }

        else if (carriageReturnSeen) {
            eraseToLastNewLine();
            carriageReturnSeen = false;
        }

        else if (s.equals(carriageReturn)) {
            carriageReturnSeen = true;
        }

        writeToDocument(s);
    }

    private void eraseToLastNewLine() {
        
        try {
            
            // Removing all text since the last new line character '\n'
            doc.remove(lastNewLinePosition, doc.getLength() - lastNewLinePosition);

            // Updating the cursor's position
            textPane.setCaretPosition(doc.getLength());
            
        } catch (BadLocationException e) {
            throw new IllegalStateException();
        }

    }
    
    private void writeToDocument(String s) {
        
        try {
            
            // Writing the string
            doc.insertString(doc.getLength(), s, doc.getStyle(style));
            
            // Moving the cursor to the end
            textPane.setCaretPosition(doc.getLength());
            
        } catch (BadLocationException e) {
            throw new IllegalStateException(e);
        }
        
    }
    

}