/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.io.File;
import java.util.*;

import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;

/**
 * This abstract class is a skeleton implemenation of text-based
 * map input files.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public abstract class AbstractMapInputFile<K, V> 
		implements MapTextInputFile<K, V> {

	// The input file
	private final TextInputFile<String> file;
	
	// The map format
	private final MapFormat<K, V> mapFormat;

	// A set of key that must be present in this file
	private final UnmodifiableSet<K> requiredKeys;

	/**
	 * Constructs a new map input file from the provided arguments.
	 * 
	 * @param fileName the name of the file to read.
	 * @param fileFormat the format of the file.
	 * @param requiredKeys the set of keys this file must contain.
	 * @param mapFormat the map format used to parse this file.
	 */
	public AbstractMapInputFile(String fileName, 
								FileFormat fileFormat,
								Set<? extends K> requiredKeys,
								MapFormat<K, V> mapFormat) {
		// Creating file
		file = TextInputFile.getBasic(fileName, fileFormat);
		
		// Checking and saving
		Util.nullCheck(mapFormat);
		this.mapFormat = mapFormat;

		// Determining appropriate size
		int size = requiredKeys.size() + 1;

		// Creating storage
		Set<K> s = new LinkedHashSet<K>(size, 1.0F);

		// Copying
		s.addAll(requiredKeys);

		// Saving
		this.requiredKeys = new UnmodifiableSet<K>(s);
	}

	/**
	 * Constructs a new map input file from the provided arguments.
	 * 
	 * @param fileName the name of the file to read.
	 * @param fileFormat the format of the file.
	 * @param mapFormat the map format used to parse this file.
	 */
	public AbstractMapInputFile(String fileName, 
								FileFormat fileFormat,
								MapFormat<K, V> mapFormat) {
		// Creating file
		file = TextInputFile.getBasic(fileName, fileFormat);
		
		// Checking and saving
		Util.nullCheck(mapFormat);
		this.mapFormat = mapFormat;

		// Getting an empty set for the required variables
		Set<K> empty = Collections.emptySet();
		
		// Saving
		this.requiredKeys = new UnmodifiableSet<K>(empty);
	}

	/**
	 * Returns the set of keys that must be contained in this
	 * file.
	 * 
	 * @return the set of keys that must be contained in this
	 * file.
	 */
	public UnmodifiableSet<K> getRequiredKeys() {
		return requiredKeys;
	}

	/**
	 * Returns the map format used to parse this file.
	 * @return the map format used to parse this file.
	 */
	public MapFormat<K, V> getMapFormat() {
		return mapFormat;
	}

    /**
     * Returns a plain text line iterator over this text file.
     * @return a plain text line iterator over this text file.
     */
	protected final TextFileLineIterator<String> lineIterator() {
		return file.lineIterator();
	}

    /**
     * Closes this file. If this file is already closed, then
     * invoking this method has no effect.
     */
	public void close() {
		file.close();
	}

    /**
     * Returns the name of this file.
     * @return the name of this file.
     */
    public final String getOriginalName() {
        return file.getOriginalName();
    }

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     * 
     * @see File#getCanonicalPath()
     */
    public final String getCanonicalName() {
        return file.getCanonicalName();
    }

    /**
     * Indicates if this file has already been closed.
     * 
     * @return {@code true} if this file is closed;
     * {@code false} othewise.
     */
	public boolean isClosed() {
		return file.isClosed();
	}
}
