/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.io.File;
import java.util.*;

import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;

/**
 * This abstract class is a skeleton implementation of 
 * {@linkplain MapTextOutputFile}. 
 *  
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public abstract class AbstractMapOutputFile<K, V> 
	implements MapTextOutputFile<K, V> {

	// The text file to write to
	private final TextOutputFile<String> file;
	
	// The map format
    private final MapFormat<K, V> mapFormat;
    
    // The keys whose value will be written
    private final UnmodifiableSet<K> keys;

    /**
     * Constructs a new map output file from the provided
     * arguments.
     * 
     * @param fileName the name of the file to create.
     * 
     * @param fileFormat the format of the file to create.
     * 
     * @param keys the keys whose values will be
     * written to this file. If {@code null}, all values
     * from every added map will be written to this file.
     * 
     * @param mapFormat the map format.
     * 
     * @throws NullPointerException if any argument except 
     * {@code variables} is {@code null}.
     */
	public AbstractMapOutputFile(String fileName,
        FileFormat fileFormat,
        Set<? extends K> keys,
        MapFormat<K, V> mapFormat) {

    	// Creating the text file
    	file = TextOutputFile.getBasic(fileName, fileFormat);

    	// Checking and saving the vector format
        Util.nullCheck(mapFormat);
        this.mapFormat = mapFormat;

        if (keys == null) {
            this.keys = null;
        }

        else {
        	
            // Copying the set while preserving original order
            Set<K> s = new LinkedHashSet<K>(keys);
            
            // Saving it
            this.keys = new UnmodifiableSet<K>(s);
        }
    }

    /**
     * Constructs a new map output file from the provided
     * arguments.
     * 
     * @param fileName the name of the file to create.
     * 
     * @param fileFormat the format of the file to create.
     * 
     * @param allVariables a string representation of the 
     * variables whose values will be written to this file. This 
     * string is parsed by the provided vector format.
     * 
     * @param mapFormat the vector format.
     * 
     * @throws NullPointerException if any argument is {@code null}.
     */
    public AbstractMapOutputFile(String fileName,
        FileFormat fileFormat,
        String allVariables,
        MapFormat<K, V> mapFormat) {

    	this(fileName,
             fileFormat,
             mapFormat.parseAllKeys(allVariables),
             mapFormat);
    }

    /**
     * Constructs a new vector output file from the provided
     * arguments. The value of each variable from every added
     * vector will be written to this file.
     * 
     * @param fileName the name of the file to create.
     * @param fileFormat the format of the file to create.
     * @param mapFormat the map format.
     * 
     * @throws NullPointerException if any argument is {@code null}.
     */
    public AbstractMapOutputFile(String fileName,
        FileFormat fileFormat,
        MapFormat<K, V> mapFormat) {

    	this(fileName, fileFormat, (Set<K>) null, mapFormat);
    }
	
	/**
	 * Returns the set of keys whose values will be
	 * written to this output file when a map is added.
	 * This set acts as a type of filter to control which
	 * values in a map are saved, and which are discarded.
	 * 
	 * @return the set of keys whose values will be
	 * written to this output file when a map is added;
	 * or {@code null} if <i>every</i> value in an added map
	 * will be written to this file.
	 */
    public UnmodifiableSet<K> getFilterKeys() {
        return keys;
    }

    /**
     * Returns the format used to write lines to this
     * file.
     * 
     * @return the format used to write lines to this
     * file.
     */
    public MapFormat<K, V> getMapFormat() {
        return mapFormat;
    }

    /**
     * Writes the provided string to this file as a line.
     * 
     * @param line the string to write as a line in this file.
     */
    protected final void writeLine(String line) {
    	file.writeLine(line);
    }

    /**
     * Closes this file. If the file is already closed, no
     * action will be taken.
     */
	public void close() {
		file.close();
	}

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     */
    public final String getName() {
        return file.getName();
    }

    /**
     * Returns the name of this file.
     * @return the name of this file.
     */
    public final String getOriginalName() {
        return file.getOriginalName();
    }

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     * 
     * @see File#getCanonicalPath()
     */
    public final String getCanonicalName() {
        return file.getCanonicalName();
    }

    /**
     * Indicates if this file has already been closed.
     * 
     * @return {@code true} if this file is closed;
     * {@code false} othewise.
     */
	public boolean isClosed() {
		return file.isClosed();
	}

    /**
     * Writes a line of {@code #} symbols to this file.
     * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
     */
	public void breakPoint() {
		file.breakPoint();
	}
}
