/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.util.*;
import java.util.Map.Entry;

import sbrt.kernel.math.algebra.SparseDoubleMatrix;
import sbrt.kernel.math.algebra.TreeMatrix;
import sbrt.shell.text.*;

/**
 * This class is used to read column-oriented map input files.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public class ComiFile<K, V> extends AbstractMapInputFile<K, V> {

    /**
     * Reads the provided column-oriented map input file and returns its
     * contents as a sparse matrix. The row indices are numbered 1 to n, and
     * column indices are the map keys.
     * 
     * @param file the file to be read.
     * 
     * @return the contents of the provided file as a matrix.
     */
    public static SparseDoubleMatrix<Integer, String> 
        getMatrix(ComiFile<String, Double> file) {
        
        TreeMatrix<Integer, String> matrix = new TreeMatrix<Integer, String>();

        Integer rowIndex = 1;
        
        for (Map<String, Double> map : file) {
            
            for (Entry<String, Double> e : map.entrySet()) {
                matrix.setElement(rowIndex, e.getKey(), e.getValue());
            }

            rowIndex = rowIndex + 1;
        }
        
        return matrix;
    }
    
    
    private Set<K> keys;
    
    /**
     * Constructs a new row-oriented map input file from the
     * provided arguments.
     * 
     * @param fileName the name of the file to read.
     * @param fileFormat the format of the file.
     * @param requiredKeys the keys that are required to be
     * in the file.
     * @param mapFormat the format used to parse the file.
     */
    public ComiFile(String fileName,
                    FileFormat fileFormat,
                    Set<? extends K> requiredKeys,
                    MapFormat<K, V> mapFormat) {
        
    	// Calling super constructor
    	super(fileName, fileFormat, requiredKeys, mapFormat);
    }

    /**
     * Constructs a new row-oriented map input file from the
     * provided arguments.
     * 
     * @param fileName the name of the file to read.
     * @param fileFormat the format of the file.
     * @param mapFormat the format used to parse the file.
     */
    public ComiFile(String fileName,
                    FileFormat fileFormat,
                    MapFormat<K, V> mapFormat) {

    	// Calling super constructor
    	super(fileName, fileFormat, mapFormat);
    }

    /**
     * Constructs a new row-oriented map input file from the
     * provided arguments. The file format is assumed to be
     * {@linkplain FileFormat#TEXT}. 
     * 
     * @param fileName the name of the file to read.
     * @param mapFormat the format used to parse the file.
     */
    public ComiFile(String fileName,
                    MapFormat<K, V> mapFormat) {

        // Calling super constructor
        super(fileName, FileFormat.TEXT, mapFormat);
    }

    /**
     * Returns the map keys (column headers) contained in this file.
     * @return the map keys (column headers) contained in this file.
     */
    public Set<K> getKeys() {
        
        if (keys != null) {
            return keys;
        }
        
        TextFileLineIterator<String> i = lineIterator();
        
        keys = getKeys(i);
        
        i.close();
        
        return keys;
    }

    private LinkedHashSet<K> getKeys(TextFileLineIterator<String> i) {

        // If no elements are present
        if (!i.hasNext()) {
            String s = "The file '" + getCanonicalName() + "' appears to be empty.";
            throw new FormatException(s);
        }

        // Getting the first line (the header row)
        String line = i.next();

        // Getting the variables in the first line
        return parseHeader(line);
    }
    
    /**
     * Returns an iterator over this file.
     * @return an iterator over this file.
     */
    public Iterator<Map<K, V>> iterator() {
        return iterator(0);
    }

    /**
     * Returns an iterator over this file beginning at the specified
     * index. Indices begin at {@code 0}, and commented or blank lines are
     * ignored.
     * 
     * @param mapIndex the index of the map at which this iterator begins.
     * 
     * @return an iterator over this file beginning at the specified
     * index.
     * 
     * @throws IndexOutOfBoundsException if the provided index exceeds the
     * number of maps contained in the file.
     */
    public Iterator<Map<K, V>> iterator(int mapIndex) {

        // Creating a plain text iterator over the file
        final TextFileLineIterator<String> i = lineIterator();

        // Getting the variables in the first line
        final LinkedHashSet<K> variables = getKeys(i);

        // Getting the vector file line format
        final MapFormat<K, V> vectorFormat = getMapFormat();

        // Throwing away unwanted lines without parsing them
        for (int index = 0; index < mapIndex; index ++) {
            
            if (!i.hasNext()) {
                throw new IndexOutOfBoundsException();
            }
            
            // Throwing away next line
            i.next();
        }
        
        return new Iterator<Map <K, V>>() {

            public Map<K, V> next() {
                
                // Getting the next line of vector values
                String nextLine = i.next();
                
                // Parsing the list of values
                return vectorFormat.parseAllValues(variables, nextLine);  
            }

            public boolean hasNext() {
                return i.hasNext();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }
    
    private LinkedHashSet<K> parseHeader(String headerString) {

        // Getting vector file line format
        MapFormat<K, V> vectorFormat = getMapFormat();

        LinkedHashSet<K> allVariables = null;
        
        // Parsing header string
        try {
            allVariables = vectorFormat.parseAllKeys(headerString);
        } catch (FormatException e) {
            String s = "An error occurred while parsing the file '" +
                        getCanonicalName() + "'. " + e.getMessage();
            throw new FormatException(s);
        }

        // Iterating thru the required variables
        for (K variable : getRequiredKeys()) {

            // If the parsed header string is missing required variables 
            if (!allVariables.contains(variable)) {

                String s = "The file '" + getCanonicalName() +
                           "' does not contain an entry for '" +
                           vectorFormat.formatKey(variable) + "'.";
                throw new FormatException(s);
            }

        }

        return allVariables;
    }

}
