/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.util.*;

import sbrt.shell.text.*;

/**
 * This class is used to write maps to text files in a 
 * column-oriented way.
 *  
 * @author This class was written and documented by 
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public class ComoFile<K, V> extends AbstractMapOutputFile<K, V> {

    /**
     * Constructs a new column-oriented map output file 
     * from the provided arguments.
     * 
     * @param fileName the name of the file to create.
     * 
     * @param fileFormat the format of the file to create.
     * 
     * @param columnHeaderVariables the variables whose values 
     * will be
     * written to this file. If {@code null}, all values
     * from every added vector will be written to this file.
     * 
     * @param vectorFormat the vector format.
     * 
     * @throws NullPointerException if any argument except 
     * {@code variables} is {@code null}.
     */
    public ComoFile(String fileName,
        	        FileFormat fileFormat,
        	        Set<? extends K> columnHeaderVariables,
        	        MapFormat<K, V> vectorFormat) {
    	
    	// Calling constructor of super class
        super(fileName, fileFormat, columnHeaderVariables, vectorFormat);
        
        // Writing column headers
        writeColumnHeaders();
    }

    /**
     * Constructs a new column-oriented vector output file 
     * from the provided arguments.
     * 
     * @param fileName the name of the file to create.
     * 
     * @param fileFormat the format of the file to create.
     * 
     * @param allColumnHeaderVariables a string representation of the 
     * variables whose values will be written to this file. This 
     * string is parsed by the provided vector format.
     * 
     * @param vectorFormat the vector format.
     * 
     * @throws NullPointerException if any argument is {@code null}.
     */
    public ComoFile(String fileName,
    		        FileFormat fileFormat,
    		        String allColumnHeaderVariables,
    		        MapFormat<K, V> vectorFormat) {
    	
    	// Calling constructor of super class
        super(fileName, fileFormat, allColumnHeaderVariables, vectorFormat);
        
        // Writing column headers
        writeColumnHeaders();
    }

    /**
     * Constructs a new text-based column-oriented map output file 
     * from the provided arguments.
     * 
     * @param fileName the name of the file to create.
     * 
     * @param columnHeaderVariables the variables whose values 
     * will be written to this file. 
     * 
     * @param vectorFormat the vector format.
     * 
     * @throws NullPointerException if any argument except 
     * {@code variables} is {@code null}.
     */
    public ComoFile(String fileName,
                    Set<? extends K> columnHeaderVariables,
                    MapFormat<K, V> vectorFormat) {
        
        // Calling constructor of super class
        super(fileName, FileFormat.TEXT, columnHeaderVariables, vectorFormat);
        
        // Writing column headers
        writeColumnHeaders();
    }

    private void writeColumnHeaders() {

    	// Getting the map format
        MapFormat<K, V> format = getMapFormat();
        
        // Getting the "filter" keys
        Set<K> keys = getFilterKeys();

        // Writing them to this file
        writeLine(format.formatKeys(keys));
    }

	/**
	 * Writes the appropriate values of the provided vector 
	 * to this file. The appropriate values are those of the
	 * variables in the set returned by {@linkplain #getFilterKeys()}.
	 * 
	 * @param map the vector to be written to this file.
	 * 
	 * @throws IllegalArgumentException if the provided
	 * vector does not contain all the variables in the
	 * set returned by {@linkplain #getFilterKeys()}.
	 * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
	 */
    public void addData(Map<? extends K, ? extends V> map) {

    	// Creating storage for values
        List<V> values = new LinkedList<V>();

        // Iterating thru each "filter" key
        for (K key : getFilterKeys()) {

        	// Getting its value
            V value = map.get(key);

            // Ensuring a value exists in the provided vector
            if (value == null) {
                String s = "no value for: " + key;
                throw new IllegalArgumentException(s);
            }

            // Adding it to the rest
            values.add(value);
        }

        // Getting the map format
        MapFormat<K, V> mapFormat = getMapFormat();
        
        // Writting the formatted values to this file
        writeLine(mapFormat.formatValues(values));
    }

}
