/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.util.*;

import sbrt.kernel.io.*;
import sbrt.kernel.utilities.*;

/**
 * This class is used to create multiple output files.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <F> the file type.
 * @param <R> the result type. 
 *
 */
public class MultiOutputFiles<F extends OutputFile<R>, R> 
	implements ResultsManager<R> {

	// The file name source
    private Iterator<F> fileSource;

    // The current file
    private F currentFile;
    
    /**
     * Constructs a new multiple output files results manager.
     * 
     * @param fileSource a source of output files.
     */
    public MultiOutputFiles(Iterator<F> fileSource) {
    	
    	// Checking and saving
        Util.nullCheck(fileSource);
        this.fileSource = fileSource;
        
        // Getting the first file to fail-fast if necessary
        currentFile = fileSource.next();
    }

    /**
     * Closes this results manager so that no further
     * results can be added. If already closed, then
     * invoking this method has no effect.
     */
    public void close() {
    	
    	if (currentFile != null) {
    		
    		currentFile.close();
    		currentFile = null;
    		
    	}
    	
        fileSource = null;
    }

    /**
     * Indicates if this results manager has already 
     * been closed.
     * 
     * @return {@code true} if results can still be
     * added to this manager; {@code false} otherwise. 
     * 
     * @see #close()
     */
    public boolean isClosed() {
        return fileSource == null;
    }

    /**
     * Adds the provided results to the current file. If a current
     * file does not already exist, one is created and the 
     * provided results are added to it.
     * 
     * @param <S> a type extending the primary results type.
     * 
     * @param results the results to add to the current file.
     */
	public synchronized <S extends R> void addData(S results) {

		// Checking the state of this results manager
		checkState();
		
		// Getting a new file if one doesn't already exist
		if (currentFile == null) {
			currentFile = fileSource.next();
		}
		
		// Adding the data to it
		currentFile.addData(results);
	}

	/**
	 * Returns the name of the current file.
	 * @return the name of the current file, or {@code null} if one
	 * does not exist.
	 */
	public synchronized String getName() {
		return currentFile == null ? null : currentFile.getName();
	}

    /**
     * Closes the current file if one exists.
     */
    public synchronized void breakPoint() {
    	
    	// Closing the current file if it exists
    	if (currentFile != null) {
    		currentFile.close();
            currentFile = null;
    	}
    	
    }

    private synchronized void checkState() {
    	
    	if (isClosed()) {
    		String s = "This results manager is already closed.";
    		throw new IllegalStateException(s);
    	}
    }
}
