/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.util.*;

import sbrt.shell.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.*;

/**
 * This class is used to represent process input files.
 * These are the files that can be given as a single command line
 * argument to {@linkplain sbrt#SystemsBiologyResearchTool}.
 * The are used to specify all the details required to construct
 * a process using a process manager.  
 *  
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class ProcessInputFile extends
		RomiFile<String, String> implements
			ManagerInput<String, String> {

	private static final Set<String> requiredKeywords;
	static {
		requiredKeywords = new LinkedHashSet<String>();
		requiredKeywords.add(ProcessManagerFactory.PROCESS);
	}

	/**
	 * Constructs a new process input file from the
	 * provided arguments.
	 * 
	 * @param fileName the name of the file to be read.
	 * @param lineFormat the format used to parse its lines.
	 */
	public ProcessInputFile(String fileName,
			AppInputFileLineFormat lineFormat) {

		super(fileName, FileFormat.TEXT, requiredKeywords, lineFormat);
	}

	/**
	 * Constructs a new process input file from the
	 * provided argument.
	 * 
	 * @param fileName the name of the file to be read.
	 */
	public ProcessInputFile(String fileName) {
		this(fileName, new AppInputFileLineFormatVersion1());
	}

    /**
     * Returns the information contained in this file as a map.
     * @return keywords mapped to their respective values.
     */
	public Map<String, String> toMap() {

		// Getting the vector of input values
		Map<String, String> vector = getMap();

		// Determining appropriate size
		int size = vector.size() + 1;
		
		// Creating storage
		Map<String, String> input =
			new LinkedHashMap<String, String>(size, 1.0F);
		
		// Copying
		input.putAll(vector);

		return input;
	}

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     */
    public final String getName() {
        return getCanonicalName();
    }

	/**
	 * Indicates if a value exists for the provided key
	 * in this manager input.
	 * 
	 * @param key the key whose presence in this input 
	 * is to be tested.
	 * 
	 * @return {@code true} if this input contains a 
	 * value for the specified key.
	 */
	public boolean hasValueFor(Object key) {
		return getMap().containsKey(key);
	}

    /**
     * Returns the value corresponding to the provided key.
     * 
     * @param key the key whose value is to be returned.
     * 
     * @return the value corresponding to the provided key, or 
     * {@code null} if the value does not exist.
     */
	public String getOptionalValue(Object key) {
		return getMap().get(key);
	}

    /**
     * Returns the value corresponding to the provided key.
     * 
     * @param key the key whose value is to be returned.
     * 
     * @return the value corresponding to the provided key.
     * 
     * @throws sbrt.shell.ApplicationException if the value does not exist.
     */
	public String getRequiredValue(Object key) {

		String entry = getMap().get(key);

		if (entry == null) {
			String s = "The input file '" + getName()
					+ "' is missing an entry for '" + key + "'.";
			throw new FormatException(s);
		}

		return entry;
	}

	/**
	 * Returns a string representation of the information contained
	 * in this file.
	 * @return a string representation of the information contained
	 * in this file.
	 */
	public String toString() {

		// Getting input vector
		Map<String, String> entries = getMap();
		
		// Creating formatter
		MapFormat<String, String> format = getMapFormat();

		// Getting system-dependent line separator
		String newLine = System.getProperty("line.separator");

		// Creating storage for entire string
		StringBuilder s = new StringBuilder(entries.size() * 1000);

		// Creating storage for just the header
		StringBuilder header = new StringBuilder(100);

		// Appending file name
		header.append("# " + "Process Specifications" + " #");

		// Padding
		while (header.length() <= 78) {
			header.insert(0, '#');
			header.append('#');
		}

		// Appending to the entire string
		s.append(header + newLine);

		// Iterating thru each file entry
		for (Map.Entry<String, String> e : entries.entrySet()) {

			// Formatting entry
			String entry = format.formatPair(e);

			// Appending to the entire string
			s.append(entry + newLine);
		}

		// Appending final line
		for (int i = 0; i < header.length() && i < 80; i++) {
			s.append('#');
		}

		// Returning formatted string
		return s.toString();
	}

}
