/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.util.*;

import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;

/**
 * This class is used to read row-oriented map input files. These files contain
 * only a single map.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <K> the key type.
 * @param <V> the value type.
 */
public class RomiFile<K, V> extends AbstractMapInputFile<K, V> {

    // The map contained in this file
    private UnmodifiableMap<K, V> map;

    /**
     * Constructs a new row-oriented map input file from the provided arguments.
     * 
     * @param fileName the name of the file to read.
     * @param fileFormat the format of the file.
     * @param requiredKeys the keys that are required to be in the file.
     * @param mapFormat the format used to parse the file.
     */
    public RomiFile(String fileName, FileFormat fileFormat,
            Set<? extends K> requiredKeys, MapFormat<K, V> mapFormat) {

        super(fileName, fileFormat, requiredKeys, mapFormat);
    }

    /**
     * Constructs a new row-oriented map input file from the provided arguments.
     * 
     * @param fileName the name of the file to read.
     * @param fileFormat the format of the file.
     * @param mapFormat the format used to parse the file.
     */
    public RomiFile(String fileName, FileFormat fileFormat,
            MapFormat<K, V> mapFormat) {

        super(fileName, fileFormat, mapFormat);
    }

    /**
     * Constructs a new row-oriented map input file from the provided arguments.
     * 
     * @param fileName the name of the file to read.
     * @param mapFormat the format used to parse the file.
     */
    public RomiFile(String fileName, MapFormat<K, V> mapFormat) {
        super(fileName, FileFormat.TEXT, mapFormat);
    }

    /**
     * Constructs a new row-oriented map input file from the provided arguments.
     * 
     * @param fileName the name of the file to read.
     * @param requiredKeys the keys that are required to be in the file.
     * @param mapFormat the format used to parse the file.
     */
    public RomiFile(String fileName, Set<? extends K> requiredKeys,
            MapFormat<K, V> mapFormat) {
        super(fileName, FileFormat.TEXT, requiredKeys, mapFormat);
    }

    /**
     * Returns the map contained in this file.
     * 
     * @return the map contained in this file.
     * 
     * @throws FormatException if the file does not contain a required variable,
     * contains a single variable multiple times, or contains any other formatting errors.
     */
    public UnmodifiableMap<K, V> getMap() {
        if (map == null) {
            map = parseFile();
        }
        return map;
    }

    /**
     * Returns the keys of the map contained in this file.
     * 
     * @return the keys of the map contained in this file.
     * 
     * @throws FormatException if the file does not contain a required variable,
     * contains a single variable multiple times, or contains any other formatting errors.
     */
    public UnmodifiableSet<K> getKeys() {
        return getMap().keySet();
    }

    private UnmodifiableMap<K, V> parseFile() {

        // Getting the formatter
        MapFormat<K, V> mapFormat = getMapFormat();

        // Getting the required keys
        Set<K> requiredKeys = getRequiredKeys();

        // Creating storage
        Map<K, V> m = new LinkedHashMap<K, V>(requiredKeys.size() + 1);

        // Iterating thru each line in the file
        for (Iterator<String> i = lineIterator(); i.hasNext();) {

            // Getting the next line
            String line = i.next();

            // Creating a field for storage
            Map.Entry<K, V> entry = null;

            try {

                // Parsing the entry
                entry = mapFormat.parsePair(line);

            } catch (FormatException e) {
                String s =
                        "A problem occurred while parsing the line '" + line
                                + "' in the file '" + getCanonicalName() + "'. "
                                + e.getMessage();
                throw new FormatException(s);
            }

            // Getting the variable
            K variable = entry.getKey();

            // If this variable was already encountered in the file
            if (m.containsKey(variable)) {
                String s =
                        "The entry '" + mapFormat.formatKey(variable)
                                + "' appears more than once in the file '"
                                + getCanonicalName() + "'.";
                throw new FormatException(s);
            }

            // Saving the pair
            m.put(variable, entry.getValue());
        }

        // Iterating thru each required variable
        for (K variable : requiredKeys) {

            // If the variable was not encountered
            if (!m.containsKey(variable)) {
                String s =
                        "The file '" + getCanonicalName()
                                + "' does not contain an entry for '"
                                + mapFormat.formatKey(variable) + "'.";
                throw new FormatException(s);
            }

        }

        // Returning the vector contained in this vector
        return new UnmodifiableMap<K, V>(m);
    }

    /**
     * Returns an iterator over this file. This iterator returns only a single
     * map.
     * 
     * @return an iterator over this file.
     * 
     * @see #getMap()
     */
    public Iterator<Map<K, V>> iterator() {

        return new Iterator<Map<K, V>>() {

            private boolean hasNext = true;

            public boolean hasNext() {
                return hasNext;
            }

            public Map<K, V> next() {

                if (!hasNext()) {
                    throw new NoSuchElementException();
                }

                hasNext = false;

                return getMap();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };

    }

}
