/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.util.*;

import sbrt.shell.text.*;

/**
 * This class is used to write maps to text files in a 
 * row-oriented way.
 *  
 * @author This class was written and documented by 
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public class RomoFile<K, V> extends AbstractMapOutputFile<K, V> {

    /**
     * Constructs a new row-oriented map output file 
     * from the provided arguments.
     * 
     * @param fileName the name of the file to create.
     * 
     * @param fileFormat the format of the file to create.
     * 
     * @param rowHeaderVariables the variables whose values will be
     * written to this file. If {@code null}, all values
     * from every added vector will be written to this file.
     * 
     * @param mapFormat the vector format.
     * 
     * @throws NullPointerException if any argument except 
     * {@code rowHeaderVariables} is {@code null}.
     */
    public RomoFile(String fileName,
        FileFormat fileFormat,
        Set<? extends K> rowHeaderVariables,
        MapFormat<K, V> mapFormat) {
    	
        super(fileName, fileFormat, rowHeaderVariables, mapFormat);
    }

    /**
     * Constructs a new row-oriented vector output file 
     * from the provided arguments.
     * 
     * @param fileName the name of the file to create.
     * 
     * @param fileFormat the format of the file to create.
     * 
     * @param allRowHeaderVariables a string representation of the 
     * variables whose values will be written to this file. This 
     * string is parsed by the provided vector format.
     * 
     * @param vectorFormat the vector format.
     * 
     * @throws NullPointerException if any argument is {@code null}.
     */
    public RomoFile(String fileName,
    		FileFormat fileFormat,
    		String allRowHeaderVariables,
    		MapFormat<K, V> vectorFormat) {
    	
        super(fileName, fileFormat, allRowHeaderVariables, vectorFormat);
    }

    /**
     * Constructs a new row-oriented vector output file from the
     * provided arguments. The value of each variable from every added
     * vector will be written to this file.
     * 
     * @param fileName the name of the file to create.
     * @param fileFormat the format of the file to create.
     * @param vectorFormat the vector format.
     * 
     * @throws NullPointerException if any argument is {@code null}.
     */
    public RomoFile(String fileName,
    		FileFormat fileFormat,
    		MapFormat<K, V> vectorFormat) {
    	
        super(fileName, fileFormat, vectorFormat);
    }

    /**
     * Constructs a new plain text, row-oriented vector output 
     * file from the provided arguments. The value of each variable 
     * from every added vector will be written to this file.
     * 
     * @param fileName the name of the file to create.
     * @param vectorFormat the vector format.
     * 
     * @throws NullPointerException if either argument is {@code null}.
     */
    public RomoFile(String fileName, MapFormat<K, V> vectorFormat) {
    	super(fileName, FileFormat.TEXT, vectorFormat);
    }

	/**
	 * Writes the provided value to this file if the variable is
	 * contained in the set returned by {@linkplain #getFilterKeys()}.
	 * 
	 * @param variableValuePair the variable-value pair.
	 * 
	 * @throws IllegalArgumentException if the provided
	 * vector does not contain all the variables in the
	 * set returned by {@linkplain #getFilterKeys()}.
	 * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
	 */
    public void addData(Map.Entry<? extends K, ? extends V> variableValuePair) {

    	// Getting the vector format
        MapFormat<K, V> vectorFormat = getMapFormat();
        
        // Getting the "filter" variables
        Set<K> variables = getFilterKeys();

        // If none exist, write the pair and return
        if (variables == null) {
            writeLine(vectorFormat.formatPair(variableValuePair));
            return;
        }

        // If the provided variable is in the "filter" set,
        // write the pair to this file
        if (variables.contains(variableValuePair.getKey())) {
            writeLine(vectorFormat.formatPair(variableValuePair));
        }

    }

	/**
	 * Writes the appropriate values of the provided vector 
	 * to this file. The appropriate values are those of the
	 * variables in the set returned by {@linkplain #getFilterKeys()}.
	 * 
	 * @param vector the vector to be written to this file.
	 * 
	 * @throws IllegalArgumentException if the provided
	 * vector does not contain all the variables in the
	 * set returned by {@linkplain #getFilterKeys()}.
	 * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
	 */
    public void addData(Map<? extends K, ? extends V> vector) {

    	// Getting the vector format
        MapFormat<K, V> vectorFormat = getMapFormat();
        
        // Getting the variables whose values are to be written
        Set<K> variables = getFilterKeys();
        
        if (variables == null) {

        	// Writing every variable-value pair
        	for (Map.Entry<? extends K, ? extends V> e : vector.entrySet()) {
                writeLine(vectorFormat.formatPair(e));
            }

        	// Exiting from this method
            return;
        }

        // Iterating thru each variable in the set
        for (K variable : variables) {

        	// Getting the corresponding value
            V value = vector.get(variable);

            // Ensuring a value exists for this variable
            if (value == null) {
                String s = "no value for: " + variable;
                throw new IllegalArgumentException(s);
            }

            // Writing the formatted variable-value line
            writeLine(vectorFormat.formatPair(variable, value));
        }

    }


}
