/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.io.*;
import java.util.*;
import java.util.zip.*;

import sbrt.shell.*;

/**
 * This class is used to iterate over the objects in a serialized
 * input file.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <E> the returned type.
 */
public final class SerializedFileIterator<E> extends FileLineIterator<E> {

	/**
	 * Indicates if the provided file format is valid for
	 * this type of file iterator.
	 * 
	 * @param fileFormat the file format to test.
	 * 
	 * @return {@code true} if the provided file format is 
	 * valid for a serialized input file; {@code false} otherwise.
	 */
    public static boolean isValidFormat(FileFormat fileFormat) {
        return fileFormat == FileFormat.SERIALIZED ||
                fileFormat == FileFormat.GZIP_SER;
    }

    // The input file
    private final SerializedInputFile<E> file;

    // The object reader
    private ObjectInputStream reader;
    
    // The next object in this file
    private Object nextObject;

    /**
     * Constructs a new serialized file iterator using the
     * provided serialized input file.
     * 
     * @param file the file to be iterated over.
     */
    public SerializedFileIterator(SerializedInputFile<E> file) {

        String fileName = file.getCanonicalName();
        
        // Creating a Java file
        java.io.File f = new java.io.File(fileName);

        // If the file does not exist
        if (!f.exists()) {
            String s = "The file '" + fileName + "' does not exist.";
            throw new ApplicationException(s);
        }

        // If the file is not normal
        if (!f.isFile()) {
            String s = "The file '" + fileName +
                       "' is not a normal file.";
            throw new ApplicationException(s);
        }

        // If the file can't be read
        if (!f.canRead()) {
            String s = "The file '" + fileName + "' cannot be read.";
            throw new ApplicationException(s);
        }

        // Storing file
        this.file = file;

        try {

            // Setting reader
            reader = getReader(f, file.getFileFormat());

            // Setting first object
            nextObject = nextObject();

        } catch (IOException e) {
            String s = "An I/O error occurred while opening " +
                       "the file '" + fileName + "'.";
            throw new ApplicationException(s, e);
        }

    }

    private ObjectInputStream getReader(java.io.File file,
                                        FileFormat format) throws IOException {

        if (format == FileFormat.SERIALIZED) {

            // Creating FileInputStream
            InputStream fileIn = new FileInputStream(file);

            // Wrapping with a BufferedInputStream
            InputStream buffIn = new BufferedInputStream(fileIn);

            // Wrapping with an ObjectInputStream
            return new ObjectInputStream(buffIn);
        }

        else if (format == FileFormat.GZIP_SER) {

            // Creating FileInputStream
            InputStream fileIn = new FileInputStream(file);

            // Wrapping with a BufferedInputStream
            InputStream buffIn = new BufferedInputStream(fileIn);

            // Wrapping with a GZIPInputStream
            InputStream gzipIn = new GZIPInputStream(buffIn);

            // Wrapping with an ObjectInputStream
            return new ObjectInputStream(gzipIn);
        }

        else {
            String s = "The file format '" + format + "' is unsupported.";
            throw new IllegalArgumentException(s);
        }

    }

    /**
     * Indicates if the file has more serialized objects.
     * @return {@code true} if the file has more objects;
     * {@code false} otherwise.
     */
    public boolean hasNext() {

        // Checking state
        if (isClosed()) {
            return false;
        }

        return nextObject != null;
    }

    /**
     * Returns the next object in the file.
     * @return the next object in the file.
     * 
     * @throws NoSuchElementException if the file 
     * has no more objects.
     */
    @SuppressWarnings("unchecked")
	public E next() {

        // Checking state
        if (!hasNext()) {
            throw new NoSuchElementException();
        }

        // Saving current object
        E temp = (E) nextObject;

        // Getting next object
        nextObject = nextObject();

        // Closing file if no more lines exist
        if (nextObject == null) {
            close();
        }

        // Returning current object
        return temp;
    }

    /**
     * Returns the next object in the file.
     * @return the next object in the file.
     * 
     * @throws ApplicationException if an error occurs while
     * the file is being read.
     * 
     * @throws NoSuchElementException if the file 
     * has no more objects.
     */
    protected final Object nextObject() {

        try {

            // Reading the next object
            Object obj = reader.readObject();

            // If it's the end-of-file marker, return null
            if (SerializedOutputFile.isEof(obj)) {
                return null;
            }

            // Otherwise just return the object
            return obj;

        } catch (EOFException e) {

            String s = "The file '" + file.getCanonicalName() + "' is corrupted.";
            throw new ApplicationException(s, e);

        } catch (IOException e) {

            String s = "An I/O error occurred while reading the " +
                       "file '" + file.getCanonicalName() + "'.";
            throw new ApplicationException(s, e);

        } catch (ClassNotFoundException ex) {

            String s = "The JVM was unable to find a class during " +
                       "deserialization of the file '" + file.getCanonicalName() + "'.";
            throw new ApplicationException(s, ex);

        }

    }

    /**
     * Throws UnsupportedOperationException.
     * @throws UnsupportedOperationException
     */
    public void remove() {
        throw new UnsupportedOperationException();
    }

    /**
     * Indicates if this iterator has been closed.
     * 
     * @return {@code true} if {@linkplain #close()} has
     * been called; {@code false} otherwise.
     */
    public boolean isClosed() {
        return reader == null;
    }

    /**
     * Closes this file iterator. If this file iterator is already 
     * closed, then invoking this method has no effect.
     * 
     * @throws ApplicationException if an I/O error occurred
     * while closing this file iterator.
     */
    public void close() {

        // Do nothing if this iterator is already closed
        if (isClosed()) {
            return;
        }

        try {

            reader.close();
            reader = null;

        } catch (IOException e) {

            String s = "An I/O error occurred while closing the file '"
                       + file.getCanonicalName() + "'.";
            throw new ApplicationException(s, e);

        }

    }

    /**
     * Disposes of system resources associated with this file
     * iterator when called by the garbage collector.
     * 
     * @throws Throwable the Exception raised by this method.
     * 
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        try {
            close();
        } finally {
            super.finalize();
        }
    }

}
