/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.io.File;
import java.util.*;

import sbrt.kernel.utilities.*;
import sbrt.shell.ApplicationException;

/**
 * This class is used to read input files containing serialized
 * objects.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <E> the type contained in instances of this interface.
 */
public class SerializedInputFile<E> implements InputFile<E> {

	/**
	 * Indicates if the provided file format is valid for
	 * a text input file.
	 * 
	 * @param fileFormat the file format to test.
	 * 
	 * @return {@code true} if the provided file format is 
	 * valid for a text input file; {@code false} otherwise.
	 */
    public static boolean isValidFormat(FileFormat fileFormat) {
        return SerializedFileIterator.isValidFormat(fileFormat);
    }

    // The name of the file being read
    private final String fileName;
    
    // The format of the file being read
    private final FileFormat fileFormat;
    
    // The list of iterators over this file
    private List<SerializedFileIterator<?>> fileIterators;

    /**
     * Constructs a new text input file using the provided argument.
     * 
     * @param fileName the name of the file to be read.
     */
    public SerializedInputFile(String fileName) {
        this(fileName, FileFormat.SERIALIZED);
    }

    /**
     * Constructs a new text input file using the provided arguments.
     * 
     * @param fileName the name of the file to be read.
     * @param fileFormat the format of the file to be read.
     */
    public SerializedInputFile(String fileName, FileFormat fileFormat) {

        // Checking inputs
        Util.nullCheck(fileName);
        Util.nullCheck(fileFormat);

        // Ensuring file format is appropriate
        if (!isValidFormat(fileFormat)) {
            String s = "Unsupported file format: " + fileFormat;
            throw new IllegalArgumentException(s);
        }

        // Storing
        this.fileName = fileName;
        this.fileFormat = fileFormat;

        // Creating storage for file iterators
        fileIterators = new LinkedList<SerializedFileIterator<?>>();
    }

    /**
     * Returns the format of this file.
     * @return the format of this file.
     */
    public FileFormat getFileFormat() {
        return fileFormat;
    }

    /**
     * Returns the original name given to this file object.
     * @return the original name given to this file object.
     */
    public final String getOriginalName() {
        return fileName;
    }

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     * 
     * @see File#getCanonicalPath()
     */
    public String getCanonicalName() {
        return AbstractSbrtFile.getCanonicalName(fileName);
    }

    /**
     * Returns an object iterator over this file.
     * @return an object iterator over this file.
     */
    protected final SerializedFileIterator<?> objectIterator() {

        // Checking state
        if (isClosed()) {
            throw new IllegalStateException("already closed: " + fileName);
        }

        // Creating a new file iterator for this file
        SerializedFileIterator<?> i = new SerializedFileIterator<E>(this);

        // Storing it
        fileIterators.add(i);

        // Returning it
        return i;
    }

    /**
     * Returns an object iterator over this file.
     * @return an object iterator over this file.
     */
    @SuppressWarnings("unchecked")
	public Iterator<E> iterator() {
        return (Iterator<E>) objectIterator();
    }

    /**
     * Closes this file. If this file is already closed, then
     * invoking this method has no effect.
     * 
     * @throws ApplicationException if an I/O error occurred
     * while closing this file.
     */
    public final void close() {

        // Do nothing if the file is already closed
        if (isClosed()) {
            return;
        }

        // Iterating thru each file iterator
        for (SerializedFileIterator<?> nextIterator : fileIterators) {

            // Closing it
            nextIterator.close();
        }

        // Destroying the reference to the list
        fileIterators = null;
    }

    /**
     * Indicates if this file has already been closed.
     * 
     * @return {@code true} if this file is closed;
     * {@code false} othewise.
     */
    public boolean isClosed() {
        return fileIterators == null;
    }

    /**
     * Disposes of system resources associated with this file
     * when called by the garbage collector.
     * 
     * @throws Throwable the Exception raised by this method.
     * 
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        try {
            close();
        } finally {
            super.finalize();
        }
    }

}
