/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.util.*;

import sbrt.kernel.utilities.*;
import sbrt.shell.*;

/**
 * This class is used to read input files containing serialized
 * maps.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public class SerializedMapInputFile<K, V> 
	extends SerializedInputFile<Map<K, V>>
        implements MapInputFile<K, V> {

	// The required keys
    private final UnmodifiableSet<K> requiredKeys;

    /**
     * Constructs a new input file using the provided arguments.
     * 
     * @param fileName the name of the file to be read.
     * @param fileFormat the format of the file to be read.
     */
    public SerializedMapInputFile(String fileName, FileFormat fileFormat) {

    	super(fileName, fileFormat);
        
        // Getting an empty set for the required variables
        Set<K> empty = Collections.emptySet();
        
        // Storing as unmodifiable
        this.requiredKeys = new UnmodifiableSet<K>(empty);
    }

    /**
     * Constructs a new input file using the provided arguments.
     * 
     * @param fileName the name of the file to be read.
     * @param fileFormat the format of the file to be read.
     * @param requiredKeys the keys that are required to be
     * in the file.
     */
    public SerializedMapInputFile(String fileName,
                                  FileFormat fileFormat,
                                  Set<? extends K> requiredKeys) {
        super(fileName, fileFormat);

        // Determining appropriate size
        int size = requiredKeys.size() + 1;
        
        // Creating storage
        Set<K> s = new LinkedHashSet<K>(size, 1.0F);
        
        // Copying
        s.addAll(requiredKeys);
        
        // Storing as unmodifiable
        this.requiredKeys = new UnmodifiableSet<K>(s);
    }

    /**
     * Returns an iterator over this file.
     * @return an iterator over this file.
     */
    @SuppressWarnings("unchecked")
	public Iterator<Map<K, V>> iterator() {

        // Getting an object iterator
        final Iterator<List<?>> i = 
        	(Iterator<List<?>>) objectIterator();

        // If no elements exist
        if (!i.hasNext()) {
            String s = "The file '" + getCanonicalName() + "' appears to be empty.";
            throw new ApplicationException(s);
        }

        // Getting the map keys (the first entry in the file)
        final List<K> mapKeys = (List<K>) i.next();


        // If the file does not contain all required variables
        if (!mapKeys.containsAll(requiredKeys)) {
            String s = "The file '" + getCanonicalName() + "' does not " +
                       "contain at least one of the provided headers.";
            throw new ApplicationException(s);
        }

        return new Iterator<Map<K, V>>() {

            public Map<K, V> next() {

            	// Determining appropriate size
            	int size = mapKeys.size() + 1;
            	
                // Creating storage for the vector
                Map<K, V> vector = new LinkedHashMap<K, V>(size, 1.0F);

                // Getting the next list of values
                List<V> values = (List<V>) i.next();

                // Getting iterators for map keys and values
                Iterator<K> i = mapKeys.iterator();
                Iterator<V> j = values.iterator();

                // Filling vector with values
                while (i.hasNext()) {

                    // If the next value does not exist
                    if (!j.hasNext()) {
                        String s = "A value is missing in the file '" +
                                   getCanonicalName() + "'. The file may be corrupted.";
                        throw new ApplicationException(s);
                    }

                    // Storing key-value pair
                    vector.put(i.next(), j.next());
                }

                // If another value still remains
                if (j.hasNext()) {
                    String s = "An extra value was found in the file '" +
                               getCanonicalName() + "'. The file may be corrupted.";
                    throw new ApplicationException(s);
                }

                // Returning the next vector
                return vector;
            }

            public boolean hasNext() {
                return i.hasNext();
            }

            public void remove() {
                i.remove();
            }

        };

    }

}
