/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.io.*;
import java.util.zip.*;

import sbrt.kernel.utilities.*;
import sbrt.shell.*;

/**
 * This class is used to write serialized objects to output files.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <E> the object type.
 */
public class SerializedOutputFile<E> implements OutputFile<E> {

    // Object used to mark end-of-file
    private static final EndOfFile EOF = EndOfFile.get();

    /**
     * Indicates if the provided object is a valid end-of-file marker.
     * 
     * @param obj the object to test.
     * 
     * @return {@code true} if the provided object is an
     * instance of {@linkplain EndOfFile}; {@code false} otherwise.
     */
    public static boolean isEof(Object obj) {
        return EndOfFile.isEof(obj);
    }

	/**
	 * Indicates if the provided file format is valid for
	 * a serialized output file.
	 * 
	 * @param fileFormat the file format to test.
	 * 
	 * @return {@code true} if the provided file format is 
	 * valid for a serialized output file; {@code false} otherwise.
	 */
    public static boolean isValidFormat(FileFormat fileFormat) {
        return fileFormat == FileFormat.SERIALIZED ||
                fileFormat == FileFormat.GZIP_SER;
    }

    // The name of this file
    private final String fileName;
    
    // The format of this file
    private final FileFormat fileFormat;
    
    // The output stream
    private ObjectOutputStream out;

    /**
     * Constructs a new serialized output file from the provided
     * arguments.
     * 
     * @param fileName the name of the file to be created.
     * @param fileFormat the format of the file to be created.
     */
    public SerializedOutputFile(String fileName, FileFormat fileFormat) {

        // Checking inputs
        Util.nullCheck(fileName);
        Util.nullCheck(fileFormat);

        // Storing
        this.fileName = fileName;
        this.fileFormat = fileFormat;

        // Getting the canonical name of the file
        String name = AbstractSbrtFile.getCanonicalName(fileName);
        
        try {

            // Creating Java file
            java.io.File f = new java.io.File(name);

            if (f.exists()) {
                String s = "The file '" + name + "' already exists.";
                throw new ApplicationException(s);
            }

            try {

                f.createNewFile();

            } catch (IOException e) {

                String s = "The file '" + name + "' cannot be created.";
                throw new ApplicationException(s);

            }

            // Setting output stream
            out = getWriter(f, fileFormat);

        } catch (IOException e) {

            String s = "An I/O error occurred while trying to " +
                       "create the file '" + name + "'.";
            throw new ApplicationException(s, e);

        } catch (SecurityException e) {

            String s = "A security violation occurred while trying to " +
                       "create the file '" + name + "'.";
            throw new ApplicationException(s, e);

        }
    }

    private ObjectOutputStream getWriter(java.io.File file,
                                         FileFormat fileFormat) throws
            IOException {

        if (fileFormat == FileFormat.SERIALIZED) {

            // Creating a FileOutputStream
            OutputStream fileOut = new FileOutputStream(file);

            // Wrapping with a BufferedOutputStream
            OutputStream buffOut = new BufferedOutputStream(fileOut);

            // Wrapping with an ObjectOutputStream
            return new ObjectOutputStream(buffOut);
        }

        else if (fileFormat == FileFormat.GZIP_SER) {

            // Creating a FileOutputStream
            OutputStream fileOut = new FileOutputStream(file);

            // Wrapping with a BufferedOutputStream
            OutputStream buffOut = new BufferedOutputStream(fileOut);

            // Wrapping with a GZIPOutputStream
            OutputStream gzipOut = new GZIPOutputStream(buffOut);

            // Wrapping with an ObjectOutputStream
            return new ObjectOutputStream(gzipOut);
        }

        else {
            String s = "The file format '" + fileFormat + "' is unsupported.";
            throw new IllegalArgumentException(s);
        }

    }

    /**
     * Returns the format of this file.
     * @return the format of this file.
     */
    public FileFormat getFileFormat() {
    	return fileFormat;
    }

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     */
    public final String getName() {
        return getCanonicalName();
    }

    /**
     * Returns the name of this file.
     * @return the name of this file.
     */
    public final String getOriginalName() {
        return fileName;
    }

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     * 
     * @see File#getCanonicalPath()
     */
    public final String getCanonicalName() {
        return AbstractSbrtFile.getCanonicalName(fileName);
    }

	/**
	 * Adds data to this output file. The provided data
	 * is serialized and written to this file.
	 * 
	 * @param <S> a type extending the primary data type
	 * contained in this output file.
	 * 
	 * @param results the results, or data, to be added
	 * to this output file.
	 * 
     * @throws ApplicationException if an I/O error occurs
     * while writing data to this file.
	 * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
	 * 
	 * @throws NullPointerException if the provided 
	 * argument is {@code null}.
	 */
    public <S extends E> void addData(S results) {
 
        // Checking state
        if (isClosed()) {
            throw new IllegalStateException("already closed: " + fileName);
        }

        // Writing to file
        writeObject(results);
    }

    /**
     * Serializes the provided object and writes it to this file.
     * 
     * @param obj the object to be serialized and written to this file.
     * 
     * @throws ApplicationException if an I/O error occurs while 
     * writing to file.
     * 
     * @throws java.io#InvalidClassException if something is wrong with a 
     * class used by serialization.
     * 
     * @throws java.io#NotSerializableException if some object to be 
     * serialized does not implement the java.io.Serializable interface.
     */
    protected final void writeObject(Object obj) {

        try {

            // Writing object and flushing
            out.writeObject(obj);
            out.flush();

        } catch (IOException e) {

            // Closing file to (hopefully) preserve already written data
            close();

            // Throwing application exception
            String s = "An I/O error occurred while writing data to the " +
                       "file '" + getName() + "'.";
            throw new ApplicationException(s, e);
        }

    }
    
    /**
     * Does nothing.
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
     */
    public void breakPoint() {

        // Checking state to satisfy the method's contract
        if (isClosed()) {
            throw new IllegalStateException("already closed: " + fileName);
        }

        // Do nothing
    }

    /**
     * Closes this file. If this file is already closed, then
     * invoking this method has no effect.
     * 
     * @throws ApplicationException if an I/O error occurred
     * while closing this file.
     */
    public void close() {

        // Checking state
        if (isClosed()) {
            return; // Do nothing if already closed
        }

        try {

            // Writing end-of-file marker
            out.writeObject(EOF);

            // Flushing and closing
            out.flush();
            out.close();

            // Changing state to 'closed'
            out = null;

        } catch (IOException e) {
            String s = "An I/O error occured while trying to close the " +
                       "file '" + fileName + "'.";
            throw new ApplicationException(s, e);
        }
    }

    /**
     * Indicates if this file has already been closed.
     * 
     * @return {@code true} if this file is closed;
     * {@code false} othewise.
     */
    public final boolean isClosed() {
        return out == null;
    }

    /**
     * Disposes of system resources associated with this file
     * when called by the garbage collector.
     * 
     * @throws Throwable the Exception raised by this method.
     * 
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        try {
            close();
        } finally {
            super.finalize();
        }
    }

}
