/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.util.*;

/**
 * This class is used to write serialized maps to output files.
 * This class works in conjunction with {@linkplain SerializedMapInputFile}.
 * 
 * @author This class was written and documented by 
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public class SmoFile<K, V> 
	extends SerializedOutputFile<Map<? extends K, ? extends V>>
		implements MapOutputFile<K, V> {

    // Storage for all keys
    private final K[] keys;

    /**
     * Constructs a new serialized map output file from the
     * provided arguments.
     * 
     * @param fileName the name of the file to be created.
     * @param fileFormat the format of the file to be created.
     * @param allKeys the set of keys to be 
     * contained in this file. {@code null} is not allowed.
     */
	@SuppressWarnings("unchecked")
	public SmoFile(String fileName,
                   FileFormat fileFormat,
                   Set<? extends K> allKeys) {
		
        super(fileName, fileFormat);

        // Creating storage for all variables
        keys = (K[]) new Object[allKeys.size()];

        // Setting start index
        int index = 0;

        // Storing provided variables
        for (K variable : allKeys) {
            keys[index++] = variable;
        }

        // Writing headers
        writeObject(keys);
    }
	
	/**
	 * Returns the set of keys whose values will be
	 * written to this output file when a map is added.
	 * This set acts as a type of filter to control which
	 * values in a map are saved, and which are discarded.
	 * 
	 * @return the set of keys whose values will be
	 * written to this output file when a map is added.
	 */
	public Set<K> getFilterKeys() {
		
		// Creating storage
		Set<K> s = new LinkedHashSet<K>(keys.length + 1, 1.0F);
		
		// Copying
		for (K variable : keys) {
			s.add(variable);
		}
		
		return s;
		
	}

	/**
	 * Writes the appropriate values of the provided map 
	 * to this file. The appropriate values are those of the
	 * keys in the set returned by {@linkplain #getFilterKeys()}.
	 * 
	 * @param map the map to be written to this file.
	 * 
	 * @throws IllegalArgumentException if the provided
	 * map does not contain all the keys in the
	 * set returned by {@linkplain #getFilterKeys()}.
	 * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
	 */
	@SuppressWarnings("unchecked")
	public void addData(Map<? extends K, ? extends V> map) {

        // Creating storage for map values
        V[] allValues = (V[]) new Object[keys.length];

        // Iterating thru each key
        for (int i = 0; i < keys.length; i++) {

            // Getting the next stored variable and its
            // value from the provided vector
            K key = keys[i];
            V value = map.get(key);

            // Ensuring a value was provided for this key
            if (value == null) {
                String s = "A value wasn't provided for: " + key;
                throw new IllegalArgumentException(s);
            }

            // Storing the value
            allValues[i] = value;
        }

        // Writing all values to file
        writeObject(allValues);
    }

}
