/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.util.*;

import sbrt.kernel.utilities.Util;

/**
 * This class is used to provided a source of serialized map output files.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public class SmoFileSource<K, V> implements Iterator<SmoFile<K, V>> {

	// The file name source
    private final Iterator<String> fileNameSource;
    
    // The file format
    private final FileFormat fileFormat;

    // The filter keys
    private final Set<? extends K> keys;
	
    /**
     * Constructs a new source of serialized map output files.
     * 
     * @param fileNameSource the file name source.
     * @param fileFormat the file format used to write files.
     * 
     * @param keys the keys whose values will be
     * written to the files produced by this source. 
     * {@code null} is not allowed.
     * 
     * @throws NullPointerException if any argument is {@code null}.
     */
	public SmoFileSource(Iterator<String> fileNameSource,
						 FileFormat fileFormat,
						 Set<? extends K> keys) {
    	
    	// Checking inputs
        Util.nullCheck(fileNameSource);
        Util.nullCheck(fileFormat);
        Util.nullCheck(keys);
        
        // Saving
        this.fileNameSource = fileNameSource;
        this.fileFormat = fileFormat;
        this.keys = keys;
    }

    /**
     * Returns the format used to create files of type {@linkplain SmoFile}.
     * @return the format used to create files of type {@linkplain SmoFile}.
     */
    public FileFormat getFileFormat() {
    	return fileFormat;
    }

    /**
     * Indicates if this iterator can return another file using the
     * {@code next()} method
     * .
     * @return {@code true} if this iterator can return another file;
     * {@code false} otherwise.
     */
	public boolean hasNext() {
		return fileNameSource.hasNext();
	}

	/**
	 * Returns the next column-oriented map output file with the next
	 * name returned by the file name source.
	 * @return the next column-oriented map output file.
	 */
	public SmoFile<K, V> next() {

		// Getting the next file name
		String fileName = fileNameSource.next();
		
		// Returning the next Romo file
		return new SmoFile<K, V>(fileName, fileFormat, keys);
	}

	/**
	 * Throws {@code UnsupportedOperationException}.
	 * @throws UnsupportedOperationException
	 */
	public void remove() {
		throw new UnsupportedOperationException();
	}
}
