/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.io.*;
import java.util.*;
import java.util.zip.*;

import sbrt.kernel.utilities.*;
import sbrt.shell.*;
import sbrt.shell.text.*;

/**
 * This class is used to iterate over the lines of an input file.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <E> the returned type.
 */
public final class TextFileLineIterator<E> extends FileLineIterator<E> {

	/**
	 * Indicates if the provided file format is valid for
	 * this type of file line iterator.
	 * 
	 * @param fileFormat the file format to test.
	 * 
	 * @return {@code true} if the provided file format is 
	 * valid for a text input file; {@code false} otherwise.
	 */
    public static boolean isValidFormat(FileFormat fileFormat) {
        return fileFormat == FileFormat.TEXT ||
                fileFormat == FileFormat.GZIP_TEXT;
    }	

    // The file of this iterator
    private final TextInputFile<E> file;
    
    // The comment characters
    private final char[] commentChars;

    // The file reader
    private BufferedReader rdr;
    
    // The next line of this file
    private String nextLine;

    /**
     * Constructs a new text file line iterator using the
     * provided text input file.
     * 
     * @param file the file to be iterated over.
     */
    public TextFileLineIterator(TextInputFile<E> file) {

        // Checking and saving input
        Util.nullCheck(file);
        this.file = file;

        // Filling comment character array
        commentChars = getCommentChars(file.getCommentChars());

        // Getting file name
        String fileName = file.getCanonicalName();
        
        // Checking file
        checkFile(fileName);

        try {
            rdr = getReader(file.getCanonicalName(), file.getFileFormat());
            nextLine = getNextLine();
        } catch (IOException e) {
            String s = "An I/O error occurred while opening the file '" +
                       file.getCanonicalName() + "'.";
            throw new ApplicationException(s, e);
        }

    }

    /**
     * Creates and returns a file reader.
     */
    private BufferedReader getReader(String fileName, 
    		FileFormat fileFormat) throws IOException {

        if (fileFormat == FileFormat.TEXT) {

            // Creating FileReader
            Reader fileIn = new FileReader(fileName);

            // Wrapping with BufferedReader
            return new BufferedReader(fileIn);
        }

        else if (fileFormat == FileFormat.GZIP_TEXT) {

            // Creating FileInputStream
            InputStream fileIn = new FileInputStream(fileName);

            // Wrapping with GZIPInputStream
            InputStream gzipIn = new GZIPInputStream(fileIn);

            // Wrapping with InputStreamReader
            Reader reader = new InputStreamReader(gzipIn);

            // Wrapping with BufferedReader
            return new BufferedReader(reader);
        }

        else {
            String s = "Unsupported file format: " + fileFormat;
            throw new IllegalArgumentException(s);
        }

    }

    /**
     * Copies a collection of comment characters into an array.
     */
    private char[] getCommentChars(Collection<Character> c) {
    	
    	// Creating storage and initializing index
        char[] array = new char[c.size()];
        int index = 0;

        // Iterating thru each character and saving
        for (Character ch : c) {
            array[index] = ch.charValue();
        }

        // Alphabetizing
        Arrays.sort(array);
        
        return array;
    }

    /**
     * Returns the next element in the file.
     * @return the next element in the file.
     * 
     * @throws NoSuchElementException if the file 
     * has no more elements.
     */
	public E next() {
    	
    	// Checking state
        if (!hasNext()) {
            throw new NoSuchElementException();
        }

        // Copying saved line into temp
        String temp = nextLine;
        
        // Looking ahead one line
        nextLine = getNextLine();

        // Automatically closing the file if no more lines exist
        if (nextLine == null) {
            close();
        }

        try {

            // Parsing temp with the file's line format
        	return file.getLineFormat().parse(temp); 

        } catch (FormatException e) {
            String s = "A problem occurred while parsing the line '" + temp +
                       "' in the file '" + file.getCanonicalName() + "'.  " +
                       e.getMessage();
            throw new FormatException(s, e);
        }

    }

	/**
	 * Returns the next line in the file.
	 */
    private String getNextLine() {

        String line;

        try {

        	// While lines still exist
            while ((line = rdr.readLine()) != null) {

            	// Trim off whitespace
            	line = line.trim();

            	// If the line isn't commented, return it
                if (!isComment(line)) {
                    return line;
                }
            }

        } catch (IOException e) {
            String s = "An I/O error occurred while reading the " +
                       "file '" + file.getCanonicalName() + "'.";
            throw new ApplicationException(s, e);
        }

        // Returning null
        return line;
    }

    /**
     * Indicates if the file has more lines.
     * @return {@code true} if the file has more lines;
     * {@code false} otherwise.
     */
    public boolean hasNext() {

        if (isClosed()) {
            return false;
        }
        return nextLine != null;
    }

    /**
     * Throws UnsupportedOperationException.
     * @throws UnsupportedOperationException
     */
    public void remove() {
        throw new UnsupportedOperationException();
    }

    /**
     * Indicates if this iterator has been closed.
     * 
     * @return {@code true} if {@linkplain #close()} has
     * been called; {@code false} otherwise.
     */
    public boolean isClosed() {
        return rdr == null;
    }

    /**
     * Disposes of system resources associated with this file
     * iterator when called by the garbage collector.
     * 
     * @throws Throwable the Exception raised by this method.
     * 
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        try {
            close();
        } finally {
            super.finalize();
        }
    }

    /**
     * Closes this file iterator. If this file iterator is already 
     * closed, then invoking this method has no effect.
     * 
     * @throws ApplicationException if an I/O error occurred
     * while closing this file iterator.
     */
    public void close() {
    	
    	// Do nothing if already closed
        if (isClosed()) {
            return;
        }

        try {
        	
        	// Closing the reader and changing the file's state to "closed"
            rdr.close();
            rdr = null;

        } catch (IOException e) {
            String s = "An I/O error occurred while closing the file '"
                       + file.getCanonicalName() + "'.";
            throw new ApplicationException(s, e);
        }

    }

    /**
     * Indicates if the provided line is a comment.
     */
    private boolean isComment(String trimmedLine) {

        if (trimmedLine.length() == 0) {
            return true;
        }

        return Arrays.binarySearch(commentChars, trimmedLine.charAt(0)) >= 0;
    }
}
