/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.io.File;
import java.util.*;

import sbrt.kernel.utilities.*;
import sbrt.shell.ApplicationException;
import sbrt.shell.text.*;

/**
 * This class is used to read input files containing text.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <E> the type contained in instances of this interface.
 */
public class TextInputFile<E> implements InputFile<E> {

    /**
     * Returns a basic text input file.
     * 
     * @param fileName the name of the file to be read.
     * 
     * @return a basic text input file.
     */
	public static TextInputFile<String> getBasic(String fileName) {
		return getBasic(fileName, FileFormat.TEXT);
	}

    /**
     * Returns a basic text input file.
     * 
     * @param fileName the name of the file to be read.
     * @param fileFormat the file format of this file.
     * 
     * @return a basic text input file.
     */
	public static TextInputFile<String> 
	getBasic(String fileName, FileFormat fileFormat) {
		
		StringFormat f = new StringFormat();
		
		return new TextInputFile<String>(fileName, fileFormat, f);
	}

	/**
	 * Indicates if the provided file format is valid for
	 * a text input file.
	 * 
	 * @param fileFormat the file format to test.
	 * 
	 * @return {@code true} if the provided file format is 
	 * valid for a text input file; {@code false} otherwise.
	 */
    public static boolean isValidFormat(FileFormat fileFormat) {
        return TextFileLineIterator.isValidFormat(fileFormat);
    }

    private static final UnmodifiableCollection<Character> defaultCommentChars;
    static {
        Set<Character> s = new TreeSet<Character>();
        s.add(Character.valueOf('#'));
        defaultCommentChars = new UnmodifiableCollection<Character>(s);
    }

    // The name of this file
    private final String fileName;
    
    // The file format of this file
    private final FileFormat fileFormat;
    
    // The line format of this file
    private final Parser<E> lineFormat;

    // The list of iterators over this file
    private List<TextFileLineIterator<E>> fileIterators;
    
    // The collection of comment characters
    private UnmodifiableCollection<Character> commentCharacters;

    /**
     * Constructs a new text input file using the provided arguments.
     * 
     * @param fileName the name of the file to be read.
     * @param fileLineFormat the format used to parse the lines
     * of this file. 
     */
    public TextInputFile(String fileName, Parser<E> fileLineFormat) {
        this(fileName, FileFormat.TEXT, fileLineFormat, defaultCommentChars);
    }

    /**
     * Constructs a new text input file using the provided arguments.
     * 
     * @param fileName the name of the file to be read.
     * @param fileFormat the format of the file to be read.
     * @param fileLineFormat the format used to parse the lines
     * of this file. 
     */
    public TextInputFile(String fileName, 
    					 FileFormat fileFormat,
    					 Parser<E> fileLineFormat) {
        this(fileName, fileFormat, fileLineFormat, defaultCommentChars);
    }

    /**
     * Constructs a new text input file using the provided arguments.
     * 
     * @param fileName the name of the file to be read.
     * @param fileFormat the format of the file to be read.
     * @param fileLineFormat the format used to parse the lines
     * of this file.
     * @param commentCharacters a collection of characters used
     * for comments in this file.
     */
    public TextInputFile(String fileName,
                         FileFormat fileFormat,
                         Parser<E> fileLineFormat,
                         Collection<Character> commentCharacters) {

        // Checking inputs
        Util.nullCheck(fileName);
        Util.nullCheck(fileFormat);
        Util.nullCheck(fileLineFormat);

        // Ensuring file format is appropriate
        if (!isValidFormat(fileFormat)) {
            String s = "Unsupported file format: " + fileFormat;
            throw new IllegalArgumentException(s);
        }

        // Saving
        this.fileName = fileName;
        this.fileFormat = fileFormat;
        this.lineFormat = fileLineFormat;
        setCommentChars(commentCharacters);

        // Creating storage for file line iterators
        fileIterators = new LinkedList<TextFileLineIterator<E>>();

        // Getting the canonical name
        String name = AbstractSbrtFile.getCanonicalName(fileName);
        
        // Checking file
        FileLineIterator.checkFile(name);
    }

    /**
     * Returns the format used to parse lines of this file.
     * @return the format used to parse lines of this file.
     */
    public Parser<E> getLineFormat() {
        return lineFormat;
    }

    /**
     * Returns the format of this file.
     * @return the format of this file.
     */
    public FileFormat getFileFormat() {
        return fileFormat;
    }

    private void setCommentChars(Collection<Character> characters) {

    	// Creating storage
        Set<Character> chars = new TreeSet<Character>();

        // Iterating thru each provided character
        for (Character ch : characters) {
            chars.add(ch);
        }

        commentCharacters = new UnmodifiableCollection<Character>(chars);
    }

    /**
     * Returns the commment characters used in this file.
     * @return the commment characters used in this file.
     */
    public final UnmodifiableCollection<Character> getCommentChars() {
        return commentCharacters;
    }

    /**
     * Returns the name of this file.
     * @return the name of this file.
     */
    public final String getOriginalName() {
        return fileName;
    }

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     * 
     * @see File#getCanonicalPath()
     */
    public final String getCanonicalName() {
        return AbstractSbrtFile.getCanonicalName(fileName);
    }

    /**
     * Returns a line iterator over this text file.
     * @return a line iterator over this text file.
     */
    protected final TextFileLineIterator<E> lineIterator() {

        // Checking state
        if (isClosed()) {
            throw new IllegalStateException("Already closed: " + fileName);
        }

        // Creating new file line iterator
        TextFileLineIterator<E> i = new TextFileLineIterator<E>(this);

        // Adding it to the list
        fileIterators.add(i);

        // Returning it
        return i;
    }

    /**
     * Returns a line iterator over this text file.
     * @return a line iterator over this text file.
     */
    public Iterator<E> iterator() {
        return lineIterator();
    }

    /**
     * Returns a line iterator over this text file beginning at the specified
     * line. Line indices begin at {@code 0}, and commented or blank lines are
     * ignored.
     * 
     * @param line the index of the line at which this iterator begins.
     * 
     * @return a line iterator over this text file beginning at the specified
     * line.
     * 
     * @throws IndexOutOfBoundsException if the provided number exceeds the
     * number of lines contained in the file.
     */
    public Iterator<E> iterator(int line) {
        
        TextFileLineIterator<E> i = lineIterator();
        
        for (int index = 0; index < line; index ++) {
            
            if (!i.hasNext()) {
                throw new IndexOutOfBoundsException("Index: " + line);
            }
            
            // Throwing away next element
            i.next();
        }
        
        return i;
    }
    
    /**
     * Closes this file. If this file is already closed, then
     * invoking this method has no effect.
     * 
     * @throws ApplicationException if an I/O error occurred
     * while closing this file.
     */
    public final void close() {

        // Do nothing if already closed
        if (isClosed()) {
            return;
        }

        // Iterating thru each file iterator
        for (TextFileLineIterator<E> iterator : fileIterators) {

            // Closing it
        	iterator.close();
        }

        // Destroying the reference to the list of iterators
        fileIterators = null;
    }

    /**
     * Indicates if this file has already been closed.
     * 
     * @return {@code true} if this file is closed;
     * {@code false} othewise.
     */
    public final boolean isClosed() {
        return fileIterators == null;
    }

    /**
     * Disposes of system resources associated with this file
     * when called by the garbage collector.
     * 
     * @throws Throwable the Exception raised by this method.
     * 
     * @see java.lang.Object#finalize()
     */
    protected final void finalize() throws Throwable {
        try {
            close();
        } finally {
            super.finalize();
        }
    }

}
