/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io;

import java.io.*;
import java.util.zip.*;

import sbrt.Sbrt;
import sbrt.kernel.utilities.*;
import sbrt.shell.*;
import sbrt.shell.text.*;

/**
 * This class is used to represent output files containing
 * plain text.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <T> the type contained in this type of output file.
 */
public class TextOutputFile<T> implements OutputFile<T> {

	/**
	 * Indicates if the provided file format is valid for
	 * a text output file.
	 * 
	 * @param fileFormat the file format to test.
	 * 
	 * @return {@code true} if the provided file format is 
	 * valid for a text output file; {@code false} otherwise.
	 */
    public static boolean isValidFormat(FileFormat fileFormat) {
        return TextInputFile.isValidFormat(fileFormat);
    }

    /**
     * Returns a basic text output file.
     * 
     * @param fileName the name with which to name the file.
     * 
     * @return a basic text output file.
     */
    public static TextOutputFile<String> getBasic(String fileName) {
    	return getBasic(fileName, FileFormat.TEXT);
    }

    /**
     * Returns a basic text output file with the provided
     * file format.
     * 
     * @param fileName the name of the file to be created.
     * @param fileFormat the file format with which to write
     * this file.
     * 
     * @return a basic text output file with the provided file
     * format.
     */
    public static TextOutputFile<String> 
    getBasic(String fileName, FileFormat fileFormat) {
    	
    	// Creating the "trivial" format
    	StringFormat f = new StringFormat();
    	
    	return new TextOutputFile<String>(fileName, fileFormat, f);
    }
    
    // This file's name
    private final String fileName;
    
    // Used to indicate the file format
    private final FileFormat fileFormat;

    // Used for formatting lines
    private final Formatter<T> lineFormat;

    // Used for writing text to this file
    private Writer wtr;

    /**
     * Constructs a new text output file using the provided
     * arguments.
     * 
     * @param fileName the name of the file to be created.
     * 
     * @param lineFormat the format used for writing formatted
     * data strings to this file.
     */
    public TextOutputFile(String fileName, Formatter<T> lineFormat) {
        this(fileName, FileFormat.TEXT, lineFormat);
    }

    /**
     * Constructs a new text output file using the provided
     * arguments.
     * 
     * @param fileName the name of the file to be created.
     * 
     * @param fileFormat the file format used to write data
     * to this file.
     * 
     * @param lineFormat the format used for writing formatted
     * data strings to this file.
     */
    public TextOutputFile(String fileName,
                          FileFormat fileFormat,
                          Formatter<T> lineFormat) {

        // Checking inputs
        Util.nullCheck(fileName);
        Util.nullCheck(fileFormat);
        Util.nullCheck(lineFormat);

        // Saving
        this.fileName = fileName;
        this.fileFormat = fileFormat;
        this.lineFormat = lineFormat;

        try {

            // Getting canonical, absolute file name
            String name = AbstractSbrtFile.getCanonicalName(fileName);
            
            java.io.File f = new java.io.File(name);

            if (f.exists()) {
                String s = "The file '" + name + "' already exists.";
                throw new ApplicationException(s);
            }

            try {
                f.createNewFile();
            } catch (IOException e) {
                String s = "The file '" + name + "' cannot be created.";
                throw new ApplicationException(s);
            }

            wtr = getWriter(f, fileFormat);

            writeHeader();
            
        } catch (IOException e) {

            String s = "An I/O error occurred while trying to " +
                       "create the file '" + fileName + "'.";
            throw new ApplicationException(s, e);

        } catch (SecurityException e) {

            String s = "A security violation occurred while trying to " +
                       "create the file '" + fileName + "'.";
            throw new ApplicationException(s, e);

        }
    }

    private void writeHeader() {

        StringBuilder s = new StringBuilder(90);
        
        s.append('#');
        
        for (int i = 1; i < 79; i ++) {
            s.append('-');
        }
        
        writeLine(s.toString());
        
        String msg = "# This file was created with the Systems Biology Research Tool " +
        "version " + Sbrt.getVersion() + ".";

        writeLine(msg);
        
        writeLine(s.toString());
        writeLine("");
    }
    
    private Writer getWriter(java.io.File file, FileFormat format) throws
            IOException {

        if (format == FileFormat.TEXT) {

            // Creating a FileWriter
            Writer fileWtr = new FileWriter(file);

            // Wrapping it with a BufferedWriter
            return new BufferedWriter(fileWtr);
        }

        else if (format == FileFormat.GZIP_TEXT) {

            // Creating FileOutputStream
            OutputStream fileOut = new FileOutputStream(file);

            // Wrapping with a BufferedOutputStream
            OutputStream buffOut = new BufferedOutputStream(fileOut);

            // Wrapping with a GZIPOutputStream
            OutputStream gzipOut = new GZIPOutputStream(buffOut);

            // Wrapping with an OutputStreamWriter
            return new OutputStreamWriter(gzipOut);
        }

        else {
            String s = "Unsupported file format: '" + fileFormat;
            throw new IllegalArgumentException(s);
        }

    }

	/**
	 * Returns the canonical name of this file.
	 * @return the canonical name of this file.
	 */
    public final String getName() {
        return getCanonicalName();
    }

    /**
     * Returns the name of this file.
     * @return the name of this file.
     */
    public final String getOriginalName() {
        return fileName;
    }

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     * 
     * @see File#getCanonicalPath()
     */
    public final String getCanonicalName() {
        return AbstractSbrtFile.getCanonicalName(fileName);
    }

    /**
     * Indicates if this file has already been closed.
     * 
     * @return {@code true} if this file is closed;
     * {@code false} othewise.
     */
    public final boolean isClosed() {
        return wtr == null;
    }

    /**
     * Returns the format used to write lines to this
     * file.
     * 
     * @return the format used to write lines to this
     * file.
     */
    public Formatter<T> getLineFormat() {
        return lineFormat;
    }

    /**
     * Closes this file. If this file is already closed, then
     * invoking this method has no effect.
     * 
     * @throws ApplicationException if an I/O error occurred
     * while closing this file.
     */
    public void close() {

        // Do nothing if already closed
        if (isClosed()) {
            return;
        }

        try {

            wtr.flush();
            wtr.close();
            wtr = null;

        } catch (IOException e) {
            String s = "An I/O error occured while trying to close the " +
                       "file '" + fileName + "'.";
            throw new ApplicationException(s, e);
        }
    }

    /**
     * Disposes of system resources associated with this file
     * when called by the garbage collector.
     * 
     * @throws Throwable the Exception raised by this method.
     * 
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        try {
            close();
        } finally {
            super.finalize();
        }
    }

    /**
     * Writes a line of {@code #} symbols to this file.
     * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
     */
    public void breakPoint() {
    	
        checkState();
        
        StringBuilder s = new StringBuilder(55);
        
        for (int i = 0; i < 55; i ++) {
        	s.append('#');
        }
        
        writeLine(s.toString());
    }

	/**
	 * Adds data to this output file. The provided data
	 * is formatted using this output file's line
	 * format (see {@linkplain #getLineFormat()}) and
	 * followed by a system-dependent new line character.
	 * 
	 * @param <S> a type extending the primary data type
	 * contained in this output file.
	 * 
	 * @param results the results, or data, to be added
	 * to this output file.
	 * 
     * @throws ApplicationException if an I/O error occurs
     * while writing data to this file.
	 * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
	 * 
	 * @throws NullPointerException if the provided 
	 * argument is {@code null}.
	 */
    public <S extends T> void addData(S results) {
    	
    	// Checking the state of this file
        checkState();

        // Creating the formatting string
        String formattedLine = lineFormat.format(results);
        
        // Writing it as a line in this file
        writeLine(formattedLine);
    }

    /**
     * Writes the provided string to this output file.
     * A system-dependent new line character is automatically 
     * appended to the end of the provided string.
     * 
     * @param line the line to write to this file.
     * 
     * @throws ApplicationException if an I/O error occurs
     * while writing data to this file.
     * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
     */
    protected final void writeLine(String line) {
    	
        checkState();

        try {

            wtr.write(line + System.getProperty("line.separator"));
            wtr.flush();

        } catch (IOException e) {
            close();
            String s = "An I/O error occurred while writing data to the " +
                       "file '" + getName() + "'.";
            throw new ApplicationException(s, e);
        }

    }

    /**
	 * Throws IllegalStateException if {@code close()}
	 * has already been called.
     */
    private void checkState() {
        if (isClosed()) {
            throw new IllegalStateException("Already closed: " + fileName);
        }
    }

}
