/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io.fba;

import java.io.File;
import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.utilities.*;
import sbrt.shell.io.*;
import sbrt.shell.text.*;
import sbrt.shell.text.fba.*;

/**
 * This class is used to read files containing flux caps.
 * Flux caps are used in Mahadevan-Schilling flux interval processes.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public class FluxCapInputFile 
	implements MapTextInputFile<String, Set<LinearComb<String>>> {

	// The row-oriented map input file
	private final RomiFile<String, Set<LinearComb<String>>> file;
    
	// The fluxome used to verify reaction names
	private final Fluxome fluxome;

    /**
     * Constructs a new flux cap input file.
     * 
     * @param fileName the name of the file to read.
     * @param lineFormat the format used to parse the data strings in this 
     * file.
     */
    public FluxCapInputFile(String fileName, FluxCapFormat lineFormat) {
    	
    	fluxome = lineFormat.getFluxome();
    	
    	file = new RomiFile<String, Set<LinearComb<String>>>
    		(fileName, lineFormat);
    }

    /**
     * Returns the fluxome used to verify reaction names.
     * @return the fluxome used to verify reaction names.
     */
    public Fluxome getFluxome() {
        return fluxome;
    }

    /**
     * Returns the flux caps contained in this file.
     * @return the flux caps contained in this file.
     */
    public UnmodifiableMap<String, Set<LinearComb<String>>> getFluxCaps() {
        return file.getMap();
    }

	/**
	 * Returns the map format used to parse this file.
	 * @return the map format used to parse this file.
	 */
	public MapFormat<String, Set<LinearComb<String>>> getMapFormat() {
		return file.getMapFormat();
	}

    /**
     * Closes this file. If this file is already closed, then
     * invoking this method has no effect.
     */
	public void close() {
		file.close();
	}

    /**
     * Returns the provided name of this file.
     * @return the provided name of this file.
     */
    public final String getOriginalName() {
        return file.getOriginalName();
    }

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     * 
     * @see File#getCanonicalPath()
     */
    public final String getCanonicalName() {
        return file.getCanonicalName();
    }

    /**
     * Indicates if this file has already been closed.
     * 
     * @return {@code true} if this file is closed;
     * {@code false} othewise.
     */
	public boolean isClosed() {
		return file.isClosed();
	}

    /**
     * Returns an iterator over this file. This iterator
     * returns only a single map.
     * 
     * @return an iterator over this file.
     * 
     * @see #getFluxCaps()
     */
	public Iterator<Map<String, Set<LinearComb<String>>>> iterator() {
		return file.iterator();
	}

}
