/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io.fba;

import java.io.File;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import sbrt.kernel.fba.FluxConstraints;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.math.HashLinearComb;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.utilities.Util;
import sbrt.shell.ApplicationException;
import sbrt.shell.io.ComiFile;
import sbrt.shell.io.FileFormat;
import sbrt.shell.io.InputFile;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.math.IntervalFormatV1;

/**
 * This class is used to read flux interval input files.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public class FluxIntervalInputFile implements InputFile<FluxConstraints> {

    // The column-oriented map input file
    private final ComiFile<LinearComb<String>, Interval> file;

    private final Fluxome fluxome;
    private final double tolerance;
    
    /**
     * Constructs a new flux-interval input file.
     * 
     * @param fileName the name of the file to be read.
     * @param fileFormat the format of the file to be read.
     * @param requiredVariables the variables that must be contained in the file
     * to be read.
     * @param mapFormat the format of data strings in the file to be read.
     * @param fluxome the fluxome for which flux intervals correspond.
     * @param tolerance the constraint tolerance.
     */
    public FluxIntervalInputFile(String fileName, 
    							 FileFormat fileFormat,
    							 Set<LinearComb<String>> requiredVariables,
    							 MapFormat<LinearComb<String>, Interval> mapFormat,
    							 Fluxome fluxome,
    							 double tolerance) {

    	if (tolerance < 0 ||
    		Double.isNaN(tolerance) ||
    		Double.isInfinite(tolerance)) {
    		
    		String s = "Invalid tolerance: " + tolerance;
    		throw new IllegalArgumentException(s);
    	}
    	
    	Util.nullCheck(fluxome);
    	
    	this.fluxome = fluxome;
    	this.tolerance = tolerance;
    	
        file = new ComiFile<LinearComb<String>, Interval>(fileName,
                        fileFormat, requiredVariables, mapFormat);

    }

    /**
     * Returns the constraint tolerance.
     * @return the constraint tolerance.
     */
    public double getTolerance() {
    	return tolerance;
    }
    
    /**
     * Closes this file. If this file is already closed, then invoking this
     * method has no effect.
     */
    public void close() {
        file.close();
    }

    /**
     * Returns the provided name of this file.
     * @return the provided name of this file.
     */
    public final String getOriginalName() {
        return file.getOriginalName();
    }

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     * 
     * @see File#getCanonicalPath()
     */
    public final String getCanonicalName() {
        return file.getCanonicalName();
    }

    /**
     * Indicates if this file has already been closed.
     * 
     * @return {@code true} if this file is closed; {@code false} othewise.
     */
    public boolean isClosed() {
        return file.isClosed();
    }

    /**
     * Returns an iterator over the flux intervals contained in this file.
     * 
     * @return an iterator over the flux intervals contained in this file.
     */
    public Iterator<FluxConstraints> iterator() {

        final Iterator<Map<LinearComb<String>, Interval>> i = file.iterator();

        return new Iterator<FluxConstraints>() {

            public FluxConstraints next() {

                Map<LinearComb<String>, Interval> vector = i.next();

                return convertConstraints(vector, fluxome, tolerance);

            }

            public boolean hasNext() {
                return i.hasNext();
            }

            public void remove() {
                i.remove();
            }
        };

    }

    /**
     * Converts a set of reaction names to a set of single-term linear
     * combinations.
     * 
     * @param rxnNames the reaction names to be used for constructing
     * single-term linear combinations.
     * 
     * @return a set of single-term linear combinations corresponding to the
     * provided reaction names.
     */
    public static Set<LinearComb<String>> convertVariables(Set<String> rxnNames) {

        // Determining appropriate size
        int size = rxnNames.size() + 1;

        // Creating storage
        Set<LinearComb<String>> set =
                new LinkedHashSet<LinearComb<String>>(size, 1.0f);

        for (String variable : rxnNames) {

            LinearComb<String> expr = HashLinearComb.singleTerm(variable);

            set.add(expr);

        }

        return set;
    }

    /**
     * Converts the provided intervals into a {@code FluxConstraints} object. If
     * a provided linear combination contains a single term, with coefficient
     * equal to 1, and constant equal to 0, the interval will be applied to the
     * variable itself (not the linear combination), which is retrievable using
     * {@linkplain FluxConstraints#getConstraint(String)} and
     * {@linkplain FluxConstraints#getFluxConstraints()}. The flux intervals are 
     * also verified against the provided fluxome.
     * 
     * @param intervals reaction name expressions mapped to their corresponding
     * interval.
     * @param fluxome the fluxome for which flux intervals correspond.
     * @param tolerance the contraint tolerance.
     * 
     * @return a {@code FluxConstraints} object corresponding to the provided
     * input.
     */
    @SuppressWarnings("unchecked")
    public static FluxConstraints 
    convertConstraints(Map<LinearComb<String>, Interval> intervals,
    				   Fluxome fluxome,
    				   double tolerance) {

        // Determining maximum size
        int size = intervals.size() + 1;

        // Creating storage for flux constraints
        Map<String, Interval> fluxConsts =
                new LinkedHashMap<String, Interval>(size, 1.0F);

        // Creating storage for rxn expression constraints
        Map<LinearComb<String>, Interval> exprConsts =
                new LinkedHashMap<LinearComb<String>, Interval>();

        // Iterating thru the constraints vector
        for (Map.Entry<LinearComb<String>, Interval> e : intervals.entrySet()) {

            // Getting the key and constraint
            LinearComb<String> key = e.getKey();
            Interval interval = e.getValue();

            // If the key can be simplified to a single reaction name
            if (key.getConstant() == 0d && key.terms() == 1) {

                LinearComb.Term<String> term =
                        key.getTermList().iterator().next();

                if (term.getCoefficient() == 1.0d) {
                    fluxConsts.put(term.getVariable(), interval);
                    continue;
                }
            }

            // Otherwise
            exprConsts.put(key, interval);

        }

        verifyFluxConstraints(fluxConsts, fluxome, tolerance);
        
        // Creating constraint set
        return new FluxConstraints(fluxConsts, exprConsts);
    }
    
    /**
     * Verifies that the provided constraints are legitimate for the provided
     * fluxome.
     * 
     * @param fluxConsts reaction named mapped to flux constraints.
     * @param fluxome the fluxome for which the flux constraints correspond.
     * @param tolerance the constraint tolerance.
     * 
     * @throws ApplicationException if the provided flux constraints are not
     * valid for the provided fluxome.
     * 
     * @see Fluxome#isValid(Interval, String, double)
     */
    public static void 
    verifyFluxConstraints(Map<String, Interval> fluxConsts,
    					  Fluxome fluxome,
    					  double tolerance) {
    	
        for (Map.Entry<String, Interval> e : fluxConsts.entrySet()) {
        	
        	String rxnName = e.getKey();
        	Interval constraint = e.getValue(); 

        	if (!fluxome.isValid(constraint, rxnName, tolerance)) {
        		badConstraint(constraint, rxnName, tolerance);
        	}
        }
    	
    }
    
    private static void badConstraint(Interval constraint, 
    								  String rxnName, 
    								  double tolerance) {
    	
    	IntervalFormatV1 intervalFormat = new IntervalFormatV1();
    	DoubleFormat numberFormat = intervalFormat.getDoubleFormat();
    	
    	String s = "The flux constraint '" + intervalFormat.format(constraint) +
    			   "' is invalid for the reaction '" + rxnName +
    			   "' using " + numberFormat.format(tolerance) +
    			   " as the constraint tolerance.";
    	
    	throw new ApplicationException(s);
    }
    
    
}
