/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io.fba;

import java.io.File;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.comp.FbaOptimizer;
import sbrt.shell.io.FileFormat;
import sbrt.shell.io.RomoFile;
import sbrt.shell.text.fba.FbaOptFileLineFormat;

/**
 * This class is used to write row-oriented output files containing FBA
 * optimization data.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 * @param <F>
 *            the fluxome type.
 * 
 */
public class RowOrFbaOptOutputFile<F extends Fluxome> 
	implements FbaOptOutputFile<F> {

	// The row-oriented map output file
	private final RomoFile<String, Double> file;
	
	/**
	 * Constructs a new row-oriented FBA optimization output file.
	 * 
	 * @param fileName the name of the file to create.
	 * @param fileFormat the format of the file to create.
	 * 
	 * @param rowHeaders a set of valid row headers that will appear
	 * in the file. This set can consist of reaction names and the
	 * objective function header. If {@code null}, all values
     * from every added set of results will be written to this file.
	 * 
	 * @param fbaFormat the map format used to write data strings
	 * to the file.
	 */
    public RowOrFbaOptOutputFile(String fileName,
                                 FileFormat fileFormat,
                                 Set<String> rowHeaders,
                                 FbaOptFileLineFormat fbaFormat) {
        
        file = new RomoFile<String, Double>(fileName, 
        									fileFormat,
        									rowHeaders, 
        									fbaFormat);
    }
	
	/**
	 * Constructs a new row-oriented FBA optimization output file.
	 * 
	 * @param fileName the name of the file to create.
	 * @param fileFormat the format of the file to create.
	 * 
	 * @param fbaFormat the map format used to write data strings
	 * to the file.
	 */
    public RowOrFbaOptOutputFile(String fileName,
                                 FileFormat fileFormat,
                                 FbaOptFileLineFormat fbaFormat) {
        
        file = new RomoFile<String, Double>(fileName, fileFormat, fbaFormat);
    }

    /**
     * Returns the FBA optimization file line format used
     * to write data strings to this file.
     * 
     * @return the FBA optimization file line format used
     * to write data strings to this file.
     */
    public FbaOptFileLineFormat getFbaOptFileLineFormat() {
        return (FbaOptFileLineFormat) file.getMapFormat();
    }
	
	/**
	 * Returns the set of row headers whose values will be
	 * written to this output file when results are added.
	 * This set acts as a type of filter to control which
	 * values in a set of results are saved, and which are discarded.
	 * 
	 * @return the set of row headers whose values will be
	 * written to this output file when a map is added;
	 * or {@code null} if <i>every</i> value in an added set of
	 * results will be written to this file.
	 */
    public Set<String> getRowHeaders() {
    	return file.getFilterKeys();
    }
    
	/**
	 * Writes the appropriate values in the provided 
	 * {@code FbaOptimizer.Results} to this file.
	 * The appropriates values are those returned by 
	 * {@linkplain #getRowHeaders()}.
	 * 
	 * @param results the results to write to this file.
	 * 
	 * @throws IllegalArgumentException if the provided
	 * results object does not contain all the variables in the
	 * set returned by {@linkplain #getRowHeaders()}.
	 * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
	 */
	public void addData(FbaOptimizer<F>.Results results) {

    	// Getting flux vector
        Map<String, Double> fluxVector = results.getFluxVector();
        
        // Determining appropriate size
        int size = fluxVector.size() + 2;

        // Creating new storage
        Map<String, Double> vector = 
        	new LinkedHashMap<String, Double>(size, 1.0F);

        // Getting the line format
        FbaOptFileLineFormat format = getFbaOptFileLineFormat();

        // Getting the object that represents the objective value header
        String objValueHeader = format.getObjValueHeader();
        
        // Getting the objective value
        double objValue = results.getObjectiveValue();

        // Copying
        vector.put(objValueHeader, objValue);
        vector.putAll(fluxVector);

        // Writing to file
        file.addData(vector);
    }

	/**
	 * Writes the appropriate values in the provided 
	 * map to this file.
	 * The appropriates values are those returned by 
	 * {@linkplain #getRowHeaders()}.
	 * 
	 * @param results the results vector to write to this file.
	 * 
	 * @throws IllegalArgumentException if the provided
	 * map does not contain all the keys in the
	 * set returned by {@linkplain #getRowHeaders()}.
	 * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
	 */
	public void addData(Map<String, Double> results) {

        // Writing to file
        file.addData(results);
    }

    /**
     * Closes this file. If the file is already closed, no
     * action will be taken.
     */
	public void close() {
		file.close();
	}

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     */
    public final String getName() {
        return file.getName();
    }

    /**
     * Returns the provided name of this file.
     * @return the provided name of this file.
     */
    public final String getOriginalName() {
        return file.getOriginalName();
    }

    /**
     * Returns the canonical name of this file.
     * @return the canonical name of this file.
     * 
     * @see File#getCanonicalPath()
     */
    public final String getCanonicalName() {
        return file.getCanonicalName();
    }

    /**
     * Indicates if this file has already been closed.
     * 
     * @return {@code true} if this file is closed;
     * {@code false} othewise.
     */
	public boolean isClosed() {
		return file.isClosed();
	}

    /**
     * Writes a line of {@code #} symbols to this file.
     * 
	 * @throws IllegalStateException if {@code close()}
	 * has already been called.
     */
	public void breakPoint() {
		file.breakPoint();
	}

}
