/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io.fba;

import java.util.*;

import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.io.*;
import sbrt.shell.text.*;
import sbrt.shell.text.fba.*;

/**
 * This class is used to read files containing a set of reaction name
 * expressions. A single expression occurs on each line of the file. If the file 
 * contains identical expressions, a {@code FormatException} is thrown.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <E> the mathematical expression type.
 * @param <S> the variable type.
 *
 */
public class RxnNamExprSetInputFile<E extends MathExpr<S>, S> 
	extends TextInputFile<E> {

	// The set of all expressions in this file
    private UnmodifiableSet<E> exprSet;
    private RxnNameExprFormat<E, S> lineFormat;

    /**
     * Constructs a new reaction name expression set input file.
     * 
     * @param fileName the name of the file to be read.
     * @param lineFormat the format used to parse data strings.
     */
    public RxnNamExprSetInputFile(String fileName,
    							  RxnNameExprFormat<E, S> lineFormat) {
        super(fileName, lineFormat);
        this.lineFormat = lineFormat;
    }

    /**
     * Returns the set of reaction name expressions contained in this fle.
     * @return the set of reaction name expressions contained in this fle.
     */
    public UnmodifiableSet<E> getRxnNameExprSet() {
    	
        if (exprSet == null) {
            exprSet = parseFile();
        }

        return exprSet;
    }

    /**
     * Returns an iterator over the set of reaction name expressions contained
     * in this file.
     * 
     * @return an iterator over the set of reaction name expressions contained
     * in this file.
     */
    public Iterator<E> iterator() {
        return getRxnNameExprSet().iterator();
    }

    private UnmodifiableSet<E> parseFile() {

    	// Creating storage for all reaction name expressions
        Set<E> set = new LinkedHashSet<E>();

        // Getting an iterator over this file
        TextFileLineIterator<E> i = lineIterator();
        
        while (i.hasNext()) {
        	
        	// Getting the next reaction name expression
            E expr = i.next();

            // Adding the expression to the rest
            if (!set.add(expr)) {
            	
            	// Creating an error message if the set already contains
            	// this reaction name expression
                String s = "The expression '" +
                           lineFormat.format(expr) +
                           "' appears more than once in the file '"
                           + getCanonicalName() + "'.";
                
                // Throwing a format exception
                throw new FormatException(s);
            }
        }

        return new UnmodifiableSet<E>(set);
    }

}
