/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.io.fba;

import java.util.*;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.utilities.Util;
import sbrt.shell.ApplicationException;
import sbrt.shell.io.FileFormat;
import sbrt.shell.text.fba.FbaOptFileLineFormat;

/**
 * This class is used to provide a source of serialized FBA optimization output
 * files.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * @param <F>
 *            the fluxome type.
 */
public class SfoofSource<F extends Fluxome>
	implements Iterator<SerFbaOptOutputFile<F>> {

	// The file name source
    private final Iterator<String> fileNameSource;
    
    // The file format
    private final FileFormat fileFormat;

    // The filter keys
    private final Set<String> headers;
    
    // The FBA optimization format
    private final FbaOptFileLineFormat fbaFormat;
	
    private final String fileNameSourceName;
    
    /**
     * Constructs a new source of serialized
     * FBA optimization output files.
     * 
     * @param fileNameSource the file name source.
     * @param fileFormat the file format used to write files.
     * 
     * @param filterKeys a set of valid keys that will appear
	 * in the files produced by this source. This set can consist of
	 * reaction names and the objective function header.
	 * 
	 * @param fbaFormat the map format used to supply a 
	 * header object for objective values.
     * 
     * @throws NullPointerException if any argument except {@code filterKeys}
     * is {@code null}.
     */
	public SfoofSource(Iterator<String> fileNameSource,
					   FileFormat fileFormat,
					   Set<String> filterKeys,
					   FbaOptFileLineFormat fbaFormat) {
	    
		this(fileNameSource, null, fileFormat, filterKeys, fbaFormat);
    }
    
    /**
     * Constructs a new source of serialized
     * FBA optimization output files.
     * 
     * @param fileNameSource the file name source.
     * @param fileNameSourceName the name of the file name source.
     * @param fileFormat the file format used to write files.
     * 
     * @param filterKeys a set of valid keys that will appear
     * in the files produced by this source. This set can consist of
     * reaction names and the objective function header.
     * 
     * @param fbaFormat the map format used to supply a 
     * header object for objective values.
     * 
     * @throws NullPointerException if any argument except {@code filterKeys}
     * is {@code null}.
     */
    public SfoofSource(Iterator<String> fileNameSource,
                        String fileNameSourceName,
                         FileFormat fileFormat,
                         Set<String> filterKeys,
                         FbaOptFileLineFormat fbaFormat) {
        
        // Checking inputs
        Util.nullCheck(fileNameSource);
        Util.nullCheck(fileFormat);
        Util.nullCheck(fbaFormat);
        
        // Saving
        this.fileNameSource = fileNameSource;
        this.fileFormat = fileFormat;
        this.headers = filterKeys;
        this.fbaFormat = fbaFormat;
        this.fileNameSourceName = fileNameSourceName;
    }

    /**
     * Returns the file format used to create files of 
     * type {@linkplain SerFbaOptOutputFile}.
     * 
     * @return the file format used to create files of 
     * type {@linkplain SerFbaOptOutputFile}.
     */
    public FileFormat getFileFormat() {
    	return fileFormat;
    }

    /**
     * Indicates if this iterator can return another file using the
     * {@code next()} method.
     * 
     * @return {@code true} if this iterator can return another file;
     * {@code false} otherwise.
     */
	public boolean hasNext() {
		return fileNameSource.hasNext();
	}

	/**
	 * Returns the next serialized map output file with the next
	 * name returned by the file name source.
	 * @return the next serialized map output file.
	 */
	public SerFbaOptOutputFile<F> next() {

        if (!fileNameSource.hasNext()) {
            String s =
                    "The file name source "
                            + (fileNameSourceName == null ? "" : "'"
                                    + fileNameSourceName
                                    + "' ")
                            + "has run out of file names.";
            throw new ApplicationException(s);
        }

		// Getting the next file name
		String f = fileNameSource.next();
		
		// Returning the next Romo file
		return new SerFbaOptOutputFile<F>(f, fileFormat, headers, fbaFormat);
	}

	/**
	 * Throws {@code UnsupportedOperationException}.
	 * @throws UnsupportedOperationException
	 */
	public void remove() {
		throw new UnsupportedOperationException();
	}
}
