/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng;

import java.io.PrintWriter;
import java.util.*;

import sbrt.kernel.*;
import sbrt.kernel.io.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.*;
import sbrt.shell.io.*;
import sbrt.shell.text.*;

/**
 * This class is a skeleton implemenation of {@linkplain KernelProcessManager}.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public abstract class AbstractProcessManager implements
        ProcessManager<String, String> {

    /**
     * A print writer that discards all data it receives.
     */
    public static final PrintWriter DEV_NULL = DevNull.getWriter();

    /**
     * The default format used for percentages.
     */
    public static final PercentFormat PERCENT = new PercentFormatVersion1();

    /**
     * The keyword used to denote the maximum numerical value that will still be
     * considered equal to zero.
     */
    public static final String ZERO_CUTOFF = "Zero Cutoff";

    /**
     * The keword used to denote the seed to a pseudorandom number generator.
     */
    public static final String SEED = "Seed";

    /**
     * The prefix used to denote the number of completed operations.
     */
    public static final String COMPLETED = "Completed:  ";

    /**
     * The prefix used to denote the completed percentage of an operation.
     */
    public static final String PERCENT_COMPLETED = "Percent completed:  ";

    /**
     * The prefix used to denote the amount of time that has elapsed since an
     * operation was started.
     */
    public static final String ELAPSED_TIME = "Elapsed time:  ";

    /**
     * The keyword used to denote an algorithm.
     */
    public static final String ALGORITHM = "Algorithm";

    /**
     * The keyword used to denote the number of iterations.
     */
    public static final String ITERATIONS = "Iterations";

    /**
     * The keyword used to denote the name of the output file.
     */
    public static final String OUTPUT_FILE_NAME = "Output File Name";

    /**
     * The keyword used to denote the name of a file containing a list of file
     * names.
     */
    public static final String INPUT_FILE_NAME_FILE = "Input File Name File";

    /**
     * The keyword used to denote the name of a file containing a list of file
     * names for other output files.
     */
    public static final String OUTPUT_FILE_NAME_FILE = "Output File Name File";

    /**
     * The keyword used to denote the data headers of, say, an output file.
     */
    public static final String DATA_HEADERS = "Data Headers";

    /**
     * The keyword used to denote the name of the input file.
     */
    public static final String INPUT_FILE = "Input File";

    /**
     * The keyword used to denote the name of a file format.
     */
    public static final String FILE_FORMAT = "File Format";

    /**
     * The keyword used to denote the name of an output file format.
     */
    public static final String OUTPUT_FILE_FORMAT = "Output File Format";

    /**
     * The keyword used to denote the name of an input file format.
     */
    public static final String INPUT_FILE_FORMAT = "Input File Format";

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = Collections.emptySet();
        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of optional keywords in the {@code ManagerInput} objects supplied
     * to instances of this class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
        Set<String> s = Collections.emptySet();
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Returns the parsed value for {@linkplain #OUTPUT_FILE_NAME_FILE} in the
     * provided input and prints a descriptive message to the provided print
     * writer.
     * 
     * @param input the input whose {@code OUTPUT_FILE_NAME_FILE} value is to be
     * parsed.
     * @param out the print writer to which a message is printed.
     * 
     * @return a text-based input file containing the names of output files.
     */
    public static TextInputFile<String> getFileNameSource(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting the provided file name
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME_FILE);

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Creating the input file
        TextInputFile<String> inputFile = TextInputFile.getBasic(fileName);

        // Printing a closing message
        out.println("DONE");

        return inputFile;
    }

    /**
     * Throws an ApplicationException if this {@code ProcessManager} cannot
     * understand the provided input.
     * 
     * @param input ManagerInput
     */
    public void check(ManagerInput<String, String> input) {

        // Getting all required and optional keys
        // for this type of process
        Set<String> requiredKeys = getRequiredKeys();
        Set<String> optionalKeys = getOptionalKeys();

        // Creating storage for all valid keys
        int size = requiredKeys.size() + optionalKeys.size() + 1;
        Set<String> validKeys = new HashSet<String>(size, 1.0F);

        // Creating a set of all valid keys
        validKeys.addAll(requiredKeys);
        validKeys.addAll(optionalKeys);

        // Removing all valid keys from the provided keys
        // to see if any extra keys are present.
        Set<String> actualKeys = input.toMap().keySet();
        actualKeys.removeAll(validKeys);

        // If extra keys are present
        if (!actualKeys.isEmpty()) {
            String name = input.getName();
            String s =
                    "The entry '"
                            + actualKeys.iterator().next()
                            + (name == null ? "" : "' in '" + name)
                            + "' is invalid for this type of process.";
            throw new ApplicationException(s);
        }

        // Ensuring all required keys are present
        for (String k : requiredKeys) {
            input.getRequiredValue(k);
        }

    }

    /**
     * Runs the provided process using the provided results manager and prints
     * descriptive progress messages to the provided print writer.
     * 
     * @param <T> the results type.
     * @param process the process to run.
     * @param resultsManager the results manager with which to supply the
     * process.
     * @param out the print writer where descriptive progress messages will be
     * written.
     */
    public static <T> void 
    runProcess(final MonitoredProcess<T> process,
    		   final ResultsManager<T> resultsManager,
    		   final PrintWriter out) {

        // Setting wait interval
        long waitInterval = 100;

        // Creating timer
        Timer t = new Timer(true);

        // Creating an anonymous inner TimerTask
        t.scheduleAtFixedRate(new TimerTask() {

            private Number previous = null;

            public void run() {

                // Getting the progress of this process
                Number n = process.getProgress().getProgress();

                // If the progress has changed
                if (!n.equals(previous)) {
                    printProgress(n, out);
                    previous = n;
                }
            }

        }, 0, waitInterval);

        try {

            // Printing separating line
            out.println();

            // Running process
            process.run(resultsManager);

        } finally {

            // Stopping the timer
            stopTimer(t, waitInterval);

            // Printing progress one last time
            printProgress(process.getProgress().getProgress(), out);

            // Printing line break after progress
            out.println();

            // Printing elapsed time
            long duration = process.getProgress().getDuration();
            // out.println(ELAPSED_TIME + duration / 1000 + " seconds");
            out.println(ELAPSED_TIME + duration + " milliseconds");

            // Printing a separating line
            out.println();
        }

    }

    /**
     * Closes the provided file and prints a descriptive message to the provided
     * print writer. If the file is already closed, nothing is done.
     * 
     * @param rm the results manager to close.
     * @param out the print writer to which a message will be printed.
     * 
     * @throws NullPointerException if either argument is null.
     */
    public static void closeRm(ResultsManager<?> rm, PrintWriter out) {

        // Throwing NullPointerException if necessary
        Util.nullCheck(out);

        // Do nothing if file is already closed
        if (rm.isClosed()) {
            return;
        }

        // Printing an opening message
        out.print("Closing '" + rm.getName() + "' ... ");
        out.flush();

        // Closing the file
        rm.close();

        // Printing a closing message
        out.println("DONE");
    }

    /**
     * Closes the provided results manager and prints a descriptive message to
     * the provided print writer. If the file is already closed, nothing is
     * done.
     * 
     * @param file the file to close.
     * @param out the print writer to which a message will be printed.
     * 
     * @throws NullPointerException if either argument is null.
     */
    public static void closeFile(SbrtFile file, PrintWriter out) {

        // Throwing NullPointerException if necessary
        Util.nullCheck(out);

        // Do nothing if file is already closed
        if (file.isClosed()) {
            return;
        }

        // Printing an opening message
        out.print("Closing '" + file.getOriginalName() + "' ... ");
        out.flush();

        // Closing the file
        file.close();

        // Printing a closing message
        out.println("DONE");
    }

    private static void stopTimer(Timer t, long interval) {

        // Telling the thread to stop running
        t.cancel();

        // Waiting for any remaining thread activities
        try {

            Thread.sleep(interval + 10);

        } catch (InterruptedException e) {
            String s = "The main thread was interrupted.";
            throw new ApplicationException(s, e);
        }
    }

    /**
     * Prints the provided progress to the provided print writer.
     * 
     * @param progress a number representing a progress.
     * @param out the print writer to which the progress will be printed.
     * 
     * @throws NullPointerException if either argument is null.
     */
    public static void printProgress(Number progress, PrintWriter out) {

        // If the provided number is a percentage
        if (progress instanceof Double) {
            printProgress((Double) progress, out);
        }

        // If it's a count
        else if (progress instanceof Integer) {
            printProgress((Integer) progress, out);
        }

        else {
            throw new ClassCastException(progress.getClass().getName());
        }

    }

    /**
     * Prints the provided progress to the provided print writer.
     * 
     * @param progress a number representing a progress.
     * @param out the print writer to which the progress will be printed.
     * 
     * @throws NullPointerException if either argument is null.
     */
    public static void printProgress(Integer progress, PrintWriter out) {
        out.print('\r');
        out.print(COMPLETED);
        out.print(progress);
        out.flush();
    }

    /**
     * Prints the provided progress to the provided print writer.
     * 
     * @param progress a number representing a progress.
     * @param out the print writer to which the progress will be printed.
     * 
     * @throws NullPointerException if either argument is null.
     */
    public static void printProgress(Double progress, PrintWriter out) {
        out.print('\r');
        out.print(PERCENT_COMPLETED);
        out.print(PERCENT.format(progress));
        out.flush();
    }

    /**
     * Parses the provided string as a {@code long}.
     * 
     * @param number the string to parse.
     * @return the {@code long} value represented by the provided string.
     */
    public static long parseLong(String number) {
        return new LongFormatV1().parsePrimitive(number);
    }

    /**
     * Parses the provided string as a {@code int}.
     * 
     * @param number the string to parse.
     * @return the {@code int} value represented by the provided string.
     */
    public static int parseInt(String number) {
        return new IntegerFormatV1().parsePrimitive(number);
    }

    /**
     * Returns the parsed value for {@linkplain #SEED} in the provided input and
     * prints a descriptive message to the provided print writer. If the
     * provided input does not contain a value for {@code SEED}, the current
     * time in milliseconds is returned instead.
     * 
     * @param input the input whose {@code SEED} value is to be parsed.
     * @param out the print writer to which a message is printed.
     * 
     * @return the parsed seed value; or {@linkplain System#currentTimeMillis()}
     * if no value is present.
     */
    public static long getSeed(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting String value
        String s = input.getOptionalValue(SEED);

        if (s == null) {
            return System.currentTimeMillis();
        }

        // Parsing
        long seed = parseLong(s);

        // Printing
        out.println("Parsed " + SEED + ": " + seed);

        // Returning seed
        return seed;
    }

    /**
     * Returns the parsed value for {@linkplain #ITERATIONS} in the provided input
     * and prints a descriptive message to the provided print writer.
     * 
     * @param input the input whose {@code ITERATIONS} value is to be parsed.
     * @param out the print writer to which a message is printed.
     * 
     * @return the parsed number of iterations.
     * 
     * @throws ApplicationException if the parsed integer is less than 0.
     */
    public static int getIterations(ManagerInput<String, String> input,
    								PrintWriter out) {

        // Getting String value
        String x = input.getRequiredValue(ITERATIONS);

        int iterations;
        
        // Parsing
        try {
        	iterations = parseInt(x);
        } catch (ApplicationException e) {
        	String s = "An error occurred while parsing the value provided for '" +
        				ITERATIONS + "'. " + e.getMessage();
        	throw new ApplicationException(s);
        }

        // Checking its value
        if (iterations < 0) {

            String s =
                    "The value '"
                            + iterations
                            + "' for the keyword '"
                            + ITERATIONS
                            + "' in '"
                            + input.getName()
                            + "' is invalid.";

            throw new ApplicationException(s);

        }

        // Printing
        out.println("Parsed " + ITERATIONS + ": " + iterations);

        // Returning iterations
        return iterations;
    }

    /**
     * Returns the parsed value for {@linkplain #FILE_FORMAT} in the provided input.
     * 
     * @param input the input whose {@code FILE_FORMAT} value is to be parsed.
     * 
     * @return the parsed file format.
     */
    public static FileFormat getFileFormat(ManagerInput<String, String> input) {

        // If no input value was provided
        if (!input.hasValueFor(FILE_FORMAT)) {
            return FileFormat.TEXT;
        }

        // Getting file format name
        String name = input.getRequiredValue(FILE_FORMAT);

        // Returning parsed name
        return getFileFormat(name);
    }

    /**
     * Returns the text file format corresponding to the provided string.
     * @param fileFormatString a string parsable by {@linkplain FileFormatFormat}.
     * @return {@linkplain FileFormat#TEXT} if the provided string is {@code null},
     * or {@linkplain FileFormat#TEXT} or {@linkplain FileFormat#GZIP_TEXT}.
     * 
     * @throws ApplicationException if the provided string cannot be parsed, or is parsed
     * as any file format other than plain or gzipped text.
     */
    public static FileFormat getTextFileFormat(String fileFormatString) {

        // Getting the file format
        FileFormat f = getFileFormat(fileFormatString);

        // Ensuring it's a text format
        if (!TextOutputFile.isValidFormat(f)) {

            // Creating file format-format
            FileFormatFormat formatter = new FileFormatFormat();

            // Creating error message
            String s =
                    "The file format '"
                            + formatter.format(f)
                            + "' is not a text-based file format.";

            throw new ApplicationException(s);
        }

        return f;
    }
    
    /**
     * Returns the parsed value for {@linkplain #FILE_FORMAT} in the provided input.
     * 
     * @param input the input whose {@code FILE_FORMAT} value is to be parsed.
     * 
     * @return the parsed file format.
     * 
     * @throws ApplicationException if the parsed format is not a valid
     * text-based file format.
     */
    public static FileFormat getTextFileFormat(ManagerInput<String, String> input) {

        // Getting the file format string
        String fileFormatString = input.getOptionalValue(FILE_FORMAT);
        
        return getTextFileFormat(fileFormatString);
    }

    /**
     * Parses and returns the file format corresponding to the provided string.
     * 
     * @param formatName the name of a file format.
     * 
     * @return the parsed file format.
     */
    public static FileFormat getFileFormat(String formatName) {

        // If null, return default format
        if (formatName == null) {
            return FileFormat.TEXT;
        }

        // Creating file format parser
        FileFormatFormat parser = new FileFormatFormat();

        // Returning parsed name
        return parser.parse(formatName);
    }

    /**
     * Returns the parsed value for {@linkplain #ZERO_CUTOFF} in the provided input.
     * 
     * @param input the input whose {@code ZERO_CUTOFF} value is to be parsed.
     * @param out the print writer to which a status message will be printed.
     * 
     * @return the parsed numerical zero cutoff; or {@code 0d} if no value for
     * {@code ZERO_CUTOFF} exists.
     * 
     * @throws ApplicationException if the parsed cutoff is infinite, NaN, or <
     * 0.
     */
    public static double getZeroCutoff(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting provided value
        String x = input.getOptionalValue(ZERO_CUTOFF);

        if (x == null) {
            return 0.0d;
        }

        // Creating double parser
        DoubleFormat formatter = new DoubleFormatV1();

        // Parsing it
        double cutoff = formatter.parsePrimitive(x);

        // Checking its value
        if (cutoff < 0 || Double.isInfinite(cutoff) || Double.isNaN(cutoff)) {

            String s =
                    "The value '"
                            + cutoff
                            + "' for '"
                            + ZERO_CUTOFF
                            + "' in '"
                            + input.getName()
                            + "' is invalid.";

            throw new ApplicationException(s);

        }

        out.println("Parsed " + ZERO_CUTOFF + ": " + formatter.format(cutoff));

        // Returning parsed value
        return cutoff;
    }

    /**
     * Returns the parsed value for {@linkplain #ZERO_CUTOFF} in the provided input.
     * 
     * @param input the input whose {@code ZERO_CUTOFF} value is to be parsed.
     * 
     * @return the parsed numerical zero cutoff.
     * 
     * @throws ApplicationException if the parsed cutoff is infinite, NaN, or <
     * 0.
     */
    public static double getZeroCutoff(ManagerInput<String, String> input) {
        return getZeroCutoff(input, DEV_NULL);
    }

    /**
     * Uses the provided input to prepare this process manager for execution.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     */
    public void setInput(ManagerInput<String, String> input) {
        setInput(input, DEV_NULL);
    }

    /**
     * Executes the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager.
     */
    public void execute() {
        execute(DEV_NULL);
    }

}
