/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng;

import java.util.*;

/**
 * This interface is used to represent the input to 
 * {@linkplain KernelProcessManager} objects. A {@code ManagerInput}
 * should contain all the information necessary to construct
 * a process.
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public interface ManagerInput<K, V> {

	/**
	 * Indicates if a value exists for the provided key
	 * in this manager input.
	 * 
	 * @param key the key whose presence in this input 
	 * is to be tested.
	 * 
	 * @return {@code true} if this input contains a 
	 * value for the specified key.
	 */
    public boolean hasValueFor(Object key);

    /**
     * Returns the value corresponding to the provided key.
     * 
     * @param key the key whose value is to be returned.
     * 
     * @return the value corresponding to the provided key.
     * 
     * @throws sbrt.shell.ApplicationException if the value does not exist.
     */
    public V getRequiredValue(Object key);

    /**
     * Returns the value corresponding to the provided key.
     * 
     * @param key the key whose value is to be returned.
     * 
     * @return the value corresponding to the provided key, or 
     * {@code null} if the value does not exist.
     */
    public V getOptionalValue(Object key);

    /**
     * Returns the information contained in this input as a map.
     * @return keys mapped to their respective values.
     */
    public Map<K, V> toMap();

    /**
     * Returns the name of this manager input.
     * @return the name of this manager input.
     */
    public String getName();

}
