/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng;

import java.util.*;

import sbrt.kernel.utilities.*;
import sbrt.shell.*;

/**
 * This class is a concrete implemenation of {@linkplain ManagerInput}. It provides 
 * a way to construct manager inputs from maps.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public class MapManagerInput<K, V> implements ManagerInput<K, V> {

    private final UnmodifiableMap<K, V> input;
    private final String name;

    /**
     * Constructs a new manager input from the provided map and
     * name.
     * 
     * @param input keys mapped to their respective values.
     * @param name the name given to this manager input.
     */
    public MapManagerInput(Map<K, V> input, String name) {

    	// Determining appropriate size
    	int size = input.size() + 1;
    	
    	// Creating storage
        Map<K, V> temp = 
        	new LinkedHashMap<K, V>(size, 1.0F);

        // Iterating thru each entry
        for (Map.Entry<K, V> e : input.entrySet()) {
            temp.put(e.getKey(), e.getValue());
        }

        // Saving
        this.input = new UnmodifiableMap<K, V>(temp);
        this.name = name;
    }

    /**
     * Constructs a new manager input from the provided map with a
     * {@code null} name.
     * 
     * @param input keys mapped to their respective values.
     */
    public MapManagerInput(Map<K, V> input) {
        this(input, null);
    }

	/**
	 * Indicates if a value exists for the provided key
	 * in this manager input.
	 * 
	 * @param key the key whose presence in this input 
	 * is to be tested.
	 * 
	 * @return {@code true} if this input contains a 
	 * value for the specified key.
	 */
    public boolean hasValueFor(Object key) {
        return input.containsKey(key);
    }

    /**
     * Returns the value corresponding to the provided key.
     * 
     * @param key the key whose value is to be returned.
     * 
     * @return the value corresponding to the provided key, or 
     * {@code null} if the value does not exist.
     */
    public V getOptionalValue(Object key) {
        return input.get(key);
    }

    /**
     * Returns the value corresponding to the provided key.
     * 
     * @param key the key whose value is to be returned.
     * 
     * @return the value corresponding to the provided key.
     * 
     * @throws sbrt.shell.ApplicationException if the value does not exist.
     */
    public V getRequiredValue(Object key) {

        V value = getOptionalValue(key);

        if (value == null) {
            String s = "No value was provided for '" + key +
                       (name == null ? "" : "' in '" + name) + "'.";
            throw new ApplicationException(s);
        }

        return value;
    }

    /**
     * Returns the information contained in this input as a map.
     * @return keys mapped to their respective values.
     */
    public Map<K, V> toMap() {
        
    	// Determining appropriate size
    	int size = input.size() + 1;
    	
    	// Creating storage
    	LinkedHashMap<K, V> input = 
    		new LinkedHashMap<K, V>(size, 1.0F);
        
    	// Copying
    	input.putAll(this.input);
        
    	return input;
    }

    /**
     * Returns the name of this manager input.
     * @return the name of this manager input.
     */
    public String getName() {
        return name;
    }

}
