/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng;

import java.io.*;
import java.util.*;

import sbrt.shell.ApplicationException;

/**
 * This interface is used to represent process managers.
 * A process manager is used to manage a process executed by the Systems Biology
 * Research Tool.
 * 
 * <P>
 * A process manager is created from a {@code ManagerInput}
 * object. Process managers require a minimum set of key-value 
 * pairs to perform their action.
 * In addition to this minimum set, a set of optional
 * key-value pairs may also be allowed. If the provided {@code ManagerInput}
 * contains any additional keys beyond the set of required and optional keys,
 * an {@code ApplicationException} will be thrown. 
 * 
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public interface ProcessManager<K, V> {

	/**
	 * Returns the set of keys the provided
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keys the provided
	 * {@code ManagerInput} object must contain.
	 */
    public Set<K> getRequiredKeys();

    /**
     * Returns the set of keys the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keys the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public Set<K> getOptionalKeys();

    /**
     * Uses the provided input to prepare this process
     * manager for execution.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     */
    public void setInput(ManagerInput<K, V> input);

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<K, V> input, PrintWriter out);

    /**
     * Executes the process this process manager controls.
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute();

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where some results or status messages
     * will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out);

    /**
     * Throws an {@linkplain ApplicationException} if this {@code ProcessManager}
     * cannot understand the provided input.
     * @param input the manager input to be checked.
     */
    public void check(ManagerInput<K, V> input);
}
