/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.algebra;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.LinearComb.Term;
import sbrt.kernel.math.algebra.DoubleMatrix;
import sbrt.kernel.math.algebra.TreeMatrix;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.ComoFile;
import sbrt.shell.io.InputFile;
import sbrt.shell.io.TextInputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.ShellProcessManager;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.NonEmptyStringFormat;
import sbrt.shell.text.math.LinearCombFormat;
import sbrt.shell.text.math.LinearCombFormatV1;

/**
 * This class is used to convert linear combinations to matrices.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class LinearCombConverterManager 
    extends AbstractProcessManager implements ShellProcessManager<String, String> {

    
	/**
	 * The set of keywords required by this process manager.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {

		Set<String> s = new TreeSet<String>();

        s.add(INPUT_FILE);
        s.add(OUTPUT_FILE_NAME);
        
		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by this process
	 * manager.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
        Set<String> s = Collections.emptySet();
        optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Returns the set of keywords required by this process manager.
	 * 
	 * @return the set of keywords required by this process manager.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
     * Returns the set of keywords understood, but not required, by this process
     * manager.
     * 
     * @return the set of keywords understood, but not required, by this process
     * manager.
     */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

    private String inputFileName;
    private String outputFileName;
    
	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input all of the information required to prepare this process
	 * manager for execution.
	 * 
	 * @param out the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

        inputFileName = input.getRequiredValue(INPUT_FILE);
        
        outputFileName = input.getRequiredValue(OUTPUT_FILE_NAME);

	}

	private void checkState() {

		if (inputFileName == null) {
			String s = "No input was provided.";
			throw new IllegalStateException(s);
		}
	}

	/**
	 * Executes the process this process manager controls and writes informative messages
	 * to the provided print writer.
	 * 
	 * @param out the print writer where informative messagse will be written.
	 * 
	 * @throws IllegalStateException if no input was provided to this process manager.
	 */
	public void execute(PrintWriter out) {

		checkState();
        
        // Creating formats
        NonEmptyStringFormat keyFormat = new NonEmptyStringFormat();
        DoubleFormat valFormat = new DoubleFormatV1();
        
        MapFormat<String, Double> mapFormat = 
            new EqualsMapFormatV1<String, Double>(keyFormat, valFormat);
        
        LinearCombFormat<String> expFormat = new LinearCombFormatV1<String>(keyFormat);
        
        
        out.print("Reading '" + inputFileName + "' ... ");
        out.flush();
        
        InputFile<LinearComb<String>> input =
            new TextInputFile<LinearComb<String>>(inputFileName, expFormat);
        
        DoubleMatrix<?, String> matrix = getMatrix(input);
        
        input.close();
        
        out.println("DONE");
        
        
        out.print("Writing '" + outputFileName + "' ... ");
        out.flush();
        
        Set<String> keys = matrix.getColumnIndices();
        
        ComoFile<String, Double> output = 
            new ComoFile<String, Double>(outputFileName, keys, mapFormat);
        
        Set<?> rowIndices = matrix.getRowIndices();
        Set<String> columnIndices = matrix.getColumnIndices();
        
        for (Object rowIndex : rowIndices) {
            
            Map<String, Double> map = makeMap(matrix, columnIndices, rowIndex);
            
            output.addData(map);
        }
        
        output.close();
        
        out.println("DONE");
	}

    private DoubleMatrix<?, String> getMatrix(InputFile<LinearComb<String>> input) {

        TreeMatrix<Integer, String> matrix = new TreeMatrix<Integer, String>();
        
        Integer rowIndex = 1;
        
        for (LinearComb<String> expr : input) {
            
            for (Term<String> t : expr) {
                
                double coeff = t.getCoefficient();
                String variable = t.getVariable();
                
                matrix.setElement(rowIndex, variable, coeff);
            }
            
            rowIndex = rowIndex + 1;
        }
        
        return matrix;
    }

    private Map<String, Double> makeMap(DoubleMatrix<?, String> matrix,
                                        Set<String> columnIndices,
                                        Object rowIndex) {
        
        Map<String, Double> map = new TreeMap<String, Double>();

        for (String colIndex : columnIndices) {
            
            double value = matrix.getElement(rowIndex, colIndex);
            
            map.put(colIndex, value);
        }
        
        return map;
    }
    
}
