/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.algebra;

import java.io.PrintWriter;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.ext.ExternalSoftwareException;
import sbrt.ext.mma.algebra.MathematicaSolver;
import sbrt.kernel.io.SingleRm;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.algebra.LinearSystemSolution;
import sbrt.kernel.math.algebra.proc.LinearSolverProcess;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.io.RomiFile;
import sbrt.shell.io.RomoFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.KernelProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.NonEmptyStringFormat;
import sbrt.shell.text.math.LinearCombFormat;
import sbrt.shell.text.math.LinearCombFormatV1;

/**
 * This class is used to manage linear system solvers.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class LinearSolverManager extends AbstractProcessManager implements
        KernelProcessManager<String, String> {

    /**
     * The keyword used to denote the Mathematica kernel link command.
     */
    public static final String KERNEL_LINK_CMD = "Kernel Link Command";

    /**
     * The keyword used to denote the name of the file containing the system of
     * linear equations.
     */
    public static final String EQUATIONS_FILE = "Equations File";

    /**
     * The set of keywords required by this process manager.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();
        s.add(EQUATIONS_FILE);
        s.add(OUTPUT_FILE_NAME);
        s.add(KERNEL_LINK_CMD);
        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by this process
     * manager.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
        Set<String> s = new TreeSet<String>();
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Returns the linear system solver specified by {@linkplain #KERNEL_LINK_CMD}
     * in the provided input.
     * 
     * @param input the input containing a value for the the keyword {@code MMA_EXECUTABLE}
     * @param out the print writer to which status messages will be printed.
     * 
     * @return the linear system solver specified by {@linkplain #KERNEL_LINK_CMD}
     * in the provided input.
     */
    public static MathematicaSolver<String> getSolver(ManagerInput<String, String> input,
            PrintWriter out) {

        out.print("Initializing Mathematica ... ");
        out.flush();

        // Getting the location of the MMa executable
        String mathKernel = input.getRequiredValue(KERNEL_LINK_CMD);

        // Creating solver
        MathematicaSolver<String> solver = null;

        try {
        	
            solver = new MathematicaSolver<String>(mathKernel);
            
        } catch (ExternalSoftwareException e) {
            throw interpret(e);
        }

        out.println("DONE");

        return solver;
    }

    /**
     * Interprets the provided external software exception. The provided
     * exception must have been thrown by the {@code LinearSystemSolver} returned by
     * {@linkplain #getSolver(ManagerInput, PrintWriter)}.
     * 
     * @param e the external software exception to be interpretted.
     * 
     * @return the provided exception wrapped as an application exception.
     */
    public static ApplicationException interpret(ExternalSoftwareException e) {

        String s =
                "Mathematica threw an exception. Its error message is '"
                        + e.getCause().getMessage().trim()
                        + "'.";

        return new ApplicationException(s, e);
    }

    private static RomoFile<String, LinearComb<String>> 
        getOutputFile(ManagerInput<String, String> input, PrintWriter out) {

        // Getting file name
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Key formatter
        NonEmptyStringFormat keyFormat = new NonEmptyStringFormat();
        
        // Value formatter
        LinearCombFormat<String> valueFormat = LinearCombFormatV1.getBasic();

        MapFormat<String, LinearComb<String>> format =
                new EqualsMapFormatV1<String, LinearComb<String>>(keyFormat,
                                                                  valueFormat);

        // Opening output file
        RomoFile<String, LinearComb<String>> file =
                new RomoFile<String, LinearComb<String>>(fileName, format);

        // Printing closing message
        out.println("DONE");

        return file;
    }

    /**
     * Returns the set of linear equations contained in the file with the
     * provided name.
     * 
     * @param input the input containing the name of the file containing the set
     * of equations.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return the set of equations contained in the file with the provided
     * name.
     */
    public static Map<LinearComb<String>, Double> getEquations(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting file name
        String fileName = input.getRequiredValue(EQUATIONS_FILE);

        return getEquations(fileName, out);
    }

    /**
     * Returns the set of linear equations contained in the file with the
     * provided name.
     * 
     * @param fileName the name of the file containing the set of equations.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return the set of equations contained in the file with the provided
     * name.
     */
    public static Map<LinearComb<String>, Double> getEquations(String fileName,
            PrintWriter out) {

        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        Map<LinearComb<String>, Double> m = getEquations(fileName);

        out.println("DONE");

        return m;
    }

    /**
     * Returns the set of linear equations contained in the file with the
     * provided name.
     * 
     * @param fileName the name of the file containing the set of equations.
     * 
     * @return the set of equations contained in the file with the provided
     * name.
     */
    public static Map<LinearComb<String>, Double> getEquations(String fileName) {

        // Creating linear combination parser
        LinearCombFormat<String> combFormat = LinearCombFormatV1.getBasic();

        // Creating value formatter
        DoubleFormat valueFormat = new DoubleFormatV1();
        
        // Creating map parser
        MapFormat<LinearComb<String>, Double> mapFormat =
                new EqualsMapFormatV1<LinearComb<String>, Double>(combFormat,
                        valueFormat);
        // Opening file
        RomiFile<LinearComb<String>, Double> file =
                new RomiFile<LinearComb<String>, Double>(fileName, mapFormat);

        // Reading equations
        Map<LinearComb<String>, Double> equations = file.getMap();

        // Closing file
        file.close();

        return equations;
    }

    // The process this process manager executes
    private LinearSolverProcess<String> process;
    private MathematicaSolver<String> solver;
    private RomoFile<String, LinearComb<String>> outputFile;

    /**
     * Returns the set of keywords required by this process manager.
     * 
     * @return the set of keywords required by this process manager.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords understood, but not required, by this process
     * manager.
     * 
     * @return the set of keywords understood, but not required, by this process
     * manager.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process controlled by this process manager.
     * @return the process controlled by this process manager.
     */
    public LinearSolverProcess<String> getProcess() {
        checkState();
        return process;
    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        Map<LinearComb<String>, Double> equations = getEquations(input, out);
        outputFile = getOutputFile(input, out);

        out.println();
        solver = getSolver(input, out);
        
        process = new LinearSolverProcess<String>();
        process.setEquations(equations);
        process.setSolver(solver);
        
    }

    private void checkState() {

        if (process == null) {
            String s = "No input was provided.";
            throw new IllegalStateException(s);
        }
    }

    /**
     * Executes the process this process manager controls and writes the results
     * to the provided print writer.
     * 
     * @param out the print writer where the results will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager.
     */
    public void execute(PrintWriter out) {
        
        checkState();

        try {

            // Creating simple results manager
            SingleRm<LinearSystemSolution<String>> rm = 
                new SingleRm<LinearSystemSolution<String>>();

            out.println();
            out.print("Solving the linear system ... ");
            out.flush();
            
            process.run(rm);        

            Map<String, LinearComb<String>> m = rm.getResults().toMap();
            
            for (Map.Entry<String, LinearComb<String>> e : m.entrySet()) {
                outputFile.addData(e);
            }

            out.println("DONE");
            out.println();

        } catch (ExternalSoftwareException e) {
            out.println();
            out.println();
            throw interpret(e);
        } finally {
            closeFile(outputFile, out);
            solver.close();
        }
    }


}
