/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.algebra;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Map.Entry;

import sbrt.kernel.math.HashLinearComb;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.HashLinearComb.Term;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.ComiFile;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.ShellProcessManager;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.NonEmptyStringFormat;
import sbrt.shell.text.math.LinearCombFormat;
import sbrt.shell.text.math.LinearCombFormatV1;

/**
 * This class is used to convert matrices to linear combinations.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class MatrixConverterManager 
    extends AbstractProcessManager implements ShellProcessManager<String, String> {

    
	/**
	 * The set of keywords required by this process manager.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {

		Set<String> s = new TreeSet<String>();

        s.add(INPUT_FILE);
        s.add(OUTPUT_FILE_NAME);
        
		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by this process
	 * manager.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
        Set<String> s = Collections.emptySet();
        optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Returns the set of keywords required by this process manager.
	 * 
	 * @return the set of keywords required by this process manager.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
     * Returns the set of keywords understood, but not required, by this process
     * manager.
     * 
     * @return the set of keywords understood, but not required, by this process
     * manager.
     */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

    private String inputFileName;
    private String outputFileName;
    
	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input all of the information required to prepare this process
	 * manager for execution.
	 * 
	 * @param out the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

        inputFileName = input.getRequiredValue(INPUT_FILE);
        
        outputFileName = input.getRequiredValue(OUTPUT_FILE_NAME);

	}

	private void checkState() {

		if (inputFileName == null) {
			String s = "No input was provided.";
			throw new IllegalStateException(s);
		}
	}

	/**
	 * Executes the process this process manager controls and writes informative messages
	 * to the provided print writer.
	 * 
	 * @param out the print writer where informative messagse will be written.
	 * 
	 * @throws IllegalStateException if no input was provided to this process manager.
	 */
	public void execute(PrintWriter out) {

		checkState();
        
        // Creating formats
        NonEmptyStringFormat keyFormat = new NonEmptyStringFormat();
        DoubleFormat valFormat = new DoubleFormatV1();
        
        MapFormat<String, Double> mapFormat = 
            new EqualsMapFormatV1<String, Double>(keyFormat, valFormat);
        
        LinearCombFormat<String> expFormat = new LinearCombFormatV1<String>(keyFormat);
        
        
        out.print("Opening '" + inputFileName + "' ... ");
        out.flush();
        
        ComiFile<String, Double> input = 
            new ComiFile<String, Double>(inputFileName, mapFormat);
        
        out.println("DONE");
        
        
        out.print("Opening '" + outputFileName + "' ... ");
        out.flush();
        
        OutputFile<LinearComb<String>> output =
            new TextOutputFile<LinearComb<String>>(outputFileName, expFormat);
        
        out.println("DONE");
        out.println();
        
        
        out.print("Converting ... ");
        out.flush();
        
        for (Map<String, Double> map : input) {
            
            LinearComb<String> expr = makeExpr(map);
            
            output.addData(expr);
        }
        
        out.println("DONE");
        out.println();
        
        closeFile(input, out);
        closeFile(output, out);
        
	}

    private LinearComb<String> makeExpr(Map<String, Double> map) {
        
        List<Term<String>> terms = new LinkedList<Term<String>>();
        
        for (Entry<String, Double> e : map.entrySet()) {
            
            String variable = e.getKey();
            Double coeff = e.getValue();
            
            // Skipping this term if the coefficient equals 0
            if (coeff.doubleValue() == 0d) {
                continue;
            }
            
            Term<String> t = new Term<String>(variable, coeff);
            
            terms.add(t);
        }
        
        return new HashLinearComb<String>(terms, 0d, 0d);
    }
    
}
