/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.algebra;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.io.ComiFile;
import sbrt.shell.io.InputFile;
import sbrt.shell.io.RomoFile;
import sbrt.shell.io.TextInputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.ShellProcessManager;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.MapFormat;

/**
 * This class is used to convert files containing multiple vectors into a series of
 * files, each containing a single vector.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class MultipleVectorsConverterManager 
    extends AbstractProcessManager implements ShellProcessManager<String, String> {

    
	/**
	 * The set of keywords required by this process manager.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {

		Set<String> s = new TreeSet<String>();

        s.add(INPUT_FILE);
        s.add(OUTPUT_FILE_NAME_FILE);
        
		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by this process
	 * manager.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
        Set<String> s = Collections.emptySet();
        optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Returns the set of keywords required by this process manager.
	 * 
	 * @return the set of keywords required by this process manager.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
     * Returns the set of keywords understood, but not required, by this process
     * manager.
     * 
     * @return the set of keywords understood, but not required, by this process
     * manager.
     */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

    private String inputFileName;
    private String outputFileNameFile;
    
	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input all of the information required to prepare this process
	 * manager for execution.
	 * 
	 * @param out the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

        inputFileName = input.getRequiredValue(INPUT_FILE);
        
        outputFileNameFile = input.getRequiredValue(OUTPUT_FILE_NAME_FILE);

	}

	private void checkState() {

		if (inputFileName == null) {
			String s = "No input was provided.";
			throw new IllegalStateException(s);
		}
	}

	/**
	 * Executes the process this process manager controls and writes informative messages
	 * to the provided print writer.
	 * 
	 * @param out the print writer where informative messagse will be written.
	 * 
	 * @throws IllegalStateException if no input was provided to this process manager.
	 */
	public void execute(PrintWriter out) {

		checkState();
        
        MapFormat<String, String> mapFormat = EqualsMapFormatV1.getBasic();

        
        out.print("Opening '" + inputFileName + "' ... ");
        out.flush();
        
        ComiFile<String, String> input = 
            new ComiFile<String, String>(inputFileName, mapFormat);
        
        out.println("DONE");
        
        
        out.print("Opening '" + outputFileNameFile + "' ... ");
        out.flush();
        
        InputFile<String> fileNames = TextInputFile.getBasic(outputFileNameFile);
        
        out.println("DONE");
        out.println();
        
        
        out.print("Converting ... ");
        out.flush();
        
        Iterator<Map<String, String>> i = input.iterator();
        Iterator<String> j = fileNames.iterator();
        
        while (i.hasNext()) {
            
            Map<String, String> map = i.next();
            
            
            if (!j.hasNext()) {
                String s = "The file '" + outputFileNameFile + 
                           "' contains too few output file names.";
                throw new ApplicationException(s);
            }
            
            
            String name = j.next();
            
            RomoFile<String, String> output =
                new RomoFile<String, String>(name, mapFormat);
            
            output.addData(map);
            
            output.close();
        }
        
        out.println("DONE");
        out.println();
        
        closeFile(input, out);
        closeFile(fileNames, out);
        
	}

    
}
