/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.comb;

import java.io.PrintWriter;
import java.util.*;

import sbrt.kernel.io.*;
import sbrt.kernel.math.comb.proc.SingleElementUnionProc;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.*;

/**
 * This class is used to manage single-element union computers.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public final class SingleElementUnionManager 
    extends AbstractProcessManager implements KernelProcessManager<String, String> {

	/**
	 * The keyword used to denote the name of the file containing the collection
	 * of sets.
	 */
	public static final String SETS_FILE = "Sets File";

	/**
	 * The set of keywords required by this process manager.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {

		Set<String> s = new TreeSet<String>();

		s.add(SETS_FILE);
		s.add(OUTPUT_FILE_NAME);

		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by this process
	 * manager.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {

		Set<String> s = Collections.emptySet();
		optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Returns the list of sets contained in the file with the provided name.
	 * The sets must be parsable using {@linkplain PipeSetFormat};
	 * 
	 * @param fileName
	 *            the name of the file containing a list of sets.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * 
	 * @return the list of sets contained in the file with the provided name.
	 */
	public static List<Set<String>> getSets(String fileName, PrintWriter out) {

		// Printing an opening message
		out.print("Reading '" + fileName + "' ... ");
		out.flush();

		List<Set<String>> sets = getSets(fileName);

		// Printing a closing message
		out.println("DONE");

		return sets;
	}

	/**
	 * Returns the list of sets contained in the file with the provided name.
	 * The sets must be parsable using {@linkplain PipeSetFormat};
	 * 
	 * @param fileName
	 *            the name of the file containing a list of sets.
	 * 
	 * @return the list of sets contained in the file with the provided name.
	 */
	public static List<Set<String>> getSets(String fileName) {

		// Creating set parser
		PipeSetFormat<String> f = PipeSetFormat.getBasic();

		// Opening file
		InputFile<LinkedHashSet<String>> file = new TextInputFile<LinkedHashSet<String>>(
				fileName, f);

		// Creating storage
		List<Set<String>> sets = new LinkedList<Set<String>>();

		// Reading file
		for (Set<String> s : file) {
			sets.add(s);
		}

		// Closing file
		file.close();

		return sets;
	}

	/**
	 * Returns the list of sets contained in the file with the name contained in the
	 * provided input.
	 * The sets must be parsable using {@linkplain PipeSetFormat};
	 * 
	 * @param input
	 *            the input containing the name of the file containing a list of sets.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * 
	 * @return the list of sets contained in the file with the provided name.
	 */
	public static List<Set<String>> 
	getSets(ManagerInput<String, String> input, PrintWriter out) {
		
		// Getting file name
		String fileName = input.getRequiredValue(SETS_FILE);
		
		return getSets(fileName, out);
	}
	
	/**
	 * Returns an output file with the name provided in the manager input.
	 * 
	 * @param input
	 *            the manager input containing the desired name of the output
	 *            file.
	 * @param out
	 *            the print writer to which a status message will be printed.
	 * 
	 * @return an output file with the provided name.
	 */
	public static OutputFile<Set<String>> getOutputFile(
			ManagerInput<String, String> input, PrintWriter out) {

		String fileName = input.getOptionalValue(OUTPUT_FILE_NAME);

		if (fileName == null) {
			return null;
		}

		return getOutputFile(fileName, out);
	}

	/**
	 * Returns an output file with the provided name.
	 * 
	 * @param fileName
	 *            the desired name of the output file.
	 * @param out
	 *            the print writer to which a status message will be printed.
	 * 
	 * @return an output file with the provided name.
	 */
	public static OutputFile<Set<String>> getOutputFile(String fileName,
			PrintWriter out) {

		out.print("Opening '" + fileName + "' ... ");
		out.flush();

		OutputFile<Set<String>> file = getOutputFile(fileName);

		out.println("DONE");

		return file;
	}

	/**
	 * Returns an output file with the provided name.
	 * 
	 * @param fileName
	 *            the desired name of the output file.
	 * 
	 * @return an output file with the provided name.
	 */
	public static OutputFile<Set<String>> getOutputFile(String fileName) {

		// Creating set formatter
		PipeSetFormat<String> f = PipeSetFormat.getBasic();

		// Wrapping as a simple format
		SimpleSetFormat<String> setFormat = new SimpleSetFormat<String>(f);

		// Opening file
		return new TextOutputFile<Set<String>>(fileName, setFormat);
	}

	// The process this process manager executes
	private SingleElementUnionProc<String> process;

	private OutputFile<Set<String>> output;

	/**
	 * Returns the set of keywords required by this process manager.
	 * 
	 * @return the set of keywords required by this process manager.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
	 * Returns the set of keywords understood, but not required, by this process
	 * manager.
	 * 
	 * @return the set of keywords understood, but not required, by this process
	 *         manager.
	 */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

	/**
	 * Returns the process controlled by this process manager.
	 * 
	 * @return the process controlled by this process manager.
	 */
	public SingleElementUnionProc<String> getProcess() {
		checkState();
		return process;
	}

	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input
	 *            all of the information required to prepare this process
	 *            manager for execution.
	 * 
	 * @param out
	 *            the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

		// Reading file
		List<Set<String>> sets = getSets(input, out);

		// Creating the single-element union process
		process = new SingleElementUnionProc<String>();
		process.setSets(sets);
		
		// Creating output file
		output = getOutputFile(input, out);
	}

	private void checkState() {

		if (process == null) {
			String s = "No input was provided.";
			throw new IllegalStateException(s);
		}
	}

	/**
	 * Executes the process this process manager controls and writes the results
	 * to the provided print writer.
	 * 
	 * @param out
	 *            the print writer where the results will be printed.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager.
	 */
	public void execute(PrintWriter out) {

		checkState();

		// Creating results manger
		SingleRm<Set<Set<String>>> rm = new SingleRm<Set<Set<String>>>();

		// Running process
		process.run(rm);

		// Retrieving results
		Set<Set<String>> results = rm.getResults();

		// Writing results
		for (Set<String> set : results) {
			output.addData(set);
		}
		
		closeFile(output, out);
	}

}
