/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.*;

import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.mng.*;
import sbrt.shell.proc.fba.BiggSbmlReader;
import sbrt.shell.text.FormatException;

/**
 * This class is used to manage the reading of SBML files obtained from the BiGG
 * Database.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public final class BiggSbmlReaderManager 
    extends AbstractProcessManager implements ShellProcessManager<String, String> {

	/**
	 * Used to denote the name of the file to which gene-reaction relationships
	 * should be written.
	 */
	public static final String GENE_FILE_NAME = "Gene File Name";

    /**
	 * Used to denote the name of the file to which chemical reactions should be
	 * written.
	 */
    public static final String RXN_FILE_NAME = "Reaction File Name";

    /**
     * The keyword used to denote the name of the SBML input file.
     */
    public static final String SBML_INPUT_FILE = "SBML Input File";
    
    
	/**
	 * The set of keywords required by this process manager.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {

		Set<String> s = new TreeSet<String>();

        s.add(SBML_INPUT_FILE);

		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by this process
	 * manager.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
        
        Set<String> s = new TreeSet<String>();
        
        s.add(RXN_FILE_NAME);
        s.add(GENE_FILE_NAME);
		
        optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Returns the set of keywords required by this process manager.
	 * 
	 * @return the set of keywords required by this process manager.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
	 * Returns the set of keywords understood, but not required, by this process
	 * manager.
	 * 
	 * @return the set of keywords understood, but not required, by this process
	 *         manager.
	 */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

    private String inputFileName;
    private String rxnFileName;
    private String geneFileName;
    
	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input all of the information required to prepare this process
	 * manager for execution.
	 * 
	 * @param out the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

        // Getting input file name
        inputFileName = input.getRequiredValue(SBML_INPUT_FILE);
        
        // Getting output file names
        rxnFileName = input.getOptionalValue(RXN_FILE_NAME);

        geneFileName = input.getOptionalValue(GENE_FILE_NAME);
	}

	private void checkState() {

		if (inputFileName == null) {
			String s = "No input was provided.";
			throw new IllegalStateException(s);
		}
	}

	/**
	 * Executes the process this process manager controls and writes informative
	 * messages to the provided print writer.
	 * 
	 * @param out
	 *            the print writer where informative messagse will be written.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager.
	 */
	public void execute(PrintWriter out) {

		checkState();
        
        out.print("Opening '" + inputFileName + "' ... ");
        out.flush();
        
        // Creating reader
        BiggSbmlReader reader = new BiggSbmlReader(inputFileName);
        
        out.println("DONE");
        out.println();
        
        try {

            writeFiles(reader, out);
        
        } catch (FormatException e) {
            
            String s = "An error occurred while parsing the SBML file. " +
                        e.getMessage();
            throw new ApplicationException(s);
            
        }
        
	}

    private void writeFiles(BiggSbmlReader reader, PrintWriter out) {
        
        
        if (rxnFileName != null) {
            
            out.print("Writing '" + rxnFileName + "' ... ");
            out.flush();
            
            reader.writeRxns(rxnFileName);
            
            out.println("DONE");
        }
        
        if (geneFileName != null) {
            
            out.print("Writing '" + geneFileName + "' ... ");
            out.flush();
            
            reader.writeGenes(geneFileName);
            
            out.println("DONE");
        }
        
        
    }
}
