/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.comp.CdHarFluxComp;
import sbrt.kernel.fba.comp.InvalidFluxVectorException;
import sbrt.kernel.fba.proc.CdHarFluxVectorProcess;
import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.math.ExceededLimitException;
import sbrt.kernel.math.InfeasibleSolutionException;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.UnboundedVariableException;
import sbrt.kernel.math.algebra.LinearSystemSolution;
import sbrt.kernel.math.geom.DimensionException;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.io.FileFormat;
import sbrt.shell.io.MapOutputFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.geom.CdHarManager;
import sbrt.shell.text.math.IntervalFormat;
import sbrt.shell.text.math.IntervalFormatV1;
import sbrt.shell.text.math.LinearCombFormat;
import sbrt.shell.text.math.LinearCombFormatV1;

/**
 * This class is used to manage FBA coordinate direction hit-and-run random flux
 * vector processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class CdHarFluxVectorManager extends RandomFluxVectorManager {

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        s.add(RXN_FILE);
        s.add(InitialPointManager.FLUX_INTERVAL_FILE);

        s.add(ITERATIONS);
        s.add(SYSTEM_SOLUTION_FILE);
        s.add(CdHarManager.INITIAL_POINT_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
    	
        Set<String> s = new TreeSet<String>();

        s.add(CdHarManager.MIN_CHORD_LENGTH);
        s.add(CdHarManager.MAX_OBSTRUCTED_MOVES);
        s.add(CdHarManager.POINTS_PER_POINT);
        s.add(SEED);
        s.add(CONSTRAINT_TOLERANCE);
        s.add(DATA_HEADERS);
        s.add(FILE_FORMAT);
        s.add(ZERO_CUTOFF);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Creates and returns a coordinate-direction hit-and-run flux vector
     * computer using the provided input.
     * 
     * @param input the input containing all necessary values.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return a coordinate-direction hit-and-run flux vector computer using the
     * provided input.
     * 
     * @see #getMixedFluxome(ManagerInput, PrintWriter)
     * @see #getConstraintTolerance(ManagerInput)
     * @see #getMixedComp(ManagerInput, PrintWriter, MixedFluxome, double)
     */
    public static CdHarFluxComp<MixedFluxome> 
    getMixedComp(ManagerInput<String, String> input,
    			 PrintWriter out) {

        // Getting the constraint tolerance
        double tolerance = getConstraintTolerance(input, out);

        out.println();
        
        // Creating fluxome
        MixedFluxome fluxome = 
            InitialPointManager.getMixedFluxome(input, out, tolerance);

        // Returning flux vector computer
        return getMixedComp(input, out, fluxome, tolerance);
    }

    /**
     * Creates and returns a coordinate-direction hit-and-run flux vector
     * computer using the provided input.
     * 
     * @param input the input containing all necessary values.
     * @param out the print writer to which status messages will be printed.
     * @param fluxome the fluxome with which the computer is created.
     * @param tolerance the constraint tolerance.
     * 
     * @return a coordinate-direction hit-and-run flux vector computer created
     * from the provided input.
     * 
     * @see #getFluxomeSolution(ManagerInput, PrintWriter, Fluxome, double)
     * @see #getSeed(ManagerInput, PrintWriter)
     * @see CdHarManager#getMinChordLength(ManagerInput, PrintWriter)
     */
    public static CdHarFluxComp<MixedFluxome> 
    getMixedComp(ManagerInput<String, String> input,
    			 PrintWriter out,
    			 MixedFluxome fluxome,
    			 double tolerance) {

    	out.println();
    	
    	int pointsPerPoint = CdHarManager.getPointsPerPoint(input, out);
        double zeroCutoff = getZeroCutoff(input, out);
        
    	out.println();
    	
            // Getting fluxome solution
        LinearSystemSolution<String> solution =
                getFluxomeSolution(input, out, fluxome, zeroCutoff);

        // Creating the computer
        CdHarFluxComp<MixedFluxome> computer =
                new CdHarFluxComp<MixedFluxome>(fluxome, solution);

        // Creating the random number generator
        Random rand = new Random(getSeed(input, out));
        computer.setRandom(rand);

        // Setting the points-per-point
        computer.setPointsPerPoint(pointsPerPoint);

        // Setting the constraint tolerance
        computer.setTolerance(tolerance);

        // Getting and setting the minimum chord length
        double minChordLength = CdHarManager.getMinChordLength(input, out);
        computer.setMinChordLength(minChordLength);
        
        return computer;
    }

    private static void checkBounds(CdHarFluxComp<MixedFluxome> comp) {
    	
    	IntervalFormat format = new IntervalFormatV1();
    	
    	Map<String, Interval> fluxBounds = comp.getFluxConstraints();
    	
    	double min = comp.getMinChordLength();
    	
    	for (Map.Entry<String, Interval> e : fluxBounds.entrySet()) {
    		
    		Interval i = e.getValue();
    		double range = i.getRange();
    		
    		if (range < min && range != 0) {
    			
    			String s = "The range of the provided flux interval '" +
    						format.format(i) + "' for the reaction '" +
    						e.getKey() + "' is less than the minimum " +
    						"chord length " + min + " but greater than " +
    						"zero.";
    			
    			throw new ApplicationException(s);
    			
    		}
    		
    		
    	}
    	
    	LinearCombFormat<String> exprFormat = LinearCombFormatV1.getBasic();
    	
    	Map<LinearComb<String>, Interval> exprBounds = comp.getRxnExprConstraints();

    	for (Map.Entry<LinearComb<String>, Interval> e : exprBounds.entrySet()) {
    		
    		Interval i = e.getValue();
    		double range = i.getRange();
    		
    		if (range < min && range != 0) {
    			
    			String s = "The range of the provided interval '" +
    						format.format(i) + "' for the expression '" +
    						exprFormat.format(e.getKey()) + "' is less than the minimum " +
    						"chord length " + min + " but greater than " +
    						"zero.";
    			
    			throw new ApplicationException(s);
    			
    		}
    		
    		
    	}
    	
    }

    /**
     * Opens the file denoted by {@linkplain CdHarManager#INITIAL_POINT_FILE} in the
     * provided manager input and returns the flux vector it contains.
     * 
     * @param input the input containing a value for {@code INITIAL_POINT_FILE}.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the flux vector corresponds.
     * @param tolerance the amount by which fluxes in the vector are allowed to
     * violate their defined constraints.
     * 
     * @return the flux vector contained in the file indicated by
     * {@code CdHarManager#INITIAL_POINT_FILE} in the provided input.
     * 
     * @see #getFluxVector(String, Fluxome, double)
     */
    public static Map<String, Double> 
    getFluxVector(ManagerInput<String, String> input,
    			  PrintWriter out,
    			  Fluxome fluxome,
    			  double tolerance) {

        // Getting the file name
        String fileName =
                input.getRequiredValue(CdHarManager.INITIAL_POINT_FILE);

        // Printing an opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Getting the flux vector
        Map<String, Double> fluxVector =
                getFluxVector(fileName, fluxome, tolerance);

        // Printing a closing message
        out.println("DONE");

        // Returning flux vector
        return fluxVector;
    }

    /**
     * Returns an empty output file that is suitable for storing multiple flux
     * vectors. The values for {@linkplain #OUTPUT_FILE_NAME},
     * {@linkplain #DATA_HEADERS}, and
     * {@linkplain #FILE_FORMAT} in the provided input are
     * parsed and used to call the method
     * {@linkplain #getMultiFluxVectorOutputFile(String, FileFormat, Set, Fluxome)}.
     * 
     * @param input the input containing values for the listed keywords.
     * @param out the print writer to which status messages will be printed.
     * @param fluxome the fluxome for which the data headers correspond.
     * 
     * @return an empty output file that is suitable for storing multiple flux
     * vectors.
     * 
     * @see #getRxnNameHeaders(ManagerInput, Fluxome)
     * @see #getFileFormat(ManagerInput)
     */
    public static MapOutputFile<String, Double> 
    	getMultiFluxVectorOutputFile(ManagerInput<String, String> input,
    								 PrintWriter out,
    								 Fluxome fluxome) {

        // Getting output file name
        String name = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing opening message
        out.print("Opening '" + name + "' ... ");
        out.flush();

        // Getting provided reaction name headers
        Set<String> rxnNames = getRxnNameHeaders(input, fluxome);

        // Getting file format
        FileFormat format = getFileFormat(input);

        // Creating variable for output file
        MapOutputFile<String, Double> file =
                getMultiFluxVectorOutputFile(name, format, rxnNames, fluxome);

        // Printing closing message
        out.println("DONE");

        // Returning file
        return file;
    }

    /**
     * Attempts to interpret the provided RuntimeException.
     * 
     * @param e the exception to interpret.
     * 
     * @return the interpretted exception.
     * 
     * @throws RuntimeException the orignal exception if it could not be
     * interpretted.
     */
    public static ApplicationException interpret(RuntimeException e) {

        if (e instanceof DimensionException) {
            return interpret((DimensionException) e);
        }

        if (e instanceof InfeasibleSolutionException) {
            return interpret((InfeasibleSolutionException) e);
        }

        if (e instanceof UnboundedVariableException) {
            return interpret((UnboundedVariableException) e);
        }

        if (e instanceof InvalidFluxVectorException) {
            return interpret((InvalidFluxVectorException) e);
        }

        if (e instanceof ExceededLimitException) {
            return interpret((ExceededLimitException) e);
        }

        throw e;
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e a dimension exception.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(DimensionException e) {

        String s =
                "The dimension of the provided flux space is zero. "
                        + "This ocurrs when the flux of each reaction lies "
                        + "on an interval of length zero. A random walk is not "
                        + "possible through this flux space, since only a single "
                        + "flux vector exists in this space.";

        return new ApplicationException(s, e);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e an infeasible solution exception.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(InfeasibleSolutionException e) {

        String s =
                "No feasible solution exists for the provided flux space. "
                        + "Check the flux interval for the reaction '"
                        + e.getDirection()
                        + "' and any "
                        + "other reactions that depend on it (i.e. reactions "
                        + "whose solutions contain it).";

        return new ApplicationException(s, e);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e an unbounded variable exception.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(UnboundedVariableException e) {

        String s =
                "The flux of the reaction '"
                        + e.getVariable()
                        + "' is unbounded. This can result from "
                        + "inappropriately constrained exchange reactions or if "
                        + "this reaction participates in a futile cycle.";

        return new ApplicationException(s, e);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e an invalid flux vector exception.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(InvalidFluxVectorException e) {

        String s = "The provided flux vector does not lie within the " +
        		   "defined flux space when " + e.getTolerance() +
        		   " is used as the constraint tolerance. Adjusting the " +
        		   "parameters for this process is probably necessary.";

        return new ApplicationException(s, e);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e an exceeded limit exception.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(ExceededLimitException e) {

        String s =
                "The maximum number of obstructed moves has been reached. "
                        + "This can happen when the random walker gets stuck in a "
                        + "corner of the flux space. There are several ways to "
                        + "solve this problem: "
                        + System.getProperty("line.separator")
                        + System.getProperty("line.separator")
                        + "1) The maximum number of obstructed moves can be increased,"
                        + System.getProperty("line.separator")
                        + "2) A different initial point can be used,"
                        + System.getProperty("line.separator")
                        + "3) The seed for the pseudo-random number generator can "
                        + "be changed, or"
                        + System.getProperty("line.separator")
                        + "4) The minimum chord length can be reduced."
                        + System.getProperty("line.separator");

        return new ApplicationException(s, e);
    }

    private CdHarFluxVectorProcess<MixedFluxome> process;

    private MapOutputFile<String, Double> outputFile;


    /**
     * Returns the set of keywords the provided {@code ManagerInput} object must
     * contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object must
     * contain.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Closing output file if necessary
        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        // Getting the computer
        CdHarFluxComp<MixedFluxome> computer = getMixedComp(input, out);

        // Getting the fluxome
        MixedFluxome fluxome = computer.getFluxome();

        // Getting the constraint tolerance
        double tolerance = computer.getTolerance();
        
        // Getting the initial flux vector
        Map<String, Double> fluxVector =
                getFluxVector(input, out, fluxome, tolerance);

        // Trying to set initial flux vector
        try {

            computer.setInitialFluxVector(fluxVector);

        } catch (InvalidFluxVectorException e) {
            throw interpret(e);
        }

        // This purposefully appears after the intial flux vector is set,
        // so that an InvalidFluxVectorException can be thrown first if
        // necessary.
        checkBounds(computer);
        
        // Creating the process
        process = new CdHarFluxVectorProcess<MixedFluxome>(computer);

        // Getting and setting the number of iterations
        int iterations = getIterations(input, out);
        process.setIterations(iterations);

        // Getting the points per point
        int pointsPerPoint = CdHarManager.getPointsPerPoint(input, DEV_NULL);
        process.setPointsPerPoint(pointsPerPoint);

        // Getting and setting the maximum number of obstructed moves
        int maxWasted = CdHarManager.getMaxObstructedMoves(input, out);
        process.setMaxWastedMoves(maxWasted);

        // Printing the basic fluxome information
        printFluxomeInfo(fluxome, out);

        // Getting the flux vector output file
        outputFile = getMultiFluxVectorOutputFile(input, out, fluxome);
    }
    
    /**
     * Returns the process this process manager controls.
     * 
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public CdHarFluxVectorProcess<MixedFluxome> getProcess() {

        // Checking state
        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

        return process;
    }

    /**
     * Executes the process this process manager controls, and writes the status
     * or results of this process to the provided print writer.
     * 
     * @param out the print writer where status messages will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {

        try {

            // Creating results manager
            Rm rm = new Rm(outputFile, out);

            // Running process
            runProcess(process, rm, out);

            // Printing closing message
            rm.printClosing();

        } catch (InvalidFluxVectorException e) {

            throw interpret(e);

        } catch (InfeasibleSolutionException e) {

            throw interpret(e);

        } catch (UnboundedVariableException e) {

            throw interpret(e);

        } catch (ExceededLimitException e) {

            throw interpret(e);

        } finally {

            closeFile(outputFile, out);

        }

    }

    private static class Rm 
    extends AbstractResultsManager<CdHarFluxComp<MixedFluxome>.Results> {

        private final MapOutputFile<String, Double> file;

        private final PrintWriter out;

        private int obstructedMoves;

        private Rm(MapOutputFile<String, Double> file, PrintWriter out) {
            this.file = file;
            this.out = out;
            obstructedMoves = 0;
        }

        /**
         * Adds data to this results manager.
         * 
         * @param <S> a type extending the primary result type.
         * @param results the results, or data, to be added to this manager.
         * 
         * @throws IllegalStateException if {@code close()} has already been
         * called.
         */
        public <S extends CdHarFluxComp<MixedFluxome>.Results> void addData(S results) {

            // Checking state
            if (isClosed()) {
                throw new IllegalStateException("Already closed");
            }

            // Incrementing wasted moves
            obstructedMoves += results.unchangedFreeVariables();

            // Writing flux vector to file
            file.addData(results.getFluxVector());
        }

        /**
         * Prints a closing message.
         */
        public void printClosing() {
            out.println("Total Obstructed Moves = " + obstructedMoves);
            out.println();
        }
    }

}
