/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.*;
import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.fba.comp.*;
import sbrt.kernel.fba.proc.*;
import sbrt.kernel.math.*;
import sbrt.kernel.math.optimization.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.ProgramSolverFactory;
import sbrt.shell.io.fba.FbaOptOutputFile;
import sbrt.shell.io.fba.FluxIntervalInputFile;
import sbrt.shell.mng.*;

/**
 * This class is used to manage constraint variation-FBA optimization 
 * processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class CvFbaOptManager extends FbaManager {

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
    	
        Set<String> s = new TreeSet<String>();

        // Basic Fba Optimization Inputs
        s.add(RXN_FILE);
        s.add(ProgramSolverFactory.SOLVER);
        s.add(FbaOptManager.OBJECTIVE_FUNCTION);
        s.add(FbaOptManager.OPTIMIZATION_SENSE);

        // process input file
        s.add(IntervalVariationManager.CV_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME);
        s.add(DATA_HEADERS);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
    	
        Set<String> s = new TreeSet<String>(FbaManager.optionalKeywords);

        s.add(ProgramSolverFactory.PARAM_FILE);
        
        // Allows user to specify default set of constraints
        s.add(FbaOptManager.CONSTRAINTS_FILE);
        s.add(FbaOptManager.SAFETY_LEVEL);
        s.add(CONSTRAINT_TOLERANCE);
        
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    private CvFbaOptProcess<MixedFluxome> process;
    private FluxIntervalInputFile constraintFile;
    private FbaOptOutputFile<MixedFluxome> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    private void closeFiles(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        if (constraintFile != null) {
            closeFile(constraintFile, out);
        }

    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Closing the previous files if necessary
        closeFiles(out);

        // Getting computer
        FbaOptimizer<MixedFluxome> computer = 
        	FbaOptManager.getMixedComp(input, out);

        // Getting and printing fluxome
        MixedFluxome fluxome = computer.getFluxome();
        printFluxomeInfo(fluxome, out);

        // Creating process
        process = new CvFbaOptProcess<MixedFluxome>(computer);

        // Getting and setting objective function
        MathExpr<String> objective = 
        	FbaOptManager.getObjFunc(input, out, fluxome);
        process.setObjectiveFunction(objective);

        // Getting and setting optimization sense
        OptSense sense = FbaOptManager.getOptSense(input);
        process.setOptimizationSense(sense);

        // Getting and setting constraint variation source
        double tolerance = computer.getTolerance();
        
        constraintFile = 
        	IntervalVariationManager.getCvInputFile(input, 
        											out, 
        											fluxome, 
        											tolerance);
        process.setConstraintSource(constraintFile);

        // Getting and setting results manager
        outputFile = 
        	FbaOptManager.getMultiResultsOutputFile(input, out, fluxome);
    }

    private void checkState() {

        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public CvFbaOptProcess<MixedFluxome> getProcess() {
        checkState();
        return process;
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();

        try {
        	
            runProcess(process, outputFile, out);
            
        } catch (ProgramSolverException e) {
        	throw FbaOptManager.interpret(e);
        } catch (FbaOptException e) {
            throw FbaOptManager.interpret(e, process);
        } catch (ConstraintViolationException e) {
            throw FbaOptManager.interpret(e, process);
        } finally {
            closeFiles(out);
        }
    }

}
