/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.util.*;
import java.io.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.fba.comp.*;
import sbrt.kernel.fba.proc.*;
import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.ProgramSolverFactory;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;

/**
 * This class is used to manage constraint variation-constrained reverse
 * reaction flux interval processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class CvRevRxnFluxIntervalManager extends FbaManager {

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Basic inputs
        s.add(RXN_FILE);
        s.add(ProgramSolverFactory.SOLVER);
        s.add(IntervalVariationManager.CV_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
        Set<String> s = new TreeSet<String>(FbaManager.optionalKeywords);

        // Allows user to specify default set of constraints
        s.add(FbaOptManager.CONSTRAINTS_FILE);

        // Enables control over numerical error
        s.add(ZERO_CUTOFF);
        s.add(FbaOptManager.SAFETY_LEVEL);
        s.add(CONSTRAINT_TOLERANCE);

        s.add(DATA_HEADERS);

        s.add(ProgramSolverFactory.PARAM_FILE);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    private CvFluxIntervalProcess<IrrevRxnFluxome> process;

    private InputFile<FluxConstraints> constraintFile;

    private ComoFile<String, Interval> outputFile;

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object must
     * contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object must
     * contain.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    private void closeFiles(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        if (constraintFile != null) {
            closeFile(constraintFile, out);
        }

    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Closing the previous files if necessary
        closeFiles(out);

        // Getting the computer
        ReverseRxnIntervalComp<IrrevRxnFluxome> computer =
        	ReverseRxnFluxIntervalManager.getComputer(input, out);

        // Getting and printing the fluxome
        IrrevRxnFluxome fluxome = computer.getFluxome();
        printFluxomeInfo(fluxome, out);

        // Creating the process
        process = new CvFluxIntervalProcess<IrrevRxnFluxome>(computer);

        // Getting and setting the constraint variation source
        double tolerance = computer.getTolerance();
        
        constraintFile =
        	IntervalVariationManager.getCvInputFile(input, 
        											out, 
        											fluxome, 
        											tolerance);
        process.setConstraintSource(constraintFile);

        // Getting results manager and setting reaction names
        outputFile =
        	SimpleFluxIntervalManager.getMultiResultsOutputFile(input, out,
                        fluxome);
        process.setReactionNames(outputFile.getFilterKeys());
    }

    private void checkState() {

        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Returns the process this process manager controls.
     * 
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public CvFluxIntervalProcess<IrrevRxnFluxome> getProcess() {
        checkState();
        return process;
    }

    /**
     * Executes the process this process manager controls, and writes the status
     * or results of this process to the provided print writer.
     * 
     * @param out the print writer where status messages will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {

        checkState();

        try {

            Rm rm = new Rm(outputFile);

            runProcess(process, rm, out);

        } catch (IllegalBoundsException e) {
            
            throw SimpleFluxIntervalManager.interpret(e);

        } catch (RuntimeException e) {

            throw FbaOptManager.interpret(e, process);

        } finally {
            closeFiles(out);
        }

    }

    private static class Rm extends
            AbstractResultsManager<Map<String, Interval>> {

        private final ComoFile<String, Interval> output;

        private Rm(ComoFile<String, Interval> output) {
            this.output = output;
        }

        /**
         * Adds data to this results manager.
         * 
         * @param results the results, or data, to be added to this manager.
         * 
         * @throws IllegalStateException if {@code close()} has already been
         * called.
         */
        public void addData(Map<String, Interval> results) {
            output.addData(results);
        }
    }

}
