/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.FluxConstraints;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.comp.FbaOptException;
import sbrt.kernel.fba.comp.FbaOptimizer;
import sbrt.kernel.fba.comp.RxnDeletionComputer;
import sbrt.kernel.fba.proc.CvRxnDeletionProcess;
import sbrt.kernel.io.ConflictingInputException;
import sbrt.kernel.math.ConstraintViolationException;
import sbrt.kernel.math.MathExpr;
import sbrt.kernel.math.optimization.OptSense;
import sbrt.kernel.math.optimization.ProgramSolverException;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.ProgramSolverFactory;
import sbrt.shell.io.InputFile;
import sbrt.shell.io.MultiOutputFiles;
import sbrt.shell.io.fba.FbaOptOutputFile;
import sbrt.shell.mng.ManagerInput;

/**
 * This class is used to manage FBA constraint variation-reaction deletion
 * processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class CvRxnDeletionManager extends FbaManager {


    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Basic Fba Optimization Inputs
        s.add(RXN_FILE);
        s.add(ProgramSolverFactory.SOLVER);
        s.add(FbaOptManager.OBJECTIVE_FUNCTION);
        s.add(FbaOptManager.OPTIMIZATION_SENSE);

        // process input files
        s.add(RxnDeletionManager.RXN_DELETION_FILE);
        s.add(IntervalVariationManager.CV_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME_FILE);
        s.add(DATA_HEADERS);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }


    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Allows user to specify default set of constraints
        s.add(FbaOptManager.CONSTRAINTS_FILE);
        s.add(FbaOptManager.SAFETY_LEVEL);
        s.add(CONSTRAINT_TOLERANCE);
        s.add(ZERO_CUTOFF);

        s.add(ProgramSolverFactory.PARAM_FILE);
        
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
	 * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e the exception.
     * @param fileName1 the name of the file containing one of the 
     * 					conflicting inputs.
     * 
     * @param fileName2 the name of the file containing one of the 
     * 					conflicting inputs.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException 
    interpret(ConflictingInputException e,
    		  String fileName1,
    		  String fileName2) {

        List<?> reactions = e.getConflicts();
        
        String s = "The files '" + fileName1 + "' and '" +
                   fileName2 + "' result in conflicting " +
                   "constraints for the reaction '" + reactions.get(0) +
                   "'.";
        
        return new ApplicationException(s, e);
    }
    
    
    private CvRxnDeletionProcess<MixedFluxome> process;
    private InputFile<String> fileNameSource;
    private InputFile<FluxConstraints> constraintFile;
    private InputFile<Set<String>> deletionFile;
    private MultiOutputFiles<? extends FbaOptOutputFile<MixedFluxome>, 
    						 FbaOptimizer<MixedFluxome>.Results> outputFiles;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public Set<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public Set<String> getOptionalKeys() {
        return optionalKeywords;
    }

    private void closeFiles(PrintWriter out) {

        if (outputFiles != null) {
            outputFiles.close();
        }

        if (fileNameSource != null) {
            closeFile(fileNameSource, out);
        }

        if (constraintFile != null) {
            closeFile(constraintFile, out);
        }

        if (deletionFile != null) {
            closeFile(deletionFile, out);
        }
    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Closing the previous files if necessary
        closeFiles(out);

        // Getting the computer
        RxnDeletionComputer<MixedFluxome> computer = 
        	RxnDeletionManager.getMixedDelComp(input, out);

        // Getting and printing the fluxome
        MixedFluxome fluxome = computer.getFluxome();
        printFluxomeInfo(fluxome, out);

        // Creating the process
        process = new CvRxnDeletionProcess<MixedFluxome>(computer);

        // Getting and setting the objective function
        MathExpr<String> objective = 
        	FbaOptManager.getObjFunc(input, out, fluxome);
        process.setObjectiveFunction(objective);

        // Getting and setting the optimization sense
        OptSense sense = FbaOptManager.getOptSense(input);
        process.setOptimizationSense(sense);

        // Getting and setting the reaction deletion source
        deletionFile = 
        	RxnDeletionManager.getInputFile(input, out, fluxome);
        process.setRxnDeletionSource(deletionFile);

        // Getting and setting the constraint variation source
        double tolerance = computer.getTolerance();
        constraintFile = IntervalVariationManager.getCvInputFile(input, out, fluxome, tolerance);
        process.setConstraintSource(constraintFile);

        // Getting the provided file name
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME_FILE);

        // Getting the results manager
        fileNameSource = getFileNameSource(input, out);
        Iterator<String> i = fileNameSource.iterator();
        outputFiles = 
        	FbaOptManager.getMultiFileOutput(input, out, i, fileName, fluxome);
    }

    private void checkState() {

        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public CvRxnDeletionProcess<MixedFluxome> getProcess() {
        checkState();
        return process;
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();

        try {
        	
            runProcess(process, outputFiles, out);
            
        } catch (FbaOptException e) {

            throw FbaOptManager.interpret(e, process);

        } catch (ConflictingInputException e) {

        	String fileName1 = constraintFile.getCanonicalName(); 
        	String fileName2 = deletionFile.getCanonicalName();
        	
            throw interpret(e, fileName1, fileName2);
            
        } catch (ProgramSolverException e) {
        	
        	throw FbaOptManager.interpret(e);
        	
        } catch (ConstraintViolationException e) {
        	
            throw FbaOptManager.interpret(e, process);
            
        } finally {
            closeFiles(out);
        }
    }

}
