/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.comp.ExtremeCurrentComputer;
import sbrt.kernel.fba.comp.SlpExCurrComputerV2;
import sbrt.kernel.fba.proc.ExtremeCurrentProcess;
import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.math.algebra.DoubleMatrix;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.ComoFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.fba.RxnNameVerifier;

/**
 * This class is used to manage FBA extreme current processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class ExtremeCurrentManager extends FbaManager {

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();
        s.add(RXN_FILE);
        s.add(OUTPUT_FILE_NAME);
        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /** 
     * The set of optional keywords in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
    	
    	// Getting empty set
    	Set<String> empty = Collections.emptySet();
    	
        optionalKeywords = new UnmodifiableSet<String>(empty);
    }


    /**
     * Returns an output file to which extreme currents can be written.
     * 
     * @param input the manager input from which the value for 
     * {@linkplain #OUTPUT_FILE_NAME} is obtained.
     * @param out the print writer to which status messages will be printed.
     * @param fluxome the fluxome in which the reactions are contained.
     * @param rxnNames the reactions names that will appear as column variables
     * in the extreme current matrix.
     * 
     * @return the output file with the name contained in {@code input}.
     */
    public static ComoFile<String, Double>
    getOutputFile(ManagerInput<String, String> input, 
    			  PrintWriter out, 
    			  Fluxome fluxome,
    			  Set<String> rxnNames) {

    	// Getting the file name
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        return getOutputFile(fileName, out, fluxome, rxnNames);
    }

    /**
     * Returns an output file to which extreme currents can be written.
     * 
     * @param fileName the name of the file to be created.
     * @param out the print writer to which status messages will be printed.
     * @param fluxome the fluxome in which the reactions are contained.
     * @param rxnNames the reactions names that will appear as column variables
     * in the extreme current matrix.
     * 
     * @return the output file with the name contained in {@code input}.
     */
    public static ComoFile<String, Double>
    getOutputFile(String fileName,
    			  PrintWriter out, 
    			  Fluxome fluxome,
    			  Set<String> rxnNames) {

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        if (!fluxome.areRxnNames(rxnNames)) {
        	throw new IllegalArgumentException("invalid reaction name");
        }
        
        // Creating the variable format
        RxnNameVerifier rxnFormat = new RxnNameVerifier(fluxome);

        // Creating the number format
        DoubleFormat numFormat = new DoubleFormatV1();

        // Creating the map format
        MapFormat<String, Double> mapFormat =
            new EqualsMapFormatV1<String, Double>(rxnFormat, numFormat);
        
        // Opening the output file
        ComoFile<String, Double> file = 
            new ComoFile<String, Double>(fileName, rxnNames, mapFormat);

        // Printing a closing message
        out.println("DONE");

        return file;
    }

    /**
     * Returns an extreme current computer created from the provided input.
     *   
     * @param input the manager input from which the necessary information is
     * obtained.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return an extreme current computer created from the provided input.
     */
    public static ExtremeCurrentComputer
            getComputer(ManagerInput<String, String> input, PrintWriter out) {

        String algorithm = input.getOptionalValue(ALGORITHM);
        
        if (algorithm == null) {
            return new SlpExCurrComputerV2();
        }

        String s = "The extreme pathway algorithm '" + algorithm +
        			"' is unrecognizable.";
        throw new FormatException(s);

    }

    private ExtremeCurrentProcess<IrrevRxnFluxome> process;
    private ComoFile<String, Double> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Closing the previous output file if necessary
        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        // Getting and printing the fluxome
        IrrevRxnFluxome fluxome = getIrrevRxnFluxome(input, out);
        printFluxomeInfo(fluxome, out);

        // Getting the computer
        ExtremeCurrentComputer computer = getComputer(input, out);

        // Creating the process
        process = new ExtremeCurrentProcess<IrrevRxnFluxome>(fluxome, computer);

        // Getting the results manager
        outputFile = getOutputFile(input, out, fluxome, fluxome.getRxnNames());
    }

    private void checkState() {

        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public ExtremeCurrentProcess<IrrevRxnFluxome> getProcess() {
        checkState();
        return process;
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();

        try {
        	
        	Rm rm = new Rm(outputFile);
        	
            runProcess(process, rm, out);
            
        } finally {
            closeFile(outputFile, out);
        }

    }

    /**
     * @author wright
     *
     */
    protected static class Rm 
    extends AbstractResultsManager<DoubleMatrix<?, String>> {
    	
    	private final ComoFile<String, Double> output;
    	
    	/**
    	 * Constructs a new results manager for writing extreme current
    	 * matrices.
    	 * @param output a column-oriented map output file.
    	 */
    	protected Rm(ComoFile<String, Double> output) {
    		this.output = output;
    	}

    	/**
    	 * Adds data to this results manager.
    	 * 
    	 * @param results the results, or data, to be added
    	 * to this manager.
    	 * 
    	 * @throws IllegalStateException if {@code close()}
    	 * has already been called.
    	 */
    	public void addData(DoubleMatrix<?, String> results) {
    		
            // Getting all column indices (reaction names)
            Set<String> colInds = results.getColumnIndices();
            
            // Iterating thru each row index
            for (Object rowIndex : results.getRowIndices()) {
                
                // Creating a new map
                Map<String, Double> pathway = new TreeMap<String, Double>();
                
                // Iterating thru each column index
                for (String colIndex : colInds) {
                    
                    // Getting matrix element
                    Double element = results.getElement(rowIndex, colIndex);
                    
                    pathway.put(colIndex, element);
                }
                
                // Adding the pathway vector to the output file
                output.addData(pathway);
            }

        }
    }
    
}
