/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.IrrevRxnFluxomeV1;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.MixedFluxomeV1;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.algebra.HashLinearSystemSolution;
import sbrt.kernel.math.algebra.LinearSystemSolution;
import sbrt.kernel.utilities.And;
import sbrt.kernel.utilities.Or;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.io.ComoFile;
import sbrt.shell.io.FileFormat;
import sbrt.shell.io.MapOutputFile;
import sbrt.shell.io.RomiFile;
import sbrt.shell.io.RomoFile;
import sbrt.shell.io.SerializedOutputFile;
import sbrt.shell.io.SmoFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.KernelProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.BoolRelFormatV1;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.chem.CatalystFormat;
import sbrt.shell.text.chem.CatalystFormatV1;
import sbrt.shell.text.fba.FluxVectorFormat;
import sbrt.shell.text.fba.FluxVectorFormatV1;
import sbrt.shell.text.fba.FluxomeSolFileLineFormatV1;
import sbrt.shell.text.fba.FluxomeSolutionFileLineFormat;
import sbrt.shell.text.fba.IrfFormatV1;
import sbrt.shell.text.fba.MixedFluxomeFormatV1;
import sbrt.shell.text.fba.RxnNameExprFormat;
import sbrt.shell.text.fba.RxnNameExprFormatV1;
import sbrt.shell.text.fba.RxnNameSetFormat;
import sbrt.shell.text.fba.RxnNameSetFormatV1;
import sbrt.shell.text.fba.RxnNameVerifier;

/**
 * This abstract class provides a foundation for Flux Balance Analysis process
 * managers. It contains many of the keywords required by FBA managers, as well
 * as a number of utility functions.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public abstract class FbaManager 
	extends AbstractProcessManager 
		implements KernelProcessManager<String, String> {

	/**
	 * The keyword used to denote an FBA reaction file.
	 */
	public static final String RXN_FILE = "Reaction File";
	
    /**
     * The keyword used to denote the name of a file containing reaction-catalyst
     * associations.
     */
    public static final String RXN_CATALYST_FILE = "Reaction-Catalyst File";
    
	/**
	 * The keyword used to denote a flux vector input file.
	 */
	public static final String FLUX_VECTOR_FILE = "Flux Vector File";

	/**
	 * The keyword used to denote the name of a flux vector output file.
	 */
	public static final String FLUX_VECTOR_FILE_NAME = "Flux Vector File Name";

	/**
	 * The keyword used to denote a file containing a solution to the system
	 * of linear equations formed by a fluxome.
	 */
	public static final String SYSTEM_SOLUTION_FILE = "System Solution File";

	/**
	 * The keyword used to denote a constraint tolerance.
	 */
	public static final String CONSTRAINT_TOLERANCE = "Constraint Tolerance";

	/**
	 * The default constraint tolerance.
	 */
	public static final double DEFAULT_CONSTRAINT_TOLERANCE = 1E-9;

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {
		Set<String> s = new TreeSet<String>();
		s.add(RXN_FILE);
		requiredKeywords = new UnmodifiableSet<String>(s);
	}

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
		Set<String> s = Collections.emptySet();
		optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Prints basic information about the provided fluxome.
	 * 
	 * @param fluxome the fluxome for which basic information will be printed.
	 * @param out the print writer to which basic information will be printed.
	 */
	public static void printFluxomeInfo(Fluxome fluxome, PrintWriter out) {

        out.println();
        out.println("Number of chemical species: " + fluxome.species());
        out.println("Number of chemical reactions: " + fluxome.rxns());
        out.println("Number of internal reactions: " + fluxome.internalRxns());
        out.println("Number of exchange reactions: " + fluxome.exchangeRxns());
        out.println("Number of user-defined flux constraints: "
                + fluxome.addedConstraints());
        out.println();
	}

    /**
     * Returns the default reaction name expression format.
     * 
     * @param f the fluxome for which the format will be created.
     * 
     * @return the default reaction name expression format.
     */
    public static RxnNameExprFormat<LinearComb<String>, String>
    getRxnNameExprFormat(Fluxome f) {
        return new RxnNameExprFormatV1(f);
    }

    /**
     * Parses and returns the value for {@linkplain #CONSTRAINT_TOLERANCE} in the 
     * provided manager input.
     * 
     * @param input keyword-value pairs.
     * 
     * @param out the print writer to which a status message will be printed.
     * 
     * @return the parsed value for {@linkplain #CONSTRAINT_TOLERANCE} in the
     * provided input.
     * 
     * @throws ApplicationException if the parsed tolerance is infinite, NaN, or < 0.
     */
	public static double 
	getConstraintTolerance(ManagerInput<String, String> input, 
						   PrintWriter out) {

		// Getting the input string
		String s = input.getOptionalValue(CONSTRAINT_TOLERANCE);
		
		// If no value was provided
		if (s == null) {
			return DEFAULT_CONSTRAINT_TOLERANCE;
		}

		// Creating a double parser
		DoubleFormat f = new DoubleFormatV1();

		// Parsing the constraint tolerance
		double t = f.parsePrimitive(s);
        
        // Checking its value
        if (t < 0 || Double.isInfinite(t) || Double.isNaN(t)) {

            String x = "The value '" + t + "' for '" + 
            		   CONSTRAINT_TOLERANCE +
            		   "' in '" + input.getName() + "' is invalid.";

            throw new ApplicationException(x);

        }

		// Printing a formatted message
		out.println("Parsed " + CONSTRAINT_TOLERANCE + ": " + f.format(t));

		return t;
	}

    /**
     * Parses and returns the value for {@linkplain #CONSTRAINT_TOLERANCE} in the 
     * provided manager input.
     * 
     * @param input keyword-value pairs.
     * 
     * @return the parsed value for {@linkplain #CONSTRAINT_TOLERANCE} in the
     * provided input.
     * 
     * @throws ApplicationException if the parsed tolerance is infinite, NaN, or < 0.
     */
	public static double 
	getConstraintTolerance(ManagerInput<String, String> input) {

		// Returning constraint tolerance
		return getConstraintTolerance(input, DEV_NULL);
	}

    /**
     * Parses and returns the value for {@linkplain #DATA_HEADERS} in the 
     * provided manager input. The format used to parse the data header
     * string is {@linkplain RxnNameSetFormatV1}. If no value exists for
     * {@code DATA_HEADERS}, the set of all reactions names from the
     * provided fluxome is returned.
     * 
     * @param input keyword-value pairs.
     * @param fluxome the fluxome for which the data headers correspond.
     * 
     * @return the parsed value for {@linkplain #DATA_HEADERS} in the
     * provided input.
     */
	public static Set<String> 
	getRxnNameHeaders(ManagerInput<String, String> input, Fluxome fluxome) {

		// Getting header string value
		String headers = input.getOptionalValue(DATA_HEADERS);

		// If no input value was provided, return an alphabetically
		// sorted set of all reaction names
		if (headers == null) {
			return new TreeSet<String>(fluxome.getRxnNames());
		}

		// Returning parsed reaction name headers
		return getRxnNameHeaders(headers, fluxome);
	}

    /**
     * Parses and returns the set of reaction names contained in the provided
     * string. The format used to parse the data header string is 
     * {@linkplain RxnNameSetFormatV1}.
     * 
     * @param dataHeaders the data header string to be parsed.
     * @param fluxome the fluxome for which the data headers correspond.
     * 
     * @return the set of reaction names contained in the provided string.
     */
	public static LinkedHashSet<String> 
	getRxnNameHeaders(String dataHeaders, Fluxome fluxome) {

		// Getting a reaction name header parser
		RxnNameSetFormat headerFormat = 
			new RxnNameSetFormatV1(fluxome);

		// Returning parsed reaction name headers
		return headerFormat.parse(dataHeaders);
	}

	/**
	 * Returns an empty output file that is suitable for storing a single flux
	 * vector. The values for {@linkplain #FLUX_VECTOR_FILE_NAME},
	 * {@linkplain #DATA_HEADERS}, and {@linkplain #FILE_FORMAT} in the
	 * provided input are parsed and used to call the method
	 * {@linkplain #getSingleFluxVectorOutputFile(String, FileFormat, Set, Fluxome)}.
	 * 
	 * @param input
	 *            keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fluxome
	 *            the fluxome for which the data headers correspond.
	 * 
	 * @return an empty output file that is suitable for storing a single flux
	 *         vector.
	 * 
	 * @see #getRxnNameHeaders(ManagerInput, Fluxome)
	 * @see #getFileFormat(ManagerInput)
	 */
	public static MapOutputFile<String, Double>
	getSingleFluxVectorOutputFile(ManagerInput<String, String> input,
								  PrintWriter out, 
								  Fluxome fluxome) {

		// Getting output file name
		String name = input.getRequiredValue(FLUX_VECTOR_FILE_NAME);

		// Printing opening message
		out.print("Opening '" + name + "' ... ");
		out.flush();

		// Getting provided reaction name headers
		Set<String> rxnNames = getRxnNameHeaders(input, fluxome);

		// Getting file format
		FileFormat format = getFileFormat(input);

		// Getting output file
		MapOutputFile<String, Double> file = 
			getSingleFluxVectorOutputFile(name, format, rxnNames, fluxome);

		// Printing closing message
		out.println("DONE");

		// Returning file
		return file;
	}

	/**
	 * Returns an empty output file that is suitable for storing a single
	 * flux vector.
	 * 
	 * @param fileName the name of the file to be created.
	 * @param fileFormat the format of the file to be created.
	 * @param rxnNameHeaders the set of reaction names to be written to the 
	 * file.
	 * @param fluxome the fluxome to which the reaction names correspond.
	 * 
	 * @return an empty output file that is suitable for storing a single
	 * flux vector.
	 */
	public static MapOutputFile<String, Double>
	getSingleFluxVectorOutputFile(String fileName,
								  FileFormat fileFormat, 
								  Set<String> rxnNameHeaders, 
								  Fluxome fluxome) {

		// If this is a serialized file format
		if (SerializedOutputFile.isValidFormat(fileFormat)) {

			// Creating serialized vector output file
			return new SmoFile<String, Double>(
					fileName, fileFormat, rxnNameHeaders);
		}

		// If this is a text file format
		else if (TextOutputFile.isValidFormat(fileFormat)) {

			// Getting flux vector file line format
			FluxVectorFormat format = new FluxVectorFormatV1(fluxome);

			// Returning row-oriented flux vector output file
			return new RomoFile<String, Double>(
					fileName, fileFormat, rxnNameHeaders, format);
		}

		// If it's neither
		else {
			String s = "The file format '" + fileFormat
					+ "' is not yet supported.";
			throw new IllegalArgumentException(s);
		}

	}

	/**
	 * Returns an empty output file that is suitable for storing multiple flux
	 * vectors. The values for {@linkplain #FLUX_VECTOR_FILE_NAME},
	 * {@linkplain #DATA_HEADERS}, and {@linkplain #FILE_FORMAT} in the
	 * provided input are parsed and used to call the method
	 * {@linkplain #getMultiFluxVectorOutputFile(String, FileFormat, Set, Fluxome)}.
	 * 
	 * @param input
	 *            the input containing values for the listed keywords.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fluxome
	 *            the fluxome for which the data headers correspond.
	 * 
	 * @return an empty output file that is suitable for storing multiple flux
	 *         vectors.
	 * 
	 * @see #getRxnNameHeaders(ManagerInput, Fluxome)
	 * @see #getFileFormat(ManagerInput)
	 */
	public static MapOutputFile<String, Double>
	getMultiFluxVectorOutputFile(ManagerInput<String, String> input,
								 PrintWriter out, 
								 Fluxome fluxome) {

		// Getting output file name
		String name = input.getRequiredValue(FLUX_VECTOR_FILE_NAME);

		// Printing opening message
		out.print("Opening '" + name + "' ... ");
		out.flush();

		// Getting provided reaction name headers
		Set<String> rxnNames = getRxnNameHeaders(input, fluxome);

		// Getting file format
		FileFormat format = getFileFormat(input);

		// Creating variable for output file
		MapOutputFile<String, Double> file = 
			getMultiFluxVectorOutputFile(name, format, rxnNames,
				fluxome);

		// Printing closing message
		out.println("DONE");

		// Returning file
		return file;
	}

	/**
	 * Returns an empty output file that is suitable for storing multiple
	 * flux vectors.
	 * 
	 * @param fileName the name of the file to be created.
	 * @param fileFormat the format of the file to be created.
	 * @param rxnNameHeaders the set of reaction names to be written to the 
	 * file.
	 * @param fluxome the fluxome to which the reaction names correspond.
	 * 
	 * @return an empty output file that is suitable for storing multiple
	 * flux vectors.
	 */
	public static MapOutputFile<String, Double>
	getMultiFluxVectorOutputFile(String fileName,
								 FileFormat fileFormat, 
								 Set<String> rxnNameHeaders, 
								 Fluxome fluxome) {

		// If this is a serialized file format
		if (SerializedOutputFile.isValidFormat(fileFormat)) {

			// Creating serialized vector output file
			return new SmoFile<String, Double>(
					fileName, fileFormat, rxnNameHeaders);
		}

		// If this is a text file format
		else if (TextOutputFile.isValidFormat(fileFormat)) {

			// Getting flux vector file line format
			FluxVectorFormat format = new FluxVectorFormatV1(fluxome);

			// Returning column-oriented flux vector output file
			return new ComoFile<String, Double>(
					fileName, fileFormat, rxnNameHeaders, format);
		}

		// If it's neither
		else {
			String s = "The file format '" + fileFormat
					+ "' is not yet supported.";
			throw new IllegalArgumentException(s);
		}

	}
	
    /**
	 * Opens the file denoted by {@linkplain #RXN_FILE} in the provided manager
	 * input and returns the fluxome it contains. The file denoted by
	 * {@linkplain #RXN_CATALYST_FILE} is also opened and applied if a value is
	 * provided.
	 * 
	 * @param input
	 *            keyword-value pairs.
	 * @param out
	 *            the print writer to which a status message will be printed.
	 * 
	 * @return the fluxome contained in the files indicated by
	 *         {@linkplain #RXN_FILE} and {@linkplain #RXN_CATALYST_FILE} in the
	 *         provided input.
	 */
	public static MixedFluxome 
	getMixedFluxome(ManagerInput<String, String> input, PrintWriter out) {

		// Getting the reaction file name
		String rxnFileName = input.getRequiredValue(RXN_FILE);

        // Getting the reaction-catalyst association file name,
        // if it exists.
        String catFileName = input.getOptionalValue(RXN_CATALYST_FILE);
        
        // If no catalyst file name was provided
        if (catFileName == null) {
            return getMixedFluxome(rxnFileName, out);
        }
        
		return getMixedFluxome(rxnFileName, catFileName, out);
	}
	
    /**
	 * Opens the file denoted by {@linkplain #RXN_FILE} in the provided manager
	 * input and returns the fluxome it contains. The file denoted by
	 * {@linkplain #RXN_CATALYST_FILE} is also opened and applied if a value is
	 * provided.
	 * 
	 * @param input
	 *            keyword-value pairs.
	 * @param out
	 *            the print writer to which a status message will be printed.
	 * 
	 * @return the fluxome contained in the files indicated by
	 *         {@linkplain #RXN_FILE} and {@linkplain #RXN_CATALYST_FILE} in the
	 *         provided input.
	 */
	public static IrrevRxnFluxome 
	getIrrevRxnFluxome(ManagerInput<String, String> input, PrintWriter out) {

		// Getting the reaction file name
		String rxnFileName = input.getRequiredValue(RXN_FILE);

        // Getting the reaction-catalyst association file name,
        // if it exists.
        String catFileName = input.getOptionalValue(RXN_CATALYST_FILE);
        
        // If no catalyst file name was provided
        if (catFileName == null) {
            return getIrrevRxnFluxome(rxnFileName, out);
        }
        
		return getIrrevRxnFluxome(rxnFileName, catFileName, out);
	}
    
    /**
	 * Returns the fluxome contained in the file with the provided name.
	 * 
	 * @param fileName
	 *            the name of a file containing a fluxome parsable with
	 *            {@linkplain IrfFormatV1}.
	 * @param out
	 *            the print writer to which a status message will be printed.
	 * @return the fluxome contained in the file with the provided name.
	 */
    public static IrrevRxnFluxome 
    getIrrevRxnFluxome(String fileName, PrintWriter out) {

        // Printing an opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Getting the fluxome
        IrrevRxnFluxome fluxome = getIrrevRxnFluxome(fileName);

        // Printing a closing message
        out.println("DONE");

        // Returning the fluxome
        return fluxome;
    }
    
    /**
	 * Returns the fluxome contained in the file with the provided name.
	 * 
	 * @param fileName
	 *            the name of a file containing a fluxome parsable with
	 *            {@linkplain MixedFluxomeFormatV1}.
	 * @param out
	 *            the print writer to which a status message will be printed.
	 * @return the fluxome contained in the file with the provided name.
	 */
    public static MixedFluxome 
    getMixedFluxome(String fileName, PrintWriter out) {

        // Printing an opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Getting the fluxome
        MixedFluxome fluxome = getMixedFluxome(fileName);

        // Printing a closing message
        out.println("DONE");

        // Returning the fluxome
        return fluxome;
    }
    
    /**
	 * Returns the fluxome contained in the file with the provided name.
	 * 
	 * @param fileName
	 *            the name of a file containing a fluxome parsable with
	 *            {@linkplain IrfFormatV1}.
	 * @return the fluxome contained in the file with the provided name.
	 */
    public static IrrevRxnFluxome getIrrevRxnFluxome(String fileName) {

        // Creating formatter
        IrfFormatV1 format = new IrfFormatV1();
        
        // Creating the file
        RomiFile<String, ChemRxn> file = 
            new RomiFile<String, ChemRxn>(fileName, format);

        // Getting the reaction name-reaction pairs
        Map<String, ChemRxn> rxns = file.getMap();
        
        // Creating the fluxome
        IrrevRxnFluxome fluxome = new IrrevRxnFluxomeV1(rxns);

        return fluxome;
    }

    /**
	 * Returns the fluxome contained in the file with the provided name.
	 * 
	 * @param fileName
	 *            the name of a file containing a fluxome parsable with
	 *            {@linkplain MixedFluxomeFormatV1}.
	 * @return the fluxome contained in the file with the provided name.
	 */
    public static MixedFluxome getMixedFluxome(String fileName) {

        // Creating formatter
        MixedFluxomeFormatV1 format = new MixedFluxomeFormatV1();
        
        // Creating the file
        RomiFile<String, ChemRxn> file = 
            new RomiFile<String, ChemRxn>(fileName, format);

        // Getting the reaction name-reaction pairs
        Map<String, ChemRxn> rxns = file.getMap();
        
        // Creating the fluxome
        MixedFluxome fluxome = new MixedFluxomeV1(rxns);

        return fluxome;
    }
    
    /**
	 * Returns the fluxome contained in the files with the provided names.
	 * 
	 * @param rxnFileName
	 *            the name of a file containing a fluxome parsable with
	 *            {@linkplain IrfFormatV1}.
	 * 
	 * @param catalystFileName
	 *            the name of a file containing reaction-catalyst associations
	 *            parsable with {@linkplain BoolRelFormatV1}
	 * 
	 * @return the "catalyzed fluxome" contained in the files with the provided
	 *         names.
	 */
    public static IrrevRxnFluxome 
    getIrrevRxnFluxome(String rxnFileName,
    				   String catalystFileName) {
        
    	return getIrrevRxnFluxome(rxnFileName, catalystFileName, DEV_NULL);
    }
    
    /**
	 * Returns the fluxome contained in the files with the provided names.
	 * 
	 * @param rxnFileName
	 *            the name of a file containing a fluxome parsable with
	 *            {@linkplain MixedFluxomeFormatV1}.
	 * 
	 * @param catalystFileName
	 *            the name of a file containing reaction-catalyst associations
	 *            parsable with {@linkplain BoolRelFormatV1}
	 * 
	 * @return the "catalyzed fluxome" contained in the files with the provided
	 *         names.
	 */
    public static MixedFluxome 
    getMixedFluxome(String rxnFileName,
    				String catalystFileName) {
    	
    	return getMixedFluxome(rxnFileName, catalystFileName, DEV_NULL);
    }
    
    /**
	 * Returns the fluxome contained in the files with the provided names.
	 * 
	 * @param rxnFileName
	 *            the name of a file containing a fluxome parsable with
	 *            {@linkplain IrfFormatV1}.
	 * 
	 * @param catalystFileName
	 *            the name of a file containing reaction-catalyst associations
	 *            parsable with {@linkplain BoolRelFormatV1}
	 * 
	 * @param out
	 *            the print writer to which descriptive message will be written.
	 * 
	 * @return the "catalyzed fluxome" contained in the files with the provided
	 *         names.
	 */
    public static IrrevRxnFluxome 
    getIrrevRxnFluxome(String rxnFileName,
    				   String catalystFileName,
    				   PrintWriter out) {

        // Creating a basic fluxome
        IrrevRxnFluxome f = getIrrevRxnFluxome(rxnFileName, out);

        // Getting the reaction-catalyst associations
        Map<String, Or<And<String>>> catalysts = 
            getCatalysts(f, catalystFileName, out);
  
        // Returning a new fluxome containing the catalysts
        return new IrrevRxnFluxomeV1(f.getRxns(), catalysts);
    }
    
    /**
	 * Returns the fluxome contained in the files with the provided names.
	 * 
	 * @param rxnFileName
	 *            the name of a file containing a fluxome parsable with
	 *            {@linkplain MixedFluxomeFormatV1}.
	 * 
	 * @param catalystFileName
	 *            the name of a file containing reaction-catalyst associations
	 *            parsable with {@linkplain BoolRelFormatV1}
	 * 
	 * @param out
	 *            the print writer to which descriptive message will be written.
	 * 
	 * @return the "catalyzed fluxome" contained in the files with the provided
	 *         names.
	 */
    public static MixedFluxome 
    getMixedFluxome(String rxnFileName,
    				String catalystFileName, 
    				PrintWriter out) {

		// Creating a basic fluxome
		MixedFluxome f = getMixedFluxome(rxnFileName, out);

		// Getting the reaction-catalyst associations
		Map<String, Or<And<String>>> catalysts = 
			getCatalysts(f, catalystFileName, out);

		// Returning a new fluxome containing the catalysts
		return new MixedFluxomeV1(f.getRxns(), catalysts);
	}
    
    /**
	 * Returns the reaction-catalyst associations contained in the file with the
	 * provided name.
	 * 
	 * @param fluxome
	 *            the fluxome with which reaction names are verified.
	 * @param fileName
	 *            the name of the catalyst file.
	 * 
	 * @return reaction names mapped to the boolean logic statements governing
	 *         their catalysis.
	 */
    public static Map<String, Or<And<String >>> 
    getCatalysts(Fluxome fluxome, 
    			 String fileName) {
        
        // Key parser (only checks if a rxn name is valid)
        RxnNameVerifier rxnFormat = new RxnNameVerifier(fluxome);
        
        CatalystFormat catFormat = new CatalystFormatV1();
        
        // Value parser
        BoolRelFormatV1<String> boolFormat = 
            new BoolRelFormatV1<String>(catFormat);

        // Map parser
        MapFormat<String, Or<And<String>>> mapFormat =
            new EqualsMapFormatV1<String, Or<And<String>>>(rxnFormat, boolFormat);
        
        RomiFile<String, Or<And<String>>> file =
            new RomiFile<String, Or<And<String>>>(fileName, mapFormat);
        
        Map<String, Or<And<String>>> catalysts = file.getMap();
        
        // Getting the names of all internal reactions
        Set<String> internalRxnNames = fluxome.getInternalRxnNames();
        
        // Ensuring all reaction names are for internal reactions
        for (String rxnName : catalysts.keySet()) {
            
            if (!internalRxnNames.contains(rxnName)) {
                
                String s = "The reaction name '" + rxnName +
                           "' in the file '" + fileName +
                           "' refers to an exchange reaction and not an " +
                           "internal reaction. Catalysts can only be specified " +
                           "for internal reactions.";
                throw new FormatException(s);
            }
            
        }
        
        file.close();
        
        return catalysts;
    }

    /**
     * Returns the reaction-catalyst associations contained in the file with the
     * provided name.
     * 
     * @param fluxome the fluxome with which reaction names are verified.
     * @param fileName the name of the catalyst file.
     * @param out the print writer to which descriptive messages will be
     * written.
     * 
     * @return reaction names mapped to the boolean logic statements governing
     * their catalysis.
     */
    public static Map<String, Or<And<String>>> 
        getCatalysts(Fluxome fluxome, String fileName, PrintWriter out) {
        
        out.print("Reading '" + fileName + "' ... ");
        out.flush();
        
        Map<String, Or<And<String>>> m = getCatalysts(fluxome, fileName);
        
        out.println("DONE");
        
        return m;
    }
    
    /**
     * Writes the provided fluxome to a file with the provided name. The
     * output file format is {@linkplain IrfFormatV1}.
     * 
     * @param fluxome the fluxome to be written to file.
     * @param fileName the name of the file to which the fluxome will be written.
     */
    public static void writeFluxome(IrrevRxnFluxome fluxome, String fileName) {
        
        // Creating formatter
        IrfFormatV1 format = new IrfFormatV1();
        
        // Creating the file
        RomoFile<String, ChemRxn> file = 
            new RomoFile<String, ChemRxn>(fileName, format);
       
        // Writing the reactions from the fluxome
        file.addData(fluxome.getRxns());

        // Closing the file
        file.close();

    }
    
    /**
     * Writes the provided fluxome to a file with the provided name. The
     * output file format is {@linkplain IrfFormatV1}.
     * 
     * @param fluxome the fluxome to be written to file.
     * @param fileName the name of the file to which the fluxome will be written.
     * @param out the print writer to which descriptive messages will be written.
     */
    public static void writeFluxome(IrrevRxnFluxome fluxome, 
    								String fileName,
    								PrintWriter out) {

		// Printing an opening message
		out.print("Writing '" + fileName + "' ... ");
		out.flush();
        
		writeFluxome(fluxome, fileName);

		// Printing a closing message
		out.println("DONE");
    }
    
    /**
     * Writes the provided fluxome to a file with the provided name. The
     * output file format is {@linkplain MixedFluxomeFormatV1}.
     * 
     * @param fluxome the fluxome to be written to file.
     * @param fileName the name of the file to which the fluxome will be written.
     */
    public static void writeFluxome(MixedFluxome fluxome, String fileName) {
        
        // Creating formatter
    	MixedFluxomeFormatV1 format = new MixedFluxomeFormatV1();
        
        // Creating the file
        RomoFile<String, ChemRxn> file = 
            new RomoFile<String, ChemRxn>(fileName, format);
       
        // Writing the reactions from the fluxome
        file.addData(fluxome.getRxns());

        // Closing the file
        file.close();

    }
    
    /**
     * Writes the provided fluxome to a file with the provided name. The
     * output file format is {@linkplain MixedFluxomeFormatV1}.
     * 
     * @param fluxome the fluxome to be written to file.
     * @param fileName the name of the file to which the fluxome will be written.
     * @param out the print writer to which descriptive messages will be written.
     */
    public static void writeFluxome(MixedFluxome fluxome, 
    								String fileName,
    								PrintWriter out) {

		// Printing an opening message
		out.print("Writing '" + fileName + "' ... ");
		out.flush();
        
		writeFluxome(fluxome, fileName);

		// Printing a closing message
		out.println("DONE");
    }
    
    /**
     * Opens the file denoted by {@linkplain #SYSTEM_SOLUTION_FILE} in the 
     * provided manager input and returns the linear system solution 
     * it contains.
     * 
     * @param input keyword-value pairs.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the solution corresponds.
     * @param coeffCutoff the smallest value that the absolute value of a coefficient
     * can attain.
     * 
     * @return the linear system solution contained in the file indicated by 
     * {@linkplain #SYSTEM_SOLUTION_FILE} in the provided input.
     * 
     *  @see LinearComb#getZeroCutoff()
     */
	public static LinearSystemSolution<String> 
	getFluxomeSolution(ManagerInput<String, String> input, 
					   PrintWriter out, 
					   Fluxome fluxome,
					   double coeffCutoff) {

		// Getting the filename
		String sysSolutionFileName = 
			input.getRequiredValue(SYSTEM_SOLUTION_FILE);

		// Printing an opening message
		out.print("Reading '" + sysSolutionFileName + "' ... ");
		out.flush();

		// Getting all reaction names
		Set<String> required = fluxome.getRxnNames();

		// Creating the file line format
		FluxomeSolutionFileLineFormat format = 
			new FluxomeSolFileLineFormatV1(fluxome, coeffCutoff);

		// Creating the file
		RomiFile<String, LinearComb<String>> solFile = 
			new RomiFile<String, LinearComb<String>>(
				sysSolutionFileName, FileFormat.TEXT, required, format);

		// Getting the solution vector
		Map<String, LinearComb<String>> vector = solFile.getMap();

		// Creating the linear system solution
		HashLinearSystemSolution<String> solution = 
			new HashLinearSystemSolution<String>(vector);

		// Closing the file
		solFile.close();

		// Printing a closing message
		out.println("DONE");

		// Returning the fluxome solution
		return solution;
	}

}
