/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.FluxConstraints;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.comp.FbaOptException;
import sbrt.kernel.fba.comp.FbaOptimizer;
import sbrt.kernel.fba.proc.FbaOptProcess;
import sbrt.kernel.fba.proc.FbaProcess;
import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.math.ConstraintViolationException;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.MathExpr;
import sbrt.kernel.math.optimization.OptSense;
import sbrt.kernel.math.optimization.ProgramSolver;
import sbrt.kernel.math.optimization.ProgramSolverException;
import sbrt.kernel.math.optimization.ProgramSolverStatus;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.kernel.utilities.Util;
import sbrt.shell.ApplicationException;
import sbrt.shell.ProgramSolverFactory;
import sbrt.shell.io.FileFormat;
import sbrt.shell.io.MultiOutputFiles;
import sbrt.shell.io.RomiFile;
import sbrt.shell.io.SerializedOutputFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.io.fba.CofoofSource;
import sbrt.shell.io.fba.ColumnOrFbaOptOutputFile;
import sbrt.shell.io.fba.FbaOptOutputFile;
import sbrt.shell.io.fba.FluxIntervalInputFile;
import sbrt.shell.io.fba.RofoofSource;
import sbrt.shell.io.fba.RowOrFbaOptOutputFile;
import sbrt.shell.io.fba.SerFbaOptOutputFile;
import sbrt.shell.io.fba.SfoofSource;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.fba.ConstraintsFileLineFormat;
import sbrt.shell.text.fba.ConstraintsFileLineFormatV1;
import sbrt.shell.text.fba.ConstraintsFileLineFormatV2;
import sbrt.shell.text.fba.FbaOptFileLineFormat;
import sbrt.shell.text.fba.FbaOptFileLineFormatV1;
import sbrt.shell.text.fba.FbaOptHeaderParser;
import sbrt.shell.text.fba.FbaOptHeaderParserV1;
import sbrt.shell.text.fba.RxnNameExprFormat;
import sbrt.shell.text.fba.RxnNameOrExprFormat;
import sbrt.shell.text.fba.RxnNameOrExprFormatV1;
import sbrt.shell.text.math.SenseFormat;

/**
 * This class is used to manage basic FBA optimization processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class FbaOptManager extends FbaManager {

    /**
	 * The default safety level used by
	 * {@linkplain #getSafetyLevel(ManagerInput, PrintWriter)}.
	 */
    public static final int DEFAULT_SAFETY_LEVEL = 1;

    /**
     * The keyword used to denote an objective function.
     */
    public static final String OBJECTIVE_FUNCTION = "Objective Function";

    /**
     * The keyword used to denote a constraints file.
     */
    public static final String CONSTRAINTS_FILE = "Constraints File";

    /**
     * The keyword used to denote the optimization sense, or direction.
     */
    public static final String OPTIMIZATION_SENSE = "Optimization Sense";

    /**
     * The keyword used to denote the safety level.
     */
    public static final String SAFETY_LEVEL = "Safety Level";

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        s.add(RXN_FILE);
        s.add(CONSTRAINTS_FILE);
        s.add(ProgramSolverFactory.SOLVER);
        s.add(OBJECTIVE_FUNCTION);
        s.add(OPTIMIZATION_SENSE);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {

        Set<String> s = new TreeSet<String>();

        s.addAll(FbaManager.optionalKeywords);
        s.add(SAFETY_LEVEL);
        s.add(CONSTRAINT_TOLERANCE);
        s.add(OUTPUT_FILE_NAME);
        s.add(DATA_HEADERS);

        s.add(ProgramSolverFactory.PARAM_FILE);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Parses and returns the value for {@linkplain #OBJECTIVE_FUNCTION} in the
     * provided manager input.
     * 
     * @param input keyword-value pairs.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the objective function corresponds.
     * 
     * @return the parsed value for {@linkplain #OBJECTIVE_FUNCTION} in the provided
     * input.
     */
    public static LinearComb<String> 
    getObjFunc(ManagerInput<String, String> input,
    					 PrintWriter out,
    					 Fluxome fluxome) {

        // Getting the objective function string
        String objFunction = input.getRequiredValue(OBJECTIVE_FUNCTION);

        // Getting an objective function parser
        RxnNameExprFormat<LinearComb<String>, String> format = 
        	getRxnNameExprFormat(fluxome);

        // Parsing the objective function
        LinearComb<String> obj = format.parse(objFunction);

        // Returning the parsed objective function
        return obj;
    }

    /**
	 * Creates and returns a fluxome based on the provided input. The values for
	 * {@linkplain #RXN_FILE}, {@linkplain #CONSTRAINTS_FILE}, and
	 * {@linkplain FbaManager#CONSTRAINT_TOLERANCE} are parsed and combined to
	 * create the returned fluxome.
	 * 
	 * @param input keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * 
	 * @return a fluxome based on the provided input.
	 * 
	 * @see FbaManager#getIrrevRxnFluxome(ManagerInput, PrintWriter)
	 * @see FbaManager#getConstraintTolerance(ManagerInput, PrintWriter)
	 * @see #getConstraints(ManagerInput, PrintWriter, Fluxome, double)
	 */
    public static IrrevRxnFluxome 
    getIrrevRxnFluxome(ManagerInput<String, String> input,
            		   PrintWriter out) {

        // Getting basic fluxome (i.e. one with no constraints)
        IrrevRxnFluxome fluxome = FbaManager.getIrrevRxnFluxome(input, out);

        // Getting constraint tolerance
        double tolerance = getConstraintTolerance(input, out);
        
        // Getting constraints
        FluxConstraints constraints = getConstraints(input, out, fluxome, tolerance);

        // Returning new fluxome
        return fluxome.replaceConstraints(constraints, tolerance);
    }

    /**
     * Creates and returns a fluxome based on the provided input. The values for
     * {@linkplain #RXN_FILE} and {@linkplain #CONSTRAINTS_FILE} are parsed and
     * combined to create the returned fluxome.
     * 
     * @param input keyword-value pairs.
     * @param out the print writer to which status messages will be printed.
     * @param tolerance the constraint tolerance.
     * 
     * @return a fluxome based on the provided input.
     * 
     * @see FbaManager#getIrrevRxnFluxome(ManagerInput, PrintWriter)
     * @see #getConstraints(ManagerInput, PrintWriter, Fluxome, double)
     * @see Fluxome#replaceConstraints(FluxConstraints, double)
     */
    public static IrrevRxnFluxome 
    getIrrevRxnFluxome(ManagerInput<String, String> input,
            		   PrintWriter out,
            		   double tolerance) {

        // Getting basic fluxome (i.e. one with no constraints)
        IrrevRxnFluxome fluxome = FbaManager.getIrrevRxnFluxome(input, out);

        // Getting constraints
        FluxConstraints constraints = getConstraints(input, out, fluxome, tolerance);

        // Returning new fluxome
        return fluxome.replaceConstraints(constraints, tolerance);
    }

    /**
	 * Creates and returns a fluxome based on the provided input. The values for
	 * {@linkplain #RXN_FILE}, {@linkplain #CONSTRAINTS_FILE}, and
	 * {@linkplain FbaManager#CONSTRAINT_TOLERANCE} are parsed and combined to
	 * create the returned fluxome.
	 * 
	 * @param input keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * 
	 * @return a fluxome based on the provided input.
	 * 
	 * @see FbaManager#getMixedFluxome(ManagerInput, PrintWriter)
	 * @see FbaManager#getConstraintTolerance(ManagerInput, PrintWriter)
	 * @see #getConstraints(ManagerInput, PrintWriter, Fluxome, double)
	 */
    public static MixedFluxome 
    getMixedFluxome(ManagerInput<String, String> input,
            		PrintWriter out) {

        // Getting basic fluxome (i.e. one with no constraints)
        MixedFluxome fluxome = FbaManager.getMixedFluxome(input, out);

        // Getting constraint tolerance
        double tolerance = getConstraintTolerance(input, out);

        // Getting constraints
        FluxConstraints constraints = getConstraints(input, out, fluxome, tolerance);

        // Returning new fluxome
        return fluxome.replaceConstraints(constraints, tolerance);
    }

    /**
     * Creates and returns a fluxome based on the provided input. The values for
     * {@linkplain #RXN_FILE} and {@linkplain #CONSTRAINTS_FILE} are parsed and
     * combined to create the returned fluxome.
     * 
     * @param input keyword-value pairs.
     * @param out the print writer to which status messages will be printed.
     * @param tolerance the constraint tolerance.
     * 
     * @return a fluxome based on the provided input.
     * 
     * @see FbaManager#getMixedFluxome(ManagerInput, PrintWriter)
     * @see #getConstraints(ManagerInput, PrintWriter, Fluxome, double)
     * @see Fluxome#replaceConstraints(FluxConstraints, double)
     */
    public static MixedFluxome 
    getMixedFluxome(ManagerInput<String, String> input,
    				PrintWriter out,
    				double tolerance) {

        // Getting basic fluxome (i.e. one with no constraints)
    	MixedFluxome fluxome = FbaManager.getMixedFluxome(input, out);

        // Getting constraints
        FluxConstraints constraints = getConstraints(input, out, fluxome, tolerance);

        // Returning new fluxome
        return fluxome.replaceConstraints(constraints, tolerance);
    }

    /**
     * Opens the file denoted by {@linkplain #CONSTRAINTS_FILE} in the provided
     * manager input and returns the set of flux constraints it contains.
     * 
     * @param input keyword-value pairs.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the constraints correspond.
     * @param tolerance the constraint tolerance.
     * 
     * @return the set of flux constraints contained in the file indicated by
     * {@code CONSTRAINTS_FILE} in the provided input; or, if the input does
     * not contain a value for {@code CONSTRAINTS_FILE}, an empty set of flux
     * constraints.
     * 
     * @see #getConstraints(String, Fluxome, double)
     */
    public static FluxConstraints 
    getConstraints(ManagerInput<String, String> input,
    			   PrintWriter out,
    			   Fluxome fluxome,
    			   double tolerance) {

        // Getting the name of the constraints file if it exists
        String fileName = input.getOptionalValue(CONSTRAINTS_FILE);

        // Returning empty constraints if no input value was provided
        if (fileName == null) {

            Map<String, Interval> fluxC = Collections.emptyMap();
            Map<MathExpr<String>, Interval> exprC = Collections.emptyMap();

            return new FluxConstraints(fluxC, exprC);
        }

        // Printing an opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Converting to constraint set
        FluxConstraints constraints = 
        	getConstraints(fileName, fluxome, tolerance);

        // Printing a closing message
        out.println("DONE");

        // Returning the flux constraints contained in the file
        return constraints;
    }

    /**
     * Returns the set of constraints contained in the file with the provided
     * name. This file must have the format of a row-oriented map file, and the
     * lines must be parsable using {@linkplain ConstraintsFileLineFormatV2}.
     * 
     * @param fileName the name of the file to be read.
     * @param fluxome the fluxome to which the constraints correspond.
     * @param tolerance the constraint tolerance.
     * 
     * @return the set of constraints contained in the file with the provided
     * name.
     * 
     * @throws NullPointerException if any argument is {@code null}.
     * 
     * @see RomiFile
     */
    public static FluxConstraints 
    getConstraints(String fileName, Fluxome fluxome, double tolerance) {

        // Getting the constraints file line format
        ConstraintsFileLineFormat<LinearComb<String>> format =
                new ConstraintsFileLineFormatV2(fluxome);

        // Getting the row-oriented constraints file
        RomiFile<LinearComb<String>, Interval> file =
                new RomiFile<LinearComb<String>, Interval>(fileName,
                        FileFormat.TEXT, format);

        // Getting the flux constraints vector
        Map<LinearComb<String>, Interval> vector = file.getMap();

        // Closing the file
        file.close();

    	// Converting to constraint set
    	FluxConstraints constraints = null;
    	
        try {
        
        	constraints =
                FluxIntervalInputFile.convertConstraints(vector, 
                										 fluxome, 
                										 tolerance);
        
        } catch (ApplicationException e) {
        	
        	String s = "An error occurred while parsing the file '" +
        			   file.getCanonicalName() + "'. " +
        			   e.getMessage();
        	throw new ApplicationException(s);
        	
        }
        
        
        // Returning the flux constraints contained in the file
        return constraints;
    }

    /**
	 * Creates and returns an FBA optimization computer based on the provided
	 * input. The values for {@linkplain #RXN_FILE},
	 * {@linkplain #CONSTRAINTS_FILE},
	 * {@linkplain FbaManager#CONSTRAINT_TOLERANCE}, {@linkplain #SAFETY_LEVEL},
	 * {@linkplain ProgramSolverFactory#SOLVER}, and
	 * {@linkplain ProgramSolverFactory#PARAM_FILE} are parsed and combined to
	 * create the returned optimization computer.
	 * 
	 * @param input
	 *            keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * 
	 * @return an FBA optimization computer based on the provided input.
	 * 
	 * @see #getIrrevRxnFluxome(ManagerInput, PrintWriter)
	 * @see #getSafetyLevel(ManagerInput, PrintWriter)
	 * @see #getConstraintTolerance(ManagerInput, PrintWriter)
	 * @see ProgramSolverFactory#make(ManagerInput, PrintWriter)
	 */
    public static FbaOptimizer<IrrevRxnFluxome> 
    getIrrevRxnComp(ManagerInput<String, String> input,
    				PrintWriter out) {

        // Getting the optimization computer safety level
        int safetyLevel = getSafetyLevel(input, out);

        // Getting the constraint tolerance
        double tolerance = getConstraintTolerance(input, out);

        // Getting the fluxome
        IrrevRxnFluxome fluxome = getIrrevRxnFluxome(input, out, tolerance);

        out.println();
        
        // Getting the program solver
        ProgramSolver<Object, Object> ps =
                ProgramSolverFactory.make(input, out);

        // Creating the optimization computer
        FbaOptimizer<IrrevRxnFluxome> c = 
        	new FbaOptimizer<IrrevRxnFluxome>(fluxome, ps);
        
        c.setSafetyLevel(safetyLevel);
        c.setTolerance(tolerance);

        // Returning the computer
        return c;
    }

    /**
	 * Creates and returns an FBA optimization computer based on the provided
	 * input. The values for {@linkplain #RXN_FILE},
	 * {@linkplain #CONSTRAINTS_FILE},
	 * {@linkplain FbaManager#CONSTRAINT_TOLERANCE}, {@linkplain #SAFETY_LEVEL},
	 * {@linkplain ProgramSolverFactory#SOLVER}, and
	 * {@linkplain ProgramSolverFactory#PARAM_FILE} are parsed and combined to
	 * create the returned optimization computer.
	 * 
	 * @param input
	 *            keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * 
	 * @return an FBA optimization computer based on the provided input.
	 * 
	 * @see #getMixedFluxome(ManagerInput, PrintWriter)
	 * @see #getSafetyLevel(ManagerInput, PrintWriter)
	 * @see #getConstraintTolerance(ManagerInput, PrintWriter)
	 * @see ProgramSolverFactory#make(ManagerInput, PrintWriter)
	 */
    public static FbaOptimizer<MixedFluxome> 
    getMixedComp(ManagerInput<String, String> input,
    			 PrintWriter out) {

        // Getting the optimization computer safety level
        int safetyLevel = getSafetyLevel(input, out);

        // Getting the constraint tolerance
        double tolerance = getConstraintTolerance(input, out);

        // Getting the fluxome
        MixedFluxome fluxome = getMixedFluxome(input, out, tolerance);

        out.println();
        
        // Getting the program solver
        ProgramSolver<Object, Object> ps =
                ProgramSolverFactory.make(input, out);

        // Creating the optimization computer
        FbaOptimizer<MixedFluxome> c =         	
        	new FbaOptimizer<MixedFluxome>(fluxome, ps);
        
        c.setSafetyLevel(safetyLevel);
        c.setTolerance(tolerance);

        // Returning the computer
        return c;
    }
    
    /**
	 * Parses and returns the value for {@linkplain #SAFETY_LEVEL} in the
	 * provided manager input.
	 * 
	 * @param input
	 *            keyword-value pairs.
	 * @param out
	 *            the print writer to which a status message will be printed.
	 * 
	 * @return the parsed value for {@linkplain #SAFETY_LEVEL} in the provided
	 *         input, or {@linkplain #DEFAULT_SAFETY_LEVEL} if the provided
	 *         input does not contain such a value.
	 */
    public static int 
    getSafetyLevel(ManagerInput<String, String> input,
    			   PrintWriter out) {

        // Getting the provided safety level string
        String s = input.getOptionalValue(SAFETY_LEVEL);

        // If no input value was provided, return the default value
        if (s == null) {
            return DEFAULT_SAFETY_LEVEL;
        }

        // Parsing the provided value
        int n = parseInt(s);

        // Printing a message
        out.println("Parsed " + SAFETY_LEVEL + ": " + n);

        // Returning the parsed safety level
        return n;
    }

    /**
     * Parses and returns the value for {@linkplain #OPTIMIZATION_SENSE} in the
     * provided manager input.
     * 
     * @param input keyword-value pairs.
     * 
     * @return the parsed value for {@linkplain #OPTIMIZATION_SENSE} in the provided
     * manager input.
     */
    public static OptSense getOptSense(ManagerInput<String, String> input) {

        // Getting the provided optimization sense string
        String value = input.getRequiredValue(OPTIMIZATION_SENSE);

        // Parsing it
        return new SenseFormat().parse(value);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e a runtime exception thrown during an FBA optimization.
     * @param process the FBA process from which the exception was thrown.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     * 
     * @throws RuntimeException the original exception if it cannot be
     * intrepretted.
     */
    public static ApplicationException 
    interpret(RuntimeException e, FbaProcess<?, ?> process) {

        if (e instanceof ProgramSolverException) {
            return interpret((ProgramSolverException) e);
        }

        if (e instanceof ConstraintViolationException) {
            return interpret((ConstraintViolationException) e, process);
        }

        if (e instanceof FbaOptException) {
            return interpret((FbaOptException) e, process);
        }

        throw e;
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e a program solver exception.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(ProgramSolverException e) {
        return ProgramSolverFactory.interpret(e);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e a constraint violation exception.
     * @param process the FBA process from which the exception was thrown.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(ConstraintViolationException e,
            FbaProcess<?, ?> process) {

        Object var = e.getVariable();
        Interval bounds = e.getBounds();
        Number value = e.getValue();

        RxnNameOrExprFormat varFormat =
                new RxnNameOrExprFormatV1(process.getFluxome());

        ConstraintsFileLineFormat<String> boundFormat =
                new ConstraintsFileLineFormatV1(process.getFluxome());

        DoubleFormat numberFormat = new DoubleFormatV1();

        String s =
                "The program solver has violated the constraints for '"
                        + varFormat.format(var)
                        + "'. The computed value '"
                        + numberFormat.format(value.doubleValue())
                        + "' does not satisfy the constraint '"
                        + boundFormat.formatValue(bounds)
                        + "'.";

        return new ApplicationException(s, e);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e an FBA optimization exception.
     * @param process the FBA process from which the exception was thrown.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException 
    interpret(FbaOptException e, FbaProcess<?, ?> process) {

        // Gathering information from the FBA optimization exception
        MathExpr<?> obj = e.getObjFunction();
        OptSense sense = e.getSense();
        ProgramSolverStatus status = e.getStatus();

        StringBuilder s = new StringBuilder(1000);

        s.append("The program solver was unable to find an optimal "
                + "solution for the objective '"
                + sense
                + ": "
                + obj
                + "'. The status was reported to be '"
                + status
                + "'.");

        if (status == ProgramSolverStatus.UNBOUNDED) {
            s.append(" This can be caused by inappropriately constrained "
                            + "exchange reactions or by the presence of cycles.");
        }

        else if (status == ProgramSolverStatus.INFEASIBLE) {
            s.append(" This can be caused by conflicting flux constraints "
                    + "or by inappropriate parameter settings for the "
                    + "program solver.");
        }

        else if (status != ProgramSolverStatus.UNKNOWN) {
            s.append(" The documentation of the linear program solver " +
            "should contain information regarding this type of status.");
        }
        
        return new ApplicationException(s.toString(), e);
    }

    /**
     * Parses and returns the value for {@linkplain #DATA_HEADERS} in the provided
     * manager input. This method uses {@linkplain FbaOptHeaderParserV1} to perform
     * the parsing.
     * 
     * @param input keyword-value pairs.
     * @param fluxome the fluxome to which the data headers correspond.
     * 
     * @return the parsed value for {@linkplain #DATA_HEADERS} in the provided
     * manager input; or the set of all possible data headers if the input
     * contains no such value.
     */
    public static Set<String> 
    getDataHeaders(ManagerInput<String, String> input,
            	   Fluxome fluxome) {

        // Getting header string value
        String headers = input.getOptionalValue(DATA_HEADERS);

        // Returning parsed reaction name headers
        return getDataHeaders(headers, fluxome);
    }

    /**
     * Parses the provided data header string and returns the corresponding set.
     * This method uses {@linkplain FbaOptHeaderParserV1} to perform the parsing.
     * 
     * @param dataHeaders a string containing data headers.
     * @param fluxome the fluxome to which the data headers correspond.
     * 
     * @return the set of data headers parsed from the provided string; or the
     * set of all possible data headers if the provided string is {@code null}.
     */
    public static LinkedHashSet<String> 
    getDataHeaders(String dataHeaders, Fluxome fluxome) {

        // Getting a reaction name header parser
        FbaOptHeaderParser headerParser = new FbaOptHeaderParserV1(fluxome);

        // If no input value was provided
        if (dataHeaders == null) {
            dataHeaders = "Objective_Value | All_Reactions";
        }

        // Returning parsed reaction name headers
        return headerParser.parse(dataHeaders);
    }

    /**
	 * Returns an empty output file that is suitable for storing multiple FBA
	 * optimization results. The values for {@linkplain #OUTPUT_FILE_NAME},
	 * {@linkplain #DATA_HEADERS}, and {@linkplain #FILE_FORMAT} in the
	 * provided input are parsed and used to call the method
	 * {@linkplain #getMultiResultsOutputFile(String, FileFormat, Set, Fluxome)}.
     * 
     * @param <F> the fluxome type.
	 * 
	 * @param input
	 *            the input containing values for the listed keywords.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fluxome
	 *            the fluxome to which the results correspond.
	 * 
	 * @return an empty output file that is suitable for storing multiple FBA
	 *         optimization results.
	 * 
	 * @see #getDataHeaders(ManagerInput, Fluxome)
	 * @see #getFileFormat(ManagerInput)
	 */
    public static <F extends Fluxome> FbaOptOutputFile<F> 
    getMultiResultsOutputFile(ManagerInput<String, String> input,
    						  PrintWriter out,
    						  F fluxome) {

        // Getting output file name
        String name = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing an opening message
        out.print("Opening '" + name + "' ... ");
        out.flush();

        // Getting data headers
        Set<String> headers = getDataHeaders(input, fluxome);

        // Getting file format
        FileFormat format = getFileFormat(input);

        // Getting output file
        FbaOptOutputFile<F> file =
                getMultiResultsOutputFile(name, format, headers, fluxome);

        // Printing a closing message
        out.println("DONE");

        // Returning the output file
        return file;
    }

    /**
     * Returns an empty output file that is suitable for storing multiple FBA
     * optimization results. This method uses {@linkplain FbaOptFileLineFormatV1}.
     * 
     * @param <F> the fluxome type. 
     * 
     * @param fileName the name of the file to be created.
     * @param fileFormat the format of the file to be created.
     * @param dataHeaders the set of data headers to be written to the file.
     * @param fluxome the fluxome to which the results correspond.
     * 
     * @return an empty output file that is suitable for storing multiple FBA
     * optimization results.
     */
    public static <F extends Fluxome> FbaOptOutputFile<F> 
    getMultiResultsOutputFile(String fileName,
    						  FileFormat fileFormat,
    						  Set<String> dataHeaders,
    						  F fluxome) {

        // Getting file line format
        FbaOptFileLineFormat fbaFormat = new FbaOptFileLineFormatV1(fluxome);

        // If this is a serialized file format
        if (SerializedOutputFile.isValidFormat(fileFormat)) {

            // Creating serialized vector output file
            return new SerFbaOptOutputFile<F>(fileName, 
            								  fileFormat, 
            								  dataHeaders,
            								  fbaFormat);
        }

        // If this is a text file format
        else if (TextOutputFile.isValidFormat(fileFormat)) {

            // Creating a column-oriented text output file
            return new ColumnOrFbaOptOutputFile<F>(fileName, 
            									   fileFormat,
            									   dataHeaders, 
            									   fbaFormat);
        }

        // If it's neither
        else {
            String s =
                    "The file format '"
                            + fileFormat
                            + "' is not yet supported.";
            throw new IllegalArgumentException(s);
        }

    }

    /**
	 * Returns an empty output file that is suitable for storing a single FBA
	 * optimization result. The values for {@linkplain #OUTPUT_FILE_NAME},
	 * {@linkplain #DATA_HEADERS}, and {@linkplain #FILE_FORMAT} in the
	 * provided input are parsed and used to call the method
	 * {@linkplain #getSingleResultOutputFile(String, FileFormat, Set, Fluxome)}.
	 * 
	 * @param <F>
	 *            the fluxome type.
	 * 
	 * @param input
	 *            the input containing values for the listed keywords.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fluxome
	 *            the fluxome to which the results correspond.
	 * 
	 * @return an empty output file that is suitable for storing a single FBA
	 *         optimization result.
	 * 
	 * @see #getDataHeaders(ManagerInput, Fluxome)
	 * @see #getFileFormat(ManagerInput)
	 */
    public static <F extends Fluxome> FbaOptOutputFile<F> 
    getSingleResultOutputFile(ManagerInput<String, String> input,
    						  PrintWriter out,
    						  F fluxome) {

        // Getting output file name
        String name = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing an opening message
        out.print("Opening '" + name + "' ... ");
        out.flush();

        // Getting data headers
        Set<String> headers = getDataHeaders(input, fluxome);

        // Getting file format
        FileFormat format = getFileFormat(input);

        // Getting output file
        FbaOptOutputFile<F> file =
                getSingleResultOutputFile(name, format, headers, fluxome);

        // Printing a closing message
        out.println("DONE");

        // Returning the output file
        return file;
    }

    /**
     * Returns an empty output file that is suitable for storing a single FBA
     * optimization result. This method uses {@linkplain FbaOptFileLineFormatV1}.
     * 
     * @param <F> the fluxome type. 
     * 
     * @param fileName the name of the file to be created.
     * @param fileFormat the format of the file to be created.
     * @param dataHeaders the set of data headers to be written to the file.
     * @param fluxome the fluxome to which the results correspond.
     * 
     * @return an empty output file that is suitable for storing a single FBA
     * optimization result.
     */
    public static <F extends Fluxome> FbaOptOutputFile<F> 
    getSingleResultOutputFile(String fileName,
    						  FileFormat fileFormat,
    						  Set<String> dataHeaders,
    						  F fluxome) {

        // Getting file line format
        FbaOptFileLineFormat fbaFormat = new FbaOptFileLineFormatV1(fluxome);

        // If this is a serialized file format
        if (SerializedOutputFile.isValidFormat(fileFormat)) {

            // Creating serialized vector output file
            return new SerFbaOptOutputFile<F>(fileName, 
            								  fileFormat, 
            								  dataHeaders,
            								  fbaFormat);
        }

        // If this is a text file format
        else if (TextOutputFile.isValidFormat(fileFormat)) {

            // Creating a column-oriented text output file
            return new RowOrFbaOptOutputFile<F>(fileName, 
            									fileFormat, 
            									dataHeaders,
            									fbaFormat);
        }

        // If it's neither
        else {
            String s =
                    "The file format '"
                            + fileFormat
                            + "' is not yet supported.";
            throw new IllegalArgumentException(s);
        }

    }

    /**
	 * Returns a source of FBA optimization output files. Each file
	 * is suitable for storing the results of multiple FBA optimizations. The
	 * value for {@linkplain #DATA_HEADERS} and {@linkplain #FILE_FORMAT} in the
	 * provided input are parsed and used.
	 * 
	 * @param <F>
	 *            the fluxome type.
	 * 
	 * @param input
	 *            keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fileNameSource
	 *            a source of desired output file names.
	 * @param fileNameSourceName
	 *            the name of the output file name source.
	 * @param fluxome
	 *            the fluxome for which the results correspond.
	 * 
	 * @return a source of FBA optimization output files.
     * 
     * @see #getMultiColOrOutputFiles(ManagerInput, PrintWriter, Iterator, String, Fluxome)
     * @see #getMultiSerOutputFiles(ManagerInput, PrintWriter, Iterator, String, Fluxome)
     */
    public static <F extends Fluxome>
    MultiOutputFiles<? extends FbaOptOutputFile<F>, 
    				 FbaOptimizer<F>.Results> 
    getMultiFileOutput(ManagerInput<String, String> input,
    				   PrintWriter out,
    				   Iterator<String> fileNameSource,
    				   String fileNameSourceName,
    				   F fluxome) {

        // Getting file format
        FileFormat fileFormat = getFileFormat(input);

        // If it's a text file format
        if (TextOutputFile.isValidFormat(fileFormat)) {
        	
            return getMultiColOrOutputFiles(input,
            								out, 
            								fileNameSource,
            								fileNameSourceName, 
            								fluxome);
        }

        // If it's a serialized file format
        else if (SerializedOutputFile.isValidFormat(fileFormat)) {
        	
            return getMultiSerOutputFiles(input, 
            							  out, 
            							  fileNameSource,
            							  fileNameSourceName, 
            							  fluxome);
        }

        // If it's neither
        else {
            String s = fileFormat + " not yet supported.";
            throw new UnsupportedOperationException(s);
        }

    }

    /**
	 * Returns a source of serialized FBA optimization output files. Each file
	 * is suitable for storing the results of multiple FBA optimizations. The
	 * value for {@linkplain #DATA_HEADERS} and {@linkplain #FILE_FORMAT} in the
	 * provided input are parsed and used.
	 * 
	 * @param <F>
	 *            the fluxome type.
	 * 
	 * @param input
	 *            keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fileNameSource
	 *            a source of desired output file names.
	 * @param fileNameSourceName
	 *            the name of the output file name source.
	 * @param fluxome
	 *            the fluxome for which the results correspond.
	 * 
	 * @return a source of serialized FBA optimization output files.
	 * 
	 * @see SerFbaOptOutputFile
	 * @see SfoofSource
	 */
    public static <F extends Fluxome> 
    MultiOutputFiles<SerFbaOptOutputFile<F>, FbaOptimizer<F>.Results> 
    getMultiSerOutputFiles(ManagerInput<String, String> input,
    					   PrintWriter out,
    					   Iterator<String> fileNameSource,
    					   String fileNameSourceName,
    					   F fluxome) {

        // Getting file format
        FileFormat fileFormat = getFileFormat(input);

        // Getting column header string
        String headerString = input.getRequiredValue(DATA_HEADERS);

        // Creating a header parser
        FbaOptHeaderParser setFormat = new FbaOptHeaderParserV1(fluxome);

        // Parsing the data headers
        Set<String> headers = setFormat.parse(headerString);

        // Creating file line format
        FbaOptFileLineFormat fbaFormat = new FbaOptFileLineFormatV1(fluxome);

        // Creating output file source
        Iterator<SerFbaOptOutputFile<F>> source =
                new SfoofSource<F>(fileNameSource, 
                				   fileNameSourceName, 
                				   fileFormat,
                				   headers, 
                				   fbaFormat);

        // Creating file
        return new MultiOutputFiles<SerFbaOptOutputFile<F>, 
        							FbaOptimizer<F>.Results>(source);
    }

    /**
	 * Returns a source of column-oriented FBA optimization output files. Each file
	 * is suitable for storing the results of multiple FBA optimizations. The
	 * value for {@linkplain #DATA_HEADERS} and {@linkplain #FILE_FORMAT} in the
	 * provided input are parsed and used.
	 * 
	 * @param <F>
	 *            the fluxome type.
	 * 
	 * @param input
	 *            keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fileNameSource
	 *            a source of desired output file names.
	 * @param fileNameSourceName
	 *            the name of the output file name source.
	 * @param fluxome
	 *            the fluxome for which the results correspond.
	 * 
	 * @return a source of source of serialized FBA optimization output files.
	 * 
	 * @see ColumnOrFbaOptOutputFile
	 * @see CofoofSource
	 */
    public static <F extends Fluxome>
    MultiOutputFiles<ColumnOrFbaOptOutputFile<F>, FbaOptimizer<F>.Results> 
    getMultiColOrOutputFiles(ManagerInput<String, String> input,
            				 PrintWriter out,
            				 Iterator<String> fileNameSource,
            				 String fileNameSourceName,
            				 F fluxome) {

        // Getting file format
        FileFormat fileFormat = getFileFormat(input);

        // Getting column header string
        String columnHeaders = input.getRequiredValue(DATA_HEADERS);

        // Creating a header parser
        FbaOptHeaderParser setFormat = new FbaOptHeaderParserV1(fluxome);

        // Parsing the data headers
        Set<String> headers = setFormat.parse(columnHeaders);

        // Getting file line format
        FbaOptFileLineFormat fbaFormat = new FbaOptFileLineFormatV1(fluxome);

        // Creating output file source
        Iterator<ColumnOrFbaOptOutputFile<F>> source =
        	new CofoofSource<F>(fileNameSource, 
        						fileNameSourceName,
        						fileFormat, 
        						headers, 
        						fbaFormat);

        // Creating file
        return new MultiOutputFiles<ColumnOrFbaOptOutputFile<F>,
        							FbaOptimizer<F>.Results>(source);
    }

    /**
	 * Returns a source of row-oriented FBA optimization output files. Each file
	 * is suitable for storing the results of a single FBA optimization. The
	 * value for {@linkplain #DATA_HEADERS} and {@linkplain #FILE_FORMAT} in the
	 * provided input are parsed and used.
	 * 
	 * @param <F>
	 *            the fluxome type.
	 * 
	 * @param input
	 *            keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fileNameSource
	 *            a source of desired output file names.
	 * @param fluxome
	 *            the fluxome for which the results correspond.
	 * 
	 * @return a source of source of serialized FBA optimization output files.
	 * 
	 * @see RowOrFbaOptOutputFile
	 * @see RofoofSource
	 */
    public static <F extends Fluxome>
    MultiOutputFiles<RowOrFbaOptOutputFile<F>, FbaOptimizer<F>.Results>
    getMultiRowOrOutputFiles(ManagerInput<String, String> input,
    						 PrintWriter out,
    						 Iterator<String> fileNameSource,
    						 F fluxome) {

        // Getting file format
        FileFormat fileFormat = getFileFormat(input);

        // Getting row header string
        String rowHeaders = input.getRequiredValue(DATA_HEADERS);

        // Creating a header parser
        FbaOptHeaderParser setFormat = new FbaOptHeaderParserV1(fluxome);

        // Parsing the data headers
        Set<String> headers = setFormat.parse(rowHeaders);

        // Getting file line format
        FbaOptFileLineFormat fbaFormat = new FbaOptFileLineFormatV1(fluxome);

        // Creating output file source
        Iterator<RowOrFbaOptOutputFile<F>> source =
                new RofoofSource<F>(fileNameSource, 
                					fileFormat, 
                					headers, 
                					fbaFormat);

        // Creating file
        return new MultiOutputFiles<RowOrFbaOptOutputFile<F>, 
        							FbaOptimizer<F>.Results>(source);
    }

    private FbaOptProcess<MixedFluxome> process;

    private FbaOptOutputFile<MixedFluxome> outputFile;

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object must
     * contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object must
     * contain.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * 
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public FbaOptProcess<MixedFluxome> getProcess() {

        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

        return process;
    }

    private <F extends Fluxome> FbaOptOutputFile<F> 
    getOutputFile(ManagerInput<String, String> input,
    			  PrintWriter out,
    			  F fluxome) {
        
    	Util.nullCheck(fluxome);

        // Getting the file name if it was provided
        String fileName = input.getOptionalValue(OUTPUT_FILE_NAME);

        // Getting the header string if it was provided
        String headerString = input.getOptionalValue(DATA_HEADERS);

        // If a header value was provided but a file name wasn't
        if (fileName == null && headerString != null) {
            String s =
                    "A value was provided for '"
                            + DATA_HEADERS
                            + "' but no value was provided for '"
                            + OUTPUT_FILE_NAME
                            + "' in '"
                            + input.getName()
                            + "'.";
            throw new FormatException(s);
        }

        // If no file name was provided, return null
        if (fileName == null) {
            return null;
        }

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting data headers
        Set<String> headers = getDataHeaders(headerString, fluxome);

        // Getting output file
        FbaOptOutputFile<F> outputFile =
                getSingleResultOutputFile(fileName, 
                						  FileFormat.TEXT, 
                						  headers,
                						  fluxome);

        // Printing a closing message
        out.println("DONE");
        out.println();

        // Returning the output file
        return outputFile;
    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Getting computer
        FbaOptimizer<MixedFluxome> computer = getMixedComp(input, out);

        // Getting fluxome
        MixedFluxome fluxome = computer.getFluxome();

        // Creating process
        process = new FbaOptProcess<MixedFluxome>(computer);

        // Getting and setting objective function
        LinearComb<String> obj = getObjFunc(input, out, fluxome);
        process.setObjectiveFunction(obj);

        // Getting and setting optimization sense
        OptSense sense = getOptSense(input);
        process.setOptimizationSense(sense);

        // Printing fluxome information
        printFluxomeInfo(fluxome, out);

        // Getting output file
        outputFile = getOutputFile(input, out, fluxome);
    }

    /**
     * Executes the process this process manager controls, and writes the status
     * or results of this process to the provided print writer.
     * 
     * @param out the print writer where status messages will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {

        Rm<MixedFluxome> rm = 
        	new Rm<MixedFluxome>(process, out, outputFile);

        try {

            process.run(rm);

        } catch (RuntimeException e) {

            throw interpret(e, process);

        } finally {

            if (outputFile != null) {
                closeFile(outputFile, out);
            }
        }

    }

    private static final class Rm<F extends Fluxome> 
    extends AbstractResultsManager<FbaOptimizer<F>.Results> {

        private final FbaOptProcess<F> process;

        private final PrintWriter out;

        private final FbaOptOutputFile<F> file;

        private Rm(FbaOptProcess<F> process, 
        		   PrintWriter out,
                   FbaOptOutputFile<F> file) {

            this.process = process;
            this.out = out;
            this.file = file;

            out.print("Computing Optimal Solution ... ");
            out.flush();

        }

        /**
         * Adds data to this results manager.
         * 
         * @param results the results, or data, to be added to this manager.
         * 
         * @throws IllegalStateException if {@code close()} has already been
         * called.
         */
        public void addData(FbaOptimizer<F>.Results results) {

            checkState();

            // Getting the objective value
            double objValue = results.getObjectiveValue();

            // Getting optimization sense
            OptSense sense = process.getOptimizationSense();

            // Getting number format
            DoubleFormat numberFormat = new DoubleFormatV1();

            // Printing a closing message
            out.println((sense == OptSense.MAX ? "Maximized"
                    : "Minimized")
                    + " objective = "
                    + numberFormat.format(objValue));

            // Writing file if necessary
            if (file != null) {

                out.println();
                file.addData(results);

            }

        }
    }

}
