/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.util.*;
import java.io.*;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.proc.*;
import sbrt.kernel.fba.proc.FluxActivityProcess.Results;
import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.utilities.*;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.fba.FluxVectorFormat;
import sbrt.shell.text.fba.FluxVectorFormatV1;

/**
 * This class is used to manage flux activity processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public final class FluxActivityManager extends FbaManager {

	/**
	 * The keyword used to denote the name of the activity count output file.
	 */
	public static final String ACTIVITY_COUNT_FILE_NAME = "Activity Count File Name";

	/**
	 * The keyword used to denote the name of the activity vector output file.
	 */
	public static final String ACTIVITY_VECTOR_FILE_NAME = "Activity Vector File Name";

	/**
	 * The keyword used to denote the format of the flux vector input file.
	 */
	public static final String FLUX_VECTOR_FILE_FORMAT = "Flux Vector File Format";

	/**
	 * The set of keywords required to be present in the {@code ManagerInput}
	 * objects supplied to instances of this class.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {
		Set<String> s = new TreeSet<String>();

		s.add(RXN_FILE);
		s.add(FLUX_VECTOR_FILE);
		s.add(FLUX_VECTOR_FILE_FORMAT);
		s.add(ZERO_CUTOFF);

		// Output file information
		s.add(ACTIVITY_COUNT_FILE_NAME);
		// s.add(ACTIVITY_VECTOR_FILE_NAME);

		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of optional keywords in the {@code ManagerInput} objects supplied
	 * to instances of this class.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
		Set<String> empty = Collections.emptySet();
		optionalKeywords = new UnmodifiableSet<String>(empty);
	}

	/**
	 * Returns the flux vector input file specified by the provided input. The
	 * values for {@linkplain #FLUX_VECTOR_FILE} and
	 * {@linkplain #FLUX_VECTOR_FILE_FORMAT} are parsed and used to call the
	 * method {@linkplain #getFluxVectorFile(String,  FileFormat, Fluxome)}.
	 * 
	 * @param input keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fluxome
	 *            the fluxome to which the flux vectors correspond.
	 * 
	 * @return the flux vector input file specified by the provided input.
	 */
	public static MapInputFile<String, Double> getFluxVectorFile(
			ManagerInput<String, String> input, PrintWriter out, Fluxome fluxome) {

		// Getting the file name
		String name = input.getRequiredValue(FLUX_VECTOR_FILE);

		// Printing an opening message
		out.print("Opening '" + name + "' ... ");
		out.flush();

		// Getting flux vector file format
		String s = input.getRequiredValue(FLUX_VECTOR_FILE_FORMAT);

		// Parsing it
		FileFormat f = getFileFormat(s);

		// Getting output file
		MapInputFile<String, Double> inputFile = 
				getFluxVectorFile(name, f, fluxome);

		// Printing a closing message
		out.println("DONE");

		// Returning the input file
		return inputFile;
	}

    /**
     * Returns the flux vector input file with the provided name. If this file 
     * is text-based, it must be a column-oriented map file that is parsable 
     * using {@linkplain FluxVectorFormatV1}.
     * 
     * @param fileName the name of the file to open.
     * @param format the format of the file to open.
     * @param fluxome the fluxome to which the flux vectors correspond.
     * 
     * @return the flux vector input file with the provided name.
     */
    public static MapInputFile<String, Double> 
    getFluxVectorFile(String fileName, FileFormat format, Fluxome fluxome) {

        // If this is a serialized file format
        if (SerializedInputFile.isValidFormat(format)) {
            
            // Creating serialized vector output file
            return new SerializedMapInputFile<String, Double>(fileName, format);
        }

        // If this is a text file format
        else if (TextOutputFile.isValidFormat(format)) {

            // Getting flux vector file line format
            FluxVectorFormat lineFormat =
                    new FluxVectorFormatV1(fluxome);

            // Returning row-oriented flux vector output file
            return new ComiFile<String, Double>(fileName, format, lineFormat);
        }

        // If it's neither
        else {
            String s = "The file format '" + format +
                       "' is not yet supported.";
            throw new IllegalArgumentException(s);
        }

    }
    
	private static OutputFile<String> 
	getActivityCountFile(ManagerInput<String, String> input, 
						 PrintWriter out) {

		// Getting the file name
		String fileName = input.getRequiredValue(ACTIVITY_COUNT_FILE_NAME);

		// Printing an opening message
		out.print("Opening '" + fileName + "' ... ");
		out.flush();

		// Getting output file
		OutputFile<String> outputFile = TextOutputFile.getBasic(fileName);

		// Printing a closing message
		out.println("DONE");

		// Returning the output file
		return outputFile;
	}

	private FluxActivityProcess<MixedFluxome> process;

	private InputFile<Map<String, Double>> inputFile;

	private OutputFile<String> fluxActivityFile;

	/**
	 * Returns the set of keywords the provided {@code ManagerInput} object must
	 * contain.
	 * 
	 * @return the set of keywords the provided {@code ManagerInput} object must
	 *         contain.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
	 * Returns the set of keywords the provided {@code ManagerInput} object does
	 * not necessarily have to contain.
	 * 
	 * @return the set of keywords the provided {@code ManagerInput} object does
	 *         not necessarily have to contain.
	 */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

	/**
	 * Returns the process this process manager controls.
	 * 
	 * @return the process this process manager controls.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager via
	 *             {@linkplain #setInput(ManagerInput)} or
	 *             {@linkplain #setInput(ManagerInput, PrintWriter)}
	 */
	public FluxActivityProcess<MixedFluxome> getProcess() {
		checkState();
		return process;
	}

	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input
	 *            all of the information required to prepare this process
	 *            manager for execution.
	 * 
	 * @param out
	 *            the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Closing previous file if necessary
		if (fluxActivityFile != null) {
			closeFile(fluxActivityFile, out);
		}

		// Checking the provided input
		check(input);

		
		// Getting fluxome
		MixedFluxome fluxome = getMixedFluxome(input, out);

		// Creating process
		process = new FluxActivityProcess<MixedFluxome>(fluxome);

		// Getting and setting zero cutoff
		double cutoff = getZeroCutoff(input, out);
		process.setZeroCutoff(cutoff);

		// Printing fluxome information
		printFluxomeInfo(fluxome, out);

		// Getting and setting flux vector source
		inputFile = getFluxVectorFile(input, out, fluxome);
		process.setFluxVectorSource(inputFile);

		// Getting flux activity output file
		fluxActivityFile = getActivityCountFile(input, out);
	}

	private void checkState() {

		if (process == null) {
			String s = "No input was provided.";
			throw new IllegalStateException(s);
		}

	}

	/**
	 * Executes the process this process manager controls, and writes the status
	 * or results of this process to the provided print writer.
	 * 
	 * @param out
	 *            the print writer where status messages will be printed.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager via
	 *             {@linkplain #setInput(ManagerInput)} or
	 *             {@linkplain #setInput(ManagerInput, PrintWriter)}
	 */
	public void execute(PrintWriter out) {

		checkState();

		Rm rm = new Rm(fluxActivityFile);

		try {
			runProcess(process, rm, out);
		} finally {
			closeFile(fluxActivityFile, out);
		}

	}

	private static class Rm extends AbstractResultsManager<Results> {

		private final OutputFile<String> fluxActivityFile;

		private Rm(OutputFile<String> fluxActivityFile) {

			// Saving
			this.fluxActivityFile = fluxActivityFile;

			// Writing simple header
			fluxActivityFile.addData("Active | Inactive");
		}

		/**
		 * Adds data to this results manager.
		 * 
		 * @param <S>
		 *            a type extending the primary result type.
		 * @param results
		 *            the results, or data, to be added to this manager.
		 * 
		 * @throws IllegalStateException
		 *             if {@code close()} has already been called.
		 */
		public <S extends Results> void addData(S results) {

			// Checking state
			if (isClosed()) {
				throw new IllegalStateException("already closed");
			}

			// Creating output line
			String s = results.active() + " | " + results.inactive();

			// Writing line to file
			fluxActivityFile.addData(s);
		}

	}
}
