/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.comp.FluxCapComputer;
import sbrt.kernel.fba.proc.FluxCapProcess;
import sbrt.kernel.io.SingleRm;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.algebra.SparseDoubleMatrix;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.ComiFile;
import sbrt.shell.io.FileFormat;
import sbrt.shell.io.MapOutputFile;
import sbrt.shell.io.RomoFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.AppInputFileLineFormat;
import sbrt.shell.text.AppInputFileLineFormatVersion1;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.PipeSetFormat;
import sbrt.shell.text.fba.RxnNameExprFormat;
import sbrt.shell.text.fba.RxnNameVerifier;

/**
 * This class is used to manage flux cap processes. Flux caps are used to
 * compute Mahadevan-Schilling flux intervals.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class FluxCapManager extends FbaManager {

	/**
	 * The keyword used to denote the steady state cycle input file.
	 */
    public static final String CYCLE_FILE = "Cycle File";

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        s.add(RXN_FILE);
        s.add(CYCLE_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
    	Set<String> s = Collections.emptySet(); 
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Opens the file denoted by {@linkplain #CYCLE_FILE} in the 
     * provided manager input and returns the list of cycles it contains.
     * This file must be a column-oriented map input file.
     * 
     * @param input the input containing a value for {@code CYCLE_FILE}.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome in which the cycles are contained.
     * 
     * @return the list of cycles contained in the file indicated by 
     * {@linkplain #CYCLE_FILE} in the provided input.
     */
    public static List<LinearComb<String>>
    getCycles(ManagerInput<String, String> input,
    				PrintWriter out,
    				Fluxome fluxome) {

    	// Getting file name
        String fileName = input.getRequiredValue(CYCLE_FILE);

        // Printing opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        RxnNameVerifier keyFormat = new RxnNameVerifier(fluxome);
        DoubleFormat valueFormat = new DoubleFormatV1();
        
        MapFormat<String, Double> mapFormat = 
            new EqualsMapFormatV1<String, Double>(keyFormat, valueFormat);
        
        ComiFile<String, Double> file = 
            new ComiFile<String, Double>(fileName, mapFormat);
        
        SparseDoubleMatrix<?, String> m = ComiFile.getMatrix(file);
        
        // Getting the cycles it contains
        List<LinearComb<String>> cycles = m.toLinearCombs(); 

        // Printing closing message
        out.println("DONE");

        return cycles;
    }

    private static MapOutputFile<String, String>
    getOutputFile(ManagerInput<String, String> input,
    			  PrintWriter out,
    			  Fluxome fluxome) {

    	// Getting file name
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Creating map format
        AppInputFileLineFormat format = new AppInputFileLineFormatVersion1();

        // Opening output file
        MapOutputFile<String, String> file =
        	new RomoFile<String, String>(fileName, FileFormat.TEXT, format);

        // Printing closing message
        out.println("DONE");

        return file;
    }

    private FluxCapProcess process;
    private MapOutputFile<String, String> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public FluxCapProcess getProcess() {
        checkState();
        return process;
    }

    private void closeFile(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }

    }

    private void checkState() {
        if (process == null) {
            throw new IllegalStateException("No manager input was provided.");
        }
    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing previous file if necessary
        closeFile(out);

        // Getting fluxome
        IrrevRxnFluxome fluxome = getIrrevRxnFluxome(input, out);

        // Getting cycles
        Collection<LinearComb<String>> cycles = 
        	getCycles(input, out, fluxome);

        // Creating computer
        FluxCapComputer computer = new FluxCapComputer(fluxome, cycles);

        process = new FluxCapProcess(computer);

        // Getting fluxome and printing basic information
        printFluxomeInfo(fluxome, out);

        // Setting output file
        outputFile = getOutputFile(input, out, fluxome);
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();

        try {

            // Creating simple results manager
            SingleRm<Map<String, Set<LinearComb<String>>>> rm = 
            	new SingleRm<Map<String, Set<LinearComb<String>>>>();

            // Adding results to this manager
            runProcess(process, rm, out);

            // Retreiving the results
            Map<String, Set<LinearComb<String>>> fluxCaps = rm.getResults();

            // Getting fluxome
            Fluxome fluxome = process.getFluxome();

            // Creating reaction name formatter
            RxnNameVerifier rxnFormat = new RxnNameVerifier(fluxome);

            // Creating reaction name expression formatter
            RxnNameExprFormat<LinearComb<String>, String> expFormat = 
            	getRxnNameExprFormat(fluxome);

            // Creating set formatter
            PipeSetFormat<LinearComb<String>> setFormat = 
            	new PipeSetFormat<LinearComb<String>>(expFormat);

            // Creating storage for formatted flux caps
            Map<String, String> formattedCaps = 
            	new LinkedHashMap<String, String>(fluxCaps.size() + 1, 1.0F);

            // Iterating thru each flux cap
            for (Map.Entry<String, Set<LinearComb<String>>> e : 
            	 fluxCaps.entrySet()) {

                // Getting the next reaction and its corresponding flux caps
                String rxnName = e.getKey();
                Set<LinearComb<String>> cap = e.getValue();

                // Formatting the reaction name and set of flux caps
                String formattedRxn = rxnFormat.format(rxnName);
                String formattedCap = setFormat.format(cap);

                // Adding formatted entries
                formattedCaps.put(formattedRxn, formattedCap);
            }

            // Writing formatted flux caps to output file and closing
            outputFile.addData(formattedCaps);

        } finally {
            closeFile(out);
        }
    }


}
