/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.util.*;
import java.io.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.fba.proc.*;
import sbrt.kernel.fba.proc.FluxPlasticityProcess.Results;
import sbrt.kernel.io.*;
import sbrt.kernel.math.Interval;
import sbrt.kernel.utilities.*;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.fba.*;

/**
 * This class is used to manage flux plasticity processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class FluxPlasticityManager extends FbaManager {

    /**
     * The keyword used to denote the name of the file containing the sets of
     * flux intervals.
     */
    public static final String FLUX_INTERVAL_FILE =
            IntervalVariationManager.FLUX_INTERVAL_FILE;

    /**
     * The keyword used to denote the name of the output file for the names of
     * always-active reactions.
     */
    public static final String ALWAYS_ACTIVE_FILE_NAME =
            "Always-Active Reaction File Name";

    /**
     * The keyword used to denote the name of the output file for the names of
     * potentially-active reactions.
     */
    public static final String POT_ACTIVE_FILE_NAME =
            "Potentially-Active Reaction File Name";

    /**
     * The keyword used to denote the name of the output file for the names of
     * never-active reactions.
     */
    public static final String NEVER_ACTIVE_FILE_NAME =
            "Never-Active Reaction File Name";

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        s.add(RXN_FILE);
        s.add(FLUX_INTERVAL_FILE);
        s.add(ZERO_CUTOFF);

        // Output file information
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {

        Set<String> s = new TreeSet<String>();

        s.add(ALWAYS_ACTIVE_FILE_NAME);
        s.add(POT_ACTIVE_FILE_NAME);
        s.add(NEVER_ACTIVE_FILE_NAME);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Returns the flux interval input with the name denoted by
     * {@linkplain #FLUX_INTERVAL_FILE} in the provided manager input.
     * 
     * @param input keyword-value pairs.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the reaction names correspond.
     * 
     * @return the flux interval input with the name denoted by
     * {@linkplain #FLUX_INTERVAL_FILE} in the provided manager input.
     */
    public static MapInputFile<String, Interval> 
    getFluxIntervalFile(ManagerInput<String, String> input,
    					PrintWriter out,
    					Fluxome fluxome) {

        // Getting the flux interval file name
        String fileName = input.getRequiredValue(FLUX_INTERVAL_FILE);

        // Printing opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting input file
        MapInputFile<String, Interval> file =
                getFluxIntervalFile(fileName, fluxome);

        // Printing the closing message
        out.println("DONE");

        // Returning the file
        return file;
    }

    /**
     * Returns the flux interval input file with the provided name.
     * 
     * @param fileName the name of the flux interval file to open.
     * @param fluxome the fluxome to which the reaction names correspond.
     * 
     * @return the flux interval input file with the provided name.
     */
    public static MapInputFile<String, Interval> 
    getFluxIntervalFile(String fileName,
    					Fluxome fluxome) {

        // Getting the file line format
        ConstraintsFileLineFormat<String> format =
                new ConstraintsFileLineFormatV1(fluxome);

        // Required variables
        Set<String> empty = Collections.emptySet();

        // Getting the column-oriented input file
        MapInputFile<String, Interval> inputFile =
                new ComiFile<String, Interval>(fileName, 
                							   FileFormat.TEXT,
                							   empty, 
                							   format);

        // Returning the file
        return inputFile;
    }

    private static OutputFile<String> 
    getOutputFile(ManagerInput<String, String> input,
    			  PrintWriter out) {

        // Getting the file name
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting output file
        OutputFile<String> outputFile = TextOutputFile.getBasic(fileName);

        // Printing a closing message
        out.println("DONE");

        // Returning the output file
        return outputFile;
    }

    private static OutputFile<Set<String>> 
    getRxnNameOutputFile(String fileName,
    					 Fluxome fluxome) {

        // Creating a reaction name formatter
        RxnNameSetFormat format = new RxnNameSetFormatV1(fluxome);

        // Creating the file
        OutputFile<Set<String>> output =
                new TextOutputFile<Set<String>>(fileName, format);

        return output;
    }

    private static OutputFile<Set<String>> 
    getRxnNameOutputFile(String fileName,
    					 PrintWriter out,
    					 Fluxome fluxome) {

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Creating the file
        OutputFile<Set<String>> output =
                getRxnNameOutputFile(fileName, fluxome);

        // Printing a closing message
        out.println("DONE");

        return output;
    }

    private static OutputFile<Set<String>> 
    getAlwaysActiveRxnFile(ManagerInput<String, String> input,
    					   PrintWriter out,
    					   Fluxome fluxome) {

        // Getting the file name
        String fileName = input.getOptionalValue(ALWAYS_ACTIVE_FILE_NAME);

        // If no file name was provided
        if (fileName == null) {
            return null;
        }

        return getRxnNameOutputFile(fileName, out, fluxome);
    }

    private static OutputFile<Set<String>> 
    getCondActiveRxnFile(ManagerInput<String, String> input,
    					 PrintWriter out,
    					 Fluxome fluxome) {

        // Getting the file name
        String fileName = input.getOptionalValue(POT_ACTIVE_FILE_NAME);

        // If no file name was provided
        if (fileName == null) {
            return null;
        }

        return getRxnNameOutputFile(fileName, out, fluxome);
    }

    private static OutputFile<Set<String>> 
    getNeverActiveRxnFile(ManagerInput<String, String> input,
    					  PrintWriter out,
    					  Fluxome fluxome) {

        // Getting the file name
        String fileName = input.getOptionalValue(NEVER_ACTIVE_FILE_NAME);

        // If no file name was provided
        if (fileName == null) {
            return null;
        }

        return getRxnNameOutputFile(fileName, out, fluxome);
    }

    private FluxPlasticityProcess<MixedFluxome> process;

    private MapInputFile<String, Interval> inputFile;

    private OutputFile<String> outputFile;

    // Optional output files
    private OutputFile<Set<String>> alwaysActiveRxnFile;

    private OutputFile<Set<String>> condActiveRxnFile;

    private OutputFile<Set<String>> neverActiveRxnFile;

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object must
     * contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object must
     * contain.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * 
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public FluxPlasticityProcess<MixedFluxome> getProcess() {
        checkState();
        return process;
    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Closing previous file if necessary
        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        // Checking the provided input
        check(input);

        // Getting fluxome
        MixedFluxome fluxome = getMixedFluxome(input, out);

        // Creating process
        process = new FluxPlasticityProcess<MixedFluxome>(fluxome);

        // Getting and setting zero cutoff
        double cutoff = getZeroCutoff(input, out);
        process.setZeroCutoff(cutoff);

        // Printing fluxome information
        printFluxomeInfo(fluxome, out);

        // Getting and setting flux interval source
        inputFile = getFluxIntervalFile(input, out, fluxome);
        process.setFluxIntervalSource(inputFile);

        // Getting the standard flux activity output file
        outputFile = getOutputFile(input, out);

        // Getting the always-active reaction file
        alwaysActiveRxnFile = getAlwaysActiveRxnFile(input, out, fluxome);

        // Getting the conditionally-active reaction file
        condActiveRxnFile = getCondActiveRxnFile(input, out, fluxome);

        // Getting the never-active reaction file
        neverActiveRxnFile = getNeverActiveRxnFile(input, out, fluxome);
    }

    private void checkState() {

        if (process == null) {
            String s = "No input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Executes the process this process manager controls, and writes the status
     * or results of this process to the provided print writer.
     * 
     * @param out the print writer where status messages will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {

        checkState();

        Rm rm = new Rm(outputFile, 
        			   alwaysActiveRxnFile, 
        			   condActiveRxnFile,
                       neverActiveRxnFile, 
                       process.getFluxome());

        try {

            runProcess(process, rm, out);

        } finally {

            closeFile(outputFile, out);
            closeFile(inputFile, out);

            // Closing optional files
            if (alwaysActiveRxnFile != null) {
                closeFile(alwaysActiveRxnFile, out);
            }
            if (condActiveRxnFile != null) {
                closeFile(condActiveRxnFile, out);
            }
            if (neverActiveRxnFile != null) {
                closeFile(neverActiveRxnFile, out);
            }
        }

    }

    private static class Rm extends AbstractResultsManager<Results> {

        private final OutputFile<String> outputFile;

        private final OutputFile<Set<String>> alwaysActiveRxnFile;

        private final OutputFile<Set<String>> condActiveRxnFile;

        private final OutputFile<Set<String>> neverActiveRxnFile;

        private Rm(OutputFile<String> outputFile,
                OutputFile<Set<String>> alwaysActiveRxnFile,
                OutputFile<Set<String>> condActiveRxnFile,
                OutputFile<Set<String>> neverActiveRxnFile, Fluxome fluxome) {

            // Saving
            this.outputFile = outputFile;
            this.alwaysActiveRxnFile = alwaysActiveRxnFile;
            this.condActiveRxnFile = condActiveRxnFile;
            this.neverActiveRxnFile = neverActiveRxnFile;

            // Creating header string
            String header =
                    "Plastic | "
                            + "Rigid | "
                            + "Always_Active | "
                            + "Potentially_Active | "
                            + "Never_Active";

            // Writing simple header
            outputFile.addData(header);

        }

        /**
         * Adds data to this results manager.
         * 
         * @param <S> a type extending the primary result type.
         * @param results the results, or data, to be added to this manager.
         * 
         * @throws IllegalStateException if {@code close()} has already been
         * called.
         */
        public <S extends Results> void addData(S results) {

            // Checking state
            if (isClosed()) {
                throw new IllegalStateException("already closed");
            }

            // Creating output line
            String s =
                    results.plastic()
                            + " | "
                            + results.rigid()
                            + " | "
                            + results.alwaysActive()
                            + " | "
                            + results.potentiallyActive()
                            + " | "
                            + results.neverActive();

            // Writing line to file
            outputFile.addData(s);

            // If an always-active reaction file was provided
            if (alwaysActiveRxnFile != null) {

                // Getting always-active reactions
                Set<String> rxns = results.getAlwaysActive();

                // Writing to file
                alwaysActiveRxnFile.addData(rxns);
            }

            // If a conditionally-active reaction file was provided
            if (condActiveRxnFile != null) {

                // Getting conditionally-active reactions
                Set<String> rxns = results.getPotActive();

                // Writing to file
                condActiveRxnFile.addData(rxns);
            }

            // If a never-active reaction file was provided
            if (neverActiveRxnFile != null) {

                // Getting never-active reactions
                Set<String> rxns = results.getNeverActive();

                // Writing to file
                neverActiveRxnFile.addData(rxns);
            }

        }

    }
}
