/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.FluxConstraints;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.comp.FbaOptException;
import sbrt.kernel.fba.comp.FbaOptimizer;
import sbrt.kernel.fba.comp.InitialPointComputer;
import sbrt.kernel.fba.proc.FbaProcess;
import sbrt.kernel.fba.proc.InitialPointGenerator;
import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.optimization.ProgramSolver;
import sbrt.kernel.math.optimization.ProgramSolverStatus;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.ProgramSolverFactory;
import sbrt.shell.io.FileFormat;
import sbrt.shell.io.MapOutputFile;
import sbrt.shell.io.RomiFile;
import sbrt.shell.io.fba.FluxIntervalInputFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.fba.RxnNameExprFormatV1;
import sbrt.shell.text.fba.RxnNameVerifier;
import sbrt.shell.text.math.BoundedIntervalFormatV1;
import sbrt.shell.text.math.IntervalFormat;

/**
 * This class is used to manage initial point processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class InitialPointManager extends FbaManager {

    /**
     * The keyword used to denote the P intercept (the value of positioning
     * variables at the edge of the convex poytope).
     */
    public static final String P_MAX = "P Max";

    /**
     * The default P intercept value.
     */
    public static final double DEFAULT_P = 1E15d;

    /**
     * The keyword used to denote the name of the C vector output file.
     */
    public static final String P_VECTOR_FILE_NAME = "P Vector File Name";

    /**
     * The keyword used to denote the name of the file containing the set of
     * flux intervals.
     */
    public static final String FLUX_INTERVAL_FILE =
            IntervalVariationManager.FLUX_INTERVAL_FILE;

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        s.add(RXN_FILE);
        s.add(FLUX_INTERVAL_FILE);
        s.add(ProgramSolverFactory.SOLVER);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        s.add(FLUX_VECTOR_FILE_NAME);
        s.add(P_MAX);
        s.add(P_VECTOR_FILE_NAME);
        s.add(FbaOptManager.SAFETY_LEVEL);
        s.add(CONSTRAINT_TOLERANCE);
        s.add(ProgramSolverFactory.PARAM_FILE);
        s.add(ZERO_CUTOFF);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Returns the parsed value for {@linkplain #P_MAX} in the provided input.
     * 
     * @param input the input whose {@code P_INTERCEPT} value is to be parsed.
     * @param out the print writer to which a status message will be printed.
     * 
     * @return the parsed numerical P intercept value; or {@linkplain #DEFAULT_P} if
     * no value for {@code P_INTERCEPT} exists.
     * 
     * @throws ApplicationException if the parsed intercept is infinite, NaN, or <
     * 10.
     */
    public static double getPIntercept(ManagerInput<String, String> input,
            PrintWriter out) {

        String value = input.getOptionalValue(P_MAX);

        if (value == null) {
            return DEFAULT_P;
        }

        DoubleFormat format = new DoubleFormatV1();

        double p = format.parsePrimitive(value);

        // Checking its value
        if (p < 10 || Double.isInfinite(p) || Double.isNaN(p)) {

            String s =
                    "The value '"
                            + p
                            + "' for '"
                            + DEFAULT_P
                            + "' in '"
                            + input.getName()
                            + "' is invalid.";

            throw new ApplicationException(s);

        }

        out.println("Parsed " + P_MAX + ": " + format.format(p));

        // Returning parsed value
        return p;
    }

    /**
     * Returns an empty output file that is suitable for storing a single flux
     * vector. The values for {@linkplain #FLUX_VECTOR_FILE_NAME} and
     * {@linkplain #DATA_HEADERS} in the provided input are
     * parsed and used to call the method
     * {@linkplain #getSingleFluxVectorOutputFile(String, FileFormat, Set, Fluxome)}.
     * 
     * @param input the input containing values for the listed keywords.
     * @param out the print writer to which status messages will be printed.
     * @param fluxome the fluxome for which the data headers correspond.
     * 
     * @return an empty output file that is suitable for storing a single flux
     * vector.
     * 
     * @see #getRxnNameHeaders(ManagerInput, Fluxome)
     */
    public static MapOutputFile<String, Double> 
    getFluxVectorFile(ManagerInput<String, String> input,
    				  PrintWriter out,
    				  Fluxome fluxome) {

        // Getting flux vector file name
        String fileName = input.getRequiredValue(FLUX_VECTOR_FILE_NAME);

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting data headers
        Set<String> rxnNames = getRxnNameHeaders(input, fluxome);

        // Getting output file
        MapOutputFile<String, Double> file =
                getSingleFluxVectorOutputFile(fileName, FileFormat.TEXT,
                        rxnNames, fluxome);

        // Printing closing message
        out.println("DONE");

        // Returning file;
        return file;
    }

    /**
     * Returns an empty output file that is suitable for storing a single C
     * vector with the name denoted by {@linkplain #P_VECTOR_FILE_NAME} in the
     * provided input.
     * 
     * @param input the input containing values for the listed keywords.
     * @param out the print writer to which status messages will be printed.
     * @param fluxome the fluxome for which the data headers correspond.
     * 
     * @return an empty output file that is suitable for storing a single C
     * vector.
     * 
     */
    public static MapOutputFile<String, Double> 
    getPVectorFile(ManagerInput<String, String> input,
    			   PrintWriter out,
    			   Fluxome fluxome) {

        // Getting flux vector file name
        String fileName = input.getRequiredValue(P_VECTOR_FILE_NAME);

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting output file
        MapOutputFile<String, Double> file =
                getSingleFluxVectorOutputFile(fileName, 
                							  FileFormat.TEXT, 
                							  null,
                							  fluxome);

        // Printing closing message
        out.println("DONE");

        // Returning file;
        return file;
    }

    /**
     * Creates and returns an FBA optimization computer based on the provided
     * input. The values for {@linkplain #RXN_FILE}, {@linkplain #FLUX_INTERVAL_FILE},
     * {@linkplain FbaManager#CONSTRAINT_TOLERANCE},
     * {@linkplain FbaOptManager#SAFETY_LEVEL}, {@linkplain ProgramSolverFactory#SOLVER},
     * and {@linkplain ProgramSolverFactory#PARAM_FILE} are parsed and combined to
     * create the returned optimization computer.
     * 
     * @param input the input containing values for the listed keywords.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return an FBA optimization computer based on the provided input.
     * 
     * @see #getMixedFluxome(ManagerInput, PrintWriter)
     * @see FbaOptManager#getSafetyLevel(ManagerInput, PrintWriter)
     * @see #getConstraintTolerance(ManagerInput, PrintWriter)
     * @see ProgramSolverFactory#make(ManagerInput, PrintWriter)
     */
    public static FbaOptimizer<MixedFluxome> 
    getMixedComp(ManagerInput<String, String> input,
    			PrintWriter out) {

        // Getting the optimization computer safety level
        int safetyLevel = FbaOptManager.getSafetyLevel(input, out);

        // Getting the constraint tolerance
        double tolerance = getConstraintTolerance(input, out);

        // Getting the fluxome
        MixedFluxome fluxome = 
        	InitialPointManager.getMixedFluxome(input, out, tolerance);

        // Getting the program solver
        ProgramSolver<Object, Object> ps =
                ProgramSolverFactory.make(input, out);

        // Creating the optimization computer
        FbaOptimizer<MixedFluxome> c = 
        	new FbaOptimizer<MixedFluxome>(fluxome, ps);
        
        c.setSafetyLevel(safetyLevel);
        c.setTolerance(tolerance);

        // Returning the computer
        return c;
    }

    /**
     * Creates and returns a fluxome based on the provided input. The values for
     * {@linkplain #RXN_FILE} and {@linkplain #FLUX_INTERVAL_FILE} are parsed and combined
     * to create the returned fluxome.
     * 
     * @param input the input containing values for the listed keywords.
     * @param out the print writer to which status messages will be printed.
     * @param tolerance the constraint tolerance.
     * 
     * @return a fluxome based on the provided input.
     * 
     * @see FbaManager#getMixedFluxome(ManagerInput, PrintWriter)
     * @see #getIntervals(ManagerInput, PrintWriter, Fluxome, double)
     * @see Fluxome#replaceConstraints(FluxConstraints, double)
     */
    public static MixedFluxome getMixedFluxome(ManagerInput<String, String> input,
            PrintWriter out,
            double tolerance) {

        // Getting basic fluxome (i.e. one with default flux constraints)
    	MixedFluxome fluxome = FbaManager.getMixedFluxome(input, out);

        // Getting constraints
        FluxConstraints intervals = getIntervals(input, out, fluxome, tolerance);

        // Returning new fluxome
        return fluxome.replaceConstraints(intervals, tolerance);
    }

    /**
     * Opens the file denoted by {@linkplain #FLUX_INTERVAL_FILE} in the provided
     * manager input and returns the set of flux intervals it contains.
     * 
     * @param input the input containing a value for {@code FLUX_INTERVAL_FILE}.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the flux intervals correspond.
     * @param tolerance the constraint tolerance.
     * 
     * @return the set of flux intervals contained in the file indicated by
     * {@code #FLUX_INTERVAL_FILE} in the provided input.
     * 
     * @throws ApplicationException if the file does not contain a flux interval
     * for each reaction in the provided fluxome, or if an interval in
     * unbounded.
     * 
     * @see #getIntervals(String, Fluxome, double)
     */
    public static FluxConstraints getIntervals(ManagerInput<String, String> input,
            PrintWriter out,
            Fluxome fluxome,
            double tolerance) {

        // Getting the name of the flux interval file
        String fileName = input.getRequiredValue(FLUX_INTERVAL_FILE);

        // Printing an opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Converting to constraint set
        FluxConstraints constraints = getIntervals(fileName, fluxome, tolerance);

        // Printing a closing message
        out.println("DONE");

        // Returning the flux constraints contained in the file
        return constraints;
    }

    /**
     * Returns the set of flux intervals contained in the file with the provided
     * name. This file must have the format of a row-oriented map file. The
     * reaction names in this file must be parsable using
     * {@linkplain RxnNameVerifier}, and the intervals must be parsable using
     * {@linkplain BoundedIntervalFormatV1}. All reaction names from the provided
     * fluxome must be present.
     * 
     * @param fileName the name of the file to be read.
     * @param fluxome the fluxome to which the flux intervals correspond.
     * @param tolerance the contraint tolerance.
     * 
     * @return the set of flux intervals contained in the file with the provided
     * name.
     * 
     * @throws ApplicationException if the file does not contain a flux interval
     * for each reaction in the provided fluxome, or if an interval in
     * unbounded.
     * 
     * @throws FormatException if any part of the file cannot be parsed.
     * @throws NullPointerException if any argument is {@code null}.
     * 
     * @see RomiFile
     */
    public static FluxConstraints 
    getIntervals(String fileName, Fluxome fluxome, double tolerance) {

        // Creating key parser
        RxnNameExprFormatV1 keyParser = new RxnNameExprFormatV1(fluxome);

        // Creating value parser
        IntervalFormat valueParser = new BoundedIntervalFormatV1();

        // Creating map parser
        MapFormat<LinearComb<String>, Interval> mapParser =
                new EqualsMapFormatV1<LinearComb<String>, Interval>(keyParser,
                        valueParser);

        // Getting all reaction names
        Set<String> rxnNames = fluxome.getRxnNames();

        // Creating required variables
        Set<LinearComb<String>> required =
                FluxIntervalInputFile.convertVariables(rxnNames);

        // Getting the row-oriented constraints file
        RomiFile<LinearComb<String>, Interval> file =
                new RomiFile<LinearComb<String>, Interval>(fileName,
                        FileFormat.TEXT, required, mapParser);

        // Getting the flux intervals vector
        Map<LinearComb<String>, Interval> vector = file.getMap();

        // Closing the file
        file.close();

        // Returning the flux constraints contained in the file
        return FluxIntervalInputFile.convertConstraints(vector, fluxome, tolerance);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e an FBA optimization exception.
     * @param process the FBA process from which the exception was thrown.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException 
    interpret(FbaOptException e, FbaProcess<?, ?> process) {

        // Gathering information from the FBA optimization exception
        ProgramSolverStatus status = e.getStatus();

        StringBuilder s = new StringBuilder(1000);

        s.append("The program solver was unable to find a minimum P. " +
                 "The status was reported to be '"
                + status
                + "'.");

        if (status == ProgramSolverStatus.UNBOUNDED) {
            s.append(" This can be caused by inappropriately constrained "
                            + "exchange reactions or by the presence of cycles.");
        }

        else if (status == ProgramSolverStatus.INFEASIBLE) {
            s.append(" This can be caused by conflicting flux constraints "
                    + "or by inappropriate parameter settings for the "
                    + "program solver.");
        }

        else if (status != ProgramSolverStatus.UNKNOWN) {
            s.append(" The documentation of the linear program solver " +
            "should contain information regarding this type of status.");
        }
        
        return new ApplicationException(s.toString(), e);
    }

    private InitialPointGenerator<MixedFluxome> generator;

    private MapOutputFile<String, Double> fluxVectorFile;

    private MapOutputFile<String, Double> pVectorFile;

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object must
     * contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object must
     * contain.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * 
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public InitialPointGenerator<MixedFluxome> getProcess() {
        checkState();
        return generator;
    }

    private void closeFiles(PrintWriter out) {

        if (fluxVectorFile != null) {
            closeFile(fluxVectorFile, out);
        }

        if (pVectorFile != null) {
            closeFile(pVectorFile, out);
        }

    }

    private void setOutputFiles(ManagerInput<String, String> input,
            PrintWriter out,
            Fluxome fluxome) {

        if (input.hasValueFor(FLUX_VECTOR_FILE_NAME)) {
            fluxVectorFile = getFluxVectorFile(input, out, fluxome);
        }

        if (input.hasValueFor(P_VECTOR_FILE_NAME)) {
            pVectorFile = getPVectorFile(input, out, fluxome);
        }

    }

    private void checkState() {
        if (generator == null) {
            throw new IllegalStateException("No manager input was provided.");
        }
    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing previous files if necessary
        closeFiles(out);

        double intercept = getPIntercept(input, out);
        double zeroCutoff = getZeroCutoff(input, out);
        
        out.println();
        
        // Getting optimization computer
        FbaOptimizer<MixedFluxome> opt =
                InitialPointManager.getMixedComp(input, out);

        // Creating computer
        InitialPointComputer<MixedFluxome> pointComputer = 
        	new InitialPointComputer<MixedFluxome>(opt);
        
        pointComputer.setPIntercept(intercept);
        pointComputer.setZeroCutoff(zeroCutoff);
        
        // Creating initial point process
        generator = new InitialPointGenerator<MixedFluxome>(pointComputer);
        
        // Getting fluxome and printing basic information
        Fluxome fluxome = opt.getFluxome();
        printFluxomeInfo(fluxome, out);

        // Setting output files
        setOutputFiles(input, out, fluxome);
    }

    /**
     * Executes the process this process manager controls, and writes the status
     * or results of this process to the provided print writer.
     * 
     * @param out the print writer where status messages will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {

        checkState();

        try {

            // Creating results manager
            Rm rm = new Rm(out, fluxVectorFile, pVectorFile);

            // Running the process
            generator.run(rm);

        } catch (FbaOptException e) { 

            out.println();
            out.println();
            throw interpret(e, generator);
        
        } catch (RuntimeException e) {

            out.println();
            out.println();
            throw FbaOptManager.interpret(e, generator);

        } finally {

            closeFiles(out);

        }
    }

    private static class Rm 
    extends AbstractResultsManager<InitialPointComputer<MixedFluxome>.Results> {

        private final PrintWriter out;

        private final MapOutputFile<String, Double> fluxVectorFile;

        private final MapOutputFile<String, Double> pVectorFile;

        private Rm(PrintWriter out,
                MapOutputFile<String, Double> fluxVectorFile,
                MapOutputFile<String, Double> pVectorFile) {

            // Printing opening message
            out.println();
            out.print("Computing Minimum P ... ");
            out.flush();

            this.out = out;
            this.fluxVectorFile = fluxVectorFile;
            this.pVectorFile = pVectorFile;
        }

        /**
         * Adds data to this results manager.
         * 
         * @param <T> a type extending the primary Results type.
         * 
         * @param results the results, or data, to be added to this manager.
         * 
         * @throws IllegalStateException if {@code close()} has already been
         * called.
         */
        public <T extends InitialPointComputer<MixedFluxome>.Results> 
        void addData(T results) {

            checkState();

            // Getting the minimum and average values of positioning variables
            double pMin = results.getPMin();
            double pAvg = results.getPAvg();

            // Creating number formatter
            DoubleFormat f = new DoubleFormatV1();

            // Printing closing message
            out.println("DONE");
            out.println();
            out.println("P variables = " + results.pVariables());
            out.println("Minimum P = " + f.format(pMin));
            out.println("Average P = " + f.format(pAvg));
            out.println();

            // Writing files if necessary
            if (fluxVectorFile != null) {
                fluxVectorFile.addData(results.getFluxVector());
            }

            if (pVectorFile != null) {
                Map<String, Double> vector = results.getPVector();
                pVectorFile.addData(vector);
            }

        }

    }

}
