/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.*;
import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.LinearComb;
import sbrt.shell.io.*;
import sbrt.shell.io.fba.FluxIntervalInputFile;
import sbrt.shell.mng.*;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.fba.*;
import sbrt.shell.text.math.BoundedIntervalFormatV1;
import sbrt.shell.text.math.IntervalFormat;

/**
 * This abstract class provides utitlity functions for managing flux interval
 * variation processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public abstract class IntervalVariationManager extends FbaManager {

    /**
     * The keyword used to denote a constraint variation input file.
     */
    public static final String CV_FILE = "Constraint Variation File";

    /**
     * The keyword used to denote a flux interval input file.
     */
    public static final String FLUX_INTERVAL_FILE = "Flux Interval File";

    /**
     * Returns the constraint variation input with the name denoted by
     * {@linkplain #CV_FILE} in the provided manager input.
     * 
     * @param input the input containing a value for {@code CV_FILE}.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the constraints correspond.
     * @param tolerance the constraint tolerance.
     * 
     * @return the constraint variation input with the name denoted by
     * {@linkplain #CV_FILE} in the provided manager input.
     */
    public static FluxIntervalInputFile 
    getCvInputFile(ManagerInput<String, String> input,
    			   PrintWriter out,
    			   Fluxome fluxome,
    			   double tolerance) {

        // Getting the constaint-variation file name
        String fileName = input.getRequiredValue(CV_FILE);

        // Printing opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting input file
        FluxIntervalInputFile file = getCvInputFile(fileName, fluxome, tolerance);

        // Printing the closing message
        out.println("DONE");

        // Returning the file
        return file;
    }

    /**
     * Returns the constraint variation input file with the provided name.
     * 
     * @param fileName the name of the constraint variation file to open.
     * @param fluxome the fluxome to which the constraints correspond.
     * @param tolerance the constraint tolerance.
     * 
     * @return the constraint variation input file with the provided name.
     */
    public static FluxIntervalInputFile 
    	getCvInputFile(String fileName, Fluxome fluxome, double tolerance) {

        // Getting the file line format
        ConstraintsFileLineFormat<LinearComb<String>> format =
                new ConstraintsFileLineFormatV2(fluxome);

        // Required variables
        Set<LinearComb<String>> empty = Collections.emptySet();

        // Getting the column-oriented input file
        FluxIntervalInputFile inputFile =
                new FluxIntervalInputFile(fileName, 
                						  FileFormat.TEXT, 
                						  empty, 
                						  format, 
                						  fluxome, 
                						  tolerance);

        // Returning the file
        return inputFile;
    }

    /**
     * Returns the space variation input with the name denoted by
     * {@linkplain #FLUX_INTERVAL_FILE} in the provided manager input.
     * 
     * @param input the input containing a value for {@code FLUX_INTERVAL_FILE}.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the constraints correspond.
     * @param tolerance the constraint tolerance.
     * 
     * @return the space variation input with the name denoted by
     * {@linkplain #FLUX_INTERVAL_FILE} in the provided manager input.
     */
    public static FluxIntervalInputFile 
    getSvInputFile(ManagerInput<String, String> input,
    			   PrintWriter out,
    			   Fluxome fluxome,
    			   double tolerance) {

        // Getting the space-variation file name
        String fileName = input.getRequiredValue(FLUX_INTERVAL_FILE);

        // Printing opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting input file
        FluxIntervalInputFile file = 
        	getIntervalInputFile(fileName, fluxome, tolerance);

        // Printing the closing message
        out.println("DONE");

        // Returning the file
        return file;
    }

    /**
     * Returns the flux interval input file with the provided name.
     * 
     * @param fileName the name of the flux interval file to open.
     * @param fluxome the fluxome to which the constraints correspond.
     * @param mapFormat the parser.
     * @param tolerance the constraint tolerance.
     * 
     * @return the flux interval input file with the provided name.
     */
    public static FluxIntervalInputFile 
    getIntervalInputFile(String fileName,
    					 Fluxome fluxome,
    					 MapFormat<LinearComb<String>, Interval> mapFormat,
    					 double tolerance) {

        // Getting all reaction names
        Set<String> rxnNames = fluxome.getRxnNames();

        // Required variables
        Set<LinearComb<String>> required =
                FluxIntervalInputFile.convertVariables(rxnNames);

        // Getting the column-oriented input file
        FluxIntervalInputFile inputFile =
                new FluxIntervalInputFile(fileName, 
                						  FileFormat.TEXT, 
                						  required,
                						  mapFormat, 
                						  fluxome, 
                						  tolerance);

        // Returning the file
        return inputFile;
    }

    /**
     * Returns the flux interval input file with the provided name.
     * 
     * @param fileName the name of the flux interval file to open.
     * @param fluxome the fluxome to which the constraints correspond.
     * @param tolerance the constraint tolerance.
     * 
     * @return the flux interval input file with the provided name.
     */
    public static FluxIntervalInputFile 
    getIntervalInputFile(String fileName,
    					 Fluxome fluxome, 
    					 double tolerance) {

        // Creating key parser
        RxnNameExprFormatV1 keyParser = new RxnNameExprFormatV1(fluxome);

        // Creating value parser
        IntervalFormat valueParser = new BoundedIntervalFormatV1();

        // Creating map parser
        MapFormat<LinearComb<String>, Interval> mapParser =
             new EqualsMapFormatV1<LinearComb<String>, Interval>(keyParser, 
            		 											 valueParser);

        // Returning the file
        return getIntervalInputFile(fileName, fluxome, mapParser, tolerance);
    }
}
