/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.InputFile;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.TextInputFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.ShellProcessManager;
import sbrt.shell.proc.fba.MetatoolInputReader;
import sbrt.shell.proc.fba.MetatoolInputReader.Results;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.chem.IrrevRxnFormatV1;
import sbrt.shell.text.chem.RevRxnFormatV1;
import sbrt.shell.text.math.LinearCombFormatV1;

/**
 * This class is used to manage the conversion of input files for Metatool
 * into SBRT FBA Reaction Files.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class MetatoolInputReaderManager 
    extends AbstractProcessManager 
    implements ShellProcessManager<String, String> {
	
	/**
	 * The set of keywords required by this process manager.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {

		Set<String> s = new TreeSet<String>();

        s.add(MetatoolOutputReaderManager.METATOOL_INPUT_FILE);
        s.add(OUTPUT_FILE_NAME);

		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by this process
	 * manager.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
        
        Set<String> s = new TreeSet<String>();
        
        optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Returns the set of keywords required by this process manager.
	 * 
	 * @return the set of keywords required by this process manager.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
	 * Returns the set of keywords understood, but not required, by this process
	 * manager.
	 * 
	 * @return the set of keywords understood, but not required, by this process
	 *         manager.
	 */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

	private String metatoolInputFile;
	private String outputFile;
    
	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input all of the information required to prepare this process
	 * manager for execution.
	 * 
	 * @param out the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

		String s = MetatoolOutputReaderManager.METATOOL_INPUT_FILE;
		
		metatoolInputFile = input.getRequiredValue(s);
		
		outputFile = input.getRequiredValue(OUTPUT_FILE_NAME);
	}

	private void checkState() {

		if (metatoolInputFile == null) {
			String s = "No input was provided.";
			throw new IllegalStateException(s);
		}
		
	}

	/**
	 * Executes the process this process manager controls and writes informative
	 * messages to the provided print writer.
	 * 
	 * @param out
	 *            the print writer where informative messagse will be written.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager.
	 */
	public void execute(PrintWriter out) {

		checkState();
		
        out.println();
        out.print("Reading '" + metatoolInputFile + "' ... ");
        out.flush();
        
        Results results = parseInputFile();
        
        out.println("DONE");
        out.println();
        out.print("Writing '" + outputFile + "' ... ");
        out.flush();
        
        writeOutputFile(results);
        
        out.println("DONE");
	}

	private Results parseInputFile() {

		InputFile<String> file = TextInputFile.getBasic(metatoolInputFile);
		
		List<String> lines = new ArrayList<String>();
		
		for (String line : file) {
			lines.add(line);
		}
		
		file.close();
		
		Results results = null;
		
		try {
			results = MetatoolInputReader.parse(lines);
		} catch (FormatException e) {
			error(e.getMessage());
		}
		
		return results;
	}
	
	
	private void writeOutputFile(Results results) {

		MapFormat<String, String> mapFormat = EqualsMapFormatV1.getBasic();
		LinearCombFormatV1<String> linCombFormat = LinearCombFormatV1.getBasic();
        
        OutputFile<String> file = TextOutputFile.getBasic(outputFile);
        
        for (Map.Entry<String, ChemRxn> e : results.getCat().entrySet()) {
        	
        	String rxnName = e.getKey();
        	ChemRxn rxn = e.getValue();
        	
        	LinearComb<String> reactants = rxn.getReactants();
        	LinearComb<String> products = rxn.getProducts();
        	
        	String rxnString = null;
        	
        	if (rxn.isReversible()) {
        		rxnString = linCombFormat.format(reactants) +
        					" " + RevRxnFormatV1.ARROW + " " + 
        					linCombFormat.format(products);
        	}
        	
        	else {
        		rxnString = linCombFormat.format(reactants) +
							" " + IrrevRxnFormatV1.ARROW + " " +
							linCombFormat.format(products);
        	}
        	
        	String line = mapFormat.formatPair(rxnName, rxnString);
        	file.addData(line);
        }
        
        file.close();
		
	}
	
	
	
	private void error(String msg) {
		String s = "An error occurred while parsing the Metatool input file '" + 
					metatoolInputFile + "'. " + msg; 
		throw new FormatException(s);
	}
}
