/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.math.algebra.DoubleMatrix;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.InputFile;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.TextInputFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.ShellProcessManager;
import sbrt.shell.proc.fba.MetatoolInputReader;
import sbrt.shell.proc.fba.MetatoolOutputReader;
import sbrt.shell.proc.fba.MetatoolInputReader.Results;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.LinkedHashSetFormat;
import sbrt.shell.text.ListFormat;
import sbrt.shell.text.PipeListFormat;
import sbrt.shell.text.PipeSetFormat;

/**
 * This class is used to manage the writing of input files for Metatool.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class MetatoolOutputReaderManager 
    extends AbstractProcessManager implements ShellProcessManager<String, String> {

	/**
	 * The keyword used to denote a Metatool input file.
	 */
	public static final String METATOOL_INPUT_FILE = "Metatool Input File";

	/**
	 * The keyword used to denote a Metatool output file.
	 */
	public static final String METATOOL_OUTPUT_FILE = "Metatool Output File";

	/**
	 * The keyword used to denote the name of a file to which elementary
	 * modes are written.
	 */
	public static final String ELEM_MODE_FILE_NAME = "Elementary Mode File Name";
	
	/**
	 * The set of keywords required by this process manager.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {

		Set<String> s = new TreeSet<String>();

        s.add(METATOOL_INPUT_FILE);
        s.add(METATOOL_OUTPUT_FILE);

		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by this process
	 * manager.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
        
        Set<String> s = new TreeSet<String>();
        s.add(ELEM_MODE_FILE_NAME);
        optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Returns the set of keywords required by this process manager.
	 * 
	 * @return the set of keywords required by this process manager.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
	 * Returns the set of keywords understood, but not required, by this process
	 * manager.
	 * 
	 * @return the set of keywords understood, but not required, by this process
	 *         manager.
	 */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

	private String metatoolInputFile;
	private String metatoolOutputFile;
    private String elemModeOutputFileName;
    
	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input all of the information required to prepare this process
	 * manager for execution.
	 * 
	 * @param out the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

		metatoolInputFile = input.getRequiredValue(METATOOL_INPUT_FILE);
		metatoolOutputFile = input.getRequiredValue(METATOOL_OUTPUT_FILE);
		
		elemModeOutputFileName = input.getOptionalValue(ELEM_MODE_FILE_NAME);
	}

	private void checkState() {

		if (metatoolInputFile == null) {
			String s = "No input was provided.";
			throw new IllegalStateException(s);
		}
		
	}

	/**
	 * Executes the process this process manager controls and writes informative
	 * messages to the provided print writer.
	 * 
	 * @param out
	 *            the print writer where informative messagse will be written.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager.
	 */
	public void execute(PrintWriter out) {

		checkState();
		
        List<String> rxnNames = parseInputFile(out);
        
        DoubleMatrix<String, String> matrix = 
        	parseOutputFile(rxnNames, out);
        
        writeElemModeFile(matrix, out);

	}


	@SuppressWarnings("null")
	private List<String> parseInputFile(PrintWriter out) {

        out.println();
        out.print("Reading '" + metatoolInputFile + "' ... ");
        out.flush();
        
		// Reading file
		
		InputFile<String> file = TextInputFile.getBasic(metatoolInputFile);
		
		List<String> lines = new ArrayList<String>();
		
		for (String line : file) {
			lines.add(line);
		}
		
		file.close();
		
		// Parsing its contents
		
		Results results = null;
		
		try {
			results = MetatoolInputReader.parse(lines);
		} catch (FormatException e) {
			error(e.getMessage());
		}

		// Combining lists of reaction names
		
		List<String> orderedRxnNames = new ArrayList<String>(); 
		
		orderedRxnNames.addAll(results.getEnzIrrev());
		orderedRxnNames.addAll(results.getEnzRev());

        out.println("DONE");
        
		return orderedRxnNames;
	}
	
	private DoubleMatrix<String, String> 
	parseOutputFile(List<String> rxnNames, PrintWriter out) {


        out.println();
        out.print("Reading '" + metatoolOutputFile + "' ... ");
        out.flush();

        InputFile<String> file1 = TextInputFile.getBasic(metatoolOutputFile);
        
        MetatoolOutputReader outputParser = new MetatoolOutputReader();

        DoubleMatrix<String, String> matrix =
        	outputParser.getElementaryModes(file1, rxnNames);
        
        out.println("DONE");
        
        return matrix;
	}
	
	
	private void writeElemModeFile(DoubleMatrix<String, String> matrix,
								 PrintWriter out) {

		if (elemModeOutputFileName == null) {
			return;
		}
		
        out.println();
        out.print("Writing '" + elemModeOutputFileName + "' ... ");
        out.flush();
        
        OutputFile<String> file2 = TextOutputFile.getBasic(elemModeOutputFileName); 
        
        ListFormat<String> listFormatter = PipeListFormat.getBasic();
        LinkedHashSetFormat<String> setFormatter = PipeSetFormat.getBasic();
        DoubleFormat numFormatter = new DoubleFormatV1();
        
        Set<String> colIndices = matrix.getColumnIndices();
        Set<String> rowIndices = matrix.getRowIndices();
        
        // Writing column headers
        file2.addData(setFormatter.format(colIndices));
        
        // Writing rows
        for (String rowIndex : rowIndices) {
        	
        	List<String> elements = new LinkedList<String>();
        	
        	for (String colIndex : colIndices) {
        		double e = matrix.getElement(rowIndex, colIndex);
        		elements.add(numFormatter.format(e));
        	}
        	
        	file2.addData(listFormatter.format(elements));
        }
        

        
        out.println("DONE");
		
	}
	
	
	private void error(String msg) {
		String s = "An error occurred while parsing the Metatool input file '" + 
					metatoolInputFile + "'. " + msg; 
		throw new FormatException(s);
	}
	
}
