/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.*;
import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.fba.comp.*;
import sbrt.kernel.fba.proc.*;
import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.*;
import sbrt.shell.io.*;
import sbrt.shell.io.fba.FluxCapInputFile;
import sbrt.shell.mng.*;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.fba.FluxCapFormat;
import sbrt.shell.text.fba.FluxCapFormatV1;
import sbrt.shell.text.fba.RxnNameOrExprFormat;
import sbrt.shell.text.fba.RxnNameOrExprFormatV1;

/**
 * This class is used to manage Mahadevan-Schilling flux interval processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class MsFluxIntervalManager extends FbaManager {

    /**
     * The keyword used to denote the flux cap input file.
     */
    public static final String FLUX_CAP_FILE = "Flux Cap File";

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Basic inputs
        s.add(RXN_FILE);
        s.add(ProgramSolverFactory.SOLVER);
        s.add(FbaOptManager.CONSTRAINTS_FILE);
        s.add(FLUX_CAP_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Enables control over numerical error
        s.add(ZERO_CUTOFF);
        s.add(FbaOptManager.SAFETY_LEVEL);
        s.add(CONSTRAINT_TOLERANCE);

        s.add(DATA_HEADERS);

        s.add(ProgramSolverFactory.PARAM_FILE);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Returns the flux caps contained in the file with the name specified by
     * {@linkplain #FLUX_CAP_FILE} in the provided input.
     * 
     * @param input the input containing a value for {@code FLUX_CAP_FILE}.
     * @param out the print writer to which status messages will be printed.
     * @param fluxome the fluxome to which the flux caps correspond.
     * 
     * @return the flux caps contained in the file with the name specified by
     * {@linkplain #FLUX_CAP_FILE} in the provided input.
     */
    public static UnmodifiableMap<String, Set<LinearComb<String>>> 
    getFluxCaps(ManagerInput<String, String> input,
    			PrintWriter out,
    			Fluxome fluxome) {

        // Getting the file name
        String fileName = input.getRequiredValue(FLUX_CAP_FILE);

        // Printing an opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Creating file line format
        FluxCapFormat format = new FluxCapFormatV1(fluxome);

        // Creating the input file
        FluxCapInputFile file = new FluxCapInputFile(fileName, format);

        // Getting the flux caps it contains
        UnmodifiableMap<String, Set<LinearComb<String>>> fluxCaps =
                file.getFluxCaps();

        // Printing a closing message
        out.println("DONE");

        return fluxCaps;
    }

    /**
     * Returns the flux interval computer specified by the provided input. This
     * method calls {@linkplain FbaOptManager#getMixedComp(ManagerInput, PrintWriter)}.
     * 
     * @param input the input containing all required keywords to construct the
     * flux interval computer.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return the flux interval computer specified by the provided input.
     */
    public static MsFluxIntervalComputer<IrrevRxnFluxome> 
    getIrrevComp(ManagerInput<String, String> input,
    			PrintWriter out) {

        // Getting zero cutoff
        double zeroCutoff = getZeroCutoff(input, out);

        // Getting optimization computer
        FbaOptimizer<IrrevRxnFluxome> computer =
                FbaOptManager.getIrrevRxnComp(input, out);

        // Getting fluxome
        IrrevRxnFluxome fluxome = computer.getFluxome();

        // Getting flux caps
        Map<String, Set<LinearComb<String>>> fluxCaps =
                getFluxCaps(input, out, fluxome);

        return new MsFluxIntervalComputer<IrrevRxnFluxome>(computer, 
        												   fluxCaps, 
        												   zeroCutoff);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e a program solver exception.
     * @param process the FBA process from which the exception was thrown.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(FluxRangeException e,
            									 FbaProcess<?, ?> process) {

        // Gathering information about the problem
        Object offender = e.getOffender();
        double rangeCutoff = e.getRangeCutoff();

        // Getting the fluxome
        Fluxome fluxome = process.getFluxome();

        // Creating reaction expression formatter
        RxnNameOrExprFormat rf = new RxnNameOrExprFormatV1(fluxome);

        // Creating double formatter
        DoubleFormat df = new DoubleFormatV1();

        // Creating the formatted error message
        String s =
                "The zero cutoff '"
                        + df.format(rangeCutoff)
                        + "' appears to be too restrictive. "
                        + "Constraining '"
                        + rf.format(offender)
                        + "' results in an infeasible solution. Please "
                        + "specify a larger zero cutoff.";

        return new ApplicationException(s, e);
    }

    private FluxIntervalProcess<IrrevRxnFluxome> process;

    private RomoFile<String, Interval> outputFile;

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object must
     * contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object must
     * contain.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Closing the previous output file if necessary
        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        // Getting the computer
        FluxIntervalComputer<IrrevRxnFluxome> computer = getIrrevComp(input, out);

        // Getting and printing the fluxome
        Fluxome fluxome = computer.getFluxome();
        printFluxomeInfo(fluxome, out);

        // Getting the results manager
        outputFile =
                SimpleFluxIntervalManager.getSingleResultOutputFile(input, out,
                        fluxome);

        // Creating the process
        process = new FluxIntervalProcess<IrrevRxnFluxome>(computer);

        // Setting the reaction names
        process.setReactionNames(outputFile.getFilterKeys());
    }

    private void checkState() {

        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Returns the process this process manager controls.
     * 
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public FluxIntervalProcess<IrrevRxnFluxome> getProcess() {
        checkState();
        return process;
    }

    /**
     * Executes the process this process manager controls, and writes the status
     * or results of this process to the provided print writer.
     * 
     * @param out the print writer where status messages will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {

        checkState();

        try {

            Rm rm = new Rm(outputFile);

            runProcess(process, rm, out);

        } catch (FluxRangeException e) {
            
            throw interpret(e, process);
            
        } catch (IllegalBoundsException e) {

            throw SimpleFluxIntervalManager.interpret(e);

        } catch (RuntimeException e) {

            throw FbaOptManager.interpret(e, process);

        } finally {

            closeFile(outputFile, out);

        }
    }

    private static class Rm extends
            AbstractResultsManager<Map<String, Interval>> {

        private final RomoFile<String, Interval> output;

        private Rm(RomoFile<String, Interval> output) {
            this.output = output;
        }

        /**
         * Adds data to this results manager.
         * 
         * @param results the results, or data, to be added to this manager.
         * 
         * @throws IllegalStateException if {@code close()} has already been
         * called.
         */
        public void addData(Map<String, Interval> results) {
            output.addData(results);
        }
    }

}
