/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.CatalyzedFluxome;
import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.proc.NetworkInfoProcess;
import sbrt.kernel.fba.proc.NetworkInfoProcess.Results;
import sbrt.kernel.io.SingleRm;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.algebra.SparseDoubleMatrix;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.MapManagerInput;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.fba.CatalystVerifier;
import sbrt.shell.text.fba.ChemSpeciesVerifier;
import sbrt.shell.text.fba.RxnNameExprFormatV1;
import sbrt.shell.text.fba.RxnNameSetFormat;
import sbrt.shell.text.fba.RxnNameSetFormatV1;
import sbrt.shell.text.fba.RxnNameVerifier;

/**
 * This class is used to manage fluxome information processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class NetworkInfoManager extends FbaManager {


	/**
	 * The keyword used to denote the name of the file to which
	 * stoichiometrically equivalent reactions will be written.
	 */
	public static final String EQUIV_RXN_FILE_NAME = 
		"Equivalent Reaction File Name";

	/**
	 * The keyword used to denote the name of the file to which the system of
	 * linear equations formed by a fluxome will be written.
	 */
	public static final String EQUATION_FILE_NAME = "Equation File Name";

	/**
	 * The keyword used to denote the name of the file to which the
	 * stoichiometry matrix of a fluxome will be written.
	 */
	public static final String S_MATRIX_FILE_NAME = "Stoichiometry Matrix File Name";

	/**
	 * The keyword used to denote the name of the file to which the chemical species
	 * of a fluxome will be written.
	 */
	public static final String CHEM_SPECIES_FILE_NAME = "Chemical Species File Name";

	/**
	 * The keyword used to denote the name of the file to which the names of
	 * reactions of a fluxome will be written.
	 */
	public static final String RXN_NAME_FILE_NAME = "Reaction Name File Name";

	/**
	 * The keyword used to denote the name of the file to which a fluxome will
	 * be written.
	 */
	public static final String RXN_OUTPUT_FILE_NAME = "Reaction Output File Name";

    /**
     * The keyword used to denote teh name of the file to which all catalysts
     * will be written.
     */
    public static final String CATALYST_FILE_NAME = "Catalyst File Name";
    
	/**
	 * The set of keywords required to be present in the {@code ManagerInput}
	 * objects supplied to instances of this class.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {
		Set<String> s = new TreeSet<String>();
		s.add(RXN_FILE);
		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by instances of this
	 * class.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
		Set<String> s = new TreeSet<String>();
        s.add(RXN_CATALYST_FILE);
        s.add(CATALYST_FILE_NAME);
		s.add(CHEM_SPECIES_FILE_NAME);
		s.add(RXN_NAME_FILE_NAME);
		s.add(EQUIV_RXN_FILE_NAME);
        s.add(EQUATION_FILE_NAME);
		s.add(S_MATRIX_FILE_NAME);
		optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Prints basic information about the provided fluxome.
	 * 
	 * @param fluxome
	 *            the fluxome for which basic information will be printed.
	 * @param out
	 *            the print writer to which basic information will be printed.
	 */
	public static void printBasicInfo(MixedFluxome fluxome, PrintWriter out) {

		out.println();
		out.println("Number of chemical species: " + fluxome.species());
		out.println("Number of chemical reactions: " + fluxome.rxns());
		out.println("Number of internal reactions: " + fluxome.internalRxns());
		out.println("Number of exchange reactions: " + fluxome.exchangeRxns());
		
		// Printing number of non-zero matrix elements
		out.print("Number of non-zero matrix elements: ");
		out.flush();
		
		int nonZeros = 0;
		
		for (ChemRxn rxn : fluxome.getRxns().values()) {
			nonZeros += rxn.getReactants().terms();
			nonZeros += rxn.getProducts().terms();
		}
		
		out.println(nonZeros);

		// Printing number of user-defined flux constraints
		out.println("Number of user-defined flux constraints: "
				+ fluxome.addedConstraints());

		out.println("Number of catalysts: " + fluxome.catalysts());

        
        out.println();

	}

	private NetworkInfoProcess<MixedFluxome> process;

	private MapManagerInput<String, String> input;

	/**
	 * Returns the set of keywords the provided {@code ManagerInput} object must
	 * contain.
	 * 
	 * @return the set of keywords the provided {@code ManagerInput} object must
	 *         contain.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
	 * Returns the set of keywords the provided {@code ManagerInput} object does
	 * not necessarily have to contain.
	 * 
	 * @return the set of keywords the provided {@code ManagerInput} object does
	 *         not necessarily have to contain.
	 */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

	/**
	 * Returns the process this process manager controls.
	 * 
	 * @return the process this process manager controls.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager via
	 *             {@linkplain #setInput(ManagerInput)} or
	 *             {@linkplain #setInput(ManagerInput, PrintWriter)}
	 */
	public NetworkInfoProcess<MixedFluxome> getProcess() {
		checkState();
		return process;
	}

	private void checkState() {
		if (process == null) {
			String s = "No manager input was provided.";
			throw new IllegalStateException(s);
		}
	}

	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input
	 *            all of the information required to prepare this process
	 *            manager for execution.
	 * 
	 * @param out
	 *            the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

		// Getting fluxome
		MixedFluxome fluxome = FbaManager.getMixedFluxome(input, out);

		// Creating process
		process = new NetworkInfoProcess<MixedFluxome>(fluxome);

		// Making a defensive copy
		this.input = new MapManagerInput<String, String>(input.toMap());

		// Printing basic information
		printBasicInfo(fluxome, out);
	}

	/**
	 * Executes the process this process manager controls, and writes the status
	 * or results of this process to the provided print writer.
	 * 
	 * @param out
	 *            the print writer where status messages will be printed.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager via
	 *             {@linkplain #setInput(ManagerInput)} or
	 *             {@linkplain #setInput(ManagerInput, PrintWriter)}
	 */
	public void execute(PrintWriter out) {

		checkState();

		// Creating simple results manager
		SingleRm<Results<MixedFluxome>> rm = 
			new SingleRm<Results<MixedFluxome>>();

		// Adding results to this manager
		process.run(rm);

		// Retreiving the results
		Results<MixedFluxome> results = rm.getResults();

		// Writing optional files
		writeOptionalFiles(results, out);

	}

	private void writeOptionalFiles(Results<MixedFluxome> results, 
									PrintWriter out) {

		writeRxnNameFile(results, out);
		writeChemSpeciesFile(results, out);
		writeCatalystFile(results, out);
		writeStoichMatrixFile(results, out);
		writeEquivalentRxnFile(results, out);
		writeEquationFile(results, out);

	}


	private void 
	writeRxnNameFile(Results<MixedFluxome> results, 
				     PrintWriter out) {

		// Getting the filename
		String fileName = input.getOptionalValue(RXN_NAME_FILE_NAME);

		// If no filename was provided, do nothing
		if (fileName == null) {
			return;
		}

		// Printing an opening message
		out.print("Writing '" + fileName + "' ... ");
		out.flush();

		// Getting the fluxome
		Fluxome fluxome = process.getFluxome();

		// Creating file line format
		RxnNameVerifier formatter = new RxnNameVerifier(fluxome);

		// Creating output file
		OutputFile<String> file = new TextOutputFile<String>(fileName,
				formatter);

		// Writing each reaction name
		for (String rxnName : results.getRxnNames()) {
			file.addData(rxnName);
		}

		// Closing the file
		file.close();

		// Printing a closing message
		out.println("DONE");
	}

	private void 
	writeChemSpeciesFile(Results<MixedFluxome> results, 
						 PrintWriter out) {

		// Getting the filename
		String fileName = input.getOptionalValue(CHEM_SPECIES_FILE_NAME);

		// If no filename was provided, do nothing
		if (fileName == null) {
			return;
		}

		// Printing an opening message
		out.print("Writing '" + fileName + "' ... ");
		out.flush();

		// Getting the fluxome
		Fluxome fluxome = process.getFluxome();

		// Creating a name checker (file line format)
		ChemSpeciesVerifier formatter = new ChemSpeciesVerifier(fluxome);

		// Creating the file
		OutputFile<String> file = 
			new TextOutputFile<String>(fileName, formatter);

		// Writing each species to the file
		for (String species : results.getChemSpecies()) {
			file.addData(species);
		}

		// Closing the file
		file.close();

		// Printing a closing messsage
		out.println("DONE");
	}

	private void 
	writeCatalystFile(Results<MixedFluxome> results, 
					  PrintWriter out) {

		// Getting the filename
		String fileName = input.getOptionalValue(CATALYST_FILE_NAME);

		// If no filename was provided, do nothing
		if (fileName == null) {
			return;
		}

		// Printing an opening message
		out.print("Writing '" + fileName + "' ... ");
		out.flush();

		// Getting the fluxome
        CatalyzedFluxome fluxome = process.getFluxome();

		// Creating an catalyst checker (file line format)
		CatalystVerifier formatter = new CatalystVerifier(fluxome);

		// Creating the file
		OutputFile<String> file = 
            new TextOutputFile<String>(fileName, formatter);

		// Writing each catalyst to the file
		for (String catalyst : results.getCatalysts()) {
			file.addData(catalyst);
		}

		// Closing the file
		file.close();

		// Printing a closing message
		out.println("DONE");
	}

	private void 
	writeEquationFile(Results<MixedFluxome> results, 
					  PrintWriter out) {

		// Getting the filename
		String fileName = input.getOptionalValue(EQUATION_FILE_NAME);

		// If no name was provided, do nothing
		if (fileName == null) {
			return;
		}

		// Printing an opening message
		out.print("Writing '" + fileName + "' ... ");
		out.flush();

		// Getting the fluxome
		Fluxome fluxome = process.getFluxome();

		// Creating rxn name expression formatter
		RxnNameExprFormatV1 formatter = new RxnNameExprFormatV1(fluxome);

		// Creating the file
		OutputFile<String> output = TextOutputFile.getBasic(fileName);

		// Getting the matrix as a list of linear combinations
		List<LinearComb<String>> exprs = 
			results.getStoichMatrix().toLinearCombs();

		// Writing each linear combination to the file
		for (LinearComb<String> e : exprs) {

			// Creating formatted line of text
			String s = formatter.format(e) + " = 0.0";

			// Writing the line
			output.addData(s);
		}

		// Closing the file
		output.close();

		// Printing a closing message
		out.println("DONE");
	}

	private void 
	writeStoichMatrixFile(Results<MixedFluxome> results, 
						  PrintWriter out) {

		// Getting the filename
		String fileName = input.getOptionalValue(S_MATRIX_FILE_NAME);

		// If no name was provided, do nothing
		if (fileName == null) {
			return;
		}

		// Printing an opening message
		out.print("Writing '" + fileName + "' ... ");
		out.flush();

		// Getting the fluxome
		Fluxome fluxome = process.getFluxome();

		// Creating the reaction name formatter
		RxnNameVerifier rxnNameFormat = new RxnNameVerifier(fluxome);

		// Creating the species name formatter
		ChemSpeciesVerifier speciesFormat = new ChemSpeciesVerifier(fluxome);

		// Creating the double formatter
		DoubleFormat numSimpleFormat = new DoubleFormatV1();

		// Creating the file
		OutputFile<String> output = TextOutputFile.getBasic(fileName);

		// Getting the stoichiometry matrix
		SparseDoubleMatrix<String, String> matrix = results.getStoichMatrix();

		// Getting all row and column variables
		Set<String> rows = matrix.getRowIndices();
		Set<String> cols = matrix.getColumnIndices();

		// Creating string builder
		StringBuilder s = new StringBuilder(fluxome.rxns() * 20);

		// Appending each column variable to the string builder
		for (String rxn : cols) {
			s.append(" | ");
			s.append(rxnNameFormat.format(rxn));
		}

		// Writing the line
		output.addData(s.toString());

		// Iterating thru each row
		for (String species : rows) {

			// Deleting the old characters
			s.delete(0, s.length());

			// Appending the formatted species
			s.append(speciesFormat.format(species));

			// Iterating thru each row element
			for (String rxn : cols) {

				// Getting the value
				double v = matrix.getElement(species, rxn);

				// Appending the formatted value
				s.append(" | ");
				s.append(numSimpleFormat.format(v));
			}

			// Writing the line
			output.addData(s.toString());
		}

		// Closing the file
		output.close();

		// Printing a closing message
		out.println("DONE");
	}

	private void 
	writeEquivalentRxnFile(Results<MixedFluxome> results, 
						   PrintWriter out) {

		// Getting the filename
		String fileName = input.getOptionalValue(EQUIV_RXN_FILE_NAME);

		// If no filename was provided, do nothing
		if (fileName == null) {
			return;
		}

		// Printing an opening message
		out.print("Writing '" + fileName + "' ... ");
		out.flush();

		// Getting the fluxome
		Fluxome fluxome = process.getFluxome();

		// Creating the reaction name formatter
		RxnNameSetFormat rxnNameFormat = new RxnNameSetFormatV1(fluxome);

		// Creating the file
		OutputFile<Set<String>> outFile = new TextOutputFile<Set<String>>(
				fileName, rxnNameFormat);

		// Writing each set of equivalent reactions
		for (Set<String> equivs : results.getEquivRxns()) {
			outFile.addData(equivs);
		}

		// Closing the file
		outFile.close();

		// Printing a closing message
		out.println("DONE");
	}

}
