/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.util.*;
import java.io.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.fba.comp.*;
import sbrt.kernel.fba.proc.*;
import sbrt.kernel.math.*;
import sbrt.kernel.math.optimization.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.ProgramSolverFactory;
import sbrt.shell.io.*;
import sbrt.shell.io.fba.FbaOptOutputFile;
import sbrt.shell.mng.*;
import sbrt.shell.text.fba.*;

/**
 * This class is used to manage objective function processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class ObjectiveFunctionManager extends FbaManager {

	/**
	 * The keyword used to denote the file containing objective functions.
	 */
    public static final String OBJ_FUNCT_FILE =
            "Objective Function File";

    /**
     * The keyword used to denote the format of the file containing objective
     * functions.
     */
    public static final String OBJ_FUNCT_FILE_FORMAT =
            "Objective Function File Format";

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Basic inputs
        s.add(RXN_FILE);
        s.add(ProgramSolverFactory.SOLVER);
        s.add(FbaOptManager.OPTIMIZATION_SENSE);
        s.add(FbaOptManager.CONSTRAINTS_FILE);

        // process input file
        s.add(OBJ_FUNCT_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME);
        s.add(DATA_HEADERS);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
        Set<String> s = new TreeSet<String>();
        
        s.add(FbaOptManager.SAFETY_LEVEL);
        s.add(CONSTRAINT_TOLERANCE);
        s.add(OBJ_FUNCT_FILE_FORMAT);
        s.add(ProgramSolverFactory.PARAM_FILE);
        
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Returns the file denoted by {@linkplain #OBJ_FUNCT_FILE} in the 
     * provided manager input. The format of this file can be specified by
     * the additional keyword {@linkplain #OBJ_FUNCT_FILE_FORMAT}. The values
     * corresponding to these keywords are parsed and used as arguments to the 
     * method {@linkplain #getInputFile(String, FileFormat, Fluxome)}.
     * 
     * @param input the input containing a value for {@code OBJ_FUNCT_FILE},
     * and possibly even {@code OBJ_FUNCT_FILE_FORMAT}.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the objective functions correspond.
     * 
     * @return the file denoted by {@linkplain #OBJ_FUNCT_FILE} in the 
     * provided manager input. 
     * 
     * @see #getFileFormat(String)
     */
    public static TextInputFile<LinearComb<String>> 
    getInputFile(ManagerInput<String, String> input,
                 PrintWriter out,
                 Fluxome fluxome) {

        // Getting the file name
        String fileName = input.getRequiredValue(OBJ_FUNCT_FILE);

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting provided file format
        String fileFormat = input.getOptionalValue(OBJ_FUNCT_FILE_FORMAT);
        
        // Parsing the file format
        FileFormat f = getFileFormat(fileFormat); 
        
        // Getting the input file
        TextInputFile<LinearComb<String>> objFile = 
        	getInputFile(fileName, f, fluxome);

        // Printing a closing message
        out.println("DONE");

        // Returning the file
        return objFile;
    }

    /**
	 * Returns the input file with the provided name. This file must have been
	 * written in either {@linkplain FileFormat#TEXT} or
	 * {@linkplain FileFormat#GZIP_TEXT} formats, and the non-commented lines
	 * must be parsable with the default reaction name expression parser. Every
	 * variable contained in each objective function must be the name of a
	 * reaction in the provided fluxome.
	 * 
	 * @param fileName
	 *            the name of the file to be opened.
	 * @param format
	 *            the format of the file to be opened. {@code FileFormat.TEXT}
	 *            and {@code FileFormat.GZIP_TEXT} are currently the only
	 *            supported choices.
	 * @param fluxome
	 *            the fluxome to which the objective functions correspond.
	 * 
	 * @return the input file with the provided name.
	 * 
	 * @see FbaOptManager#getRxnNameExprFormat(Fluxome)
	 */
    public static TextInputFile<LinearComb<String>> 
    getInputFile(String fileName, 
    			 FileFormat format, 
    			 Fluxome fluxome) {

        // Getting a reaction name expression format
        RxnNameExprFormat<LinearComb<String>, String> lineFormat = 
        	getRxnNameExprFormat(fluxome);
        
        // Returning input file
        return new TextInputFile<LinearComb<String>>(fileName, 
        											 format, 
        											 lineFormat);
    }

    /**
     * Returns the input file with the provided
     * name. This file must have been written in {@linkplain FileFormat#TEXT}
     * format, and the non-commented lines
     * must be parsable with the default reaction name expression parser. Every
     * variable contained in each objective function must be the name of a
     * reaction in the provided fluxome.
     * 
     * @param fileName the name of the file to be opened.
     * @param fluxome the fluxome to which the objective functions correspond.
     * 
     * @return the input file with the provided name.
     *
     * @see FbaOptManager#getRxnNameExprFormat(Fluxome)
     */
    public static TextInputFile<LinearComb<String>> 
    getInputFile(String fileName, Fluxome fluxome) {
    	return getInputFile(fileName, FileFormat.TEXT, fluxome);
    }

    private ObjFunAnalysis<MixedFluxome> process;
    private InputFile<LinearComb<String>> objectiveFile;
    private FbaOptOutputFile<MixedFluxome> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    private void closeFiles(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        if (objectiveFile != null) {
            closeFile(objectiveFile, out);
        }

    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Closing the previous files if necessary
        closeFiles(out);

        // Getting the computer
        FbaOptimizer<MixedFluxome> computer = 
        	FbaOptManager.getMixedComp(input, out);

        // Getting and printing the fluxome
        MixedFluxome fluxome = computer.getFluxome();
        printFluxomeInfo(fluxome, out);

        // Creating the process
        process = new ObjFunAnalysis<MixedFluxome>(computer);

        // Getting and setting the optimization sense
        OptSense sense = 
        	FbaOptManager.getOptSense(input);
        process.setOptimizationSense(sense);

        // Getting and setting the objective function source
        objectiveFile = getInputFile(input, out, fluxome);
        process.setObjectiveFunctionSource(objectiveFile);

        // Getting the output file
        outputFile = 
        	FbaOptManager.getMultiResultsOutputFile(input, out, fluxome);
    }

    private void checkState() {

        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public ObjFunAnalysis<MixedFluxome> getProcess() {
        checkState();
        return process;
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();

        try {
        	
            runProcess(process, outputFile, out);
            
        } catch (ProgramSolverException e) {
        	
        	throw FbaOptManager.interpret(e);
        	
        } catch (FbaOptException e) {
        	
            throw FbaOptManager.interpret(e, process);
            
        } catch (ConstraintViolationException e) {
        	
            throw FbaOptManager.interpret(e, process);
            
        } finally {
            closeFiles(out);
        }
    }

}
