/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.proc.RandConstGeneratorV1;
import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.math.Interval;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.io.ComoFile;
import sbrt.shell.io.FileFormat;
import sbrt.shell.io.MapOutputFile;
import sbrt.shell.io.MapTextOutputFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.fba.ConstraintsFileLineFormat;
import sbrt.shell.text.fba.ConstraintsFileLineFormatV1;
import sbrt.shell.text.math.IntervalFormatV1;

/**
 * This class is used to manage random constraints generators.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class RandomConstraintsManager extends FbaManager {

	/**
	 * The keyword used to denote the constraint for reversible reactions.
	 */
    public static final String REV_CONSTRAINT = "Reversible Constraint";

	/**
	 * The keyword used to denote the constraint for irreversible reactions.
	 */
    public static final String IRREV_CONSTRAINT = "Irreversible Constraint";
    
    /**
     * The keyword used to denote the inactive constraint.
     */
    public static final String INACTIVE = "Inactive Constraint";

    /**
     * The keyword used to denote the number of active fluxes.
     */
    public static final String NUM_ACTIVE_FLUXES = "Number of Active Fluxes";

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
    	
    	// Creating storage
        Set<String> s = new TreeSet<String>();

        // Basic FBA inputs
        s.add(RXN_FILE);

        // process specific inputs
        s.add(ITERATIONS);
        s.add(SEED);
        s.add(NUM_ACTIVE_FLUXES);
        s.add(REV_CONSTRAINT);
        s.add(IRREV_CONSTRAINT);
        s.add(INACTIVE);

        // Output file information
        s.add(DATA_HEADERS);
        s.add(OUTPUT_FILE_NAME);

        // Saving as unmodifiable
        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
    	
    	// Getting empty set
    	Set<String> empty = Collections.emptySet();
    	
        // Saving as unmodifiable
        optionalKeywords = new UnmodifiableSet<String>(empty);
    }

    /**
     * Parses and returns the value for {@linkplain #IRREV_CONSTRAINT} in the 
     * provided manager input. The string must be parsable using
     * {@linkplain IntervalFormatV1}.
     * 
     * @param input keyword-value pairs.
     * @return the parsed value for {@linkplain #IRREV_CONSTRAINT} in the
     * provided input.
     */
    public static Interval getIrrevConstraint(ManagerInput<String, String> input) {

    	// Getting the provided value
        String on = input.getRequiredValue(IRREV_CONSTRAINT);

        // Creating the parser
        IntervalFormatV1 format = new IntervalFormatV1();

        // Returning the parsed value
        return format.parse(on);
    }

    /**
     * Parses and returns the value for {@linkplain #REV_CONSTRAINT} in the 
     * provided manager input. The string must be parsable using
     * {@linkplain IntervalFormatV1}.
     * 
     * @param input keyword-value pairs.
     * @return the parsed value for {@linkplain #REV_CONSTRAINT} in the
     * provided input.
     */
    public static Interval getRevConstraint(ManagerInput<String, String> input) {

    	// Getting the provided value
        String on = input.getRequiredValue(REV_CONSTRAINT);

        // Creating the parser
        IntervalFormatV1 format = new IntervalFormatV1();

        // Returning the parsed value
        return format.parse(on);
    }

    /**
     * Parses and returns the value for {@linkplain #INACTIVE} in the 
     * provided manager input. The string must be parsable using
     * {@linkplain IntervalFormatV1}.
     * 
     * @param input keyword-value pairs.
     * @return the parsed value for {@linkplain #INACTIVE} in the
     * provided input.
     */
    public static Interval 
    getInactive(ManagerInput<String, String> input) {

    	// Getting the provided value
        String off = input.getRequiredValue(INACTIVE);

        // Creating the parser
        IntervalFormatV1 format = new IntervalFormatV1();

        // Returning the parsed value
        return format.parse(off);
    }

    /**
     * Parses and returns the value for {@linkplain #NUM_ACTIVE_FLUXES} in the 
     * provided manager input.
     * 
     * @param input keyword-value pairs.
     * @param out the print writer to which a status message will be printed.
     * 
     * @return the parsed value for {@linkplain #NUM_ACTIVE_FLUXES} in the
     * provided input.
     */
    public static int 
    getNumActiveFluxes(ManagerInput<String, String> input, PrintWriter out) {
    	
    	// Getting string
        String n = input.getRequiredValue(NUM_ACTIVE_FLUXES);
        
        // Returing parsed value
        int i = parseInt(n);
        
        if (i < 0) {
        	String s = "The value '" + n + "' provided for the keyword '" +
        			   NUM_ACTIVE_FLUXES + 
                       "' in '" +
                       input.getName() +
                       "' is invalid.";
        	throw new ApplicationException(s);
        }
        
        return i;
    }

	/**
	 * Returns an empty output file that is suitable for storing multiple
	 * sets of flux constraints. The values for
	 * {@linkplain #OUTPUT_FILE_NAME}, 
	 * {@linkplain #DATA_HEADERS}, and
	 * {@linkplain #FILE_FORMAT} in the provided input are parsed
	 * and used to call the method 
	 * {@linkplain #getOutputFile(String, FileFormat, Set, IrrevRxnFluxome)}.
	 * 
     * @param input the input containing values for the listed keywords.
     * @param out the print writer to which status messages will be printed.
	 * @param fluxome the fluxome to which the constraints correspond.
	 * 
	 * @return an empty output file that is suitable for storing multiple
	 * sets of flux constraints.
	 */
    public static MapTextOutputFile<String, Interval>
            getOutputFile(ManagerInput<String, String> input,
                          PrintWriter out,
                          Fluxome fluxome) {

    	// Getting output filename
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting reaction names
        Set<String> headers = getRxnNameHeaders(input, fluxome);

        // Getting file format
        FileFormat f = getTextFileFormat(input);
        
        // Creating output file
        MapTextOutputFile<String, Interval> output =
                getOutputFile(fileName, f, headers, fluxome);

        // Printing a closing message
        out.println("DONE");

        // Returning the output file
        return output;
    }

	/**
	 * Returns an empty output file that is suitable for storing multiple
	 * sets of flux constraints. The format used to write data to this file
	 * is {@linkplain ConstraintsFileLineFormatV1}.
	 * 
	 * @param fileName the name of the file to be created.
	 * @param format the format of the file to be created.
	 * {@code FileFormat.TEXT} and {@code FileFormat.GZIP_TEXT} are currently 
     * the only supported choices.
	 * @param headers the set of objects whose constraints will be written to 
	 * the file.
	 * @param fluxome the fluxome to which the constraints correspond.
	 * 
	 * @return an empty output file that is suitable for storing multiple
	 * sets of flux constraints.
	 */
    public static MapTextOutputFile<String, Interval>
            getOutputFile(String fileName, 
            			  FileFormat format,
            			  Set<String> headers,
                          Fluxome fluxome) {

        // Creating constraints file line format
        ConstraintsFileLineFormat<String> f =
                new ConstraintsFileLineFormatV1(fluxome);

        // Creating output file
        ComoFile<String, Interval> output =
        	new ComoFile<String, Interval>(fileName, format, headers, f);

        // Returning the output file
        return output;
    }

	/**
	 * Returns an empty output file that is suitable for storing multiple
	 * sets of flux constraints. The format used to write data to this text file
	 * is {@linkplain ConstraintsFileLineFormatV1}.
	 * 
	 * @param fileName the name of the file to be created.
	 * @param headers the set of objects whose constraints will be written to 
	 * the file.
	 * @param fluxome the fluxome to which the constraints correspond.
	 * 
	 * @return an empty output file that is suitable for storing multiple
	 * sets of flux constraints.
	 */
    public static MapTextOutputFile<String, Interval>
    getOutputFile(String fileName,
    			  Set<String> headers,
                  IrrevRxnFluxome fluxome) {
    	
    	return getOutputFile(fileName, FileFormat.TEXT, headers, fluxome);
    }
    
    private RandConstGeneratorV1<MixedFluxome> generator;
    private MapOutputFile<String, Interval> outputFile;

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing previous file if necessary
        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        // Getting fluxome
        MixedFluxome fluxome = getMixedFluxome(input, out);

        // Creating generator
        generator = new RandConstGeneratorV1<MixedFluxome>(fluxome);

        // Getting and setting iterations
        int iterations = getIterations(input, out);
        generator.setIterations(iterations);

        // Getting and setting ON constraint
        Interval irrev = getIrrevConstraint(input);
        generator.setIrrevConstraint(irrev);

        // Getting and setting ON constraint
        Interval rev = getRevConstraint(input);
        generator.setRevConstraint(rev);

        // Getting and setting ON constraint
        Interval off = getInactive(input);
        generator.setInactive(off);

        // Getting and setting number of ON fluxes
        int activeFluxes = getNumActiveFluxes(input, out);
        generator.setActiveFluxes(activeFluxes);

        // Getting and setting random number generator
        Random rand = new Random(getSeed(input, out));
        generator.setRandom(rand);

        // Printing basic fluxome information
        printFluxomeInfo(fluxome, out);

        // Getting and setting results manager
        outputFile = getOutputFile(input, out, fluxome);
        
        // Getting reaction names
        Set<String> rxnNames = outputFile.getFilterKeys();
        
        if (rxnNames.size() < activeFluxes) {
        	
        	String s = "The specified number of active fluxes (" +
        				activeFluxes + ") exceeds the number of " +
        			   "reactions specified as data headers (" +
        			   rxnNames.size() + ").";

        	throw new ApplicationException(s);
        }
        
        // Getting and setting reaction names from output file
        generator.setReactions(rxnNames);

    }

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public RandConstGeneratorV1<MixedFluxome> getProcess() {
        checkState();
        return generator;
    }

    private void checkState() {

        if (generator == null) {
            String s = "Input was never set.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();

        try {
        	
        	Rm rm = new Rm(outputFile);
        	
            runProcess(generator, rm, out);
            
        } finally {
            closeFile(outputFile, out);
        }

    }

    
    private static class Rm extends 
    AbstractResultsManager<Map<String, Interval>> {
    	
    	private final MapOutputFile<String, Interval> output;
    	
    	private Rm(MapOutputFile<String, Interval> output) {
    		this.output = output;
    	}

    	/**
    	 * Adds data to this results manager.
    	 * 
    	 * @param results the results, or data, to be added
    	 * to this manager.
    	 * 
    	 * @throws IllegalStateException if {@code close()}
    	 * has already been called.
    	 */
    	public void addData(Map<String, Interval> results) {
    		output.addData(results);
    	}
    }
    

}
