/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.*;
import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.shell.*;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.fba.*;

/**
 * This abstract class provides a foundation for random flux vector
 * managers. It contains a keyword as well as a set of utility functions.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public abstract class RandomFluxVectorManager extends FbaManager {

	/**
	 * The keyword used to denote a random flux vector algorithm.
	 */
    public static final String ALGORITHM = "Random Flux Vector Algorithm";

    /**
     * Opens the file denoted by {@linkplain #FLUX_VECTOR_FILE} in the 
     * provided manager input and returns the flux vector it contains.
     * 
     * @param input the input containing a value for {@code FLUX_VECTOR_FILE}.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the flux vector corresponds.
     * @param tolerance the amount by which fluxes in the vector are allowed to
     * violate their defined constraints.
     * 
     * @return the flux vector contained in the file indicated by 
     * {@code #FLUX_VECTOR_FILE} in the provided input.
     * 
     * @see #getFluxVector(String, Fluxome, double)
     */
    public static Map<String, Double> 
    getFluxVector(ManagerInput<String, String> input,
    				PrintWriter out,
    				Fluxome fluxome,
    				double tolerance) {

        // Getting the file name
        String fileName = input.getOptionalValue(FLUX_VECTOR_FILE);

        // Printing an opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Getting the flux vector
        Map<String, Double> fluxVector = 
        	getFluxVector(fileName, fluxome, tolerance);

        // Printing a closing message
        out.println("DONE");

        // Returning flux vector
        return fluxVector;
    }

    /**
     * Returns the flux vector contained in the file with the provided name.
     * This file must have the format of a row-oriented map file, the lines must
     * be parsable using {@linkplain FluxVectorFormatV1}, and a flux must be present
     * for every reaction in the provided fluxome. The flux vector must also lie
     * within the flux space defined by the provided fluxome.
     * 
     * @param fileName the name of the file to be read.
     * @param fluxome the fluxome to which the fluxes correspond.
     * @param tolerance the amount by which fluxes in the vector are allowed to
     * violate their defined constraints.
     * 
     * @return the flux vector contained in the file with the provided name.
     * 
     * @see RomiFile
     * @see Fluxome#isFluxVector(Map, double)
     * 
     * @throws ApplicationException if the flux vector in the file does not lie
     * within the flux space defined by the provided fluxome.
     */
    public static Map<String, Double> 
    getFluxVector(String fileName, Fluxome fluxome, double tolerance) {

        // Getting flux vector file line format
        FluxVectorFormat format = new FluxVectorFormatV1(fluxome);

        // Getting all required variables
        Set<String> variables = fluxome.getRxnNames();

        // Creating row-oriented vector input file
        RomiFile<String, Double> inFile =
                new RomiFile<String, Double>(fileName,
                                             FileFormat.TEXT,
                                             variables, 
                                             format);

        // Getting the flux vector contained in the file
        Map<String, Double> fluxVector = inFile.getMap();

        // Closing the input file
        inFile.close();

        // Ensuring the flux vector is within the defined flux space
        if (!fluxome.isFluxVector(fluxVector, tolerance)) {
            String s = "The flux vector in the file '" + fileName +
                       "' is not a valid point in flux space " + 
            		   "when " + tolerance + " is used as the " +
            		   "constraint tolerance.";
            throw new ApplicationException(s);
        }

        // Returning the vector
        return fluxVector;
    }

    /**
     * Opens the file denoted by {@linkplain FbaOptManager#CONSTRAINTS_FILE} in the 
     * provided manager input and returns the set of flux constraints 
     * it contains.
     * 
     * @param input keyword-value pairs.
     * @param out the print writer to which a status message will be printed.
     * @param fluxome the fluxome to which the constraints correspond.
     * @param tolerance the constraint tolerance.
     * 
     * @return the set of flux constraints contained in the file indicated by 
     * {@code #CONSTRAINTS_FILE} in the provided input.
     * 
     * @see #getConstraints(String, Fluxome, double)
     */
    public static FluxConstraints 
    getConstraints(ManagerInput<String, String> input,
                   PrintWriter out,
                   Fluxome fluxome,
                   double tolerance) {

        // Getting constraints file name
        String keyword = FbaOptManager.CONSTRAINTS_FILE;
        String fileName = input.getRequiredValue(keyword);

        // Printing opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Getting flux constraints
        FluxConstraints constraints = 
        	getConstraints(fileName, fluxome, tolerance);

        // Printing closing message
        out.println("DONE");

        // Returning all flux constraints contained in the file
        return constraints;
    }

    /**
     * Returns the set of constraints contained in the file with the provided
     * name. This file must have the format of a row-oriented map file, the
     * lines must be parsable using {@linkplain ConstraintsFileLineFormatV1}, and
     * a constraint must be present for every reaction in the provided fluxome.
     * 
     * @param fileName the name of the file to be read.
     * @param fluxome the fluxome to which the constraints correspond.
     * @param tolerance the constraint tolerance.
     * 
     * @return the set of constraints contained in the file with the provided
     * name. 
     *
     * @throws NullPointerException if either argument is {@code null}.
     *
     * @see #getConstraints(ManagerInput, PrintWriter, Fluxome, double)
     * @see FbaOptManager#getConstraints(String, Fluxome, double)
     * @see RomiFile
     */
    public static FluxConstraints 
    getConstraints(String fileName, Fluxome fluxome, double tolerance) {

        // Getting flux constraints
        FluxConstraints c = 
        	FbaOptManager.getConstraints(fileName, fluxome, tolerance);
        
        // Getting provided reaction names
        Set<String> rxnNames = c.getFluxConstraints().keySet();
        
        // Ensuring a constraint was provided for each reaction name
        for (String name : fluxome.getRxnNames()) {
        	
        	if (!rxnNames.contains(name)) {
        		String s = "A flux constraint was not provided for '" +
        		name + "' in the file '" + fileName + "'.";
        		throw new ApplicationException(s);
        	}
        	
        }

        return c;
    }

	/**
	 * Creates and returns a fluxome based on the provided input.
	 * The values for {@linkplain #RXN_FILE}, 
	 * {@linkplain FbaOptManager#CONSTRAINTS_FILE}, and
	 * {@linkplain #CONSTRAINT_TOLERANCE}
	 * are parsed and combined to create the returned fluxome.
	 * 
     * @param input the input containing values for the listed keywords.
     * @param out the print writer to which status messages will be printed.
	 * 
	 * @return a fluxome based on the provided input.
	 * 
	 * @see FbaManager#getIrrevRxnFluxome(ManagerInput, PrintWriter)
	 * @see FbaManager#getConstraintTolerance(ManagerInput, PrintWriter)
	 * @see #getConstraints(ManagerInput, PrintWriter, Fluxome, double)
	 * @see Fluxome#replaceConstraints(FluxConstraints, double)
	 */
    public static IrrevRxnFluxome 
    getIrrevRxnFluxome(ManagerInput<String, String> input, PrintWriter out) {

        // Getting basic fluxome (i.e. one with default flux constraints)
    	IrrevRxnFluxome fluxome = FbaManager.getIrrevRxnFluxome(input, out);

        // Getting constraint tolerance from the provided fluxome
        double tolerance = getConstraintTolerance(input, out);

        // Getting constraints for *all* fluxes
        FluxConstraints constraints = getConstraints(input, out, fluxome, tolerance);

        // Returning new fluxome
        return fluxome.replaceConstraints(constraints, tolerance);
    }


}
