/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.proc.RandomObjectiveGenerator;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.optimization.ProgramSolverException;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.FileFormat;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.fba.RxnNameExprFormat;

/**
 * This class is used to manage random objective function generators.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class RandomObjectiveManager extends FbaManager {

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Basic FBA input
        s.add(RXN_FILE);

        // process specific inputs
        s.add(ITERATIONS);
        s.add(SEED);

        // Output file information
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
    	
    	Set<String> s = new TreeSet<String>();
    	
    	s.add(FILE_FORMAT);
    	
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Returns an empty output file that is suitable for storing randomly 
     * generated objective functions. This method uses the default format
     * for mathematical expressions of reaction names.
     * 
	 * @param fileName the name of the file to be created.
	 * @param format the format of the file to be created.
     * @param out the print writer to which status messages will be printed.
     * @param fluxome the fluxome to which the linear combinations of
     * reaction names correspond.
     * 
     * @return an empty output file that is suitable for storing randomly 
     * generated objective functions.
     * 
     * @see #getRxnNameExprFormat(Fluxome)
     */
    public static TextOutputFile<LinearComb<String>> 
    getOutputFile(String fileName,
    			  FileFormat format,
    			  PrintWriter out,
    			  Fluxome fluxome) {

        // Printing opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting file line format
        RxnNameExprFormat<LinearComb<String>, String> f =
        	getRxnNameExprFormat(fluxome);
        
        // Creating output file
        TextOutputFile<LinearComb<String>> output =
                new TextOutputFile<LinearComb<String>>(fileName, format, f);

        // Printing closing message
        out.println("DONE");

        // Returning file
        return output;
    }

	/**
	 * Returns an empty output file that is suitable for storing randomly
	 * generated objective functions. The values for {@linkplain #OUTPUT_FILE_NAME}
	 * and {@linkplain #FILE_FORMAT} in the provided input are
	 * parsed and used to call the method
	 * {@linkplain #getOutputFile(String, FileFormat, PrintWriter, IrrevRxnFluxome)}.
	 * 
	 * @param input the input containing values for the listed keywords.
	 * @param out the print writer to which status messages will be printed.
	 * @param fluxome the fluxome for which the data headers correspond.
	 * 
	 * @return an empty output file that is suitable for storing randomly
	 * generated objective functions.
	 * 
	 * @see #getFileFormat(ManagerInput)
	 */
    public static TextOutputFile<LinearComb<String>> 
    getOutputFile(ManagerInput<String, String> input, 
    			  PrintWriter out,
    			  Fluxome fluxome) {

        // Getting filename
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Getting file format
        FileFormat f = getTextFileFormat(input);
        
        // Returning file
        return getOutputFile(fileName, f, out, fluxome);
    }

    private RandomObjectiveGenerator<MixedFluxome> generator;
    private OutputFile<LinearComb<String>> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing previous file if necessary
        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        // Getting and printing fluxome
        MixedFluxome fluxome = getMixedFluxome(input, out);
        printFluxomeInfo(fluxome, out);

        // Creating generator
        generator = new RandomObjectiveGenerator<MixedFluxome>(fluxome);

        // Getting and setting iterations
        int iterations = getIterations(input, out);
        generator.setIterations(iterations);

        // Getting and setting random number generator
        Random rand = new Random(getSeed(input, out));
        generator.setRandom(rand);

        // Getting and setting results manager
        out.println();
        outputFile = getOutputFile(input, out, fluxome);
    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public RandomObjectiveGenerator<MixedFluxome> getProcess() {

        // Ensuring input was provided
        checkState();
        return generator;
    }

    private void checkState() {

        // Checking generator
        if (generator == null) {
            String s = "Input was never set.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {

        // Ensuring input was provided
        checkState();

        try {
        	
            // Running process
            runProcess(generator, outputFile, out);
            
        } catch (ProgramSolverException e) {
        	throw FbaOptManager.interpret(e);
        } finally {
            closeFile(outputFile, out);
        }

    }


}
