/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.proc.RedundRxnRemoverProcess;
import sbrt.kernel.io.SingleRm;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.RomoFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.fba.MixedFluxomeFormatV1;

/**
 * This class is used to manage redundant reaction remover processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class RedundRxnRemoverManager extends FbaManager {

    
	/**
	 * The set of keywords required to be present in the {@code ManagerInput}
	 * objects supplied to instances of this class.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {
		Set<String> s = new TreeSet<String>();
		s.add(RXN_FILE);
		s.add(OUTPUT_FILE_NAME);
		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by instances of this
	 * class.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
		Set<String> s = new TreeSet<String>();
		optionalKeywords = new UnmodifiableSet<String>(s);
	}

	private RedundRxnRemoverProcess process;

	private String outputName;

	/**
	 * Returns the set of keywords the provided {@code ManagerInput} object must
	 * contain.
	 * 
	 * @return the set of keywords the provided {@code ManagerInput} object must
	 *         contain.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
	 * Returns the set of keywords the provided {@code ManagerInput} object does
	 * not necessarily have to contain.
	 * 
	 * @return the set of keywords the provided {@code ManagerInput} object does
	 *         not necessarily have to contain.
	 */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

	/**
	 * Returns the process this process manager controls.
	 * 
	 * @return the process this process manager controls.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager via
	 *             {@linkplain #setInput(ManagerInput)} or
	 *             {@linkplain #setInput(ManagerInput, PrintWriter)}
	 */
	public RedundRxnRemoverProcess getProcess() {
		checkState();
		return process;
	}

	private void checkState() {
		if (process == null) {
			String s = "No manager input was provided.";
			throw new IllegalStateException(s);
		}
	}

	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input
	 *            all of the information required to prepare this process
	 *            manager for execution.
	 * 
	 * @param out
	 *            the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

		// Getting fluxome
		MixedFluxome fluxome = FbaManager.getMixedFluxome(input, out);

		// Making a defensive copy
		outputName = input.getRequiredValue(OUTPUT_FILE_NAME);

		// Creating process
		process = new RedundRxnRemoverProcess(fluxome);

		// Printing basic information
		printFluxomeInfo(fluxome, out);
	}

	/**
	 * Executes the process this process manager controls, and writes the status
	 * or results of this process to the provided print writer.
	 * 
	 * @param out
	 *            the print writer where status messages will be printed.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager via
	 *             {@linkplain #setInput(ManagerInput)} or
	 *             {@linkplain #setInput(ManagerInput, PrintWriter)}
	 */
	public void execute(PrintWriter out) {

		checkState();

		// Printing an opening message
		out.print("Writing '" + outputName + "' ... ");
		out.flush();

		// Creating simple results manager
		SingleRm<Fluxome> rm = new SingleRm<Fluxome>();

		// Adding results to this manager
		process.run(rm);

		// Retreiving the results
		Fluxome fluxome = rm.getResults();

		// Creating file line format
        MixedFluxomeFormatV1 format = new MixedFluxomeFormatV1();

		// Creating output file
        RomoFile<String, ChemRxn> file = 
            new RomoFile<String, ChemRxn>(outputName, format);

		// Writing the reactions
		file.addData(fluxome.getRxns());

		// Closing the file
		file.close();

		// Printing a closing message
		out.println("DONE");
	}


}
