/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.proc.RevRxnBreakerProcess;
import sbrt.kernel.io.SingleRm;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.MapOutputFile;
import sbrt.shell.io.RomoFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.fba.IrfFormatV1;

/**
 * This class is used to manage reversible reaction breakers. 
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class RevRxnBreakerManager extends FbaManager {

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        s.add(RXN_FILE);
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
    	Set<String> s = Collections.emptySet(); 
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    private static MapOutputFile<String, ChemRxn>
    getOutputFile(ManagerInput<String, String> input,
    			  PrintWriter out) {

    	// Getting file name
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Creating map format
        IrfFormatV1 format = new IrfFormatV1();
        
        // Opening output file
        RomoFile<String, ChemRxn> file = 
        	new RomoFile<String, ChemRxn>(fileName, format);

        // Printing closing message
        out.println("DONE");

        return file;
    }

    private RevRxnBreakerProcess process;
    private MapOutputFile<String, ChemRxn> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public RevRxnBreakerProcess getProcess() {
        checkState();
        return process;
    }

    private void closeFile(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }
    }

    private void checkState() {
        if (process == null) {
            throw new IllegalStateException("No manager input was provided.");
        }
    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing previous file if necessary
        closeFile(out);

        // Getting fluxome
        MixedFluxome fluxome = getMixedFluxome(input, out);

        // Getting fluxome and printing basic information
        printFluxomeInfo(fluxome, out);

        // Getting all reactions it contains
        Map<String, ChemRxn> rxns = fluxome.getRxns();
        
        // Creating reaction breaker
        process = new RevRxnBreakerProcess();
        process.setRxns(rxns);

        // Setting output file
        outputFile = getOutputFile(input, out);
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();

        try {
        	
        	out.println();
        	out.print("Breaking reversible reactions ...");
        	out.flush();
        	
            // Creating simple results manager
            SingleRm<Map<String, ChemRxn>> rm = 
            	new SingleRm<Map<String, ChemRxn>>();

            process.run(rm);

            // Writing results to the file
            outputFile.addData(rm.getResults());

            out.println("DONE");
            out.println();
            
        } finally {
            closeFile(out);
        }
    }


}
