/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.*;
import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.fba.comp.*;
import sbrt.kernel.fba.proc.*;
import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.*;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;

/**
 * This class is used to manage constrained-reverse reaction flux interval processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class ReverseRxnFluxIntervalManager extends FbaManager {

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Basic inputs
        s.add(RXN_FILE);
        s.add(ProgramSolverFactory.SOLVER);
        s.add(FbaOptManager.CONSTRAINTS_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Enables control over numerical error
        s.add(ZERO_CUTOFF);
        s.add(FbaOptManager.SAFETY_LEVEL);
        s.add(CONSTRAINT_TOLERANCE);

        s.add(DATA_HEADERS);

        s.add(ProgramSolverFactory.PARAM_FILE);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Returns the flux interval computer specified by the provided input. This
     * method calls {@linkplain FbaOptManager#getMixedComp(ManagerInput, PrintWriter)}.
     * 
     * @param input the input containing all required keywords to construct the
     * flux interval computer.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return the flux interval computer specified by the provided input.
     */
    public static ReverseRxnIntervalComp<IrrevRxnFluxome> 
    getComputer(ManagerInput<String, String> input,
    			PrintWriter out) {

        // Getting zero cutoff
        double zeroCutoff = getZeroCutoff(input, out);
        
        // Getting optimization computer
        FbaOptimizer<IrrevRxnFluxome> computer =
                FbaOptManager.getIrrevRxnComp(input, out);

        return new ReverseRxnIntervalComp<IrrevRxnFluxome>(computer, zeroCutoff);
    }

    private FluxIntervalProcess<IrrevRxnFluxome> process;

    private RomoFile<String, Interval> outputFile;

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object must
     * contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object must
     * contain.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Closing the previous output file if necessary
        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        // Getting the computer
        FluxIntervalComputer<IrrevRxnFluxome> computer = getComputer(input, out);

        // Getting and printing the fluxome
        IrrevRxnFluxome fluxome = computer.getFluxome();
        printFluxomeInfo(fluxome, out);

        // Getting the results manager
        outputFile = SimpleFluxIntervalManager.getSingleResultOutputFile(input, out, fluxome);

        // Creating the process
        process = new FluxIntervalProcess<IrrevRxnFluxome>(computer);

        // Setting the reaction names
        process.setReactionNames(outputFile.getFilterKeys());
    }

    private void checkState() {

        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Returns the process this process manager controls.
     * 
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public FluxIntervalProcess<IrrevRxnFluxome> getProcess() {
        checkState();
        return process;
    }

    /**
     * Executes the process this process manager controls, and writes the status
     * or results of this process to the provided print writer.
     * 
     * @param out the print writer where status messages will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {

        checkState();

        try {

            Rm rm = new Rm(outputFile);

            runProcess(process, rm, out);

        } catch (IllegalBoundsException e) {

            throw SimpleFluxIntervalManager.interpret(e);

        } catch (RuntimeException e) {

            throw FbaOptManager.interpret(e, process);

        } finally {

            closeFile(outputFile, out);

        }
    }

    private static class Rm extends
            AbstractResultsManager<Map<String, Interval>> {

        private final RomoFile<String, Interval> output;

        private Rm(RomoFile<String, Interval> output) {
            this.output = output;
        }

        /**
         * Adds data to this results manager.
         * 
         * @param results the results, or data, to be added to this manager.
         * 
         * @throws IllegalStateException if {@code close()} has already been
         * called.
         */
        public void addData(Map<String, Interval> results) {
            output.addData(results);
        }
    }

}
