/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.comp.FbaOptException;
import sbrt.kernel.fba.comp.FbaOptimizer;
import sbrt.kernel.fba.comp.RxnDeletionComputer;
import sbrt.kernel.fba.proc.RxnDeletionProcess;
import sbrt.kernel.math.ConstraintViolationException;
import sbrt.kernel.math.MathExpr;
import sbrt.kernel.math.optimization.OptSense;
import sbrt.kernel.math.optimization.ProgramSolverException;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ProgramSolverFactory;
import sbrt.shell.io.InputFile;
import sbrt.shell.io.TextInputFile;
import sbrt.shell.io.fba.FbaOptOutputFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.fba.RxnNameSetFormat;
import sbrt.shell.text.fba.RxnNameSetFormatV1;

/**
 * This class is used to manage FBA reaction deletion processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class RxnDeletionManager extends FbaManager {

	/**
	 * The keyword used to denote the name of a file containing sets of
	 * names of reactions whose effect upon deletion will be computed.
	 */
    public static final String RXN_DELETION_FILE = "Reaction Deletion File";

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
    	
        Set<String> s = new TreeSet<String>();

        // Basic Fba Optimization Inputs
        s.add(RXN_FILE);
        s.add(ProgramSolverFactory.SOLVER);
        s.add(FbaOptManager.OBJECTIVE_FUNCTION);
        s.add(FbaOptManager.OPTIMIZATION_SENSE);
        s.add(FbaOptManager.CONSTRAINTS_FILE);

        // process input file
        s.add(RXN_DELETION_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME);
        s.add(DATA_HEADERS);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /** 
     * The set of optional keywords in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
    	
        Set<String> s = new TreeSet<String>();

        s.add(CONSTRAINT_TOLERANCE);
        s.add(ZERO_CUTOFF);
        s.add(FbaOptManager.SAFETY_LEVEL);
        s.add(ProgramSolverFactory.PARAM_FILE);
        
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
	 * Parses and returns the value for {@linkplain #RXN_DELETION_FILE} in the
	 * provided manager input. The non-commented lines of this file must be
	 * parsable with {@linkplain RxnNameSetFormatV1}.
	 * 
	 * @param input keyword-value pairs.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fluxome
	 *            the fluxome in which the reactions are contained.
	 * 
	 * @return the reaction name input file with the name contained in
	 *         {@code input}.
	 */
    public static InputFile<Set<String>>
    getInputFile(ManagerInput<String, String> input,
    			 PrintWriter out,
    			 Fluxome fluxome) {

        // Getting the name of the reaction name file
        String fileName = input.getRequiredValue(RXN_DELETION_FILE);

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting the input file
        InputFile<Set<String>> inputFile = getInputFile(fileName, fluxome);

        // Printing a closing message
        out.println("DONE");

        // Returning the file
        return inputFile;
    }

    /**
     * Returns the reaction name input file with the provided name. The
     * non-commented lines of this file must be parsable with 
     * {@linkplain RxnNameSetFormatV1}.
     * 
     * @param fileName the name of the file containing sets of reaction names.
     * @param fluxome the fluxome in which the reactions are contained.
     * 
     * @return the reaction name input file with the provided name.
     */
    public static InputFile<Set<String>> 
    getInputFile(String fileName, Fluxome fluxome) {

        // Getting a reaction name file line format
        RxnNameSetFormat rxnFormat = new RxnNameSetFormatV1(fluxome);

        // Returning the input file
        return new TextInputFile<Set<String>>(fileName, rxnFormat);
    }
    
    /**
     * Returns a reaction deletion computer created from the provided input.
     *   
     * @param input the manager input from which the necessary information is
     * obtained.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return a reaction deletion computer created from the provided input.
     * 
     * @see #getMixedDelComp(ManagerInput, PrintWriter)
     * @see #getZeroCutoff(ManagerInput, PrintWriter)
     */
    public static RxnDeletionComputer<MixedFluxome>
    getMixedDelComp(ManagerInput<String, String> input, PrintWriter out) {

        double zeroCutoff = getZeroCutoff(input, out);
        
        // Getting optimization computer
        FbaOptimizer<MixedFluxome> computer = 
        	FbaOptManager.getMixedComp(input, out);
        
        RxnDeletionComputer<MixedFluxome> rdc = 
        	new RxnDeletionComputer<MixedFluxome>(computer);

        rdc.setZeroCutoff(zeroCutoff);
        
        return rdc;
    }

    private RxnDeletionProcess<MixedFluxome> process;
    private InputFile<Set<String>> deletionFile;
    private FbaOptOutputFile<MixedFluxome> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    private void closeFiles(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        if (deletionFile != null) {
            closeFile(deletionFile, out);
        }
    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Closing the previous files if necessary
        closeFiles(out);

        // Getting the computer
        RxnDeletionComputer<MixedFluxome> computer = 
        	getMixedDelComp(input, out);

        // Getting and printing the fluxome
        MixedFluxome fluxome = computer.getFluxome();
        printFluxomeInfo(fluxome, out);

        // Getting the objective function
        MathExpr<String> objective = 
        	FbaOptManager.getObjFunc(input, out, fluxome);

        // Getting the optimization sense (i.e. direction)
        OptSense sense = FbaOptManager.getOptSense(input);

        // Getting the reaction name input file
        deletionFile = getInputFile(input, out, fluxome);

        // Getting the column oriented output file
        outputFile = 
        	FbaOptManager.getMultiResultsOutputFile(input, out, fluxome);

        // Creating the process
        process = new RxnDeletionProcess<MixedFluxome>(computer);

        process.setObjectiveFunction(objective);
        process.setOptimizationSense(sense);
        process.setRxnDeletionSource(deletionFile);
    }

    private void checkState() {

        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public RxnDeletionProcess<MixedFluxome> getProcess() {
        checkState();
        return process;
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();

        try {
        	
            runProcess(process, outputFile, out);
            
        } catch (FbaOptException e) {
        	
            throw FbaOptManager.interpret(e, process);
            
        } catch (ProgramSolverException e) {
        	
        	throw FbaOptManager.interpret(e);
        	
        } catch (ConstraintViolationException e) {
        	
            throw FbaOptManager.interpret(e, process);
            
        } finally {
            closeFiles(out);
        }

    }

}
