/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.*;
import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.fba.comp.*;
import sbrt.kernel.fba.proc.*;
import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.*;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.fba.*;

/**
 * This class is used to manage simple flux interval processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class SimpleFluxIntervalManager extends FbaManager {

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Basic inputs
        s.add(RXN_FILE);
        s.add(ProgramSolverFactory.SOLVER);
        s.add(FbaOptManager.CONSTRAINTS_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        // Enables control over numerical error
        s.add(ZERO_CUTOFF);
        s.add(FbaOptManager.SAFETY_LEVEL);
        s.add(CONSTRAINT_TOLERANCE);

        s.add(DATA_HEADERS);

        s.add(ProgramSolverFactory.PARAM_FILE);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
	 * Returns an empty output file that is suitable for storing a single set of
	 * flux interval results. The values for {@linkplain #OUTPUT_FILE_NAME} and
	 * {@linkplain #DATA_HEADERS} in the provided input
	 * are parsed and used to call the method
	 * {@linkplain #getMultiResultsOutputFile(String, Set, Fluxome)}.
	 * 
	 * @param input
	 *            the input containing values for the listed keywords.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fluxome
	 *            the fluxome for which the data headers correspond.
	 * 
	 * @return an empty output file that is suitable for storing a single set of
	 *         flux interval results.
	 * 
	 * @see #getRxnNameHeaders(ManagerInput, Fluxome)
	 */
    public static RomoFile<String, Interval> 
    getSingleResultOutputFile(ManagerInput<String, String> input,
    						  PrintWriter out,
    						  Fluxome fluxome) {

        // Getting the file name
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Parsing the headers
        Set<String> headers = getRxnNameHeaders(input, fluxome);

        // Creating the file
        RomoFile<String, Interval> output =
                getSingleResultOutputFile(fileName, headers, fluxome);

        // Printing a closing message
        out.println("DONE");

        return output;
    }

    /**
	 * Returns an empty output file that is suitable for storing a single set of
	 * flux interval results. Data will be written to this file using
	 * {@linkplain ConstraintsFileLineFormatV1}.
	 * 
	 * @param fileName
	 *            the name of the file to be created.
	 * @param headers
	 *            the set of objects whose intervals are to be written to the
	 *            file. These objects can be reaction names or mathematical
	 *            expressions of reaction names.
	 * @param fluxome
	 *            the fluxome to which the intervals correspond.
	 * 
	 * @return an empty output file that is suitable for storing a single set of
	 *         flux interval results.
	 */
    public static RomoFile<String, Interval> 
    getSingleResultOutputFile(String fileName,
    						  Set<String> headers,
    						  Fluxome fluxome) {

        // Creating the file line format
        ConstraintsFileLineFormat<String> format =
                new ConstraintsFileLineFormatV1(fluxome);

        // Creating the output file
        RomoFile<String, Interval> output =
                new RomoFile<String, Interval>(fileName, 
                							   FileFormat.TEXT,
                							   headers, 
                							   format);
        
        return output;
    }

    /**
	 * Returns an empty output file that is suitable for storing multiple sets
	 * of flux interval results. The values for {@linkplain #OUTPUT_FILE_NAME}
	 * and {@linkplain #DATA_HEADERS} in the provided input are parsed and used
	 * to call the method
	 * {@linkplain #getMultiResultsOutputFile(String, Set, Fluxome)}.
	 * 
	 * @param input
	 *            the input containing values for the listed keywords.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * @param fluxome
	 *            the fluxome for which the data headers correspond.
	 * 
	 * @return an empty output file that is suitable for storing multiple sets
	 *         of flux interval results.
	 * 
	 * @see #getRxnNameHeaders(ManagerInput, Fluxome)
	 */
    public static ComoFile<String, Interval> 
    getMultiResultsOutputFile(ManagerInput<String, String> input,
    						  PrintWriter out,
    						  Fluxome fluxome) {

        // Getting the file name
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Parsing the headers
        Set<String> headers = getRxnNameHeaders(input, fluxome);

        // Creating the file
        ComoFile<String, Interval> output =
                getMultiResultsOutputFile(fileName, headers, fluxome);

        // Printing a closing message
        out.println("DONE");

        return output;
    }

    /**
	 * Returns an empty output file that is suitable for storing multiple sets
	 * of flux interval results. Data will be written to this file using
	 * {@linkplain ConstraintsFileLineFormatV1}.
	 * 
	 * @param fileName
	 *            the name of the file to be created.
	 * @param headers
	 *            the set of objects whose intervals are to be written to the
	 *            file. These objects can be reaction names or mathematical
	 *            expressions of reaction names.
	 * @param fluxome
	 *            the fluxome to which the intervals correspond.
	 * 
	 * @return an empty output file that is suitable for storing multiple sets
	 *         of flux interval results.
	 */
    public static ComoFile<String, Interval> 
    getMultiResultsOutputFile(String fileName,
    						  Set<String> headers,
    						  Fluxome fluxome) {

        // Creating formatter
        ConstraintsFileLineFormat<String> format =
                new ConstraintsFileLineFormatV1(fluxome);

        // Creating output file
        ComoFile<String, Interval> output =
                new ComoFile<String, Interval>(fileName, FileFormat.TEXT,
                        headers, format);
        return output;
    }

    /**
	 * Returns the flux interval computer specified by the provided input. This
	 * method calls
	 * {@linkplain FbaOptManager#getMixedComp(ManagerInput, PrintWriter)}.
	 * 
	 * @param input
	 *            the input containing all required keywords to construct the
	 *            flux interval computer.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * 
	 * @return the flux interval computer specified by the provided input.
	 */
    public static SimpleFluxIntervalComputer<MixedFluxome> 
    getMixedComp(ManagerInput<String, String> input,
    			 PrintWriter out) {

        // Getting zero cutoff
        double zeroCutoff = getZeroCutoff(input, out);
        
        // Getting optimization computer
        FbaOptimizer<MixedFluxome> computer =
                FbaOptManager.getMixedComp(input, out);

        return new SimpleFluxIntervalComputer<MixedFluxome>(computer, zeroCutoff);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e a program solver exception.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(IllegalBoundsException e) {

        String rxn = (String) e.getVariable();
        Number min = e.getLower();
        Number max = e.getUpper();

        String s =
                "The program solver computed an invalid interval for '"
                        + rxn
                        + "'. The lower bound '"
                        + min
                        + "' is greater than the upper bound '"
                        + max
                        + "' by more than the specified zero cutoff.";

        return new ApplicationException(s, e);
    }

    private FluxIntervalProcess<MixedFluxome> process;

    private RomoFile<String, Interval> outputFile;

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object must
     * contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object must
     * contain.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking the provided input
        check(input);

        // Closing the previous output file if necessary
        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        // Getting the computer
        FluxIntervalComputer<MixedFluxome> computer = getMixedComp(input, out);

        // Getting and printing the fluxome
        Fluxome fluxome = computer.getFluxome();
        printFluxomeInfo(fluxome, out);

        // Getting the results manager
        outputFile = getSingleResultOutputFile(input, out, fluxome);

        // Creating the process
        process = new FluxIntervalProcess<MixedFluxome>(computer);

        // Setting the reaction names
        process.setReactionNames(outputFile.getFilterKeys());
    }

    private void checkState() {

        if (process == null) {
            String s = "No manager input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Returns the process this process manager controls.
     * 
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public FluxIntervalProcess<MixedFluxome> getProcess() {
        checkState();
        return process;
    }

    /**
     * Executes the process this process manager controls, and writes the status
     * or results of this process to the provided print writer.
     * 
     * @param out the print writer where status messages will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {

        checkState();

        try {

            Rm rm = new Rm(outputFile);

            runProcess(process, rm, out);

        } catch (IllegalBoundsException e) {

            throw interpret(e);

        } catch (RuntimeException e) {

            throw FbaOptManager.interpret(e, process);

        } finally {

            closeFile(outputFile, out);

        }
    }

    private static class Rm extends
            AbstractResultsManager<Map<String, Interval>> {

        private final RomoFile<String, Interval> output;

        private Rm(RomoFile<String, Interval> output) {
            this.output = output;
        }

        /**
         * Adds data to this results manager.
         * 
         * @param results the results, or data, to be added to this manager.
         * 
         * @throws IllegalStateException if {@code close()} has already been
         * called.
         */
        public void addData(Map<String, Interval> results) {
            output.addData(results);
        }
    }

}
