/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.ShellProcessManager;
import sbrt.shell.proc.fba.SimpleRxnFileReader;

/**
 * This class is used to manage the translation of simple reaction files
 * into Systems Biology Research Tool (SBRT) FBA Reaction Files.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class SimpleRxnFileTransManager extends AbstractProcessManager
        implements ShellProcessManager<String, String> {

    /**
     * The keyword used to denote the name of the input reaction file.
     */
    public static final String INPUT_RXN_FILE = "Input Reaction File";

    /**
     * The set of keywords required by this process manager.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {

        Set<String> s = new TreeSet<String>();

        s.add(INPUT_RXN_FILE);
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by this process
     * manager.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {

        Set<String> s = Collections.emptySet();
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Returns the set of keywords required by this process manager.
     * 
     * @return the set of keywords required by this process manager.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords understood, but not required, by this process
     * manager.
     * 
     * @return the set of keywords understood, but not required, by this process
     * manager.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    private String inputFileName;
    private String outputFileName;

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Getting input file name
        inputFileName = input.getRequiredValue(INPUT_RXN_FILE);

        // Getting output file name
        outputFileName = input.getRequiredValue(OUTPUT_FILE_NAME);

    }

    private void checkState() {

        if (inputFileName == null) {
            String s = "No input was provided.";
            throw new IllegalStateException(s);
        }
    }

    /**
     * Executes the process this process manager controls and writes informative
     * messages to the provided print writer.
     * 
     * @param out the print writer where informative messagse will be written.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager.
     */
    public void execute(PrintWriter out) {

        checkState();

        // Printing introduction
        out.print("Translating '"
                + inputFileName
                + "' into '"
                + outputFileName
                + "' ... ");
        out.flush();

        // Creating translater
        SimpleRxnFileReader translater = new SimpleRxnFileReader();

        // Running process
        translater.translate(inputFileName, outputFileName);

        // Printing conclusion
        out.println("DONE");

    }

}
