/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;

import java.io.PrintWriter;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.comp.ExtremeCurrentComputer;
import sbrt.kernel.fba.comp.SlpCycleComputer;
import sbrt.kernel.fba.proc.CycleIdProcess;
import sbrt.kernel.math.optimization.ProgramSolverException;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.ComoFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.fba.ExtremeCurrentManager.Rm;

/**
 * This class is used to manage steady state cycle identification.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class SlpCycleIdManager extends FbaManager {

    /**
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        s.add(RXN_FILE);
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {

        Set<String> s = new TreeSet<String>();

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    private CycleIdProcess<IrrevRxnFluxome> process;
    private IrrevRxnFluxome fluxome;
    private ComoFile<String, Double> outputFile;

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object must
     * contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object must
     * contain.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     * 
     * @return the set of keywords the provided {@code ManagerInput} object does
     * not necessarily have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * 
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public CycleIdProcess<IrrevRxnFluxome> getProcess() {
        checkState();
        return process;
    }

    private void checkState() {
        if (process == null) {
            throw new IllegalStateException("No manager input was provided.");
        }
    }

    private ComoFile<String, Double> 
    	getOutputFile(ManagerInput<String, String> input, 
    				  PrintWriter out, 
    				  IrrevRxnFluxome fluxome) {
    	
    	// The column variables of the extreme current matrix
    	// will necessarily be the names of all internal reactions.
    	Set<String> rxnNames = fluxome.getInternalRxnNames();
    	
    	return ExtremeCurrentManager.getOutputFile(input, out, fluxome, rxnNames);
    }
    
    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing the previous output file if necessary
        if (outputFile != null) {
            closeFile(outputFile, out);
        }

        // Getting fluxome
        fluxome = getIrrevRxnFluxome(input, out);

        // Getting extreme pathway computer
        ExtremeCurrentComputer exPathComp =
                ExtremeCurrentManager.getComputer(input, out);

        // Creating cycle computer
        SlpCycleComputer computer = new SlpCycleComputer(exPathComp);

        // Creating cycle process
        process = new CycleIdProcess<IrrevRxnFluxome>(fluxome, computer);

        // Getting fluxome and printing basic information
        printFluxomeInfo(fluxome, out);

        // Setting output file
        outputFile = getOutputFile(input, out, fluxome);
    }

    /**
     * Executes the process this process manager controls, and writes the status
     * or results of this process to the provided print writer.
     * 
     * @param out the print writer where status messages will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {

        checkState();

        try {

            // Using the ExtremeCurrentManager's result manager
            Rm rm = new Rm(outputFile);

            runProcess(process, rm, out);
            
        } catch (ProgramSolverException e) {
            throw FbaOptManager.interpret(e);
        } finally {
        	closeFile(outputFile, out);
        }
    }

}
