/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.fba;


import java.io.PrintWriter;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.comp.WwFluxomeReducer;
import sbrt.kernel.fba.proc.FluxomeReducerProcess;
import sbrt.kernel.io.SingleRm;
import sbrt.kernel.math.optimization.ProgramSolver;
import sbrt.kernel.math.optimization.ProgramSolverException;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ProgramSolverFactory;
import sbrt.shell.io.RomoFile;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.fba.IrfFormatV1;

/**
 * This class is used to manage the WW fluxome reduction algorithm.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class WwFluxomeReducerManager extends FbaManager {

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String> s = new TreeSet<String>();

        s.add(RXN_FILE);
        s.add(ProgramSolverFactory.SOLVER);

        // Output file information
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
    	
        Set<String> s = new TreeSet<String>();
        
        s.add(ProgramSolverFactory.PARAM_FILE);
        
        optionalKeywords = new UnmodifiableSet<String>(s);
    }
    
    /**
     * Returns a file to which the chemical reactions of a fluxome can be 
     * written.
     * @param fileName the name of the file to be opened.
     * @return a file to which the chemical reactions of a fluxome can be 
     * written.
     */
    public static RomoFile<String, ChemRxn> getOutputFile(String fileName) {
        
        // Creating formatter
        IrfFormatV1 format = new IrfFormatV1();
        
        // Creating the file
        RomoFile<String, ChemRxn> file = 
            new RomoFile<String, ChemRxn>(fileName, format);
    	
    	return file;
    }
    
    /**
     * Returns a file to which the chemical reactions of a fluxome can be 
     * written.
     * @param fileName the name of the file to be opened.
     * @param out the print writer to which descriptive messages will be written.
     * @return a file to which the chemical reactions of a fluxome can be 
     * written.
     */
    public static RomoFile<String, ChemRxn> 
    	getOutputFile(String fileName, PrintWriter out) {
    
    	out.print("Opening '" + fileName + "' ... ");
    	out.flush();
    	
    	RomoFile<String, ChemRxn> file = getOutputFile(fileName);
    	
    	out.println("DONE");
    	
    	return file;
    }
    
    
    private FluxomeReducerProcess<IrrevRxnFluxome> process;
    private RomoFile<String, ChemRxn> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public FluxomeReducerProcess<IrrevRxnFluxome> getProcess() {
        checkState();
        return process;
    }

    private void checkState() {
        if (process == null) {
            throw new IllegalStateException("No manager input was provided.");
        }
    }

    /**
     * Returns a WW fluxome reducer created from the provided input.
     * 
     * @param input keyword-value pairs.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return a WW fluxome reducer.
     * 
     * @see ProgramSolverFactory#make(ManagerInput, PrintWriter)
     */
    private WwFluxomeReducer getReducer(ManagerInput<String, String> input,
    									PrintWriter out) {
 
        // Making the program solver
        ProgramSolver<Object, Object> ps = ProgramSolverFactory.make(input, out);

        // Creating fluxome reducer
        WwFluxomeReducer ww = new WwFluxomeReducer(ps);
        
        return ww;
    }

    private void closeFiles(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }
        
    }
    
    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing the previous files if necessary
        closeFiles(out);

        IrrevRxnFluxome fluxome = getIrrevRxnFluxome(input, out);

        WwFluxomeReducer reducer = getReducer(input, out);
    	
        process = new FluxomeReducerProcess<IrrevRxnFluxome>(fluxome, reducer);

        printFluxomeInfo(fluxome, out);

        // Creating the output file
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);
        
        outputFile = getOutputFile(fileName, out);
    }

    /**
     * Executes the process this process manager controls, and writes the status
     * or results of this process to the provided print writer.
     * 
     * @param out the print writer where status messages will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager via {@linkplain #setInput(ManagerInput)} or
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();

        try {
        	
        	SingleRm<IrrevRxnFluxome> rm = new SingleRm<IrrevRxnFluxome>();
        	
            runProcess(process, rm, out);
         
            outputFile.addData(rm.getResults().getRxns());
            
        } catch (ProgramSolverException e) {
        	throw FbaOptManager.interpret(e);
        } finally {
        	closeFiles(out);
        }
    }

}
