/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.geom;

import java.io.PrintWriter;
import java.util.*;

import sbrt.kernel.math.geom.comp.CdHarComp;
import sbrt.kernel.io.*;
import sbrt.kernel.math.ExceededLimitException;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.UnboundedVariableException;
import sbrt.kernel.math.geom.ConvexPolytope;
import sbrt.kernel.math.geom.HashConvexPolytope;
import sbrt.kernel.math.geom.ParallelHyperplanes;
import sbrt.kernel.math.geom.proc.CdHarProcess;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.*;
import sbrt.shell.text.math.IntervalFormat;
import sbrt.shell.text.math.IntervalFormatV1;
import sbrt.shell.text.math.LinearCombFormat;
import sbrt.shell.text.math.LinearCombFormatV1;

/**
 * This class is used to manage coordinate direction hit-and-run algorithms.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class CdHarManager extends AbstractProcessManager implements KernelProcessManager<String, String> {

    /**
     * The default constraint tolerance.
     */
    public static final double DEFAULT_CONSTRAINT_TOLERANCE = 1E-9;

    /**
     * The default constraint tolerance.
     */
    public static final double DEFAULT_MIN_CHORD_LENGTH = 1E-6;

    /**
     * The keyword used to denote the name of the file containing the collection
     * of sets.
     */
    public static final String PLANES_FILE = "Parallel Hyperplanes File";

    /**
     * The keyword used to denote the constraint tolerance.
     */
    public static final String CONSTRAINT_TOLERANCE = "Constraint Tolerance";

    /**
     * The keyword used to denote the initial point file.
     */
    public static final String INITIAL_POINT_FILE = "Initial Point File";

    /**
     * The keyword used to denote the minimum chord length.
     */
    public static final String MIN_CHORD_LENGTH = "Minimum Chord Length";

    /**
     * The keyword used to denote the number of points generated for each point
     * returned.
     */
    public static final String POINTS_PER_POINT = "Points Per Point";

    /**
     * The keyword used to denote the maximum number of obstructed moves to
     * allow.
     */
    public static final String MAX_OBSTRUCTED_MOVES =
            "Maximum Obstructed Moves";

    /**
     * The set of keywords required by this process manager.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {

        Set<String> s = new TreeSet<String>();

        // Files
        s.add(PLANES_FILE);
        s.add(OUTPUT_FILE_NAME);
        s.add(INITIAL_POINT_FILE);

        // Parameters
        s.add(ITERATIONS);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by this process
     * manager.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {

        Set<String> s = new TreeSet<String>();

        // Parameters
        s.add(CONSTRAINT_TOLERANCE);
        s.add(POINTS_PER_POINT);
        s.add(MIN_CHORD_LENGTH);
        s.add(SEED);
        s.add(MAX_OBSTRUCTED_MOVES);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e an unbounded variable exception.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(UnboundedVariableException e) {

        String s = "The value of '" + e.getVariable() + "' is unbounded.";

        return new ApplicationException(s, e);
    }

    /**
     * Returns an {@code ApplicationException} whose message describes the error
     * that ocurred.
     * 
     * @param e an exceeded limit exception.
     * 
     * @return an {@code ApplicationException} wrapping the provided exception.
     */
    public static ApplicationException interpret(ExceededLimitException e) {

        String s =
                "The maximum number of obstructed moves has been reached. "
                        + "This can happen when the random walker gets stuck in a "
                        + "corner of the polytope. There are several ways to "
                        + "solve this problem: "
                        + System.getProperty("line.separator")
                        + System.getProperty("line.separator")
                        + "1) The maximum number of obstructed moves can be increased,"
                        + System.getProperty("line.separator")
                        + "2) A different initial point can be used,"
                        + System.getProperty("line.separator")
                        + "3) The seed for the pseudo-random number generator can "
                        + "be changed, or"
                        + System.getProperty("line.separator")
                        + "4) The minimum chord length can be reduced."
                        + System.getProperty("line.separator")
                        + System.getProperty("line.separator")
                        + "Also ensure that the polytope is truly full dimensional."
                        + System.getProperty("line.separator");

        return new ApplicationException(s, e);
    }

    /**
     * Opens the file with the name denoted by {@linkplain #INITIAL_POINT_FILE} and
     * return the vector it contains.
     * 
     * @param input the input containing a value for {@code INITIAL_POINT_FILE}.
     * @param out the print writer to which informative messages will be
     * written.
     * @param variables the variables required to be present in the file.
     * 
     * @return the vector contained in the file with the provided name.
     * 
     * @throws FormatException if the file does not contain a required variable,
     * contains a single variable multiple times, or contains any other
     * formatting errors.
     */
    public static Map<String, Double> getInitialPoint(ManagerInput<String, String> input,
            PrintWriter out,
            Set<String> variables) {

        // Gettting file name
        String fileName = input.getRequiredValue(INITIAL_POINT_FILE);

        return getInitialPoint(fileName, out, variables);
    }

    /**
     * Opens the file with the provided name and return the vector it contains.
     * 
     * @param fileName the name of the file containing the initial point, that is, vector.
     * @param out the print writer to which informative messages will be written. 
     * @param variables the variables required to be present in the file.
     * 
     * @return the vector contained in the file with the provided name.
     * 
     * @throws FormatException if the file does not contain a required variable,
     * contains a single variable multiple times, or contains any other formatting errors.
     */
    public static Map<String, Double> getInitialPoint(String fileName,
            PrintWriter out,
            Set<String> variables) {

        // Printing an opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Getting the point
        Map<String, Double> point = getInitialPoint(fileName, variables);

        // Printing a closing message
        out.println("DONE");

        return point;
    }

    /**
     * Opens the file with the provided name and return the vector it contains.
     * 
     * @param fileName the name of the file containing the initial point, that is, vector.
     * @param variables the variables required to be present in the file.
     * 
     * @return the vector contained in the file with the provided name.
     * 
     * @throws FormatException if the file does not contain a required variable,
     * contains a single variable multiple times, or contains any other formatting errors.
     */
    public static Map<String, Double> getInitialPoint(String fileName,
            Set<String> variables) {

        // Creating key format
        NonEmptyStringFormat keyF = new NonEmptyStringFormat();

        // Creating value format
        DoubleFormat valueF = new DoubleFormatV1();

        // Creating map format
        MapFormat<String, Double> format =
                new EqualsMapFormatV1<String, Double>(keyF, valueF);

        return getInitialPoint(fileName, format, variables);
    }

    /**
     * Opens the file with the provided name and return the vector it contains.
     * 
     * @param fileName the name of the file containing the initial point, that is, vector.
     * @param format the format object used to parse the file.
     * @param variables the variables required to be present in the file.
     * 
     * @return the vector contained in the file with the provided name.
     * 
     * @throws FormatException if the file does not contain a required variable,
     * contains a single variable multiple times, or contains any other formatting errors.
     */
    public static Map<String, Double> getInitialPoint(String fileName,
            MapFormat<String, Double> format,
            Set<String> variables) {

        // Opening file
        RomiFile<String, Double> file =
                new RomiFile<String, Double>(fileName, variables, format);

        // Getting the map it contains
        Map<String, Double> point = file.getMap();

        // Closing
        file.close();

        return point;
    }

    /**
     * Parses and returns the value for {@linkplain #POINTS_PER_POINT} in the
     * provided manager input.
     * 
     * @param input the input containing a value for {@code POINTS_PER_POINT}.
     * @param out the print writer to which a status message will be printed.
     * 
     * @return the parsed value for {@linkplain #POINTS_PER_POINT} in the provided
     * input.
     */
    public static int getPointsPerPoint(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting the String value
        String s = input.getOptionalValue(POINTS_PER_POINT);

        if (s == null) {
            return 1;
        }

        // Parsing the value
        int x = parseInt(s);

        // Printing the value
        out.println("Parsed " + POINTS_PER_POINT + ": " + x);

        // Returning value
        return x;
    }

    /**
     * Parses and returns the value for {@linkplain #MAX_OBSTRUCTED_MOVES} in the
     * provided manager input.
     * 
     * @param input the input containing a value for
     * {@code MAX_OBSTRUCTED_MOVES}.
     * @param out the print writer to which a status message will be printed.
     * 
     * @return the parsed value for {@linkplain #MAX_OBSTRUCTED_MOVES} in the
     * provided input.
     */
    public static int getMaxObstructedMoves(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting the String value
        String s = input.getOptionalValue(MAX_OBSTRUCTED_MOVES);

        if (s == null) {
            return 0;
        }

        // Parsing the value
        int x = parseInt(s);

        // Printing the value
        out.println("Parsed " + MAX_OBSTRUCTED_MOVES + ": " + x);

        // Returning value
        return x;
    }

    /**
     * Parses and returns the value for {@linkplain #MIN_CHORD_LENGTH} in the
     * provided manager input.
     * 
     * @param input the input containing a value for {@code MIN_CHORD_LENGTH}.
     * @param out the print writer to which a status message will be printed.
     * 
     * @return the parsed value for {@linkplain #MIN_CHORD_LENGTH} in the provided
     * input.
     */
    public static double getMinChordLength(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting the String value
        String s = input.getOptionalValue(MIN_CHORD_LENGTH);

        if (s == null) {
            return DEFAULT_MIN_CHORD_LENGTH;
        }

        // Creating the number formatter
        DoubleFormat df = new DoubleFormatV1();

        // Parsing the value
        double length = df.parsePrimitive(s);

        // Checking its value
        if (length < 0 || Double.isInfinite(length) || Double.isNaN(length)) {

            String x =
                    "The value '"
                            + s
                            + "' for '"
                            + MIN_CHORD_LENGTH
                            + "' in '"
                            + input.getName()
                            + "' is invalid.";

            throw new ApplicationException(x);

        }

        // Printing parsed value
        out.println("Parsed " + MIN_CHORD_LENGTH + ": " + df.format(length));

        // Returning value
        return length;
    }

    /**
     * Parses and returns the value for {@linkplain #CONSTRAINT_TOLERANCE} in the
     * provided manager input.
     * 
     * @param input the input containing a value for
     * {@code CONSTRAINT_TOLERANCE}.
     * @param out the print writer to which a status message will be printed.
     * 
     * @return the parsed value for {@linkplain #CONSTRAINT_TOLERANCE} in the
     * provided input.
     */
    public static double getConstraintTolerance(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting the input string
        String s = input.getOptionalValue(CONSTRAINT_TOLERANCE);

        // If no value was provided
        if (s == null) {
            return DEFAULT_CONSTRAINT_TOLERANCE;
        }

        // Creating a double parser
        DoubleFormat f = new DoubleFormatV1();

        // Parsing the constraint tolerance
        double t = f.parsePrimitive(s);

        // Printing a formatted message
        out.println("Parsed " + CONSTRAINT_TOLERANCE + ": " + f.format(t));

        return t;
    }

    /**
     * Returns the list of sets contained in the file with the provided name.
     * The sets must be parsable using {@linkplain PipeSetFormat};
     * 
     * @param fileName the name of the file containing a list of sets.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return the list of sets contained in the file with the provided name.
     */
    public static Set<ParallelHyperplanes<String>> getPlanes(String fileName,
            PrintWriter out) {

        // Printing an opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        Set<ParallelHyperplanes<String>> planes = getPlanes(fileName);

        // Printing a closing message
        out.println("DONE");

        return planes;
    }

    /**
     * Returns the list of sets contained in the file with the provided name.
     * The sets must be parsable using {@linkplain PipeSetFormat};
     * 
     * @param fileName the name of the file containing a list of sets.
     * 
     * @return the list of sets contained in the file with the provided name.
     */
    public static Set<ParallelHyperplanes<String>> getPlanes(String fileName) {

        // Creating linear combination parser
        LinearCombFormat<String> combFormat = LinearCombFormatV1.getBasic();

        // Creating interval parser
        IntervalFormat intFormat = new IntervalFormatV1();

        // Creating map parser
        MapFormat<LinearComb<String>, Interval> mapFormat =
                new EqualsMapFormatV1<LinearComb<String>, Interval>(combFormat,
                        intFormat);

        // Opening file
        RomiFile<LinearComb<String>, Interval> file =
                new RomiFile<LinearComb<String>, Interval>(fileName, mapFormat);

        // Converting the map to parallel hyperplanes
        Set<ParallelHyperplanes<String>> planes = makePlanes(file.getMap());

        // Closing file
        file.close();

        return planes;
    }

    private static void checkPlanes(Set<ParallelHyperplanes<String>> planes,
            double minChordLength) {

        // Creating linear combination parser
        LinearCombFormat<String> combFormat = LinearCombFormatV1.getBasic();

        // Creating interval parser
        IntervalFormat intFormat = new IntervalFormatV1();

        for (ParallelHyperplanes<String> p : planes) {

            Interval b = p.getBounds();

            if (b.getUpper() - b.getLower() < minChordLength) {

                String s =
                        "The range of the bounds '"
                                + intFormat.format(b)
                                + "' for the hyperplane '"
                                + combFormat.format(p.getPlane())
                                + "' equals zero. The polytopes supplied to this process must be "
                                + "full dimensional.";

                throw new ApplicationException(s);

            }

        }

    }

    private static Set<ParallelHyperplanes<String>> makePlanes(Map<LinearComb<String>, Interval> vector) {

        // Determining an appropriate size
        int size = vector.size() + 1;

        // Creating storage
        Set<ParallelHyperplanes<String>> planes =
                new LinkedHashSet<ParallelHyperplanes<String>>(size, 1F);

        // Iterating thru each map entry
        for (Map.Entry<LinearComb<String>, Interval> e : vector.entrySet()) {

            ParallelHyperplanes<String> p =
                    new ParallelHyperplanes<String>(e.getKey(), e.getValue());

            planes.add(p);
        }

        return planes;
    }

    /**
     * Returns the list of sets contained in the file with the name contained in
     * the provided input. The sets must be parsable using {@linkplain PipeSetFormat};
     * 
     * @param input the input containing the name of the file containing a list
     * of sets.
     * @param out the print writer to which status messages will be printed.
     * 
     * @return the list of sets contained in the file with the provided name.
     */
    public static Set<ParallelHyperplanes<String>> getPlanes(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting file name
        String fileName = input.getRequiredValue(PLANES_FILE);

        return getPlanes(fileName, out);
    }

    /**
     * Returns an output file with the name provided in the manager input.
     * 
     * @param input the manager input containing the desired name of the output
     * file.
     * @param out the print writer to which a status message will be printed.
     * @param headers the variables whose values will be written to the output file.
     * 
     * @return an output file with the provided name.
     */
    public static ComoFile<String, Double> getOutputFile(ManagerInput<String, String> input,
            PrintWriter out,
            Set<String> headers) {

        String fileName = input.getOptionalValue(OUTPUT_FILE_NAME);

        if (fileName == null) {
            return null;
        }

        return getOutputFile(fileName, out, headers);
    }

    /**
     * Returns an output file with the provided name.
     * 
     * @param fileName the desired name of the output file.
     * @param out the print writer to which a status message will be printed.
     * @param headers the variables whose values will be written to the output file.
     * 
     * @return an output file with the provided name.
     */
    public static ComoFile<String, Double> getOutputFile(String fileName,
            PrintWriter out,
            Set<String> headers) {

        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        ComoFile<String, Double> file = getOutputFile(fileName, headers);

        out.println("DONE");

        return file;
    }

    /**
     * Returns an output file with the provided name.
     * 
     * @param fileName the desired name of the output file.
     * @param headers the data headers.
     * 
     * @return an output file with the provided name.
     */
    public static ComoFile<String, Double> getOutputFile(String fileName,
            Set<String> headers) {

        // Creating key formatter
        NonEmptyStringFormat keyF = new NonEmptyStringFormat();

        // Creating value formatter
        DoubleFormat valueF = new DoubleFormatV1();

        // Creating map formatter
        MapFormat<String, Double> mapF =
                new EqualsMapFormatV1<String, Double>(keyF, valueF);

        // Opening file
        return new ComoFile<String, Double>(fileName, headers, mapF);
    }

    // The process this process manager executes
    private CdHarProcess<String> process;

    private ComoFile<String, Double> output;

    /**
     * Returns the set of keywords required by this process manager.
     * 
     * @return the set of keywords required by this process manager.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords understood, but not required, by this process
     * manager.
     * 
     * @return the set of keywords understood, but not required, by this process
     * manager.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process controlled by this process manager.
     * 
     * @return the process controlled by this process manager.
     */
    public CdHarProcess<String> getProcess() {
        checkState();
        return process;
    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Getting parameters
        int iterations = getIterations(input, out);
        int maxObstructedMoves = getMaxObstructedMoves(input, out);
        int pointsPerPoint = getPointsPerPoint(input, out);
        long seed = getSeed(input, out);
        double minChordLength = getMinChordLength(input, out);
        double tolerance = getConstraintTolerance(input, out);

        // Getting the hyperplanes
        Set<ParallelHyperplanes<String>> p = getPlanes(input, out);
        checkPlanes(p, minChordLength);

        // Creating the polytope
        ConvexPolytope<String> polytope = new HashConvexPolytope<String>(p);

        // Getting the polytope's variables
        Set<String> variables = polytope.getVariables();

        // Getting the initial point
        Map<String, Double> startPoint = getInitialPoint(input, out, variables);

        // Ensuring the provided point is within the polytope
        if (!polytope.isWithin(startPoint, tolerance)) {

            String s =
                    "The point in the file '"
                            + input.getRequiredValue(INITIAL_POINT_FILE)
                            + "' does not lie within the polytope defined in the file '"
                            + input.getRequiredValue(PLANES_FILE)
                            + "'.";

            throw new ApplicationException(s);

        }

        // Creating the random numer generator
        Random random = new Random(seed);

        // Getting the output file
        output = getOutputFile(input, out, variables);

        // Creating the process
        process = new CdHarProcess<String>(polytope);
        process.setConstraintTolerance(tolerance);
        process.setIterations(iterations);
        process.setMaxObstructedMoves(maxObstructedMoves);
        process.setMinChordLength(minChordLength);
        process.setPoint(startPoint);
        process.setPointsPerPoint(pointsPerPoint);
        process.setRandom(random);

    }

    private void checkState() {

        if (process == null) {
            String s = "No input was provided.";
            throw new IllegalStateException(s);
        }
    }

    /**
     * Executes the process this process manager controls and writes the results
     * to the provided print writer.
     * 
     * @param out the print writer where the results will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager.
     */
    public void execute(PrintWriter out) {

        checkState();

        // Creating results manger
        Rm rm = new Rm(output, out);

        try {

            // Running process
            runProcess(process, rm, out);

        } catch (ExceededLimitException e) {

            throw interpret(e);

        }

        closeFile(output, out);
    }

    private static class Rm 
    extends AbstractResultsManager<CdHarComp<String>.Results> {

        private final MapOutputFile<String, Double> file;

        private final PrintWriter out;

        private int wastedMoves;

        private Rm(MapOutputFile<String, Double> file, PrintWriter out) {
            this.file = file;
            this.out = out;
            wastedMoves = 0;
        }

        /**
         * Adds data to this results manager.
         * 
         * @param <S> a type extending the primary result type.
         * @param results the results, or data, to be added to this manager.
         * 
         * @throws IllegalStateException if {@code close()} has already been
         * called.
         */
        public <S extends CdHarComp<String>.Results> void addData(S results) {

            // Checking state
            if (isClosed()) {
                throw new IllegalStateException("Already closed");
            }

            // Incrementing wasted moves
            wastedMoves += results.unchangedVariables();

            Map<String, Double> point = results.getPoint();
            
            // Writing flux vector to file
            file.addData(point);
        }

        /**
         * Prints a closing message.
         */
        public void printClosing() {
            out.println("Total Wasted Moves = " + wastedMoves);
            out.println();
        }
    }
}
