/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.graph_theory;

import java.util.Set;
import java.io.PrintWriter;
import java.util.*;

import sbrt.kernel.io.*;
import sbrt.kernel.math.graph_theory.*;
import sbrt.kernel.math.graph_theory.proc.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.*;

/**
 * This class is used to manage simple cycle identification processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class CycleIdManager extends AbstractProcessManager implements KernelProcessManager<String, String> {

	/**
	 * The keyword used to denote the input file containing a set of edges.
	 */
    public static final String EDGE_FILE = "Edge File";

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
    	
        Set<String>  s = new TreeSet<String> ();

        s.add(EDGE_FILE);

        requiredKeywords = new UnmodifiableSet<String> (s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String>  optionalKeywords;
    static {

    	Set<String> s = new TreeSet<String>();

    	// Output file information
        s.add(OUTPUT_FILE_NAME);
        s.add(OUTPUT_FILE_FORMAT);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }
    
    /**
	 * Creates a returns an output file if one is specified in the provided
	 * input.
	 * 
	 * @param input
	 *            the input for which the
	 *            {@linkplain #OUTPUT_FILE_NAME} and
	 *            {@linkplain #OUTPUT_FILE_FORMAT} values
	 *            are optionally present.
     * @param out the print writer to which descriptive messages will be written.
	 *            
	 * @return an output file if one is specified in the provided input, {@code null}
	 * otherwise.
	 */
    public static TextOutputFile<List<String>> 
    getOutputFile(ManagerInput<String, String> input, PrintWriter out) {

    	// Getting the file name
        String fileName = input.getOptionalValue(OUTPUT_FILE_NAME);

        // If no file is to be written
        if (fileName == null) {
        	return null;
        }
        
        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();
        
        // Creating the output file
        TextOutputFile<List<String>> file = getOutputFile(input);

        // Printing a closing message
        out.println("DONE");

        return file;
    }
    
    /**
	 * Creates a returns an output file if one is specified in the provided
	 * input.
	 * 
	 * @param input
	 *            the input for which the
	 *            {@linkplain #OUTPUT_FILE_NAME} and
	 *            {@linkplain #OUTPUT_FILE_FORMAT} values
	 *            are optionally present.
	 *            
	 * @return an output file if one is specified in the provided input, {@code null}
	 * otherwise.
	 */
    public static TextOutputFile<List<String>> 
    getOutputFile(ManagerInput<String, String> input) {

    	// Getting the file name
        String fileName = input.getOptionalValue(OUTPUT_FILE_NAME);

        // If no file is to be written
        if (fileName == null) {
        	return null;
        }

        // Getting the file format
        String s = input.getOptionalValue(OUTPUT_FILE_FORMAT);
        
        FileFormat fileFormat = getTextFileFormat(s);
        
        // Creating list format
        PipeListFormat<String> format = PipeListFormat.getBasic();
        
        SimpleListFormat<String> lineFormat = new SimpleListFormat<String>(format);
        
        // Creating the output file
        TextOutputFile<List<String>> file = 
        	new TextOutputFile<List<String>>(fileName, fileFormat, lineFormat);

        return file;
    }
    
    private CycleIdProcess<String> process;
    private OutputFile<List<String>> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public CycleIdProcess<String> getProcess() {
        checkState();
        return process;
    }

    private void closeFile(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }

    }

    private void checkState() {
        if (process == null) {
            throw new IllegalStateException("No manager input was provided.");
        }
    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing previous file if necessary
        closeFile(out);

        // Creating graph
        DirectedGraph<String> g = PathIdManager.getGraph(input, out);

        // Creating process
        process = new CycleIdProcess<String>();
        process.setGraph(g);

        // Setting output file
        outputFile = getOutputFile(input, out);
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();
        
        try {

            // Creating results manager
            Rm rm = new Rm(outputFile);
        	
            // Running process
            runProcess(process, rm, out);

            // Closing results manager
            rm.close(out);
            
        } finally {
            closeFile(out);
        }
    }

    private final class Rm extends AbstractResultsManager<Path<String>> {
    	
    	private final OutputFile<List<String>> output;
    	private int cycles;
    	
    	private Rm(OutputFile<List<String>> output) {
    		this.output = output;
    		cycles = 0;
    	}

    	/**
    	 * Adds data to this results manager.
    	 * 
    	 * @param cycle the results, or data, to be added
    	 * to this manager.
    	 * 
    	 * @throws IllegalStateException if {@code close()}
    	 * has already been called.
    	 */
    	public void addData(Path<String> cycle) {
    		
    		++cycles;
    		
    		if (output != null) {
    			output.addData(cycle.toList());
    		}
    	}
    	
    	/**
    	 * Prints a closing message.
    	 * @param out the print writer to which the closing message is printed.
    	 */
    	public void close(PrintWriter out) {

    		out.println("Total cycles = " + cycles);
    		
    		if (outputFile != null) {
    			out.println();
    		}

    		super.close();
    	}
    }
}
