/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.graph_theory;

import java.io.PrintWriter;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.math.graph_theory.DirectedGraph;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.MapManagerInput;

/**
 * This class is used to manage a graph information gatherer.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class GraphInfoManager 
	extends AbstractProcessManager {

    /**
     * The keyword used to denote the name of the file to which all nodes
     * will be written.
     */
    public static final String NODE_FILE_NAME = "Node File Name";
    
	/**
	 * The set of keywords required to be present in the {@code ManagerInput}
	 * objects supplied to instances of this class.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {
		Set<String> s = new TreeSet<String>();
		s.add(PathIdManager.EDGE_FILE);
		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by instances of this
	 * class.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
		Set<String> s = new TreeSet<String>();
		s.add(NODE_FILE_NAME);
		optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Prints basic information about the provided graph.
	 * 
	 * @param graph
	 *            the graph for which basic information will be printed.
	 * @param out
	 *            the print writer to which basic information will be printed.
	 */
	public static void printBasicInfo(DirectedGraph<?> graph, PrintWriter out) {

		out.println();
		out.println("Number of nodes: " + graph.nodes());
		out.println("Number of edges: " + graph.edges());        
        out.println();

	}

	private DirectedGraph<String> graph;
	private MapManagerInput<String, String> input;

	/**
	 * Returns the set of keywords the provided {@code ManagerInput} object must
	 * contain.
	 * 
	 * @return the set of keywords the provided {@code ManagerInput} object must
	 *         contain.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
	 * Returns the set of keywords the provided {@code ManagerInput} object does
	 * not necessarily have to contain.
	 * 
	 * @return the set of keywords the provided {@code ManagerInput} object does
	 *         not necessarily have to contain.
	 */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}
	
	private void checkState() {
		if (input == null) {
			String s = "No manager input was provided.";
			throw new IllegalStateException(s);
		}
	}

	/**
	 * Uses the provided input to prepare this process manager for execution. As
	 * the input is processed, informative messages are printed to the provided
	 * print writer.
	 * 
	 * @param input
	 *            all of the information required to prepare this process
	 *            manager for execution.
	 * 
	 * @param out
	 *            the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

		// Getting graph
		graph = PathIdManager.getGraph(input, out);

		// Getting keyword-value pairs
		this.input = new MapManagerInput<String, String>(input.toMap());

		// Printing basic information
		printBasicInfo(graph, out);
	}

	/**
	 * Executes the process this process manager controls, and writes the status
	 * or results of this process to the provided print writer.
	 * 
	 * @param out
	 *            the print writer where status messages will be printed.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager via
	 *             {@linkplain #setInput(ManagerInput)} or
	 *             {@linkplain #setInput(ManagerInput, PrintWriter)}
	 */
	public void execute(PrintWriter out) {

		checkState();
		
		// Writing optional files
		writeOptionalFiles(out);

	}

	private void writeOptionalFiles(PrintWriter out) {
		writeNodeFile(out);
	}

	private void writeNodeFile(PrintWriter out) {

		// Getting the filename
		String fileName = input.getOptionalValue(NODE_FILE_NAME);

		// If no filename was provided, do nothing
		if (fileName == null) {
			return;
		}

		// Printing an opening message
		out.print("Writing '" + fileName + "' ... ");
		out.flush();

		// Creating output file
        OutputFile<String> file =
        	TextOutputFile.getBasic(fileName);
        
		// Writing the nodes
        for (String node : graph.getNodes()) {
    		file.addData(node);
        }

		// Closing the file
		file.close();

		// Printing a closing message
		out.println("DONE");

	}


}
