/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.graph_theory;

import java.util.Set;
import java.io.PrintWriter;
import java.util.*;

import sbrt.kernel.io.*;
import sbrt.kernel.math.graph_theory.*;
import sbrt.kernel.math.graph_theory.proc.*;
import sbrt.kernel.math.graph_theory.DirectedGraph.Edge;
import sbrt.kernel.utilities.*;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.*;
import sbrt.shell.text.math.*;

/**
 * This class is used to manage simple path identification processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class PathIdManager extends AbstractProcessManager implements KernelProcessManager<String, String> {

	/**
	 * The keyword used to denote the input file containing a set of edges.
	 */
    public static final String EDGE_FILE = "Edge File";

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
    	
        Set<String>  s = new TreeSet<String> ();

        s.add(EDGE_FILE);

        requiredKeywords = new UnmodifiableSet<String> (s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String>  optionalKeywords;
    static {

    	Set<String> s = new TreeSet<String>();

    	// Output file information
        s.add(OUTPUT_FILE_NAME);
        s.add(OUTPUT_FILE_FORMAT);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
	 * Opens the file denoted by {@linkplain #EDGE_FILE} in the provided manager
	 * input and returns the directed graph it contains. The file must be a 
	 * text file, and each non-commented line
     * must be parsable with {@linkplain DirEdgeFormatV1}. 
	 * 
	 * @param input the input containing a value for {@code EDGE_FILE}.
	 * @param out the print writer to which a status message will be printed.
	 * 
	 * @return the directed graph contained in the file with the provided name.
	 */
    public static DirectedGraph<String>
    getGraph(ManagerInput<String, String> input, PrintWriter out) {

    	// Getting file name
        String fileName = input.getRequiredValue(EDGE_FILE);

        // Printing opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Getting the graph
        DirectedGraph<String> g = getGraph(fileName);
        
        // Printing a closing message
        out.println("DONE");

        return g;
    }

    /**
     * Parses the file with the provided name, and returns the set of edges it
     * contains. The file must be a text-file, and each non-commented line
     * must be parsable with {@linkplain DirEdgeFormatV1}. 
     * 
     * @param fileName the name of the file to be parsed.
     * 
     * @return the set of edges contained in the file with the provided name.
     */
    public static Set<Edge<String>> getEdges(String fileName) {
       	
        // Creating egde format
    	DirEdgeFormat<String> format = DirEdgeFormatV1.getBasic();

    	return getEdges(fileName, format);
    }

    /**
     * Parses the file with the provided name, and returns the set of edges it
     * contains. The file must be a text-file, and each non-commented line
     * must be parsable with the provided format. 
     * 
     * @param <N> the node type.
     * @param fileName the name of the file to be parsed.
     * @param format the format object used to parse the lines of the file.
     * 
     * @return the set of edges contained in the file with the provided name.
     */
    public static <N> Set<Edge<N>> 
    getEdges(String fileName, DirEdgeFormat<N> format) {

        // Opening input file
        InputFile<Edge<N>> file = 
        	new TextInputFile<Edge<N>>(fileName, format);

        // Creating storage
        Set<Edge<N>> edges = new LinkedHashSet<Edge<N>>();

        // Iterating thru the file
        for (Edge<N> e : file) {

        	// Trying to add the edge to the rest
        	if (!edges.add(e)) {
        		
        		String s = "The edge '" + format.format(e) + 
        		"' appears more than once in the file '" + fileName + "'.";
        		
        		throw new FormatException(s);
        		
        	}
        }
    	
        return edges;
    }

    /**
     * Parses the file with the provided name, and returns the directed graph it
     * contains. The file must be a text-file, and each non-commented line
     * must be parsable with {@linkplain DirEdgeFormatV1}. 
     * 
     * @param fileName the name of the file to be parsed.
     * 
     * @return the directed graph contained in the file with the provided name.
     */
    public static DirectedGraph<String> getGraph(String fileName) {

        // Getting edges
        Set<Edge<String>> edges = getEdges(fileName);
        
        // Creating graph
    	return new DirectedGraph<String>(edges);
    }

    /**
     * Parses the file with the provided name, and returns the directed graph it
     * contains. The file must be a text-file, and each non-commented line
     * must be parsable with the provided format. 
     * 
     * @param <N> the node type.
     * @param fileName the name of the file to be parsed.
     * @param format the format object used to parse the lines of the file.
     * 
     * @return the directed graph contained in the file with the provided name.
     */
    public static <N> DirectedGraph<N> 
    getGraph(String fileName, DirEdgeFormat<N> format) {

        // Getting edges
        Set<Edge<N>> edges = getEdges(fileName, format);
        
        // Creating graph
        return new DirectedGraph<N>(edges);
    }

    private PathIdProcess<String> process;
    private OutputFile<List<String>> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public PathIdProcess<String> getProcess() {
        checkState();
        return process;
    }

    private void closeFile(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }

    }

    private void checkState() {
        if (process == null) {
            throw new IllegalStateException("No manager input was provided.");
        }
    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing previous file if necessary
        closeFile(out);

        // Creating graph
        DirectedGraph<String> g = getGraph(input, out);

        // Creating process
        process = new PathIdProcess<String>();
        process.setGraph(g);

        // Setting output file
        outputFile = CycleIdManager.getOutputFile(input, out);
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();
        
        try {

            // Creating results manager
            Rm rm = new Rm(outputFile);
        	
            // Running process
            runProcess(process, rm, out);

            // Closing results manager
            rm.close(out);
            
        } finally {
            closeFile(out);
        }
    }

    private final class Rm extends AbstractResultsManager<Path<String>> {
    	
    	private final OutputFile<List<String>> output;
    	private int paths;
    	
    	private Rm(OutputFile<List<String>> output) {
    		this.output = output;
    		paths = 0;
    	}

    	/**
    	 * Adds data to this results manager.
    	 * 
    	 * @param path the results, or data, to be added
    	 * to this manager.
    	 * 
    	 * @throws IllegalStateException if {@code close()}
    	 * has already been called.
    	 */
    	public void addData(Path<String> path) {
    		
    		++paths;
    		
    		if (output != null) {
    			output.addData(path.toList());
    		}
    	}
    	
    	/**
    	 * Prints a closing message.
    	 * @param out the print writer to which the closing message is printed.
    	 */
    	public void close(PrintWriter out) {

    		out.println("Total paths = " + paths);
    		
    		if (outputFile != null) {
    			out.println();
    		}

    		super.close();
    	}
    }
}
