/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.graph_theory;

import java.io.PrintWriter;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.FileFormat;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.TextInputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.mng.ShellProcessManager;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.PipeListFormat;
import sbrt.shell.text.SimpleListFormat;

/**
 * This class is used to identify the unique cycles in files that contain 
 * collections of cycles. 
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class UniqueCycleIdManager 
	extends AbstractProcessManager 
	implements ShellProcessManager<String, String> {


    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
    	
        Set<String>  s = new TreeSet<String> ();

        s.add(INPUT_FILE);

        requiredKeywords = new UnmodifiableSet<String> (s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String>  optionalKeywords;
    static {
    	Set<String> s = new TreeSet<String>();
    	
    	s.add(OUTPUT_FILE_NAME);
        s.add(OUTPUT_FILE_FORMAT);
        s.add(INPUT_FILE_FORMAT);
        
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    private static SimpleListFormat<String> cycleFormat;
    
    private static SimpleListFormat<String> getCycleFormat() {

    	if (cycleFormat == null) {
    		cycleFormat = new SimpleListFormat<String>(PipeListFormat.getBasic());
    	}
    	
    	return cycleFormat;
    }


    private static TextInputFile<List<String>> 
    getInputFile(ManagerInput<String, String> input, PrintWriter out) {

    	// Getting the file name
        String fileName = input.getRequiredValue(INPUT_FILE);
        
        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Getting list format
        SimpleListFormat<String> lineF = getCycleFormat();
        
        // Getting file format
        String s = input.getOptionalValue(INPUT_FILE_FORMAT);
        FileFormat fileF = getTextFileFormat(s);
        
        // Creating the output file
        TextInputFile<List<String>> file =
        	new TextInputFile<List<String>>(fileName, fileF, lineF);

        // Printing a closing message
        out.println("DONE");

        return file;
    }
    
    private TextInputFile<List<String>> inputFile;
    private OutputFile<List<String>> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }


    private void closeFile(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }

    }

    private void checkState() {
        if (inputFile == null) {
            throw new IllegalStateException("No manager input was provided.");
        }
    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing previous output file if necessary
        closeFile(out);

        // Getting the input file of cycles
        inputFile = getInputFile(input, out);

        // Setting output file
        outputFile = CycleIdManager.getOutputFile(input, out);
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();
        
        try {

        	out.println();
        	out.print("Identifying unique cycles ... ");
        	out.flush();
        	
        	int uniqueCycles = run();
        	
        	out.println("DONE");

        	out.println();
        	out.println("Unique cycles = " + uniqueCycles);
            
        } catch (RuntimeException e) {

        	out.println();
        	throw e;
        	
        } finally {
        	out.println();
            closeFile(outputFile, out);
            closeFile(inputFile, out);
        }
    }

    private int run() {
    	
    	// Creating counter for unique cycles
    	int uniqueCycles = 0;
    	
    	// Creating a set to hold the line indices that should be ignored
    	Set<Integer> ignorableIndices = new TreeSet<Integer>();
    	
    	// Getting an iterator over the input file
    	Iterator<List<String>> i = inputFile.iterator();
    	
    	// Iterating over each line of the input file, 
    	// keeping track of the current line index.
    	for (int currentIndex = 0; i.hasNext(); currentIndex ++) {
    		
    		// Getting the cycle from the current line
    		List<String> currentCycle = i.next();

    		// Ignoring the current line if it should be ignored
    		if (ignorableIndices.contains(currentIndex)) {
    			continue;
    		}
    		
    		// Ensuring the cycle is properly formed
    		checkCycle(currentCycle);
    		
    		// Writing it to the output file
    		outputFile.addData(currentCycle);
    		
    		// Finding any additional indices that should be ignored
    		Set<Integer> moreIgnorableIndices = findIgnorableIndices(currentCycle, currentIndex);
    	
    		// Adding these indices to the rest
    		ignorableIndices.addAll(moreIgnorableIndices);

    		// Incrementing unique cycles
    		++ uniqueCycles;
    	}
    	
    	return uniqueCycles;
    }
    
    
    private Set<Integer> findIgnorableIndices(List<String> currentCycle, int currentIndex) {
    	
    	// Creating an iterator over the input file starting from the 
    	// current index.
    	Iterator<List<String>> i = inputFile.iterator(currentIndex);
    	
    	// Skipping the first cycle, since it is the 'currentCycle'
    	if (i.hasNext()) {
    		i.next();
    	}
    	
    	// Creating a set to hold the indices of redundant cycles
    	Set<Integer> indices = new TreeSet<Integer>();
    	
    	// Looping over the subsequent cycles in the input file
    	for (int index = currentIndex + 1; i.hasNext(); index ++) {
    		
    		// Getting the next cycle in the file
    		List<String> cycle = i.next();
    		
    		// Ensuring it is properly formed
    		checkCycle(cycle);
    		
    		// If the two cycles are equivalent, add the index 
    		// of the *second* cycle to the set of ignorable indices.
    		if (sameCycles(currentCycle, cycle)) {
    			indices.add(index);
    		}
    	}

    	return indices;
    }
    
    
    private void checkCycle(List<String> cycle) {
    	
    	int n = cycle.size();
    	
    	if (n < 2) {
    		
    		String s = "A problem occurred while parsing the file '" + inputFile.getCanonicalName() +
    		    	   "'. The list '" + getCycleFormat().format(cycle) + 
    				   "' does not contain enough nodes to be a legitimate cycle.";
    		throw new FormatException(s);
    		
    	}
    	
    	if (!cycle.get(0).equals(cycle.get(n - 1))) {
    		
    		String s = "A problem occurred while parsing the file '" + inputFile.getCanonicalName() +
	    	   		   "'. The list '" + getCycleFormat().format(cycle) +
    				   "' does not begin and end with the same node.";
    		throw new FormatException(s);
    		
    	}
    	
    }
    
    
    private static boolean sameCycles(List<String> cycle1, List<String> cycle2) {
    	
    	int n = cycle1.size();
    	
    	if (n != cycle2.size()) {
    		return false;
    	}
    	
    	// If both cycles are (somehow) empty, they are the same
    	if (n == 0) {
    		return true;
    	}
    	
    	// Updating node count to reflect the fact that the last node
    	// in a cycle is the same as the first.
    	--n;
    	
    	// Getting the first node from cycle1
    	Object node = cycle1.get(0);
    	
    	// Getting its index in cycle2, if it exists
    	int index = cycle2.indexOf(node);
    	
    	if (index < 0) {
    		return false;
    	}
    	
    	// Comparing the nodes from each cycle, in order
    	for (int i = 0; i < n; i ++, index ++) {
    		
    		Object node1 = cycle1.get(i);
    		Object node2 = cycle2.get(index % n);
    		
    		if (!node1.equals(node2)) {
    			return false;
    		}
    	}
    	
    	return true;
    }


}
