/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.stat;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.ext.R.JriException;
import sbrt.ext.R.R;
import sbrt.ext.R.stat.RCorrelator;
import sbrt.ext.R.stat.RKendall;
import sbrt.ext.R.stat.RPearson;
import sbrt.ext.R.stat.RSpearman;
import sbrt.ext.R.stat.RCorrelator.Result;
import sbrt.kernel.io.SingleRm;
import sbrt.kernel.math.stat.proc.CorrelationAnalysis;
import sbrt.kernel.utilities.UnmodifiableMap;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.RomiFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.KernelProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.ListFormat;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.NonEmptyStringFormat;
import sbrt.shell.text.PipeListFormat;
import sbrt.shell.text.SimpleListFormat;

/**
 * This class is used to manage correlation processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public class CorrelationManager extends AbstractProcessManager implements
        KernelProcessManager<String, String> {

    /**
     * The keyword used to denote the type of correlation coefficient to be
     * computed.
     */
    public static final String CORR_COEFF = "Correlation Coefficient";

    /**
     * The value used to denote a Kendall's tau correlation coefficient.
     */
    public static final String KENDALLS_TAU = "Kendall's tau";

    /**
     * The value used to denote a Pearson's r correlation coefficient.
     */
    public static final String PEARSONS_R = "Pearson's r";

    /**
     * The value used to denote a Spearman's rho correlation coefficient.
     */
    public static final String SPEARMANS_RHO = "Spearman's rho";

    /**
     * The keyword used to denote the name of the file containing the x values.
     */
    public static final String X_FILE_NAME = "X Value File Name";

    /**
     * The keyword used to denote the name of the file containing the y values.
     */
    public static final String Y_FILE_NAME = "Y Value File Name";

    /**
     * The set of keywords required by this process manager.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {

        Set<String> s = new TreeSet<String>();

        s.add(CORR_COEFF);
        s.add(X_FILE_NAME);
        s.add(Y_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by this process
     * manager.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {

        Set<String> s = new TreeSet<String>();

        s.add(OUTPUT_FILE_NAME);

        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Parses the file with the provided name and returns the vector it
     * contains. The file must be a row-oriented map file whose variable-value
     * pairs are parsable using {@linkplain EqualsMapFormatV1}. The variables are
     * not parsed further, and the values are parsed using
     * {@linkplain DoubleFormatV1}.
     * 
     * @param fileName the name of the file to be read.
     * 
     * @return variables mapped to their corresponding value.
     */
    public static UnmodifiableMap<String, Double> getVector(String fileName) {

        // Creating variable parser
        NonEmptyStringFormat varFormat = new NonEmptyStringFormat();

        // Creating number parser
        DoubleFormat numFormat = new DoubleFormatV1();

        // Creating line parser
        MapFormat<String, Double> lineFormat =
                new EqualsMapFormatV1<String, Double>(varFormat, numFormat);

        // Opening input file
        RomiFile<String, Double> input =
                new RomiFile<String, Double>(fileName, lineFormat);

        // Parsing file
        UnmodifiableMap<String, Double> vector = input.getMap();

        // Closing file
        input.close();

        // Returning vector
        return vector;
    }

    /**
     * Parses the file with the provided name, prints a status message to the
     * provided print writer, and returns the vector the file contains. The file
     * must be a row-oriented map file whose variable-value pairs are parsable
     * using {@linkplain EqualsMapFormatV1}. The variables are not parsed further,
     * and the values are parsed using {@linkplain DoubleFormatV1}.
     * 
     * @param fileName the name of the file to be read.
     * @param out the print writer to which a status message will be written.
     * 
     * @return variables mapped to their corresponding value.
     */
    public static UnmodifiableMap<String, Double> getVector(String fileName,
            PrintWriter out) {

        // Printing an opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Getting vector
        UnmodifiableMap<String, Double> vector = getVector(fileName);

        // Printing a closing message
        out.println("DONE");

        return vector;
    }

    /**
     * Returns the parsed value for {@linkplain #X_FILE_NAME} in the provided input
     * and prints a descriptive message to the provided print writer.
     * 
     * @param input the input whose {@code X_FILE_NAME} value is to be parsed.
     * @param out the print writer to which a message is printed.
     * 
     * @return variables mapped to their corresponding value.
     * 
     * @see #getVector(String, PrintWriter)
     */
    public static UnmodifiableMap<String, Double> getXValues(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting the x value file name
        String fileName = input.getRequiredValue(X_FILE_NAME);

        // Returning the vector contained in this file
        return getVector(fileName, out);
    }

    /**
     * Returns the parsed value for {@linkplain #Y_FILE_NAME} in the provided input
     * and prints a descriptive message to the provided print writer.
     * 
     * @param input the input whose {@code Y_FILE_NAME} value is to be parsed.
     * @param out the print writer to which a message is printed.
     * 
     * @return variables mapped to their corresponding value.
     * 
     * @see #getVector(String, PrintWriter)
     */
    public static UnmodifiableMap<String, Double> getYValues(ManagerInput<String, String> input,
            PrintWriter out) {

        // Getting the y value file name
        String fileName = input.getRequiredValue(Y_FILE_NAME);

        // Returning the vector contained in this file
        return getVector(fileName, out);
    }
    
    /**
     * Writes the intersection of the provided vectors to a file with the specified
     * name.
     * 
     * @param fileName the desired name of the output file.
     * @param xValues the x values.
     * @param yValues the y values.
     * @param out the print writer to which descriptive messages will be written.
     */
    public static void 
    writeIntersection(String fileName,
                      Map<String, Double> xValues, 
                      Map<String, Double> yValues,
                      PrintWriter out) {
        
        // Printing an opening message
        out.print("Writing '" + fileName + "' ... ");
        out.flush();
        
        // Creating list format
        ListFormat<String> listFormat = PipeListFormat.getBasic();
        
        // Converting the list format into a file line formatter
        SimpleListFormat<String> lineFormat = new SimpleListFormat<String>(listFormat);
        
        // Creating number formatter
        DoubleFormat numFormat = new DoubleFormatV1();
        
        // Opening output file
        OutputFile<List<String>> output = 
            new TextOutputFile<List<String>>(fileName, lineFormat);
        
        // Creating temporary storage
        List<String> line = new ArrayList<String>(4);
        
        // Writing file
        for (Map.Entry<String, Double> e : xValues.entrySet()) {
            
        	String variable = e.getKey();
        	
            // Getting value from y if it exists
            Double yVal = yValues.get(variable);
            
            // If there is no y value, move on to the next
            if (yVal == null) {
                continue;
            }
            
            // Formatting values
            String x = numFormat.format(e.getValue());
            String y = numFormat.format(yVal);
            
            // Storing line
            line.clear();
            line.add(variable);
            line.add(x);
            line.add(y);
            
            // Writing line
            output.addData(line);
        }
        
        // Closing
        output.close();
        
        // Printing a closing message
        out.println("DONE");
    }
    
    /**
     * Returns the correlation computer corresponding to the provided string
     * value.
     * 
     * @param coeff a string equal to {@linkplain #KENDALLS_TAU},
     * {@linkplain #SPEARMANS_RHO}, or {@linkplain #PEARSONS_R}.
     * 
     * @return the correlation computer corresponding to the provided string
     * value.
     * 
     * @throws ApplicationException if the provided string does not equal
     * {@linkplain #KENDALLS_TAU}, {@linkplain #SPEARMANS_RHO}, or
     * {@linkplain #PEARSONS_R}.
     */
    public static RCorrelator<String> parseCorrelator(String coeff) {

        if (coeff.equals(PEARSONS_R)) {
            return new RPearson<String>();
        }

        else if (coeff.equals(KENDALLS_TAU)) {
            return new RKendall<String>();
        }

        else if (coeff.equals(SPEARMANS_RHO)) {
            return new RSpearman<String>();
        }

        else {
            String s =
                    "The correlation coefficient '"
                            + coeff
                            + "' is unrecognizable.";
            throw new ApplicationException(s);
        }
    }

    /**
     * Starts the R engine and writes an informative message to the provided
     * print writer.
     * @param out the print writer to which a message will be written.
     */
    public static void startR(PrintWriter out) {
        
        out.println();
        out.print("Initializing R ... ");
        out.flush();

        try {
            R.startR();
        } catch (Throwable t) {
            
            String s = "R could not be started. Please ensure that the " +
                       "appropriate files are located in the SBRT's 'lib' " +
                       "directory and that the environmental variable 'R_HOME' " +
                       "is properly set.";
            throw new ApplicationException(s, t);
        }
        
        out.println("DONE");
        
    }
    
    /**
     * Wraps the provided JRI exception as an application exception.
     * @param e the JRI exception.
     * @return an application exception.
     */
    public static ApplicationException interpret(JriException e) {
        
        String s = "The Java to R interface threw an exception. " +
                   "Its message is: " + e.getMessage();
        
        return new ApplicationException(s, e);
    }
    
    // The process this process manager executes
    private CorrelationAnalysis<String, Result> process;

    /**
     * Returns the set of keywords required by this process manager.
     * 
     * @return the set of keywords required by this process manager.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords understood, but not required, by this process
     * manager.
     * 
     * @return the set of keywords understood, but not required, by this process
     * manager.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process controlled by this process manager.
     * 
     * @return the process controlled by this process manager.
     */
    public CorrelationAnalysis<String, Result> getProcess() {
        checkState();
        return process;
    }

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Getting the correlator
        String stat = input.getRequiredValue(CORR_COEFF);
        RCorrelator<String> corr = parseCorrelator(stat);

        // Getting the x and y vectors
        Map<String, Double> x = getXValues(input, out);
        Map<String, Double> y = getYValues(input, out);

        // Writing output file
        writeOutput(input, x, y, out);

        // Creating the correlator
        process = new CorrelationAnalysis<String, Result>(corr);
        process.setXValues(x);
        process.setYValues(y);
    }

    private void writeOutput(ManagerInput<String, String> input,
                             Map<String, Double> xValues,
                             Map<String, Double> yValues,
                             PrintWriter out) {

        // Getting the output file name
        String fileName = input.getOptionalValue(OUTPUT_FILE_NAME);

        if (fileName != null) {
            writeIntersection(fileName, xValues, yValues, out);
        }
        
    }

    private void checkState() {

        if (process == null) {
            String s = "No input was provided.";
            throw new IllegalStateException(s);
        }
    }

    /**
     * Executes the process this process manager controls and writes the results
     * to the provided print writer.
     * 
     * @param out the print writer where the results will be printed.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager.
     */
    public void execute(PrintWriter out) {

        checkState();
        
        // Starting the R engine in a monitorable manner
        startR(out);
        
        // Creating results manger
        SingleRm<Result> rm = new SingleRm<Result>();

        try {

            // Running process
            process.run(rm);

        } catch (JriException e) {
            throw interpret(e);
        }

        // Retrieving results
        Result results = rm.getResults();

        // Writing results
        out.println();
        out.println("Coefficient = " + results.getCoeff());
        out.println("P-Value     = " + results.getP());

    }

}
