/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.stat;

import java.io.PrintWriter;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.io.SingleRm;
import sbrt.kernel.math.stat.comp.KendallTau;
import sbrt.kernel.math.stat.comp.KendallTauA;
import sbrt.kernel.math.stat.comp.KendallTauB;
import sbrt.kernel.math.stat.comp.TauResult;
import sbrt.kernel.math.stat.proc.CorrelationAnalysis;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.ApplicationException;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.KernelProcessManager;
import sbrt.shell.mng.ManagerInput;

/**
 * This class is used to manage correlation process whose correlation 
 * computers are of type {@linkplain KendallTau}.
 *  
 * @author This class was written and documented by Jeremiah Wright while in
 * the Wagner lab.
 *
 */
public class KendallTauManager extends AbstractProcessManager
        implements KernelProcessManager<String, String> {


	/**
	 * The value used to denote a Kendall's Tau<sub>A</sub> correlation
	 * coefficient. 
	 */
	public static final String TAU_A = "Kendall's Tau A";

	/**
	 * The value used to denote a Kendall's Tau<sub>B</sub> correlation
	 * coefficient. 
	 */
	public static final String TAU_B = "Kendall's Tau B";
	
	/**
	 * The set of keywords required by this process manager.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {

		Set<String> s = new TreeSet<String>();

		s.add(CorrelationManager.CORR_COEFF);
		s.add(CorrelationManager.X_FILE_NAME);
		s.add(CorrelationManager.Y_FILE_NAME);
		
		requiredKeywords = new UnmodifiableSet<String>(s);
	}
	
	/**
	 * The set of keywords understood, but not required, by this process
	 * manager.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {
		
		Set<String> s = new TreeSet<String>();
		
		s.add(OUTPUT_FILE_NAME);
		
		optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Returns the Kendall's Tau statistic computer corresponding to the 
	 * provided string value.
	 * 
	 * @param stat a string equal to either {@linkplain #TAU_A} or
	 * {@linkplain #TAU_B}.
	 * 
	 * @return the Kendall's Tau computer corresponding to the provided
	 * string value.
	 * 
	 * @throws ApplicationException if the provided string does not equal
	 * {@linkplain #TAU_A} or {@linkplain #TAU_B}.
	 */
	public static KendallTau<String> parseCorrelator(String stat) {
		
		if (stat.equals(TAU_A)) {
			return new KendallTauA<String>();
		}
		
		else if (stat.equals(TAU_B)) {
			return new KendallTauB<String>();
		}
		
		else {
			String s = "The correlation statistic '" + stat + 
					   "' is unrecognizable.";
			throw new ApplicationException(s);
		}
	}

	// The process this process manager executes
	private CorrelationAnalysis<String, TauResult> process;
	
	/**
	 * Returns the set of keywords required by this process manager.
	 * @return the set of keywords required by this process manager.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}
	
	/**
	 * Returns the set of keywords understood, but not required, by this 
	 * process manager.
	 * @return the set of keywords understood, but not required, by this 
	 * process manager.
	 */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

	/**
	 * Returns the process controlled by this process manager.
	 * @return the process controlled by this process manager.
	 */
	public CorrelationAnalysis<String, TauResult> getProcess() {
		checkState();
		return process;
	}

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
        check(input);

        // Getting the correlator
        String stat = input.getRequiredValue(CorrelationManager.CORR_COEFF);
        KendallTau<String> corr = parseCorrelator(stat); 
        
        // Getting the x and y vectors
        Map<String, Double> x = CorrelationManager.getXValues(input, out);
        Map<String, Double> y = CorrelationManager.getYValues(input, out);

        // Writing output file
        writeOutput(input, x, y, out);
        
		// Creating the process
        process = new CorrelationAnalysis<String, TauResult>(corr);
        process.setXValues(x);
        process.setYValues(y);
	} 
	
    private void writeOutput(ManagerInput<String, String> input, 
                             Map<String, Double> xValues, 
                             Map<String, Double> yValues, 
                             PrintWriter out) {

        // Getting the output file name
        String fileName = input.getOptionalValue(OUTPUT_FILE_NAME);
        
        if (fileName != null) {
            CorrelationManager.writeIntersection(fileName, xValues, yValues, out);
        }

    }


	private void checkState() {
		
		if (process == null) {
			String s = "No input was provided.";
			throw new IllegalStateException(s);
		}
	}
	
    /**
     * Executes the process this process manager controls
     * and writes the results to the provided print writer.
     *
     * @param out the print writer where the results will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager.
     */
	public void execute(PrintWriter out) {

		checkState();
		
		// Creating results manger
		SingleRm<TauResult> rm = new SingleRm<TauResult>();
		
		// Running process
		process.run(rm);
		
		// Retrieving results
		TauResult results = rm.getResults();
		 
		// Writing results
		out.println();
		out.println("Statistic = " + results.getCoeff());
		out.println("Approx. Z = " + results.getApproxZ());
		out.println("N         = " + results.getN());
	}

}
