/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.stat;

import java.io.PrintWriter;
import java.util.*;

import sbrt.kernel.io.*;
import sbrt.kernel.math.stat.comp.*;
import sbrt.kernel.math.stat.proc.MwUAnalysis;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.*;

/**
 * This class is used to manage Mann-Whitney tests.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public final class MannWhitneyUManager extends AbstractProcessManager implements KernelProcessManager<String, String> {

	/**
	 * The keyword used to denote the name of the file containing sample A.
	 */
	public static final String SAMPLE_A_FILE = "Sample A File";

	/**
	 * The keyword used to denote the name of the file containing sample B.
	 */
	public static final String SAMPLE_B_FILE = "Sample B File";

	/**
	 * The set of keywords required by this process manager.
	 */
	public static final UnmodifiableSet<String> requiredKeywords;
	static {

		Set<String> s = new TreeSet<String>();

		s.add(SAMPLE_A_FILE);
		s.add(SAMPLE_B_FILE);

		requiredKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * The set of keywords understood, but not required, by this process
	 * manager.
	 */
	public static final UnmodifiableSet<String> optionalKeywords;
	static {

		Set<String> s = Collections.emptySet();
		optionalKeywords = new UnmodifiableSet<String>(s);
	}

	/**
	 * Returns the list of values contained in the file with the provided name.
	 * The values must be parsable using {@linkplain DoubleFormatV1};
	 * 
	 * @param fileName
	 *            the name of the file containing a list of values.
	 * @param out
	 *            the print writer to which status messages will be printed.
	 * 
	 * @return the list of values contained in the file with the provided name.
	 */
	public static List<Double> getSample(String fileName, PrintWriter out) {

		// Printing an opening message
		out.print("Reading '" + fileName + "' ... ");
		out.flush();

		List<Double> sample = getSample(fileName);

		// Printing a closing message
		out.println("DONE");

		return sample;
	}

	/**
	 * Returns the list of values contained in the file with the provided name.
	 * The values must be parsable using {@linkplain DoubleFormatV1};
	 * 
	 * @param fileName
	 *            the name of the file containing a list of values.
	 * 
	 * @return the list of values contained in the file with the provided name.
	 */
	public static List<Double> getSample(String fileName) {

		// Creating number parser
		DoubleFormat f = new DoubleFormatV1();

		// Opening file
		InputFile<Double> file = new TextInputFile<Double>(fileName, f);

		// Creating storage
		List<Double> values = new LinkedList<Double>();

		// Reading file
		for (Double v : file) {
			values.add(v);
		}

		// Closing file
		file.close();

		return values;
	}

	// The process this process manager executes
	private MwUAnalysis<String> process;

	/**
	 * Returns the set of keywords required by this process manager.
	 * 
	 * @return the set of keywords required by this process manager.
	 */
	public UnmodifiableSet<String> getRequiredKeys() {
		return requiredKeywords;
	}

	/**
	 * Returns the set of keywords understood, but not required, by this
	 * process manager.
	 * 
	 * @return the set of keywords understood, but not required, by this
	 *         process manager.
	 */
	public UnmodifiableSet<String> getOptionalKeys() {
		return optionalKeywords;
	}

	/**
	 * Returns the process controlled by this process manager.
	 * 
	 * @return the process controlled by this process manager.
	 */
	public MwUAnalysis<String> getProcess() {
		checkState();
		return process;
	}

	/**
	 * Uses the provided input to prepare this process manager for
	 * execution. As the input is processed, informative messages are printed to
	 * the provided print writer.
	 * 
	 * @param input
	 *            all of the information required to prepare this process
	 *            manager for execution.
	 * 
	 * @param out
	 *            the print writer to which all messages will be printed.
	 */
	public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

		// Getting sample file names
		String fileA = input.getRequiredValue(SAMPLE_A_FILE);
		String fileB = input.getRequiredValue(SAMPLE_B_FILE);

		// Reading files
		List<Double> sampleA = getSample(fileA, out);
		List<Double> sampleB = getSample(fileB, out);

		// Creating the Mann-Whitney U tester
		process = new MwUAnalysis<String>();

		process.setSampleA(sampleA);
		process.setSampleB(sampleB);
	}

	private void checkState() {
		
		if (process == null) {
			String s = "No input was provided.";
			throw new IllegalStateException(s);
		}
	}
	
	/**
	 * Executes the process this process manager controls and writes the
	 * results to the provided print writer.
	 * 
	 * @param out
	 *            the print writer where the results will be printed.
	 * 
	 * @throws IllegalStateException
	 *             if no input was provided to this process manager.
	 */
	public void execute(PrintWriter out) {

		checkState();
		
		// Creating results manger
		SingleRm<MannWhitneyResults> rm = new SingleRm<MannWhitneyResults>();

		// Running process
		process.run(rm);

		// Retrieving results
		MannWhitneyResults results = rm.getResults();

		// Writing results
		out.println();
		out.println("U  = " + results.getU());
		out.println("n1 = " + results.getN1());
		out.println("n2 = " + results.getN2());

		if (!Double.isNaN(results.getApproxT())) {
			out.println("Approx. t = " + results.getApproxT());	
		}
		
	}

}
