/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.utilities;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.io.AbstractResultsManager;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.IntervalComparer.Diff;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.kernel.utilities.proc.IntervalComparison;
import sbrt.shell.io.ComiFile;
import sbrt.shell.io.MapInputFile;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.KernelProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.ListFormat;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.NonEmptyStringFormat;
import sbrt.shell.text.PipeListFormat;
import sbrt.shell.text.SimpleListFormat;
import sbrt.shell.text.math.IntervalFormat;
import sbrt.shell.text.math.IntervalFormatV1;

/**
 * This class is used to compare the intervals contained in separate files.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class IntervalComparisonManager extends AbstractProcessManager
        implements KernelProcessManager<String, String> {

    /**
     * The string used to indicate the name of the first file.
     */
    public static final String FILE_1 = "File 1";

    /**
     * The string used to indicate the name of the second file.
     */
    public static final String FILE_2 = "File 2";

    /**
     * The set of keywords required by this process manager.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {

        Set<String> s = new TreeSet<String>();

        s.add(FILE_1);
        s.add(FILE_2);
        s.add(ZERO_CUTOFF);
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by this process
     * manager.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {

        Set<String> s = Collections.emptySet();
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    private static IntervalFormat getIntervalFormat() {
        return new IntervalFormatV1();
    }
    
    private static MapFormat<String, Interval> getFormat() {
        
        // Creating text parsers
        NonEmptyStringFormat keyFormat = new NonEmptyStringFormat();
        IntervalFormat valueFormat = getIntervalFormat();
        
        return new EqualsMapFormatV1<String, Interval>(keyFormat, 
                                                       valueFormat);
    }
    
    private static MapInputFile<String, Interval> 
        getInputFile(String fileName, PrintWriter out) {

        out.print("Opening '" + fileName + "' ... ");

        MapFormat<String, Interval> mapFormat = getFormat();
        
        MapInputFile<String, Interval> file = 
            new ComiFile<String, Interval>(fileName, mapFormat);
        
        out.println("DONE");
        
        return file;
    }
    
    private static OutputFile<List<String>> 
        getOutputFile(String fileName, PrintWriter out) {

        out.print("Opening '" + fileName + "' ... ");
        
        ListFormat<String> listFormat = PipeListFormat.getBasic();

        SimpleListFormat<String> format = 
            new SimpleListFormat<String>(listFormat);
        
        OutputFile<List<String>> file = 
            new TextOutputFile<List<String>>(fileName, format);
        
        out.println("DONE");
        
        return file;
    }
    
    /**
     * Returns the set of keywords required by this process manager.
     * 
     * @return the set of keywords required by this process manager.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords understood, but not required, by this process
     * manager.
     * 
     * @return the set of keywords understood, but not required, by this process
     * manager.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    private MapInputFile<String, Interval> file1;
    private MapInputFile<String, Interval> file2;
    private OutputFile<List<String>> outputFile;
    
    private IntervalComparison<String> process;

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

		// Checking input
		check(input);

        // Getting input file name 1
        String fileName1 = input.getRequiredValue(FILE_1);
        
        // Getting input file name 1
        String fileName2 = input.getRequiredValue(FILE_2);

        // Getting output file name
        String outputFileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        double zeroCutoff = getZeroCutoff(input, out);
 
        out.println();
        
        file1 = getInputFile(fileName1, out);
        file2 = getInputFile(fileName2, out);
    
        outputFile = getOutputFile(outputFileName, out);
        
        // Creating process
        process = new IntervalComparison<String>();
        process.setTolerance(zeroCutoff);
        process.setIntervalSources(file1, file2);
        
    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public IntervalComparison<String> getProcess() {
        return process;
    }
    
    private void checkState() {

        if (file1 == null) {
            String s = "No input was provided.";
            throw new IllegalStateException(s);
        }
        
    }

    /**
     * Executes the process this process manager controls and writes informative
     * messages to the provided print writer.
     * 
     * @param out the print writer where informative messagse will be written.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager.
     */
    public void execute(PrintWriter out) {

        checkState();

        Rm rm = new Rm(outputFile);
        
        runProcess(process, rm, out);
        
        closeFile(outputFile, out);
        closeFile(file2, out);
        closeFile(file1, out);

    }

    private static class Rm extends AbstractResultsManager<Set<Diff<String>>> {

        private OutputFile<List<String>> outputFile;
        private int n;
        
        private IntervalFormat format;
        
        private List<String> list;
        
        /**
         * Constructs a new results manager.
         * @param outputFile the output file.
         */
        protected Rm(OutputFile<List<String>> outputFile) {
            this.outputFile = outputFile;
            n = 0;
            format = new IntervalFormatV1();
            list = new ArrayList<String>();
        }

        /**
         * Adds data to this results manager.
         * 
         * @param <S> a type extending the primary result type.
         * @param results the results, or data, to be added
         * to this manager.
         * 
         * @throws IllegalStateException if {@code close()}
         * has already been called.
         */
        public <S extends Set<Diff<String>>> void addData(S results) {

            ++ n;
            
            if (results.isEmpty()) {
                return;
            }
            
            for (Diff<String> d : results) {

                list.clear();
                
                list.add("Vector_" + n);
                list.add(d.getVariable());
                list.add(format.format(d.getInterval1()));
                list.add(format.format(d.getInterval2()));
                
                outputFile.addData(list);
            }

        }
        
    }
}
