/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.utilities;

import java.util.Set;
import java.io.PrintWriter;
import java.util.*;

import sbrt.kernel.io.*;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.utilities.*;
import sbrt.kernel.utilities.proc.*;
import sbrt.shell.io.*;
import sbrt.shell.mng.*;
import sbrt.shell.text.*;
import sbrt.shell.text.math.*;

/**
 * This class is used to manage variable participation processes.
 * 
 * @author This class was written and documented by Jeremiah Wright while
 * in the Wagner lab.
 *
 */
public final class VariableParticipationManager extends AbstractProcessManager implements KernelProcessManager<String, String> {

	/**
	 * The keyword used to denote the input file containing mathematical
	 * expressions.
	 */
    public static final String EXPRESSION_FILE = "Expression File";

    /** 
     * The set of keywords required to be present in the {@code ManagerInput}
     * objects supplied to instances of this class.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {
        Set<String>  s = new TreeSet<String> ();

        s.add(EXPRESSION_FILE);

        // Output file information
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String> (s);
    }

    /** 
     * The set of keywords understood, but not required, by instances of this
     * class.
     */
    public static final UnmodifiableSet<String>  optionalKeywords;
    static {
    	Set<String> s = Collections.emptySet();
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * Opens the file denoted by {@linkplain #EXPRESSION_FILE} in the 
     * provided manager input and returns the list of mathematical expressions
     * it contains. The file must be a text file and the expressions it
     * contains must be parsable with {@linkplain LinearCombFormatV1}.
     * 
     * @param input the input containing a value for {@code EXPRESSION_FILE}.
     * @param out the print writer to which a status message will be printed.
     * 
     * @return the list of mathematical expressions contained in the file 
     * indicated by {@linkplain #EXPRESSION_FILE} in the provided input.
     */
    public static List<LinearComb<String>> 
    getExpressions(ManagerInput<String, String> input, PrintWriter out) {

    	// Getting file name
        String fileName = input.getRequiredValue(EXPRESSION_FILE);

        // Printing opening message
        out.print("Reading '" + fileName + "' ... ");
        out.flush();

        // Creating linear combination format
        LinearCombFormat<String> expFormat =
            new LinearCombFormatV1<String>(new NonEmptyStringFormat());

        // Opening input file
        InputFile<LinearComb<String>> file = 
        	new TextInputFile<LinearComb<String>>(fileName, expFormat);

        // Creating storage
        List<LinearComb<String>> expressions = 
        	new LinkedList<LinearComb<String>>();

        // Iterating thru the file
        for (LinearComb<String> e : file) {
            expressions.add(e);
        }

        // Printing a closing message
        out.println("DONE");

        // Returning the linear combinations contained in the file
        return expressions;
    }

    private static OutputFile<String> 
    getOutputFile(ManagerInput<String, String> input,
    		 	  PrintWriter out) {

    	// Getting the file name
        String fileName = input.getRequiredValue(OUTPUT_FILE_NAME);

        // Printing an opening message
        out.print("Opening '" + fileName + "' ... ");
        out.flush();

        // Creating the output file
        OutputFile<String> file = TextOutputFile.getBasic(fileName);

        // Printing a closing message
        out.println("DONE");

        return file;
    }

    private VariableParticipation<LinearComb<String>, String> process;
    private OutputFile<String> outputFile;

	/**
	 * Returns the set of keywords the provided 
	 * {@code ManagerInput} object must contain.
	 *  
	 * @return the set of keywords the provided
	 * {@code ManagerInput} object must contain.
	 */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     * 
     * @return the set of keywords the provided
     * {@code ManagerInput} object does not necessarily
     * have to contain.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    /**
     * Returns the process this process manager controls.
     * @return the process this process manager controls.
     * 
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public VariableParticipation<LinearComb<String>, String> getProcess() {
        checkState();
        return process;
    }

    private void closeFile(PrintWriter out) {

        if (outputFile != null) {
            closeFile(outputFile, out);
        }

    }

    private void checkState() {
        if (process == null) {
            throw new IllegalStateException("No manager input was provided.");
        }
    }

    /**
     * Uses the provided input to prepare this process
     * manager for execution. As the input is processed,
     * informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to
     * prepare this process manager for execution.
     * 
     * @param out the print writer to which all messages
     * will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        // Closing previous file if necessary
        closeFile(out);

        // Getting expressions
        List<LinearComb<String>> expressions = getExpressions(input, out);
        
        // Creating process
        process = new VariableParticipation<LinearComb<String>, String>();
        process.setExprs(expressions);

        // Setting output file
        outputFile = getOutputFile(input, out);
    }

    /**
     * Executes the process this process manager controls,
     * and writes the status or results of this process to the
     * provided print writer.
     *
     * @param out the print writer where status messages will be printed. 
     *
     * @throws IllegalStateException if no input was provided
     * to this process manager via {@linkplain #setInput(ManagerInput)} or 
     * {@linkplain #setInput(ManagerInput, PrintWriter)}
     */
    public void execute(PrintWriter out) {
    	
        checkState();

        try {

            // Creating simple results manager
            Rm rm = new Rm(outputFile);

            // Running process
            process.run(rm);

        } finally {
            closeFile(out);
        }
    }

    private static final class Rm 
    	extends AbstractResultsManager<Map<String, List<LinearComb<String>>>> {
    	
    	private final OutputFile<String> outputFile;
    	
    	private Rm(OutputFile<String> outputFile) {
    		this.outputFile = outputFile;
    	}

    	/**
    	 * Adds data to this results manager.
    	 * 
    	 * @param participation the map of variable participation.
    	 * 
    	 * @throws IllegalStateException if {@code close()}
    	 * has already been called.
    	 */
    	public void 
    	addData(Map<String, List<LinearComb<String>>> participation) {

            // Creating reaction name expression formatter
            LinearCombFormat<String> expFormat =
                    new LinearCombFormatV1<String>(new NonEmptyStringFormat());

            // Writing results to file
            for (Map.Entry<String, List<LinearComb<String>>> e : 
            	participation.entrySet()) {

                // Getting next entry
                String variable = e.getKey();
                List<LinearComb<String>> expressions = e.getValue();

                // Writing the variable
                outputFile.addData(variable + ":");

                // Writing the expressions in which it is contained
                for (LinearComb<String> j : expressions) {
                    outputFile.addData(expFormat.format(j));
                }

                // Writing a blank line
                outputFile.addData("");
            }

    	}
    }

}
