/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.mng.utilities;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import sbrt.kernel.io.ListResultsManager;
import sbrt.kernel.utilities.UnmodifiableSet;
import sbrt.kernel.utilities.proc.VectorComparison;
import sbrt.shell.io.ComiFile;
import sbrt.shell.io.ComoFile;
import sbrt.shell.io.MapOutputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.mng.KernelProcessManager;
import sbrt.shell.mng.ManagerInput;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapFormatV1;
import sbrt.shell.text.MapFormat;
import sbrt.shell.text.NonEmptyStringFormat;

/**
 * This class is used to compare the intervals contained in separate files.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class VectorComparisonManager extends AbstractProcessManager
        implements KernelProcessManager<String, String> {

    /**
     * The string used to indicate the name of the vector file.
     */
    public static final String FILE = "File";

    /**
     * The set of keywords required by this process manager.
     */
    public static final UnmodifiableSet<String> requiredKeywords;
    static {

        Set<String> s = new TreeSet<String>();

        s.add(FILE);
        s.add(ZERO_CUTOFF);
        s.add(OUTPUT_FILE_NAME);

        requiredKeywords = new UnmodifiableSet<String>(s);
    }

    /**
     * The set of keywords understood, but not required, by this process
     * manager.
     */
    public static final UnmodifiableSet<String> optionalKeywords;
    static {
        Set<String> s = Collections.emptySet();
        optionalKeywords = new UnmodifiableSet<String>(s);
    }

    private static MapFormat<String, Double> getFormat() {

        // Creating text parsers
        NonEmptyStringFormat keyFormat = new NonEmptyStringFormat();
        DoubleFormat valueFormat = new DoubleFormatV1();

        return new EqualsMapFormatV1<String, Double>(keyFormat, valueFormat);
    }

    private static ComiFile<String, Double> getInputFile(String fileName,
            PrintWriter out) {

        out.print("Opening '" + fileName + "' ... ");

        MapFormat<String, Double> mapFormat = getFormat();

        ComiFile<String, Double> file =
                new ComiFile<String, Double>(fileName, mapFormat);

        out.println("DONE");

        return file;
    }

    private static MapOutputFile<String, Double> getOutputFile(String fileName,
            Set<String> columnHeaders,
            PrintWriter out) {

        out.print("Opening '" + fileName + "' ... ");

        MapFormat<String, Double> format =
                new EqualsMapFormatV1<String, Double>(new NonEmptyStringFormat(),
                        new DoubleFormatV1());

        MapOutputFile<String, Double> file =
                new ComoFile<String, Double>(fileName, columnHeaders, format);

        out.println("DONE");

        return file;
    }

    /**
     * Returns the set of keywords required by this process manager.
     * 
     * @return the set of keywords required by this process manager.
     */
    public UnmodifiableSet<String> getRequiredKeys() {
        return requiredKeywords;
    }

    /**
     * Returns the set of keywords understood, but not required, by this process
     * manager.
     * 
     * @return the set of keywords understood, but not required, by this process
     * manager.
     */
    public UnmodifiableSet<String> getOptionalKeys() {
        return optionalKeywords;
    }

    private ComiFile<String, Double> file1;

    private MapOutputFile<String, Double> output;

    private VectorComparison<String> process;

    /**
     * Uses the provided input to prepare this process manager for execution. As
     * the input is processed, informative messages are printed to the provided
     * print writer.
     * 
     * @param input all of the information required to prepare this process
     * manager for execution.
     * 
     * @param out the print writer to which all messages will be printed.
     */
    public void setInput(ManagerInput<String, String> input, PrintWriter out) {

        // Checking input
        check(input);

        double zeroCutoff = getZeroCutoff(input, out);

        out.println();

        // Getting input file name
        String fileName = input.getRequiredValue(FILE);
        file1 = getInputFile(fileName, out);

        String outputName = input.getRequiredValue(OUTPUT_FILE_NAME);
        output = getOutputFile(outputName, file1.getKeys(), out);

        process = new VectorComparison<String>();
        process.setTolerance(zeroCutoff);
        process.setVectorSource(file1);
    }

    /**
     * Returns the process managed by this object.
     * @return the process managed by this object.
     */
    public VectorComparison<String> getProcess() {
        return process;
    }

    private void checkState() {

        if (file1 == null) {
            String s = "No input was provided.";
            throw new IllegalStateException(s);
        }

    }

    /**
     * Executes the process this process manager controls and writes informative
     * messages to the provided print writer.
     * 
     * @param out the print writer where informative messagse will be written.
     * 
     * @throws IllegalStateException if no input was provided to this process
     * manager.
     */
    public void execute(PrintWriter out) {

        checkState();

        ListResultsManager<Map<String, Double>> rm =
                new ListResultsManager<Map<String, Double>>();

        runProcess(process, rm, out);

        for (Map<String, Double> m : rm) {
            output.addData(m);
        }

        closeFile(output, out);
        closeFile(file1, out);
    }
    
}
