/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.proc.fba;


import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.IrrevRxn;
import sbrt.kernel.fba.RevRxn;
import sbrt.kernel.math.LinearComb;
import sbrt.shell.io.InputFile;
import sbrt.shell.io.TextInputFile;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.WhitespaceListFormat;
import sbrt.shell.text.math.LinearCombFormatV2;

/**
 * This class is used to parse Metatool input files for the purpose of
 * converting them to FBA Reaction Files.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 *
 */
public class MetatoolInputReader {
	
	/**
	 * The Metatool keyword {@code -ENZREV}.
	 */
	public static final String ENZREV = "-ENZREV";

	/**
	 * The Metatool keyword {@code -ENZIRREV}.
	 */
	public static final String ENZIRREV = "-ENZIRREV";

	/**
	 * The Metatool keyword {@code -METINT}.
	 */
	public static final String METINT = "-METINT";

	/**
	 * The Metatool keyword {@code -METEXT}.
	 */
	public static final String METEXT = "-METEXT";

	/**
	 * The Metatool keyword {@code -CAT}.
	 */
	public static final String CAT = "-CAT";

	private static final String[] KEYWORDS =
		{ ENZREV, ENZIRREV, METINT, METEXT, CAT };
	
	private static final WhitespaceListFormat<String> listFormat;
	private static final LinearCombFormatV2<String> linCombFormat;
	
	static {
		listFormat = WhitespaceListFormat.getBasic();
		linCombFormat = LinearCombFormatV2.getBasic();
	}
	
	private MetatoolInputReader() {
		// Prevents instantiation
	}
	
	/**
	 * Parses the lines of a Metatool input file.
	 * @param fileLines a list of all lines contained in a Metatool input file.
	 * @return the results of parsing.
	 */
	public static Results parse(List<String> fileLines) {
		
		// Ensuruing all keywords are present and in their 
		// appropriate locations
		checkKeywords(fileLines);
		
		// Initializing storage
		List<String> enzRev = new ArrayList<String>();
		List<String> enzIrrev = new ArrayList<String>();
		List<String> metInt = new ArrayList<String>();
		List<String> metExt = new ArrayList<String>();
		Map<String, ChemRxn> cat = new LinkedHashMap<String, ChemRxn>();
		
		// Initializing a list index
		int i = 0;
		
		// Parsing ENZREV
		i = parseList(i, fileLines, ENZREV, ENZIRREV, enzRev);

		// Parsing ENZIRREV
		i = parseList(i, fileLines, ENZIRREV, METINT, enzIrrev);
		
		// Parsing METINT
		i = parseList(i, fileLines, METINT, METEXT, metInt);
		
		// Parsing METEXT
		i = parseList(i, fileLines, METEXT, CAT, metExt);
		
		// Parsing CAT
		parseCat(i, fileLines, enzRev, enzIrrev, cat);
		
		// Returning the results of parsing
		return new Results(enzRev, enzIrrev, metInt, metExt, cat);
	}
	
	private static boolean is(List<String> lines, int index, String keyword) {
		
		if (!validIndex(index, lines)) {
			return false;
		}
		
		return keyword.equals(lines.get(index));
	}
	
	private static void checkKeywords(List<String> lines) {

		int[] keywordIndices = new int[KEYWORDS.length];
		
		for (int i = 0; i < KEYWORDS.length; i ++) {
			
			// Finding the index of the 1st occurrence of this keyword
			keywordIndices[i] = lines.indexOf(KEYWORDS[i]); 

			if (keywordIndices[i] < 0) {
				missingKeyword(KEYWORDS[i]);
			}
			
			if (keywordIndices[i] != lines.lastIndexOf(KEYWORDS[i])) {
				duplicateKeyword(KEYWORDS[i]);
			}
			
		}
		
		int largestIndex = -1;
		
		for (int index : keywordIndices) {
			
			if (index > largestIndex) {
				largestIndex = index;
			}
			
			else {
				outOfOrder(lines.get(index), lines.get(largestIndex));
			}
		}
	
		
		
	}
	
	
	private static boolean validIndex(int index, List<String> lines) {
		return index < lines.size() && index >= 0;
	}
	
	
	private static int parseList(int startIndex,
						  		 List<String> fileLines,
						  		 String keyword,
						  		 String nextKeyword,
						  		 List<String> list) {
		
		int index = startIndex;
		
		if ( !is(fileLines, index, keyword)) {
			misplacedKeyword(keyword);
		}

		++ index;

		if (!validIndex(index, fileLines)) {
			return index;
		}
		
		if ( is(fileLines, index, nextKeyword)) {
			return index;
		}

		List<String> entries = listFormat.parse(fileLines.get(index));
		
		list.addAll(entries);
		
		++ index;
		
		return index;
	}
	
	
	
	private static void
	parseCat(int startIndex, 
			 List<String> lines,
			 List<String> enzRev,
			 List<String> enzIrrev,
			 Map<String, ChemRxn> cat) {
		
		int index = startIndex;
		
		if ( !is(lines, index, CAT)) {
			misplacedKeyword(CAT);
		}

		++ index;
		
		if (!validIndex(index, lines)) {
			return;
		}
		
		Iterator<String> i = lines.listIterator(index);
		
		while(i.hasNext()) {
			parseRxn(i.next(), enzRev, enzIrrev, cat); 
		}
		
	}
	
	private static void parseRxn(String line, 
			 			  		 List<String> enzRev, 
			 			  		 List<String> enzIrrev,
			 			  		 Map<String, ChemRxn> rxns) {
		
		String[] parts = line.split("\\s+:\\s+");
		
		if (parts.length != 2) {
			badRxn(line);
		}
		
		String rxnName = parts[0];
		
		parts = parts[1].split("\\s+=\\s+");
		
		if (parts.length != 2) {
			badRxn(line);
		}
		
		LinearComb<String> reactants = linCombFormat.parse(parts[0]);
		LinearComb<String> products = linCombFormat.parse(parts[1]);

		ChemRxn rxn = null;
		
		if (enzRev.contains(rxnName)) {
			rxn = new RevRxn(reactants, products);
		}
		
		else if (enzIrrev.contains(rxnName)) {
			rxn = new IrrevRxn(reactants, products);
		}
		
		else {
			badRxnName(rxnName);
		}

		if (rxns.put(rxnName, rxn) != null) {
			duplicateRxnName(rxnName);
		}

	}
	
	private static void duplicateKeyword(String keyword) {
		error("The keyword '" + keyword + 
		  	  "' appears more than once.");
	}
	
	private static void missingKeyword(String keyword) {
		error("The keyword '" + keyword + "' is not present.");
	}
	
	private static void outOfOrder(String keyword1, String keyword2) {
		error("The keywords '" + keyword1 + "' and '" + keyword2 +
		   	  "' are out of order.");
	}
	
	private static void misplacedKeyword(String keyword) {
		error("The keyword '" + keyword + 
			  "' is not in its expected location.");
	}
	
	private static void duplicateRxnName(String name) {
		error("The reaction name '" + name + 
			  "' occurs more than once in the '" +
			  CAT + "' section.");
	}
	
	private static void badRxnName(String name) {
		error("The reaction name '" + name + "' in the '" + CAT +
				"' section is not listed under '" +
				ENZREV + "' or '" + ENZIRREV + "'.");
	}
	
	private static void badRxn(String line) {
		error("The reaction '" + line + "' is improperly formatted.");
	}
	
	private static void error(String msg) {
		throw new FormatException(msg);
	}
	
	/**
	 * This class is used to store the results of parsing a Metatool
	 * input file. 
	 * 
	 * @author This class was written and documented by Jeremiah Wright 
	 * while in the Wagner lab.
	 */
	public static class Results {

		private final List<String> enzRev;
		private final List<String> enzIrrev;
		private final List<String> metInt;
		private final List<String> metExt;
		private final Map<String, ChemRxn> cat;
		
		private Results(List<String> enzRev,
						List<String> enzIrrev,
						List<String> metInt,
						List<String> metExt,
						Map<String, ChemRxn> cat) {
			
			this.enzRev = enzRev;
			this.enzIrrev = enzIrrev;
			this.metInt = metInt;
			this.metExt = metExt;
			this.cat = cat;
		}
		
		/**
		 * Returns the reaction names that are listed in the {@code -ENZREV}
		 * section of the input file.
		 * 
		 * @return the reaction names that are listed in the {@code -ENZREV}
		 * section of the input file.
		 */
		public List<String> getEnzRev() {
			return new ArrayList<String>(enzRev);
		}
		
		/**
		 * Returns the reaction names that are listed in the {@code -ENZIRREV}
		 * section of the input file.
		 * 
		 * @return the reaction names that are listed in the {@code -ENZIRREV}
		 * section of the input file.
		 */
		public List<String> getEnzIrrev() {
			return new ArrayList<String>(enzIrrev);
		}
		
		/**
		 * Returns the chemical species that are listed in the {@code -METINT}
		 * section of the input file.
		 * 
		 * @return the reaction names that are listed in the {@code -METINT}
		 * section of the input file.
		 */
		public List<String> getMetInt() {
			return new ArrayList<String>(metInt);
		}

		/**
		 * Returns the chemical species that are listed in the {@code -METEXT}
		 * section of the input file.
		 * 
		 * @return the reaction names that are listed in the {@code -METEXT}
		 * section of the input file.
		 */
		public List<String> getMetExt() {
			return new ArrayList<String>(metExt);
		}
		
		/**
		 * Returns the reactions that are listed in the {@code -CAT}
		 * section of the input file.
		 * 
		 * @return reaction names mapped to their corresponding chemical
		 * reaction.
		 */
		public Map<String, ChemRxn> getCat() {
			return new LinkedHashMap<String, ChemRxn>(cat);
		}
		
	}
	
	
	/**
	 * Used for testing purposes.
	 * @param args
	 */
	public static void main(String[] args) {
		
		InputFile<String> file = TextInputFile.getBasic(args[0]);
		
		List<String> lines = new ArrayList<String>();
		
		for (String line : file) {
			lines.add(line);
		}
		
		file.close();
		
		Results results = parse(lines);
		
		System.out.println(results.getEnzRev());
		System.out.println(results.getEnzIrrev());
		System.out.println(results.getMetInt());
		System.out.println(results.getMetExt());
		System.out.println(results.getCat());
	}
}
