/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.proc.fba;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.IrrevRxn;
import sbrt.kernel.fba.IrrevRxnFluxome;
import sbrt.kernel.fba.MixedFluxome;
import sbrt.kernel.fba.RevExchRxn;
import sbrt.kernel.fba.RevRxn;
import sbrt.kernel.fba.SinkRxn;
import sbrt.kernel.fba.SourceRxn;
import sbrt.kernel.math.LinearComb;
import sbrt.shell.io.OutputFile;
import sbrt.shell.text.math.LinearCombFormatV2;

/**
 * This class is used to write input files for Metatool.
 */
public class MetatoolInputWriter {

	/** 
	 * The string used to denote the surroundings.
	 */
	public final static String SURR = "Surroundings";
    
	private final LinearCombFormatV2<String> exprFormat;
	
	/**
	 * Creates a new Metatool input file writer. 
	 */
	public MetatoolInputWriter() {
		exprFormat = LinearCombFormatV2.getBasic();
	}
	
    /**
     * Writes a metatool input file.
     * @param fluxome the network to be written to file.
     * @param file the metatool input file.
     */
    public void writeFile(IrrevRxnFluxome fluxome, OutputFile<String> file) {
        
        Map<String, ChemRxn> rxns = fluxome.getRxns();
        Set<String> species = fluxome.getSpecies();
        
        writeEnzRev(file, rxns);
        writeEnzIrrev(file, rxns);
        writeMetInt(file, species);
        writeMetExt(file);
        writeCat(file, rxns);
        
    }
    
    /**
     * Writes a metatool input file.
     * @param fluxome the network to be written to file.
     * @param file the metatool input file.
     */
    public void writeFile(MixedFluxome fluxome, OutputFile<String> file) {
        
        Map<String, ChemRxn> rxns = fluxome.getRxns();
        Set<String> species = fluxome.getSpecies();
        
        writeEnzRev(file, rxns);
        writeEnzIrrev(file, rxns);
        writeMetInt(file, species);
        writeMetExt(file);
        writeCat(file, rxns);
    }
    
    private void writeMetExt(OutputFile<String> file) {
        file.addData("-METEXT");
        file.addData(SURR);
        file.addData("");
    }
    
    
    private void writeMetInt(OutputFile<String> file, Set<String> species) {
    	
        file.addData("-METINT");

    	StringBuilder sb = new StringBuilder(1000);
        
        Iterator<String> i = species.iterator();
        
        while (i.hasNext()) {
            
            sb.append(i.next());
            
            if (i.hasNext()) {
                sb.append(" ");
            }
        }
        
        file.addData(sb.toString());
        file.addData("");
    }
    
    private void writeCat(OutputFile<String> file, Map<String, ChemRxn> rxns) {
        
        file.addData("-CAT");

        for (Map.Entry<String, ChemRxn> e: rxns.entrySet()) {
        	writeRxn(e.getKey(), e.getValue(), file);
        }

    }

    private void writeRxn(String id, ChemRxn rxn, OutputFile<String> file) {
    	
    	if (rxn instanceof IrrevRxn) {
    		writeRxn(id, (IrrevRxn) rxn, file);
    	}
    	
    	else if (rxn instanceof RevRxn) {
    		writeRxn(id, (RevRxn) rxn, file);
    	}
    	
    	else if (rxn instanceof SourceRxn) {
    		writeRxn(id, (SourceRxn) rxn, file);
    	}
    	
    	else if (rxn instanceof SinkRxn) {
    		writeRxn(id, (SinkRxn) rxn, file);
    	}
    	
    	else if (rxn instanceof RevExchRxn) {
    		writeRxn(id, (RevExchRxn) rxn, file);
    	}
    	
    	else {
    		throw new IllegalStateException();
    	}
    	
    }
    
    private void writeRxn(String id, RevExchRxn rxn, OutputFile<String> file) {

        StringBuilder sb = new StringBuilder(100);
    	
    	sb.append(id);
    	sb.append(" : ");
    	
    	LinearComb<String> reactants = rxn.getReactants();
    	
    	if (reactants.hasVariables()) {
    		writeStoichExpr(sb, reactants);
    		sb.append(" = ");
    		sb.append(SURR);
    	}

    	else {
    		sb.append(SURR);
    		sb.append(" = ");
    		writeStoichExpr(sb, rxn.getProducts());
    	}
    	
    	file.addData(sb.toString());
    }
    
    private void writeRxn(String id, SinkRxn rxn, OutputFile<String> file) {

        StringBuilder sb = new StringBuilder(100);
    	
    	sb.append(id);
    	sb.append(" : ");
    	
    	writeStoichExpr(sb, rxn.getReactants());
    	
    	sb.append(" = ");
    	sb.append(SURR);
    	
    	file.addData(sb.toString());
    }
    
    private void writeRxn(String id, SourceRxn rxn, OutputFile<String> file) {

        StringBuilder sb = new StringBuilder(100);
    	
    	sb.append(id);
    	sb.append(" : ");
    	sb.append(SURR);
    	sb.append(" = ");
    	
    	writeStoichExpr(sb, rxn.getProducts());
    	
    	file.addData(sb.toString());
    }
    
    private void writeRxn(String id, RevRxn rxn, OutputFile<String> file) {

        StringBuilder sb = new StringBuilder(100);
    	
    	sb.append(id);
    	sb.append(" : ");
    	
    	writeStoichExpr(sb, rxn.getReactants());
    	
    	sb.append(" = ");
    	
    	writeStoichExpr(sb, rxn.getProducts());
    	
    	file.addData(sb.toString());
    }
    
    private void writeRxn(String id, IrrevRxn rxn, OutputFile<String> file) {

        StringBuilder sb = new StringBuilder(100);
    	
    	sb.append(id);
    	sb.append(" : ");
    	
    	writeStoichExpr(sb, rxn.getReactants());
    	
    	sb.append(" = ");
    	
    	writeStoichExpr(sb, rxn.getProducts());
    	
    	file.addData(sb.toString());
    }
    
    private void writeStoichExpr(StringBuilder sb, LinearComb<String> expr) {
    	sb.append(exprFormat.format(expr));
    }
    
    private void writeEnzIrrev(OutputFile<String> file, Map<String, ChemRxn> rxns) {
        
        file.addData("-ENZIRREV");

        StringBuilder sb = new StringBuilder(5000);
        
        Iterator<Map.Entry<String, ChemRxn>> i = 
        	rxns.entrySet().iterator();
        
        while (i.hasNext()) {
            
            Map.Entry<String, ChemRxn> e = i.next();
        	
            // Skipping reversible reactions
            if (e.getValue().isReversible()) {
            	continue;
            }
            
        	sb.append(e.getKey());
            
            if (i.hasNext()) {
                sb.append(" ");
            }
        }
        
        file.addData(sb.toString());
        file.addData("");
    }
    
    private void writeEnzRev(OutputFile<String> file, Map<String, ChemRxn> rxns) {
    	
        file.addData("-ENZREV");

        StringBuilder sb = new StringBuilder(5000);
        
        Iterator<Map.Entry<String, ChemRxn>> i = 
        	rxns.entrySet().iterator();
        
        while (i.hasNext()) {
            
            Map.Entry<String, ChemRxn> e = i.next();
        	
            // Skipping irreversible reactions
            if (!e.getValue().isReversible()) {
            	continue;
            }
            
        	sb.append(e.getKey());
            
            if (i.hasNext()) {
                sb.append(" ");
            }
        }
        
        file.addData(sb.toString());
        file.addData("");
    }


    
    
}
