/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.proc.fba;

import java.util.Iterator;
import java.util.List;

import sbrt.kernel.math.algebra.DoubleMatrix;
import sbrt.kernel.math.algebra.HashMatrix;
import sbrt.shell.io.InputFile;
import sbrt.shell.text.DoubleFormat;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.IntegerFormat;
import sbrt.shell.text.IntegerFormatV1;
import sbrt.shell.text.ListFormat;
import sbrt.shell.text.WhitespaceListFormat;

/**
 * This class is used to parse Metatool output files for the purpose of
 * creating SBRT compliant files.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 *
 */
public class MetatoolOutputReader {

	
	private static final String ELEMENTARY_MODES = "ELEMENTARY MODES";
	
	private static final String MATRIX_DIM = "matrix dimension r";

	private final IntegerFormat intParser;
	private final DoubleFormat numParser;
	private final ListFormat<Double> lineParser;
	private String fileName;
	
	/**
	 * Creates a new Metatool output file reader.
	 */
	public MetatoolOutputReader() {
		intParser = new IntegerFormatV1();
		numParser = new DoubleFormatV1();
		lineParser = new WhitespaceListFormat<Double>(numParser);
	}
	
	/**
	 * Returns the elementary mode matrix contained in the provided Metatool
	 * output file.
	 * @param file a file produced by Metatool.
	 * @param orderedRxnNames the reactions corresponding to each matrix column.
	 * @return the elementary mode matrix contained in the provided Metatool
	 * output file.
	 */
	public DoubleMatrix<String, String> 
	getElementaryModes(InputFile<String> file, List<String> orderedRxnNames) {
		
		fileName = file.getCanonicalName();
		boolean modesSection = false;
		
		int rows = -1;
		
		Iterator<String> i = file.iterator();
		
		while(i.hasNext()) {
			
			String line = i.next();
			
			if (line.startsWith(ELEMENTARY_MODES)) {
				modesSection = true;
			}
			
			if (modesSection && line.contains(MATRIX_DIM)) {

				line = line.replace(MATRIX_DIM, "");
				
				String[] parts = line.split("\\s+");

				if (parts.length != 3) {
					error();
				}

				// Getting the number of rows in the matrix
				rows = intParser.parsePrimitive(parts[0]);

				return parseMatrix(rows, i, orderedRxnNames);
			}
			
		}
		
		// If the file doesn't contain the elementary mode matrix
		error();
		
		return null; // Never reached
	}
	
	
	private DoubleMatrix<String, String> 
	parseMatrix(int rows, Iterator<String> i, List<String> rxnNames) {
		

		HashMatrix<String, String> m = new HashMatrix<String, String>();
		
		for (int r = 0; r < rows; r ++) {
			
			if (!i.hasNext()) {
				error();
			}
			
			// Parsing the line
			List<Double> elements = lineParser.parse(i.next());
			
			Iterator<String> j = rxnNames.iterator();
			Iterator<Double> k = elements.iterator();
			
			String rowIndex = "r" + r;
			
			// Adding to the matrix
			while (j.hasNext() && k.hasNext()) {
				
				String name = j.next();
				Double value = k.next();
				
				m.setElement(rowIndex, name, value);

			}
			
			// If any elements remain in either list
			if (j.hasNext() || k.hasNext()) {
				error();
			}
			
		}
		
		return m;
	}
	
	
	private void error() {
		String s = "The Metatool output file '" + fileName + 
				   "' could not be parsed.";
		throw new FormatException(s);
	}
	
}
