/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.proc.fba;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import sbrt.kernel.math.Interval;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.utilities.And;
import sbrt.kernel.utilities.Or;
import sbrt.shell.io.AbstractSbrtFile;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.io.XmlFileFactory;
import sbrt.shell.text.BoolRelFormatV1;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapEntryFormatV1;
import sbrt.shell.text.chem.IrrevRxnFormatV1;
import sbrt.shell.text.chem.RevRxnFormatV1;
import sbrt.shell.text.fba.PalssonRxnNodeParser;
import sbrt.shell.text.math.IntervalFormatV1;
import sbrt.shell.text.math.LinearCombFormatV1;

/**
 * This class is used to read SBML files downloaded directly from 
 * Dr. Bernhard Palsson's website.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public class PalssonSbmlReader {

	/**
	 * The string used to denote that a particular value was not provided
	 * in an SBML input file.
	 */
	public static final String NOT_PROVIDED = "Not_Provided";
	
    private final String forwardArrow;
    private final String reversibleArrow;
    
    private final LinearCombFormatV1<String> linFormat;
    private final EqualsMapEntryFormatV1<String, String> entryFormat;
    private final PalssonRxnNodeParser nodeParser;
    private final BoolRelFormatV1<String> boolFormat;
    private final DoubleFormatV1 numberFormat;
    private final IntervalFormatV1 intervalFormat;
    
    private final Document sbmlFile;
    
    /**
     * Constructs a new SBML file reader.
     * @param sbmlFileName the canonical name of the SBML file to be read.
     * @see AbstractSbrtFile#getCanonicalName(String)
     */
    public PalssonSbmlReader(String sbmlFileName) {
        
        sbmlFile = XmlFileFactory.getDOM(sbmlFileName);
        
        forwardArrow = IrrevRxnFormatV1.ARROW;
        reversibleArrow = RevRxnFormatV1.ARROW;
        
        linFormat = LinearCombFormatV1.getBasic();
        entryFormat = EqualsMapEntryFormatV1.getBasic();
        boolFormat = BoolRelFormatV1.getBasic();
        nodeParser = new PalssonRxnNodeParser();
        numberFormat = new DoubleFormatV1();
        intervalFormat = new IntervalFormatV1();
    }
    
    /**
     * Writes the reactions contained in the SBML file to a file with the
     * provided name.
     * 
     * @param rxnFileName the name of the file to generated.
     */
    public void writeRxns(String rxnFileName) {

        // Opening the output file
        OutputFile<String> file = TextOutputFile.getBasic(rxnFileName);
        
        // Getting list of all reaction nodes
        NodeList listOfReactions = sbmlFile.getElementsByTagName("reaction");

        // Getting the total number of reactions
        int length = listOfReactions.getLength();

        // Iterating thru each reaction contained in the SBML file
        for (int i = 0; i < length; i++) {

            // Getting the next reaction node
            Node reaction = listOfReactions.item(i);

            // Getting reaction name (or id)
            String rxnName = nodeParser.getRxnId(reaction);
            
            // Determining reversibility
            boolean reversible = nodeParser.isReversible(reaction);

            // Determing reactants and products
            LinearComb<String> reactants = nodeParser.getReactants(reaction);
            LinearComb<String> products = nodeParser.getProducts(reaction);

            if (reversible) {
                writeRevRxn(file, rxnName, reactants, products);
            }
            
            else {
                writeIrrevRxn(file, rxnName, reactants, products);
            }
        }

    }

    /**
     * Writes the genes contained in the SBML file to a file with the
     * provided name.
     * 
     * @param geneFileName the name of the file to generated.
     */
    public void writeGenes(String geneFileName) {

        // Opening the output file
        OutputFile<String> file = TextOutputFile.getBasic(geneFileName);
        
        // Getting list of all reaction nodes
        NodeList listOfReactions = sbmlFile.getElementsByTagName("reaction");

        // Getting the total number of reactions
        int length = listOfReactions.getLength();

        // Iterating thru each reaction contained in the SBML file
        for (int i = 0; i < length; i++) {

            // Getting the next reaction node
            Node reaction = listOfReactions.item(i);

            // Getting reaction name (or id)
            String rxnName = nodeParser.getRxnId(reaction);

            // Determing boolean logic rules for genes
            Or<And<String>> geneRules = nodeParser.getGenes(reaction);
            
            writeGenes(file, rxnName, geneRules);
        }

    }

    /**
     * Writes the proteins contained in the SBML file to a file with the
     * provided name.
     * 
     * @param proteinFileName the name of the file to generated.
     */
    public void writeProteins(String proteinFileName) {

        // Opening the output file
        OutputFile<String> file = TextOutputFile.getBasic(proteinFileName);
        
        // Getting list of all reaction nodes
        NodeList listOfReactions = sbmlFile.getElementsByTagName("reaction");

        // Getting the total number of reactions
        int length = listOfReactions.getLength();

        // Iterating thru each reaction contained in the SBML file
        for (int i = 0; i < length; i++) {

            // Getting the next reaction node
            Node reaction = listOfReactions.item(i);

            // Getting reaction name (or id)
            String rxnName = nodeParser.getRxnId(reaction);

            // Determing boolean logic rules for proteins
            Or<And<String>> proteinRules = nodeParser.getProteins(reaction);
            
            writeProteins(file, rxnName, proteinRules);
        }

    }

    /**
     * Writes the flux vector contained in the SBML file to a file with the
     * provided name.
     * 
     * @param fluxVectorFileName the name of the file to generated.
     */
    public void writeFluxVector(String fluxVectorFileName) {

        // Opening the output file
        OutputFile<String> file = TextOutputFile.getBasic(fluxVectorFileName);
        
        // Getting list of all reaction nodes
        NodeList listOfReactions = sbmlFile.getElementsByTagName("reaction");

        // Getting the total number of reactions
        int length = listOfReactions.getLength();

        // Iterating thru each reaction contained in the SBML file
        for (int i = 0; i < length; i++) {

            // Getting the next reaction node
            Node reaction = listOfReactions.item(i);

            // Getting reaction name (or id)
            String rxnName = nodeParser.getRxnId(reaction);

            double flux = nodeParser.getFluxValue(reaction);
            
            writeFlux(file, rxnName, flux);
        }

    }

    /**
     * Writes the flux bounds contained in the SBML file to a file with the
     * provided name.
     * 
     * @param fluxBoundsFileName the name of the file to generated.
     */
    public void writeFluxBounds(String fluxBoundsFileName) {

        // Opening the output file
        OutputFile<String> file = TextOutputFile.getBasic(fluxBoundsFileName);
        
        // Getting list of all reaction nodes
        NodeList listOfReactions = sbmlFile.getElementsByTagName("reaction");

        // Getting the total number of reactions
        int length = listOfReactions.getLength();

        // Iterating thru each reaction contained in the SBML file
        for (int i = 0; i < length; i++) {

            // Getting the next reaction node
            Node reaction = listOfReactions.item(i);

            // Getting reaction name (or id)
            String rxnName = nodeParser.getRxnId(reaction);

            Interval bounds = nodeParser.getFluxBounds(reaction);
            
            writeFluxBounds(file, rxnName, bounds);
        }

    }
    

    private void writeRevRxn(OutputFile<String> file,
                             String rxnName,
                             LinearComb<String> reactants,
                             LinearComb<String> products) {

        String rxn = linFormat.format(reactants) +
                     " " + reversibleArrow + " " +
                     linFormat.format(products);
        
        String line = entryFormat.formatPair(rxnName, rxn);
        
        file.addData(line);                         
    }
    
    private void writeIrrevRxn(OutputFile<String> file,
                               String rxnName,
                               LinearComb<String> reactants,
                               LinearComb<String> products) {
        

        String rxn = linFormat.format(reactants) +
                     " " + forwardArrow + " " +
                     linFormat.format(products);
        
        String line = entryFormat.formatPair(rxnName, rxn);
        
        file.addData(line);
    }
    
    private void writeGenes(OutputFile<String> file,
                            String rxnName, 
                            Or<And<String>> geneRules) {

        String rules = boolFormat.format(geneRules);
        
        if (rules.length() == 0) {
            rules = NOT_PROVIDED;
        }
        
        String line = entryFormat.formatPair(rxnName, rules);
        
        file.addData(line);
    }

    
    private void writeProteins(OutputFile<String> file,
                               String rxnName, 
                               Or<And<String>> protRules) {

        String rules = boolFormat.format(protRules);
        
        if (rules.length() == 0) {
            rules = NOT_PROVIDED;
        }
        
        String line = entryFormat.formatPair(rxnName, rules);
        
        file.addData(line);
    }
    
    private void writeFlux(OutputFile<String> file,
                           String rxnName, 
                           double flux) {

        String fluxString = numberFormat.format(flux);
        
        String line = entryFormat.formatPair(rxnName, fluxString);
        
        file.addData(line);
    }
    
    private void writeFluxBounds(OutputFile<String> file,
                                 String rxnName, 
                                 Interval fluxBounds) {

        String bounds = intervalFormat.format(fluxBounds);
        
        String line = entryFormat.formatPair(rxnName, bounds);
        
        file.addData(line);
    }

}
