/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.proc.fba;

import java.io.PrintWriter;
import java.util.Map.Entry;

import sbrt.kernel.fba.ChemRxn;
import sbrt.shell.io.InputFile;
import sbrt.shell.io.OutputFile;
import sbrt.shell.io.TextInputFile;
import sbrt.shell.io.TextOutputFile;
import sbrt.shell.mng.AbstractProcessManager;
import sbrt.shell.text.EqualsMapEntryFormatV1;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.fba.IrrevRxnFormatV2;
import sbrt.shell.text.fba.RevRxnFormatV2;
import sbrt.shell.text.fba.RxnFormatV1;

/**
 * This class is used to translate simple reaction files from, say,
 * spreadsheets into FBA Reaction Files.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public class SimpleRxnFileReader {

    private final RevRxnFormatV2 inputRevFormat;
    private final IrrevRxnFormatV2 inputIrrevFormat;
    
    private final RxnFormatV1 outputFormat;
    
    private final EqualsMapEntryFormatV1<String, String> entryFormat;
    
    /**
     * Constructs a new simple reaction file reader.
     */
    public SimpleRxnFileReader() {
        
    	inputRevFormat = new RevRxnFormatV2();
        inputIrrevFormat = new IrrevRxnFormatV2();
        
        outputFormat = new RxnFormatV1();
        
        entryFormat = EqualsMapEntryFormatV1.getBasic();
    }
    
    /**
     * Returns the parser used for reversible reactions in the provided
     * input file.
     * 
     * @return the parser used for reversible reactions in the provided
     * input file.
     */
    public RevRxnFormatV2 getInputRevRxnFormat() {
        return inputRevFormat;
    }
    
    /**
     * Returns the parser used for irreversible reactions in the provided
     * input file.
     * 
     * @return the parser used for irreversible reactions in the provided
     * input file.
     */
    public IrrevRxnFormatV2 getInputIrrevRxnFormat() {
        return inputIrrevFormat;
    }
    
    /**
     * Returns the formatter used for writing reactions to the
     * output file.
     * 
     * @return the formatter used for writing reactions to the
     * output file.
     */
    public RxnFormatV1 getOutputRxnFormat() {
        return outputFormat;
    }
    
    /**
     * Returns the format used for individual lines in the input and 
     * output files.
     * 
     * @return the format used for individual lines in the input and 
     * output files.
     */
    public EqualsMapEntryFormatV1<String, String> getLineFormat() {
        return entryFormat;
    }
    
    /**
     * Translates a simple reaction file into an FBA Reaction File.
     * 
     * @param simpleRxnFileName the name of the simple reaction file.
     * @param outputFileName the desired name of the FBA Reaction File.
     */
    public void translate(String simpleRxnFileName, String outputFileName) {
        translate(simpleRxnFileName, 
        		  outputFileName, 
        		  AbstractProcessManager.DEV_NULL);
    }
    
    /**
     * Translates a simple reaction file into an FBA Reaction File.
     * 
     * @param simpleRxnFileName the name of the simple reaction file.
     * @param outputFileName the desired name of the FBA Reaction File.
     * @param out the print writer to which descriptive messages will be written.
     */
    public void translate(String simpleRxnFileName, 
    					  String outputFileName, 
    					  PrintWriter out) {

        // Opening files
        InputFile<String> input = getInputFile(simpleRxnFileName, out);
        OutputFile<String> output = getOutputFile(outputFileName, out);
        
        for (String line : input) {
            translateLine(line, output);
        }
        
        // Closing files
        AbstractProcessManager.closeFile(input, out);
        AbstractProcessManager.closeFile(output, out);
    }

    private void translateLine(String line, OutputFile<String> file) {
        
        Entry<String, String> e = entryFormat.parsePair(line);
        
        String name = e.getKey();
        String rxnString = e.getValue();
        
        if (rxnString.contains(IrrevRxnFormatV2.ARROW)) {
            transIrrevRxn(name, rxnString, file);
        }
        
        else if (rxnString.contains(RevRxnFormatV2.ARROW)) {
            transRevRxn(name, rxnString, file);
        }
        
        else {
            String s = "The reaction '" + rxnString + 
            "' does not contain a recognizable arrow.";
            throw new FormatException(s);
        }
    }
    
    private void transRevRxn(String name, 
    						 String rxnString, 
    						 OutputFile<String> file) {
        
        // Parsing
        ChemRxn rxn = inputRevFormat.parse(rxnString); 
        
        // Reformatting
        rxnString = outputFormat.format(rxn);
        
        // Writing to output file
        file.addData(entryFormat.formatPair(name, rxnString));
    }

    private void transIrrevRxn(String name, 
    						   String rxnString, 
    						   OutputFile<String> file) {
        
        // Parsing
        ChemRxn rxn = inputIrrevFormat.parse(rxnString); 
        
        // Reformatting
        rxnString = outputFormat.format(rxn);
        
        // Writing to output file
        file.addData(entryFormat.formatPair(name, rxnString));
    }
    
    private InputFile<String> getInputFile(String fileName, PrintWriter out) {
        
        out.print("Opening '" + fileName + "' ... ");
        out.flush();
        
        InputFile<String> file = TextInputFile.getBasic(fileName);
        
        out.println("DONE");
        
        return file;
    }

    private OutputFile<String> getOutputFile(String fileName, PrintWriter out) {
        
        out.print("Opening '" + fileName + "' ... ");
        out.flush();
        
        OutputFile<String> file = TextOutputFile.getBasic(fileName);
        
        out.println("DONE");
        
        return file;
    }
    
}
