/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import java.util.*;
import java.util.regex.*;

import sbrt.kernel.utilities.*;

/**
 * This abstract class is a skeleton implementation of 
 * {@linkplain LinkedHashSetFormat}. A concrete class with a
 * defined format can be
 * created by simply overriding the constructor of this class.
 * 
 * @author This class was written and documented by 
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <E> the type of element.
 */
public abstract class AbstractLinkedHashSetFormat<E> 
	implements LinkedHashSetFormat<E> {

    private final String delimiter;
    private final Pattern regex;
    private final Formatter<E> elementFormatter;
    private final Parser<E> elementParser;
    
    /**
     * Constructs a new {@code LinkedHashSet} format from
     * the provided arguments.
     * 
     * @param delimiter the delimiter used to separate elements
     * of sets when producing a formatted string
     * representation.
     * 
     * @param regex the regular expression pattern used to
     * parse a string representation of a set with this format.
     * The method call {@code regex.split(String)} should return
     * an array containing the string representation of the elements
     * of the provided set string.
     * 
     * @param elementFormat the format used for elements the 
     * of sets with this format.
     * 
     * @throws NullPointerException if any provided argument
     * is {@code null}.
     */
    public AbstractLinkedHashSetFormat(String delimiter, 
    								   Pattern regex,
    								   SimpleFormat<E> elementFormat) {

    	this(delimiter, regex, elementFormat, elementFormat);
    }
    
    /**
     * Constructs a new {@code LinkedHashSet} format from
     * the provided arguments.
     * 
     * @param delimiter the delimiter used to separate elements
     * of sets when producing a formatted string
     * representation.
     * 
     * @param regex the regular expression pattern used to
     * parse a string representation of a set with this format.
     * The method call {@code regex.split(String)} should return
     * an array containing the string representation of the elements
     * of the provided set string.
     * 
     * @param elementFormatter the formatter used for elements the 
     * of sets with this format.
     * @param elementParser the parser used for elements the 
     * of sets with this format.
     * 
     * @throws NullPointerException if any provided argument
     * is {@code null}.
     */
    public AbstractLinkedHashSetFormat(String delimiter, 
    								   Pattern regex,
    								   Formatter<E> elementFormatter,
    								   Parser<E> elementParser) {
        
    	// Checking inputs
    	Util.nullCheck(delimiter);
    	Util.nullCheck(regex);
    	Util.nullCheck(elementFormatter);
    	Util.nullCheck(elementParser);
        
        // Saving
        this.delimiter = delimiter;
        this.regex = regex;
        this.elementFormatter = elementFormatter;
        this.elementParser = elementParser;
    }

    /**
	 * Returns the delimiter used to separate elements
	 * when producing a formatted string.
	 * 
	 * @return the delimiter used to separate elements
	 * when producing a formatted string.
	 */
	public String getDelimiter() {
		return delimiter;
	}
	
	/**
	 * Returns the regular expression used to separate
	 * elements when parsing a formatted string.
	 * 
	 * @return the regular expression used to separate
	 * elements when parsing a formatted string.
	 */
	public Pattern getRegex() {
		return regex;
	}
	
    /**
     * Returns the formatter for the elements of this 
     * {@code LinkedHashSet} format.
     * 
     * @return the formatter for the elements of this 
     * {@code LinkedHashSet} format.
     */
	public Formatter<E> getElementFormatter() {
		return elementFormatter;
	}

    /**
     * Returns the parser for 
     * the elements of this set format.
     * 
     * @return the parser for 
     * the elements of this set format.
     */
    public Parser<E> getElementParser() {
    	return elementParser;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided collection.

	 * @param c the collection to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided collection.
	 */
	public String format(Set<? extends E> c) {

    	// Creating string builder
        StringBuilder b = new StringBuilder(c.size() * 100);

        // Iterating thru each element
        for (Iterator<? extends E> i = c.iterator(); i.hasNext(); ) {

        	// Getting the next element
            E obj = i.next();
            
            // Formatting and adding to the builder
            b.append(elementFormatter.format(obj));

            // If another element is present, append the provided delimiter
            if (i.hasNext()) {
                b.append(delimiter);
            }
        }

        // Returning formatted string
        return b.toString();
    }

    /**
     * Parses the provided string and returns its 
     * corresponding collection of elements as a 
     * {@code LinkedHashSet}.
     * 
     * @param set the string to be parsed.
     * 
     * @return the provided string and returns its 
     * corresponding collection of elements as a 
     * {@code LinkedHashSet}.
     */
    public LinkedHashSet<E> parse(String set) {

    	// Splitting on the provided regular expression
        String[] elements = regex.split(set, -1);

        // Creating storage
        LinkedHashSet<E> parsedElements =
            new LinkedHashSet<E>(elements.length + 1, 1.0F);

        // Iterating thru each string
        for (int i = 0; i < elements.length; i ++) {

        	// Parsing element 
        	E element = elementParser.parse(elements[i]);

        	// Trying to add the element to the set
            if (!parsedElements.add(element)) {
            	
            	// If this element already exists in the set
                String s = "'" + elements[i] +
                           "' appears more than once in '" + set + "'.";
                throw new FormatException(s);
            }
        }

        return parsedElements;
    }


}
