/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import java.util.*;
import java.util.regex.*;

import sbrt.kernel.utilities.*;

/**
 * This abstract class is a skeleton implementation of
 * {@linkplain MapEntryFormat}.  A concrete class with a definite
 * format can be created by simply overriding the
 * constructor of this class.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public abstract class AbstractMapEntryFormat<K, V> implements MapEntryFormat<K, V> {

    private final Pattern delimiter;
    private final String leftPadding;
    private final String rightPadding;
    
    /**
     * Constructs a new map entry format from the provided arguments.
     * 
     * @param leftPadding a string of whitespace characters
     * placed between the key and delimiter when producing
     * a formatted string representation of a map entry.
     * 
     * @param delimiter the delimiter used to separate
     * key-value pairs when producing and parsing a formatted
     * string representation of a map entry.
     * 
     * @param rightPadding a string of whitespace characters
     * placed between the delimiter and value when producing
     * a formatted string representation of a map entry.
     * 
     * @throws IllegalArgumentException if the left or right
     * paddings contain non-whitespace characters.
     * 
     * @throws NullPointerException if any provided argument
     * is {@code null}.
     * 
     * @throws PatternSyntaxException if the provided delimiter
     * cannot be compiled into a regular expression.
     */
    public AbstractMapEntryFormat(String leftPadding,
    		String delimiter,
    		String rightPadding) {
    	
    	// Checking inputs
        Util.nullCheck(delimiter);

        // Ensuring left padding contains only whitespace
        // characters
        if (leftPadding.trim().length() != 0) {
        	String s = "Invalid left padding: " + leftPadding;
        	throw new IllegalArgumentException(s);
        }

        // Ensuring right padding contains only whitespace
        // characters
        if (rightPadding.trim().length() != 0) {
        	String s = "Invalid right padding: " + rightPadding;
        	throw new IllegalArgumentException(s);
        }
        
        // Setting
        this.delimiter = Pattern.compile(delimiter);
        this.leftPadding = leftPadding;
        this.rightPadding = rightPadding;
    }
    
    /**
     * Constructs a new map entry format using the provided delimiter
     * and zero-length strings as left and right paddings.
     * 
     * @param delimiter the delimiter used to separate
     * key-value pairs when producing and parsing a formatted
     * string representation of a map entry.
     * 
     * @throws NullPointerException if the provided argument
     * is {@code null}.
     * 
     * @throws PatternSyntaxException if the provided delimiter
     * cannot be compiled into a regular expression.
     * 
     * @see #AbstractMapEntryFormat(String, String, String)
     */
    public AbstractMapEntryFormat(String delimiter) {
        this("", delimiter, "");
    }

    /**
     * Returns the delimiter used to separate the key from the
     * value in a key-value pair.
     * 
     * @return the delimiter used to separate the key from the
     * value in a key-value pair.
     */
    public final String getDelimiter() {
        return delimiter.pattern();
    }

    /**
     * Returns the left padding used to create formatted
     * string representations of key-value pairs. The left
     * padding is inserted between the key and delimiter.
     * In other words, it appears to the left of the delimiter
     * in key-value strings.
     * 
     * @return the left padding used to create formatted
     * string representations of key-value pairs. 
     */
    public final String getLeftPadding() {
        return leftPadding;
    }

    /**
     * Returns the right padding used to create formatted
     * string representations of key-value pairs. The right
     * padding is inserted between the delimiter and value.
     * In other words, it appears to the right of the delimiter
     * in key-value strings.
     * 
     * @return the right padding used to create formatted
     * string representations of key-value pairs. 
     */
    public final String getRightPadding() {
        return rightPadding;
    }

    /**
     * Returns a string representation of the provided
     * key-value pair.
     * 
     * @param <S> a type extending the primary key type.
     * @param <T> a type extending the primary value type.
     * 
     * @param key the key.
     * @param value the value.
     * 
     * @return a string representation of the provided
     * key-value pair.
     */
    public <S extends K, T extends V> String formatPair(S key, T value) {
    	
    	// Formatting both
    	String k = formatKey(key);
    	String v = formatValue(value);
    	
    	// Returning formatted key-value pair
        return k + leftPadding + getDelimiter() + rightPadding + v;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided map entry.
	 * 
	 * @param entry the map entry.
	 * 
	 * @return a formatted string representation of the
	 * provided map entry.
	 */
    public String formatPair(Map.Entry<? extends K, ? extends V> entry) {

    	// Returning formatted key-value pair
        return formatPair(entry.getKey(), entry.getValue());
    }

    /**
     * Parses the provided map entry string, and
     * returns its corresponding map entry. The string is split on the 
     * <em>first</em> occurrence of the delimiter.
     * 
     * @param mapEntry the string representation of a
     * single map entry.
     * 
     * @return a singleton map corresponding to the 
     * provided string. 
     */
    public Map.Entry<K, V> parsePair(String mapEntry) {
    	
    	// Trimming whitespace
        mapEntry = mapEntry.trim();

        // If the string starts with the delimiter
        if (mapEntry.startsWith(getDelimiter())) {
            String s = "'" + mapEntry +
                       "' is missing a left-hand side.";
            throw new FormatException(s);
        }

        // If the string ends with the delimiter
        if (mapEntry.endsWith(getDelimiter())) {
            String s = "'" + mapEntry +
                       "' is missing a right-hand side.";
            throw new FormatException(s);
        }

        // Splitting on the first delimiter
        String[] parts = delimiter.split(mapEntry, 2);

        // If only one part exists
        if (parts.length == 1) {
            String s = "The delimiter '" + getDelimiter() +
                       "' is not present in '" + mapEntry + "'.";
            throw new FormatException(s);
        }

        // If more than one part exists
        /*
        if (parts.length > 2) {
            String s = "The delimiter '" + getDelimiter() +
                       "' appears more than once in '" + mapEntry + "'.";
            throw new FormatException(s);
        }
        */
        
        // Trimming whitespace from both parts
        String left = parts[0].trim();
        String right = parts[1].trim();

        // If the left hand side is empty
        if (left.length() == 0) {
            String s = "'" + mapEntry + "' is missing a left-hand side.";
            throw new FormatException(s);
        }

        // If the right hand side is empty
        if (right.length() == 0) {
            String s = "'" + mapEntry + "' is missing a right-hand side.";
            throw new FormatException(s);
        }

        // Parsing the key and value
        K key = parseKey(left);
        V value = parseValue(right);
        
        // Creating a singleton map
        Map<K, V> m = Collections.singletonMap(key, value); 
        
        // Returning a map entry
        return m.entrySet().iterator().next();
    }

}
