/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import java.util.*;

/**
 * This abstract class is a skeleton implementation of
 * {@linkplain MapFormat}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public abstract class AbstractMapFormat<K, V> implements MapFormat<K, V> {

    /**
     * Returns the formatter used for keys.
     * @return the formatter used for keys.
     */
    public Formatter<K> getKeyFormatter() {
    	return getSetFormat().getElementFormatter();
    }
    
    /**
     * Returns the parser used for keys.
     * @return the parser used for keys.
     */
    public Parser<K> getKeyParser() {
    	return getSetFormat().getElementParser();
    }

    /**
     * Returns the formatter used for values.
     * @return the formatter used for values.
     */
    public Formatter<V> getValueFormatter() {
    	return getListFormat().getElementFormatter();
    }

    /**
     * Returns the parser used for values.
     * @return the parser used for values.
     */
    public Parser<V> getValueParser() {
    	return getListFormat().getElementParser();
    }
    
    /**
     * Returns the format used for variable-value pairs. This
     * format must be a basic format, this is, one that does not
	 * format or parse the keys or values supplied to it.
	 * 
     * @return the format used for variable-value pairs.
     */
    public abstract MapEntryFormat<String, String> getMapEntryFormat();
    
    /**
     * Returns the format used for sets of keys.
     * @return the format used for sets of keys.
     */
    public abstract LinkedHashSetFormat<K> getSetFormat();

    /**
     * Returns the format used for lists of values.
     * @return the format used for lists of values.
     */
    public abstract ListFormat<V> getListFormat();

	/**
	 * Returns a formatted string representation of the
	 * provided key.
	 * 
	 * @param <S> a type that extends the primary key
	 * type of this format.
	 * 
	 * @param key the key to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided key.
	 */
	public <S extends K> String formatKey(S key) {
		return getKeyFormatter().format(key);
	}

	/**
	 * Returns a formatted string representation of the
	 * provided value.
	 * 
	 * @param <T> a type that extends the primary value
	 * type of this format.
	 * 
	 * @param value the value to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided value.
	 */
	public <T extends V> String formatValue(T value) {
		return getValueFormatter().format(value);
	}

	/**
	 * Returns a formatted string representation of the
	 * provided set of keys.
	 * 
	 * @param keys the set of keys to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided set of keys.
	 */
	public String formatKeys(Set<? extends K> keys) {
		return getSetFormat().format(keys);
	}

	/**
	 * Returns a string representation of the
	 * provided list of values.
	 * 
	 * @param values the list of values to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided list of values.
	 */
	public String formatValues(List<? extends V> values) {
		return getListFormat().format(values);
	}

	/**
	 * Returns a string representation of the
	 * values in the provided map formatted as a list. The order of
	 * occurrence of values in the returned string depends
	 * on the <i>order</i> of the map. See 
	 * {@code java.util.Map} for further details.
	 * 
	 * @param map keys mapped to their respective values.
	 * 
	 * @return a string representation of the
	 * values in the provided map formatted as a list.
	 */
    public String formatValues(Map<? extends K, ? extends V> map) {

    	// Creating storage for values
        List<V> values = new LinkedList<V>();

        // Iterating thru each variable-value pair
        for (Map.Entry<? extends K, ? extends V> e : map.entrySet()) {

        	// Adding value to the rest
        	values.add(e.getValue());
        }

        return getListFormat().format(values);
    }

    /**
     * Returns a string representation of the provided
     * key-value pair.
     * 
     * @param <S> a type extending the primary key type.
     * @param <T> a type extending the primary value type.
     * 
     * @param key the key.
     * @param value the value.
     * 
     * @return a string representation of the provided
     * key-value pair.
     */
    public <S extends K, T extends V> String formatPair(S key, T value) {
    	
    	// Formatting both
    	String k = formatKey(key);
    	String v = formatValue(value);
    	
    	// Returning formatted key-value pair
        return getMapEntryFormat().formatPair(k, v);
    }

	/**
	 * Returns a formatted string representation of the
	 * provided map entry.
	 * 
	 * @param entry the map entry.
	 * 
	 * @return a formatted string representation of the
	 * provided map entry.
	 */
    public String formatPair(Map.Entry<? extends K, ? extends V> entry) {

    	// Returning formatted key-value pair
        return formatPair(entry.getKey(), entry.getValue());
    }

    /**
     * Parses the provided string and returns its 
     * corresponding key.
     * 
     * @param key a string representation of a key.
     * 
     * @return the key corresponding to the 
     * provided string. 
     */
	public K parseKey(String key) {
		return getKeyParser().parse(key);
	}

    /**
     * Parses the provided string and returns its 
     * corresponding value.
     * 
     * @param value a string representation of a value.
     * 
     * @return the value corresponding to the 
     * provided string. 
     */
	public V parseValue(String value) {
		return getValueParser().parse(value);
	}

    /**
     * Parses the provided map entry string, and
     * returns its corresponding map entry.
     * 
     * @param mapEntry the string representation of a
     * single map entry.
     * 
     * @return a singleton map corresponding to the 
     * provided string. 
     */
    public Map.Entry<K, V> parsePair(String mapEntry) {
    	
    	Map.Entry<String, String> e = getMapEntryFormat().parsePair(mapEntry); 
    	
    	K key = getKeyParser().parse(e.getKey());
    	V value = getValueParser().parse(e.getValue());
    	
    	Map<K, V> m = Collections.singletonMap(key, value);
    	
    	return m.entrySet().iterator().next();
    }

    /**
     * Parses the provided string and returns its 
     * corresponding set of keys.
     * 
     * @param allKeys a string representation of a set
     * of keys.
     * 
     * @return a set of keys corresponding to the 
     * provided string. 
     */
	public LinkedHashSet<K> parseAllKeys(String allKeys) {
		return getSetFormat().parse(allKeys);
	}

    /**
     * Parses the provided string and returns its 
     * corresponding list of values.
     * 
     * @param allValues a string representation of a list
     * of values.
     * 
     * @return a list of values corresponding to the 
     * provided string. 
     */
    public List<V> parseAllValues(String allValues) {
        return getListFormat().parse(allValues);
    }

    /**
     * Parses the provided string of values and returns a 
     * corresponding map of variable-value pairs using the
     * provided set of keys. The association between
     * keys and values is determined from their order of
     * occurrence in their respective collection. In other
     * words, the first variable returned by the set's iterator
     * is associated with the first parsed value, and so on.
     *
     * @param allkeys the set of keys with which 
     * to associate the parsed values.
     * 
     * @param allValues a string representation of a list
     * of values.
     * 
     * @return keys mapped to their respective values.
     */
    public Map<K, V> 
    parseAllValues(LinkedHashSet<? extends K> allkeys,
    			   String allValues) {

    	// Creating storage for values
        List<V> values = parseAllValues(allValues);

        // Getting sizes
        int varSize = allkeys.size();
        int valSize = values.size();

        // Comparing
        if (varSize != valSize) {

            String s = "There are " + varSize + " keys in '" +
                       formatKeys(allkeys) + "' but there are " +
                       valSize + " values in '" + formatValues(values) + "'.";
            throw new FormatException(s);

        }

        // Creating storage for map
        Map<K, V> map = new LinkedHashMap<K, V>(varSize + 1);

        // Getting iterators over keys and values
        Iterator<? extends K> i = allkeys.iterator();
        Iterator<V> j = values.iterator();

        // While another variable exists
        while (i.hasNext()) {
            map.put(i.next(), j.next());
        }

        return map;
    }

    /**
     * Parses the provided set of keys and list of values
     * and returns a corresponding map of variable-value pairs.
     * The association between keys and values is determined
     * from their order of occurrence in their respective collection.
     * In other words, the first parsed variable is
     * associated with the first parsed value, and so on. 
     * 
     * @param allkeys a string representation of a set
     * of keys.
     * 
     * @param allValues a string representation of a list of
     * values.
     * 
     * @return keys mapped to their respective values.
     */
    public Map<K, V> parseAllPairs(String allkeys, String allValues) {

    	// Parsing all keys
    	LinkedHashSet<K> keys = parseAllKeys(allkeys);

    	// Parsing all values
    	return parseAllValues(keys, allValues);
    }

}
