/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import java.util.*;

import sbrt.kernel.utilities.And;
import sbrt.kernel.utilities.Or;

/**
 * This interface is used to represent boolean relationship
 * formats.
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <E> the type of element.
 */
public class BoolRelFormatV1<E> implements SimpleFormat<Or<And<E>>> {

    /**
     * Returns an {@code BoolRelFormatV1} that does not format or parse the
     * elements supplied to it. In other words, it returns an
     * {@code BoolRelFormatV1} with the most basic possible functionality.
     * 
     * @return an {@code BoolRelFormatV1} that does not format or parse the
     * elements supplied to it.
     */
    public static BoolRelFormatV1<String> getBasic() {
        return new BoolRelFormatV1<String>(new NonEmptyStringFormat());
    }
    
    private PipeSetFormat<String> orFormat;
    private AmpersandSetFormat<E> andFormat;
    private SimpleFormat<E> elementFormat;
    
    /**
     * Constructs a new boolean relation format object that uses the
     * provided format for its elements.
     * @param elementFormat the format to be used for elements.
     */
    public BoolRelFormatV1(SimpleFormat<E> elementFormat) {
        
        this.elementFormat = elementFormat;
        
        orFormat = PipeSetFormat.getBasic();
        andFormat = new AmpersandSetFormat<E>(elementFormat);
    }
    
    
	/**
	 * Returns a formatted string representation of the
	 * provided boolean relationship.

	 * @param boolRel the boolean relationship to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided set.
	 */
    public String format(Or<And<E>> boolRel) {
         
        Set<String> andGroupStrings = new LinkedHashSet<String>();
        
        for (And<E> and : boolRel) {
            
            String s = andFormat.format(and.getConditions());
            
            andGroupStrings.add(s);
        }
        
        return orFormat.format(andGroupStrings);
    }

    /**
     * Parses the provided string and returns its 
     * corresponding collection of elements as a 
     * {@code LinkedHashSet}.
     * 
     * @param set the string to be parsed.
     * 
     * @return the set of elements 
     * corresponding to the provided string. 
     */	
    public Or<And<E>> parse(String set) {
        
        Set<String> andGroupStrings = orFormat.parse(set);

        int size = andGroupStrings.size() + 1;
        
        Set<And<E>> andGroups = new LinkedHashSet<And<E>>(size, 1f);
        
        for (String s : andGroupStrings) {
            
            Set<E> group = andFormat.parse(s);
            
            And<E> and = new And<E>(group);
            
            andGroups.add(and);
        }
        
        return new Or<And<E>>(andGroups);
    }

    /**
     * Returns the formatter for 
     * the elements of this set format.
     * 
     * @return the formatter for 
     * the elements of this set format.
     */
    public Formatter<E> getElementFormatter() {
        return elementFormat;
    }

    /**
     * Returns the parser for 
     * the elements of this set format.
     * 
     * @return the parser for 
     * the elements of this set format.
     */
    public Parser<E> getElementParser() {
        return elementFormat;
    }


}
