/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import sbrt.kernel.utilities.Util;

/**
 * This class is a concrete implementation of {@linkplain MapEntryFormat}.
 * The string representation of map entries is:<P>
 * <i>key: value</i>. 
 * <P>Instances of this class are immutable.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public final class ColonMapEntryFormatV1<K, V> 
	extends AbstractMapEntryFormat<K, V> {

	/**
	 * Returns a {@code ColonMapEntryFormatV1} that does not
	 * format or parse the keys or values supplied to it. In other
	 * words, it returns a {@code ColonMapEntryFormatV1} with
	 * the most basic possible functionality. Empty strings are not 
	 * allowed however.
	 *  
	 * @return a {@code ColonMapEntryFormatV1} that does not
	 * format or parse the keys and values supplied to it.
	 */
	public static final ColonMapEntryFormatV1<String, String> getBasic() {

		// Creating the "trivial" format
		NonEmptyStringFormat f = new NonEmptyStringFormat();
		
		return new ColonMapEntryFormatV1<String, String>(f, f);
	}
	
	private final SimpleFormat<K> keyFormat;
	private final SimpleFormat<V> valueFormat;
	
	/**
	 * Constructs a new map entry format using the provided key and
	 * value formats.
	 * 
	 * @param keyFormat the format to use for keys.
	 * @param valueFormat the format to use for values.
	 * 
	 * @throws NullPointerException if either argument is {@code null}.
	 */
    public ColonMapEntryFormatV1(SimpleFormat<K> keyFormat,
    							 SimpleFormat<V> valueFormat) {
        super("", ":", " ");
        
        // Checking input
        Util.nullCheck(keyFormat);
        Util.nullCheck(valueFormat);
        
        // Saving
        this.keyFormat = keyFormat;
        this.valueFormat = valueFormat;
    }

    /**
     * Returns the format used for keys.
     * @return the format used for keys.
     */
    public SimpleFormat<K> getKeyFormat() {
    	return keyFormat;
    }

    /**
     * Returns the format used for values.
     * @return the format used for values.
     */
    public SimpleFormat<V> getValueFormat() {
    	return valueFormat;
    }

	/**
	 * Returns a formatted string representation of the 
	 * provided key object.
	 * 
	 * @param <S> a type extending the primary key type.
	 * 
	 * @param key the key object to format.
	 * 
	 * @return a formatted string representation of the 
	 * provided key object.
	 */
	public <S extends K> String formatKey(S key) {
		return keyFormat.format(key);
	}

	/**
	 * Returns a formatted string representation of the 
	 * provided value object.
	 * 
	 * @param <T> a type extending the primary value type.
	 * 
	 * @param value the value object to format.
	 * 
	 * @return a formatted string representation of the 
	 * provided value object.
	 */
	public <T extends V> String formatValue(T value) {
		return valueFormat.format(value);
	}
	
    /**
     * Parses the provided string representation of a key
     * object.
     * 
     * @param key the string representation of a
     * key object.
     * 
     * @return the key object corresponding to the 
     * provided string. 
     */
    public K parseKey(String key) {
    	return keyFormat.parse(key);
    }
    
    /**
     * Parses the provided string representation of a value
     * object.
     * 
     * @param value the string representation of a
     * value object.
     * 
     * @return the value object corresponding to the 
     * provided string. 
     */
    public V parseValue(String value) {
    	return valueFormat.parse(value);
    }
    
    
}
