/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

/**
 * This class is a concrete implementation of {@linkplain DoubleFormat}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 */
public class DoubleFormatV1 implements DoubleFormat {

    private static final String POS_INF = "Infinity";
    private static final String NEG_INF = "-Infinity";
    private static final String NAN = "NaN";

	/**
	 * Returns a formatted string representation of the
	 * provided double precision number. See
	 * {@linkplain java.lang.Double#toString()} for the 
	 * exact formatting rules.
	 * 
	 * @param d the double precision number to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided double precision number.
	 */
    public String format(Double d) {
        return format(d.doubleValue());
    }

    /**
     * Parses the provided string and returns its 
     * corresponding double precision number. See
     * {@linkplain java.lang.Double#valueOf(String)}
     * for the exact formatting rules.
     * 
     * @param d the string to be parsed.
     * 
     * @return the double precision number corresponding 
     * to the provided string. 
     * 
     * @throws FormatException if the provided string
     * is not parsable as a double precision number.
     * 
     * @throws NullPointerException if the provided 
     * argument is {@code null}.
     */
    public String format(double d) {

    	if (Double.isNaN(d)) {
    		return NAN;
    	}
    	
    	else if (d == Double.POSITIVE_INFINITY) {
            return POS_INF;
        } 
    	
    	else if (d == Double.NEGATIVE_INFINITY) {
            return NEG_INF;
        }

        return Double.toString(d);
    }

    /**
     * Parses the provided string and returns its 
     * corresponding double precision number. See
     * {@linkplain java.lang.Double#valueOf(String)}
     * for the exact formatting rules.
     * 
     * @param d the string to be parsed.
     * 
     * @return the double precision number corresponding 
     * to the provided string. 
     * 
     * @throws FormatException if the provided string
     * is not parsable as a double precision number.
     * 
     * @throws NullPointerException if the provided 
     * argument is {@code null}.
     */
    public Double parse(String d) {
    	
    	// Parsing
    	double x = parsePrimitive(d);
    	
    	// Returning as Double
        return Double.valueOf(x);
    }

    /**
     * Parses the provided string and returns its 
     * corresponding double precision number. See
     * {@linkplain java.lang.Double#valueOf(String)}
     * for the exact formatting rules.
     * 
     * @param d the string to be parsed.
     * 
     * @return the double precision number corresponding 
     * to the provided string. 
     * 
     * @throws FormatException if the provided string
     * is not parsable as a double precision number.
     * 
     * @throws NullPointerException if the provided 
     * argument is {@code null}.
     */
    public double parsePrimitive(String d) {

    	// Removing leading and trailing whitespace
        String trimmed = d.trim();

        if (trimmed.endsWith(NAN)) {
        	return Double.NaN;
        }
        
        if (trimmed.equals(POS_INF)) {
            return Double.POSITIVE_INFINITY;
        } 
        
        else if (trimmed.equals(NEG_INF)) {
            return Double.NEGATIVE_INFINITY;
        }

        try {
        	
            return Double.valueOf(trimmed);
            
        } catch (Throwable t) {
            String s = "'" + d + "' is not parsable as a number.";
            throw new FormatException(s, t);
        }
    }

}

