/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import sbrt.kernel.utilities.Util;

/**
 * This class is a concrete implementation of {@linkplain MapFormat}.
 * The string representation of map entries is:<P>
 * <i>key = value</i>. 
 * <P>Instances of this class are immutable.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public final class EqualsMapFormatV1<K, V> extends AbstractMapFormat<K, V> {

	/**
	 * Returns an {@code EqualsMapFormatV1} that does not
	 * format or parse the keys or values supplied to it. In other
	 * words, it returns an {@code EqualsMapFormatV1} with
	 * the most basic possible functionality. Empty strings are not 
	 * allowed however.
	 *  
	 * @return an {@code EqualsMapFormatV1} that does not
	 * format or parse the keys and values supplied to it.
	 */
	public static final EqualsMapFormatV1<String, String> getBasic() {

		// Creating the "trivial" format
		NonEmptyStringFormat f = new NonEmptyStringFormat();
		
		return new EqualsMapFormatV1<String, String>(f, f);
	}
	
	private final EqualsMapEntryFormatV1<String, String> entryFormat;
	private final PipeSetFormat<K> setFormat;
	private final PipeListFormat<V> listFormat;
	
	/**
	 * Constructs a new map format using the provided key and
	 * value formats.
	 * 
	 * @param keyFormatter the formatter to use for keys.
	 * @param keyParser the parser to use for keys.
	 * @param valueFormatter the formatter to use for values.
	 * @param valueParser the parser to use for values.
	 * 
	 * @throws NullPointerException if either argument is {@code null}.
	 */
    public EqualsMapFormatV1(Formatter<K> keyFormatter,
    						 Parser<K> keyParser,
    						 Formatter<V> valueFormatter,
    						 Parser<V> valueParser) {
        
        // Checking input
        Util.nullCheck(keyFormatter);
        Util.nullCheck(keyParser);
        Util.nullCheck(valueFormatter);
        Util.nullCheck(valueParser);
        
        // Creating formats
        entryFormat = EqualsMapEntryFormatV1.getBasic();
        setFormat = new PipeSetFormat<K>(keyFormatter, keyParser);
        listFormat = new PipeListFormat<V>(valueFormatter, valueParser);
    }
    
	/**
	 * Constructs a new map format using the provided key and
	 * value formats.
	 * 
	 * @param keyFormat the format to use for keys.
	 * @param valueFormat the format to use for values.
	 * 
	 * @throws NullPointerException if either argument is {@code null}.
	 */
    public EqualsMapFormatV1(SimpleFormat<K> keyFormat,
    						 SimpleFormat<V> valueFormat) {

    	this(keyFormat, keyFormat, valueFormat, valueFormat);
    }
    
    /**
     * Returns the format used for variable-value pairs.
     * @return the format used for variable-value pairs.
     */
    public EqualsMapEntryFormatV1<String, String> getMapEntryFormat() {
        return entryFormat;
    }

    /**
     * Returns the format used for sets of variables.
     * @return the format used for sets of variables.
     */
    public PipeSetFormat<K> getSetFormat() {
        return setFormat;
    }

    /**
     * Returns the format used for lists of values.
     * @return the format used for lists of values.
     */
    public PipeListFormat<V> getListFormat() {
        return listFormat;
    }

}
