/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import java.util.*;

import sbrt.kernel.utilities.*;
import sbrt.shell.io.*;

/**
 * This class is used to represent formats for types of 
 * file formats.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 */
public class FileFormatFormat implements SimpleFormat<FileFormat> {

    private static final 
    	UnmodifiableMap<String, FileFormat> fileFormatNames;

    static {

    	// Creating storage
        Map<String, FileFormat> m = new TreeMap<String, FileFormat>();

        // Defining the names of file formats
        m.put("Text", FileFormat.TEXT);
        m.put("Gzipped Text", FileFormat.GZIP_TEXT);
        m.put("Serialized", FileFormat.SERIALIZED);
        m.put("Gzipped Serialized", FileFormat.GZIP_SER);

        // Storing as unmodifiable
        fileFormatNames = new UnmodifiableMap<String, FileFormat>(m);
    }

	/**
	 * Returns a formatted string representation of the
	 * provided file format.
	 * 
	 * @param fileFormat the file format to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided file format.
	 */
    public String format(FileFormat fileFormat) {

        // Getting all name-file format pairs
        Set<Map.Entry<String, FileFormat>> entries = 
        	fileFormatNames.entrySet();

        // Iterating thru each pair
        for (Map.Entry<String, FileFormat> e : entries) {

            // If the provided format equals this format, return this key
            if (fileFormat == e.getValue()) {
                return e.getKey();
            }

        }

        // If the provided format doesn't match the known formats
        String s = "The file format '" + fileFormat + 
        			"' is not yet supported.";
        throw new IllegalArgumentException(s);
    }

    /**
     * Parses the provided string and returns its 
     * corresponding file format.
     * 
     * @param source the string to be parsed.
     * 
     * @return the file format corresponding to the 
     * provided string. 
     */
    public FileFormat parse(String source) {

        // Trimming leading and trailing whitespace
        String trimmed = source.trim();

        // Getting name-file format pairs
        Set<Map.Entry<String, FileFormat>> entries = 
        	fileFormatNames.entrySet();

        // Iterating thru each pair
        for (Map.Entry<String, FileFormat> e : entries) {

            // Getting the defined name
            String name = e.getKey();

            // If the trimmed string equals the defined name,
            // return the corresponding FileFormat
            if (trimmed.equals(name)) {
                return e.getValue();
            }
        }

        // If the provided name didn't match any defined names
        String s = "'" + source + "' is not parsable as a file format.";
        throw new FormatException(s);
    }
}
