/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import java.util.*;

/**
 * This interface is used to represent formats for map entries.
 * 
 * @author This interface was written and documented by 
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public interface MapEntryFormat<K, V> extends Format {

	/**
	 * Returns a formatted string representation of the
	 * provided key.
	 * 
	 * @param <S> a type that extends the primary key
	 * type of this format.
	 * 
	 * @param key the key to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided key.
	 */
    public <S extends K> String formatKey(S key);

	/**
	 * Returns a formatted string representation of the
	 * provided value.
	 * 
	 * @param <T> a type that extends the primary value
	 * type of this format.
	 * 
	 * @param value the value to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided value.
	 */
    public <T extends V> String formatValue(T value);

	/**
	 * Returns a formatted string representation of the
	 * provided key-value pair.
	 * 
	 * @param <S> a type extending the primary key type.
	 * @param <T> a type extending the primary value type.
	 * 
	 * @param key the key.
	 * @param value the value.
	 * 
	 * @return a formatted string representation of the
	 * provided key-value pair.
	 */
    public <S extends K, T extends V> String formatPair(S key, T value);

	/**
	 * Returns a formatted string representation of the
	 * provided key-value pair.
	 * 
	 * @param keyValuePair the key-value pair
	 * to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided key-value pair.
	 */
    public String 
    formatPair(Map.Entry<? extends K, ? extends V> keyValuePair);

    /**
     * Parses the provided string and returns its 
     * corresponding key.
     * 
     * @param key a string representation of a key.
     * 
     * @return the key corresponding to the 
     * provided string. 
     */
    public K parseKey(String key);

    /**
     * Parses the provided string and returns its 
     * corresponding value.
     * 
     * @param value a string representation of a value.
     * 
     * @return the value corresponding to the 
     * provided string. 
     */
    public V parseValue(String value);

    /**
     * Parses the provided string and returns a 
     * corresponding map entry.
     * 
     * @param keyValuePair a string representation of
     * a key-value pair, or in other words, a map entry.
     * 
     * @return a map entry containing the key-value 
     * pair corresponding to the provided string. 
     */
    public Map.Entry<K, V> parsePair(String keyValuePair);

}
