/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import java.util.*;

/**
 * This interface is used to represent formats for maps.
 * <P>
 * There are multiple ways to represent maps as strings.
 * One way is to store the key-value pairs as a single
 * unit, such as <i>key = value</i>. Another way is to
 * store the keys separately from their values, such
 * as
 * <P>
 * <i>key<sub>1</sub> ... key<sub>n</sub></i><P>
 * <i>value<sub>1</sub> ... value<sub>n</sub></i><P>
 * 
 * This interface provides methods to deal with these
 * possibilities.
 * 
 * @author This interface was written and documented by 
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <K> the key type.
 * @param <V> the value type.
 */
public interface MapFormat<K, V> extends Format {

	/**
	 * Returns a formatted string representation of the
	 * provided key.
	 * 
	 * @param <T> a type that extends the primary key
	 * type of this format.
	 * 
	 * @param key the key to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided key.
	 */
    public <T extends K> String formatKey(T key);

	/**
	 * Returns a formatted string representation of the
	 * provided value.
	 * 
	 * @param <T> a type that extends the primary value
	 * type of this format.
	 * 
	 * @param value the value to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided value.
	 */
    public <T extends V> String formatValue(T value);

	/**
	 * Returns a formatted string representation of the
	 * provided set of keys.
	 * 
	 * @param keys the set of keys to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided set of keys.
	 */
    public String formatKeys(Set<? extends K> keys);

	/**
	 * Returns a string representation of the
	 * provided list of values.
	 * 
	 * @param values the list of values to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided list of values.
	 */
    public String formatValues(List<? extends V> values);

	/**
	 * Returns a string representation of the
	 * values in the provided map formatted as a list. The order of
	 * occurrence of values in the returned string depends
	 * on the <i>order</i> of the map. See 
	 * {@code java.util.Map} for further details.
	 * 
	 * @param map keys mapped to their respective values.
	 * 
	 * @return a string representation of the
	 * values in the provided map formatted as a list.
	 */
    public String formatValues(Map<? extends K, ? extends V> map);

	/**
	 * Returns a formatted string representation of the
	 * provided key-value pair.
	 * 
	 * @param <S> a type extending the primary key type.
	 * @param <T> a type extending the primary value type.
	 * 
	 * @param key the key.
	 * @param value the value.
	 * 
	 * @return a formatted string representation of the
	 * provided key-value pair.
	 */
    public <S extends K, T extends V> String formatPair(S key, T value);

	/**
	 * Returns a formatted string representation of the
	 * provided key-value pair.
	 * 
	 * @param keyValuePair the key-value pair
	 * to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided key-value pair.
	 */
    public String 
    formatPair(Map.Entry<? extends K, ? extends V> keyValuePair);

    /**
     * Parses the provided string and returns its 
     * corresponding key.
     * 
     * @param key a string representation of a key.
     * 
     * @return the key corresponding to the 
     * provided string. 
     */
    public K parseKey(String key);

    /**
     * Parses the provided string and returns its 
     * corresponding value.
     * 
     * @param value a string representation of a value.
     * 
     * @return the value corresponding to the 
     * provided string. 
     */
    public V parseValue(String value);

    /**
     * Parses the provided string and returns a 
     * corresponding map entry.
     * 
     * @param keyValuePair a string representation of
     * a key-value pair, or in other words, a map entry.
     * 
     * @return a map entry containing the key-value 
     * pair corresponding to the provided string. 
     */
    public Map.Entry<K, V> parsePair(String keyValuePair);

    /**
     * Parses the provided string and returns its 
     * corresponding set of keys.
     * 
     * @param allKeys a string representation of a set
     * of keys.
     * 
     * @return a set of keys corresponding to the 
     * provided string. 
     */
    public LinkedHashSet<K> parseAllKeys(String allKeys);

    /**
     * Parses the provided string and returns its 
     * corresponding list of values.
     * 
     * @param allValues a string representation of a list
     * of values.
     * 
     * @return a list of values corresponding to the 
     * provided string. 
     */
    public List<V> parseAllValues(String allValues);

    /**
     * Parses the provided string of values and returns a 
     * corresponding map of key-value pairs using the
     * provided set of keys. The association between
     * keys and values is determined from their order of
     * occurrence in their respective collection. In other
     * words, the first key returned by the set's iterator
     * is associated with the first parsed value, and so on.
     *
     * @param keys the set of keys with which 
     * to associate the parsed values.
     * 
     * @param allValues a string representation of a list
     * of values.
     * 
     * @return keys mapped to their respective values.
     */
    public Map<K, V> parseAllValues(LinkedHashSet<? extends K> keys,
    								String allValues);

    /**
     * Parses the provided set of keys and list of values
     * and returns a corresponding map of key-value pairs.
     * The association between keys and values is determined
     * from their order of occurrence in their respective collection.
     * In other words, the first parsed key is
     * associated with the first parsed value, and so on. 
     * 
     * @param allKeys a string representation of a set
     * of variables.
     * 
     * @param allValues a string representation of a list of
     * values.
     * 
     * @return keys mapped to their respective values.
     */
    public Map<K, V> parseAllPairs(String allKeys, String allValues);

}
