/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

/**
 * This class is a concrete implementation of 
 * {@linkplain PercentFormat}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public class PercentFormatVersion1 implements PercentFormat {

	private static final String PERCENT = "%";
	
    private final DoubleFormatV1 doubleFormat;

    /**
     * Constructs a new percent format.
     */
    public PercentFormatVersion1() {
        doubleFormat = new DoubleFormatV1();
    }

    /**
     * Returns the {@code DoubleFormat} used to 
     * parse and format the numerical percentage values.
     * 
     * @return the {@code DoubleFormat} used to 
     * parse and format the numerical percentage values.

     */
    public DoubleFormatV1 getDoubleFormat() {
        return doubleFormat;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided percentage. The percentage is formatted
	 * as a double precision number, truncated after
	 * the second decimal place, and followed by the 
	 * {@code %} sign.
	 * 
	 * @param percentage the percentage to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided percentage.
	 */
    public String format(double percentage) {

        if (percentage < 0.01) {
            return "0.00%";
        }

        // Formatting as a double
        StringBuilder s = new StringBuilder(doubleFormat.format(percentage));

        // Getting the decimal point position
        int point = s.indexOf(".");
        
        // Getting the string length
        int length = s.length();
        
        // Truncating if necessary
        if (point + 3 <= length) {
            s.delete(point + 3, length);
        }
        
        // Padding if necessary
        for (; length - point < 3; length ++) {
            s.append("0");
        }

        // Returning the formatted percentage
        return s + "%";
    }

	/**
	 * Returns a formatted string representation of the
	 * provided percentage. The percentage is formatted
	 * as a double precision number, truncated after
	 * the second decimal place, and followed by the 
	 * {@code %} sign.
	 * 
	 * @param percentage the percentage to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided percentage.
	 */
    public String format(Double percentage) {
        return format(percentage.doubleValue());
    }

    /**
     * Parses the provided string and returns its 
     * corresponding percentage.
     * 
     * @param percentage the string to be parsed.
     * 
     * @return the double precision number corresponding 
     * to the provided string. 
     */
    public Double parse(String percentage) {
    	
    	// Trimming whitespace
    	String trimmed = percentage.trim();
    	
    	// Removing percent sign if it exists
    	if (trimmed.endsWith(PERCENT)) {
    		trimmed = trimmed.substring(0, trimmed.length() - 1);
    	}
    	
    	// Returning double value
    	return doubleFormat.parse(trimmed);
    }

    /**
     * Parses the provided string and returns its 
     * corresponding percentage.
     * 
     * @param percentage the string to be parsed.
     * 
     * @return the double precision number corresponding 
     * to the provided string. 
     */
    public double parsePrimitive(String percentage) {
    	
    	// Trimming whitespace
    	String trimmed = percentage.trim();
    	
    	// Removing percent sign if it exists
    	if (trimmed.endsWith(PERCENT)) {
    		trimmed = trimmed.substring(0, trimmed.length() - 1);
    	}
    	
    	// Returning double value
    	return doubleFormat.parsePrimitive(trimmed);
    }
}
