/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import java.util.regex.*;

/**
 * This class is used to define a format for pipe-delimited
 * sets. If a set is composed of elements 
 * <i>a, b, c</i>,
 * its string representation would be 
 * "{@code a | b | c}".
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <E> the element type.
 */
public final class PipeSetFormat<E> extends AbstractLinkedHashSetFormat<E> { 

	/**
	 * Returns a {@code PipeSetFormat} that does not
	 * format or parse the elements supplied to it. In other
	 * words, it returns a {@code PipeSetFormat} with
	 * the most basic possible functionality. Empty strings are not 
	 * allowed however.
	 *  
	 * @return an {@code PipeSetFormat} that does not
	 * format or parse the elements supplied to it.
	 */
	public static final PipeSetFormat<String> getBasic() {

		// Creating the "trivial" format
		NonEmptyStringFormat f = new NonEmptyStringFormat();
		
		return new PipeSetFormat<String>(f);
	}
	
	/**
	 * Constructs a new pipe-delimited set format with the 
	 * provided element format.
	 * 
	 * @param elementFormat the format of the set-elements.
	 */
    public PipeSetFormat(SimpleFormat<E> elementFormat) {
        this(elementFormat, elementFormat);
    }
	
	/**
	 * Constructs a new pipe-delimited set format with the 
	 * provided element format.
	 * 
	 * @param elementFormatter the formatter for the set-elements.
	 * @param elementParser the parser for the set-elements.
	 */
    public PipeSetFormat(Formatter<E> elementFormatter, Parser<E> elementParser) {
        super(" | ", Pattern.compile("\\s*\\|\\s*"), elementFormatter, elementParser);
    }

}
