/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import java.util.*;

import sbrt.kernel.utilities.Util;

/**
 * This interface is used to wrap {@code ListFormat} objects as {@code SimpleFormat}
 * objects.  
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <E> the type of element.
 */
public class SimpleListFormat<E> implements SimpleFormat<List<E>> {
    
	private final ListFormat<E> listFormat;
	
	/**
	 * Constructs a new simple list format from the provided list format.
	 * @param listFormat the list format to be wrapped.
	 */
	public SimpleListFormat(ListFormat<E> listFormat) {
		Util.nullCheck(listFormat);
		this.listFormat = listFormat;
	}
	
	/**
	 * Returns a formatted string representation of the
	 * provided list.

	 * @param list the list to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided list.
	 */
    public String format(List<E> list) {
    	return listFormat.format(list);
    }
    
    /**
     * Parses the provided string and returns its 
     * corresponding collection of elements as a list.
     * 
     * @param list the string to be parsed.
     * 
     * @return the list of elements 
     * corresponding to the provided string. 
     */	
    public List<E> parse(String list) {
    	return listFormat.parse(list);
    }

    /**
     * Returns the underlying list format.
     * @return the underlying list format.
     */
    public ListFormat<E> getListFormat() {
    	return listFormat;
    }
    
    /**
     * Returns the formatter for 
     * the elements of this list format.
     * 
     * @return the formatter for 
     * the elements of this list format.
     */
    public Formatter<E> getElementFormatter() {
    	return listFormat.getElementFormatter();
    }

    /**
     * Returns the parser for 
     * the elements of this list format.
     * 
     * @return the parser for 
     * the elements of this list format.
     */
    public Parser<E> getElementParser() {
    	return listFormat.getElementParser();
    }
}
