/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

import java.util.*;

import sbrt.kernel.utilities.Util;

/**
 * This interface is used to wrap {@code ListFormat} objects as {@code SimpleFormat}
 * objects.  
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <E> the type of element.
 */
public class SimpleSetFormat<E> implements SimpleFormat<Set<E>> {

	private final LinkedHashSetFormat<E> setFormat;
	
	/**
	 * Constructs a new simple set format from the provided set format.
	 * @param setFormat the set format to be wrapped.
	 */
	public SimpleSetFormat(LinkedHashSetFormat<E> setFormat) {
		
		Util.nullCheck(setFormat);
		this.setFormat = setFormat;
	}
	
	/**
	 * Returns a formatted string representation of the
	 * provided set.

	 * @param set the set to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided set.
	 */
    public String format(Set<E> set) {
    	return setFormat.format(set);
    }
    
    /**
     * Parses the provided string and returns its 
     * corresponding collection of elements as a set.
     * 
     * @param set the string to be parsed.
     * 
     * @return the set of elements 
     * corresponding to the provided string. 
     */	
    public Set<E> parse(String set) {
    	return setFormat.parse(set);
    }

    /**
     * Returns the underlying set format.
     * @return the underlying set format.
     */
    public LinkedHashSetFormat<E> getSetFormat() {
    	return setFormat;
    }
    
    /**
     * Returns the formatter for 
     * the elements of this set format.
     * 
     * @return the formatter for 
     * the elements of this set format.
     */
    public Formatter<E> getElementFormatter() {
    	return setFormat.getElementFormatter();
    }

    /**
     * Returns the parser for 
     * the elements of this set format.
     * 
     * @return the parser for 
     * the elements of this set format.
     */
    public Parser<E> getElementParser() {
    	return setFormat.getElementParser();
    }
}
