/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text;

/**
 * This class is used to define a format for true/false values. 
 * The string
 * "true" is considered here equivalent to the boolean value
 * true, and "false" is considered here equivalent to the 
 * boolean value false. Lower and upper case characters are treated as
 * equivalent.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wager lab.
 */
public final class TrueFalseFormat implements BooleanFormat {

    private static final String trueString = "true";
    private static final String falseString = "false";

    /**
     * Constructs a new true/false format object.
     */
    public TrueFalseFormat() {
    	super();
    }
    
    /**
     * Returns "true" or "false" depending on the provided
     * boolean value.
     * 
	 * @param b the boolean value to be formatted. 
	 * 
     * @return "true" if {@code b} is {@code true}; 
     * "false" if {@code b} is {@code false}.
     */
    public String format(Boolean b) {
        return format(b.booleanValue());
    }

    /**
     * Returns "true" or "false" depending on the provided
     * boolean value.
     * 
	 * @param b the boolean value to be formatted. 
	 * 
     * @return "true" if {@code b} is {@code true}; 
     * "false" if {@code b} is {@code false}.
     */
    public String format(boolean b) {
        return b ? trueString : falseString;
    }

    /**
     * Parses the provided string and returns its
     * corresponding boolean value. The string is first
     * stripped of leading and trailing whitespace and
     * then compared to the strings "true" and "false".
     * The case is ignored during this comparison.
     * 
     * @param b the string to be parsed.
     * 
     * @return {@code true} if {@code b} matches "true";
     * {@code false} if {@code b} matches "false".
     * 
     * @throws FormatException if the provided string 
     * does not match "true" or "false".
     */
    public Boolean parse(String b) {
        return Boolean.valueOf(parsePrimitive(b));
    }

    /**
     * Parses the provided string and returns its
     * corresponding boolean value. The string is first
     * stripped of leading and trailing whitespace and
     * then compared to the strings "true" and "false".
     * The case is ignored during this comparison.
     * 
     * @param b the string to be parsed.
     * 
     * @return {@code true} if {@code b} matches "true";
     * {@code false} if {@code b} matches "false".
     * 
     * @throws FormatException if the provided string 
     * does not match "true" or "false".
     */
    public boolean parsePrimitive(String b) {

        b = b.trim();
        boolean result;

        if (b.equalsIgnoreCase(trueString)) {
            result = true;
        }

        else if (b.equalsIgnoreCase(falseString)) {
            result = false;
        }

        else {
            String s = "'" + b + "' is neither '" + trueString + "' nor '" + falseString + "'.";
            throw new FormatException(s);
        }

        return result;
    }

}
