/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.chem;

import java.util.regex.*;

import sbrt.shell.text.FormatException;
import sbrt.shell.text.SpecialChars;

/**
 * This class is a concrete implemenation of {@linkplain ChemSpeciesFormat}.
 * This format considers chemical species to be {@code String} objects, which
 * are referred to as <i>species names</i>. Species names 
 * cannot contain any whitespace characters (spaces, tabs, or newlines)
 * and their length must be greater than 0.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 */
public final class ChemSpeciesFormatV1 
	implements ChemSpeciesFormat {
	
    private static final Pattern whitespace = Pattern.compile("\\s+");

	/**
	 * Returns the provided chemical species name after 
	 * ensuring it conforms to this format.
	 * 
	 * @param speciesName the species name to be checked.
	 * 
	 * @return the provided species name.
	 * 
	 * @throws IllegalArgumentException if the provided
	 * species name does not conform to this format.
	 * 
	 * @throws NullPointerException if the provided
	 * string is {@code null}.
	 */
    public String format(String speciesName) {

        if (speciesName.length() == 0) {
            String s = "The chemical species name is an empty string";
            throw new IllegalArgumentException(s);
        }

        if (whitespace.matcher(speciesName).matches()) {
            String s = "The chemical species name '" + speciesName +
                       "' contains whitespace characters.";
            throw new IllegalArgumentException(s);
        }

        if (hasIllegalChar(speciesName)) {
        	String s = "The chemical species '" + speciesName +
        			   "' contains an illegal character.";	
        	throw new IllegalArgumentException(s);
        }
        
        return speciesName;
    }

    /**
     * Parses the provided string and returns its 
     * corresponding species name.
     * 
     * @param speciesName the string to be parsed.
     * 
     * @return a copy of the provided string with leading and 
     * trailing white space removed, or the provided string if 
     * it has no leading or trailing white space.
     * 
     * @throws FormatException if the provided string
     * is not a valid species name.
     * 
     * @throws NullPointerException if the provided string
     * is {@code null}.
     */
    public String parse(String speciesName) {

        String trimmedName = speciesName.trim();

        if (trimmedName.length() == 0) {
            String s = "The chemical species name '" + speciesName +
                       "' in invalid.";
            throw new FormatException(s);
        }

        if (whitespace.matcher(trimmedName).matches()) {
            String s = "The chemical species '" + trimmedName +
                       "' contains whitespace characters.";
            throw new FormatException(s);
        }

        if (hasIllegalChar(trimmedName)) {
        	String s = "The chemical species '" + trimmedName +
        			   "' contains an illegal character.";	
        	throw new FormatException(s);
        }
        
        return trimmedName.intern();
    }
    
    private static boolean hasIllegalChar(String s) {
    	
    	for (char ch : SpecialChars.getArray()) {
    		if (s.indexOf(ch) >= 0) {
    			return true;
    		}
    	}
    	
    	return false;
    }
}
