/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.chem;

import java.util.regex.*;

import sbrt.shell.text.FormatException;
import sbrt.shell.text.SpecialChars;

/**
 * This class is a concrete implementation of {@linkplain RxnNameFormat}.
 * This format considers reaction names to be {@code String} objects.
 * Reaction names cannot contain any whitespace characters 
 * (spaces, tabs, or newlines) and their length must be greater than 0.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class RxnNameFormatV1 implements RxnNameFormat<String> {
	
    private static final Pattern whitespace = Pattern.compile("\\s+");

	/**
	 * Returns the provided reaction name after ensuring it 
	 * conforms to this format.
	 * 
	 * @param reactionName the reaction name to be checked.
	 * 
	 * @return the provided reaction name.
	 * 
	 * @throws IllegalArgumentException if the provided
	 * reaction name does not conform to this format.
	 * 
	 * @throws NullPointerException if the provided
	 * string is {@code null}.
	 */
    public String format(String reactionName) {

        if (reactionName.length() == 0) {
            String s = "The reaction name is an empty string.";
            throw new IllegalArgumentException(s);
        }

        if (whitespace.matcher(reactionName).matches()) {
            String s = "The reaction name '" + reactionName +
                       "' contains whitespace characters.";
            throw new IllegalArgumentException(s);
        }

        if (hasIllegalChar(reactionName)) {
        	String s = "The reaction name '" + reactionName +
        			   "' contains an illegal character.";	
        	throw new IllegalArgumentException(s);
        }
        
        return reactionName;
    }

    /**
     * Parses the provided string and returns its 
     * corresponding reaction name.
     * 
     * @param reactionName the string to be parsed.
     * 
     * @return a copy of the provided string with leading and 
     * trailing white space removed, or the provided string if 
     * it has no leading or trailing white space.
     * 
     * @throws FormatException if the provided string
     * is not a valid reaction name.
     * 
     * @throws NullPointerException if the provided string
     * is {@code null}.
     */
    public String parse(String reactionName) {

        String trimmed = reactionName.trim();

        if (trimmed.length() == 0) {
            String s = "The reaction name '" + reactionName +
                       "' is invalid.";
            throw new FormatException(s);
        }

        if (whitespace.matcher(trimmed).matches()) {
            String s = "The reaction name '" + trimmed +
                       "' contains whitespace characters.";
            throw new FormatException(s);
        }

        if (hasIllegalChar(reactionName)) {
        	String s = "The reaction name '" + reactionName +
        			   "' contains an illegal character.";	
        	throw new FormatException(s);
        }
        
        return trimmed.intern();
    }
    
    private static boolean hasIllegalChar(String s) {
    	
    	for (char ch : SpecialChars.getArray()) {
    		if (s.indexOf(ch) >= 0) {
    			return true;
    		}
    	}
    	
    	return false;
    }
}
