/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Pattern;

import sbrt.kernel.utilities.And;
import sbrt.kernel.utilities.Or;

/**
 * This class is used to parse bracketed expressions contained in SBML files
 * from the BiGG Database. These expressions are used to represent
 * gene-protein-reaction associations using boolean logic. An example would be:
 * </p>
 * 
 * <center>((b3951 and b3952) or b3114 or (b0902 and b0903)).</center>
 * 
 * <p>
 * All provided expressions are assumed to be formatted like the above example
 * and free of errors. If this is not the case, the results of parsing are
 * unpredictable.
 * </p>
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 */
public class BiggBracketParserV1 {

    private static final char openBracket = '(';

    private static final char closeBracket = ')';
    
    private static final Pattern and = Pattern.compile("\\s+and\\s+");

    private static final Pattern or = Pattern.compile("\\s+or\\s+");

    /**
     * Used for testing purposes.
     * 
     * @param args ignored
     */
    public static void main(String[] args) {

        Map<String, Integer> tests = new LinkedHashMap<String, Integer>();

        tests.put("AckA", 1);
        
        // With spaces around brackets
        tests.put("( b1900 ) and ( b1898 and b1899 ) and ( b1901 )", 1); 
        tests.put("( b1817 and b1818 and b1819 and b2415 and b2416 )", 1);
        tests.put(" ( IlvB ) or ( IlvG ) or ( IlvH )", 3);
        tests.put("( ( b3670 and b3671 ) or ( b3767 and b3768 and b3769 ) or ( b0077 and b0078 ) )",
                        3);
        tests.put("( NagE and PtsH and PtsI ) or ( Crr and PtsG and PtsH and PtsI )",
                        2);
        tests.put("( ( b0679 and b2415 and b2416 ) or ( b2417 and b1101 and b2415 and b2416 ) )",
                        2);
        tests.put("( b2530 and b3992 and ( b3990 and b3991 ) and b0423 and b4407 )",
                        1);
        tests.put("( b1814 or b2797 or b3708 or ( b3111 and b3112 ) )", 4);
        tests.put("( SdaAec ) or ( SdaB ) or ( TnaA ) or ( TdcG )", 4);
        tests.put("( ( b3951 and b3952 ) or b3114 or ( b0902 and b0903 ) )", 3);
        tests.put("( b2530 and b3992 and ( b3990 and b3991 ) and b0423 and b4407 )",
                        1);
        tests.put("( PotA and PotB and PotC and PotDec ) or ( YdcS and YdcT and YdcU and YdcV )",
                        2);
        tests.put("( ( x ) and ( y and ( z and q ) ) and ( ( w ) and ( r ) ) )",
                1);
        tests.put("( ( x ) or ( y and ( z and q ) ) or ( ( w ) and ( r ) ) )",
                3);

        // Without spaces around brackets
        tests.put("(b1900) and (b1898 and b1899) and (b1901)", 1); 
        tests.put("(b1817 and b1818 and b1819 and b2415 and b2416)", 1);
        tests.put("(IlvB) or (IlvG) or (IlvH)", 3);
        tests.put("((b3670 and b3671) or (b3767 and b3768 and b3769) or (b0077 and b0078))",
                        3);
        tests.put("(NagE and PtsH and PtsI) or (Crr and PtsG and PtsH and PtsI)",
                        2);
        tests.put("((b0679 and b2415 and b2416) or (b2417 and b1101 and b2415 and b2416))",
                        2);
        tests.put("(b2530 and b3992 and (b3990 and b3991) and b0423 and b4407)",
                        1);
        tests.put("(b1814 or b2797 or b3708 or (b3111 and b3112))", 4);
        tests.put("(SdaAec) or (SdaB) or (TnaA) or (TdcG )", 4);
        tests.put("((b3951 and b3952) or b3114 or (b0902 and b0903))", 3);
        tests.put("(b2530 and b3992 and (b3990 and b3991) and b0423 and b4407)",
                        1);
        tests.put("(PotA and PotB and PotC and PotDec) or (YdcS and YdcT and YdcU and YdcV)",
                        2);
        tests.put("((x) and (y and (z and q)) and ((w) and (r)))",
                1);
        tests.put("((x) or (y and (z and q)) or ((w) and (r)))",
                3);
        
        BiggBracketParserV1 parser = new BiggBracketParserV1();

        for (Map.Entry<String, Integer> e : tests.entrySet()) {

            Or<And<String>> pieces = parser.parse(e.getKey());

             //System.out.println("'"+pieces+"'");

            if (pieces.elements() != e.getValue().intValue()) {
                
                String s1 = "The string '" + e.getKey() + "' could not be parsed.";
                String s2 = "Parsed pieces = " + pieces.elements();
                
                
                System.err.println(s1);
                System.err.println(s2);
                System.exit(1);
            }
        }

        System.out.println("No errors were detected.");
    }

    /**
     * Parses the provided BiGG style bracketed expression.
     * 
     * @param string a BiGG-type bracketed expressions with parantheses used
     * as brackets.
     * 
     * @return the boolean rules represented by the string.
     */
    public Or<And<String>> parse(String string) {

        StringBuilder sb = new StringBuilder(string);

        // Stripping away all brackets
        removeBrackets(sb);
        
        // Splitting on "or"
        String[] groups = or.split(sb, -1); 
        
        // Creating storage for 'and' groups
        List<And<String>> andGroups = new LinkedList<And<String>>();
        
        // Collecting all 'and' groups together
        for (String g : groups) {
            andGroups.add(getAnd(g));
        }

        return new Or<And<String>>(andGroups);
    }

    private void removeBrackets(StringBuilder sb) {
        
        for (int i = 0, size = sb.length(); i < size; i ++) {
            
            char c = sb.charAt(i);
            
            if (c == openBracket || c == closeBracket) {
                sb.setCharAt(i, ' ');
            }
            
        }
        
    }

    private And<String> getAnd(String string) {
        
        String[] elements = and.split(string.trim(), -1);
        
        Set<String> s = new TreeSet<String>();
        
        for (String e : elements) {
            s.add(e);
        }
        
        return new And<String>(s);
    }
    
}
