/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import sbrt.kernel.math.HashLinearComb;
import sbrt.kernel.math.Interval;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.math.LinearComb.Term;
import sbrt.kernel.utilities.And;
import sbrt.kernel.utilities.Or;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.TrueFalseFormat;

/**
 * This class is used to make a direct translation from SBML 'reaction' nodes into
 * Systems Biology Research Tool objects. The names of reactions or chemical species 
 * are not checked or processed in any way during this translation. This class works
 * for SBML files downloaded directly from the BiGG Database.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public class BiggRxnNodeParser {

    /**
     * The name of the nodes parsed by this class ('reaction').
     */
    public static final String RXN_NODE_NAME = "reaction";
    
    private static final Or<And<String>> emptyRules;
    
    static {
        Set<And<String>> and = Collections.emptySet(); 
        emptyRules = new Or<And<String>>(and);
    }
    
    private final DoubleFormatV1 numFormat;
    private final TrueFalseFormat trueFalseFormat;
    private final BiggBracketParserV1 rulesParser;

    /**
     * Constructs a new SBML reaction node parser.
     */
    public BiggRxnNodeParser() {
        numFormat = new DoubleFormatV1();
        trueFalseFormat = new TrueFalseFormat();
        rulesParser = new BiggBracketParserV1();
    }
    
    /**
     * Returns the reaction ID of the provided reaction node.
     * @param reactionNode the reaction node whose ID is to be returned.
     * @return the reaction ID of the provided reaction node.
     */
    public String getRxnId(Node reactionNode) {

        // Ensuring the provided node is a 'reaction' node
        checkNodeName(reactionNode, RXN_NODE_NAME);
        
        // Getting reaction attributes
        NamedNodeMap attributes = getAttributes(reactionNode);

        // Getting reaction name (or id)
        return parseRxnId(attributes);
    }
    
    @SuppressWarnings("null")
	private String parseRxnId(NamedNodeMap rxnAttributes) {

        Node id = rxnAttributes.getNamedItem("id");

        if (id == null) {
            String s = "A 'reaction' node does not contain an 'id' attribute.";
            formatError(s);
        }

        // Getting the reaction ID from the file
        String value = id.getNodeValue();

        if (value == null) {
            String s =
                    "The 'id' attribute for a 'reaction' node does not have a value.";
            formatError(s);
        }

        return value;
    }

    /**
     * Returns the flux bounds of the provided reaction node.
     * @param reactionNode the reaction node whose flux bounds are to be returned.
     * @return the flux bounds of the provided reaction node.
     */
    public Interval getFluxBounds(Node reactionNode) {

        double lower = getLowerBound(reactionNode);
        double upper = getUpperBound(reactionNode);
        
        if (lower > upper) {
            String s = "The lower bound '" + lower + 
                       "' is greater than the upper bound '" + upper +
                       "'.";
            formatError(s);
        }
        
        return new Interval(lower, upper);
    }

    /**
     * Returns the lower flux bound of the provided reaction node.
     * @param reactionNode the reaction node whose lower flux bound is to be returned.
     * @return the lower flux bound of the provided reaction node.
     */
    public double getLowerBound(Node reactionNode) {

        // Ensuring the provided node is a 'reaction' node
        checkNodeName(reactionNode, RXN_NODE_NAME);

        NodeList parameterNodes = getParameterNodes(reactionNode);
        
        return getParameterValue(parameterNodes, "LOWER_BOUND");
    }

    /**
     * Returns the upper flux bound of the provided reaction node.
     * @param reactionNode the reaction node whose upper flux bound is to be returned.
     * @return the upper flux bound of the provided reaction node.
     */
    public double getUpperBound(Node reactionNode) {

        // Ensuring the provided node is a 'reaction' node
        checkNodeName(reactionNode, RXN_NODE_NAME);

        NodeList parameterNodes = getParameterNodes(reactionNode);
        
        return getParameterValue(parameterNodes, "UPPER_BOUND");
    }

    /**
     * Returns the flux value of the provided reaction node.
     * @param reactionNode the reaction node whose flux value are to be returned.
     * @return the flux value of the provided reaction node.
     */
    public double getFluxValue(Node reactionNode) {

        // Ensuring the provided node is a 'reaction' node
        checkNodeName(reactionNode, RXN_NODE_NAME);

        NodeList parameterNodes = getParameterNodes(reactionNode);
        
        return getParameterValue(parameterNodes, "FLUX_VALUE");
    }
    
    private double getParameterValue(NodeList parameters, String id) {
        
        int size = parameters.getLength();
        
        // Iterating thru each parameter node
        for (int i = 0; i < size; i ++) {
            
            Node param = parameters.item(i);

            if (!param.getNodeName().equals("parameter")) {
                continue;
            }
            
            // Getting parameter attributes
            NamedNodeMap attributes = getAttributes(param);
            
            // Getting the id node
            Node idNode = attributes.getNamedItem("id");
            
            // If this node's id matches the provided id string
            if (idNode.getNodeValue().equals(id)) {
                
                // Getting the value node
                Node valueNode = attributes.getNamedItem("value");
                
                // Getting the string value of the value node
                String number = valueNode.getNodeValue();
                
                // Parsing as a double and returning
                return numFormat.parsePrimitive(number);
            }
        }
        
        String s = "A group of 'parameter' nodes does not contain the id '" +
                   id + "'.";
        formatError(s);
        
        return Double.NaN; // never reached
    }
    
    @SuppressWarnings("null")
	private NodeList getParameterNodes(Node reactionNode) {

        checkNodeName(reactionNode, RXN_NODE_NAME);
        
        NodeList childNodes = reactionNode.getChildNodes();

        Node kineticLawNode = getNode(childNodes, "kineticLaw");

        if (kineticLawNode == null) {
            String s =
                    "A 'reaction' node does not have a 'kineticLaw' child node.";
            formatError(s);
        }

        childNodes = kineticLawNode.getChildNodes();
        
        Node parametersNode = getNode(childNodes, "listOfParameters");
        
        if (parametersNode == null) {
            String s = "A 'kineticLaw' node does not have a 'listOfParameters' child node.";
            formatError(s);
        }
        
        return parametersNode.getChildNodes();
    }

    
    /**
     * Returns the reactants of the provided reaction node.
     * @param reactionNode the reaction node whose reactants are to be returned.
     * @return the reactants of the provided reaction node.
     */
    public LinearComb<String> getReactants(Node reactionNode) {

        // Ensuring the provided node is a 'reaction' node
        checkNodeName(reactionNode, RXN_NODE_NAME);
        
        // Getting the reactants
        return parseReactants(reactionNode);
    }

    private LinearComb<String> parseReactants(Node reaction) {

        NodeList childNodes = reaction.getChildNodes();

        Node reactantsNode = getNode(childNodes, "listOfReactants");

        if (reactantsNode == null) {
            String s =
                    "A 'reaction' node does not have a 'listOfReactants' child node.";
            formatError(s);
        }

        LinearComb<String> reactants = parseStoichExpr(reactantsNode);
        
        if (!reactants.hasVariables()) {
            String s = "No reactants were provided for the reaction '" +
                        getRxnId(reaction) + "'.";
            formatError(s);
        }

        return reactants;
    }
    
    /**
     * Returns the products of the provided reaction node.
     * @param reactionNode the reaction node whose products are to be returned.
     * @return the products of the provided reaction node.
     */
    public LinearComb<String> getProducts(Node reactionNode) {

        // Ensuring the provided node is a 'reaction' node
        checkNodeName(reactionNode, RXN_NODE_NAME);
        
        // Getting the reactants
        return parseProducts(reactionNode);
    }

    private LinearComb<String> parseProducts(Node reaction) {

        NodeList childNodes = reaction.getChildNodes();

        Node productsNode = getNode(childNodes, "listOfProducts");

        if (productsNode == null) {
            String s =
                    "A 'reaction' node does not have a 'listOfProducts' child node.";
            formatError(s);
        }

        LinearComb<String> products = parseStoichExpr(productsNode);
        
        if (!products.hasVariables()) {
            String s = "No products were provided for the reaction '" +
            			getRxnId(reaction) + "'.";
            formatError(s);
        }
        
        return products;
    }

    /**
     * Indicates if the reaction in the provided reaction node is reversible.
     * @param reactionNode the reaction node whose reversibility is to be determined.
     * @return {@code true} if the reaction node indicates a reversible reaction;
     * {@code false} if the reaction node indicates an irreversible reaction.
     */
    public boolean isReversible(Node reactionNode) {

        // Ensuring the provided node is a 'reaction' node
        checkNodeName(reactionNode, RXN_NODE_NAME);
        
        // Getting reaction attributes
        NamedNodeMap attributes = getAttributes(reactionNode);
        
        // Getting the reactants
        return parseReversible(attributes);
    }

    @SuppressWarnings("null")
	private boolean parseReversible(NamedNodeMap rxnAttributes) {

        Node node = rxnAttributes.getNamedItem("reversible");

        if (node == null) {
            String s =
                    "A 'reaction' node does not contain a 'reversible' attribute.";
            formatError(s);
        }

        String reversible = node.getNodeValue();

        if (reversible == null) {
            String s =
                    "The 'reversible' attribute for a 'reaction' node does not have a value.";
            formatError(s);
        }

        return trueFalseFormat.parsePrimitive(reversible);
    }
    
    /**
     * Calls {@linkplain #getCatalysts(Node, String)} with "{@code GENE ASSOCIATION:}"
     * as the {@code prefix}.
     * @param reactionNode the reaction node for which the genes are to be returned.
     * @return the boolean logic rules for the genes of the provided reaction node.
     */
    public Or<And<String>> getGenes(Node reactionNode) {
        return getCatalysts(reactionNode, "GENE ASSOCIATION:");
    }

    /**
     * Returns the boolean logic rules for the catalysts of the provided
     * reaction node. 
     * 
     * @param reactionNode the reaction node for which the catalysts are to be returned.
     * @param prefix the string used to denote these catalysts in the SBML file.
     * 
     * @return the boolean logic rules for the catalysts of the provided
     * reaction node. 
     */
    public Or<And<String>> getCatalysts(Node reactionNode, String prefix) {

        // Ensuring the provided node is a 'reaction' node
        checkNodeName(reactionNode, RXN_NODE_NAME);

        // Getting child nodes
        NodeList childNodes = reactionNode.getChildNodes();

        // Getting gene and protein associations
        Node notes = getNode(childNodes, "notes");

        if (notes == null) {
            String s =
                    "A 'reaction' node does not have a child node called 'notes'.";
            formatError(s);
        }

        return parseCatalystAssoc(notes, prefix);
    }
    
    @SuppressWarnings("null")
	private Or<And<String>> parseCatalystAssoc(Node notes, String prefix) {

        checkNodeName(notes, "notes");

        // Getting HTML paragraph nodes
        NodeList notesNodes = notes.getChildNodes();

        int length = notesNodes.getLength();

        int prefixLength = prefix.length();

        // Iterating thru each <html:p> node
        for (int i = 0; i < length; i++) {

            Node p = notesNodes.item(i);

            // Getting the text is contains
            String text = p.getTextContent().trim();

            if (text == null) {
                String s = "An 'html:p' node does not contain a text value.";
                formatError(s);
            }

            text = text.trim();

            // If it matches the provided prefix
            if (text.startsWith(prefix)) {

                // Getting the post-prefix section
                String suffix = text.substring(prefixLength).trim();
                
                if (suffix.length() == 0) {
                    return emptyRules;
                }

                return rulesParser.parse(suffix);
            }

        }

        // If the appropriate paragraph node does not exist
        return emptyRules;
    }

    private NamedNodeMap getAttributes(Node node) {
        
        NamedNodeMap attributes = node.getAttributes();

        if (attributes == null) {
            String s = "A '" + node.getNodeName() + "' node does not contain attributes.";
            formatError(s);
        }

        return attributes;
    }
    
    private void checkNodeName(Node node, String expectedName) {

        String nodeName = node.getNodeName();
        
        // Checking the nodes name
        if (!nodeName.equals(expectedName)) {
            
            String s = "The provided node is named '" + nodeName + 
            "' instead of '" + expectedName + "'";
            
            throw new IllegalArgumentException(s);
            
        }
        
    }



    private LinearComb<String> parseStoichExpr(Node listOfSpecies) {

        NodeList speciesList = listOfSpecies.getChildNodes();

        return parseStoichExpr(speciesList);
    }

    private LinearComb<String> parseStoichExpr(NodeList speciesList) {

        // Creating storage for all terms of the stoichiometric expression
        List<Term<String>> terms =
                new LinkedList<Term<String>>();

        int length = speciesList.getLength();

        // Iterating thru each node
        for (int i = 0; i < length; i++) {

            Node n = speciesList.item(i);

            // If it is a speciesReference node
            if (n.getNodeName().equals("speciesReference")) {

                LinearComb.Term<String> term = parseStoichTerm(n);

                terms.add(term);
            }
        }

        // Returning stoichiometric expression made from terms
        return new HashLinearComb<String>(terms, 0d, 0d);
    }

    @SuppressWarnings("null")
	private LinearComb.Term<String> parseStoichTerm(Node speciesReference) {

        NamedNodeMap attributes = speciesReference.getAttributes();

        if (attributes == null) {
            String s = "A 'speciesReference' node does not contain attributes.";
            formatError(s);
        }

        Node speciesNode = attributes.getNamedItem("species");
        Node stoichNode = attributes.getNamedItem("stoichiometry");

        if (speciesNode == null) {
            String s =
                    "A 'speciesReference' node does not contain a 'species' attribute.";
            formatError(s);
        }

        if (stoichNode == null) {
            String s =
                    "A 'speciesReference' node does not contain a 'stoichiometry' attribute.";
            formatError(s);
        }

        String speciesValue = speciesNode.getNodeValue();
        String stoichValue = stoichNode.getNodeValue();

        if (speciesValue == null) {
            String s =
                    "The 'species' attribute for a 'speciesReference' node "
                            + "does not have a value.";
            formatError(s);
        }

        if (stoichValue == null) {
            String s =
                    "The 'stoichiometry' attribute for a 'speciesReference' node "
                            + "does not have a value.";
            formatError(s);
        }

        // Parsing stoichiometric coefficient
        double x = numFormat.parsePrimitive(stoichValue);

        return new HashLinearComb.Term<String>(speciesValue, x);
    }

    private Node getNode(NodeList nodeList, String name) {

        int length = nodeList.getLength();

        for (int i = 0; i < length; i++) {

            Node n = nodeList.item(i);

            if (n.getNodeName().equals(name)) {
                return n;
            }

        }

        return null;
    }

    private void formatError(String msg) {
        throw new FormatException(msg);
    }

}
