/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;

/**
 * This class is a concrete implemenation of {@linkplain CatalystSetFormat}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class CatalystSetFormatV1
        implements CatalystSetFormat {

    private final CatalystVerifier catalystChecker;
    private final PipeSetFormat<String> setFormat;

    /**
     * Constructs a new catalyst name collection format
     * using the provided fluxome.
     * 
     * @param fluxome the fluxome used to verify catalyst
     * names.
     */
    public CatalystSetFormatV1(CatalyzedFluxome fluxome) {

    	// Checking input
    	Util.nullCheck(fluxome);
        
    	// Creating internal formats
    	catalystChecker = new CatalystVerifier(fluxome);
        setFormat = new PipeSetFormat<String>(catalystChecker);
    }

    /**
     * Returns the fluxome used to verify catalyst names.
     * @return the fluxome used to verify catalyst names.
     */
    public CatalyzedFluxome getFluxome() {
        return catalystChecker.getFluxome();
    }

    /**
     * Returns the format used for sets of catalyst names.
     * @return the format used for sets of catalyst names.
     */
    public PipeSetFormat<String> getSetFormat() {
        return setFormat;
    }

    /**
     * Returns the catalyst name verifier.
     * @return the catalyst name verifier.
     */
    public CatalystVerifier getEnzymeChecker() {
        return catalystChecker;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided set of catalyst names.

	 * @param catalystNames the set of catalyst names to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided set of catalyst names.
	 */
    public String format(Set<String> catalystNames) {
        return setFormat.format(catalystNames);
    }

    /**
     * Parses the provided string and returns the
     * set of catalyst names it contains.
     * 
     * @param catalystNames the string to be parsed.
     * 
     * @return the set of catalyst names 
     * contained in the provided string. 
     */
    public LinkedHashSet<String> parse(String catalystNames) {
        return setFormat.parse(catalystNames);
    }

}
