/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import sbrt.kernel.fba.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;
import sbrt.shell.text.chem.*;

/**
 * This class is used to ensure catalysts exist in a given {@code CatalyzedFluxome}. 
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public class CatalystVerifier implements CatalystFormat {

    private final CatalyzedFluxome fluxome;

    /**
     * Constructs a new catalyst name verifier for the provided fluxome.
     * 
     * @param fluxome the fluxome whose catalysts will be used for comparison.
     */
    public CatalystVerifier(CatalyzedFluxome fluxome) {
        Util.nullCheck(fluxome);
        this.fluxome = fluxome;
    }

    /**
     * Returns the fluxome used to create this catalyst name verifier.
     * 
     * @return the fluxome used to create this catalyst name verifier.
     */
    public CatalyzedFluxome getFluxome() {
        return fluxome;
    }

    /**
     * Returns the provided catalyst name after ensuring it
     * exists in the corresponding fluxome of this catalyst name verifier.
     * 
     * @param catalyst the name of a catalyst in this
     * catalyst name verifier's fluxome.
     * 
     * @return the provided catalyst name.
     * 
     * @throws IllegalArgumentException if the provided string
     * is not the name of a catalyst in this catalyst name verifier's fluxome.
     */
    public String format(String catalyst) {

        if (!fluxome.isCatalyst(catalyst)) {
            String s = "Not a valid catalyst name: " + catalyst;
            throw new IllegalArgumentException(s);
        }

        return catalyst;
    }

    /**
     * Parses the provided string and returns its 
     * corresponding catalyst name.
     * 
     * @param catalystString the string to be parsed.
     * 
     * @return a copy of the provided string with leading and 
     * trailing white space removed, or the provided string if 
     * it has no leading or trailing white space.
     * 
     * @throws FormatException if the provided string
     * is not a valid catalyst name in the fluxome used to
     * create this catalyst name verifier.
     * 
     * @throws NullPointerException if the provided string
     * is {@code null}.
     */
    public String parse(String catalystString) {

        String trimmed = catalystString.trim();

        if (!fluxome.isCatalyst(trimmed)) {
            String s = "A catalyst named '" + trimmed +
                       "' does not exist in the provided reaction network.";
            throw new FormatException(s);
        }

        return trimmed.intern();
    }

}
