/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import sbrt.kernel.fba.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;
import sbrt.shell.text.chem.*;

/**
 * This class is used to ensure chemical species exist in a 
 * given {@code Fluxome}. 
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public class ChemSpeciesVerifier implements ChemSpeciesFormat {

    private final Fluxome fluxome;

    /**
     * Constructs a new chemical species verifier for the 
     * provided fluxome.
     * 
     * @param fluxome the fluxome whose chemical species
     * will be used for comparison.
     */
    public ChemSpeciesVerifier(Fluxome fluxome) {
        Util.nullCheck(fluxome);
        this.fluxome = fluxome;
    }

    /**
     * Returns the fluxome used to create this chemical species verifier.
     * 
     * @return the fluxome used to create this chemical species verifier.
     */
    public Fluxome getFluxome() {
        return fluxome;
    }

    /**
     * Returns the provided chemical species after ensuring it
     * exists in the corresponding fluxome of this chemical species verifier.
     * 
     * @param validChemSpecies the name of a chemical species in this
     * metabolite name checker's fluxome.
     * 
     * @return the provided chemical species after ensuring it
     * exists in the corresponding fluxome of this chemical species verifier.
     * 
     * @throws IllegalArgumentException if the provided string
     * is not the name of a chemical species in this chemical species verifier's fluxome.
     */
    public String format(String validChemSpecies) {

        if (!fluxome.isSpecies(validChemSpecies)) {
            String s = "Not a valid chemical species: " + validChemSpecies;
            throw new IllegalArgumentException(s);
        }

        return validChemSpecies;
    }

    /**
     * Parses the provided string and returns its 
     * corresponding chemical species.
     * 
     * @param sourceChemSpecies the string to be parsed.
     * 
     * @return a copy of the provided string with leading and 
     * trailing white space removed, or the provided string if 
     * it has no leading or trailing white space.
     * 
     * @throws FormatException if the provided string
     * is not a valid chemical species in the fluxome used to
     * create this chemical species verifier.
     * 
     * @throws NullPointerException if the provided string
     * is {@code null}.
     */
    public String parse(String sourceChemSpecies) {

        String trimmed = sourceChemSpecies.trim();

        if (!fluxome.isSpecies(trimmed)) {
            String s = "A chemical species named '" + trimmed +
                       "' does not exist in the provided network.";
            throw new FormatException(s);
        }

        return trimmed;
    }

}
