/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.shell.text.*;

/**
 * This class is a concrete implementation of 
 * {@linkplain FbaOptFileLineFormat}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class FbaOptFileLineFormatV1 
	implements FbaOptFileLineFormat {

	private final FbaOptHeaderParserV1 headerParser;
	private final FbaOptHeaderFormatterV1 headerFormatter;
	
	private final DoubleFormatV1 doubleFormat;
	private final EqualsMapEntryFormatV1<String, String> entryFormat;
	private final PipeSetFormat<String> setFormat;
	private final PipeListFormat<Double> listFormat;

	/**
	 * Constructs a new FBA optimization file line format.
	 * 
	 * @param fluxome the fluxome used to verify reaction 
	 * names.
	 */
    public FbaOptFileLineFormatV1(Fluxome fluxome) {
        
        // Creating internal formats
        doubleFormat = new DoubleFormatV1();
        headerParser = new FbaOptHeaderParserV1(fluxome);
        headerFormatter = new FbaOptHeaderFormatterV1(fluxome);
    	
        entryFormat = EqualsMapEntryFormatV1.getBasic();
        setFormat = PipeSetFormat.getBasic();
        listFormat = new PipeListFormat<Double>(doubleFormat);
    }

    /**
     * Indicates if the provided string is the header used
     * to indicate the value of the objective function.
     * 
     * @param s the string to compare.
     * 
     * @return {@code true} if the provided string is equal
     * to the string returned by {@linkplain #getObjValueHeader()};
     * {@code false} otherwise.
     */
    public boolean isObjValueHeader(String s) {
        return headerParser.isObjValueHeader(s);
    }

	/**
	 * Returns the header used to indicate the value of the
	 * objective function.
	 * 
	 * @return the header used to indicate the value of the
	 * objective function.
	 */
    public String getObjValueHeader() {
        return headerParser.getObjValueHeader();
    }

    /**
     * Returns the format used for doubles.
     * @return the format used for doubles.
     */
    public DoubleFormatV1 getDoubleFormat() {
        return doubleFormat;
    }

    /**
     * Returns the fluxome used to verify reaction names.
     * @return the fluxome used to verify reaction names.
     */
    public Fluxome getFluxome() {
        return headerParser.getFluxome();
    }

	/**
	 * Returns a formatted string representation of the
	 * provided key.
	 * 
	 * @param key the key to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided key.
	 */
	public String formatKey(String key) {
		return headerFormatter.format(key);
	}

	/**
	 * Returns a formatted string representation of the
	 * provided value.
	 * 
	 * @param value the value to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided value.
	 */
	public String formatValue(Double value) {
		return doubleFormat.format(value);
	}

	/**
	 * Returns a formatted string representation of the
	 * provided set of keys.
	 * 
	 * @param keys the set of keys to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided set of keys.
	 */
	public String formatKeys(Set<? extends String> keys) {

		// Determining appropriate size
		int size = keys.size() + 1;
		
		// Creating temporary storage
		Set<String> formatted = new LinkedHashSet<String>(size, 1.0F); 
		
		// Formatting each provided key
		for (String k : keys) {
			formatted.add(headerFormatter.format(k));
		}
		
		// Returning formatted set
		return setFormat.format(formatted);
	}

	/**
	 * Returns a string representation of the
	 * provided list of values.
	 * 
	 * @param values the list of values to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided list of values.
	 */
	public String formatValues(List<? extends Double> values) {
		return listFormat.format(values);
	}

	/**
	 * Returns a string representation of the
	 * values in the provided map formatted as a list. The order of
	 * occurrence of values in the returned string depends
	 * on the <i>order</i> of the map. See 
	 * {@code java.util.Map} for further details.
	 * 
	 * @param map keys mapped to their respective values.
	 * 
	 * @return a string representation of the
	 * values in the provided map formatted as a list.
	 */
    public String formatValues(Map<? extends String, ? extends Double> map) {

    	// Creating storage for values
        List<Double> values = new LinkedList<Double>();

        // Iterating thru each variable-value pair
        for (Map.Entry<? extends String, ? extends Double> e : map.entrySet()) {

        	// Adding value to the rest
        	values.add(e.getValue());
        }

        return listFormat.format(values);
    }

    /**
     * Returns a string representation of the provided
     * key-value pair.
     * 
     * @param key the key.
     * @param value the value.
     * 
     * @return a string representation of the provided
     * key-value pair.
     */
    public String formatPair(String key, Double value) {
    	
    	// Formatting both
    	String k = formatKey(key);
    	String v = formatValue(value);
    	
    	// Returning formatted key-value pair
        return entryFormat.formatPair(k, v);
    }

	/**
	 * Returns a formatted string representation of the
	 * provided map entry.
	 * 
	 * @param entry the map entry.
	 * 
	 * @return a formatted string representation of the
	 * provided map entry.
	 */
    public String 
    formatPair(Map.Entry<? extends String, ? extends Double> entry) {

    	// Returning formatted key-value pair
        return formatPair(entry.getKey(), entry.getValue());
    }

    /**
     * Parses the provided string and returns its 
     * corresponding key.
     * 
     * @param key a string representation of a key.
     * 
     * @return the key corresponding to the 
     * provided string. 
     */
	public String parseKey(String key) {

		// Parsing string
		Set<String> parsed = headerParser.parse(key);
		
		// Checking size
		if (parsed.size() != 1) {
			
			String s = "The string '" + key + "' is neither '" +
			headerParser.getObjValueHeader() + 
			"' nor a valid reaction name in the provided fluxome.";

			throw new FormatException(s);
		}
		
		// Returning the only element in the set
		return parsed.iterator().next();
	}

    /**
     * Parses the provided string and returns its 
     * corresponding value.
     * 
     * @param value a string representation of a value.
     * 
     * @return the value corresponding to the 
     * provided string. 
     */
	public Double parseValue(String value) {
		return doubleFormat.parse(value);
	}

    /**
     * Parses the provided map entry string, and
     * returns its corresponding map entry.
     * 
     * @param mapEntry the string representation of a
     * single map entry.
     * 
     * @return a singleton map corresponding to the 
     * provided string. 
     */
    public Map.Entry<String, Double> parsePair(String mapEntry) {
    	
    	// Breaking apart the pair
    	Map.Entry<String, String> e = entryFormat.parsePair(mapEntry); 
    	
    	// Parsing each
    	String key = parseKey(e.getKey());
    	Double value = parseValue(e.getValue());
    	
    	// Creating a singelton map
    	Map<String, Double> m = Collections.singletonMap(key, value);
    	
    	// Returning its only entry
    	return m.entrySet().iterator().next();
    }

    /**
     * Parses the provided string and returns its 
     * corresponding set of keys.
     * 
     * @param allKeys a string representation of a set
     * of keys.
     * 
     * @return a set of keys corresponding to the 
     * provided string. 
     */
	public LinkedHashSet<String> parseAllKeys(String allKeys) {
		return headerParser.parse(allKeys);
	}

    /**
     * Parses the provided string and returns its 
     * corresponding list of values.
     * 
     * @param allValues a string representation of a list
     * of values.
     * 
     * @return a list of values corresponding to the 
     * provided string. 
     */
    public List<Double> parseAllValues(String allValues) {
        return listFormat.parse(allValues);
    }

    /**
     * Parses the provided string of values and returns a 
     * corresponding map of variable-value pairs using the
     * provided set of keys. The association between
     * keys and values is determined from their order of
     * occurrence in their respective collection. In other
     * words, the first variable returned by the set's iterator
     * is associated with the first parsed value, and so on.
     *
     * @param allkeys the set of keys with which 
     * to associate the parsed values.
     * 
     * @param allValues a string representation of a list
     * of values.
     * 
     * @return keys mapped to their respective values.
     */
    public Map<String, Double> 
    parseAllValues(LinkedHashSet<? extends String> allkeys,
    			   String allValues) {

    	// Creating storage for values
        List<Double> values = parseAllValues(allValues);

        // Getting sizes
        int varSize = allkeys.size();
        int valSize = values.size();

        // Comparing
        if (varSize != valSize) {

            String s = "There are " + varSize + " keys in '" +
                       formatKeys(allkeys) + "' but there are " +
                       valSize + " values in '" + formatValues(values) + "'.";
            throw new FormatException(s);

        }

        // Creating storage for map
        Map<String, Double> map = new LinkedHashMap<String, Double>(varSize + 1);

        // Getting iterators over keys and values
        Iterator<? extends String> i = allkeys.iterator();
        Iterator<Double> j = values.iterator();

        // While another variable exists
        while (i.hasNext()) {
            map.put(i.next(), j.next());
        }

        return map;
    }

    /**
     * Parses the provided set of keys and list of values
     * and returns a corresponding map of variable-value pairs.
     * The association between keys and values is determined
     * from their order of occurrence in their respective collection.
     * In other words, the first parsed variable is
     * associated with the first parsed value, and so on. 
     * 
     * @param allkeys a string representation of a set
     * of keys.
     * 
     * @param allValues a string representation of a list of
     * values.
     * 
     * @return keys mapped to their respective values.
     */
    public Map<String, Double> parseAllPairs(String allkeys, String allValues) {

    	// Parsing all keys
    	LinkedHashSet<String> keys = parseAllKeys(allkeys);

    	// Parsing all values
    	return parseAllValues(keys, allValues);
    }

}
