/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.utilities.Util;
import sbrt.shell.text.*;

/**
 * This class is a concrete implemenation of {@linkplain FbaOptHeaderFormatter}.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public class FbaOptHeaderFormatterV1 implements FbaOptHeaderFormatter {

	/**
	 * The header used to indicate the value of the objective function.
	 */
	public static final String OBJ_VALUE = 
		FbaOptHeaderParserV1.OBJ_VALUE;

	private final Fluxome fluxome;
	private final RxnNameSetFormatV1 rxnSetFormat;
	private final PipeSetFormat<String> setFormat;
	
	/**
	 * Constructs a new FBA optimization header formatter.
	 * 
	 * @param fluxome
	 *            the fluxome used to verify reaction names.
	 */
	public FbaOptHeaderFormatterV1(Fluxome fluxome) {

		// Checking and saving input
		Util.nullCheck(fluxome);
		this.fluxome = fluxome;

		// Creating internal formats
		rxnSetFormat = new RxnNameSetFormatV1(fluxome);
		setFormat = new PipeSetFormat<String>(new StringFormat());
	}

	/**
	 * Returns the fluxome used to verify reaction names.
	 * 
	 * @return the fluxome used to verify reaction names.
	 */
	public Fluxome getFluxome() {
		return fluxome;
	}

	/**
	 * Returns the format used for reaction name sets.
	 * 
	 * @return the format used for reaction name sets.
	 */
	public RxnNameSetFormatV1 getRxnNameSetFormat() {
		return rxnSetFormat;
	}

	/**
	 * Returns the format used for sets of strings.
	 * 
	 * @return the format used for sets of strings.
	 */
	public PipeSetFormat<String> getPipeSetFormat() {
		return setFormat;
	}

	/**
	 * Returns the header used to indicate the value of the objective function.
	 * 
	 * @return the header used to indicate the value of the objective function.
	 */
	public String getObjValueHeader() {
		return OBJ_VALUE;
	}

	/**
	 * Indicates if the provided string is the header used to indicate the value
	 * of the objective function.
	 * 
	 * @param s
	 *            the string to compare.
	 * 
	 * @return {@code true} if the provided string is equal to the string
	 *         returned by {@linkplain #getObjValueHeader()}; {@code false}
	 *         otherwise.
	 */
	public boolean isObjValueHeader(String s) {
		return OBJ_VALUE.equals(s);
	}

	/**
	 * Returns a formatted string representation of the
	 * provided header.
	 * 
	 * @param header the header to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided header.
	 */
	public String format(String header) {
		
		if (OBJ_VALUE.equals(header)) {
			return OBJ_VALUE;
		}
		
		if (fluxome.isRxnName(header)) {
			return header;
		}

		// Creating error message
		String s = "The header '" + header + "' is not " +
		OBJ_VALUE + " or a valid reaction name.";
		
		// Throwing exception
		throw new IllegalArgumentException(s);
	}
}
