/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;

/**
 * This class is a concrete implemenation of {@linkplain FbaOptHeaderParser}.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 *         Wagner lab.
 * 
 */
public final class FbaOptHeaderParserV1 implements FbaOptHeaderParser {

	/**
	 * The header used to indicate the value of the objective function.
	 */
	public static final String OBJ_VALUE = "Objective_Value";

	private final Fluxome fluxome;
	private final RxnNameSetFormatV1 rxnSetFormat;
	private final PipeSetFormat<String> setFormat;
	
	/**
	 * Constructs a new FBA optimization header parser.
	 * 
	 * @param fluxome
	 *            the fluxome used to verify reaction names.
	 */
	public FbaOptHeaderParserV1(Fluxome fluxome) {

		// Checking and saving input
		Util.nullCheck(fluxome);
		this.fluxome = fluxome;

		// Creating internal formats
		rxnSetFormat = new RxnNameSetFormatV1(fluxome);
		setFormat = new PipeSetFormat<String>(new StringFormat());
	}

	/**
	 * Returns the fluxome used to verify reaction names.
	 * 
	 * @return the fluxome used to verify reaction names.
	 */
	public Fluxome getFluxome() {
		return fluxome;
	}

	/**
	 * Returns the format used for reaction name sets.
	 * 
	 * @return the format used for reaction name sets.
	 */
	public RxnNameSetFormatV1 getRxnNameSetFormat() {
		return rxnSetFormat;
	}

	/**
	 * Returns the format used for sets of strings.
	 * 
	 * @return the format used for sets of strings.
	 */
	public PipeSetFormat<String> getPipeSetFormat() {
		return setFormat;
	}

	/**
	 * Returns the header used to indicate the value of the objective function.
	 * 
	 * @return the header used to indicate the value of the objective function.
	 */
	public String getObjValueHeader() {
		return OBJ_VALUE;
	}

	/**
	 * Indicates if the provided string is the header used to indicate the value
	 * of the objective function.
	 * 
	 * @param s
	 *            the string to compare.
	 * 
	 * @return {@code true} if the provided string is equal to the string
	 *         returned by {@linkplain #getObjValueHeader()}; {@code false}
	 *         otherwise.
	 */
	public boolean isObjValueHeader(String s) {
		return OBJ_VALUE.equals(s);
	}

	/**
	 * Parses the provided string and returns its corresponding set of FBA
	 * optimization headers as a {@code LinkedHashSet}.
	 * 
	 * @param headerSet
	 *            the string to be parsed.
	 * 
	 * @return the set of FBA optimization headers corresponding to the provided
	 *         string.
	 */
	@SuppressWarnings("unchecked")
	public LinkedHashSet<String> parse(String headerSet) {

		// Parsing into a set of strings
		LinkedHashSet<String> headerStrings = setFormat.parse(headerSet);

		// Determining an appropriate size
		int size = headerStrings.size() + 1;

		// Creating a set
		LinkedHashSet<String> set = new LinkedHashSet<String>(size, 1.0F);

		// Iterating thru each header string
		for (String header : headerStrings) {

			// Parsing the header string
			Object headerObj = parseHeader(header);

			// If the header is ALL_RXNS
			if (headerObj instanceof Set) {
				set.addAll((Set<String>) headerObj);
			}

			// If it was OBJ_VALUE or a reaction name
			else {
				set.add(header);
			}
		}

		// Returning the parsed set of headers
		return set;
	}

	/**
	 * Parses the provided header string.
	 */
	private Object parseHeader(String header) {

		// If it's OBJ_VALUE
		if (header.trim().equals(OBJ_VALUE)) {
			return OBJ_VALUE;
		}

		// Parsing it with the reaction header format
		try {
			return rxnSetFormat.parse(header);
		} catch (FormatException e) {
			String s = "'" + header + "' is not parsable as '" + OBJ_VALUE
					+ "', '" + RxnNameSetFormatV1.ALL_RXNS
					+ "', or as a reaction name in the provided fluxome.";
			throw new FormatException(s, e);
		}

	}

}
