/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import java.util.*;

import sbrt.kernel.fba.Fluxome;
import sbrt.kernel.math.LinearComb;
import sbrt.shell.text.*;

/**
 * This class is a concrete implementation of {@linkplain FluxCapFormat}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public class FluxCapFormatV1 
	extends AbstractMapFormat<String, Set<LinearComb<String>>> 
	implements FluxCapFormat {

	private final Fluxome fluxome;
	
	private final ColonMapEntryFormatV1<String, String> entryFormat;
	private final PipeSetFormat<String> setFormat;
	private final PipeListFormat<Set<LinearComb<String>>> listFormat;
	
	/**
	 * Constructs a new flux cap file line format using the 
	 * provided fluxome.
	 * 
	 * @param fluxome the fluxome with which to verify reaction
	 * names.
	 * 
	 * @throws NullPointerException if the provided argument 
	 * is {@code null}.
	 */
	public FluxCapFormatV1(Fluxome fluxome) {

		this.fluxome = fluxome;
		
		// Creating basic entry format
		entryFormat = ColonMapEntryFormatV1.getBasic();
		
		// Creating reaction name expression parser
		RxnNameExprFormatV1 expFormat = new RxnNameExprFormatV1(fluxome);

		// Getting its verifier
		RxnNameVerifier rxnFormat = expFormat.getRxnNameChecker();
		
		// Creating reaction name set parser
		setFormat = new PipeSetFormat<String>(rxnFormat);
		
		// Creating flux cap set parser
		PipeSetFormat<LinearComb<String>> capFormat =
			new PipeSetFormat<LinearComb<String>>(expFormat);
		
		// Wrapping the flux cap set parser as a simple format
		SimpleSetFormat<LinearComb<String>> simpleSetFormat = 
			new SimpleSetFormat<LinearComb<String>>(capFormat);
		
		// Creating value list parser
		listFormat = new PipeListFormat<Set<LinearComb<String>>>(simpleSetFormat);
	}
	
	/**
	 * Returns the fluxome used to verify reaction names.
	 * @return the fluxome used to verify reaction names.
	 */
    public Fluxome getFluxome() {
    	return fluxome;
    }
	
    /**
     * Returns the format used for variable-value pairs.
     * @return the format used for variable-value pairs.
     */
    public ColonMapEntryFormatV1<String, String> getMapEntryFormat() {
        return entryFormat;
    }

    /**
     * Returns the format used for sets of variables.
     * @return the format used for sets of variables.
     */
    public PipeSetFormat<String> getSetFormat() {
        return setFormat;
    }

    /**
     * Returns the format used for lists of values.
     * @return the format used for lists of values.
     */
    public PipeListFormat<Set<LinearComb<String>>> getListFormat() {
        return listFormat;
    }


}
