/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import sbrt.kernel.fba.*;
import sbrt.kernel.math.*;
import sbrt.shell.text.*;

/**
 * This class is a concrete implemenation of 
 * {@linkplain FluxomeSolutionFileLineFormat}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class FluxomeSolFileLineFormatV1 
	extends AbstractMapFormat<String, LinearComb<String>> 
		implements FluxomeSolutionFileLineFormat {

	private final Fluxome fluxome;
	private final EqualsMapEntryFormatV1<String, String> entryFormat;
	private final PipeSetFormat<String> setFormat;
	private final PipeListFormat<LinearComb<String>> listFormat;
	
	/**
	 * Constructs a new fluxome solution file line format.
	 * 
	 * @param fluxome the fluxome to use to verify
	 * reaction names.
	 */
    public FluxomeSolFileLineFormatV1(Fluxome fluxome) {

    	RxnNameExprFormatV1 e = new RxnNameExprFormatV1(fluxome);
    	RxnNameVerifier r = e.getRxnNameChecker();
    	
    	this.fluxome = fluxome;
    	entryFormat = EqualsMapEntryFormatV1.getBasic();
    	setFormat = new PipeSetFormat<String>(r);
    	listFormat = new PipeListFormat<LinearComb<String>>(e);
    }

	/**
	 * Constructs a new fluxome solution file line format.
	 * 
	 * @param fluxome the fluxome to use to verify reaction names.
     * @param coeffCutoff the smallest value that the absolute value of a coefficient
     * can attain.
     * 
     *  @see LinearComb#getZeroCutoff()
	 */
    public FluxomeSolFileLineFormatV1(Fluxome fluxome, double coeffCutoff) {

    	RxnNameExprFormatV1 e = new RxnNameExprFormatV1(fluxome, coeffCutoff);
    	RxnNameVerifier r = e.getRxnNameChecker();
    	
    	this.fluxome = fluxome;
    	entryFormat = EqualsMapEntryFormatV1.getBasic();
    	setFormat = new PipeSetFormat<String>(r);
    	listFormat = new PipeListFormat<LinearComb<String>>(e);
    }
    
    /**
     * Returns the fluxome used to verify reaction names.
     * @return the fluxome used to verify reaction names.
     */
    public Fluxome getFluxome() {
        return fluxome;
    }

    /**
     * Returns the format used for variable-value pairs.
     * @return the format used for variable-value pairs.
     */
    public EqualsMapEntryFormatV1<String, String> getMapEntryFormat() {
        return entryFormat;
    }

    /**
     * Returns the format used for sets of variables.
     * @return the format used for sets of variables.
     */
    public PipeSetFormat<String> getSetFormat() {
        return setFormat;
    }

    /**
     * Returns the format used for lists of values.
     * @return the format used for lists of values.
     */
    public PipeListFormat<LinearComb<String>> getListFormat() {
        return listFormat;
    }
}
