/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import java.util.Collections;
import java.util.Set;
import java.util.regex.Pattern;

import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.IrrevRxn;
import sbrt.kernel.fba.SinkRxn;
import sbrt.kernel.fba.SourceRxn;
import sbrt.kernel.math.LinearComb;
import sbrt.kernel.utilities.Util;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.SimpleFormat;
import sbrt.shell.text.chem.ChemSpeciesFormatV1;
import sbrt.shell.text.math.LinearCombFormatV2;

/**
 * This class is used to format irreversible reactions for use in flux balance analysis.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class IrrevRxnFormatV2 implements SimpleFormat<ChemRxn> {
    
    /**
     * The string used to denote the <i>surroundings</i>. This is used as a 
     * pseudo-species when representing exchange reactions.
     */
    public static final String SURR = "Surroundings";
    
    /**
     * The arrow used to signify an irreversible reaction.
     */
    public static final String ARROW = "-->";
    
    private static final Pattern ARROW_PATTERN = Pattern.compile("\\s*-->\\s*");
    
    private final ChemSpeciesFormatV1 speciesFormat;
    private final LinearCombFormatV2<String> linearFormat;
    
    /**
     * Constructs a new irreversible reaction format.
     */
    public IrrevRxnFormatV2() {
        speciesFormat = new ChemSpeciesFormatV1();
        linearFormat = new LinearCombFormatV2<String>(speciesFormat);
    }
    
    /**
     * Returns the format used for chemical species.
     * @return the format used for chemical species.
     */
    public ChemSpeciesFormatV1 getChemSpeciesFormat() {
        return speciesFormat;
    }
    
    /**
     * Returns the format used for stoichiometric expressions.
     * @return the format used for stoichiometric expressions.
     */
    public LinearCombFormatV2<String> getLinearCombFormat() {
        return linearFormat;
    }

    /**
     * Returns a formatted string representation of the
     * provided irreversible chemical reaction.
     * 
     * @param rxn the reaction to be formatted.
     * 
     * @return a formatted string representation of the
     * provided irreversible chemical reaction.
     * 
     * @throws IllegalArgumentException if {@code rxn} is not an 
     * {@linkplain IrrevRxn}, {@linkplain SourceRxn}, or {@linkplain SinkRxn}.
     */
    public String format(ChemRxn rxn) {

        if ((rxn instanceof IrrevRxn)) {
            return format((IrrevRxn) rxn);
        }

        if ((rxn instanceof SourceRxn)) {
            return format((SourceRxn) rxn);
        }

        if ((rxn instanceof SinkRxn)) {
            return format((SinkRxn) rxn);
        }
        
        String s = rxn + " is not an understood type: " + 
                   rxn.getClass().getCanonicalName();
        
        throw new IllegalArgumentException(s);
    }
    
	/**
	 * Returns a formatted string representation of the
	 * provided irreversible chemical reaction.
	 * 
	 * @param rxn the reaction to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided irreversible chemical reaction.
	 */
    public String format(IrrevRxn rxn) {

    	// Creating storage for formatted string
    	StringBuilder sb = new StringBuilder(100);

    	// Getting the pieces
        LinearComb<String> reactants = rxn.getReactants();
        LinearComb<String> products = rxn.getProducts();

        if (reactants.hasVariable(SURR) || products.hasVariable(SURR)) {
            String s = "Invalid species '" + SURR + "' in: " + rxn;
            throw new IllegalArgumentException(s);
        }
        
        // Appending the reaction
        sb.append(linearFormat.format(reactants));
        sb.append(" --> ");
        sb.append(linearFormat.format(products));

        // Returning the formatted string
        return sb.toString();
    }

    /**
     * Returns a formatted string representation of the
     * provided source.
     * 
     * @param source the reaction to be formatted.
     * 
     * @return a formatted string representation of the
     * provided source.
     */
    public String format(SourceRxn source) {

        // Creating storage for formatted string
        StringBuilder s = new StringBuilder(100);

        // Getting the pieces
        LinearComb<String> products = source.getProducts();

        // Appending the reaction
        s.append(SURR);
        s.append(" --> ");
        s.append(linearFormat.format(products));

        // Returning the formatted string
        return s.toString();
    }

    /**
     * Returns a formatted string representation of the
     * provided sink.
     * 
     * @param sink the reaction to be formatted.
     * 
     * @return a formatted string representation of the
     * provided sink.
     */
    public String format(SinkRxn sink) {

        // Creating storage for formatted string
        StringBuilder s = new StringBuilder(100);

        // Getting the pieces
        LinearComb<String> reactants = sink.getReactants();

        // Appending the reaction
        s.append(linearFormat.format(reactants));
        s.append(" --> ");
        s.append(SURR);

        // Returning the formatted string
        return s.toString();
    }

    /**
     * Parses the provided string and returns its 
     * corresponding chemical reaction. 
     * 
     * @param reaction the string to be parsed.
     * 
     * @return the chemical reaction corresponding to the 
     * provided string. The type of this reaction will be either
     * {@linkplain IrrevRxn}, {@linkplain SourceRxn}, or 
     * {@linkplain SinkRxn}.
     */
    public ChemRxn parse(String reaction) {

        reaction = reaction.trim();
        
        // Ensuring the provided string contains an arrow
        if (!reaction.contains(ARROW)) {
            String s = "The reaction '" + reaction + 
                       "' does not contain the arrow '" + ARROW + "'.";
            throw new FormatException(s);
        }
        
        // Splitting on the arrow
        String[] parts = ARROW_PATTERN.split(reaction, 0);
        
        // If more than one arrow is present
        if (parts.length > 2) {
            String s = "The reaction '" + reaction + 
                       "' contains more than one arrow.";
            throw new FormatException(s);
        }
        
        // Parsing as a source exchange reaction 
        if (parts[0].equals(SURR)) {
            
            String species = speciesFormat.parse(parts[1]);
            return new SourceRxn(species);
        }
        
        // Parsing as a sink exchange reaction 
        if (parts[1].equals(SURR)) {

            String species = speciesFormat.parse(parts[0]);
            return new SinkRxn(species);
        }
        
        // Parsing as a normal irreversible reaction
        LinearComb<String> reactants = linearFormat.parse(parts[0]);
        LinearComb<String> products = linearFormat.parse(parts[1]);
        
        checkLinComb(parts[0], reactants);
        checkLinComb(parts[1], products);
        
        checkBothSides(reaction, reactants, products);
        
        return new IrrevRxn(reactants, products);
    }


    private static void checkLinComb(String expression, LinearComb<String> expr) {

        for (LinearComb.Term<String> t : expr) {

            double coeff = t.getCoefficient();

            if (coeff <= 0 || Double.isInfinite(coeff) || Double.isNaN(coeff)) {
                String s = "The coefficient '" + coeff + "' is invalid.";
                throw new FormatException(s);
            }
            
            if (t.getVariable().equals(SURR)) {
                String s = "The pseudo-species '" + SURR +
                           "' does not appear by itself in the stoichiometric " +
                           "expression '" + expression + "'.";
                throw new FormatException(s);
            }

        }

        if (!Util.isZero(expr.getConstant())) {
            String s = "The stoichiometric expression '" + expression +
                       "' contains a coefficient with no corresponding " +
                       "chemical species.";
            throw new FormatException(s);
        }
        
    }

    private static void checkBothSides(String reaction,
                                       LinearComb<String> reactants,
                                       LinearComb<String> products) {

        Set<String> rVars = reactants.getVariables();
        Set<String> pVars = products.getVariables();

        if (rVars.isEmpty()) {
            String s = "The reaction '" + reaction + 
                       "' does not contain reactants.";
            throw new FormatException(s);
        }

        if (pVars.isEmpty()) {
            String s = "The reaction '" + reaction + 
                       "' does not contain products.";
            throw new FormatException(s);
        }
        
        if (!Collections.disjoint(rVars, pVars)) {
            String s = "The reactants and products in '" + reaction +
                       "' have at least one chemical species in common.";
            throw new FormatException(s);
        }

    }

 }
