/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import sbrt.kernel.fba.ChemRxn;
import sbrt.shell.text.AbstractMapFormat;
import sbrt.shell.text.EqualsMapEntryFormatV1;
import sbrt.shell.text.PipeListFormat;
import sbrt.shell.text.PipeSetFormat;
import sbrt.shell.text.chem.RxnNameFormatV1;

/**
 * This class is used to format irreversible reaction fluxome files.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class MixedFluxomeFormatV1 extends AbstractMapFormat<String, ChemRxn> {


	private final EqualsMapEntryFormatV1<String, String> entryFormat;
	private final PipeSetFormat<String> setFormat;
	private final PipeListFormat<ChemRxn> listFormat;

    private final RxnNameFormatV1 rxnNameFormat;
    private final RxnFormatV1 rxnFormat;
    
	/**
	 * Constructs a new irreversible reaction fluxome format.
	 */
    public MixedFluxomeFormatV1() {

        // Key format
        rxnNameFormat = new RxnNameFormatV1();
        
        // Value format
        rxnFormat = new RxnFormatV1();

        // Creating formats
        entryFormat = EqualsMapEntryFormatV1.getBasic();
        setFormat = new PipeSetFormat<String>(rxnNameFormat);
        listFormat = new PipeListFormat<ChemRxn>(rxnFormat);
    }
    
    /**
     * Returns the format used for reaction names.
     * @return the format used for reaction names.
     */
    public RxnNameFormatV1 getRxnNameFormat() {
        return rxnNameFormat;
    }

    /**
     * Returns the format used for reactions.
     * @return the format used for reactions.
     */
    public RxnFormatV1 getRxnFormat() {
        return rxnFormat;
    }
    
    /**
     * Returns the format used for reaction name-reaction pairs.
     * @return the format used for reaction name-reaction pairs.
     */
    public EqualsMapEntryFormatV1<String, String> getMapEntryFormat() {
        return entryFormat;
    }

    /**
     * Returns the format used for sets of reaction names.
     * @return the format used for sets of reaction names.
     */
    public PipeSetFormat<String> getSetFormat() {
        return setFormat;
    }

    /**
     * Returns the format used for lists of reactions.
     * @return the format used for lists of reactions.
     */
    public PipeListFormat<ChemRxn> getListFormat() {
        return listFormat;
    }

}
