/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import sbrt.kernel.fba.ChemRxn;
import sbrt.kernel.fba.IrrevRxn;
import sbrt.kernel.fba.RevExchRxn;
import sbrt.kernel.fba.RevRxn;
import sbrt.kernel.fba.SinkRxn;
import sbrt.kernel.fba.SourceRxn;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.SimpleFormat;
import sbrt.shell.text.chem.ChemSpeciesFormatV1;

/**
 * This class is used to format reactions for use in flux balance analysis.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class RxnFormatV1 implements SimpleFormat<ChemRxn> {
    
    private final IrrevRxnFormatV1 irrevRxnFormat;
    private final RevRxnFormatV1 revRxnFormat;
    
    /**
     * Constructs a new irreversible reaction format.
     */
    public RxnFormatV1() {
    	irrevRxnFormat = new IrrevRxnFormatV1();
    	revRxnFormat = new RevRxnFormatV1();
    }
    
    /**
     * Returns the format used for chemical species.
     * @return the format used for chemical species.
     */
    public ChemSpeciesFormatV1 getChemSpeciesFormat() {
        return revRxnFormat.getChemSpeciesFormat();
    }

    /**
     * Returns the format used for irreversible reactions.
     * @return the format used for irreversible reactions.
     */
    public IrrevRxnFormatV1 getIrrevRxnFormat() {
    	return irrevRxnFormat;
    }

    /**
     * Returns the format used for reversible reactions.
     * @return the format used for reversible reactions.
     */
    public RevRxnFormatV1 getRevRxnFormat() {
    	return revRxnFormat;
    }
    
    /**
     * Returns a formatted string representation of the
     * provided irreversible chemical reaction.
     * 
     * @param rxn the reaction to be formatted.
     * 
     * @return a formatted string representation of the
     * provided irreversible chemical reaction.
     * 
     * @throws IllegalArgumentException if {@code rxn} is not an 
     * {@linkplain IrrevRxn}, {@linkplain RevRxn}, {@linkplain SourceRxn}, 
     * {@linkplain SinkRxn}, or {@linkplain RevExchRxn}.
     */
    public String format(ChemRxn rxn) {

    	if (rxn instanceof IrrevRxn) {
    		return format((IrrevRxn) rxn);
    	}
    	
    	if (rxn instanceof SourceRxn) {
    		return format((SourceRxn) rxn);
    	}
    	
    	if (rxn instanceof SinkRxn) {
    		return format((SinkRxn) rxn);
    	}
    	
        if (rxn instanceof RevRxn) {
            return format((RevRxn) rxn);
        }

        if (rxn instanceof RevExchRxn) {
            return format((RevExchRxn) rxn);
        }
        
        String s = rxn + " is not an understood type: " + 
                   rxn.getClass().getCanonicalName();
        
        throw new IllegalArgumentException(s);
    }

    /**
     * Returns a formatted string representation of the
     * provided irreversible reaction.
     * 
     * @param rxn the reaction to be formatted.
     * 
     * @return a formatted string representation of the
     * provided irreversible reaction.
     */
    public String format(IrrevRxn rxn) {
    	return irrevRxnFormat.format(rxn);
    }

    /**
     * Returns a formatted string representation of the
     * provided source reaction.
     * 
     * @param rxn the reaction to be formatted.
     * 
     * @return a formatted string representation of the
     * provided source reaction.
     */
    public String format(SourceRxn rxn) {
    	return irrevRxnFormat.format(rxn);
    }

    /**
     * Returns a formatted string representation of the
     * provided sink reaction.
     * 
     * @param rxn the reaction to be formatted.
     * 
     * @return a formatted string representation of the
     * provided sink reaction.
     */
    public String format(SinkRxn rxn) {
    	return irrevRxnFormat.format(rxn);
    }

    /**
     * Returns a formatted string representation of the
     * provided reversible reaction.
     * 
     * @param rxn the reaction to be formatted.
     * 
     * @return a formatted string representation of the
     * provided reversible reaction.
     */
    public String format(RevRxn rxn) {
    	return revRxnFormat.format(rxn);
    }

    /**
     * Returns a formatted string representation of the
     * provided reversible exchange reaction.
     * 
     * @param rxn the reaction to be formatted.
     * 
     * @return a formatted string representation of the
     * provided reversible exchange reaction.
     */
    public String format(RevExchRxn rxn) {
    	return revRxnFormat.format(rxn);
    }
    


    /**
     * Parses the provided string and returns its 
     * corresponding chemical reaction. 
     * 
     * @param reaction the string to be parsed.
     * 
     * @return the chemical reaction corresponding to the 
     * provided string. The type of this reaction will be either
     * {@linkplain IrrevRxn}, {@linkplain SourceRxn}, or 
     * {@linkplain SinkRxn}.
     */
    public ChemRxn parse(String reaction) {

    	if (reaction.contains(IrrevRxnFormatV1.ARROW)) {
    		return irrevRxnFormat.parse(reaction);
    	}
    	
    	if (reaction.contains(RevRxnFormatV1.ARROW)) {
    		return revRxnFormat.parse(reaction);
    	}
    	
    	String s = "The reaction '" + reaction + 
    			   "' does not contain a recognizable arrow.";
    	throw new FormatException(s);
    }



 }
