/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import sbrt.kernel.fba.*;
import sbrt.kernel.math.*;
import sbrt.shell.text.math.*;

/**
 * This class is used to format linear combinations of reaction names.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class RxnNameExprFormatV1 
	implements RxnNameExprFormat<LinearComb<String>, String> {

    private final RxnNameVerifier rxnNameChecker;
    private final LinearCombFormatV1<String> linearFormat;

    /**
     * Constructs a new reaction name expression format
     * using the provided fluxome.
     * 
     * @param fluxome
     */
    public RxnNameExprFormatV1(Fluxome fluxome) {
        rxnNameChecker = new RxnNameVerifier(fluxome);
        linearFormat = new LinearCombFormatV1<String>(rxnNameChecker);
    }

    /**
     * Constructs a new reaction name expression format
     * using the provided fluxome.
     * 
     * @param fluxome
     * @param coeffCutoff the smallest value that the absolute value of a coefficient
     * can attain.
     * 
     *  @see LinearComb#getZeroCutoff()
     */
    public RxnNameExprFormatV1(Fluxome fluxome, double coeffCutoff) {
        rxnNameChecker = new RxnNameVerifier(fluxome);
        linearFormat = new LinearCombFormatV1<String>(rxnNameChecker, coeffCutoff);
    }

    /**
     * Returns the fluxome used to verify reaction names.
     * @return the fluxome used to verify reaction names.
     */
    public Fluxome getFluxome() {
        return rxnNameChecker.getFluxome();
    }

    /**
     * Returns the reaction name verifier.
     * @return the reaction name verifier.
     */
    public RxnNameVerifier getRxnNameChecker() {
        return rxnNameChecker;
    }

    /**
     * Returns the format used for linear combination.
     * @return the format used for linear combination.
     */
    public LinearCombFormatV1<String> getLinerCombFormat() {
        return linearFormat;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided linear combination of reaction names.
	 * 
	 * @param rxnNameExpr the linear combination to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided linear combination of reaction names.
	 */
    public String format(LinearComb<String> rxnNameExpr) {
        return linearFormat.format(rxnNameExpr);
    }

    /**
     * Parses the provided string and returns its 
     * corresponding linear combination of reaction names.
     * 
     * @param rxnNameExpr the string to be parsed.
     * 
     * @return the linear combination of reaction names
     * corresponding to the provided string. 
     */
    public LinearComb<String> parse(String rxnNameExpr) {
        return linearFormat.parse(rxnNameExpr);
    }


}
