/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import sbrt.kernel.fba.*;
import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;

/**
 * This class is a concrete implemenation of {@linkplain RxnNameOrExprFormat}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class RxnNameOrExprFormatV1 implements RxnNameOrExprFormat {

	// The reaction name verifier
    private final RxnNameVerifier rxnNameChecker;
    
    // The expression format
    private final RxnNameExprFormatV1 exprFormat;
    
    // The fluxome used for name verification
    private final Fluxome fluxome;

    /**
     * Constructs a new reaction name and expression format.
     * 
     * @param fluxome the fluxome used to verify reaction names.
     */
    public RxnNameOrExprFormatV1(Fluxome fluxome) {
    	
    	// Checking and savign
        Util.nullCheck(fluxome);
        this.fluxome = fluxome;

        // Creating internal formats
        rxnNameChecker = new RxnNameVerifier(fluxome);
        exprFormat = new RxnNameExprFormatV1(fluxome);
    }

	/**
	 * Returns the fluxome used to verify reaction names.
	 * @return the fluxome used to verify reaction names.
	 */
    public Fluxome getFluxome() {
        return fluxome;
    }

    /**
     * Returns the reaction name verifier.
     * @return the reaction name verifier.
     */
    public RxnNameVerifier getRxnNameChecker() {
        return rxnNameChecker;
    }

    /**
     * Returns the reaction name expression format.
     * @return the reaction name expression format.
     */
    public RxnNameExprFormatV1 getLinearCombFormat() {
        return exprFormat;
    }

    /**
     * Returns the string representation of the provided
     * object.
     * 
     * @param rxnNameOrExpression a reaction name or
     * {@linkplain LinearComb} of reaction names.
     * 
     * @return the string representation of the provided
     * object.
     * 
     */
    @SuppressWarnings("unchecked")
	public String format(Object rxnNameOrExpression) {

        String result = null;

        if (rxnNameOrExpression instanceof String) {
        	
        	String rxnName = (String) rxnNameOrExpression;
        	
            result = rxnNameChecker.format(rxnName);
        }

        else if (rxnNameOrExpression instanceof LinearComb) {
            LinearComb<String> expr = 
            	(LinearComb<String>) rxnNameOrExpression;
            result = exprFormat.format(expr);
        }

        else {
            throw new ClassCastException(rxnNameOrExpression.toString());
        }

        return result;
    }

    /**
     * Parses the provided string and returns its corresponding
     * object.
     * 
     * @param rxnNameOrExpression a string representation of a
     * reaction name or expression.
     * 
     * @return the reaction name or expression contained in the
     * provided string.
     */
    public Object parse(String rxnNameOrExpression) {

        String trimmed = rxnNameOrExpression.trim();

        if (trimmed.length() == 0) {
            String s = "'" + rxnNameOrExpression + "' is neither a reaction " +
                       "name nor a linear combination of reaction names.";
            throw new FormatException(s);
        }

        Object result = null;

        try {
            result = rxnNameChecker.parse(rxnNameOrExpression);
        } catch (FormatException e1) {

            try {
                result = exprFormat.parse(trimmed);
            } catch (FormatException e2) {
                String s = "'" + trimmed + "' is neither a reaction " +
                           "name nor a linear combination of reaction names.";
                throw new FormatException(s);
            }

        }

        return result;
    }

}
