/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.fba;

import java.util.*;

import sbrt.kernel.fba.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;

/**
 * This class is a concrete implemenation of {@linkplain RxnNameSetFormat}.
 * 
 * @author This interface was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class RxnNameSetFormatV1 implements RxnNameSetFormat {

	/**
	 * The string used to indicate every reaction in a fluxome.
	 */
	public static final String ALL_RXNS = "All_Reactions";

    private final RxnNameVerifier rxnNameChecker;
    private final PipeSetFormat<String> setFormat;

    /**
     * Constructs a new reaction name collection format
     * using the provided fluxome.
     * 
     * @param fluxome the fluxome used to verify reaction
     * names.
     */
    public RxnNameSetFormatV1(Fluxome fluxome) {

    	// Checking input
    	Util.nullCheck(fluxome);
        
    	// Creating internal formats
    	rxnNameChecker = new RxnNameVerifier(fluxome);
        setFormat = new PipeSetFormat<String>(new StringFormat());
    }

    /**
     * Returns the fluxome used to verify reaction names.
     * @return the fluxome used to verify reaction names.
     */
    public Fluxome getFluxome() {
        return rxnNameChecker.getFluxome();
    }

    /**
     * Returns the format used for sets of reaction names.
     * @return the format used for sets of reaction names.
     */
    public PipeSetFormat<String> getSetFormat() {
        return setFormat;
    }

    /**
     * Returns the reaction name verifier.
     * @return the reaction name verifier.
     */
    public RxnNameVerifier getRxnNameChecker() {
        return rxnNameChecker;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided set of reaction names.

	 * @param reactionNames the set of reaction names to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided set of reaction names.
	 */
    public String format(Set<String> reactionNames) {
        return setFormat.format(reactionNames);
    }

    /**
     * Parses the provided string and returns its 
     * corresponding collection of reaction names as a 
     * {@code LinkedHashSet}. If the string {@code ALL_RXNS}
     * is present the set will contain the name of
     * every reaction in this format's fluxome. 
     * 
     * @param set the string to be parsed.
     * 
     * @return the set of reaction names 
     * corresponding to the provided string. 
     */	
	@SuppressWarnings("unchecked")
	public LinkedHashSet<String> parse(String set) {

		// Parsing headers into a set of strings
		LinkedHashSet<String> strings = setFormat.parse(set);

		// Determining appropriate size
		int size = strings.size() + 1;

		// Creating an ordered storage space
		LinkedHashSet<String> rxnNames = 
			new LinkedHashSet<String>(size, 1.0F);

		// Iterating thru each header string
		for (String setString : strings) {

			// Parsing the header
			Object header = parseString(setString);

			// If the header is ALL_RXNS
			if (header instanceof Set) {
				rxnNames.addAll((Set<String>) header);
			}

			// If the header is a reaction name
			else {
				rxnNames.add((String) header);
			}
		}

		// Returning the set of parsed headers
		return rxnNames;
	}

	private Object parseString(String string) {

		// If the string is ALL_RXNS
		if (string.trim().equals(ALL_RXNS)) {

			// Getting all reaction names
			Set<String> names = getFluxome().getRxnNames();

			// Returning them sorted alphabetically
			return new TreeSet<String>(names);
		}

		// Treating the string as a reaction name 
		try {
			return rxnNameChecker.parse(string);
		} catch (FormatException e) {
			String s = "'" + string + "' is not parsable as '" + ALL_RXNS
					+ "' or as a reaction name in the provided fluxome.";
			throw new FormatException(s, e);
		}

	}

}
