/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.math;

import sbrt.kernel.math.*;
import sbrt.shell.text.*;

/**
 * This class is a concrete implementation of {@linkplain IntervalFormat}. The
 * string representation of intervals used by this class is:
 * {@code [LowerBound; UpperBound]}. The bounds must be finite.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 */
public final class BoundedIntervalFormatV1 implements IntervalFormat {

    private final IntervalFormatV1 intervalFormat;

    /**
     * Constructs a new interval format.
     */
    public BoundedIntervalFormatV1() {
        intervalFormat = new IntervalFormatV1();
    }
    
    /**
     * Returns the interval format used by this format.
     * @return the interval format used by this format.
     */
    public IntervalFormatV1 getIntervalFormat() {
        return intervalFormat;
    }

    /**
     * Returns the double format used for bounds.
     * 
     * @return the double format used for bounds.
     */
    public DoubleFormatV1 getDoubleFormat() {
        return intervalFormat.getDoubleFormat();
    }

    /**
     * Returns a formatted string representation of the provided interval.
     * 
     * @param interval the interval to be formatted.
     * 
     * @return a formatted string representation of the provided interval.
     * 
     * @throws IllegalArgumentException if the provided interval is unbounded.
     */
    public String format(Interval interval) {

        if (Double.isInfinite(interval.getLower())
                || Double.isInfinite(interval.getUpper())) {

            String s = "Unbounded interval: " + interval;
            throw new IllegalArgumentException(s);

        }

        // Returning the complete interval string
        return intervalFormat.format(interval);
    }

    /**
     * Parses the provided string and returns its corresponding interval.
     * 
     * @param interval the string to be parsed.
     * 
     * @return the interval corresponding to the provided string.
     * 
     * @throws FormatException if the interval cannot be parsed, or if a bound of
     * the parsed interval is infinite.
     */
    public Interval parse(String interval) {

        Interval i = intervalFormat.parse(interval);
        
        if (Double.isInfinite(i.getLower()) || Double.isInfinite(i.getUpper())) {
            String s = "The interval '" + interval + "' is unbounded.";
            throw new FormatException(s);
        }
        
        // Returning the parsed interval
        return i;
    }
}
