/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.math;

import sbrt.shell.text.*;

/**
 * This class is a concrete implementation of 
 * {@linkplain CoeffFormat}. The string representation of
 * a coefficient is: <i>numerical value</i>.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class CoeffFormatV2 implements CoeffFormat {

    private final DoubleFormatV1 doubleFormat;

    /**
     * Constructs a new coefficient format object.
     */
    public CoeffFormatV2() {
        doubleFormat = new DoubleFormatV1();
    }

    /**
     * Returns the format used for numerical values.
     * @return the format used for numerical values.
     */
    public DoubleFormatV1 getNumberFormat() {
        return doubleFormat;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided coefficient.
	 * 
	 * @param coefficient the coefficient to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided coefficient.
	 * 
	 * @throws NullPointerException if the provided
	 * coefficient is {@code null}.
	 */
    public String format(Double coefficient) {
        return format(coefficient.doubleValue());
    }

	/**
	 * Returns a formatted string representation of the
	 * provided coefficient.
	 * 
	 * @param coefficient the coefficient to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided coefficient.
	 * 
	 * @throws NullPointerException if the provided
	 * coefficient is {@code null}.
	 */
    public String format(double coefficient) {
        return doubleFormat.format(coefficient);
    }

    /**
     * Parses the provided string and returns its 
     * corresponding coefficient.
     * 
     * @param coefficient the string to be parsed.
     * 
     * @return the coefficient corresponding to the 
     * provided string.
     * 
	 * @throws NullPointerException if the provided
	 * coefficient is {@code null}.
     */
    public Double parse(String coefficient) {
    	
    	// Parsing the string
    	double x = parsePrimitive(coefficient);
    	
    	// Returning Double
    	return Double.valueOf(x);
    }

    /**
     * Parses the provided string and returns its 
     * corresponding coefficient.
     * 
     * @param coeff the string to be parsed.
     * 
     * @return the coefficient corresponding to the 
     * provided string.
     * 
	 * @throws NullPointerException if the provided
	 * coefficient is {@code null}.
     */
    public double parsePrimitive(String coeff) {

        // Parsing the number
        double x = doubleFormat.parsePrimitive(coeff);
        
        // Checking its value
        if (Double.isInfinite(x) || Double.isNaN(x)) {
            String s = "The coefficient '" + doubleFormat.format(x) + "' is invalid.";
            throw new FormatException(s);
        }
        
        // Returning the parsed number
        return x; 
    }


}
