/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.math;

import java.util.regex.*;

import sbrt.kernel.math.graph_theory.DirectedGraph.Edge;
import sbrt.kernel.utilities.Util;
import sbrt.shell.text.*;

/**
 * This class is a concrete implementation of {@linkplain DirEdgeFormat}.
 * The syntax used by this format is: {@code Source_Node -> Sink_Node}.
 * Whitespace surrounding the arrow is optional.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <N> the node type.
 *
 */
public final class DirEdgeFormatV1<N> implements DirEdgeFormat<N> {

	/**
	 * Returns a {@code DirEdgeFormatV1} that does not
	 * format or parse the nodes of the edges supplied to it. In other
	 * words, it returns a {@code DirEdgeFormatV1} with
	 * the most basic possible functionality. Empty strings are not 
	 * allowed however.
	 *  
	 * @return a {@code DirEdgeFormatV1} that does not
	 * format or parse the nodes of the edges supplied to it.
	 */
	public static DirEdgeFormatV1<String> getBasic() {
		return new DirEdgeFormatV1<String>(new NonEmptyStringFormat());
	}
	
    private static final Pattern arrow = Pattern.compile("\\s*->\\s*");

    private final SimpleFormat<N> nodeFormat;
    
    /**
     * Constructs a new directed edge format.
     * 
     * @param nodeFormat the format to use for nodes.
     */
    public DirEdgeFormatV1(SimpleFormat<N> nodeFormat) {
    	
    	Util.nullCheck(nodeFormat);
    	this.nodeFormat = nodeFormat;
    	
    }

	/**
	 * Returns a formatted string representation of the
	 * provided edge.
	 * 
	 * @param edge the edge to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided edge.
	 */
    public String format(Edge<N> edge) {

    	// Formatting the nodes
        String source = nodeFormat.format(edge.getSource());
        String sink = nodeFormat.format(edge.getSink());

        // Returning the complete edge string
        return source + " -> " + sink;
    }

    /**
     * Parses the provided string and returns its 
     * corresponding directed edge.
     * 
     * @param edge the string to be parsed.
     * 
     * @return the directed edge corresponding to the 
     * provided string. 
     */
    public Edge<N> parse(String edge) {

    	// Trimming whitespace
        String trimmed = edge.trim();

        // If the string is empty
        if (trimmed.length() == 0) {
            String s = "The edge '" + edge + "' is invalid.";
            throw new FormatException(s);
        }

        // Splitting on arrow
        String[] nodes = arrow.split(edge, -1);

        // If less than two strings are present
        if (nodes.length < 2) {
            String s = "The edge '" + trimmed + "' is improperly formatted.";
            throw new FormatException(s);
        }

        // If more than two bounding value are present
        if (nodes.length > 2) {
            String s = "The edge '" + trimmed +
                       "' contains too many arrows.";
            throw new FormatException(s);
        }

        // Ensuring a source node exists
        if (nodes[0].length() == 0) {
        	String s = "The edge '" + trimmed + 
        				"' does not appear to have a source node.";
        	throw new FormatException(s);
        }

        // Ensuring a sink node exists
        if (nodes[1].length() == 0) {
        	String s = "The edge '" + trimmed + 
        				"' does not appear to have a sink node.";
        	throw new FormatException(s);
        }

        // Parsing the nodes
        N source = nodeFormat.parse(nodes[0]);
        N sink = nodeFormat.parse(nodes[1]);
        
        // Returning the parsed interval
        return new Edge<N>(source, sink);
    }


}
