/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.math;

import java.util.List;

import sbrt.kernel.math.algebra.IndexPair;
import sbrt.kernel.utilities.Util;
import sbrt.shell.text.FormatException;
import sbrt.shell.text.ListFormat;
import sbrt.shell.text.PipeListFormat;
import sbrt.shell.text.SimpleFormat;

/**
 * This class is a concrete implementation of {@linkplain IndexPairFormat}.
 * The string representation of indices used by this class is:
 * {@code [RowIndex | ColumnIndex]}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <R> the row index type. 
 * @param <C> the column index type.
 *
 */
public final class IndexPairFormatV1<R, C> implements IndexPairFormat<R, C> {


    private final SimpleFormat<R> rowIndexFormat;
    private final SimpleFormat<C> colIndexFormat;

    /**
     * Constructs a new index pair format.
     * @param rowIndexFormat the row index formatter.
     * @param columnIndexFormat the column index formatter.
     */
    public IndexPairFormatV1(SimpleFormat<R> rowIndexFormat, 
                             SimpleFormat<C> columnIndexFormat) {
        
        Util.nullCheck(rowIndexFormat);
        Util.nullCheck(columnIndexFormat);
        
        this.rowIndexFormat = rowIndexFormat;
        this.colIndexFormat = columnIndexFormat;
    }

    /**
     * Returns the format used for row indices.
     * @return the format used for row indices.
     */
    public SimpleFormat<R> getRowIndexFormat() {
        return rowIndexFormat;
    }

    /**
     * Returns the format used for column indices.
     * @return the format used for column indices.
     */
    public SimpleFormat<C> getColumnIndexFormat() {
        return colIndexFormat;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided index pair.
	 * 
	 * @param indexPair the index pair to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided index pair.
	 */
    public String formatPair(IndexPair<? extends R, ? extends C> indexPair) {

    	// Formatting the indices
        String rowString = rowIndexFormat.format(indexPair.getRowIndex());
        String colString = colIndexFormat.format(indexPair.getColumnIndex());

        // Returning the complete index string
        return "[" + rowString + " | " + colString + "]";
    }

    /**
     * Returns a formatted string representation of the
     * provided index pair.
     * 
     * @param indexPair the index pair to be formatted.
     * 
     * @return a formatted string representation of the
     * provided index pair.
     */
    public String format(IndexPair<R, C> indexPair) {
        return formatPair(indexPair);
    }
    
    /**
     * Parses the provided string and returns its 
     * corresponding index pair.
     * 
     * @param indexPair the string to be parsed.
     * 
     * @return the index pair corresponding to the 
     * provided string. 
     */
    public IndexPair<R, C> parse(String indexPair) {

    	// Trimming whitespace
        String trimmed = indexPair.trim();

        // If the string is empty
        if (trimmed.length() == 0) {
            String s = "The index pair '" + indexPair + "' is invalid.";
            throw new FormatException(s);
        }

        // If the string does not begin with a bracket
        if (!trimmed.startsWith("[")) {
            String s = "The index pair '" + trimmed +
                       "' is missing a left-hand bracket.";
            throw new FormatException(s);
        }

        // If the string does not end with a bracket
        if (!trimmed.endsWith("]")) {
            String s = "The index pair '" + trimmed +
                       "' is missing a right-hand bracket.";
            throw new FormatException(s);
        }

        // Getting the sub-string between the brackets
        int end = trimmed.length() - 1;
        String sub = trimmed.substring(1, end);
        
        // Parsing the sub-string
        ListFormat<String> listFormat = PipeListFormat.getBasic();
        List<String> indices = listFormat.parse(sub);

        // Checking the size of the list
        int size = indices.size();
        
        if (size < 2) {
            String s = "The the index pair '" + trimmed + 
                       "' contains too few parts.";
            throw new FormatException(s);
        }

        if (size > 2) {
            String s = "The the index pair '" + trimmed + 
                       "' contains too many parts.";
            throw new FormatException(s);
        }

        // Parsing the row and column indices
        R rowIndex = rowIndexFormat.parse(indices.get(0));
        C colIndex = colIndexFormat.parse(indices.get(1));

        // Returning the parsed index pair
        return new IndexPair<R, C>(rowIndex, colIndex);
    }


}
