/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.math;

import java.util.regex.*;

import sbrt.kernel.math.*;
import sbrt.shell.text.*;

/**
 * This class is a concrete implementation of {@linkplain IntervalFormat}.
 * The string representation of intervals used by this class is:
 * {@code [LowerBound; UpperBound]}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 *
 */
public final class IntervalFormatV1 implements IntervalFormat {

    private static final Pattern semicolon = Pattern.compile("\\s*;\\s*");

    private static final String POS_INFINITY = "Infinity";
    private static final String NEG_INFINITY = "-Infinity";

    private final DoubleFormatV1 doubleFormat;

    /**
     * Constructs a new interval format.
     */
    public IntervalFormatV1() {
        doubleFormat = new DoubleFormatV1();
    }

    /**
     * Returns the double format used for bounds.
     * @return the double format used for bounds.
     */
    public DoubleFormatV1 getDoubleFormat() {
        return doubleFormat;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided interval.
	 * 
	 * @param interval the interval to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided interval.
	 */
    public String format(Interval interval) {

    	// Formatting the bounding values
        String lowerString = formatLower(interval.getLower());
        String upperString = formatUpper(interval.getUpper());

        // Returning the complete interval string
        return lowerString + "; " + upperString;
    }

    /**
     * Formats the lower bound.
     */
    private String formatLower(double x) {
    	
    	if (x == Double.NEGATIVE_INFINITY) {
    		return "(" + NEG_INFINITY;
    	}
    	
        if (x == Double.POSITIVE_INFINITY) {
        	return "(" + POS_INFINITY;
        }
        
        return "[" + doubleFormat.format(x);
    }

    /**
     * Formats the upper bound.
     */
    private String formatUpper(double x) {
        
    	if (x == Double.NEGATIVE_INFINITY) {
    		return NEG_INFINITY + ")";
    	}
    	
    	if (x == Double.POSITIVE_INFINITY) {
            return POS_INFINITY + ")";
        }

    	return doubleFormat.format(x) + "]";
    }

    /**
     * Parses the provided string and returns its 
     * corresponding interval.
     * 
     * @param interval the string to be parsed.
     * 
     * @return the interval corresponding to the 
     * provided string. 
     */
    public Interval parse(String interval) {

    	// Trimming whitespace
        String trimmed = interval.trim();

        // If the string is empty
        if (trimmed.length() == 0) {
            String s = "The interval '" + interval + "' is invalid.";
            throw new FormatException(s);
        }

        // Getting first and last indices of semicolon
        int first = trimmed.indexOf(';');
        int last = trimmed.lastIndexOf(';');

        // If a semicolon does not exist
        if (first < 0) {
            String s = "The character ';' is missing from the interval '" +
                       trimmed + "'.";
            throw new FormatException(s);
        }

        // If more than one semicolon exists
        if (first != last) {
            String s = "The character ';' appears more than once in the " +
                       "interval '" + trimmed + "' .";
            throw new FormatException(s);
        }

        // If the string does not begin with a paren or bracket
        if (!trimmed.startsWith("(") && !trimmed.startsWith("[")) {
            String s = "The interval '" + trimmed +
                       "' is missing a left-hand bracket.";
            throw new FormatException(s);
        }

        // If the string does not end with a paren or bracket
        if (!trimmed.endsWith(")") && !trimmed.endsWith("]")) {
            String s = "The interval '" + trimmed +
                       "' is missing a right-hand bracket.";
            throw new FormatException(s);
        }

        // Getting the length of the string
        int length = trimmed.length();
        
        // Splitting on the semicolon, excluding brackets
        String[] bounds = semicolon.split(trimmed.substring(1, length - 1), -1);

        // If less than two bounding values are present
        if (bounds.length < 2) {
            String s = "The interval '" + trimmed + "' is missing terms.";
            throw new FormatException(s);
        }

        // If more than two bounding value are present
        if (bounds.length > 2) {
            String s = "The interval '" + trimmed +
                       "' contains too many terms.";
            throw new FormatException(s);
        }

        // Parsing the lower and upper bounds
        double lowerBound = doubleFormat.parsePrimitive(bounds[0]);
        double upperBound = doubleFormat.parsePrimitive(bounds[1]);

        // Checking values
        if (Double.isNaN(lowerBound) || Double.isNaN(upperBound)) {
            String s = "The value 'NaN' in the interval '" + interval +
            "' is invalid.";
            throw new FormatException(s);
        }
        
        // If the lower bound is greater than the upper bound
        if (lowerBound > upperBound) {
            String s = "The interval '" + trimmed +
                       "' has contradictory bounds.";
            throw new FormatException(s);
        }

        if (lowerBound == Double.POSITIVE_INFINITY || 
        		upperBound == Double.NEGATIVE_INFINITY) {
        	String s = "The interval '" + trimmed + "' is invalid.";
        	throw new FormatException(s);
        }
        
        // Returning the parsed interval
        return new Interval(lowerBound, upperBound);
    }


}
