/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.math;

import java.util.regex.*;

import sbrt.kernel.math.*;
import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;

/**
 * This class is a concrete implementation of {@linkplain LinearCombTermFormat}.
 * Terms have the following format: {@code coefficient variable}.
 * 
 * 
 * @author This class was written and documented by 
 * Jeremiah Wright while in the Wagner lab.
 *
 * @param <V> the variable type.
 */
public final class LinCombTermFormatV2<V> 
	implements LinearCombTermFormat<V> {

	// Pattern for splitting on whitespace
    private static final Pattern whitespace = Pattern.compile("\\s+");

    // Internal formats
    private final SimpleFormat<V> variableFormat;
    private final CoeffFormatV2 coeffFormat;

    /**
     * Constructs a new linear combination term format using
     * the provided formats.
     * 
     * @param variableFormat the format to use for variables.
     */
    public LinCombTermFormatV2(SimpleFormat<V> variableFormat) {

    	// Checking inputs
    	Util.nullCheck(variableFormat);

        // Setting
        this.variableFormat = variableFormat;
        this.coeffFormat = new CoeffFormatV2();
    }

    /**
     * Returns the format used for variables.
     * @return the format used for variables.
     */
    public SimpleFormat<V> getVariableFormat() {
        return variableFormat;
    }

    /**
     * Returns the format used for coefficients.
     * @return the format used for coefficients.
     */
    public CoeffFormatV2 getCoeffFormat() {
        return coeffFormat;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided term.
	 * 
	 * @param term the term to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided term.
	 */
    public String format(LinearComb.Term<? extends V> term) {

    	// Getting the variable and coefficient
        V variable = term.getVariable();
        double coeff = term.getCoefficient();

        // Creating the coefficient string
        String coeffString = 
        	(coeff == 1d ? "" : coeffFormat.format(coeff) + " ");
        
        // Creating the variable string
        String variableString = variableFormat.format(variable);

        // Returning the formatted term
        return coeffString + variableString;
    }

    /**
     * Parses the provided string and returns its 
     * corresponding term.
     * 
     * @param term the string to be parsed.
     * 
     * @return the term corresponding to the 
     * provided string. 
     */
    public LinearComb.Term<V> parse(String term) {

    	// Trimming whitespace
        term = term.trim();

        // If the string is empty
        if (term.length() == 0) {
            String s = "The provided term is empty.";
            throw new FormatException(s);
        }

        // Splitting on whitespace
        String[] parts = whitespace.split(term, -1);

        if (parts.length > 2) {
            String s = "The term '" + term + "' contains too many parts " +
                       "after splitting on whitespace.";
            throw new FormatException(s);
        }

        // Creating fields for storage
        V variable = null;
        double coefficient = 1;

        // If only one part is present, it must be the variable
        if (parts.length == 1) {
            variable = variableFormat.parse(parts[0]);
        }

        // If two parts are present
        else if (parts.length == 2) {
            coefficient = coeffFormat.parsePrimitive(parts[0]);
            variable = variableFormat.parse(parts[1]);
        }

        // Returning the parsed term
        return new HashLinearComb.Term<V>(coefficient, variable);
    }

    /**
     * Parses the provided strings and returns their 
     * corresponding term.
     * 
     * @param coefficient the string of the coefficient to be parsed.
     * @param variable the string of the variable to be parsed.
     * 
     * @return the term corresponding to the 
     * provided strings. 
     */
    public LinearComb.Term<V> parse(String coefficient, String variable) {
        
        double coeff = coeffFormat.parsePrimitive(coefficient);
        V var = variableFormat.parse(variable);
        
        // Returning the parsed term
        return new HashLinearComb.Term<V>(coeff, var);
    }
}
