/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.math;

import java.util.*;
import java.util.regex.*;

import sbrt.kernel.math.*;
import sbrt.kernel.math.LinearComb.Term;
import sbrt.kernel.utilities.*;
import sbrt.shell.text.*;

/**
 * This class is a concrete implementation of {@linkplain LinearCombFormat}. The
 * linear combination format defined by this class looks like:
 * <p>
 * <i>(coefficient<sub>1</sub>) variable<sub>1</sub> + (constant)</i>
 * </p>
 * <p>
 * Coefficients and constants must be enclosed by parentheses. In terms, the
 * coefficients must appear before the variable, and the two must be separated
 * by at least one whitespace character. Terms must be separated by {@code +}
 * signs. Coefficients, constants, variables, and {@code +} signs must all be
 * separated from each other by at least one whitespace character, which means
 * variable strings cannot contain any whitespace characters. The constant in a
 * linear combination string can occurr in any sensible position.
 * 
 * @author This class was written and documented by Jeremiah Wright while in the
 * Wagner lab.
 * 
 * @param <V> the variable type.
 */
public final class LinearCombFormatV1<V> implements LinearCombFormat<V> {

    // Pattern for splitting on + surrounded by whitespace
    private static final Pattern plus = Pattern.compile("\\s+\\+\\s+");

    // Pattern for splitting on whitespace
    private static final Pattern whitespace = Pattern.compile("\\s+");

    /**
     * Returns a {@code LinearCombFormatV1} that does not
     * format or parse the variables supplied to it. In other
     * words, it returns a {@code LinearCombFormatV1} with
     * the most basic possible functionality. Empty strings are not 
	 * allowed however.
     *  
     * @return a {@code LinearCombFormatV1} that does not
     * format or parse the variables supplied to it.
     */
    public static LinearCombFormatV1<String> getBasic() {

		// Creating the "trivial" format
		NonEmptyStringFormat f = new NonEmptyStringFormat();

        return new LinearCombFormatV1<String>(f, 0.0d);
    }

    // Internal formats
    private final SimpleFormat<V> variableFormat;

    private final LinCombTermFormatV1<V> termFormat;

    private final CoeffFormatV1 coeffFormat;

    // The zero cutoff
    private final double zeroCutoff;

    /**
     * Constructs a new linear combination format using the provided arguments.
     * 
     * @param variableFormat the format to use for variables.
     * @param zeroCutoff the zero cutoff for constants and coefficients.
     * 
     * @see LinearComb#getZeroCutoff()
     */
    public LinearCombFormatV1(SimpleFormat<V> variableFormat, double zeroCutoff) {

        // Checking and saving
        Util.nullCheck(variableFormat);
        this.variableFormat = variableFormat;
        this.zeroCutoff = zeroCutoff;

        // Creating internal formats
        termFormat = new LinCombTermFormatV1<V>(variableFormat);
        coeffFormat = termFormat.getCoeffFormat();
    }

    /**
     * Constructs a new linear combination format using the provided format. The
     * zero cutoff is set to 0.0d.
     * 
     * @param variableFormat the format to use for variables.
     * 
     * @see LinearComb#getZeroCutoff()
     */
    public LinearCombFormatV1(SimpleFormat<V> variableFormat) {
        this(variableFormat, 0.0d);
    }

    /**
     * Returns the format used for variables.
     * 
     * @return the format used for variables.
     */
    public SimpleFormat<V> getVariableFormat() {
        return variableFormat;
    }

    /**
     * Returns the format used for terms.
     * 
     * @return the format used for terms.
     */
    public LinCombTermFormatV1<V> getTermFormat() {
        return termFormat;
    }

    /**
     * Returns the format used for coefficients.
     * 
     * @return the format used for coefficients.
     */
    public CoeffFormatV1 getCoeffFormat() {
        return coeffFormat;
    }

    /**
     * Returns a formatted string representation of the provided linear
     * combination.
     * 
     * @param expression the linear combination to be formatted.
     * 
     * @return a formatted string representation of the provided linear
     * combination.
     */
    public String format(LinearComb<V> expression) {

        // If no terms exist, return the formatted constant
        if (expression.terms() == 0) {
            return coeffFormat.format(expression.getConstant());
        }

        // Determining appropriate size
        int size = expression.terms() * 100;

        // Creating storage for formatted linear combination string
        StringBuilder s = new StringBuilder(size * 100);

        // Creating iterator over expression
        Iterator<Term<V>> i = expression.iterator();

        // Iterating thru each term
        while (i.hasNext()) {

            // Getting the next term
            Term<V> term = i.next();

            // Appening formatted term
            s.append(termFormat.format(term));

            // If another term exists, append a plus sign
            if (i.hasNext()) {
                s.append(" + ");
            }
        }

        // Getting the constant
        double constant = expression.getConstant();

        // If the constant does not equal zero
        if (!Util.isZero(constant)) {
            s.append(" + ");
            s.append(coeffFormat.format(constant));
        }

        // Returning the formatted linear combination
        return s.toString();
    }

    /*
    private LinearComb<V> parseOld(String expression) {

        // Splitting on plus sign surrounded by whitespace
        String[] termStrings = plus.split(expression, -1);

        // Creating storage for variables
        Set<V> variables = new LinkedHashSet<V>();

        // Creating storage for terms
        List<Term<V>> termList = new LinkedList<Term<V>>();

        // Creating field for storage
        Double constant = null;

        // Iterating thru each term string
        for (String term : termStrings) {

            // If the term string is empty
            if (term.length() == 0) {
                String s =
                        "An empty term is present in the linear combination '" + expression
                                + "'.";
                throw new FormatException(s);
            }

            // Attempting to parse as a constant
            Double x = parseAsConstant(term, constant, expression);

            // If it was successful
            if (x != null) {
                constant = x;
            }

            // If it wasn't
            else {

                // Attempting to parse as a term
                LinearComb.Term<V> t = parseAsTerm(term);

                // Getting its variable
                V var = t.getVariable();

                // Attempting to add the variable to the rest
                if (!variables.add(var)) {

                    String s =
                            "The variable '" + variableFormat.format(var)
                                    + "' occurs more than once in the linear combination '"
                                    + expression
                                    + "'.";
                    throw new FormatException(s);

                }

                // Adding the term to the rest
                termList.add(t);

            }

        }

        // If no constant was provided
        if (constant == null) {
            constant = 0.0d;
        }

        // Returning the parsed linear combination
        return new HashLinearComb<V>(termList, constant, zeroCutoff);
    }
    
    private LinearComb.Term<V> parseAsTerm(String string) {

        // Attempting to parse as a term
        return termFormat.parse(string);
    }
*/

    private Double parseAsConstant(String string,
            Double currentConstant,
            String expression) {

        // Attempting to parse as a constant
        Double c = coeffFormat.parse(string);

        // If it's the first constant, return it
        if (currentConstant == null) {
            return c;
        }

        // If it's the second constant found, complain
        String s = "The linear combination '" + expression +
                   "' contains more than one constant.";
            throw new FormatException(s);

    }

    /**
     * Parses the provided string and returns its corresponding linear
     * combination.
     * 
     * @param expression the string to be parsed.
     * 
     * @return the linear combination corresponding to the provided string.
     */
    public LinearComb<V> parse(String expression) {

        // Creating storage for variables
        Set<V> variables = new LinkedHashSet<V>();

        // Creating storage for terms
        List<Term<V>> termList = new LinkedList<Term<V>>();

        // Creating field for storage
        Double constant = null;

        // Splitting on +
        String[] terms = plus.split(expression.trim(), -1);

        // Iterating thru each term
        for (String t : terms) {

            // Splitting on whitespace
            String[] pieces = whitespace.split(t, -1);

            switch (pieces.length) {

                case 0: {
                    String s =
                            "An empty term is present in the linear combination '" + expression
                                    + "'.";
                    throw new FormatException(s);
                }

                case 1: {

                    // If it seems to be a constant, parse it that way
                    if (pieces[0].startsWith("(") && pieces[0].endsWith(")")) {
                        constant =
                                parseAsConstant(pieces[0], constant, expression);
                        break;
                    }

                    // If it seems to be a variable, parse it that way
                    // Attempting to parse as a term
                    LinearComb.Term<V> term = termFormat.parse(pieces[0]);

                    // Adding it to the rest
                    addTerm(term, variables, termList, expression);
                }
                    break;

                case 2: {

                    // Parsing as a term
                    LinearComb.Term<V> term =
                            termFormat.parse(pieces[0], pieces[1]);

                    // Adding it to the rest
                    addTerm(term, variables, termList, expression);
                }
                    break;

            } // End of switch

            if (pieces.length > 2) {
                String s =
                        "A term in the linear combination '" + expression
                                + "' contains too many parts.";
                throw new FormatException(s);
            }

        } // for loop

        // If no constant was provided
        if (constant == null) {
            constant = 0.0d;
        }

        // Returning the parsed linear combination
        return new HashLinearComb<V>(termList, constant, zeroCutoff);
    }

    private void addTerm(LinearComb.Term<V> term,
            Set<V> variables,
            List<LinearComb.Term<V>> termList,
            String expression) {

        // Getting its variable
        V var = term.getVariable();

        // Attempting to add the variable to the rest
        if (!variables.add(var)) {

            String s =
                    "The variable '" + variableFormat.format(var)
                            + "' occurs more than once in the linear combination '"
                            + expression
                            + "'.";
            throw new FormatException(s);

        }

        // Adding the term to the rest
        termList.add(term);
    }

}
