/*----------------------------------------------------------------------
-- Copyright (C) 2008 Jeremiah Wright. All rights reserved.
-- E-mail: j.wright@bioc.uzh.ch
--
-- This file is part of the Systems Biology Research Tool (SBRT).
--
-- The SBRT is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2, or (at your option)
-- any later version.
--
-- The SBRT is distributed in the hope that it will be useful, but WITHOUT
-- ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with the SBRT; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
----------------------------------------------------------------------*/

package sbrt.shell.text.math;

import java.util.Map.Entry;

import sbrt.kernel.math.algebra.IndexPair;
import sbrt.kernel.math.algebra.MatrixElement;
import sbrt.shell.text.DoubleFormatV1;
import sbrt.shell.text.EqualsMapEntryFormatV1;
import sbrt.shell.text.MapEntryFormat;
import sbrt.shell.text.SimpleFormat;

/**
 * This class is a concrete implementation of {@linkplain MatrixElementFormat}.
 * The string representation of matrix elements used by this class is:
 * {@code [RowIndex | ColumnIndex] = Element}.
 * 
 * @author This class was written and documented by
 * Jeremiah Wright while in the Wagner lab.
 * 
 * @param <R> the row index type. 
 * @param <C> the column index type.
 *
 */
public final class MatrixElementFormatV1<R, C> 
    implements MatrixElementFormat<R, C, Double> {


    private final IndexPairFormatV1<R, C> indexFormat;
    private final DoubleFormatV1 elementFormat;

    /**
     * Constructs a new matrix element format.
     * @param rowIndexFormat the row index formatter.
     * @param columnIndexFormat the column index formatter.
     */
    public MatrixElementFormatV1(SimpleFormat<R> rowIndexFormat, 
                                 SimpleFormat<C> columnIndexFormat) {
        
        indexFormat = new IndexPairFormatV1<R, C>(rowIndexFormat, 
                                                  columnIndexFormat);
        elementFormat = new DoubleFormatV1();
    }

    /**
     * Returns the format used for matrix elements.
     * @return the format used for matrix elements.
     */
    public IndexPairFormatV1<R, C> getIndexFormat() {
        return indexFormat;
    }

    /**
     * Returns the format used for matrix elements.
     * @return the format used for matrix elements.
     */
    public DoubleFormatV1 getElementFormat() {
        return elementFormat;
    }

	/**
	 * Returns a formatted string representation of the
	 * provided matrix element.
	 * 
	 * @param element the matrix element to be formatted.
	 * 
	 * @return a formatted string representation of the
	 * provided matrix element.
	 */
    public String
    formatElement(MatrixElement<? extends R, ? extends C, Double> element) {

    	// Formatting the matrix element
        String indices = indexFormat.formatPair(element.getIndices());

        // Formatting the element value
        String value = elementFormat.format(element.getElement());
        
        // Returning the complete index string
        return indices + " = " + value;
    }

    /**
     * Returns a formatted string representation of the
     * provided matrix element.
     * 
     * @param indexPair the matrix element to be formatted.
     * 
     * @return a formatted string representation of the
     * provided matrix element.
     */
    public String format(MatrixElement<R, C, Double> indexPair) {
        return formatElement(indexPair);
    }
    
    /**
     * Parses the provided string and returns its 
     * corresponding matrix element.
     * 
     * @param element the string to be parsed.
     * 
     * @return the matrix element corresponding to the 
     * provided string. 
     */
    public MatrixElement<R, C, Double> parse(String element) {

        // Creating map entry parser
        MapEntryFormat<IndexPair<R, C>, Double> format = 
            new EqualsMapEntryFormatV1<IndexPair<R, C>, Double>(
                    indexFormat, elementFormat);
        
        // Parsing as a map entry
        Entry<IndexPair<R, C>, Double> entry = format.parsePair(element);
        
        // Returning the parsed matrix element
        return new MatrixElement<R, C, Double>(entry.getKey(), entry.getValue());
    }


}
